;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information.

(in-package 'user)

;; This file lists the code for the rule-based reactive system 
;; example introduced in the chapter on symbolic programming. 
;; See the text for detailed comments and supplementary material.

;; Sensors - (name . values)
;; (forward near away far)
;; (jright near away far)
;; (jleft near away far)
;; (right near away far)
;; (left near away far)
;; (rear near away far)

;; Control parameters - (name . values)
;; (speed zero slow fast)
;; (turn left straight right)

;; Abstraction for parameter/value tuples
(defun make-TUPLE (param value) (list param value))
(defun TUPLE-param (tuple) (first tuple))
(defun TUPLE-value (tuple) (second tuple))   

;; Abstraction for time-stamped sensor data
(defun make-REPORT (tuple stamp) (list tuple stamp))
(defun REPORT-tuple (report) (first report))
(defun REPORT-stamp (report) (second report))
(defun REPORT-param (report) (TUPLE-param (first report)))

;; COLLECT is supposed to provide all of the queued up sensor 
;; data. In this simulation, it returns the same data every time.
(defun collect () 
  (list 
   (make-REPORT (make-TUPLE 'forward 'far) 0)
   (make-REPORT (make-TUPLE 'left 'away) 1)
   (make-REPORT (make-TUPLE 'right 'near) 1)
   (make-REPORT (make-TUPLE 'rear 'far) 1)
   (make-REPORT (make-TUPLE 'speed 'slow) 1)))

;; What would (assoc 'left '((left far) (right near))) return?
;; What would (assoc '(left far) '(((left far) 1))) return?
;; How about (assoc '(left far) '(((left far) 1)) :test #'equal)?
;; How about (assoc '(left far) '(((left near) 1)) :test #'equal)?

;; ASSOC will not quite for reasons having to do with our data
;; abstractions. Instead, we use a generalization of ASSOC called FIND.
;; Using FIND, we define two functions that perform associations on lists
;; of reports, maintaining the data abstractions.  The first function
;; searches for a report with a tuple that has a particular parameter;
;; the second function searches for a report with a particular tuple.
(defun param-assoc (param reports)
  (find param reports :key #'REPORT-param))
(defun tuple-assoc (tuple reports)
  (find tuple reports :test #'equal :key #'REPORT-tuple))

;; Replace each sensor reading with the new reading if present.
(defun update-variant (new old) 
  (mapcar #'(lambda (item)
              (or (param-assoc (REPORT-param item) new)
                  item))
          old))

;; Replace an old sensor reading if there is a new sensor reading
;; with a later time stamp. 
(defun update (new old) 
  (dolist (item new old) (setq old (fuse item old))))

(defun fuse (report old)
  (cond ((null old) (list report))
        ((eq (REPORT-param report) (REPORT-param (first old)))
         (if (> (REPORT-stamp report) (REPORT-stamp (first old)))
           (cons report (rest old)) old))
        (t (cons (first old) (fuse report (rest old))))))

(defun fuse-variant (report old) 
  (let ((a (param-assoc (REPORT-param report) old)))
    (cond ((null a) (cons report old))
          ((< (REPORT-stamp report) (REPORT-stamp a)) old)
          (t (cons a (remove a old))))))

;; Abstraction for rules
(defun make-RULE (conditions action) (list conditions action))
(defun RULE-conditions (rule) (first rule))
(defun RULE-action (rule) (second rule))

(defun controls ()
  (list 
   (make-RULE '((forward near) (jleft far)) '(turn left))
   (make-RULE '((forward near) (jright far)) '(turn right))
   (make-RULE '((forward near) (jleft near) (jright near)) '(speed stopped))))

;; Determine if a particular rule applies for a given set of reports.
(defun applicablep (rule reports) 
  (every #'(lambda (tuple) (tuple-assoc tuple reports))
         (RULE-conditions rule)))

(defun react (rules reports) 
  (dolist (rule rules reports)
    (if (applicablep rule reports) 
	(setq reports (fuse (act (RULE-action rule)) reports)))))

;; Try (react (controls) (collect)) to test the above functions.
(defun act (action) (list action (get-internal-real-time)))

;; React to the environments for 100 iterations. 
;; Try (run (controls) (collect)) to test this function. 
(defun run (rules reports)
  (dotimes (index 100 reports)
    (setq reports (react rules (update (collect) reports)))))

(defun run-variant (rules reports)
  (aux-update rules (collect) reports 0))
(defun aux-update (rules new old i)
  (if (< i 100) (aux-react rules (update new old) i)))
(defun aux-react (rules reports i) 
  (aux-update rules (collect) (react rules reports) (+ i 1)))


;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.


