;; Takes a database of propositional variable / value pairs and a list
;; of rules, where each rule has two parts: a test and a list of new
;; propositional variable / value pairs. It generates new rules, checks
;; for conflict against existing values, and runs itself again to generate
;; even more variables, until it cannot generate any more variables.
;; Returns the bigger database.

(defun update-database (database rules)
  (update-aux database rules rules))

(defun update-aux (database full-rules rules)
  (cond ((null rules) database)
	(t (let ((newdata (match-rule database (first rules))))
	     (cond (newdata (update-database newdata
					     (remove-rule (first rules)
							  full-rules)))
		   (t (update-aux database
				  full-rules (rest rules))))))))

(defun remove-rule (rule rules)
  (cond ((null rules) nil)
	((equal rule (first rules)) (rest rules))
	(t (append (list (first rules)) (remove-rule rule (rest rules))))))


;; Matches the database against the rules and returns any new information
;; if there is a match, first checking to make sure there is no conflict
;; against anything in the database now. Returns the bigger database if
;; a match, or nil if no match or conflict.

(defun match-rule (database rule)
  (if (equal (apply-rule database (first rule)) t) ;; not unknown or false
      (combine-database database (first (rest rule)))
      nil))

;; Returns true if the expression is true given the information in the
;; database. If the expression is false, returns nil. Finally, if there
;; is not enough information in the database to evaluate the expression,
;; returns '?

(defun apply-rule (database expression)
  (cond ((equal expression t) expression)
	((equal expression nil) expression)
	((symbolp expression) (lookup database expression))
        (t (let ((op (first expression))
		 (opand1 (apply-rule database
				     (first (rest expression))))
		 (opand2 (apply-rule database
				     (first (rest (rest expression))))))
             (cond ((equal op 'and) (cond ((and (equal opand1 t)
						(equal opand2 t)) t)
					  ((or  (equal opand1 '?)
						(equal opand2 '?)) '?)
					  (t nil)))
		   ((equal op 'or)  (cond ((or (equal opand1 t)
					       (equal opand2 t)) t)
					  ((or (equal opand1 '?)
					       (equal opand2 '?)) '?)
					  (t nil)))

		   ((equal op 'not) (cond ((equal opand1 '?) '?)
				       (t (not opand1)))))))))

;; Checks to make sure that one given database of information does not
;; conflict with the other database. Returns the combined databases or
;; returns nil if a conflict.

(defun combine-database (db1 db2)
  (cond ((null db2) db1)
	((combine-data db1 (first db2))
	 (combine-database (combine-data db1 (first db2))
			   (rest db2)))
	(t nil)))

;; Combines one element of a database with a database. Returns the 
;; combined database, or nil if there is a conflict.

(defun combine-data (db1 element)
  (cond ((equal (lookup db1 (first element)) '?)   ;; means no entry
	 (append (list element) db1))
	((equal (lookup db1 (first element))
		(second element))
	 db1)
	(t nil)))

;; Looks up the variable in the database and returns its boolean value
;; or returns '? if not found

(defun lookup (database variable)
  (cond ((null database) '?)
	((equal variable (first (first database)))
	 (first (rest (first database))))
	(t (lookup (rest database) variable))))
