;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information
(in-package 'user)

;; In this file, we provide a goal reduction algorithm which is 
;; complete and runs in time linear in the size of the database. 

;; Let the status of a proposition be one of (NEW WORKING THEOREM NIL);
;; THEOREM indicates a proposition is a theorem and NIL that it is not.
;; WORKING means we are in the process of trying to prove a proposition.  

;; In the following, we assume that the three operations corresponding 
;; to determining the status of a proposition, setting the status of a 
;; proposition, and retrieving the list of possible reductions licensed 
;; by the various rules associated with a given proposition can be done 
;; in constant time irregardless of the number of proposition symbols or 
;; the size of the database of rules. 

;; This is not true of the following implementations but with a little
;; extra effort we could provide such constant time operations. 
;; SET is like SETQ except that it evaluates its first argument which
;; must evaluate to a symbol.
(defun status (p) (eval p))
(defun set-status (p s) (set p s))
(defun reductions (p)
  (cdr (assoc p '((P (U V) (Q R))
		  (Q (S))
		  (U (P W) (S U))
		  (S ()) 
		  (R ())))))

;; Initialize all of the proposition symbols to have status NEW.   
(defun init ()
  (mapc #'(lambda (p) (set-status p 'new)) '(P Q R S U V W)))

;; In the following, (EVERY #'F L) is non-NIL just in case F applied
;; to each element of L is non-NIL. Similarly (SOME #'F L) is non-NIL
;; just in case F applied to at least one element of L is non-NIL.
(defun theorem (p)
  (cond ((eq (status p) 'theorem) t)
	((not (eq (status p) 'new)) nil)
	(t (set-status p 'working)
	   (if (some #'(lambda (c) (every #'theorem c))
		     (reductions p))
	       (set-status p 'theorem)
	     (set-status p nil)))))

;; THEOREM runs in time linear in the size of the database. To see 
;; that this is the case, note that REDUCTIONS is called no more than
;; one time for each propositional symbol. THEOREM is complete in the
;; sense that if P is a consequence of a theory T corresponding to a
;; database of rules then THEOREM will determine that this is the case.
;; To see this, note that if P is a consequence of T then there must be
;; some rule in the database of the form (P if Q1 Q2 ... Qn) and each of
;; the Qi must be theorems, but this is exactly what THEOREM checks for.
;; The justification for rejecting conjunctions with a conjunct that
;; corresponds to a proposition we are already working on is as follows. 
;; Suppose that you have reduced proving P to proving the conjunction 
;; (Q1 Q2 ... Qn) and suppose that you can reduce Q1 to (P R1 R2 ... Rm).
;; In this case, you can safely ignore the reduction since proving 
;; (R1 R2 ... Rm Q2 ... Qn) will not provide you with a proof of P.

;; Show the status of all of the propositions.	   
(defun show ()
  (mapcar #'(lambda (p) (list p (status p))) '(P Q R S U V W)))

;; Here's a simple demonstration of THEOREM. 
;; > (init)
;; (P Q R S U V W)
;; > (theorem 'p)
;; THEOREM
;; > (show)
;; ((P THEOREM) (Q THEOREM) (R THEOREM) (S THEOREM) (U NIL) (V NEW) (W NEW))

;; Copyright 1994, Brown University, Providence, RI
;; 
;; Permission to use and modify this software and its documentation for any
;; purpose other than its incorporation into a commercial product is hereby
;; granted without fee.  Permission to copy and distribute this software and its
;; documentation only for non-commercial use is also granted without fee,
;; provided, however, that the above copyright notice appear in all copies, that
;; both that copyright notice and this permission notice appear in supporting
;; documentation, that the name of Brown University not be used in advertising or
;; publicity pertaining to distribution of the software without specific, written
;; prior permission, and that the person doing the distribution notify Brown
;; University of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this software for
;; any purpose.  It is provided "as is" without express or implied warranty.
;; Brown University requests notification of any modifications to this software
;; or its documentation.
;; 
;; Send the following redistribution information:
;; 
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;; 
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;; 
;; 		or
;; 
;; 	brusd@cs.brown.edu
;; 
;; We will acknowledge all electronic notifications.
