;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information.

;; (in-package 'user)

;; In this file, we list the data dependency algorithm 
;; given in the appendix of Chapter 3.

;; An instance of the abstract data type node consists
;; of a name, a label IN or OUT, a list of its justifications, and
;; a list of the nodes that it justifies. 
;; Each justification is represented as a list of justifiers.

(defun make-NODE (name) (list name 'OUT () ()))
(defun NODE-label (node) (second node))
(defun NODE-justifications (node) (third node))
(defun NODE-justificands (node) (fourth node))

;; All of the elements of a node are settable except
;; name which is specified when an instance of type node
;; is created.

(defun set-NODE-label (node l) (setf (second node) l))
(defun set-NODE-justifications (node j) (setf (third node) j))
(defun set-NODE-justificands (node j) (setf (fourth node) j))

;; Add a justification to an existing dependency node.

(defun add-justification (node justification)
  (set-NODE-justifications node
    (cons justification (NODE-justifications node)))
  (mapc #'(lambda (n) 
            (set-NODE-justificands n
               (cons node (NODE-justificands node)))) 
        justification)
  (update-dependency-graph node))

;; Delete a justification from an existing dependency node.
;; We use remove and its optional :test argument to invoke EQUAL 
;; to find and remove a justification from a list of justifications 
;; and a justificand from a list of justificands.
;; This simple approach has some limitations that you are challenged 
;; to overcome in the exercises at the end of Chapter 3.

(defun delete-justification (node justification)
  (set-NODE-justifications node
    (remove justification (NODE-justifications node)
            :test #'equal))
  (mapc #'(lambda (n) 
            (set-NODE-justificands n
              (remove node (NODE-justificands node) 
                      :test #'equal))) 
        justification)
  (update-dependency-graph node))

;; Update a dependency graph in two passes starting from node.
;; The first pass erases the existing labels of all reachable nodes and 
;; relabels them as visited.
;; The second pass assigns new labels to all reachable nodes.

(defun update-dependency-graph (node)
  (erase-labels node)
  (relabel node))

;; Label as visited all nodes reachable from node 
;; by traversing justifications from justifiers to justificands.

(defun erase-labels (node)
  (cond ((eq (NODE-label node) 'VISITED) nil)
        (t (set-NODE-label node 'VISITED)
           (mapc #'erase-labels (NODE-justificands node)))))

;; Propagate labels forward along justifications
;; from justifiers to justificands starting from node. 
;; If the label at a node does not require changing, then
;; stop propagating labels forward.

(defun relabel (node)
  (let ((label (if (consistent node) 'IN 'OUT)))
    (cond ((eq (NODE-label node) label) nil)
          (t (set-NODE-label node label)
             (mapc #'relabel (NODE-justificands node))))))

;; It is consistent to label a node IN if it has some justification all 
;; of whose justifiers are labeled , otherwise it should be labeled OUT.

(defun consistent (node)
  (some #'justified (NODE-justifications node)))

;; Check to see if all of the justifiers of a justification
;; are labeled IN.

(defun justified (justification)
  (every #'(lambda (node) (eq (NODE-label node) 'IN)) 
         justification))

;; With the above code, we can build the dependency graph of
;; shown in the text and demonstrate the update algorithms.

(let ((D1 (make-NODE 'D1))
      (D2 (make-NODE 'D2))
      (D3 (make-NODE 'D3))
      (D4 (make-NODE 'D4)))
  (add-justification D1 ())
  (add-justification D2 ())
  (add-justification D3 (list D1 D2))
  (add-justification D4 (list D3))
  (add-justification D3 (list D4))
  (defun test ()
    (let (test1-result test2-result)
      ;; The label of D3 should be IN.
      (setq test1-result (eq 'IN (NODE-label D3)))
      ;; Delete the only justification from D1.
      (delete-justification D1 ())
      ;; Now the label of D3 should be OUT.
      (setq test2-result (eq 'OUT (NODE-label D3)))
      ;; The function test should return t.
      (and test1-result test2-result))))

;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.


