;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;; In this file, we list the data dependency algorithm 
;; given in the appendix of Chapter 3.

;; An instance of the abstract data type node consists
;; of a name, a label IN or OUT, a list of its justifications, and
;; a list of the nodes that it justifies. 
;; Each justification is represented as a list of justifiers.

(define (make-node name) (list name 'out '() '()))

(define (node-label node) (second node))

(define (node-justifications node) (third node))

(define (node-justificands node) (fourth node))


;; All of the elements of a node are settable except
;; name which is specified when an instance of type node
;; is created.

(define (set-node-label node l) (set-car! (list-tail node 1) l))

(define (set-node-justifications node j) (set-car! (list-tail node 2) j))

(define (set-node-justificands node j) (set-car! (list-tail node 3) j))


;; Add a justification to an existing dependency node.

(define (add-justification node justification)
  (set-node-justifications
    node
    (cons justification (node-justifications node)))
  (for-each
    (lambda (n) 
      (set-node-justificands n (cons node (node-justificands node))))
    justification)
  (update-dependency-graph node))


;; Delete a justification from an existing dependency node.
;; We use remove and its optional :test argument to invoke EQUAL 
;; to find and remove a justification from a list of justifications 
;; and a justificand from a list of justificands.
;; This simple approach has some limitations that you are challenged 
;; to overcome in the exercises at the end of Chapter 3.

(define (delete-justification node justification)
  (set-node-justifications
    node
    (remove2 justification (node-justifications node) 
	     equal? (lambda (x) x)))
  (for-each
    (lambda (n)
      (set-node-justificands
        n
        (remove2 node (node-justificands node) 
		 equal? (lambda (x) x))))
    justification)
  (update-dependency-graph node))


;; Update a dependency graph in two passes starting from node.
;; The first pass erases the existing labels of all reachable nodes and 
;; relabels them as visited.
;; The second pass assigns new labels to all reachable nodes.

(define (update-dependency-graph node) 
  (erase-labels node) (relabel node))


;; Label as visited all nodes reachable from node 
;; by traversing justifications from justifiers to justificands.

(define (erase-labels node)
  (cond ((eq? (node-label node) 'visited) #f)
        (else
         (set-node-label node 'visited)
         (for-each erase-labels (node-justificands node)))))


;; Propagate labels forward along justifications
;; from justifiers to justificands starting from node. 
;; If the label at a node does not require changing, then
;; stop propagating labels forward.

(define (relabel node)
  (let ((label (if (consistent node) 'in 'out)))
    (cond ((eq? (node-label node) label) #f)
          (else
           (set-node-label node label)
           (for-each relabel (node-justificands node))))))


;; It is consistent to label a node IN if it has some justification all 
;; of whose justifiers are labeled , otherwise it should be labeled OUT.

(define (consistent node) (some2 justified (node-justifications node)))


;; Check to see if all of the justifiers of a justification
;; are labeled IN.

(define (justified justification)
  (every2 (lambda (node) (eq? (node-label node) 'in)) justification))


;; With the above code, we can build the dependency graph of
;; shown in the text and demonstrate the update algorithms.

(let ((d1 (make-node 'd1))
      (d2 (make-node 'd2))
      (d3 (make-node 'd3))
      (d4 (make-node 'd4)))
  (add-justification d1 '())
  (add-justification d2 '())
  (add-justification d3 (list d1 d2))
  (add-justification d4 (list d3))
  (add-justification d3 (list d4))
  (set! test
    (lambda ()
      (let ((test1-result #f) (test2-result #f))
        (set! test1-result (eq? 'in (node-label d3)))
        (delete-justification d1 '())
        (set! test2-result (eq? 'out (node-label d3)))
        (and test1-result test2-result)))))


;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

