;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;; Matching and unification from the logic chapter.

(defun make-VAR (var) (list '? var))
(defun is-VAR (x) (and (consp x) (eq (first x) '?)))

;; Determine if a constant and a pattern match.
(defun match (const pat)
  (aux-match const pat '((match t))))

;; Keep track of the bindings established so far.
(defun aux-match (const pat bdgs)
  (cond ((not bdgs) nil)
        ((atom pat) (if (eq pat const) bdgs nil))
        ;; If the pattern is a bound variable, match the
        ;; constant and whatever the variable is bound to.
        ;; If the pattern is a variable and not bound, 
        ;; then bind the variable to the constant.
        ((is-VAR pat)
         (let ((bdg (assoc pat bdgs :test #'equal)))
           (cond (bdg (aux-match const (second bdg) bdgs))
                 (t (cons (list pat const) bdgs)))))
        ((or (atom const) (null const)) nil)
        (t (aux-match (first const) 
                      (first pat) 
                      (aux-match (rest const)
                                 (rest pat) bdgs)))))

(defun unify (p q)
  (aux-unify p q '((unify t))))

;; Keep track of the bindings established so far.
(defun aux-unify (p q bdgs)
  (let ((d (disagreement p q)))
    ;; If there is no disagreement, then signal success.
    (cond ((not d) bdgs)
          ;; Assume that the first element of the disagreement
          ;; is a variable. If the first element appears in the
          ;; second, then signal failure. If the assumption turns
          ;; out to be false, then the following clause will cope.
          ((occursp (first d) (second d)) nil)
          ((not (is-VAR (first d))) nil)
          ;; Resolve the disagreement by making appropriate
          ;; substitutions and then recurse on the result.
          (t (aux-unify 
              (subst (second d) (first d) p :test #'equal)
              (subst (second d) (first d) q :test #'equal)
              (cons d bdgs))))))

;; Return subexpressions where P and Q first disagree.
(defun disagreement (p q)
  ;; If either P or Q are variables
  ;; and they are not EQUAL, then you are done.
  (cond ((is-VAR p) (if (equal p q) nil (list p q)))
        ((is-VAR q) (if (equal p q) nil (list q p)))
        ;; If either P or Q is not NIL and not a
        ;; list and P and Q are not EQUAL, then
        ;; you are done, but P and Q will not unify.
        ((or (not (consp p)) (not (consp q)))
         (if (eq p q) nil (list p q)))
        (t (or (disagreement (first p) (first q))
               (disagreement (rest p) (rest q))))))

;;  Check to see of if X appears anywhere in Y.
(defun occursp (x y)
  (cond ((or (null y) (not (consp y))) nil)
        ((is-VAR y) (equal x y))
        (t (or (occursp x (first y)) (occursp x (rest y))))))

;; Test MATCH and UNIFY.
(defun test () 
  (and (match '(loves (dog fred) fred)
	      '(loves (? x) (? y)))
       (match '(loves (dog fred) mary)
	      '(loves (dog (? x)) (? y)))
       (not (match '(loves (dog fred) fred)
		   '(loves (? x) (? x))))
       (not (match '(loves (dog fred) mary)
		   '(loves (dog (? x)) (? x))))
       (unify '(loves (dog (? x)) (dog fred))
	      '(loves (? z) (? z)))))


;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

