;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information
(in-package 'user)

;; Load the search routines from the book.
(load "search")

;; We begin with some useful data abstractions.

;; A NODE in the search space is a LOCation and a 
;; path represented as a list of LOCations. 
(defun make-NODE (loc path) (list loc path))
(defun NODE-loc (node) (first node))
(defun NODE-path (node) (second node))

;; Data abstraction for grid LOCations.
(defun make-LOC (i j) (list (list i j) nil nil))
(defun LOC-coord (loc) (first loc))
;; The OCC slot is used to indicate occupied LOCations.
(defun LOC-occ (loc) (second loc))
(defun set-LOC-occ (loc val) (setf (second loc) val))
;; The MARK slot is used in searching graphs.
(defun LOC-mark (loc) (third loc))
(defun set-LOC-mark (loc mark) (setf (third loc) mark))

;; Data abstraction for coordinate pairs.
(defun make-COORD (i j) (list i j))
;; Returns the x coordinate given a coordinate pair.
(defun COORD-x (coord) (first coord))
;; Returns the y coordinate given a coordinate pair.
(defun COORD-y (coord) (second coord))

;; Data abstraction for grids. 
;; Note that GRID LOCations are zero based.
(defun make-GRID (n m)
  (do ((i 0 (+ i 1)) (rows ()))
      ((= i n) (reverse rows))
    (setq rows
          (cons (do ((j 0 (+ j 1)) (cols ()))
                    ((= j m) (reverse cols))
                  (setq cols 
                        (cons (make-LOC i j) cols)))
                rows))))
;; Return LOCation (i j) in a GRID.
(defun GRID-ref (grid i j)
  (let ((num-rows (length grid))
	(num-cols (length (first grid))))
    ;; Make sure the indices are within bounds.
    (if (or (>= i num-rows) (< i 0)
	    (>= j num-cols) (< j 0))
	nil
      (nth j (nth i grid)))))
;; Reset all the LOCation marks to nil.
(defun GRID-clear (grid)
  (dotimes (i (length grid))
    (dotimes (j (length (first grid)))
      (set-LOC-mark (GRID-ref grid i j) nil))))
;; Display the grid indicating obstacles with an X.
(defun GRID-display (grid)
  (format t "~%       ")
  (dotimes (i (length (first grid))) (format t "~D   " i))
  (dotimes (i (length grid))
    (format t "~%    ~D  " i)
    (dotimes (j (length (first grid)))
      (if (LOC-occ (GRID-ref grid i j))
        (format t "X   ")
        (format t "+   ")))))
;; Display the grid indicating marked locations with an M.
(defun GRID-marks (grid)
  (format t "~%       ")
  (dotimes (i (length (first grid))) (format t "~D   " i))
  (dotimes (i (length grid))
    (format t "~%    ~D  " i)
    (dotimes (j (length (first grid)))
      (if (LOC-occ (GRID-ref grid i j))
        (format t "X   ")
        (if (LOC-mark (GRID-ref grid i j))
          (format t "M   ")
          (format t "+   "))))))
;; Display the Manhattan distance to the goal at each location.
(defun GRID-surface (grid goal)
  (format t "~%         ")
  (dotimes (i (length (first grid))) (format t "~D   " i))
  (dotimes (i (length grid))
    (format t "~%    ~D  " i)
    (dotimes (j (length (first grid)))
      (if (LOC-occ (GRID-ref grid i j))
        (format t "  X ")
        (format t "~3D " (manhattan (LOC-coord (GRID-ref grid i j))
				      goal))))))

;; Compute the Manhattan distance between two COORDinates.
(defun manhattan (coord1 coord2)
  (+ (abs (- (COORD-x coord1) (COORD-x coord2)))
     (abs (- (COORD-y coord1) (COORD-y coord2)))))

;; Variant of iterative deepening that marks NODEs to avoid 
;; duplicating search effort.  This version will not search
;; deeper than 50 nodes thereby guaranteeing termination.
(defun aux-ids (start goalp next depth grid)
  (if (>= depth 50) nil
    (progn (GRID-clear grid)
	   (or (dfs-fd start goalp next 0 depth)
	       ;; Then look deeper if unnecessary
	       (aux-ids start goalp next (+ 1 depth) grid)))))

;; Create a 10 X 10 GRID.
(setq culdesac (make-GRID 10 10))

;; Set the obstacle LOCations to be t.
(set-LOC-occ (GRID-ref culdesac 2 1) t)
(set-LOC-occ (GRID-ref culdesac 3 1) t)
(set-LOC-occ (GRID-ref culdesac 4 1) t)
(set-LOC-occ (GRID-ref culdesac 5 1) t)
(set-LOC-occ (GRID-ref culdesac 5 2) t)
(set-LOC-occ (GRID-ref culdesac 5 3) t)
(set-LOC-occ (GRID-ref culdesac 5 4) t)
(set-LOC-occ (GRID-ref culdesac 5 5) t)
(set-LOC-occ (GRID-ref culdesac 4 5) t)
(set-LOC-occ (GRID-ref culdesac 3 5) t)
(set-LOC-occ (GRID-ref culdesac 2 5) t)
(set-LOC-occ (GRID-ref culdesac 8 5) t)
(set-LOC-occ (GRID-ref culdesac 7 6) t)
(set-LOC-occ (GRID-ref culdesac 6 7) t)
(set-LOC-occ (GRID-ref culdesac 5 8) t)

;; Display the grid.
;; > (GRID-display culdesac)
;;     0  1  2  3  4  5  6  7  8  9 
;;  0  +  +  +  +  +  +  +  +  +  + 
;;  1  +  +  +  +  +  +  +  +  +  + 
;;  2  +  X  +  +  +  X  +  +  +  + 
;;  3  +  X  +  +  +  X  +  +  +  + 
;;  4  +  X  +  +  +  X  +  +  +  + 
;;  5  +  X  X  X  X  X  +  +  X  + 
;;  6  +  +  +  +  +  +  +  X  +  + 
;;  7  +  +  +  +  +  +  X  +  +  + 
;;  8  +  +  +  +  +  X  +  +  +  + 
;;  9  +  +  +  +  +  +  +  +  +  + 

;; Here is a solution to the first part of the homework.

;; Return the unoccupied and unmarked neighbors of a GRID LOCation.
(defun GRID-neighbors (grid coord)
  (let ((x (COORD-x coord)) (y (COORD-y coord)))
    (mapcan #'(lambda (i j) 
                (let ((loc (GRID-ref grid (+ x i) (+ y j))))
                  (if (or (null loc) (LOC-mark loc) (LOC-occ loc))
                    nil (list loc))))
            '(-1 0 1 0) '(0 -1 0 1))))

;; Here are some routines for keeping track of the number of nodes 
;; expanded during the invocation of a particular search routine. 
(let ((num-nodes 0))
  (defun report-num-nodes () num-nodes)
  (defun increment-num-nodes () (setq num-nodes (+ 1 num-nodes)))
  (defun reset-num-nodes () (setq num-nodes 0)))

;; The following generic path planner works with any one of four search
;; functions: depth-first search (#'dfs), breadth-first search (#'bfs), 
;; iterative deepening (#'ids) and best-first search (#'best).
;; In the case of the latter two, the arguments have to be massaged 
;; somewhat and in the case of iterative deepening we use a modified
;; version that marks the GRID to avoid duplicating search effort.
(defun plan (grid start goal searchfun)
  (GRID-clear grid)   ;; Set all the marks in the GRID to nil.
  (reset-num-nodes)   ;; Set the number of nodes expanded to zero.
  (if (eq searchfun #'ids)
    (setq searchfun #'(lambda (nodes goalp next)
                        (aux-ids (first nodes) goalp next 1 grid))))
  (if (eq searchfun #'best)
    (setq searchfun #'(lambda (nodes goalp next)
                        (best nodes goalp next
                              #'(lambda (node1 node2)
                                  (< (manhattan (LOC-coord (NODE-loc node1)) 
						goal)
                                     (manhattan (LOC-coord (NODE-loc node2)) 
						goal)))))))
  ;; The start and goal are specified as COORDinates.
  (let ((loc (GRID-ref grid (COORD-x start) (COORD-y start))))
    (funcall searchfun 
             (list (make-NODE loc (list loc)))
             #'(lambda (node)
                 ;; Increment the number of NODEs expanded.
                 (increment-num-nodes)
                 (equal (LOC-coord (NODE-loc node)) goal))
             #'(lambda (node)
                 (mapcar #'(lambda (loc)
                             ;; Mark the LOCation as visited.  
                             (set-LOC-mark loc t)
                             (make-NODE loc 
                                        (cons loc (NODE-path node))))
                         (GRID-neighbors grid (LOC-coord (NODE-loc node))))))))

;; Test the path planner out with the culdesac GRID. 
(defun test (searchfun)
  (setq *print-pretty* t)
  (princ (plan culdesac (make-COORD 0 0) (make-COORD 9 9) searchfun))
  (GRID-marks culdesac)
  (format t "~%Total number of nodes expanded: ~D~%" (report-num-nodes)))


;; Copyright 1994, Brown University, Providence, RI
;; 
;; Permission to use and modify this software and its documentation for any
;; purpose other than its incorporation into a commercial product is hereby
;; granted without fee.  Permission to copy and distribute this software and its
;; documentation only for non-commercial use is also granted without fee,
;; provided, however, that the above copyright notice appear in all copies, that
;; both that copyright notice and this permission notice appear in supporting
;; documentation, that the name of Brown University not be used in advertising or
;; publicity pertaining to distribution of the software without specific, written
;; prior permission, and that the person doing the distribution notify Brown
;; University of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this software for
;; any purpose.  It is provided "as is" without express or implied warranty.
;; Brown University requests notification of any modifications to this software
;; or its documentation.
;; 
;; Send the following redistribution information:
;; 
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;; 
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;; 
;; 		or
;; 
;; 	brusd@cs.brown.edu
;; 
;; We will acknowledge all electronic notifications.
