;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;; Here is the code for a simple implementation of discrimination trees.
;; For completeness, we begin with the abstract data types for variables.
(defun make-VAR (var) (list '? var))
(defun is-VAR (x) (and x (listp x) (eq (car x) '?)))

;; Indices are abstracted even though they are relatively simple.
(defun make-INDEX (key dt) (cons key dt))
(defun INDEX-key (dt) (car dt))
(defun INDEX-dt (dt) (cdr dt))

;; We also make explicit our use of special symbols for indicating
;; specific types of discrimination keys.
(defun is-var-key (key) (eq key '*var*))
(defun is-begin-key (key) (eq key '*begin*))
(defun is-end-key (key) (eq key '*end*))

;; Here is the abstract type for representing discrimination trees.
(defun make-DTREE () (list 'DTREE () ()))
(defun is-DTREE (x) (and x (listp x) (eq (car x) 'DTREE)))
(defun DTREE-links (dt) (cadr dt))
(defun DTREE-contents (dt) (caddr dt))
(defun set-DTREE-links (dt links) (setf (cadr dt) links))
(defun set-DTREE-contents (dt contents) (setf (caddr dt) contents))

;; Other basic operations on discrimination trees are recorded here in 
;; an effort to keep most of the discrimination tree specific code 
;; centralized. To simplify the code, we make use of the Common Lisp 
;; iterative construct "do", described in Chapter 2.
(defun DTREE-link-existing-index (dt key)
  (let ((link (assoc key (DTREE-links dt))))
    (if link (INDEX-dt link) nil)))

(defun DTREE-rawcontents (dt)
  (append 
   (DTREE-contents dt)
   (do ((dts (mapcar #'INDEX-dt (DTREE-links dt)) (cdr dts))
        (contents () (append contents 
                              (DTREE-rawcontents (car dts)))))
       ((null dts) contents))))

(defun DTREE-link-index (dt key)
  (let ((link (assoc key (DTREE-links dt))))
    (if link (INDEX-dt link)
      (progn (setq link (make-INDEX key (make-DTREE)))
             (set-DTREE-links dt (cons link (DTREE-links dt)))
             (INDEX-dt link)))))

;; Make-key-list takes a pattern expression and returns a flattened list
;; of discrimination keys
(defun make-key-list (pat)
  (cond ((null pat) (list ()))
        ((is-VAR pat) (list '*var*))
        ((symbolp pat) (list pat))
        (t (cons '*begin*
                 (nconc (mapcan #'make-key-list pat)
                        (list '*end*))))))

;; Given a pattern and a discrimination tree, index returns a terminal
;; discrimination tree corresponding to that pattern.
(defun index (pattern dt)
  (aux-index (make-key-list pattern) dt))
(defun aux-index (keys dt)
  (if (null keys) dt
    (aux-index (cdr keys) 
	       (DTREE-link-index dt (car keys)))))

;; Given a pattern and a discrimination tree, extract returns all 
;; terminal nodes that might possibly match. Note that extract is 
;; not supposed to perform unification. In a complete system, the 
;; resulting list would be filtered using a more discriminating 
;; form of matching to eliminate false matches. For example, 
;; unification would eliminate (on (? x) (? x)) as a match for (on a b).
(defun extract (pattern dt)
  (aux-extract (make-key-list pattern) dt))
(defun aux-extract (keys dt)
  (cond ((null keys) (list dt))
        ((is-var-key (car keys))
         (var-extract (cdr keys) dt))
        ((is-begin-key (car keys))
         (subexp-extract keys dt))
        ((is-end-key (car keys))
         (key-extract (car keys) (cdr keys) dt))
        (t (nconc (key-extract (car keys) (cdr keys) dt)
                  (key-extract '*var* (cdr keys) dt)))))

;; Key-extract takes a key, a list of keys, and a discrimination tree and
;; returns a list of discrimination trees.
(defun key-extract (key keys dt)
  (let ((dt-maybe (DTREE-link-existing-index dt key)))
    (if dt-maybe (aux-extract keys dt-maybe) ())))

;; Subexp-extract takes a list of keys and a discrimination tree and 
;; returns a list of discrimination trees. If the key indicates the 
;; beginning of a subexpression, traverse the *begin* and *var* links. 
;; In the latter case, eliminate the keys corresponding to the 
;; subexpression.
(defun subexp-extract (keys dt)
  (nconc 
   (key-extract (car keys) (cdr keys) dt)
   (key-extract '*var* (strip-off-subexp-keys (cdr keys) 1) dt)))

;; Var-extract takes a list of keys and a discrimination network and 
;; traverses all atomic indices plus all sets of links corresponding to 
;; subexpressions. Var-extract returns a list of discrimination networks.
(defun var-extract (keys dt)
  (mapcan #'(lambda (link)
              (cond ((is-begin-key (INDEX-key link))
                     (skip-subexp keys (INDEX-dt link) 1))
                    ((is-end-key (INDEX-key link))
                     (error "VAR-EXTRACT: Ill formed list"))
                    (t (aux-extract keys (INDEX-dt link)))))
          (DTREE-links dt)))

;; Subexp-extract is called with its first argument, keys, bound to the 
;; list (*begin* ...), and its second argument, dt, bound to some 
;; discrimination tree. To extract discrimination trees from the *var* 
;; branch of a discrimination tree, we ignore the keys corresponding to 
;; the indicated subexpression. To do so, invoke strip-off-subexp-keys 
;; with (rest keys) and 1.
(defun strip-off-subexp-keys (keys level)
  (cond ((= level 0) keys)
        ((is-begin-key (car keys)) 
         (strip-off-subexp-keys (cdr keys) (+ level 1)))
        ((is-end-key (car keys)) 
         (strip-off-subexp-keys (cdr keys) (- level 1)))
        (t (strip-off-subexp-keys (cdr keys) level))))

;; In discriminating on a variable key, you have to skip over (or 
;; suspend discrimination on) all sets of indices corresponding to 
;; subexpressions. The level argument to skip-subexp indicates the 
;; level of nesting (i.e. the number of *begin* keys encountered). 
;; Skip-subexp is called while discriminating on a variable key when 
;; a *begin* key is encountered and hence the level argument is 
;; initialized to 1.
(defun skip-subexp (keys dt level)
  (if (= level 0) (aux-extract keys dt)
    (mapcan 
     #'(lambda (link)
          (cond ((is-begin-key (INDEX-key link))
                 (skip-subexp keys (INDEX-dt link) (+ level 1)))
                ((is-end-key (INDEX-key link))
                 (skip-subexp keys (INDEX-dt link) (- level 1)))
                (t (skip-subexp keys (INDEX-dt link) level))))
     (DTREE-links dt))))

;; Here are some routines with a local variable corresponding to
;; a discrimination tree used for storing logical formulas. 
;; Again we make use of the Common Lisp iterative construct "do". 
;; The function "add" inserts a pattern into the database and the 
;; function "fetch" retrieves all objects matching the pattern.
(let ((dt* (make-DTREE)))
  (defun add (pat) 
    (let ((dt (index pat dt*)))
      (if (member pat (DTREE-contents dt)) pat
	(set-DTREE-contents dt (cons pat (DTREE-contents dt))))))
  (defun fetch (pat)
    (do ((dts (extract pat dt*) (cdr dts)) 
         (results () (append results (DTREE-contents (car dts)))))
        ((null dts) results))))

;; The following constructs the database described in the chapter.
(defun test ()
  (mapc #'add '((on block2 (floor room17))
		(on block1 table2)
		(on (? x) (? y)) 
		(on (? z) (? z)) 
		(instance block block1)
		(instance block block2)))
  (fetch '(on (? x) (floor (? y)))))

;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu

;;
;; We will acknowledge all electronic notifications.

