;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;; A chromosome is represented as a list of alleles. An allele 
;; is a boolean value, 0 or 1. An individual is an instance of 
;; a particular chromosome, called the individual's genotype. 
   
;; Given an existing population, we compute the next generation by
;; selecting pairs of individuals according to the above distribution.
;; Each pair is given an opportunity to exchange genetic material by
;; calling the function crossover, which, with some probability,
;; destructively modifies the list structures corresponding to the
;; chromosomes of the two individuals.
(defun reproduce (population)
  (let ((offspring ()) (d (distribution population)))
    (dotimes (i (/ (length population) 2))
      (let ((x (select d)) (y (select d)))
        (crossover x y)
        (setq offspring (nconc (list x y) offspring))))
    offspring))

;; Select takes a distribution and returns an individual chosen in 
;; accord with the given distribution. Select also invokes a procedure 
;; that serves to mutate the selected genotype in constructing the 
;; new individual.
(defun select (distribution)
  (let ((rnd (random 1.0)) (prob 0) genotype)
    (some #'(lambda (pair)
              (setq prob (+ prob (car pair)))
              (if (> rnd prob) nil
                  (progn (setq genotype (cdr pair))
			 t)))
          distribution)
    (mutate genotype)))

;; Copy a bit string inverting bits 3% of the time.
;; Mutation is carried out by creating a new string in which each bit is 
;; copied from the old to the new without change with probability 0.97
;; and a bit is inverted with probability 0.03.
(defun mutate (genotype)
  (mapcar #'(lambda (x)
              (if (> (random 1.0) 0.03) x (if (= x 1) 0 1)))
          genotype))

;; We are intereseted in finding strings that have a high value with 
;; respect to a given evaluation function. In keeping with the natural 
;; selection motivation, the evaluation function is called fitness. 
;; The distribution for selecting individuals that are allowed to 
;; reproduce is described as follows.
(defun distribution (population)
  (let* ((genotypes (noduplicates population))
         (sum (apply #'+ (mapcar #'fitness genotypes))))
    (mapcar #'(lambda (x) (cons (/ (fitness x) sum) x))
            genotypes)))

;; Crossover at a random site 60% of the time.  Crossover corresponds to
;; two individuals exchanging genetic material.  In this implementation,
;; crossover occurs between a selected pair of individuals with
;; probability 0.6. A crossover sit is an integer specifying an element
;; of a list representing a bit string. Given two strings, with
;; probability 0.6, crossover randomly chooses a crossover site,and
;; exchanges the substrings beginning immediately following the site in
;; each of the strings. For instance, given (0 1 1 0 1) and (1 1 0 1 1),
;; crossover might choose the (0-based) site location of 1 and then
;; destructively modify the two strings to obtain (0 1 0 1 1) and 
;; (1 1 1 0 1).  The crossover rate of 0.6 and the mutation rate of 0.03
;; are recommended as initial values for these parameters [Goldberg,
;; 1989] but often it is necessary to experiment to find values that work
;; well for a given application.
(defun crossover (x y)
  (if (> (random 1.0) 0.6) (list x y)
      (let* ((site (random (length x)))
             (swap (cdr (nthcdr site x))))
        (setf (cdr (nthcdr site x)) (cdr (nthcdr site y)))
        (setf (cdr (nthcdr site y)) swap))))

;; Generate a population of individuals.
(defun population (n)
  (let ((population ()))
    (dotimes (i n) (setq population 
                         (cons (random-individual) population)))
    population))

;; Generate a random individual with 30 alleles.
(defun random-individual ()
  (let ((individual ()))
    (dotimes (i 30) (setq individual
                          (cons (random 2) individual)))
    individual))

;; Generate a random individual with 30 alleles most of them 0. 
(defun random-small-individual ()
  (let ((individual ()))
    (dotimes (i 30) (if (> i 10) 
		     (setq individual (cons (random 2) individual))
		     (setq individual (cons 0 individual))))
    individual))

;; f(x) = (x/(2^{30}-1))^10
(defun fitness (individual) 
  (raise (/ (decimal individual) (- (raise 2 30) 1.0)) 10))

(defun gaussian (x c w)
  (exp (- (/ (* (- x c) (- x c)) (* 2 w)))))

(defun multiple (individual)
  (let ((x (/ (decimal individual) (- (raise 2 30) 1.0))))
    (+ (* 0.50 (gaussian x 0.125 0.003125))
       (* 0.25 (gaussian x 0.375 0.003125))
       (* 1.00 (gaussian x 0.625 0.003125))
       (* 0.25 (gaussian x 0.875 0.003125)))))

;; Determine the highest rating for any individual.
(defun fittest (population)
  (apply #'max (mapcar #'fitness population)))

(defun average (population)
  (/ (apply #'+ (mapcar #'fitness population))
     (length population)))

;; Test functions

(defun test ()
  (evolve (population 30) 10))

(defun evolve (population generations)
  (format t "{~6,4F, ~6,4F}, ~%" 
          (fittest population) (average population))
  (if (= generations 0) nil
    (evolve (reproduce population) (- generations 1))))

;; Utilities

(defun raise (x n) (if (= n 0) 1 (* x (raise x (- n 1)))))

(defun decimal (bits)
  (let ((i -1))
    (apply #'+ (mapcar #'(lambda (b) 
			   (setq i (+ i 1))
                           (* b (raise 2 i)))
                      bits))))

(defun noduplicates (list)
  (let ((set ()))
    (mapc #'(lambda (x) 
              (setq set (adjoin x set :test #'equal)))
          list)
    set))
                                     

;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu

;;
;; We will acknowledge all electronic notifications.

