;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;; A chromosome is represented as a list of alleles. An allele 
;; is a boolean value, 0 or 1. An individual is an instance of 
;; a particular chromosome, called the individual's genotype. 
   
;; Given an existing population, we compute the next generation by
;; selecting pairs of individuals according to the above distribution.
;; Each pair is given an opportunity to exchange genetic material by
;; calling the function crossover, which, with some probability,
;; destructively modifies the list structures corresponding to the
;; chromosomes of the two individuals.
(define (reproduce population)
  (let ((offspring '()) (d (distribution population)))
    (do ((i 1 (+ 1 i)))
        ((> i (/ (length population) 2)))
      (let ((x (select d)) (y (select d)))
        (crossover x y)
        (set! offspring (append (list x y) offspring))))
    offspring))


;; Select takes a distribution and returns an individual chosen in 
;; accord with the given distribution. Select also invokes a procedure 
;; that serves to mutate the selected genotype in constructing the 
;; new individual.
(define (select distribution)
  (let ((rnd (random 1.)) (prob 0) (genotype #f))
    (some2 (lambda (pair)
             (set! prob (+ prob (car pair)))
             (if (> rnd prob) #f (begin (set! genotype (cdr pair)) #t)))
           distribution)
    (mutate genotype)))


;; Copy a bit string inverting bits 3% of the time.
;; Mutation is carried out by creating a new string in which each bit is 
;; copied from the old to the new without change with probability 0.97
;; and a bit is inverted with probability 0.03.
(define (mutate genotype)
  (map (lambda (x) (if (> (random 1.) .03) x (if (= x 1) 0 1))) genotype))


;; We are intereseted in finding strings that have a high value with 
;; respect to a given evaluation function. In keeping with the natural 
;; selection motivation, the evaluation function is called fitness. 
;; The distribution for selecting individuals that are allowed to 
;; reproduce is described as follows.
(define (distribution population)
  (let* ((genotypes (noduplicates population))
         (sum (apply + (map fitness genotypes))))
    (map (lambda (x) (cons (/ (fitness x) sum) x)) genotypes)))


;; Crossover at a random site 60% of the time.  Crossover corresponds to
;; two individuals exchanging genetic material.  In this implementation,
;; crossover occurs between a selected pair of individuals with
;; probability 0.6. A crossover sit is an integer specifying an element
;; of a list representing a bit string. Given two strings, with
;; probability 0.6, crossover randomly chooses a crossover site,and
;; exchanges the substrings beginning immediately following the site in
;; each of the strings. For instance, given (0 1 1 0 1) and (1 1 0 1 1),
;; crossover might choose the (0-based) site location of 1 and then
;; destructively modify the two strings to obtain (0 1 0 1 1) and 
;; (1 1 1 0 1).  The crossover rate of 0.6 and the mutation rate of 0.03
;; are recommended as initial values for these parameters [Goldberg,
;; 1989] but often it is necessary to experiment to find values that work
;; well for a given application.
(define (crossover x y)
  (if (> (random 1.) .6)
    (list x y)
    (let* ((site (random (length x))) (swap (cdr (nthcdr site x))))
      (set-cdr! (nthcdr site x) (cdr (nthcdr site y)))
      (set-cdr! (nthcdr site y) swap))))


;; Generate a population of individuals.
(define (population n)
  (let ((population '()))
    (do ((i 1 (+ 1 i)))
        ((> i n))
      (set! population (cons (random-individual) population)))
    population))


;; Generate a random individual with 30 alleles.
(define (random-individual)
  (let ((individual '()))
    (do ((i 1 (+ 1 i)))
        ((> i 30))
      (set! individual (cons (random 2) individual)))
    individual))


;; Generate a random individual with 30 alleles most of them 0. 
(define (random-small-individual)
  (let ((individual '()))
    (do ((i 1 (+ 1 i)))
        ((> i 30))
      (if (> i 10)
        (set! individual (cons (random 2) individual))
        (set! individual (cons 0 individual))))
    individual))


;; f(x) = (x/(2^{30}-1))^10
(define (fitness individual)
  (raise (/ (decimal individual) (- (raise 2 30) 1.)) 10))


(define (gaussian x c w) (exp (- (/ (* (- x c) (- x c)) (* 2 w)))))


(define (multiple individual)
  (let ((x (/ (decimal individual) (- (raise 2 30) 1.))))
    (+ (* .5 (gaussian x .125 .003125))
       (* .25 (gaussian x .375 .003125))
       (* 1. (gaussian x .625 .003125))
       (* .25 (gaussian x .875 .003125)))))


;; Determine the highest rating for any individual.
(define (fittest population) (apply max (map fitness population)))


(define (average population)
  (/ (apply + (map fitness population)) (length population)))


;; Test functions

(define (test) (evolve (population 30) 10))


(define (evolve population generations)
  (begin
    (display "{")
    (display (fittest population))
    (display ", ")
    (display (average population))
    (display "}, ")
    (newline))
  (if (= generations 0) #f (evolve (reproduce population) (- generations 1))))


;; Utilities

(define (raise x n) (if (= n 0) 1 (* x (raise x (- n 1)))))


(define (decimal bits)
  (let ((i -1))
    (apply + (map (lambda (b) (set! i (+ i 1)) (* b (raise 2 i))) bits))))


(define (noduplicates list)
  (let ((set '()))
    (for-each (lambda (x) (set! set (adjoin2 x set equal?))) list)
    set))

                                     

;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;;      Name:
;;      Organization:
;;      Address (postal and/or electronic):
;;
;; To:
;;      Software Librarian
;;      Computer Science Department, Box 1910
;;      Brown University
;;      Providence, RI 02912
;;
;;              or
;;
;;      brusd@cs.brown.edu

;;
;; We will acknowledge all electronic notifications.

