(in-package 'user)

;; This file contains all of the code implementing the decision
;; tree method described in the chapter on learning.

(setq *print-pretty* t)

;; A feature is an attribute together with a value.
(defun FEATURE-attribute (feature) (first feature))
(defun FEATURE-value (feature) (first (rest feature)))

;; A dimension is an attribute together with a set of possible values.
(defun make-DIMENSION (attribute values) (list attribute values))
(defun DIMENSION-attribute (dimension) (first dimension))
(defun DIMENSION-values (dimension) (first (rest dimension)))

;; A decision tree node consists of examples and child nodes.
(defun make-NODE (examples) (list nil examples nil))
(defun NODE-attribute (node) (first node))
(defun set-NODE-attribute (node attribute)
  (setf (first node) attribute))
(defun NODE-examples (node) (first (rest node)))
(defun NODE-children (node) (first (rest (rest node))))
(defun set-NODE-children (node children) 
  (setf (first (rest (rest node))) children))

;; An example is an identifier together with a set of features 
;; that describe the example and a class indicator.
(defun make-EXAMPLE (id features class) (list id features class))
(defun EXAMPLE-id (example) (first example))
(defun EXAMPLE-features (example) (first (rest example)))
(defun EXAMPLE-class (example) (first (rest (rest example))))

;;
;; buildtree
;;
;; This is the recursive function that do the building of the decision
;; tree (hence the name). The algorithm follows the algorithm described
;; in the text pg. 257-258.
;; Make sure you understand how this work before you modify it.
;; Understanding this function is the main part of this assignment
;;

(defun dtree (examples)
  (let ((root-node (make-NODE examples)))
    (buildtree root-node)
    root-node))

(defun buildtree (node)
  (format t "Class distribution: ~A~%"
          (mapcar #'length (classify (NODE-examples node))))
  (if (< 1 (length (findall (classify (NODE-examples node)) 
			    #'(lambda (s) (not (null s)))))) 
      (let ((dim (splitter node)))
	(format t "Dimension chosen: ~A~%" dim)
	(set-NODE-attribute node (DIMENSION-attribute dim))
	(set-NODE-children node
			   (mapcar #'make-NODE 
				   (partition (NODE-examples node) dim)))
	(mapc #'buildtree (NODE-children node)))
    (format t "Terminal node.~%" )))

;;
;; classify
;;

(defun classify (examples)
  (mapcar #'(lambda (class)
	      (findall examples #'(lambda (s) 
				    (eq class (EXAMPLE-class s)))))
	  (classes)))

;;
;; findall
;;

(defun findall (list test)
  (do ((items list (cdr items))
       (results nil))
      ((null items) results)
      (if (funcall test (car items)) 
	  (setq results (adjoin (car items) results
				:test #'equal)))))


;;
;; partition
;;

(defun partition (examples dimension)
  (let ((attribute (DIMENSION-attribute dimension)))
    (mapcar #'(lambda (value) 
		(findall examples
			 #'(lambda (example)
			     (member (list attribute value)
				     (EXAMPLE-features example)
				     :test #'equal))))
	    (DIMENSION-values dimension))))

;;
;; splitter
;;

(defun splitter (node) 
  (do* ((dims (cdr (dimensions)) (cdr dims))
	(value (evaluate node (car (dimensions))))
	(splitter (car (dimensions))) (min value))
       ((null dims) splitter)
       (setq value (evaluate node (car dims)))
       (if (< value min)
	   (setq splitter (car dims) min value))))

;;
;; evaluate
;;

(defun evaluate (node dimension)
  (let ((total (length (NODE-examples node))))
    (apply #'+
	   (mapcar #'(lambda (subset)
		       (* (/ (length subset) total)
			  (disorder (classify subset))))
		   (partition (NODE-examples node) dimension)))))
  
;;
;; disorder
;;

(defun disorder (partition)
  (let* ((sizes (mapcar #'length partition))
	 (total (apply #'+ sizes)))
    (if (= total 0) 0
      (- (log total 2)
	 (apply #'+ (mapcar #'(lambda (s) 
				(if (= s 0) 0 
				  (* (/ s total) (log s 2))))
			    sizes))))))

;; Examples

(defun sunburn ()
  (let* ((examples
	  (list
	   (make-EXAMPLE 'Sarah
			 '((hair blonde) (height average) 
			   (weight light) (lotion no)) 'yes)
	   (make-EXAMPLE 'Dana
			 '((hair blonde) (height tall) 
			   (weight average) (lotion yes)) 'no)
	   (make-EXAMPLE 'Alex
			 '((hair brown) (height short) 
			   (weight average) (lotion yes)) 'no)
	   (make-EXAMPLE 'Annie
			 '((hair blonde) (height short) 
			   (weight average) (lotion no)) 'yes)
	   (make-EXAMPLE 'Emily
			 '((hair red) (height average) 
			   (weight heavy) (lotion no)) 'yes)
	   (make-EXAMPLE 'Pete
			 '((hair brown) (height tall) 
			   (weight heavy) (lotion no)) 'no)
	   (make-EXAMPLE 'John
			 '((hair brown) (height average) 
			   (weight heavy) (lotion no)) 'no)
	   (make-EXAMPLE 'Katie
			 '((hair blonde) (height short) 
			   (weight light) (lotion yes)) 'no)))
	 (dtree (make-NODE examples)))
    (defun dimensions ()
      (list (make-DIMENSION 'hair '(blonde brown red))
	    (make-DIMENSION 'height '(short tall average))
	    (make-DIMENSION 'weight '(light heavy average))
	    (make-DIMENSION 'lotion '(yes no))))
    (defun classes () '(yes no))
    (buildtree dtree)
    dtree))

(defun robots ()
  (let* ((examples
	 (list 
	  (make-EXAMPLE '307 '((status faculty) (floor three) 
			       (dept ee) (size large)) 'no)
	  (make-EXAMPLE '309 '((status staff) (floor three) 
			       (dept ee) (size small)) 'no)
	  (make-EXAMPLE '408 '((status faculty) (floor four) 
			       (dept cs) (size medium)) 'yes)
	  (make-EXAMPLE '415 '((status student) (floor four) 
			       (dept ee) (size large)) 'yes)
	  (make-EXAMPLE '509 '((status staff) (floor five) 
			       (dept cs) (size medium)) 'no)
	  (make-EXAMPLE '517 '((status faculty) (floor five) (dept cs) 
			       (size large)) 'yes)
	  (make-EXAMPLE '316 '((status student) (floor three) 
			       (dept ee) (size small)) 'yes)
	  (make-EXAMPLE '420 '((status staff) (floor four) (dept cs) 
			       (size medium)) 'no )))
	 (dtree (make-NODE examples)))
    (defun dimensions ()
      (list (make-DIMENSION 'status '(faculty staff student))
	    (make-DIMENSION 'floor '(three four five))
	    (make-DIMENSION 'dept '(cs ee))
	    (make-DIMENSION 'size '(large medium small))))
    (defun classes () '(yes no))
    (buildtree dtree)
    dtree))

(defun final ()
  (let* ((examples (list (make-EXAMPLE 1 '((a 1) (b 0) (c 1)) 1)
			(make-EXAMPLE 2 '((a 1) (b 1) (c 0)) 1)
			(make-EXAMPLE 3 '((a 1) (b 1) (c 1)) 0)
			(make-EXAMPLE 4 '((a 0) (b 0) (c 0)) 0)
			(make-EXAMPLE 5 '((a 1) (b 1) (c 0)) 1)
			(make-EXAMPLE 6 '((a 1) (b 0) (c 0)) 0)))
	 (dtree (make-NODE examples)))
    (defun dimensions ()
      (list (make-DIMENSION 'a '(0 1))
	    (make-DIMENSION 'b '(0 1))
	    (make-DIMENSION 'c '(0 1))))
    (defun classes () '(0 1))
    (buildtree dtree)
    dtree))
