(in-package 'user)

;; Performance is defined as the percentage of correctly classified examples.
(defun performance (node examples)
  (let ((m 0))
    (mapc #'(lambda (example) 
              (if (correctp node example) (setq m (+ 1 m))))
          examples)
    (/ m (length examples))))

;; A decision tree is correct with respect to an example if the example
;; is stored in a terminal of the tree marked with the same class.  
(defun correctp (node example)
  (if (null (NODE-children node))
      (equal (EXAMPLE-class example) (NODE-class node))
    (correctp (car (findall (NODE-children node)
                            #'(lambda (c)
                                (member (NODE-feature c)
                                        (EXAMPLE-features example)
                                        :test #'equal))))
              example)))

;; Compute the number of nodes in a decision tree.
(defun size (node)
  (+ 1 (apply #'+ (mapcar #'size (NODE-children node)))))

;; Compute the maximum depth of a decision tree. 
(defun depth (node)
  (if (null (NODE-children node)) 0
    (+ 1 (apply #'max (mapcar #'depth (NODE-children node))))))

;; Generate a given number of examples from the same concept. 
(defun examples (n)
  (do ((i 0 (+ i 1)) 
       (examples nil (cons (cexample) examples)))
      ((= i n) examples)))

;; Here are routines for running the experiments.
(defun test (threshold &optional (n 1))
  (set-threshold threshold)
  (set-distribution 0.5)
  (set-noise '((DEPT 0.8) (FLOOR 0.9)))
  (let ((total 0))
    (set-concept '(AND (OR (DEPT CS)) (OR (STATUS FACULTY))))
    (dotimes (i n) (setq total (+ (run) total)))
    (set-concept '(AND (OR (DEPT CS)) (OR (STATUS FACULTY) (STATUS STUDENT))))
    (dotimes (i n) (setq total (+ (run) total)))
    (set-concept '(AND (OR (FLOOR FIVE))))
    (dotimes (i n) (setq total (+ (run) total)))
    (format t "Average performance: ~4,2F~%" (/ total (* 3 n)))))
(defun run ()
  (let* ((training (examples 40))
         (evaluation (examples 40)) 
         (dtree (make-NODE training)) score (verbose nil))
    (buildtree dtree)
    (setq score (performance dtree evaluation))
    (if verbose
	(format t "Size: ~D~%Depth: ~D~%Performance: ~4,2F~%" 
		(size dtree) (depth dtree) score))
    score))

;; Exclusive or is typically difficult for decision trees. 
(defun cxor ()
  (set-concept '(OR (AND (STATUS FACULTY) (NOT (DEPT CS)))
		    (AND (NOT (STATUS FACULTY)) (DEPT CS)))))
