(in-package 'user)

;; We assume that implications and equivalences have already been
;; eliminated and that the infix forms have been converted to 
;; prefix forms. Disjunction, conjunction, and negation are
;; represented as, respectively, OR, AND, and NOT. 

;; A literal is an atomic expression or its negation. Ignore double negatives.
(defun literalp (q)
  (or (atomicp q) (and (eq (car q) 'not) (literalp (cadr q)))))

(defun atomicp (q)
  (or (not (consp q))
      (not (member (car q) '(OR AND NOT IF FI IFF)))))

(setq l1 '(not (not p))
      l2 '(not q))

;; Reverse OR and AND.
(defun revorand (f)
  (if (eq f 'and) 'or 'and))

;; Push negation in as far as possible, eliminate double negatives,
;; and represent literals as disjunctions with a single disjunct.
;; This last step will simplify dealing with special cases. We
;; will filter the result to eliminate unnecessary connectives. 
(defun negin (q)
  (cond ((and (consp q) (eq (car q) 'not) (consp (cadr q))
	      (eq (caadr q) 'not)) (negin (cadadr q)))
	((literalp q) (list 'or q))
	((eq (car q) 'not) (setq q (cadr q))
	 (cons (revorand (car q))
	       (mapcar #'negin
		       (mapcar #'(lambda (p) (list 'not p))
			       (cdr q)))))
	(t (cons (car q)
		 (mapcar #'negin (cdr q))))))

(setq c2 '(not (and (not p) (not (or q r))))
      c1 '(or (not p) (not (and (or p q) r))))

;; Combine nested conjunctions and disjunctions where appropriate.
(defun simplify (q)
  (car (aux-simplify q (revorand (car q)))))
(defun aux-simplify (q f)
  (cond ((literalp q) (list q))
	((eq (car q) f)
	 (mapcan #'(lambda (p) (aux-simplify p f))
		 (cdr q)))
	(t (list (cons (car q)
		       (mapcan #'(lambda (p) (aux-simplify p (revorand f)))
			       (cdr q)))))))

;; Determine if an expression is in conjunctive normal form.
(defun cnfp (q)
  (and (eq (car q) 'and)
       (every #'(lambda (p) (and (eq (car p) 'or)
				 (every #'literalp (cdr p))))
	      (cdr q))))

(setq t1 '(and (or p q (not r)) (or p) (or (not p) q))
      t2 '(and (or (and p q) r) (or (not q) (not p))))

;; If possible, find a conjunction in a disjunction and 
;; apply distribution to obtain a conjunction. 
(defun conjunctify (j)
  (let ((d (findand j))) 
    (if d (distribute d (remove d j)) j)))

;; Find a disjunct in a disjunction which is a conjunction.
(defun findand (j)				    
  (cond ((null j) nil)
	((and (consp (car j)) (eq (caar j) 'and)) (car j))
	(t (findand (cdr j)))))

;; Distribute a conjunction or disjunction according to de Morgan.
(defun distribute (x y)
  (cons (car x)
	(mapcar #'(lambda (z) (cons (car y) (cons z (cdr y))))
		(cdr x))))

(setq j1 '(or (and p q) r s)
      j2 '(or p q (and s t u) (and v w)))
      
;; Convert a formula into conjunctive normal form.
(defun cnf (q)
  (aux-cnf (simplify (list 'and (negin q)))))
(defun aux-cnf (q)
  (if (cnfp q) (filter q)
    (aux-cnf (simplify (cons 'and (mapcar #'conjunctify (cdr q)))))))
	
(defun cnf-nofilter (q)
  (aux-cnf-nofilter (simplify (list 'and (negin q)))))
(defun aux-cnf-nofilter (q)
  (if (cnfp q) q
    (aux-cnf-nofilter (simplify (cons 'and (mapcar #'conjunctify (cdr q)))))))

;; Eliminate any conjunctions or disjunctions with only one element.
(defun filter (q)
  (cond ((null q) nil)
	((literalp q) q)
	((and (member (car q) '(or and))
	      (= 1 (length (cdr q)))) (filter (cadr q)))
	(t (cons (filter (car q)) (filter (cdr q))))))

(setq f1 '(not (or p (and q r) (and s t)))
      f2 '(or (and p (or w v)) (and s t) (or u v))
      f3 '(not (and (or p (and q r)) s (or (t (and u v) w)))))

;; Test cases. 
;;
;; > l1
;; (NOT (NOT P))
;; > (cnf l1)
;; P
;; > c1
;; (OR (NOT P) (NOT (AND (OR P Q) R)))
;; > (cnf c1)
;; (AND (OR (NOT P) (NOT P) (NOT R)) (OR (NOT Q) (NOT P) (NOT R)))
;; > c2
;; (NOT (AND (NOT P) (NOT (OR Q R))))
;; > (cnf c2)
;; (OR P Q R)
;; > f1
;; (NOT (OR P (AND Q R) (AND S T)))
;; > (cnf f1)
;; (AND (NOT P) (OR (NOT Q) (NOT R)) (OR (NOT S) (NOT T)))
;; > f2
;; (OR (AND P (OR W V)) (AND S T) (OR U V))
;; > (cnf f2)
;; (AND (OR S P U V) (OR T P U V) (OR S W V U V) (OR T W V U V))
;; > f3
;; (NOT (AND (OR P (AND Q R)) S (OR (T (AND U V) W))))
;; > (cnf f3)
;; (AND (OR (NOT U) (NOT V) (NOT P) (NOT S)) 
;;      (OR (NOT W) (NOT P) (NOT S)) 
;;      (OR (NOT U) (NOT V) (NOT Q) (NOT R) (NOT S)) 
;;      (OR (NOT W) (NOT Q) (NOT R) (NOT S)))
