;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information
(in-package 'user)

;; The functions in this file implement simple classifier (or rule 
;; based) systems. These functions serve to evaluate the rules in a
;; set of rules using two different strategies (static and dynamic).
;; They also serve to evaluate the performance of a complete system
;; of rules and to classify inputs according to a system of rules.

;; Using the functions in this file and modifications to the 
;; functions in genetic.lisp, you can build a learning system for
;; boolean functions employing a hypothsis space of classifier 
;; systems and genetic algorithms to search for a good hypothesis.
;; You'll need to modify genetic.lisp to handle strings over the
;; alphabet {0, 1, *}. You should also consider a more conservative
;; approach to producing a new generation; you may want to avoid
;; replacing or mutating the best rules. See the end of this file
;; for an alternative mutation operator for classifier systems. 

;; Abstract data type for classifiers (or rules).
(defun classifier (condition action) (list 0 (cons action condition)))
(defun condition (classifier) (cdr (second classifier)))
(defun action (classifier) (car (second classifier)))
(defun fitness (classifier) (first classifier))
(defun genotype (classifier) (second classifier))
(defun set-fitness (classifier value) (setf (first classifier) value))

;; Here is a small set of rules for testing purposes.
(setq rules (list (classifier '(1 * * 1) 0) 
                  (classifier '(1 1 * *) 1)
                  (classifier '(1 0 * 1) 1)
                  (classifier '(1 0 * *) 0)))

;; Abstract data type for input/output pairs.
(defun pair (input output) (list input output))
(defun input (pair) (first pair))
(defun output (pair) (second pair))

;; The function that returns 1 if at least two inputs are 1.
(setq two (list (pair '(0 0 0 0) 0) (pair '(0 0 0 1) 0)
                (pair '(1 0 0 0) 0) (pair '(1 0 0 1) 1)
                (pair '(0 1 0 0) 0) (pair '(0 1 0 1) 1)
                (pair '(1 1 0 0) 1) (pair '(1 1 0 1) 1)
                (pair '(0 0 1 0) 0) (pair '(0 0 1 1) 1)
                (pair '(1 0 1 0) 1) (pair '(1 0 1 1) 1)
                (pair '(0 1 1 0) 1) (pair '(0 1 1 1) 1)
                (pair '(1 1 1 0) 1) (pair '(1 1 1 1) 1)))

;; Competitive evaluation is a linear combination of specificity 
;; and the current fitness value. 
(let ((alpha 1) (beta 1))
  (defun evaluate (classifier)
    (+ (* alpha (specificity classifier))
       (* beta (fitness classifier)))))

;; The specificity of a classifier is a number between 1 and 0
;; that is largest when the condition of the classifier applies
;; to only one input and smallest when it applies to all inputs.
(defun specificity (classifier)
  (- 1 (/ (count '* (condition classifier)) 
          (length (condition classifier)))))

;; Count the number of times that an item appears in a list.
;; (defun count (item list)
;;   (apply #'+ (mapcar #'(lambda (x) 
;;                             (if (eq x item) 1 0)) 
;;                        list)))

;; Matching with wildcards.
(defun match (input condition)
  (if (null input) t
      (if (or (eq (car condition) '*)
              (eq (car condition) (car input)))
         (match (cdr input) (cdr condition)) nil)))

;; The basic algorithm for dynamic evaluation:
;; Set all classifiers to have the same initial value.
;; For i = 1 to max_number_of_trials
;;    Choose a random example (input/output pair).
;;    Choose a classifier from those that match the input.
;;    If output = the action of the chosen classifier
;;       then increment the fitness by delta 
;;       else decrement the fitness by delta
;; Dynamic evaluation relies on six adjustable parameters.
;;   alpha   specificity multipler used in competitive evaluation
;;   beta    fitness multiplier used in competitive evaluation
;;   gamma   probability the most highly valued rule is adjusted
;;   delta   amount that fitness is adjusted per trial
;;   sigma   initial fitness value for rules 
;;   epochs  multiplier for number of trials 
(let ((epochs 10) (delta 1) (sigma 10))
  (defun dynamic (rules pairs)
    (dolist (rule rules) (set-fitness rule sigma))
    (let (pair possible rule (m (length pairs)))
      (dotimes (i (* epochs (length pairs)))
        (setq pair (nth (random m) pairs))
        (setq possible 
              (findall rules #'(lambda (rule) 
                                 (match (input pair) 
                                        (condition rule)))))
        (if possible 
          (progn (setq rule (choose possible))
                 (if (eq (action rule) (output pair))
                   (set-fitness rule (+ (fitness rule) delta))
                   (set-fitness rule (- (fitness rule) delta)))))))))

;; Most of the time choose the most highly rated classifier 
;; but some of the time choose a random classifier.
(let ((gamma 0.7))
  (defun choose (rules)
    (if (> (random 1.0) gamma)
      (findmax rules #'evaluate)
      (nth (random (length rules)) rules))))

;; Calculate the mean of the absolute value of the error over
;; all input/output pairs. The following performance 
;; measure has a maximum of one and a minimum of zero.
(defun performance (rules pairs)
  (let ((total 0))
    (dolist (pair pairs)
      (setq total (+ total (abs (- (classify rules 
                                             (input pair))
                                   (output pair))))))
    (- 1 (/ total (length pairs)))))

;; Determine the output for a particular input and classifier system.
;; If no rules apply for a given input, return a default output of 0. 
;; This default strategy amounts to including in every classifier
;; system a rule with the condition (* * * ... *) and the action 0.
(defun classify (rules input)
  (let ((possible (findall rules 
                           #'(lambda (rule)
                               (match input
                                      (condition rule))))))
    (if (null possible) 0
        (action (findmax possible #'specificity)))))

(defun findmax (list evaluator)
  (let ((best (car list)) value 
        (max (funcall evaluator (car list))))
    (dolist (item list)
      (setq value (funcall evaluator item))
      (if (> value max) (setq max value best item)))
    best))

;; Find all items in list such that test returns non-NIL. 
(defun findall (list test)
  (let ((results nil))
    (dolist (item list)
        (if (funcall test item) 
          (setq results (cons item results))))
    results))
  
;; A simple static evaluator. Note that this evaluator has no
;; adjustable parameters in contrast to the dynamic evaluator
;; described above which has six adjustable parameters. 
(defun static (rules pairs)
  (dolist (rule rules) (set-fitness rule 0))
  (dolist (pair pairs)
    (let ((pos nil) (neg nil) posreinf negreinf)
      (dolist (rule rules)
        (if (match (input pair) (condition rule))
          (if (eq (output pair) (action rule))
            (setq pos (cons rule pos))
            (setq neg (cons rule neg)))))
      (if pos
        (progn (setq posreinf (/ 1 (length pos)))
               (dolist (rule pos)
                 (set-fitness rule (+ (fitness rule) posreinf)))))
      (if neg
        (progn (setq negreinf (/ -1 (length neg)))
               (dolist (rule neg)
                 (set-fitness rule (+ (fitness rule) negreinf))))))))

;; > (performance rules two)
;; 11/16
;; > (classify rules '(1 0 0 1))
;; 1
;; > (static rules two)
;; NIL
;; > rules
;; ((-3 (0 1 * * 1)) (4 (1 1 1 * *))
;;                   (2 (1 1 0 * 1))
;;                   (-1 (0 1 0 * *)))
;; > (dynamic rules two)
;; NIL
;; > rules 
;; ((-8 (0 1 * * 1)) (38 (1 1 1 * *))
;;                   (22 (1 1 0 * 1))
;;                   (2 (0 1 0 * *)))

;; A mutation operator for classifier systems. 
(let ((arate 0.01) (crate 0.03))
  (defun mutate (rule)
    (classifier (mapcar #'(lambda (x)
                            (if (> (random 1.0) crate) x
                                (if (= x 1) 0 1)))
                        (condition rule))
                (if (> (random 1.0) arate) 
                  (action rule)
                  (if (= (action rule) 1) 0 1)))))


;; Copyright 1994, Brown University, Providence, RI
;; 
;; Permission to use and modify this software and its documentation for any
;; purpose other than its incorporation into a commercial product is hereby
;; granted without fee.  Permission to copy and distribute this software and its
;; documentation only for non-commercial use is also granted without fee,
;; provided, however, that the above copyright notice appear in all copies, that
;; both that copyright notice and this permission notice appear in supporting
;; documentation, that the name of Brown University not be used in advertising or
;; publicity pertaining to distribution of the software without specific, written
;; prior permission, and that the person doing the distribution notify Brown
;; University of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this software for
;; any purpose.  It is provided "as is" without express or implied warranty.
;; Brown University requests notification of any modifications to this software
;; or its documentation.
;; 
;; Send the following redistribution information:
;; 
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;; 
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;; 
;; 		or
;; 
;; 	brusd@cs.brown.edu
;; 
;; We will acknowledge all electronic notifications.
