;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information
(in-package 'user)

;; This file depends on the routines in normal.lisp to convert
;; formulae specifying concepts into conjunctive normal form.
(load "/u/tld/papers/books/intro/logic/lisp/normal")

;; This file contains functions for generating training examples for 
;; concepts expressed as boolean combinations of features. The generator 
;; provides control over the frequency of positive and negative examples, 
;; and allows for the introduction of noise by corrupting features according 
;; to a specified distribution. To set things up to generate examples,
;; you have to set the list of dimensions, the noise model, the concept
;; formula and the distribution of positive and negative examples. Here are
;; the default values that are automatically set when this file is loaded.
;;
;; Specify the dimensions for the example space.
;;
;; (set-dimensions (list (make-DIMENSION 'floor '(four five))
;;                       (make-DIMENSION 'dept '(cs ee))
;;                       (make-DIMENSION 'status '(student faculty staff))
;;                       (make-DIMENSION 'size '(large medium small))))
;;
;; Specify noisy features. The following indicates that the attributes
;; FLOOR and DEPT are corrupted, respectively, 20% and 10% of the time; 
;; all other attributes are assumed to be uncorrupted.
;;
;; (set-noise '((floor 0.8) (dept 0.9)))
;; 
;; Specify the formula describing the concept in terms of any formulae
;; constructed from AND, OR, NOT, and features of the sample space:
;;
;; (set-concept '(and (or (dept cs))))
;;
;; Specify the distribution governing the positive and negative examples;
;; the following indicates that 60% of the examples are positive.
;;
;; (set-distribution 0.6)
;;
;; Now we can generate examples using CEXAMPLE as follows.
;;
;; > (cexample)
;; (6 ((FLOOR FOUR) (DEPT EE) (STATUS STUDENT) (SIZE MEDIUM)) NO)
;; > (cexample)
;; (7 ((FLOOR FIVE) (DEPT CS) (STATUS FACULTY) (SIZE SMALL)) YES)

;; Specify the sample space and the transmission noise parameters.
(let ((dimensions (list (make-DIMENSION 'floor '(four five))
                        (make-DIMENSION 'dept '(cs ee))
                        (make-DIMENSION 'status '(student faculty staff))
                        (make-DIMENSION 'size '(large medium small))))
      (classes (list 'yes 'no)) 
      (noise '((floor 0.8) (dept 0.9)))
      (attributes '(floor dept status size)))
  (defun set-dimensions (specifications)
    (setq dimensions specifications
          attributes (mapcar #'DIMENSION-attribute specifications)))
  (defun dimensions () dimensions)
  (defun classes () classes)
  (defun attributes () attributes)
  (defun set-noise (characteristics) (setq noise characteristics))
  (defun noise () noise))

;; Generate a training example for a set of classes specified as boolean
;; formulae in conjunctive normal form according to a specified distribution.
;; Each CNF formula is specified as a list (conjunction) of lists 
;; (disjunctions) of features (positive literals). Negative literals can
;; be represented as a disjunction of features (see POSITIVIZE). 
(defun example (formulae classes distribution)
  (aux-example formulae classes distribution (random 1.0)))
(defun aux-example (formulae classes distribution rn)
  (if (< rn (car distribution)) 
    (make-EXAMPLE (new-example-id)
                  (transmit (generate (car formulae) nil))
                  (car classes))
    (aux-example (cdr formulae) (cdr classes) (cdr distribution) rn)))
(let ((identifier 0))
  (defun new-example-id () (setq identifier (+ 1 identifier))))

;; Given a list of disjunctions of features and a set of features, choose, 
;; according to a uniform distribution, from the first disjunction 
;; a disjunct consistent with the set of features, and recurse on the 
;; remaining disjunctions and features augmented by the chosen disjunct.
;; It is possible that AUX-GENERATE will fail to find an example.
;; For instance in the case of (((dept cs) (floor four)) ((dept ee))),
;; if we have already assigned (dept cs) there is no way to satisfy
;; the remaining disjunction ((dept ee)). Of course, (((dept cs) 
;; (floor four)) ((dept ee))) could be simplified to (((floor four)))
;; but we do not perform arbitrary simplifications on formulae. If
;; AUX-GENERATE fails, then GENERATE tries again noting its failure.
;; If the conjunction cannot be satisfied then GENERATE will not return.
(defun generate (conjunction features)
  (cond ((aux-generate conjunction features))
        (t (generate conjunction features))))
(defun aux-generate (conjunction features)
  (if (null conjunction) (complete features)
    (let ((remains (eliminate features (car conjunction))))
      (if (null remains) 
          (and (format t "Failed to find an example!~%") nil)
	(aux-generate (cdr conjunction)
		      (cons (nth (random (length remains)) remains)
			    features))))))

;; Eliminate disjuncts inconsistent with the given features
(defun eliminate (features disjunction)
  (mapcan #'(lambda (f)
              (if (assoc (FEATURE-attribute f) features) nil (list f)))
          disjunction))
                   
(defun complete (features)
  (mapcar #'(lambda (d)
              (or (assoc (DIMENSION-attribute d) features)
                  (list (DIMENSION-attribute d)
                        (nth (random (length (DIMENSION-values d)))
                             (DIMENSION-values d)))))
          (dimensions)))

;; Transmit the features on a noisy channel.
(defun transmit (features)
  (let (pair)
    (mapcar #'(lambda (f)
                (setq pair (assoc (FEATURE-attribute f) (noise)))
                (if (null pair) f
		  (if (< (random 1.0) (cadr pair)) f
		    (alternative f))))
            features)))

(defun alternative (f)
  (let ((alternatives (DIMENSION-assoc (FEATURE-attribute f))))
    (setq alternatives (remove (FEATURE-value f) alternatives))
    (list (FEATURE-attribute f)
          (nth (random (length alternatives))
               alternatives))))

;; Given a formulae describing a concept and the probability of generating
;; a positive example, generate a new example according to that distribution.
(let ((formulae '((((dept cs))) (((dept ee)))))
      (distribution (list 0.6 1.0)))
  (defun set-distribution (probability) 
    (setq distribution (list probability 1.0)))
  (defun set-concept (formula) 
    (setq formulae (list (normalize formula)
                         (normalize (list 'not formula)))))
  (defun cexample ()
    (example formulae (classes) distribution)))

;; Convert Q to a formulae in conjunctive normal form replacing disjuncts
;; of the form (NOT (ATTRIBUTE VALUE)) with one or more disjuncts of the
;; form (ATTRIBUTE ALTERNATIVE-VALUE) where the ALTERNATIVE-VALUEs correspond
;; to all those attribute values of ATTRIBUTE except VALUE. For instance,
;; (NOT (STATUS STAFF)) is replaced with (STATUS FACULTY) and (STATUS STUDENT). 
(defun normalize (q)
  (positivize (remoperators (cnf-nofilter q))))
(defun positivize (conjunction)
  (mapcar #'aux-positivize conjunction))
(defun aux-positivize (disjunction)
  (mapcan #'(lambda (j)
              (if (not (eq (car j) 'not)) (list j)
		(let ((f (cadr j)))
		  (mapcar #'(lambda (v) (list (FEATURE-attribute f) v))
			  (remove (FEATURE-value f)
				  (DIMENSION-assoc (FEATURE-attribute f)))))))
          disjunction))

(defun remoperators (q)
  (cond ((null q) nil)
        ((not (consp q)) q)
        ((member (car q) '(or and)) 
         (remoperators (cdr q)))
        (t (cons (remoperators (car q))
                 (remoperators (cdr q))))))


;; Copyright 1994, Brown University, Providence, RI
;; 
;; Permission to use and modify this software and its documentation for any
;; purpose other than its incorporation into a commercial product is hereby
;; granted without fee.  Permission to copy and distribute this software and its
;; documentation only for non-commercial use is also granted without fee,
;; provided, however, that the above copyright notice appear in all copies, that
;; both that copyright notice and this permission notice appear in supporting
;; documentation, that the name of Brown University not be used in advertising or
;; publicity pertaining to distribution of the software without specific, written
;; prior permission, and that the person doing the distribution notify Brown
;; University of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this software for
;; any purpose.  It is provided "as is" without express or implied warranty.
;; Brown University requests notification of any modifications to this software
;; or its documentation.
;; 
;; Send the following redistribution information:
;; 
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;; 
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;; 
;; 		or
;; 
;; 	brusd@cs.brown.edu
;; 
;; We will acknowledge all electronic notifications.
