;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information
(in-package 'user)

;; Vectors are represented as lists, e.g., (1 2 3). 
;; We assume the same extreme values for all n dimensions. 
;; The space is defined by {*min*, *min* + 1,..., *max* - 1, *max*}^n.
(let ((*min* 0) (*max* 9))
  (defun maximum (x) *max*)
  (defun minimum (x) *min*)
  (defun mmax (l) (if (null l) *min* (apply #'max (cons *min* l))))
  (defun mmin (l) (if (null l) *max* (apply #'min (cons *max* l)))))

;; Given a set of positive examples and a set of negative examples, 
;; return the boundaries for two hyperrectangles one which minimally 
;; encloses all of the positive examples and the other which encloses the
;; first hyperrectangle and minimally excludes all of the negative examples. 
(defun iso (pos neg)
  (let ((pmin (piso pos #'min)) (pmax (piso pos #'max)))
          ;; Upper bounds for the inner hyperrectangle
    (list pmin 
	  ;; Lower bounds for the inner hyperrectangle
	  pmax
	  ;; Compute the upper bounds for the outer hyperrectangle
	  (niso neg pmax #'>= #'<= #'maximum)
	  ;; Compute the lower bounds for the outer hyperrectangle
	  (niso neg pmin #'<= #'>= #'minimum))))

;; Given a list of vectors and an mfun (e.g., max or min), 
;; PISO returns a vector in which the ith component is
;; the mfun of the ith components of the vectors. The invocations
;; (PISO POS #'MAX) and (PISO POS #'MIN), where POS is the list
;; of positive examples, determine the boundaries of the smallest 
;; hyperrectangle enclosing the positive examples.

(defun piso (vecs mfun)
  (let ((rec (car vecs)))
    (mapc #'(lambda (vec) (setq rec (mapcar mfun vec rec)))
	  (cdr vecs))
    rec))

;; NISO computes for each dimension i the min (max) of ith components
;; for negative examples such that those components are >= (<=) the
;; corresponding max (min) of the positive examples.
(defun niso (vecs extrema cfun1 cfun2 extremefun)
  (let ((rec (mapcar extremefun (car vecs))))
    (mapc #'(lambda (vec) 
	      (setq rec (mapcar #'(lambda (v r e)
				    (if (and (funcall cfun1 v e)
					     (funcall cfun2 v r)) v r))
				vec rec extrema)))
	  vecs)
    rec))

;; Example for 2 dimensions:
;;
;; > (setq pos2 '((1 1) (1 3) (3 1) (3 3)))
;; ((1 1) (1 3) (3 1) (3 3))
;; > (setq neg2 '((2 4) (4 2) (0 2) (2 0)))
;; ((2 4) (4 2) (0 2) (2 0))
;; > (iso pos2 neg2)
;; ((1 1) (3 3) (0 0) (4 4))

;; The hyperrectangles computed by ISO do not shrink monotonically,
;; and hence ISO does NOT return a representation that includes the 
;; set of all isothetic hyperrectangles consistent with the examples. 
;; The following function does however. 
(defun riso (pos neg)
  (let ((pmin (piso pos #'min)) (pmax (piso pos #'max)))
    (do ((i (length pmin) (- i 1)) (nmin nil) (nmax nil))
	((= i 0) (list pmin pmax nmin nmax))
	(setq nmin 
	      (cons (mmin (mapcar #'(lambda (vec) 
				     (if (constrainsp vec i #'>= pmin pmax)
					 (nth (- i 1) vec)
				       (maximum 0)))
				     
				 neg))
		    nmin))
	(setq nmax 
	      (cons (mmax (mapcar #'(lambda (vec) 
				     (if (constrainsp vec i #'<= pmin pmax)
					 (nth (- i 1) vec)
				       (minimum 0)))
				 neg))
		    nmax)))))

;; Given a vector corresponding to a negative example, an integer
;; indicating a particular dimension, a comparison function <= or >=, 
;; and two lists representing the boundaries of smallest hyperrectangle
;; enclosing the positive examples, return non-NIL iff in all dimensions
;; except the indicated dimension the vector is within the specified
;; boundaries and in the indicated dimension the vector is either >= or
;; <= the corresponding boundaries depending on the comparison function.
;; Another way of thinking about this is as follows. By ignoring one
;; dimension we are projecting the n-dimensional hyperrectangle into an
;; n-1-dimensional subspace; CONSTRAINSP return non-NIL iff the vector 
;; is contained in the hyperrectangle in this n-1-dimensional subspace 
;; and positioned on the appropriate side of the hyperrectangle in the
;; n-dimensional space.

(defun constrainsp (vec j cfun pmin pmax)
  (let ((i 0))
    (every #'(lambda (x min max)
	       (setq i (+ i 1))
	       (if (= i j) 
		   (and (funcall cfun x min) (funcall cfun x max))
		 (and (>= x min) (<= x max))))
	   vec pmin pmax)))

;; Example illustrating the difference between ISO and RISO:
;;		    
;; > (setq neg2 '((2 4) (4 2) (0 2) (2 0) (0.5 0.5) (3.5 4)))
;; ((2 4) (4 2) (0 2) (2 0) (0.5 0.5) (3.5 4))
;; > (setq pos2 '((1 1) (1 3) (3 1) (3 3)))
;; ((1 1) (1 3) (3 1) (3 3))
;; > (riso pos2 neg2)
;; ((1 1) (3 3) (4 4) (0 0))
;; > (iso pos2 neg2)
;; ((1 1) (3 3) (0.5 0.5) (3.5 4))

;; Copyright 1994, Brown University, Providence, RI
;; 
;; Permission to use and modify this software and its documentation for any
;; purpose other than its incorporation into a commercial product is hereby
;; granted without fee.  Permission to copy and distribute this software and its
;; documentation only for non-commercial use is also granted without fee,
;; provided, however, that the above copyright notice appear in all copies, that
;; both that copyright notice and this permission notice appear in supporting
;; documentation, that the name of Brown University not be used in advertising or
;; publicity pertaining to distribution of the software without specific, written
;; prior permission, and that the person doing the distribution notify Brown
;; University of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this software for
;; any purpose.  It is provided "as is" without express or implied warranty.
;; Brown University requests notification of any modifications to this software
;; or its documentation.
;; 
;; Send the following redistribution information:
;; 
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;; 
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;; 
;; 		or
;; 
;; 	brusd@cs.brown.edu
;; 
;; We will acknowledge all electronic notifications.
