;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information
(in-package 'user)

;; This assignment makes use of the following four files and their 
;; corresponding compiled versions. 
;; 
;; decision.lisp  contains the implementation for learning decision trees
;;                described in the book. 
;; examples.lisp  contains the functions for generating examples for 
;;                concepts corresponding to boolean combinations of features.
;;   normal.lisp  contains the functions for converting arbitrary boolean 
;;                formulae to conjunctive normal form.
;;    noisy.lisp  an implementation for learning decision trees that stops 
;;                splitting nodes when the most discriminating attribute
;;                is less discriminating than is indicated by a threshold
;;                which can be set using (SET-THRESHOLD VALUE).
;; 
;; Choose three concepts for the space of examples specified in
;; examples.lisp.  For at least two of these concepts, (DEPT CS) must be
;; true.  Assume the following noise model, ((DEPT 0.8) (FLOOR 0.9)).  
;; See the comments at the beginning of examples.lisp for how to specify
;; concepts and generate examples.
;; 
;; For each concept, generate two sets of examples: one for training and
;; one for evaluating the performance of the learning algorithm. All of the
;; sets of examples should be the same size, small enough to allow
;; reasonably short experiments but large enough to obtain a representative
;; sample.  Set the probability of generating a positive example to 0.5.
;; 
;; Write a function PERFORMANCE that takes a decision tree and a set of
;; examples and returns the percentage of correctly classified examples.
;; Write a function SIZE that computes the number of nodes in a decision
;; tree and a function DEPTH that computes the maximum depth of a decision
;; tree, where a decision tree with no children has zero depth.
;; 
;; For each of your three concepts, carry out the following steps.  Apply
;; the modified BUILDTREE in noisy.lisp to the training examples for that 
;; concept to obtain a decision tree.   Using the evaluation examples for 
;; that concept and the function PERFORMANCE, determine the percentage of 
;; correctly classified examples.
;; 
;; Use the average performance over the three concepts as a measure of how
;; good your algorithm is and try to maximize this measure by adjusting the
;; threshold parameter.  Provide a table indicating the PERFORMANCE, SIZE,
;; and DEPTH of the resulting decision tree for various settings of the
;; threshold used by BUILDTREE. Indicate what you consider the best setting.
;; If thresholding appears to result in no significant improvement, say so. 
;; 
;; To get a more accurate assessment of performance, you may want to run
;; additional experiments using several sets of examples for each concept 
;; and averaging the performance over all of the the experiments.
;; 
;; In addition to the table, turn in the code for PERFORMANCE, DEPTH, SIZE, 
;; and the functions that you employed to carry out your experiments. 
;;
;; Extra credit: (5 homework points) Implement a version of BUILDTREE
;; that makes use of statistical method described in noisy.tex (see the
;; text starting with ``Thresholding is not particularly effective ...'').
;; Compare this method with the method implemented below; use your 
;; functions PERFORMANCE, DEPTH, and SIZE for evaluation purposes. 

(setq *print-pretty* t)

;; A feature is an attribute together with a value.
(defun FEATURE-attribute (feature) (car feature))
(defun FEATURE-value (feature) (cadr feature))

;; A dimension is an attribute together with a set of possible values.
;; DIMENSION-assoc allows us to easily find values for a given attribute.
(defun make-DIMENSION (attribute values) (list attribute values))
(defun DIMENSION-attribute (dimension) (car dimension))
(defun DIMENSION-values (dimension) (cadr dimension))
(defun DIMENSION-assoc (attribute) 
  (DIMENSION-values (assoc attribute (dimensions))))

;; The abstract data type for decision tree nodes is a modification of the
;; one listed in the book.  In this modified version, each NODE has three
;; additional slots: the first indicates the FEATURE that the node is
;; associated with if such a feature exists (there is no feature associated
;; with the root node) and the second indicates the class of the majority
;; of the examples if the node is terminal and the third is a list of the
;; attributes on the path from the root to the node.
(defun make-NODE (examples) (list nil examples nil nil nil nil))
(defun NODE-attribute (node) (first node))
(defun NODE-examples (node) (second node))
(defun NODE-children (node) (third node))
(defun NODE-feature (node) (fourth node))
(defun NODE-class (node) (fifth node))
(defun NODE-attributes (node) (sixth node))
(defun set-NODE-attribute (node attribute) (setf (first node) attribute))
(defun set-NODE-children (node children) (setf (third node) children))
(defun set-NODE-feature (node feature) (setf (fourth node) feature))
(defun set-NODE-class (node class) (setf (fifth node) class))
(defun set-NODE-attributes (node attributes) (setf (sixth node) attributes))
  
;; An example is an identifier together with a set of features 
;; that describe the example and a class indicator.
(defun make-EXAMPLE (id features class) (list id features class))
(defun EXAMPLE-id (example) (car example))
(defun EXAMPLE-features (example) (cadr example))
(defun EXAMPLE-class (example) (caddr example))

;; Load the routines for generating examples. 
(load "/u/tld/papers/books/intro/learning/lisp/example")

;; This version of BUILDTREE annotates each NODE to keep track of the 
;; (i) the attribute that the NODE was split on, (ii) the list of 
;; attributes on the path from the root to the NODE, (iii) the feature
;; determining that subset of the examples of the NODE's parent that 
;; are stored at the NODE, (iv) and the class name corresponding to 
;; the majority class for those examples stored at a terminal NODE.
;; Further splitting is prevented when all of the attributes have 
;; already been used or when all of the examples are in the same class. 
;; The threshold parameter should be set to a value between 0.0 and 1.0. 
;; (SET-VERBOSE T) allows you to see the score for the attribute that is
;; being considered to split the current node. No splitting occurs if the 
;; score is greater than the THRESHOLD; a high score indicates that even 
;; the best attribute does not do a very good job of discrimination. A 
;; THRESHOLD of 0 never splits; a THRESHOLD of 1 always splits if all 
;; attributes have not been used and all examples are not in the same class. 
(let ((threshold 0.1) (verbose nil))
  (defun set-threshold (value) (setq threshold value))
  (defun set-verbose (value) (setq verbose value))
  (defun buildtree (node)
    (if (>= 1 (length (findall (classify (NODE-examples node))
			       #'(lambda (s) (not (null s))))))
	(set-NODE-class node (majority (NODE-examples node)))
      (let ((dim (splitter node)) examples nodes)
	(if verbose (format t "Score: ~4,2F~%" (evaluate node dim)))
        (if (or (> (evaluate node dim) threshold)
		(member (DIMENSION-attribute dim) (NODE-attributes node))
		(eq (DIMENSION-attribute dim) (NODE-attribute node)))
	    (set-NODE-class node (majority (NODE-examples node)))
          (progn (setq examples (NODE-examples node)
                       nodes (mapcar #'make-NODE
                                     (partition examples dim)))
                 (set-NODE-attribute node (DIMENSION-attribute dim))
                 (set-NODE-children node nodes)
                 (mapc #'(lambda (new value)
                           (set-NODE-attributes 
                            new (cons (DIMENSION-attribute dim)
                                      (NODE-attributes node)))
                           (set-NODE-feature 
                            new (list (DIMENSION-attribute dim) value)))
                       nodes
                       (DIMENSION-values dim))
                 (mapc #'buildtree nodes)))))))

;; Find the class name of the largest class of examples. 
(defun majority (examples)
  (let ((classes (classify examples)))
    (aux-majority (cdr classes) (car classes) (length (car classes)))))
(defun aux-majority (classes class old)
  (if (null classes) 
      (EXAMPLE-class (car class))
    (let ((new (length (car classes))))
      (if (> new old) (aux-majority (cdr classes) (car classes) new)
	(aux-majority (cdr classes) class old)))))

;; The following functions are taken directly from decision.lisp.

(defun classify (examples)
  (mapcar #'(lambda (class)
	      (findall examples #'(lambda (s) 
				    (eq class (EXAMPLE-class s)))))
	  (classes)))

(defun findall (list test)
  (do ((items list (cdr items)) (results nil))
      ((null items) results)
      (if (funcall test (car items)) 
	  (setq results (adjoin (car items) results
				:test #'equal)))))

(defun partition (examples dimension)
  (let ((attribute (DIMENSION-attribute dimension)))
    (mapcar #'(lambda (value) 
		(findall examples 
			 #'(lambda (example) 
			     (member (list attribute value) 
				     (EXAMPLE-features example)
				     :test #'equal))))
	    (DIMENSION-values dimension))))

(defun splitter (node) 
  (do* ((dims (cdr (dimensions)) (cdr dims)) 
	(value (evaluate node (car (dimensions))))
	(splitter (car (dimensions))) (min value))
       ((null dims) splitter)
       (setq value (evaluate node (car dims)))
       (if (< value min)
	   (setq splitter (car dims) min value))))

(defun evaluate (node dimension)
  (let ((total (length (NODE-examples node))))
    (if (= total 0) 0
      (apply #'+
	     (mapcar #'(lambda (subset)
			 (* (/ (length subset) total)
			    (disorder (classify subset))))
		     (partition (NODE-examples node) dimension))))))

(defun disorder (partition)
  (let* ((sizes (mapcar #'length partition))
	 (total (apply #'+ sizes)))
    (if (= total 0) 0
      (- (log total 2)
	 (apply #'+ (mapcar #'(lambda (s) 
				(if (= s 0) 0 
				  (* (/ s total) (log s 2))))
			    sizes))))))






;; Copyright 1994, Brown University, Providence, RI
;; 
;; Permission to use and modify this software and its documentation for any
;; purpose other than its incorporation into a commercial product is hereby
;; granted without fee.  Permission to copy and distribute this software and its
;; documentation only for non-commercial use is also granted without fee,
;; provided, however, that the above copyright notice appear in all copies, that
;; both that copyright notice and this permission notice appear in supporting
;; documentation, that the name of Brown University not be used in advertising or
;; publicity pertaining to distribution of the software without specific, written
;; prior permission, and that the person doing the distribution notify Brown
;; University of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this software for
;; any purpose.  It is provided "as is" without express or implied warranty.
;; Brown University requests notification of any modifications to this software
;; or its documentation.
;; 
;; Send the following redistribution information:
;; 
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;; 
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;; 
;; 		or
;; 
;; 	brusd@cs.brown.edu
;; 
;; We will acknowledge all electronic notifications.
