;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;; This file contains all of the code implementing the decision
;; tree method described in the chapter on learning.

(setq *print-pretty* t)

;; A feature is an attribute together with a value.
(defun FEATURE-attribute (feature) (car feature))
(defun FEATURE-value (feature) (cadr feature))

;; A dimension is an attribute together with a set of possible values.
(defun make-DIMENSION (attribute values) (list attribute values))
(defun DIMENSION-attribute (dimension) (car dimension))
(defun DIMENSION-values (dimension) (cadr dimension))

;; A decision tree node consists of examples and child nodes.
(defun make-NODE (examples) (list nil examples (list)))
(defun NODE-attribute (node) (car node))
(defun set-NODE-attribute (node attribute) 
  (setf (car node) attribute))
(defun NODE-examples (node) (cadr node))
(defun NODE-children (node) (caddr node))
(defun set-NODE-children (node children) 
  (setf (caddr node) children))

;; An example is an identifier together with a set of features 
;; that describe the example and a class indicator.
(defun make-EXAMPLE (id features class) (list id features class))
(defun EXAMPLE-id (example) (car example))
(defun EXAMPLE-features (example) (cadr example))
(defun EXAMPLE-class (example) (caddr example))

;; Buildtree implements the basic recursive function for constructing
;; decision trees.
(defun buildtree (node)
  (format t "Class distribution: ~A~%"
          (mapcar #'length (classify (NODE-examples node))))
  (if (< 1 (length (findall (classify (NODE-examples node)) 
			    #'(lambda (s) (not (null s)))))) 
      (let ((dim (splitter node)))
	(format t "Dimension chosen: ~A~%" dim)
	(set-NODE-attribute node (DIMENSION-attribute node))
	(set-NODE-children node
			   (mapcar #'make-NODE 
				   (partition (NODE-examples node) dim)))
	(mapc #'buildtree (NODE-children node)))
    (format t "Terminal node.~%")))

;; Given a set of examples, classify partitions the examples 
;; into classes.
(defun classify (examples)
  (mapcar #'(lambda (class)
	      (findall examples #'(lambda (s) 
				    (eq class (EXAMPLE-class s)))))
	  (classes)))

(defun findall (items test)
  (do ((items items (cdr items)) (results (list)))
      ((null items) results)
      (if (funcall test (car items)) 
	  (setq results (adjoin (car items) results
				:test #'equal)))))

;; Partition takes a set of examples and partitions by the attribute 
;; of a dimension according to the values for that attribute.
(defun partition (examples dimension)
  (let ((attribute (DIMENSION-attribute dimension)))
    (mapcar #'(lambda (value) 
		(findall examples 
			 #'(lambda (example) 
			     (member (list attribute value) 
				     (EXAMPLE-features example)
				     :test #'equal))))
	    (DIMENSION-values dimension))))

;; Splitter finds the attribute that minimizes a particular evaluation
;; function implemented by evaluate.
(defun splitter (node) 
  (do* ((dims (cdr (dimensions)) (cdr dims)) 
	(value (evaluate node (car (dimensions))))
	(splitter (car (dimensions))) (min value))
       ((null dims) splitter)
       (setq value (evaluate node (car dims)))
       (if (< value min)
	   (setq splitter (car dims) min value))))

;; The function evaluate computes the weighted sum of the disorder over 
;; all class partitions for a given attribute partition.
(defun evaluate (node dimension)
  (let ((total (length (NODE-examples node))))
    (apply #'+
	   (mapcar #'(lambda (subset)
		       (* (/ (length subset) total)
			  (disorder (classify subset))))
		   (partition (NODE-examples node) dimension)))))

;; Disorder computes the information value (disorder) associated with a
;; particular class partition according to the formula listed in the 
;; section on decision trees. 
(defun disorder (partition)
  (let* ((sizes (mapcar #'length partition))
	 (total (apply #'+ sizes)))
    (if (= total 0) 0
      (- (log total 2)
	 (apply #'+ (mapcar #'(lambda (s) 
				(if (= s 0) 0 
				  (* (/ s total) (log s 2))))
			    sizes))))))

;; Examples

(defun sunburn ()
  (let ((examples
	 (list
	  (make-EXAMPLE 'Sarah
			'((hair blonde) (height average) 
			  (weight light) (lotion no)) 'yes)
	  (make-EXAMPLE 'Dana
			'((hair blonde) (height tall) 
			  (weight average) (lotion yes)) 'no)
	  (make-EXAMPLE 'Alex
			'((hair brown) (height short) 
			  (weight average) (lotion yes)) 'no)
	  (make-EXAMPLE 'Annie
			'((hair blonde) (height short) 
			  (weight average) (lotion no)) 'yes)
	  (make-EXAMPLE 'Emily
			'((hair red) (height average) 
		       (weight heavy) (lotion no)) 'yes)
	  (make-EXAMPLE 'Pete
			'((hair brown) (height tall) 
			  (weight heavy) (lotion no)) 'no)
	  (make-EXAMPLE 'John
			'((hair brown) (height average) 
			  (weight heavy) (lotion no)) 'no)
	  (make-EXAMPLE 'Katie
			'((hair blonde) (height short) 
			  (weight light) (lotion yes)) 'no))))
    (defun dimensions ()
      (list (make-DIMENSION 'hair '(blonde brown red))
	    (make-DIMENSION 'height '(short tall average))
	    (make-DIMENSION 'weight '(light heavy average))
	    (make-DIMENSION 'lotion '(yes no))))
    (defun classes () '(yes no))
    (buildtree (make-NODE examples))))

(defun robots ()
  (let ((examples
	 (list 
	  (make-EXAMPLE '307 '((status faculty) (floor three) 
			       (dept ee) (size large)) 'no)
	  (make-EXAMPLE '309 '((status staff) (floor three) 
			       (dept ee) (size small)) 'no)
	  (make-EXAMPLE '408 '((status faculty) (floor four) 
			       (dept cs) (size medium)) 'yes)
	  (make-EXAMPLE '415 '((status student) (floor four) 
			       (dept ee) (size large)) 'yes)
	  (make-EXAMPLE '509 '((status staff) (floor five) 
			       (dept cs) (size medium)) 'no)
	  (make-EXAMPLE '517 '((status faculty) (floor five) (dept cs) 
			       (size large)) 'yes)
	  (make-EXAMPLE '316 '((status student) (floor three) 
			       (dept ee) (size small)) 'yes)
	  (make-EXAMPLE '420 '((status staff) (floor four) (dept cs) 
			       (size medium)) 'no ))))
    (defun dimensions ()
      (list (make-DIMENSION 'status '(faculty staff student))
	    (make-DIMENSION 'floor '(three four five))
	    (make-DIMENSION 'dept '(cs ee))
	    (make-DIMENSION 'size '(large medium small))))
    (defun classes () '(yes no))
    (buildtree (make-NODE examples))))

(defun final ()
  (let ((examples (list (make-EXAMPLE 1 '((a 1) (b 0) (c 1)) 1)
			(make-EXAMPLE 2 '((a 1) (b 1) (c 0)) 1)
			(make-EXAMPLE 3 '((a 1) (b 1) (c 1)) 0)
			(make-EXAMPLE 4 '((a 0) (b 0) (c 0)) 0)
			(make-EXAMPLE 5 '((a 1) (b 1) (c 0)) 1)
			(make-EXAMPLE 6 '((a 1) (b 0) (c 0)) 0))))
    (defun dimensions ()
      (list (make-DIMENSION 'a '(0 1))
	    (make-DIMENSION 'b '(0 1))
	    (make-DIMENSION 'c '(0 1))))
    (defun classes () '(0 1))
    (buildtree (make-NODE examples))))

;; Test function.
(defun test ()
  (sunburn)
  (robots)
  (final))

;; Compute some summary statistics on a decision tree.
(defun split-statistics (node)
  (mapcar #'(lambda (d)
              (format t "~A ~4,3F ~A~%"
                      (DIMENSION-attribute d)
                      (evaluate node d)
                      (mapcar #'(lambda (x) 
                                  (mapcar #'(lambda (y) 
                                              (mapcar #'EXAMPLE-id y)) 
                                          (classify x)))
                              (partition (NODE-examples node) d))))
          (dimensions)))


;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

