;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;; This file contains all of the code implementing the decision
;; tree method described in the chapter on learning.

(set! *print-pretty* #t)


;; A feature is an attribute together with a value.
(define (feature-attribute feature) (car feature))

(define (feature-value feature) (cadr feature))


;; A dimension is an attribute together with a set of possible values.
(define (make-dimension attribute values) (list attribute values))

(define (dimension-attribute dimension) (car dimension))

(define (dimension-values dimension) (cadr dimension))


;; A decision tree node consists of examples and child nodes.
(define (make-node examples) (list #f examples (list)))

(define (node-attribute node) (car node))

(define (set-node-attribute node attribute) (set-car! node attribute))

(define (node-examples node) (cadr node))

(define (node-children node) (caddr node))

(define (set-node-children node children) (set-car! (cddr node) children))


;; An example is an identifier together with a set of features 
;; that describe the example and a class indicator.
(define (make-example id features class) (list id features class))

(define (example-id example) (car example))

(define (example-features example) (cadr example))

(define (example-class example) (caddr example))


;; Buildtree implements the basic recursive function for constructing
;; decision trees.
(define (buildtree node)
  (begin
    (display "Class distribution: ")
    (display (map length (classify (node-examples node))))
    (newline))
  (if (< 1
         (length (findall (classify (node-examples node))
                          (lambda (s) (not (null? s))))))
    (let ((dim (splitter node)))
      (begin (display "Dimension chosen: ") (display dim) (newline))
      (set-node-attribute node (dimension-attribute node))
      (set-node-children
        node
        (map make-node (partition (node-examples node) dim)))
      (for-each buildtree (node-children node)))
    (begin (display "Terminal node.") (newline))))


;; Given a set of examples, classify partitions the examples 
;; into classes.
(define (classify examples)
  (map (lambda (class)
         (findall examples (lambda (s) (eq? class (example-class s)))))
       (classes)))


(define (findall items test)
  (do ((items items (cdr items)) (results (list)))
      ((null? items) results)
    (if (test (car items))
      (set! results (adjoin2 (car items) results equal?)))))


;; Partition takes a set of examples and partitions by the attribute 
;; of a dimension according to the values for that attribute.
(define (partition examples dimension)
  (let ((attribute (dimension-attribute dimension)))
    (map (lambda (value)
           (findall examples
                    (lambda (example)
                      (member2 (list attribute value)
                               (example-features example)
                               equal?
                               (lambda (x) x)))))
         (dimension-values dimension))))


;; Splitter finds the attribute that minimizes a particular evaluation
;; function implemented by evaluate.
(define (splitter node)
  (let ((dims (dimensions)))
    (aux-splitter node (cdr dims) (car dims) 
                  (evaluate node (car dims)))))
(define (aux-splitter node dims dim old)
  (if (null? dims) dim
      (let ((new (evaluate node (car dims))))
        (if (< new old) (aux-splitter node (cdr dims) (car dims) new)
            (aux-splitter node (cdr dims) dim old)))))

;; The function evaluate computes the weighted sum of the disorder over 
;; all class partitions for a given attribute partition.
(define (evaluate node dimension)
  (let ((total (length (node-examples node))))
    (apply +
           (map (lambda (subset)
                  (* (/ (length subset) total) (disorder (classify subset))))
                (partition (node-examples node) dimension)))))


;; Disorder computes the information value (disorder) associated with a
;; particular class partition according to the formula listed in the 
;; section on decision trees. 
(define (disorder partition)
  (let* ((sizes (map length partition)) (total (apply + sizes)))
    (if (= total 0)
      0
      (- (/ (log total) (log 2))
         (apply +
                (map (lambda (s) (if (= s 0) 0 (* (/ s total) 
                                                  (/ (log s) (log 2)))))
                     sizes))))))


;; Examples

(define (sunburn)
  (let ((examples
          (list (make-example
                  'sarah
                  '((hair blonde) (height average) (weight light) (lotion no))
                  'yes)
                (make-example
                  'dana
                  '((hair blonde) (height tall) (weight average) (lotion yes))
                  'no)
                (make-example
                  'alex
                  '((hair brown) (height short) (weight average) (lotion yes))
                  'no)
                (make-example
                  'annie
                  '((hair blonde) (height short) (weight average) (lotion no))
                  'yes)
                (make-example
                  'emily
                  '((hair red) (height average) (weight heavy) (lotion no))
                  'yes)
                (make-example
                  'pete
                  '((hair brown) (height tall) (weight heavy) (lotion no))
                  'no)
                (make-example
                  'john
                  '((hair brown) (height average) (weight heavy) (lotion no))
                  'no)
                (make-example
                  'katie
                  '((hair blonde) (height short) (weight light) (lotion yes))
                  'no))))
    (set! dimensions
      (lambda ()
        (list (make-dimension 'hair '(blonde brown red))
              (make-dimension 'height '(short tall average))
              (make-dimension 'weight '(light heavy average))
              (make-dimension 'lotion '(yes no)))))
    (set! classes (lambda () '(yes no)))
    (buildtree (make-node examples))))


(define (robots)
  (let ((examples
          (list (make-example
                  '307
                  '((status faculty) (floor three) (dept ee) (size large))
                  'no)
                (make-example
                  '309
                  '((status staff) (floor three) (dept ee) (size small))
                  'no)
                (make-example
                  '408
                  '((status faculty) (floor four) (dept cs) (size medium))
                  'yes)
                (make-example
                  '415
                  '((status student) (floor four) (dept ee) (size large))
                  'yes)
                (make-example
                  '509
                  '((status staff) (floor five) (dept cs) (size medium))
                  'no)
                (make-example
                  '517
                  '((status faculty) (floor five) (dept cs) (size large))
                  'yes)
                (make-example
                  '316
                  '((status student) (floor three) (dept ee) (size small))
                  'yes)
                (make-example
                  '420
                  '((status staff) (floor four) (dept cs) (size medium))
                  'no))))
    (set! dimensions
      (lambda ()
        (list (make-dimension 'status '(faculty staff student))
              (make-dimension 'floor '(three four five))
              (make-dimension 'dept '(cs ee))
              (make-dimension 'size '(large medium small)))))
    (set! classes (lambda () '(yes no)))
    (buildtree (make-node examples))))


(define (final)
  (let ((examples
          (list (make-example 1 '((a 1) (b 0) (c 1)) 1)
                (make-example 2 '((a 1) (b 1) (c 0)) 1)
                (make-example 3 '((a 1) (b 1) (c 1)) 0)
                (make-example 4 '((a 0) (b 0) (c 0)) 0)
                (make-example 5 '((a 1) (b 1) (c 0)) 1)
                (make-example 6 '((a 1) (b 0) (c 0)) 0))))
    (set! dimensions
      (lambda ()
        (list (make-dimension 'a '(0 1))
              (make-dimension 'b '(0 1))
              (make-dimension 'c '(0 1)))))
    (set! classes (lambda () '(0 1)))
    (buildtree (make-node examples))))


;; Test function.
(define (test) (sunburn) (robots) (final))


;; Compute some summary statistics on a decision tree.
(define (split-statistics node)
  (map (lambda (d)
         (begin
           (display (dimension-attribute d))
           (display " ")
           (display (evaluate node d))
           (display " ")
           (display (map (lambda (x)
                           (map (lambda (y) (map example-id y)) (classify x)))
                         (partition (node-examples node) d)))
           (newline)))
       (dimensions)))



;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;;      Name:
;;      Organization:
;;      Address (postal and/or electronic):
;;
;; To:
;;      Software Librarian
;;      Computer Science Department, Box 1910
;;      Brown University
;;      Providence, RI 02912
;;
;;              or
;;
;;      brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

