;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;; In the following, we assume a four-state dynamical system 
;; with states notated 0, 1, 2, and 3. 
;; Given a state, return the next state for a particular 
;; deterministic transition model represented as a list of 
;; pairs of the form (state next-state).
(defun deterministic (state)
  (cadr (assoc state '((0 1) (1 2) (2 3) (3 0)))))

;; Given a state, return the next state for a particular 
;; stochastic transition model represented as a list of lists 
;; of the form (state probability-of-most-likely-next-state 
;; most-likely-next-state next-most-likely-next-state) where we 
;; assume any given state has at most two possible next states. 
(defun stochastic (state)
  (let ((dist (cdr (assoc state '((0 0.8 1 0) (1 0.7 2 3) 
				  (2 0.7 3 0) (3 0.8 0 3))))))
    (if (< (random 1.0) (car dist)) (cadr dist) (caddr dist))))

;; Given a state return the reward obtained when in that state
;; for the particular reward function used in the examples.
(defun reward (state)
  (cadr (assoc state '((0 0.0) (1 0.0) (2 0.0) (3 10.0)))))

;; Update the reward function by backing up values one step
;; at a time. This is the analogous to the step-wise
;; refinement typical in dynamic programming; first compute
;; the value function for one-step lookahead and then use this 
;; to compute the value function for a two step lookahead. 
(defun update-deterministic (state next values discount)
  (setf (nth state values)
	(+ (reward next) (* discount (nth next values)))))

;; (= (+ (reward next) (* discount (nth next values)))
;;    (+ (nth state values)
;;       (- (+ (reward next) 
;; 	       (* discount (nth next values)))
;; 	    (nth state values)))

;; The stochastic case requires that we introduce a fixed 
;; rate to govern how fast values are adjusted.  This method
;; does not converge to the true values in the case of 
;; stochastic transitions, but varies about the expected 
;; value. A rate close to one results in greater variance
;; but quicker convergence to the region of variance.  A
;; smaller rate results in a more stable measurement of the 
;; expected value but a longer time to achieve this. 
(defun update-stochastic (state next values discount rate)
  (setf (nth state values)
	(+ (nth state values)
	   (* rate (- (+ (reward next) 
			 (* discount (nth next values)))
		      (nth state values))))))
	
;; Experiments

(defun test-deterministic ()
  (do ((n 1 (+ n 1)) (values '(0.0 0.0 0.0 0.0)) 
       (state 0 next) (next 1 (deterministic next)))
      ((= n 1000) values)
      (update-deterministic state next values 0.8)))

(defun test-stochastic-with-deterministic ()
  (do ((n 1 (+ n 1)) (values '(0.0 0.0 0.0 0.0)) 
       (state 0 next) (next 1 (deterministic next)))
      ((= n 1000) values)
      (update-stochastic state next values 0.8 0.7)))

(defun test-stochastic ()
  (do ((n 1 (+ n 1)) (values '(0.0 0.0 0.0 0.0)) 
       (state 0 next) (next 1 (stochastic next)))
      ((= n 1000) values)
      (update-stochastic state next values 0.8 0.1)))

;; > (test-deterministic)
;; (10.840108401084013 13.550135501355016
;;                     16.93766937669377
;;                     8.672086720867212)
;; > (test-stochastic-with-deterministic)
;; (10.840108401084013 13.550135501355016
;;                     16.93766937669377
;;                     8.672086720867212)
;; > (test-stochastic)
;; (11.701883966440893 17.945842125733712
;;                     13.72701018155669
;;                     11.369491311075088)

;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

