;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

(setq or  '(((0 0) 0) ((1 0) 1) ((0 1) 1) ((1 1) 1))
      and '(((0 0) 0) ((1 0) 0) ((0 1) 0) ((1 1) 1))
      xor '(((0 0) 0) ((1 0) 1) ((0 1) 1) ((1 1) 0)))

;; Sum the contributions to the gradient for each example at a point
;; in parameter space specified by the weights.
(defun gradient (examples weights)
  (reduce #'(lambda (v w) (mapcar #'+ v w))
          (mapcar #'(lambda (pair) 
                      (delta (car pair) (cadr pair) weights))
                  examples)))

(defun trial (target)
  (aux-trial target (init) 0))
(let ((rate 2) (freq 10) (stat nil) (max 1000))
  (defun set-rate (r) (setq rate r))
  (defun set-freq (f) (setq freq f))
  (defun set-stat (s) (setq stat s))
  (defun aux-trial (target weights n)
    (cond ((successp target weights) 
	   (if stat n
	     (format t "~%Weights: ~6,3F ~6,3F ~6,3F ~6,3F ~6,3F ~6,3F ~6,3F"
		     (nth 0 weights) (nth 1 weights) (nth 2 weights)
		     (nth 3 weights) (nth 4 weights) (nth 5 weights)
		     (nth 6 weights))))
	  ((= n max) nil)
	  (t (if (and (= 0 (mod n freq)) (not stat))
		 (format t "~%Adjustments: ~D~
		            ~%Error: ~6,3F" n (mse target weights)))
	     (aux-trial target 
			(mapcar #'+ weights 
				(mapcar #'(lambda (w) (* rate w))
					(gradient target weights)))
			(+ n 1))))))

;; Attempt to train the network several times, and average the number
;; of adjustments it took to train each successful time.
(defun average (target attempts)
  (set-stat t) 
  (let ((total 0) (successes 0) adjustments)
    (dotimes (i attempts) 
	     (setq adjustments (trial target))
	     (if (numberp adjustments)
		 (progn (setq total (+ total adjustments))
			(setq successes (+ 1 successes)))))
    (format t "The average is ~4,1F for ~D out of ~D attempts."
	    (if (> successes 0) (/ (* total 1.0) successes) 0)
	    successes attempts)))
	     
(defun train (target weights)
  (if (successp target weights) (princ weights)
    (train target (mapcar #'+ weights (gradient target weights)))))

;; For a given input, the sigmoid function corresponding to the output of
;; the network provides a value between zero and one. A given set of
;; weights is considered to have successfully implemented a target
;; function if the difference between the actual and the desired outputs
;; is less than 0.4.  All that is absolutely necessary is that the
;; difference is strictly less than 0.5, so that the rounded output is
;; equal to the desired output.  The function ABS returns the absolute 
;; value of its single argument.
(defun successp (target weights)
  (every #'(lambda (x)
             (< (abs (- (cadr x) 
                        (net (car x) weights))) 0.4))
         target))

(defun mse (target weights)
  (apply #'+
         (mapcar #'(lambda (x)
                     (square (- (cadr x) 
                                (net (car x) weights))))
                 target)))

(defun square (x) (* x x))


;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

