;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;; Here is a straight-through logic threshold unit with a fixed
;; threshold of one for any length of input and weight vectors.
(defun ltu (input weights)
  (if (< 1 (apply #'+ (mapcar #'* input weights))) 1 0))

;; Here is an implementation of the perceptron learning rule.
(defun adjust (input output weights)
  (cond ((= (ltu input weights) output) weights)
	((= (ltu input weights) 0) (mapcar #'+ weights input))
	(t (mapcar #'- weights input))))

;; A perceptron has successfully learned a given target function 
;; when the perceptron output equals the target output on all inputs.
(defun successp (target weights)
  (every #'(lambda (x) (= (ltu (car x) weights) (cadr x))) target))

;; Training consists of choosing a random input/output pair from the
;; target and adjusting the weights according to the perceptron 
;; learning rule until the target is successfully learned. 
(defun train (target weights) 
  (if (successp target weights) nil
    (let ((pair (nth (random (length target)) target)))
      (train target (adjust (car pair) (cadr pair) weights)))))

;; Here is a test version with running commentary and summary statistics.
(defun test (target weights)
  (verbose-train target weights 0))
(defun verbose-train (target weights adjustments)
  (if (successp target weights) 
      (progn (princ "Number of adjustments: ") 
	     (princ adjustments) (terpri) nil)
    (let ((pair (nth (random (length target)) target)))
      (princ "Training example: ") (princ pair) (terpri)
      (verbose-train target 
		     (verbose-adjust (car pair) (cadr pair) weights)
		     (+ 1 adjustments)))))
(defun verbose-adjust (input output weights)
  (cond ((= (ltu input weights) output))
	((= (ltu input weights) 0) 
	 (setq weights (mapcar #'+ weights input)))
	(t (setq weights (mapcar #'- weights input))))
  (princ "Adjusted weights: ") (princ weights) (terpri) weights)

;; Examples

;; disjunction
;; faculty or cs

(setq or '(((0 0) 0)
	   ((1 0) 1)
	   ((0 1) 1)
	   ((1 1) 1)))

;; conjunction
;; faculty and cs
(setq and '(((0 0) 0)
	    ((1 0) 0)
	    ((0 1) 0)
	    ((1 1) 1)))

;; exclusive or 
;; (faculty or cs) and not (faculty and cs))

(setq xor '(((0 0) 0)
	    ((1 0) 1)
	    ((0 1) 1)
	    ((1 1) 0)))

;; equivalence
;; (faculty and cs) or (not faculty} and not cs)

(setq equal '(((0 0) 1)
	      ((1 0) 0)
	      ((0 1) 0)
	      ((1 1) 1)))

;; This function demonstrates the need for adjustable thresholds.

(setq trouble '(((0 0) 1) 
		((1 0) 0) 
		((0 1) 1) 
		((1 1) 0)))

;; two or more out of three
(setq two '(((0 0 0) 0)
	    ((1 0 0) 0)
	    ((0 1 0) 0)
	    ((1 1 0) 1)
	    ((0 0 1) 0)
	    ((1 0 1) 1)
	    ((0 1 1) 1)
	    ((1 1 1) 1)))

;; Experiments
	  
;; Ran TRAIN on OR twenty times.
;; Here are the statistics for the number of adjustments required.
;; > (setq trials '(14 8 8 5 5 18 11 15 5 5 5 7 3 6 5 10 6 11 14 6.0))
;; (14 8 8 5 5 18 11 15 5 5 5 7 3 6 5 10 6 11 14 6.0)
;; > (setq mean (/ (apply #'+ trials) (length trials)))
;; 8.35
;; > (apply #'min trials)
;; 3
;; > (apply #'max trials)
;; 18

;; > (test or '(0 0))
;; Training example: ((1 1) 1)
;; Adjusted weights: (1 1)
;; Training example: ((0 1) 1)
;; Adjusted weights: (1 2)
;; Training example: ((0 1) 1)
;; Adjusted weights: (1 2)
;; Training example: ((1 0) 1)
;; Adjusted weights: (2 2)
;; Number of adjustments: 4


;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

