;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;; Gaussian function used as a radial basis function.
(defun gaussian (x c w)
  (exp (- (* (/ 1 (* 2 w)) (square (- x c))))))

;; Generate n training examples in the range -1 to 1. 
(defun examples (n)
  (do ((i 1 (+ i 1)) (examples (list)) x)
      ((= i n) examples)
      (setq x (- (random 2.0) 1))
      (setq examples (cons (list x (target x))
			   examples))))

;; Target function for the learning example:
(defun target (x)
  (+ 1.20 (* x -0.20) (* x x 3.10) (* x x x -0.90)))

(defun raise (x n)
  (if (= n 0) 1 (* x (raise x (- n 1)))))

(defun square (x) (* x x))

(defun mse (pairs weights)
  (apply #'+
	 (mapcar #'(lambda (x)
		     (square (- (cadr x) 
				(net (car x) weights))))
		 pairs)))

;; Here is a test version with running commentary and summary statistics.
(let ((max 100) (examples (examples 40)) (online nil) (verbose nil)
      (ctr1 -1.00) (ctr2 -0.50)  (ctr3 0.0) (ctr4 0.5) (width 1.0) 
      (rate 0.1))
  (defun wc (a b c d w)
    (setq ctr1 a  ctr2 b  ctr3 c  ctr4 d  width w))
  (defun set-verbose (v) (setq verbose v))
  (defun set-online (v) (setq online v))
  (defun sample () 
    (if online (list (nth (random (length examples)) examples))
      examples))
  (defun patience (adjustments) (setq max adjustments))
  (defun set-rate (x) (if (numberp x) (setq rate x) rate))
  (defun initw () '(0.0 0.0 0.0 0.0))
  ;; Compute the output of the network for a given input and weights.
  (defun net (input weights)
    (+ (* (nth 0 weights) (gaussian input ctr1 width))
       (* (nth 1 weights) (gaussian input ctr2 width))
       (* (nth 2 weights) (gaussian input ctr3 width))
       (* (nth 3 weights) (gaussian input ctr4 width))))
  ;; The gradient adjustment method 
  (defun adjust (examples weights)
    (let ((d (/ 1 (length examples))))
    (list 
     (+ (nth 0 weights)	
	(* rate 2 d (apply #'+ (mapcar #'(lambda (pair)
					   (* (- (cadr pair) 
						 (net (car pair) weights))
					     (gaussian (car pair) ctr1 width)))
				       examples))))
     (+ (nth 1 weights)	
	(* rate 2 d (apply #'+ (mapcar #'(lambda (pair)
					   (* (- (cadr pair) 
						 (net (car pair) weights))
					     (gaussian (car pair) ctr2 width)))
				       examples))))
     (+ (nth 2 weights)
	(* rate 2 d (apply #'+ (mapcar #'(lambda (pair)
					   (* (- (cadr pair) 
						 (net (car pair) weights))
					     (gaussian (car pair) ctr3 width)))
				       examples))))
     (+ (nth 3 weights)
	(* rate 2 d (apply #'+ (mapcar #'(lambda (pair)
					   (* (- (cadr pair) 
						 (net (car pair) weights))
					     (gaussian (car pair) ctr4 width)))
				       examples)))))))
  (defun test () (train (initw) 0))
  (defun train (weights adjustments)
    (if (= adjustments max) 
	(format t "~%Plot[~6,3F Gaussian[x,~6,2F,~6,2F] + (* RATE ~F *)~
	           ~%     ~6,3F Gaussian[x,~6,2F,~6,2F] + (* MSE ~6,3F *)~
	           ~%     ~6,3F Gaussian[x,~6,2F,~6,2F] + (* STEPS ~D *)~
	           ~%     ~6,3F Gaussian[x,~6,2F,~6,2F],{x,-1,1}]~%"
		(nth 0 weights) ctr1 width rate
		(nth 1 weights) ctr2 width (mse examples weights)
		(nth 2 weights) ctr3 width max
		(nth 3 weights) ctr4 width)
      (progn (if verbose (format t "~6,3F, " (mse examples weights)))
	     (train (adjust (sample) weights)
		    (+ 1 adjustments))))))

;; Experiments for learning chapter:

(defun data ()
  (patience 500)
  (terpri) (wc -0.75 -0.25 0.25  0.75 1.0) (test)
  (terpri) (wc -0.75 -0.25 0.25  0.75 0.5) (test)
  (terpri) (wc -0.75 -0.375 0.0  0.50 1.0) (test)
  (terpri) (wc -0.75 -0.375 0.0  0.50 0.5) (test)
  (terpri) (wc -1.00 -0.500 0.0  1.00 1.0) (test)
  (terpri) (wc -1.00 -0.500 0.0  1.00 0.5) (test))


;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

