;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;; This file includes all of the code for implementing the 
;; version spaces method described in the learning chapter.
;; The code also includes a solution to the exercise that 
;; prunes redundant concepts from the boundaries.

;; Abstract data types

;; A FEATURE is an attribute together with a value.
(define (feature-attribute feature) (car feature))

(define (feature-value feature) (cadr feature))


;; A DIMENSION is an attribute together with a set of possible values.
(define (make-dimension attribute values) (list attribute values))

(define (dimension-attribute dimension) (car dimension))

(define (dimension-values dimension) (cadr dimension))


;; An EXAMPLE is an identifier together with a set of features 
;; that describe the example and a class indicator.
(define (make-example id features class) (list id features class))

(define (example-id example) (car example))

(define (example-features example) (cadr example))

(define (example-class example) (caddr example))


;; The function FINDALL takes a list and a function and returns 
;; all items in the list such that the function returns non NIL 
;; when applied to the item.
(define (findall items test)
  (do ((items items (cdr items)) (results (list)))
      ((null? items) results)
    (if (test (car items))
      (set! results (adjoin2 (car items) results equal?)))))


;; REFINE the general and specific version-space boundaries.  In the
;; following, we provide code for modifying the boundaries in a version
;; space given a new training example. The code provided below is not
;; guaranteed to converge to a unique concept. In the exercises, you are
;; shown how to modify the code to implement the complete algorithm given
;; in the text. Refine takes a single training example and two lists of
;; concepts representing the general and specific boundaries. A
;; conjunctive concept is represented as a list of features. Refine
;; determines whether the training example is positive, yes, or not, no,
;; and then proceeds to, respectively, generalize the concepts in the
;; specific bounary or specialize the concepts in the general boundary.
(define (refine example general specific)
  (if (eq? (example-class example) 'yes)
    (simplify-specific
      (prune-general example general)
      (generalize-specific example specific))
    (simplify-general
      (specialize-general example general)
      (prune-specific example specific))))


;; Here is a simplified version of REFINE that produces boundaries
;; that include redundant and useless concepts.  This version does 
;; not guarantee convergence.  An example is within the boundaries
;; just in case it is consistent with one concept in the general 
;; boundary and one concept in the specific boundary such that 
;; the specific concept specializes the general concept.
;; (defun refine (example general specific) 
;;   (if (eq (EXAMPLE-class example) 'yes)
;;       (list general (generalize-specific example specific))
;;     (list (specialize-general example general) specific)))

;; Include in the general boundary only those concepts that 
;; are consistent with the positive examples.
(define (prune-general example general)
  (findall general
           (lambda (concept) (consistent (example-features example) concept))))


;; Include in the specific boundary only those concepts that are 
;; not consistent with the negative examples.
(define (prune-specific example specific)
  (findall specific
           (lambda (concept)
             (not (consistent (example-features example) concept)))))


;; Positive examples serve to generalize the specific boundary.
;; Generalizing the specific boundary consists of generalizing each 
;; concept in the boundary until it is consistent with the (positive) 
;; example.
(define (generalize-specific example boundary)
  (apply append
         (map (lambda (c) (aux-generalize-specific example c)) boundary)))

(define (aux-generalize-specific example concept)
  (if (consistent (example-features example) concept)
    (list concept)
    (generalize-specific example (generalize concept))))


;; Negative examples serve to specialize the general boundary.
;; Specializing the general boundary consists of specializing 
;; each concept in the boundary until it is not consistent 
;; with the (negative) example.
(define (specialize-general example boundary)
  (apply append
         (map (lambda (c) (aux-specialize-general example c)) boundary)))

(define (aux-specialize-general example concept)
  (if (not (consistent (example-features example) concept))
    (list concept)
    (specialize-general example (specialize concept))))


;; Determine whether a description is consistent with a given concept.  
;; A concept is consistent with an example if for each feature (conjunct)
;; in the concept, either the attribute of the feature is not mentioned
;; in the list of features describing the example, or, if it is
;; mentioned, the corresponding feature has the same attribute value. 
;; For instance, if an example corresponds to an office on the third 
;; floor, then any concept consistent with that example must either 
;; not mention a floor or indicate the third floor.
(define (consistent features concept)
  (every2 (lambda (f)
            (or (not (assq (feature-attribute f) features))
                (member2 f features equal? (lambda (x) x))))
          concept))


;; To generalize a concept you might add a disjunct or drop a conjunct.
;; Since we are restricted to conjunctions we only drop conjuncts.
;; For the restricted hypothesis space consisting of conjunctions of 
;; positive literals, generalization corresponds to removing features 
;; from a concept represented as a list of features.
(define (generalize concept)
  (map (lambda (feature) (remove2 feature concept equal? (lambda (x) x)))
       concept))


;; To specialize a concept you add a conjunct or drop a disjunct.
;; Since we are restricted to conjunctions we only add conjuncts.
;; Specialization corresponds to adding features that are consistent 
;; with the concept. Both specialize and generalize return concepts 
;; that are, respectively, minimally more specific and minimally more
;; general than the concept provided as an argument to the function.
(define (specialize concept)
  (map (lambda (feature) (cons feature concept))
       (findall (features)
                (lambda (feature)
                  (and (consistent (list feature) concept)
                       (not (member2 feature
                                     concept
                                     equal?
                                     (lambda (x) x))))))))


;; Construct the set of all possible features.
(define (features)
  (apply append
         (map (lambda (dim)
                (map (lambda (value) (list (dimension-attribute dim) value))
                     (dimension-values dim)))
              (dimensions))))


;; Simplification for the special case of conjunctions of 
;; positive literals:
(define (simplify-general general specific)
  (set! general
    (findall general
             (lambda (concept1)
               (some2 (lambda (concept2)
                        (or (more-specific concept2 concept1)
                            (equivalent concept1 concept2)))
                      specific))))
  (set! general
    (findall general
             (lambda (concept1)
               (not (some2 (lambda (concept2)
                             (more-specific concept1 concept2))
                           general)))))
  (list general specific))


(define (simplify-specific general specific)
  (set! specific
    (findall specific
             (lambda (concept1)
               (some2 (lambda (concept2)
                        (or (more-specific concept1 concept2)
                            (equivalent concept1 concept2)))
                      general))))
  (set! specific
    (findall specific
             (lambda (concept1)
               (not (some2 (lambda (concept2)
                             (more-specific concept2 concept1))
                           specific)))))
  (list general specific))


(define (more-specific concept1 concept2)
  (or (equal? concept1 '(or))
      (and (subsetp2 concept2 concept1 equal?)
           (not (subsetp2 concept1 concept2 equal?)))))


(define (equivalent concept1 concept2)
  (and (subsetp2 concept2 concept1 equal?)
       (subsetp2 concept1 concept2 equal?)))


;; Example from the text

(define (dimensions)
  (list (make-dimension 'status '(faculty staff))
        (make-dimension 'floor '(four five))
        (make-dimension 'dept '(cs ee))))


(define (classes) '(yes no))


;; Versions takes a set of training examples and a pair of initial
;; boundaries and refines the boundaries by processing each training 
;; example in turn.
(define (versions examples boundaries)
  (for-each
    (lambda (example)
      (set! boundaries (refine example (first boundaries) (second boundaries)))
      (begin
        (display "Example: ")
        (display example)
        (newline)
        (display "General: ")
        (display (first boundaries))
        (newline)
        (display "Specific: ")
        (display (second boundaries))
        (newline)))
    examples))


;; Test function.
(let* ((examples
         (list (make-example
                 '412
                 '((status faculty) (floor four) (dept cs))
                 'yes)
               (make-example '509 '((status staff) (floor five) (dept cs)) 'no)
               (make-example
                 '517
                 '((status faculty) (floor five) (dept cs))
                 'yes)
               (make-example
                 '507
                 '((status faculty) (floor five) (dept ee))
                 'no)))
       (boundaries (list (list (list)) (map example-features examples))))
  (set! test (lambda () (versions examples boundaries))))



;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;;      Name:
;;      Organization:
;;      Address (postal and/or electronic):
;;
;; To:
;;      Software Librarian
;;      Computer Science Department, Box 1910
;;      Brown University
;;      Providence, RI 02912
;;
;;              or
;;
;;      brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

