;; File: tm.lisp
;; Date: 11/2/93
;; Description: Automated temporal reasoning 

(in-package 'user) 

(setq *print-pretty* t)

;; ***********
;; Data Types:
;; ***********

;; A CRULE (for causal rule) consists of a list of antecedent conditions,
;; an expression indicating the type of the triggering event, and a list
;; of consequent effects. 
(defun make-CRULE (antecedents trigger effects)
  (list 'CRULE antecedents trigger effects))
(defun CRULE-antecedents (rule) (cadr rule))
(defun CRULE-trigger (rule) (caddr rule))
(defun CRULE-effects (rule) (cadddr rule))

;; An example CRULE:
;; (make-CRULE '((location Ralph room101)
;;	         (not (on (light room101))))
;;	       '(toggle Ralph (switch (light room101)))
;;	       '(make-EFFECT 'FACT '(on (light room101)) 0))

;; EFFECTs correspond to potential events or facts.
;; In addition to a slot indicating whether the EFFECT is a potential event or 
;; fact, each EFFECT specifies a type and a delay from its triggering event.
;; Since the only thing you do with EFFECTs is create new ones or instantiate
;; existing ones, we violate the abstractions somewhat in defining INSTANTIATE.
(defun make-EFFECT (event-or-fact type delay)
  (list event-or-fact type delay))
(defun instantiate (effect time) 
  (cond ((eq (car effect) 'EVENT) 
         (make-EVENT (cadr effect) (+ time (caddr effect))))
        ((eq (car effect) 'FACT)
         (make-FACT (cadr effect) (+ time (caddr effect))))))

;; All events are point events
(defun make-EVENT (type time) 
  (list 'EVENT type time))
(defun is-EVENT (x) (and x (listp x) (eq (car x) 'EVENT)))
(defun EVENT-type (event) (cadr event))
(defun EVENT-time (event) (caddr event))

;; All facts are initially assumed to persist forever
(defun make-FACT (type begin) 
  (list 'FACT type begin '*infty*))
(defun is-FACT (x) (and x (listp x) (eq (car x) 'FACT)))
(defun FACT-type (fact) (cadr fact))
(defun FACT-begin (fact) (caddr fact))
(defun FACT-end (fact) (cadddr fact))
(defun set-FACT-end (fact time) (setf (cadddr fact) time))

(defun project (initial-conditions initial-events rules)
  (initialize initial-conditions initial-events)
  (aux-project rules))

(defun initialize (initial-conditions initial-events)
  (reset-events) (reset-facts)
  (mapc #'(lambda (type) (add-fact (make-FACT type 0)))
        initial-conditions)
  (mapc #'(lambda (spec)
            (add-event (make-EVENT (car spec) (cadr spec))))
        initial-events))

(defun aux-project (rules)
  (let ((event (pop-event)))
    (if (null event) nil
      (progn (mapc #'(lambda (rule) (apply-rule event rule))
		   (fetch-rules event rules))
	     (aux-project rules)))))

(defun fetch-rules (event rules)
  (cond ((null rules) nil)
        ((equal (EVENT-type event) (CRULE-trigger (car rules)))
         (cons (car rules) (fetch-rules event (cdr rules))))
        (t (fetch-rules event (cdr rules)))))

(defun apply-rule (event rule)
  (let ((time (EVENT-time event)))
    (if (check-antecedents (CRULE-antecedents rule) time)
	(mapc #'(lambda (token)
		  (if (is-FACT token) 
		      (add-fact token)
		    (add-event token)))
	      (mapcar #'(lambda (effect) 
			  (instantiate effect time))
		      (CRULE-effects rule))))))

(defun check-antecedents (antecedents time)
  (cond ((null antecedents) t)
        ((not (holds time (car antecedents))) nil)
        (t (check-antecedents (cdr antecedents) time))))

;; ***********************************
;; Keeping track of what is true when:
;; ***********************************

(let ((*facts* nil))
  (defun reset-facts () (setq *facts* nil))
  (defun holds (time type)
    (find type *facts* 
	  :test #'(lambda (x y) (is-true x y time))))
  (defun add-fact (fact)
    (if (holds (FACT-begin fact) (FACT-type fact)) nil
      (let (contra (time (FACT-begin fact)))
	(setq contra (find-contra (FACT-type fact) time *facts*))
;;	(setq contra (find (contradicts (FACT-type fact)) *facts* 
;;			   :test #'(lambda (x y) (is-true x y time))))
	(if contra (set-FACT-end contra (FACT-begin fact)))
	(setq *facts* (cons fact *facts*)))))
  (defun facts () *facts*))

(defun contradicts (type)
  (if (and type (listp type) (eq (car type) 'NOT)) 
      (cadr type) (list 'NOT type)))

(defun find-contra (type time facts)
  (cond ((null facts) nil)
	((and (contradictory type (FACT-type (first facts)))
	      (or (t< (FACT-begin (first facts)) time)
		  (= (FACT-begin (first facts)) time))
	      (t< time (FACT-end (first facts)))) (first facts))
	(t (find-contra type time (rest facts)))))

(defun contradictory (type1 type2)
  (cond ((and (listp type1) (eq (first type1) 'NOT))
	 (equal (second type1) type2))
	((and (listp type2) (eq (car type2) 'NOT))
	 (equal (second type2) type1))
	((and (listp type1) (listp type2) 
	      (eq (first type1) 'LOCATION)
	      (eq (first type2) 'LOCATION)
	      (equal (second type1) (second type2))
	      (not (equal (third type1) (third type2)))))
	(t nil)))

(defun is-true (type fact time)
  (and (equal type (FACT-type fact))
       (or (t< (FACT-begin fact) time)
	   (= (FACT-begin fact) time))
       (t< time (FACT-end fact))))

(defun t< (time1 time2)
  (or (eq time2 '*infty*)
      (and (integerp time1)
           (integerp time2)
           (< time1 time2))))

;; *******************************
;; Keeping track of future events:
;; *******************************

(let ((*events* nil) (*display* nil))
  (defun reset-events () 
    (setq *events* nil)
    (setq *display* nil))
  (defun add-event (event)
    (setq *display* (cons event *display*))
    (setq *events*
	  (sort (cons event *events*)
		#'(lambda (f g) (t< (EVENT-time f) 
				    (EVENT-time g))))))
  (defun pop-event ()
    (let ((item (car *events*)))
      (setq *events* (cdr *events*))
      item))
  (defun events () *display*))

;; **********************************
;; Display the results of projection:
;; **********************************

(let ((*text-width* 75))
  (defun display ()
    (mapc #'(lambda (token)
	      (if (is-EVENT token) 
		  (display-event token) 
		(display-fact token)))
	  (sort (append (facts) (events))
		#'(lambda (x y) (< (token-begin x)
				   (token-begin y)))))
    (terpri))
  (defun token-begin (token)
    (if (is-EVENT token) (EVENT-time token) (FACT-begin token)))
  (defun display-event (event)
    (terpri) (princ (EVENT-type event)) (terpri)
    (cond ((> (EVENT-time event) (+ 2 *text-width*))
	   (tab (- *text-width* 2)) (princ ">>"))
	  (t (tab (EVENT-time event)) (princ "|"))))
  (defun display-fact (fact)
    (terpri) (princ (FACT-type fact)) (terpri)
    (cond ((> (FACT-begin fact) (+ 2 *text-width*))
	   (tab (- *text-width* 2)) (princ ">>"))
	  (t (tab (FACT-begin fact)) (princ "|")))
    (cond ((or (eq (FACT-end fact) '*infty*)
	       (> (FACT-end fact) *text-width*))
	   (printn "-" (- *text-width* (+ (FACT-begin fact) 1)))
	   (princ ">"))
	  (t (printn "-" (- (FACT-end fact) (+ (FACT-begin fact) 1)))
	     (princ "|")))))

(defun printn (string n)
  (cond ((> n 0)
         (princ string) 
         (printn string (- n 1)))
        (t nil)))

(defun tab (n)
  (cond ((or (= n 0) (< n 0)) nil)
        (t (princ " ")
           (tab (- n 1)))))

;; ********
;; Testing:
;; ********

(defun test-project () 
  (let ((initial-conditions '(P Q))
	(initial-events '((E 10)))
	(theory 
	 (list 
	  (make-CRULE '(P) 'E (list (make-EFFECT 'FACT 'R 0)))
	  (make-CRULE '(R Q) 'F (list (make-EFFECT 'FACT '(NOT P) 0)))
	  (make-CRULE '(Q) 'E (list (make-EFFECT 'EVENT 'F 5))))))
    (project initial-conditions initial-events theory)
    (display)))

;; ********************************************
;; Translating causal rules into internal form:
;; ********************************************

;; In this section, we provide conventions for conveniently notating 
;; causal rules and a procedure for translating rules in this notation
;; into a canoncial internal form. All causal rules include antecedent 
;; conditions and triggering events. A list in the antecedent conditions 
;; position whose CAR is EQ to AND indicates a conjunction of fact types. 
;; Any other object (it could be an atom so be careful checking; you can't 
;; take the CAR of an atom) is treated as a single fact type. 
;;  
;; PCAUSE rules have FACTs as consequences. FACTs are specified in a 
;; rather free form; a type is always provided, the delay is optional
;; and defaults to 0 if not stated explicitly. Conjunctions are 
;; allowed (i.e., lists whose CAR is EQ to AND). Processing conjunctions
;; of free form FACT specifications is made simpler by requiring that 
;; delays must be numbers while FACT types cannot be numbers.
;; Example:
;;          (PCAUSE (AND (VALUE (INPUT 1 DEVICE1) 0) 
;;                       (VALUE (INPUT 2 DEVICE1) 1))
;;                  (TRANSITION (CLOCK DEVICE1) 0 1)
;;                  (AND (VALUE (OUTPUT 1 DEVICE1) 1) 3 
;;                       (VALUE (OUTPUT 2 DEVICE1) 1) 
;;                       (VALUE (OUTPUT 3 DEVICE1) 1) 2))
;;          =>
;;          (CRULE ((VALUE (INPUT 1 DEVICE1) 0) 
;;                  (VALUE (INPUT 2 DEVICE1) 1))
;;                 (TRANSITION (CLOCK DEVICE1) 0 1)
;;                 ((FACT (VALUE (OUTPUT 1 DEVICE1) 1) 3)
;;                  (FACT (VALUE (OUTPUT 2 DEVICE1) 1) 0)))
;;                  (FACT (VALUE (OUTPUT 3 DEVICE1) 1) 3)))
;; 
;; ECAUSE rules have EVENTs as consequences. EVENTs are specified in  
;; the same general format as FACTs are specified in PCAUSE rules.
;; Example:
;;          (ECAUSE (VALUE (INPUT 1 DEVICE1) 0)
;;                  (TRANSITION (CLOCK DEVICE1) 0 1)
;;                  (TRANSITION (OUTPUT 3 DEVICE1) 0 1) 2)
;;          =>
;;          (CRULE ((VALUE (INPUT 1 DEVICE1) 0))
;;                 (TRANSITION (CLOCK DEVICE1) 0 1)
;;                 ((EVENT (TRANSITION (OUTPUT 3 DEVICE1) 0 1) 2)))

(defun translate-rule (rule)
  (make-crule (cond ((and (cadr rule) 
                          (listp (cadr rule))
                          (member (car (cadr rule)) '(AND and)))
                     (cdr (cadr rule)))
                    (t (list (cadr rule))))
              (caddr rule)
              (cond ((member (car rule) '(ECAUSE ecause))
                     (mapcar #'(lambda (x) 
                                  (make-effect 'EVENT (car x) (cdr x)))
                             (extract-consequents rule)))
                    (t (mapcar #'(lambda (x) 
                                  (make-effect 'FACT (car x) (cdr x)))
                             (extract-consequents rule))))))

;; Extract a list of types and delays from the consequents of a rule.
;; If RULE = (E/PCAUSE ANTEC TRIG . CONSEQ) then CONSEQ = (cdddr RULE)
;; CONSEQ = ((AND TYPE ... )) or (TYPE) or (TYPE DELAY)
(defun extract-consequents (rule)
  (compile-types-and-delays
        (cond ((and (car (cdddr rule))
                    (listp (car (cdddr rule)))
                    (member (caar (cdddr rule)) '(AND and)))
               (cdar (cdddr rule)))
              (t (cdddr rule)))))
(defun compile-types-and-delays (x)
  (cond ((null x) nil)
        ((and (cdr x) 
              (numberp (cadr x)))
         (cons (cons (car x) (cadr x))
               (compile-types-and-delays (cddr x))))
        (t (cons (cons (car x) 0)
               (compile-types-and-delays (cdr x))))))

;; An variant of PROJECT that accepts the alternative rule notation:
(defun tm (initial-conditions initial-events causal-theory)
  (project initial-conditions 
           initial-events
           (mapcar #'translate-rule causal-theory)))

;; **************
;; Test routines:
;; **************

;; Test the projection routines with rule translation:
(defun test-tm ()
  (reset-events) (reset-facts)
  (let ((initial-conditions '(P Q))
        (initial-events '((E 10)))
        (causal-theory '((PCAUSE P E R)
                         (PCAUSE (AND R Q) F (NOT P))
                         (ECAUSE Q E F 5))))
       (tm initial-conditions initial-events causal-theory)
       (display)))

