;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user) 

;; ***********
;; Data Types:
;; ***********

;; A CRULE (for causal rule) consists of a list of antecedent conditions,
;; an expression indicating the type of the triggering event, and a list
;; of consequent effects. 
(define (make-crule antecedents trigger effects)
  (list 'crule antecedents trigger effects))

(define (crule-antecedents rule) (cadr rule))

(define (crule-trigger rule) (caddr rule))

(define (crule-effects rule) (cadddr rule))


;; Here is the Lisp representation of the causal rule that whenever a 
;; robot toggles the switch for a light that is off, the light will 
;; become on.
;; (make-CRULE '((location Ralph room101)
;;	         (not (on (light room101))))
;;	       '(toggle Ralph (switch (light room101)))
;;	       '(make-EFFECT 'FACT '(on (light room101)) 0))

;; EFFECTs correspond to potential events or facts.  In addition to a
;; slot indicating whether the EFFECT is a potential event or fact, each
;; EFFECT specifies a type and a delay from its triggering event.  Since
;; the only thing you do with EFFECTs is create new ones or instantiate
;; existing ones, we violate the abstractions somewhat in defining
;; INSTANTIATE. 

;; The effects specified in a causal rule correspond to potential events
;; or propositions (facts) that become true and persist for some time.
;; The abstract data type for effects encodes whether the effect is a
;; potential event or fact, the type of event or fact, and a delay
;; indicating how much time elapses from when the triggering event occurs
;; to when the fact or event becomes true. An effect is instantiated in
;; the process of projection when a causal rule is found to apply to a
;; particular event. The only thing we do with effects is create new 
;; ones or instantiate existing ones.
(define (make-effect event-or-fact type delay-time) 
  (list event-or-fact type delay-time))

(define (instantiate effect time)
  (cond ((eq? (car effect) 'event)
         (make-event (cadr effect) (+ time (caddr effect))))
        ((eq? (car effect) 'fact)
         (make-fact (cadr effect) (+ time (caddr effect))))))


;; All events are point events. Having differentiated between facts 
;; and events, we have to introduce abstract data types for each. 
;; All events are point events, so we need only an event type and 
;; time of occurrence.
(define (make-event type time) (list 'event type time))

(define (is-event x) 
  (and (not (null? x)) (list? x) (eq? (car x) 'event)))

(define (event-type event) (cadr event))

(define (event-time event) (caddr event))


;; All facts are initially assumed to persist forever.  Facts are similar
;; to events except that they persist over intervals of time. All facts
;; are initially assumed to persist indefinitely. To represent indefinite
;; persistence, we introduce a special symbol infty meant to indicate
;; that there is no upper bound on the duration of the corresponding
;; interval. Since we may have to revise this bound in the light of
;; additional information, we make the ends of facts settable.
(define (make-fact type time) (list 'fact type time '*infty*))

(define (is-fact x) (and (not (null? x)) (list? x) (eq? (car x) 'fact)))

(define (fact-type fact) (cadr fact))

(define (fact-begin fact) (caddr fact))

(define (fact-end fact) (cadddr fact))

(define (set-fact-end fact time) (set-car! (cdddr fact) time))


;; Now we can list the rest of the code for projection and persistence
;; clipping. We proceed in a top-down manner starting with the function
;; project that takes a list of initial facts true at time 0, a list of 
;; events known to occur at various times, and a list of causal rules. 
;; Project does not return anything of interest; it modifies certain 
;; internal state variables to construct a time map in the form of a 
;; list of tokens representing propositions that are true over 
;; intervals of time.
(define (project initial-conditions initial-events rules)
  (initialize initial-conditions initial-events)
  (aux-project rules))


(define (initialize initial-conditions initial-events)
  (reset-events)
  (reset-facts)
  (for-each (lambda (type) (add-fact (make-fact type 0))) 
	    initial-conditions)
  (for-each
    (lambda (spec) (add-event (make-event (car spec) (cadr spec))))
    initial-events))


;; The events are stored in a sorted queue in the order of their
;; occurrence, earliest to latest. In aux-project, the earliest occurring
;; event is removed from the list of events and all of the causal rules
;; with the corresponding triggering event type are checked to see if
;; their antecedents are true at the time of the event.
(define (aux-project rules)
  (let ((event (pop-event)))
    (if (not event)
      #f
      (begin
        (for-each
          (lambda (rule) (apply-rule event rule))
          (fetch-rules event rules))
        (aux-project rules)))))


;; Fetch-rules takes an event and a set of causal rules and returns all 
;; rules with triggering event type that corresponds to the type of the 
;; event given as the first argument.
(define (fetch-rules event rules)
  (cond ((null? rules) (list))
        ((equal? (event-type event) (crule-trigger (car rules)))
         (cons (car rules) (fetch-rules event (cdr rules))))
        (else (fetch-rules event (cdr rules)))))


;; Apply-rule takes an event and a causal rule and, if the antecedents of
;; the rule are satisfied at the time of the event, then it adds to the
;; database the appropriate events and facts corresponding to the effects
;; supplied in consequences of the rule.
(define (apply-rule event rule)
  (let ((time (event-time event)))
    (if (check-antecedents (crule-antecedents rule) time)
      (for-each
        (lambda (token)
          (if (is-fact token) (add-fact token) (add-event token)))
        (map (lambda (effect) (instantiate effect time))
             (crule-effects rule))))))


;; To determine if the antecedents of a rule are satisfied at a
;; particular time, check-antecedents invokes the function holds which
;; determines whether or not a proposition is true at a time by looking
;; at the facts in the database. Projection proceeds by sweeping forward
;; in time so that, when a causal rule is considered with respect to a
;; particular triggering event, the database is complete up until the
;; time of that event.
(define (check-antecedents antecedents time)
  (cond ((null? antecedents) #t)
        ((not (holds time (car antecedents))) #f)
        (else (check-antecedents (cdr antecedents) time))))


;; ***********************************
;; Keeping track of what is true when:
;; ***********************************

;; Now we describe the database and the operations on the database. We
;; introduce local state variables for the functions that operate on
;; facts and events. We can reset the list of facts to the empty list,
;; add a new fact or determine if there is a fact of a particular type in
;; the database whose interval spans a given time point. In adding a new
;; fact, we clip the persistence of any existing fact of contradictory
;; type. We use the Common Lisp find function to seach through the list
;; of facts to find one with a given type that begins at or before a
;; given time and persists beyond that time. Find returns the first
;; element of its second argument such that the test function applied to
;; this element and the first argument returns true.
(let ((*facts* (list)))
  (set! reset-facts (lambda () (set! *facts* (list))))
  (set! holds
    (lambda (time type)
      (let ((result-of-find
              (member2 type
                       *facts*
                       (lambda (x y) (is-true x y time))
                       (lambda (x) x))))
        (if result-of-find (car result-of-find) #f))))
  (set! add-fact
    (lambda (fact)
      (if (holds (fact-begin fact) (fact-type fact))
        #f
        (let ((contra #f) (time (fact-begin fact)))
          (set! contra (find-contra (fact-type fact) time *facts*))
          (if contra (set-fact-end contra (fact-begin fact)))
          (set! *facts* (cons fact *facts*))))))
  (set! facts (lambda () *facts*)))


;; The following method of determining a contradictory type will suffice
;; for our examples.
(define (contradicts type)
  (if (and (not (null? type)) (list? type) (eq? (car type) 'not))
    (cadr type)
    (list 'not type)))


(define (find-contra type time facts)
  (cond ((null? facts) #f)
        ((and (contradictory type (fact-type (first facts)))
              (or (t< (fact-begin (first facts)) time)
                  (= (fact-begin (first facts)) time))
              (t< time (fact-end (first facts))))
         (first facts))
        (else (find-contra type time (rest facts)))))


(define (contradictory type1 type2)
  (cond ((and (list? type1) (eq? (first type1) 'not))
         (equal? (second type1) type2))
        ((and (list? type2) (eq? (car type2) 'not))
         (equal? (second type2) type1))
        ((and (list? type1)
              (list? type2)
              (eq? (first type1) 'location)
              (eq? (first type2) 'location)
              (equal? (second type1) (second type2))
              (not (equal? (third type1) (third type2)))))
        (else #f)))


;; A proposition (type) is true at a time given a fact (token) if the 
;; type of the fact is the same as the proposition and the fact begins 
;; before or coincident with the time and ends prior to the time.
(define (is-true type fact time)
  (and (equal? type (fact-type fact))
       (or (t< (fact-begin fact) time) (= (fact-begin fact) time))
       (t< time (fact-end fact))))


;; Temporal precedence is complicated somewhat by the symbol, infty, 
;; used to represent the indefinite upper bound on the duration of 
;; fact tokens.
(define (t< time1 time2)
  (or (eq? time2 '*infty*)
      (and (integer? time1) (integer? time2) (< time1 time2))))


;; *******************************
;; Keeping track of future events:
;; *******************************

;; In this implementation, we only keep track of events to use in
;; predicting additional facts and events. We maintain the list of events
;; in a queue ordered by time. Since events can only cause future events
;; and facts, we pop the events off the queue one at a time in performing
;; our sweep forward in time for projection.
(let ((*events* (list)) (*display* (list)))
  (set! reset-events
    (lambda () (set! *events* (list)) (set! *display* (list))))
  (set! add-event
    (lambda (event)
      (set! *display* (cons event *display*))
      (set! *events*
        (sort2 (cons event *events*)
               (lambda (f g) (t< (event-time f) (event-time g)))
               (lambda (x) x)))))
  (set! pop-event
    (lambda ()
      (if (null? *events*) #f
	  (let ((item (car *events*))) 
	    (set! *events* (cdr *events*)) item))))
  (set! events (lambda () *display*)))


;; **********************************
;; Display the results of projection:
;; **********************************

(let ((*text-width* 75))
  (set! display-tm
    (lambda ()
      (for-each
        (lambda (token)
	  (if (is-event token) 
	      (display-event token) 
	      (display-fact token)))
        (sort2 (append (facts) (events))
               (lambda (x y) (< (token-begin x) (token-begin y)))
               (lambda (x) x)))
      (newline)))
  (set! token-begin
    (lambda (token)
      (if (is-event token) (event-time token) (fact-begin token))))
  (set! display-event
    (lambda (event)
      (newline)
      (display (event-type event))
      (newline)
      (cond ((> (event-time event) (+ 2 *text-width*))
             (tab (- *text-width* 2))
             (display ">>"))
            (else (tab (event-time event)) (display "|")))))
  (set! display-fact
    (lambda (fact)
      (newline)
      (display (fact-type fact))
      (newline)
      (cond ((> (fact-begin fact) (+ 2 *text-width*))
             (tab (- *text-width* 2))
             (display ">>"))
            (else (tab (fact-begin fact)) (display "|")))
      (cond ((or (eq? (fact-end fact) '*infty*)
                 (> (fact-end fact) *text-width*))
             (printn "-" (- *text-width* (+ (fact-begin fact) 1)))
             (display ">"))
            (else
             (printn "-" (- (fact-end fact) (+ (fact-begin fact) 1)))
             (display "|"))))))


(define (printn string n)
  (if (> n 0) (begin (display string) (printn string (- n 1)))))


(define (tab n) (if (> n 0) (begin (display " ") (tab (- n 1)))))


;; ********
;; Testing:
;; ********

(define (test-project)
  (let ((initial-conditions '(p q))
        (initial-events '((e 10)))
        (theory (list (make-crule '(p) 'e 
				  (list (make-effect 'fact 'r 0)))
                      (make-crule '(r q) 'f
				  (list (make-effect 'fact '(not p) 0)))
                      (make-crule '(q) 'e (list 
					   (make-effect 'event 'f 5))))))
    (project initial-conditions initial-events theory)
    (display-tm)))


;; ********************************************
;; Translating causal rules into internal form:
;; ********************************************

;; In this section, we provide conventions for conveniently notating 
;; causal rules and a procedure for translating rules in this notation
;; into a canoncial internal form. All causal rules include antecedent 
;; conditions and triggering events. A list in the antecedent conditions 
;; position whose CAR is EQ to AND indicates a conjunction of fact types.
;; Any other object (it could be an atom so be careful checking; you 
;; can't take the CAR of an atom) is treated as a single fact type. 
;;  
;; PCAUSE rules have FACTs as consequences. FACTs are specified in a 
;; rather free form; a type is always provided, the delay is optional
;; and defaults to 0 if not stated explicitly. Conjunctions are 
;; allowed (i.e., lists whose CAR is EQ to AND). Processing conjunctions
;; of free form FACT specifications is made simpler by requiring that 
;; delays must be numbers while FACT types cannot be numbers.
;; Example:
;;          (PCAUSE (AND (VALUE (INPUT 1 DEVICE1) 0) 
;;                       (VALUE (INPUT 2 DEVICE1) 1))
;;                  (TRANSITION (CLOCK DEVICE1) 0 1)
;;                  (AND (VALUE (OUTPUT 1 DEVICE1) 1) 3 
;;                       (VALUE (OUTPUT 2 DEVICE1) 1) 
;;                       (VALUE (OUTPUT 3 DEVICE1) 1) 2))
;;          =>
;;          (CRULE ((VALUE (INPUT 1 DEVICE1) 0) 
;;                  (VALUE (INPUT 2 DEVICE1) 1))
;;                 (TRANSITION (CLOCK DEVICE1) 0 1)
;;                 ((FACT (VALUE (OUTPUT 1 DEVICE1) 1) 3)
;;                  (FACT (VALUE (OUTPUT 2 DEVICE1) 1) 0)))
;;                  (FACT (VALUE (OUTPUT 3 DEVICE1) 1) 3)))
;; 
;; ECAUSE rules have EVENTs as consequences. EVENTs are specified in  
;; the same general format as FACTs are specified in PCAUSE rules.
;; Example:
;;          (ECAUSE (VALUE (INPUT 1 DEVICE1) 0)
;;                  (TRANSITION (CLOCK DEVICE1) 0 1)
;;                  (TRANSITION (OUTPUT 3 DEVICE1) 0 1) 2)
;;          =>
;;          (CRULE ((VALUE (INPUT 1 DEVICE1) 0))
;;                 (TRANSITION (CLOCK DEVICE1) 0 1)
;;                 ((EVENT (TRANSITION (OUTPUT 3 DEVICE1) 0 1) 2)))

(define (translate-rule rule)
  (make-crule
    (cond ((and (not (null? (cadr rule)))
                (list? (cadr rule))
                (eq? (car (cadr rule)) 'and))
           (cdr (cadr rule)))
          (else (list (cadr rule))))
    (caddr rule)
    (cond ((eq? (car rule) 'ecause)
           (map (lambda (x) (make-effect 'event (car x) (cdr x)))
                (extract-consequents rule)))
          (else
           (map (lambda (x) (make-effect 'fact (car x) (cdr x)))
                (extract-consequents rule))))))


;; Extract a list of types and delays from the consequents of a rule.
;; If RULE = (E/PCAUSE ANTEC TRIG . CONSEQ) then CONSEQ = (cdddr RULE)
;; CONSEQ = ((AND TYPE ... )) or (TYPE) or (TYPE DELAY)
(define (extract-consequents rule)
  (compile-types-and-delays
    (cond ((and (not (null? (car (cdddr rule))))
                (list? (car (cdddr rule)))
                (eq? (caar (cdddr rule)) 'AND))
           (cdar (cdddr rule)))
          (else (cdddr rule)))))

(define (compile-types-and-delays x)
  (cond ((null? x) (list))
        ((and (not (null? (cdr x))) (number? (cadr x)))
         (cons (cons (car x) (cadr x)) 
	       (compile-types-and-delays (cddr x))))
        (else (cons (cons (car x) 0) 
		    (compile-types-and-delays (cdr x))))))

;; An variant of PROJECT that accepts the alternative rule notation:
(define (tm initial-conditions initial-events causal-theory)
  (project initial-conditions
           initial-events
           (map translate-rule causal-theory)))


;; **************
;; Test routines:
;; **************

;; Test the projection routines with rule translation:
(define (test-tm)
  (reset-events)
  (reset-facts)
  (let ((initial-conditions '(p q))
        (initial-events '((e 10)))
        (causal-theory
          '((pcause p e r) 
	    (pcause (and r q) f (not p)) 
	    (ecause q e f 5))))
    (tm initial-conditions initial-events causal-theory)
    (display-tm)))


;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

