;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information
(in-package 'user)

;; State-space operators for the Eight Puzzle.

;; Fluent conditions: (tile (? n) (? x) (? y))
;;                    (blank (? x) (? y))

;; Static conditions: (incr (? x) (? y)) 
;;                    (decr (? x) (? y)) 
;;                    (not (= (? x) (? y)))

;; Every state must include the following static conditions.
(setq static '((incr 1 2) (incr 2 3)
	       (decr 3 2) (decr 2 1)
	       (not (= 2 3)) (not (= 1 3)) 
	       (not (= 3 1)) (not (= 2 1))))


;; right and top to bottom with the blank in the lower right hand corner. 
(setq goal '((tile 1 1 1) (tile 2 1 2) (tile 3 1 3)
	     (tile 4 2 1) (tile 5 2 2) (tile 6 2 3)
	     (tile 7 3 1) (tile 8 3 2) (blank 3 3)))

;; Here is an easy case with only two tiles out of place. 
(setq easy 
      (append static
	      '((tile 1 1 1) (tile 2 1 2) (tile 3 1 3)
		(tile 4 2 1) (blank  2 2) (tile 5 2 3) 
		(tile 7 3 1) (tile 8 3 2) (tile 6 3 3))))

;; Move the blank tile up.
;;      x  |  n  |  x        x  |     |  x
;;      x  |     |  x	=>   x  |  n  |  x
;;      x  |  x  |  x	     x  |  x  |  x
(setq up '(((blank (? x) (? y))         ;; Note that order of preconditions
	    (not (= (? x) 1))           ;; is critical for efficiency. 
	    (decr (? x) (? z))          ;; Consider the increased number of
	    (tile (? n) (? z) (? y)))   ;; unifications if the TILE condition
	   ((blank (? z) (? y))         ;; occurred as the first precondition. 
	    (tile (? n) (? x) (? y)))
	   ((tile (? n) (? z) (? y))
	    (blank (? x) (? y)))))

;; Move the blank tile down.
;;      x  |  x  |  x        x  |  x  |  x
;;      x  |     |  x	=>   x  |  n  |  x
;;      x  |  n  |  x	     x  |     |  x
(setq down '(((blank (? x) (? y))     
	      (not (= (? x) 3))         
	      (incr (? x) (? z))        
	      (tile (? n) (? z) (? y))) 
	     ((blank (? z) (? y))       
	      (tile (? n) (? x) (? y)))
	     ((tile (? n) (? z) (? y))
	      (blank (? x) (? y)))))

;; Move the blank tile left.
;;      x  |  x  |  x        x  |  x  |  x
;;      n  |     |  x	=>      |  n  |  x
;;      x  |  x  |  x	     x  |  x  |  x
(setq left '(((blank (? x) (? y))
	      (not (= (? y) 1))
	      (decr (? y) (? z)) 
	      (tile (? n) (? x) (? z)))
	     ((blank (? x) (? z))
	      (tile (? n) (? x) (? y)))
	     ((blank (? x) (? y))
	      (tile (? n) (? z) (? y)))))

;; Move the blank tile right.
;;      x  |  x  |  x        x  |  x  |  x
;;      x  |     |  n	=>   x  |  n  |  
;;      x  |  x  |  x	     x  |  x  |  x
(setq right '(((blank (? x) (? y))
	       (not (= (? y) 3))
	       (incr (? y) (? z)) 
	       (tile (? n) (? x) (? z)))
	      ((blank (? x) (? z))
	       (tile (? n) (? x) (? y)))
	      ((blank (? x) (? y))
	       (tile (? n) (? z) (? y)))))

;; There are four operator schemas.
(setq operators (list up down right left))

;; Count the number of tiles out of place. 
(defun distance (state goal)
  (let ((total 0))
    (mapc #'(lambda (x)
	      (if (member x state :test #'equal) nil
		(setq total (+ 1 total))))
	  goal)
    total))

;; Means-ends analysis and breadth-first search on an easy problem.
;;
;; > (mea easy)
;; Unifications: 524
;; > (sss easy)
;; Unifications: 2876

;; Copyright 1994, Brown University, Providence, RI
;; 
;; Permission to use and modify this software and its documentation for any
;; purpose other than its incorporation into a commercial product is hereby
;; granted without fee.  Permission to copy and distribute this software and its
;; documentation only for non-commercial use is also granted without fee,
;; provided, however, that the above copyright notice appear in all copies, that
;; both that copyright notice and this permission notice appear in supporting
;; documentation, that the name of Brown University not be used in advertising or
;; publicity pertaining to distribution of the software without specific, written
;; prior permission, and that the person doing the distribution notify Brown
;; University of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this software for
;; any purpose.  It is provided "as is" without express or implied warranty.
;; Brown University requests notification of any modifications to this software
;; or its documentation.
;; 
;; Send the following redistribution information:
;; 
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;; 
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;; 
;; 		or
;; 
;; 	brusd@cs.brown.edu
;; 
;; We will acknowledge all electronic notifications.
