;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;; Least commitment planning

;; In the following, a plan consists of a set of steps corresponding to
;; operations, constraints on the order in which these steps are to be
;; carried out, and information regarding how the steps depend on one
;; another.  The dependency information is of three different types.
;; First, there are records called requirements indicating propositions
;; that are required in order that certain operators have their desired
;; effect. Next, there are records called links describing how the
;; consequences of one operator are used to satisfy the requirements of
;; another. Finally, there are records called conflicts describing
;; potential undesirable interactions between operators. As an example of
;; a conflict, one operator might delete a proposition that another
;; operator adds, where the deleted proposition is required for the plan
;; to succeed.
(define (make-plan steps constraints conflicts links requirements)
  (list steps constraints conflicts links requirements))

(define (plan-steps plan) (car plan))

(define (plan-constraints plan) (cadr plan))

(define (plan-conflicts plan) (caddr plan))

(define (plan-links plan) (cadddr plan))

(define (plan-requirements plan) (cadddr (cdr plan)))


;; A link involves two steps and a condition. One step, called the
;; producer, has the condition in the list of additions for its
;; associated operator.  The producer makes the condition true. The other
;; step, called the consumer, has the condition in the list of
;; preconditions for its associated operator.  The consumer requires that
;; the condition be true. Links are created to satisfy requirements.
(define (make-link producer condition consumer)
  (list producer condition consumer))

(define (link-producer link) (car link))

(define (link-condition link) (cadr link))

(define (link-consumer link) (caddr link))


;; A conflict involves a link and a step. The step, called the clobberer,
;; has the condition of the link in the list of deletions for its
;; associated operator. A conflict arises when a step is added to a plan
;; and that step might be carried out after the producer of a link and
;; before the consumer, preventing the link from satisfying the
;; requirement that the link was introduced to satisfy. When a conflict
;; occurs the clobberer is said to threaten link.
(define (make-conflict link clobberer) (list link clobberer))

(define (conflict-link conflict) (car conflict))

(define (conflict-clobberer conflict) (cadr conflict))


;; We represent constraints in terms of their begin and end steps, and 
;; assign each step a unique integer to distinguish different steps 
;; employing the same operator.
(let ((n 0))
  (set! make-step 
	(lambda (operator) 
	  (set! n (+ n 1))
	  (list n operator))))

(define (step-id step) (car step))

(define (step-operator step) (cadr step))


;; As in previous section, we describe a state as a set of conditions.
;; A requirement corresponds to a step and a condition that must be true
;; immediately prior to that step. The condition of the requirement
;; corresponds to a precondition of the operator indicated in the step.
(define (make-requirement step condition) (list step condition))

(define (requirement-step req) (car req))

(define (requirement-condition req) (cadr req))


(define (make-operator preconditions additions deletions)
  (list preconditions additions deletions))

(define (preconditions operator) (car operator))

(define (additions operator) (car (cdr operator)))

(define (deletions operator) (car (cdr (cdr operator))))


;; The Lisp implementation for refinements is provided as follows. Recall
;; that, unless the plan satisfies the goal, it has either conflicts or
;; requirements.
(define (refinements plan operators)
  (if (not (null? (plan-conflicts plan)))
    (resolve-conflict (car (plan-conflicts plan)) plan)
    (append (resolve-req-new-step
              (car (plan-requirements plan))
              operators
              plan)
            (resolve-req-existing-step 
	     (car (plan-requirements plan)) plan))))


;; Conflicts are resolved by constraining the clobberer to occur before
;; the producer of the associated link or after the consumer of the link.
(define (resolve-conflict conflict plan)
  (let ((link (conflict-link conflict)) (step (conflict-clobberer conflict)))
    (append (constrain step (link-producer link) plan)
            (constrain (link-consumer link) step plan))))


;; A new plan is created with one step to occur before another if the two
;; steps are not already constrained to occur in the opposite order.
(define (constrain step1 step2 plan)
  (if (precedes step2 step1 (plan-constraints plan))
    (list)
    (list (make-plan
            (plan-steps plan)
            (adjoin2 (list step1 step2) (plan-constraints plan) equal?)
            (cdr (plan-conflicts plan))
            (plan-links plan)
            (plan-requirements plan)))))


;; Precedes determines if one step precedes another given a set of 
;; constraints.
(define (precedes step1 step2 constraints)
  (or (equal? step1 step2)
      (some2 (lambda (c)
               (and (equal? step1 (car c))
                    (precedes (cadr c) step2 constraints)))
             constraints)))


;; To eliminate a requirement by adding a new step, create a new plan for
;; each applicable operator.
(define (resolve-req-new-step req operators plan)
  (apply append (map (lambda (p) (applicablep p req plan)) operators)))


;; An operator is applicable just in case the condition of the
;; requirement is added by the operator. If applicable, create a new plan
;; from the old one by adding a new step, constraining the new step to
;; precede the step of the resolved requirement, elimiating this
;; requirement, adding a link resolving the requirement with the new step
;; as producer, adding a new set of requirements corresponding to the
;; preconditions of the operator, and updating the set of conflicts.
;; Conflicts can arise when the deletions of the new step threaten
;; existing links or when existing steps threaten the new link.
(define (applicablep operator req plan)
  (if (not (member2 (requirement-condition req)
                    (additions operator)
                    equal?
                    (lambda (x) x)))
    (list)
    (let* ((step (make-step operator))
           (constraint (list step (requirement-step req)))
           (link (make-link
                   step
                   (requirement-condition req)
                   (requirement-step req))))
      (list (make-plan
              (cons step (plan-steps plan))
              (adjoin2 constraint (plan-constraints plan) equal?)
              (append (link-conflicts link plan)
                      (step-conflicts step plan)
                      (plan-conflicts plan))
              (cons link (plan-links plan))
              (append (generate-requirements operator step)
                      (cdr (plan-requirements plan))))))))


(define (resolve-req-existing-step req plan)
  (apply append (map (lambda (s) (linkablep s req plan)) (plan-steps plan))))


;; An existing step can be linked to satisfy a requirement just in case
;; its associated operator adds the condition of the requirement and the
;; existing step is not constrained to follow the step of the
;; requirement. If this criterion is met, create a new plan from the old
;; one by constraining the existing step to the step of the resolved
;; requirement, eliminating this requirement, adding a link resolving the
;; requirement with the existing step as producer, and updating the set
;; of conflicts.
(define (linkablep step req plan)
  (if (or (not (member2 (requirement-condition req)
                        (additions (step-operator step))
                        equal?
                        (lambda (x) x)))
          (precedes (requirement-step req) step (plan-constraints plan)))
    (list)
    (let ((link (make-link
                  step
                  (requirement-condition req)
                  (requirement-step req)))
          (constraint (list step (requirement-step req))))
      (list (make-plan
              (plan-steps plan)
              (adjoin2 constraint (plan-constraints plan) equal?)
              (append (link-conflicts link plan) (plan-conflicts plan))
              (cons link (plan-links plan))
              (cdr (plan-requirements plan)))))))


;; When a link is added, we find all steps that might conflict with it.
;; This function extends the data abstraction for objects of type LINK.

(define (link-conflicts link plan)
  (apply append (map (lambda (step) (conflictp link step)) (plan-steps plan))))


;; When a step is added, we find all links that might conflict with it.
;; This function extends the data abstraction for objects of type STEP.

(define (step-conflicts step plan)
  (apply append (map (lambda (link) (conflictp link step)) (plan-links plan))))


;; A link and a step conflict whenever the operator of the step deletes 
;; the condition of the link, unless the step is the consumer of the 
;; link.
(define (conflictp link step)
  (if (and (not (equal? step (link-consumer link)))
           (member2 (link-condition link)
                    (deletions (step-operator step))
                    equal?
                    (lambda (x) x)))
    (list (make-conflict link step))
    (list)))


;; The operator of a step has associated with it one requirement for each
;; of its preconditions.
(define (generate-requirements operator step)
  (map (lambda (p) (make-requirement step p)) (preconditions operator)))


(define (best states goalp next comparator)
  (cond ((null? states) #f)
        ((goalp (car states)) (car states))
        (else
         (best (sort2 (append (next (car states)) (cdr states))
                      comparator
                      (lambda (x) x))
               goalp
               next
               comparator))))


;; Simple abstract test:

;; (setq start (make-step (make-operator () '(P Q) ())))
;; (setq finish (make-step (make-operator '(R) () ())))
;; (setq reqs (list (make-requirement finish 'R)))
;; (setq operators (list (make-operator '(P) '(R) ())
;; 		      (make-operator '(Q) '(S) ())))
;; (setq plan (make-plan (list start finish)
;; 		      (list (list start finish)) () () 
;; 		      reqs))

;; Sussman's anomaly:

;; The start step uses a pseudo operator to encode the initial 
;; conditions.
(set! start
  (make-step
    (make-operator
      '()
      '((on c a) (on a table) (on b table) (clear c) (clear b))
      '())))


;; The finish step uses a pseudo operator to encode the goal conditions
(set! finish (make-step (make-operator '((on a b) (on b c)) '() '())))


;; The initial requirements correspond to the goal conditions
(set! requirements
  (list (make-requirement finish '(on a b))
        (make-requirement finish '(on b c))))


;; The initial plan steps consists of the start and the finish steps.
(set! plan
  (make-plan
    (list start finish)
    (list (list start finish))
    '()
    '()
    requirements))


;; We show only the operators needed for the example.
(set! operators
  (list (make-operator
          '((on c a) (clear c))
          '((on c table) (clear a))
          '((on c a)))
        (make-operator
          '((on a table) (clear a) (clear b))
          '((on a b))
          '((on a table) (clear b)))
        (make-operator
          '((on b table) (clear b) (clear c))
          '((on b c))
          '((on b table) (clear c)))))


;; Test function:

(define (test)
  (if (best (list plan)
            (lambda (p)
              (if (and (null? (plan-conflicts p))
                       (null? (plan-requirements p)))
                (begin (plan-print p) #t)
                #f))
            (lambda (p) (refinements p operators))
            (lambda (p q)
              (< (length (plan-requirements p))
                 (length (plan-requirements q)))))
    #t
    #f))


(define (partial plan)
  (map (lambda (step1)
         (cons (step-id step1)
               (apply append
                      (map (lambda (step2)
                             (if (or (precedes
				      step1
				      step2
				      (plan-constraints plan))
				     (precedes
				      step2
				      step1
				      (plan-constraints plan)))
				 (list)
				 (list (step-id step2))))
                           (plan-steps plan)))))
       (sort2 (plan-steps plan)
              (lambda (x y) (precedes x y (plan-constraints plan)))
              (lambda (x) x))))


;; Pretty printer for plans:

(define (plan-print p)
  (set! *print-pretty* #t)
  (newline)
  (display "Steps:")
  (newline)
  (do ((steps (plan-steps p) (cdr steps)))
      ((null? steps) #f)
    (display (car steps))
    (newline))
  (display "Constraints:")
  (newline)
  (do ((constraints (plan-constraints p) (cdr constraints)))
      ((null? constraints) #f)
    (display (step-id (car (car constraints))))
    (display " precedes ")
    (display (step-id (cadr (car constraints))))
    (newline))
  (display "Links:")
  (newline)
  (do ((links (plan-links p) (cdr links)))
      ((null? links) #f)
    (link-print (car links))
    (newline))
  (display "Conflicts:")
  (newline)
  (do ((conflicts (plan-conflicts p) (cdr conflicts)))
      ((null? conflicts) #f)
    (display (step-id (conflict-clobberer (car conflicts))))
    (display " conflicts with ")
    (link-print (conflict-link (car conflicts)))
    (newline))
  (display "Requirements:")
  (newline)
  (do ((reqs (plan-requirements p) (cdr reqs)))
      ((null? reqs) #f)
    (display (step-id (requirement-step (car reqs))))
    (display " requires ")
    (display (requirement-condition (car reqs)))
    (newline))
  (display "Partial order:")
  (newline)
  (display (partial p)))


(define (link-print link)
  (display (step-id (link-producer link)))
  (display " produces ")
  (display (link-condition link))
  (display " for ")
  (display (step-id (link-consumer link))))



;; > (test)
;; 
;; Steps:
;; (5 (((ON B TABLE) (CLEAR B) (CLEAR C)) 
;;     ((ON B C))
;;     ((ON B TABLE) (CLEAR C))))                                   
;; (4 (((ON C A) (CLEAR C)) 
;;     ((ON C TABLE) (CLEAR A))
;;     ((ON C A))))
;; (3 (((ON A TABLE) (CLEAR A) (CLEAR B)) 
;;     ((ON A B))
;;     ((ON A TABLE) (CLEAR B))))
;; (2 (((ON A B) (ON B C)) 
;;     NIL 
;;     NIL))
;; (1 (NIL 
;;     ((ON C A) (ON A TABLE) (ON B TABLE) (CLEAR C) (CLEAR B))
;;     NIL))
;; Constraints:
;; 5 precedes 3
;; 1 precedes 5
;; 4 precedes 5
;; 5 precedes 2
;; 1 precedes 4
;; 4 precedes 3
;; 1 precedes 3
;; 3 precedes 2
;; 1 precedes 2
;; Links:
;; 1 produces (ON B TABLE) for 5
;; 1 produces (CLEAR B) for 5
;; 1 produces (CLEAR C) for 5
;; 5 produces (ON B C) for 2
;; 1 produces (ON A TABLE) for 3
;; 1 produces (ON C A) for 4
;; 1 produces (CLEAR C) for 4
;; 4 produces (CLEAR A) for 3
;; 1 produces (CLEAR B) for 3
;; 3 produces (ON A B) for 2
;; Conflicts:
;; Requirements:


;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

