;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;; Complete the exercise at the end of chapter.  Redefine the
;; function DISTANCE to compute the metric described in the 
;; search chapter.  Create an easy instance of the Eight 
;; Puzzle with only a few tiles out of place, e.g., 
;;      1  |  2  |  3       1  |  2  |  3       1  |  2  |  3
;;      4  |     |  5	=>  4  |  5  |      =>  4  |  5  |  6
;;      7  |  8  |  6	    7  |  8  |  6 	7  |  8  |
;; and demonstrate MEA and SSS solving the instance using your
;; state-space representation for operators, states and goals.

;; Data abstraction for search operators.
(define (make-operator preconditions additions deletions)
  (list preconditions additions deletions))

(define (preconditions operator) (car operator))

(define (additions operator) (car (cdr operator)))

(define (deletions operator) (car (cdr (cdr operator))))


;; Data abstraction for logic variables.
(define (make-var var) (list '? var))

(define (is-var x) (and (pair? x) (eq? (car x) '?)))


;; Given a set of operators and an initial state, return the set of all
;; resulting states obtained by applying operators whose preconditions 
;; are satisfied in the initial state.
(define (apply-operators operators state)
  (if (null? operators)
    (list)
    (let ((p (car operators)) (q #f) (subs #f))
      (set! subs (instances p state))
      (append (map (lambda (sub)
                     (set! q (varsubst p sub))
                     (union2 (additions q)
                             (set-difference2 state (deletions q) equal?)
                             equal?))
                   subs)
              (apply-operators (cdr operators) state)))))


;; Find all instances of a given operator satisfied in a given state.
(define (instances operator state)
  (aux-instances (preconditions operator) state '((bogus #t))))

(define (aux-instances preconditions state bdgs)
  (if (null? preconditions)
    (list bdgs)
    (let ((bdgss (findem (car preconditions) state bdgs)))
      (apply append
             (map (lambda (bdg) 
		    (aux-instances (cdr preconditions) state bdg))
                  bdgss)))))


;; Determine all instances of a pattern in a list of expressions.
(define (findem pat consts bdgs)
  (apply append
         (map (lambda (const)
                (let ((bdgs (match const pat bdgs))) 
		  (if bdgs (list bdgs) (list))))
              consts)))


;; Find a substitution instance for pattern and a constant.
;; The local variable is used to log summary statistics.
(let ((unifications 0))
  (set! reset (lambda () (set! unifications 0)))
  (set! get-unifications (lambda () unifications))
  (set! match
    (lambda (const pat bdgs)
      (set! unifications (+ 1 unifications))
      (aux-match const pat bdgs)))
  (set! aux-match
    (lambda (const pat bdgs)
      (cond ((not bdgs) #f)
            ((null? pat) (if (null? const) bdgs #f))
            ((atom pat) (if (eq? pat const) bdgs #f))
            ((is-var pat)
             (let ((bdg (assoc2 pat bdgs equal? (lambda (x) x))))
               (cond (bdg (aux-match const (cadr bdg) bdgs))
                     (else (cons (list pat const) bdgs)))))
            ((or (atom const) (null? const)) #f)
            (else
             (aux-match
               (car const)
               (car pat)
               (aux-match (cdr const) (cdr pat) bdgs)))))))


;; Compute a subsitution from a pattern and set of bindings.
(define (varsubst p bdgs)
  (if (null? bdgs) p
      (let ((bdg (car bdgs)))
	(varsubst
	 (subst2 (cadr bdg) (car bdg) p equal? (lambda (x) x))
	 (cdr bdgs)))))


;; State-space search using breadth-first search.
(define (sss-bfs states goal operators)
  (bfs states
       (lambda (s) (subsetp2 goal s equal?))
       (lambda (s) (apply-operators operators s))))


(define (bfs states goalp next)
  (cond ((null? states) #f)
        ((goalp (car states)) (car states))
        (else (bfs (append (cdr states) 
			   (next (car states))) 
		   goalp 
		   next))))


;; State-space search using best-first search and means-ends analysis.
(define (sss-mea states goal operators)
  (best states
        (lambda (s) (subsetp2 goal s equal?))
        (lambda (s) (apply-operators operators s))
        (lambda (u v) (< (distance u goal) (distance v goal)))))


;; Prefer states with fewer differences from the goal.
(define (distance state goal) 
  (length (set-difference2 goal state equal?)))


(define (best states goalp next comparator)
  (cond ((null? states) #f)
        ((goalp (car states)) (car states))
        (else
         (best (sort2 (append (next (car states)) (cdr states))
                      comparator
                      (lambda (x) x))
               goalp
               next
               comparator))))


;; Operators for block-stacking problems.
(set! operators
  '((((on (? x) (? z)) (clear (? x)) (block (? z)))
     ((on (? x) table) (clear (? z)))
     ((on (? x) (? z))))
    (((on (? x) table) (clear (? x)) (clear (? y)) 
		       (not (equal? (? x) (? y))))
     ((on (? x) (? y)))
     ((on (? x) table) (clear (? y))))
    (((on (? x) (? z))
      (clear (? x))
      (clear (? y))
      (block (? z))
      (not (equal? (? x) (? y))))
     ((on (? x) (? y)) (clear (? z)))
     ((on (? x) (? z)) (clear (? y))))))



;; The goal for sample block-stacking problems.
(set! goal '((on a b) (on b c)))



;; The initial state for an instance of a block-stacking problem.
(set! state
  '((on c a)
    (on a table)
    (on b table)
    (clear c)
    (clear b)
    (block a)
    (block b)
    (block c)
    (not (equal? a b))
    (not (equal? b a))
    (not (equal? a c))
    (not (equal? c a))
    (not (equal? b c))
    (not (equal? c b))))


;; A seemingly easy problem that turns out to be hard for means-ends analsis.
(set! hard
  '((on c table)
    (on a table)
    (on b table)
    (clear a)
    (clear c)
    (clear b)
    (block a)
    (block b)
    (block c)
    (not (equal? a b))
    (not (equal? b a))
    (not (equal? a c))
    (not (equal? c a))
    (not (equal? b c))
    (not (equal? c b))))


;; The same state as EASY but with the conditions rearranged.
(set! easy
  '((on b table)
    (on c table)
    (on a table)
    (clear c)
    (clear a)
    (clear b)
    (block a)
    (block b)
    (block c)
    (not (equal? a b))
    (not (equal? b a))
    (not (equal? a c))
    (not (equal? c a))
    (not (equal? b c))
    (not (equal? c b))))


;; State-space search with best-first
(define (mea state)
  (reset)
  (display (sss-mea (list state) goal operators))
  (begin (newline) 
	 (display "Unifications: ") 
	 (display (get-unifications))))


;; State-space search with breadth-first and the above goal 
;; and operators.
(define (sss state)
  (reset)
  (display (sss-bfs (list state) goal operators))
  (begin (newline) 
	 (display "Unifications: ") 
	 (display (get-unifications))))


;; Here are some examples illustrating SSS and MEA.
;;
;; > (sss state)
;; Unifications: 10994
;; > (sss easy)
;; Unifications: 3186
;; > (sss hard)
;; Unifications: 8701
;; > (mea easy)
;; Unifications: 921
;; > (mea hard) ... this never returns


;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

