;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;; In this file we provide two implementations for computing the
;; posterior distributions of random variables in probabilistic networks
;; given evidence of the form of instantiated variables. First, we
;; implement an algorithm for exact computation of posterior
;; distributions by propagating causal and diagnostic support. Second, we
;; implement the likelihood-weighting algorithm for estimating posterior
;; distributions using stochastic simulation. We begin with the data
;; structures for representing probabilistic networks which will be used
;; in both implementations.

;; In the following implementations, we assume that all random variables
;; are boolean valued and that only boundary nodes are instantiated as
;; evidence. The value of a random variable is represented as either 0
;; (False) or 1 (True). An instance of the abstract data type NODE
;; consists of a name for the associated random variable, a list of
;; parents, a list of children, causal and diagnostic support vectors, a
;; number representing the probability that the associated random
;; variable is True given the evidence, a distribution providinc the
;; conditional probability of the random variable given its parents, a
;; value used to keep track of the value assigned to evidence and for
;; generating samples in stochastic simulation, and a pair of scores used
;; in stochastic simulation to keep track of the accumulated scores for
;; the values 0 and 1.

;; A node has a name, parents, children, distribution (prior 
;; conditional distribution function), causal and diagnostic support
;; (vectors), a probability (posterior distribution), a value if the 
;; corresponding random variable is instantiated as evidence, and a
;; pair of scores for use in stochastic approximation.
;; The name of a node is a symbol corresponding to the name of the
;; associated random variable.
;; In the following implementation, we assume boolean-valued random 
;; variables and evidence only at boundary nodes. 

(defun make-NODE (name) 
  (list name  ;; boolean random variable
	(list)   ;; parents
	(list)   ;; children
	(list)   ;; prior distribution
	(list)   ;; causal support
	(list)   ;; diagnostic support
	(list)   ;; probability given the evidence
	(list)   ;; boolean value for evidence
        (list))) ;; scores for stochastic simulation
(defun NODE-name (n) (first n))
(defun NODE-parents (n) (second n))
(defun NODE-children (n) (third n))
(defun NODE-distribution (n) (fourth n))
(defun NODE-causal (n m) (second (assoc m (fifth n))))
(defun NODE-diagnostic (n m) (second (assoc m (sixth n))))
(defun NODE-probability (n) (seventh n))
(defun NODE-value (n) (eighth n))
(defun NODE-scores (n) (ninth n))

;; All the elements of a node are settable except its name which is
;; specified when an instance of the type NODE is created. At the time
;; that the parents and children are set, association lists are created
;; that map nodes contributing support to causal and diagnostic support
;; vectors represented by lists of numbers. Causal and diagnostic support
;; vectors are accessed and set by specifying the node receiving the
;; support and the node contributing the support.

(defun set-NODE-parents (n p) 
  ;; Create an association list for accessing causal support.
  (setf (fifth n) (mapcar #'(lambda (m) (list m (list))) p))
  (setf (second n) p))
(defun set-NODE-children (n c) 
  ;; Create an association list for accessing diagnostic support.
  (setf (sixth n) (mapcar #'(lambda (m) (list m (list))) c))
  (setf (third n) c))
(defun set-NODE-distribution (n d) (setf (fourth n) d))
(defun set-NODE-causal (n m s) 
  (setf (second (assoc m (fifth n))) s))
(defun set-NODE-diagnostic (n m s) 
  (setf (second (assoc m (sixth n))) s))
(defun set-NODE-probability (n b) (setf (seventh n) b))
(defun set-NODE-value (n v) (setf (eighth n) v))
(defun set-NODE-scores (n s) (setf (ninth n) s))

;; A conditional distribution is represented as a nested list. There
;; is one level of nesting for each random variable. The following
;; is a conditional distribution for a variable with two parents. 
;;                       '(((0.2 0.8)
;;                          (0.3 0.7))
;;                         ((0.1 0.9)
;;                          (0.8 0.2)))
;; Since we represent truth values by 0 and 1 the listing of 
;; probabilities is reversed from how it appears in the text.

;; The following propagation code for applies to any singly-connected 
;; network. However, the code for computing causal and diagnostic 
;; support applies only to tree-structured networks. As an exercise,
;; you are asked to extend the code to handle the singly-connected case.

;; Propagate diagnostic support up from a sender node to its parent 
;; receiver. If the diagnostic support from the sender is unchanged or
;; the receiver is an instantiated root node, then quit,
;; otherwise update the diagnostic support from the sender,
;; propagate diagnostic support up to the parents, and 
;; propagate causal support down to each child except the sender. 

(defun up (recipient sender support)
  ;; Don't propagate support if it has not changed
  ;; or the recipient is an instantiated root node.
  (if (or (equal support (NODE-diagnostic recipient sender))
          (and (null (NODE-parents recipient))
               (NODE-value recipient)))
      nil
    (progn 
      (set-NODE-diagnostic recipient sender support)
      ;; Propagate support up to each parent.
      (mapc #'(lambda (n) 
                (up n recipient (diagnostic n recipient)))
            (NODE-parents recipient))
      ;; Propagate support down to each child but sender.
      (mapc #'(lambda (n) 
                (down n recipient (causal n recipient)))
            (remove sender (NODE-children recipient))))))

;; Propagate causal support down from sender to receiver. 
;; If the causal support from the sender is unchanged, then quit,
;; otherwise update the causal support from the sender.
;; If the receiver is an uninstantiated terminal node, then quit,
;; otherwise propagate diagnostic support up to each parent except 
;; the sender and propagate causal support down to each child. 
;; Note that using the path-based definition of independence we
;; could implement further optimizations terminating propagation.

(defun down (recipient sender support)
  (if (not (equal support (NODE-causal recipient sender)))
    (progn 
      (set-NODE-causal recipient sender support)
      ;; Don't propagate through an uninstantiated terminal node. 
      (if (or (NODE-children recipient)
	      (NODE-value recipient))
        (progn
          ;; Propagate support up to each parent except sender.
          (mapc #'(lambda (n) 
                    (up n recipient (diagnostic n recipient)))
                (remove sender (NODE-parents recipient)))
          ;; Propagate support down to each child.
          (mapc #'(lambda (n) 
                    (down n recipient (causal n recipient)))
                (NODE-children recipient)))))))

;; Compute the diagnostic support originating in n for recipient r.
;; We don't have to know the recipient for the tree-structured case,
;; since the recipient is always the node's single parent.

(defun diagnostic (r n)
  ;; If the node is terminal, then treat it specially.
  (if (null (NODE-children n))
    (if (null (NODE-value n)) '(1.0 1.0)
      (mapcar #'(lambda (v) (nth (NODE-value n)  
                                 (nth v (NODE-distribution n))))
              '(0 1)))
    (mapcar #'(lambda (v)
                (sum #'(lambda (u) 
                         (* (nth u (nth v (NODE-distribution n)))
                            (prod #'(lambda (m) 
                                      (nth u (NODE-diagnostic n m)))
                                  (NODE-children n))))
                     '(0 1)))
            '(0 1))))

;; Compute the causal support originating in n for the recipient r.

(defun causal (r n)
  ;; Handle instantiated root nodes specially.
  (if (and (null (NODE-parents n)) 
	   (NODE-value n))
    (nth (NODE-value n) (NODE-distribution r))
    (let ((terms
           (mapcar #'(lambda (v)
                       (* (nth v (causal-support n))
                          (prod #'(lambda (m) 
                                    (nth v (NODE-diagnostic n m)))
                                (remove r (NODE-children n)))))
                   '(0 1))) normalizer distribution)
      (setq normalizer (apply #'+ terms))
      (setq distribution
            (mapcar #'(lambda (term) (/ term normalizer)) terms))
      (mapcar #'(lambda (v)
                  (sum #'(lambda (u)
                           (* (nth v (nth u (NODE-distribution r)))
                              (nth u distribution)))
                       '(0 1)))
              '(0 1)))))

;; Fetch the causal support for a node. In the case of singly-connected
;; networks, note that each node has at most one parent. The causal 
;; support for a root node corresponds to the prior distribution.
(defun causal-support (node)
  (if (NODE-parents node)
    (NODE-causal node (first (NODE-parents node)))
    (NODE-distribution node)))

;; Compute the probability that the associated random variable is False 
;; given the evidence. The probability that the variable is True is 
;; just one minus the probability it is False. The function posterior 
;; is called when evidence propagation is complete.

(defun posterior (n)
  (if (NODE-value n)
    (if (= (NODE-value n) 1) 
      (set-NODE-probability n 0.0)
      (set-NODE-probability n 1.0))
    (let ((terms
           (mapcar #'(lambda (v)
                       (* (nth v (causal-support n))
                          (prod #'(lambda (m) 
                                    (nth v (NODE-diagnostic n m)))
                                (NODE-children n))))
                   '(0 1))))
      (set-NODE-probability n (/ (first terms) (apply #'+ terms))))))

(defun display-node (node)
  (format t "Node ~A has distribution (~4,2F, ~4,2F).~%" 
          (NODE-name node) 
          (NODE-probability node)
          (- 1 (NODE-probability node))))


;; Initialize the support for each node in the network assigning
;; diagnostic support to each node and propagating the priors from
;; the root nodes down. In the case of tree-structured networks, there 
;; will be exactly one root node.
(defun initialize-priors (nodes)
  ;; Initialize the diagnostic support for each node.
  (mapc #'(lambda (n) 
	    (mapc #'(lambda (m)
		      (set-NODE-diagnostic n m '(1.0 1.0)))
                  (NODE-children n)))
	nodes)
  ;; Eliminate any evidence in the network.
  (mapc #'(lambda (n) (set-NODE-value n (list))) nodes)
  ;; Propagate priors down from the root nodes. In the case of 
  ;; tree-structured networks there will be exactly one root node.
  (mapc #'(lambda (root)
            (mapc #'(lambda (n) (down n root (causal n root)))
                  (NODE-children root)))
        (mapcan #'(lambda (n) (if (null (NODE-parents n)) 
				  (list n ) 
				(list)))
                nodes)))

;; Propagate evidence down from root nodes and up from terminal nodes.
;; only boundary nodes are allowed to be assigned values as evidence.
(defun propagate (evidence)
  (let ((n (first evidence)) (v (second evidence)))
    (set-NODE-value n v)
    (cond ((null (NODE-parents n))
           ;; Propagate support down to each child.
           (mapc #'(lambda (m) 
                     (down m n (causal m n)))
                 (NODE-children n)))
          ((null (NODE-children n))
           ;; Propagate support up to each parent. 
           (mapc #'(lambda (m) 
                     (up m n (diagnostic m n)))
                 (NODE-parents n)))
          (t (princ "Illegal evidence node!")))))

(defun sum (function terms)
  (apply #'+ (mapcar function terms)))

(defun prod (function terms)
  (apply #'* (mapcar function terms)))

;; The description so far has focused on the details of propagating
;; evidence and computing causal and diagnostic support. Now the
;; description turns top down with a top-level function that computes 
;; the posterior distributions for all the nodes in a tree-structured
;; network, given evidence in the form of a list of pairs assigning
;; values to nodes.

(defun exact (nodes evidence)
  (initialize-priors nodes)
  (mapc #'propagate evidence)
  (mapc #'posterior nodes)
  (mapc #'display-node nodes) t)

;; 
;; Approximate inference using stochastic simulation
;; 

;; Now we provide the functions implementing the likelihood-weighting
;; algorithm. Since the likelihood-weighting algorithm applies to
;; arbitrary networks and in particular to networks with nodes that have
;; more than one parent, we will find it useful to define functions for
;; handling conditional probability distributions with more than one
;; conditioning variable.
;; 
;; Index takes the distribution of a node and a list of values. For our
;; purposes, the length of the list of values should be equal to one plus
;; the length of the parents of the node. Index returns the probability
;; given by the distribution for the case in which the values in the list
;; are assigned to the variables. The last value in the list is assigned
;; to the variable corresponding to the node and the earlier values in
;; the list are in the order determined by the parents of the node.
(defun index (distribution values)
  (if (null values) distribution 
    (index (nth (first values) distribution) 
	   (rest values))))

;; Compute the probability that the variable corresponding to a node
;; takes on a particular value given the values of its parents. Note the
;; way that mapcar is used to build a list of the values of the parents
;; and then the specified value is added to the end of this list using
;; nconc.
(defun probability (node value)
  (index (NODE-distribution node)
         (nconc (mapcar #'NODE-value (NODE-parents node))
                (list value))))

;; Estimate the posterior probabilities using likelihood weighting. 
;; Evidence is provided as a list of pairs assigning values to nodes.
(defun approximate (nodes evidence iterations)
  (initialize-scores nodes)
  (dotimes (i iterations)
    (initialize-values nodes evidence)
    (simulate nodes)
    (update-scores nodes evidence))
  (set-evidence nodes evidence)
  (mapc #'estimate nodes)
  (mapc #'display-node nodes) t)

;; The scores for each node are initialized to a pair of zeros.
(defun initialize-scores (nodes)
  (mapc #'(lambda (n) (set-NODE-scores n (list 0 0))) nodes))

;; The value of each node is initialized to nil unless the node is 
;; assigned a value by the evidence and the node has no parents. 
;; Evidence nodes with parents are assigned values by stochastic 
;; simulation but the resulting assignments are weighted.
(defun initialize-values (nodes evidence)
  (mapc #'(lambda (n) 
            (let ((pair (assoc n evidence)))
              (if (and pair (null (NODE-parents n)))
		(set-NODE-value n (second pair))
                (set-NODE-value n (list)))))
        nodes))

;; Clean up the values of nodes after stochastic simulation. 
;; Only evidence nodes should have values after the clean up.
(defun set-evidence (nodes evidence)
  (mapc #'(lambda (node)
	    (let ((pair (assoc node evidence)))
	      (if pair
		  (set-NODE-value node (second pair))
		  (set-NODE-value node (list)))))
        nodes))

;; Update the scores of each node using likelihood weighting.
(defun update-scores (nodes evidence)
  (let ((w (sample-weight evidence)))
    (mapc #'(lambda (n)
              (let ((s (NODE-scores n)))
                (if (= 0 (NODE-value n))
                  (setf (first s) (+ w (first s)))
                  (setf (second s) (+ w (second s))))))
          nodes)))

;; Compute the likelihood weighting assigned to the evidence 
;; given the sample.
(defun sample-weight (evidence)
  (apply #'* (mapcar #'(lambda (assignment)
                         (apply #'probability assignment))
                     evidence)))

;; Simulate by propagating forward through the network.
(defun simulate (nodes)
  (cond ((null nodes) nil)
        ;; Ignore a node if it already has a value or
        ;; if it has one or more parents without values.
        ((or (NODE-value (first nodes))
             (some #'(lambda (n) (null (NODE-value n))) 
                   (NODE-parents (first nodes))))
         (simulate (rest nodes)))
        (t (let ((number (random 1.0)) (n (first nodes)))
             (if (< number (probability n 0))
               (set-NODE-value n 0) (set-NODE-value n 1))
             (simulate (append (rest nodes)
                               (NODE-children n)))))))

;; Estimate the posterior probability for the case in which the node 
;; is assigned False. The probability for the case in which the node is
;; assigned True is just one minus the probability it is False.
(defun estimate (n)
  (if (NODE-value n)
    ;; No sense in estimating evidence nodes. 
    (if (eq (NODE-value n) 1) 
      (set-NODE-probability n 0.0) (set-NODE-probability n 1.0))
    (set-NODE-probability n (/ (first (NODE-scores n))
                               (+ (first (NODE-scores n)) 
                                  (second (NODE-scores n)))))))

;;
;; Examples
;;

;; Simplified version of the buggy Lisp code example
(let ((H (make-NODE 'H))     ;;                   H       B
      (B (make-NODE 'B))     ;;                     \   /
      (L (make-NODE 'L))     ;;                       L
      nodes evidence)
  (set-NODE-children H (list L))
  (set-NODE-children B (list L))
  (set-NODE-parents L (list H B))
  (set-NODE-distribution H '(0.7 0.3))
  (set-NODE-distribution B '(0.2 0.8))
  (set-NODE-distribution L '(((0.2 0.8)
                              (0.3 0.7))
                             ((0.1 0.9)
                              (0.8 0.2))))
  (setq nodes (list H L B)
        evidence (list (list H 0)))
  (defun buggy-approximate (n) (approximate nodes evidence n)))

;; Sonya example in the text
(let ((O (make-NODE 'O))     ;;                        O
      (C (make-NODE 'C))     ;;                      /   \
      (L (make-NODE 'L))     ;;                    L       C
      nodes evidence)
  (set-NODE-children O (list C L))
  (set-NODE-parents C (list O))
  (set-NODE-parents L (list O))
  (set-NODE-distribution O '(0.6 0.4))
  (set-NODE-distribution C '((0.7 0.3) 
                             (0.2 0.8)))
  (set-NODE-distribution L '((0.9 0.1)
                             (0.4 0.6)))
  (setq nodes (list O C L)
        evidence (list (list L 1)))
  (defun sonya-exact () (exact nodes evidence))
  (defun sonya-approximate (n) (approximate nodes evidence n)))

;; Variation on the Sonya example.
(let ((O (make-NODE 'O))     ;;                        O
      (C (make-NODE 'C))     ;;                      /   \
      (L (make-NODE 'L))     ;;                    L       C
      (Q (make-NODE 'Q))     ;;                      \
      nodes evidence)        ;;                        Q
  (set-NODE-children O (list C L))
  (set-NODE-parents C (list O))
  (set-NODE-parents L (list O))
  (set-NODE-children L (list Q))
  (set-NODE-parents Q (list L))
  (set-NODE-distribution O '(0.6 0.4))
  (set-NODE-distribution C '((0.7 0.3) 
                             (0.2 0.8)))
  (set-NODE-distribution L '((0.9 0.1)
                             (0.4 0.6)))
  (set-NODE-distribution Q '((0.9 0.1)
                             (0.4 0.6)))
  (setq nodes (list O C L Q)
        evidence (list (list Q 1)))
  (defun var-sonya-exact () 
    (exact nodes evidence))
  (defun var-sonya-approximate (n) 
    (approximate nodes evidence n)))

;; Somewhat more complex singly-connected network
(let ((A (make-NODE 'A))     ;;
      (B (make-NODE 'B))     ;;               A       B
      (C (make-NODE 'C))     ;;             /   \   /   
      (D (make-NODE 'D))     ;;           C       D       E
      (E (make-NODE 'E))     ;;                 /   \   /
      (F (make-NODE 'F))     ;;               F       G
      (G (make-NODE 'G))     ;;
      nodes evidence)
  (set-NODE-children A (list C D))
  (set-NODE-children B (list D))
  (set-NODE-children D (list F G))
  (set-NODE-children E (list G))
  (set-NODE-parents C (list A))
  (set-NODE-parents D (list A B))
  (set-NODE-parents F (list D))
  (set-NODE-parents G (list D E) )
  (set-NODE-distribution A '(0.6 0.4))
  (set-NODE-distribution B '(0.6 0.4))
  (set-NODE-distribution C '((0.7 0.3) 
                             (0.2 0.8)))
  (set-NODE-distribution D '(((0.2 0.8)
                              (0.3 0.7))
                             ((0.1 0.9)
                              (0.8 0.2))))
  (set-NODE-distribution E '(0.6 0.4))
  (set-NODE-distribution F '((0.7 0.3) 
                             (0.4 0.6)))
  (set-NODE-distribution G '(((0.2 0.8)
                              (0.3 0.7))
                             ((0.1 0.9)
                              (0.8 0.2))))
  (setq nodes (list A B C D E F G)
        evidence (list (list F 1)))
  (defun singly-approximate (n) 
    (approximate nodes evidence n))) 

;; Simple test
(defun test ()
  (sonya-exact)
  (sonya-approximate 100))


;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

