;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;; In this file we provide two implementations for computing the
;; posterior distributions of random variables in probabilistic networks
;; given evidence of the form of instantiated variables. First, we
;; implement an algorithm for exact computation of posterior
;; distributions by propagating causal and diagnostic support. Second, we
;; implement the likelihood-weighting algorithm for estimating posterior
;; distributions using stochastic simulation. We begin with the data
;; structures for representing probabilistic networks which will be used
;; in both implementations.

;; In the following implementations, we assume that all random variables
;; are boolean valued and that only boundary nodes are instantiated as
;; evidence. The value of a random variable is represented as either 0
;; (False) or 1 (True). An instance of the abstract data type NODE
;; consists of a name for the associated random variable, a list of
;; parents, a list of children, causal and diagnostic support vectors, a
;; number representing the probability that the associated random
;; variable is True given the evidence, a distribution providinc the
;; conditional probability of the random variable given its parents, a
;; value used to keep track of the value assigned to evidence and for
;; generating samples in stochastic simulation, and a pair of scores used
;; in stochastic simulation to keep track of the accumulated scores for
;; the values 0 and 1.

;; A node has a name, parents, children, distribution (prior 
;; conditional distribution function), causal and diagnostic support
;; (vectors), a probability (posterior distribution), a value if the 
;; corresponding random variable is instantiated as evidence, and a
;; pair of scores for use in stochastic approximation.
;; The name of a node is a symbol corresponding to the name of the
;; associated random variable.
;; In the following implementation, we assume boolean-valued random 
;; variables and evidence only at boundary nodes. 

(define (make-node name)
  (list name (list) (list) (list) (list) (list) (list) (list) (list)))
 ;; scores for stochastic simulation
(define (node-name n) (first n))

(define (node-parents n) (second n))

(define (node-children n) (third n))

(define (node-distribution n) (fourth n))

(define (node-causal n m) 
  (second (assoc2 m (fifth n) eq? (lambda (x) x))))

(define (node-diagnostic n m) 
  (second (assoc2 m (sixth n) eq? (lambda (x) x))))

(define (node-probability n) (seventh n))

(define (node-value n) (eighth n))

(define (node-scores n) (ninth n))


;; All the elements of a node are settable except its name which is
;; specified when an instance of the type NODE is created. At the time
;; that the parents and children are set, association lists are created
;; that map nodes contributing support to causal and diagnostic support
;; vectors represented by lists of numbers. Causal and diagnostic support
;; vectors are accessed and set by specifying the node receiving the
;; support and the node contributing the support.

(define (set-node-parents n p)
  (set-car! (list-tail n 4) (map (lambda (m) (list m (list))) p))
  (set-car! (list-tail n 1) p))

(define (set-node-children n c)
  (set-car! (list-tail n 5) (map (lambda (m) (list m (list))) c))
  (set-car! (list-tail n 2) c))

(define (set-node-distribution n d) (set-car! (list-tail n 3) d))

(define (set-node-causal n m s)
  (set-car! (list-tail (assoc2 m (fifth n) eq? (lambda (x) x)) 1) s))

(define (set-node-diagnostic n m s)
  (set-car! (list-tail (assoc2 m (sixth n) eq? (lambda (x) x)) 1) s))

(define (set-node-probability n b) (set-car! (list-tail n 6) b))

(define (set-node-value n v) (set-car! (list-tail n 7) v))

(define (set-node-scores n s) (set-car! (list-tail n 8) s))


;; A conditional distribution is represented as a nested list. There
;; is one level of nesting for each random variable. The following
;; is a conditional distribution for a variable with two parents. 
;;                       '(((0.2 0.8)
;;                          (0.3 0.7))
;;                         ((0.1 0.9)
;;                          (0.8 0.2)))
;; Since we represent truth values by 0 and 1 the listing of 
;; probabilities is reversed from how it appears in the text.

;; The following propagation code for applies to any singly-connected 
;; network. However, the code for computing causal and diagnostic 
;; support applies only to tree-structured networks. As an exercise,
;; you are asked to extend the code to handle the singly-connected case.

;; Propagate diagnostic support up from a sender node to its parent 
;; receiver. If the diagnostic support from the sender is unchanged or
;; the receiver is an instantiated root node, then quit,
;; otherwise update the diagnostic support from the sender,
;; propagate diagnostic support up to the parents, and 
;; propagate causal support down to each child except the sender. 

(define (up recipient sender support)
  (if (or (equal? support (node-diagnostic recipient sender))
          (and (null? (node-parents recipient))
               (not (null? (node-value recipient)))))
    #f
    (begin
      (set-node-diagnostic recipient sender support)
      (for-each
        (lambda (n) (up n recipient (diagnostic n recipient)))
        (node-parents recipient))
      (for-each
        (lambda (n) (down n recipient (causal n recipient)))
        (remove2 sender (node-children recipient) eq? (lambda (x) x))))))


;; Propagate causal support down from sender to receiver. 
;; If the causal support from the sender is unchanged, then quit,
;; otherwise update the causal support from the sender.
;; If the receiver is an uninstantiated terminal node, then quit,
;; otherwise propagate diagnostic support up to each parent except 
;; the sender and propagate causal support down to each child. 
;; Note that using the path-based definition of independence we
;; could implement further optimizations terminating propagation.

(define (down recipient sender support)
  (if (not (equal? support (node-causal recipient sender)))
    (begin
      (set-node-causal recipient sender support)
      (if (or (not (null? (node-children recipient)))
              (not (null? (node-value recipient))))
        (begin
          (for-each
            (lambda (n) (up n recipient (diagnostic n recipient)))
            (remove2 sender (node-parents recipient) eq? (lambda (x) x)))
          (for-each
            (lambda (n) (down n recipient (causal n recipient)))
            (node-children recipient)))))))


;; Compute the diagnostic support originating in n for recipient r.
;; We don't have to know the recipient for the tree-structured case,
;; since the recipient is always the node's single parent.

(define (diagnostic r n)
  (if (null? (node-children n))
    (if (null? (node-value n))
      '(1. 1.)
      (map (lambda (v) 
	     (nth (node-value n) (nth v (node-distribution n))))
           '(0 1)))
    (map (lambda (v)
           (sum (lambda (u)
                  (* (nth u (nth v (node-distribution n)))
                     (prod (lambda (m) (nth u (node-diagnostic n m)))
                           (node-children n))))
                '(0 1)))
         '(0 1))))


;; Compute the causal support originating in n for the recipient r.

(define (causal r n)
  (if (and (null? (node-parents n)) (not (null? (node-value n))))
    (nth (node-value n) (node-distribution r))
    (let ((terms (map (lambda (v)
                        (* (nth v (causal-support n))
                           (prod (lambda (m) 
				   (nth v (node-diagnostic n m)))
                                 (remove2 r
                                          (node-children n)
                                          eq?
                                          (lambda (x) x)))))
                      '(0 1)))
          (normalizer #f)
          (distribution #f))
      (set! normalizer (apply + terms))
      (set! distribution (map (lambda (term) (/ term normalizer)) terms))
      (map (lambda (v)
             (sum (lambda (u)
                    (* (nth v (nth u (node-distribution r)))
                       (nth u distribution)))
                  '(0 1)))
           '(0 1)))))


;; Fetch the causal support for a node. In the case of singly-connected
;; networks, note that each node has at most one parent. The causal 
;; support for a root node corresponds to the prior distribution.
(define (causal-support node)
  (if (not (null? (node-parents node)))
    (node-causal node (first (node-parents node)))
    (node-distribution node)))


;; Compute the probability that the associated random variable is False 
;; given the evidence. The probability that the variable is True is 
;; just one minus the probability it is False. The function posterior 
;; is called when evidence propagation is complete.

(define (posterior n)
  (if (not (null? (node-value n)))
    (if (= (node-value n) 1)
      (set-node-probability n 0.)
      (set-node-probability n 1.))
    (let ((terms (map (lambda (v)
                        (* (nth v (causal-support n))
                           (prod (lambda (m) 
				   (nth v (node-diagnostic n m)))
                                 (node-children n))))
                      '(0 1))))
      (set-node-probability n (/ (first terms) (apply + terms))))))


(define (display-node node)
  (begin
    (display "Node ")
    (display (node-name node))
    (display " has distribution (")
    (display (node-probability node))
    (display ", ")
    (display (- 1 (node-probability node)))
    (display ").")
    (newline)))



;; Initialize the support for each node in the network assigning
;; diagnostic support to each node and propagating the priors from
;; the root nodes down. In the case of tree-structured networks, there 
;; will be exactly one root node.
(define (initialize-priors nodes)
  (for-each
    (lambda (n)
      (for-each
        (lambda (m) (set-node-diagnostic n m '(1. 1.)))
        (node-children n)))
    nodes)
  (for-each (lambda (n) (set-node-value n (list))) nodes)
  (for-each
    (lambda (root)
      (for-each
        (lambda (n) (down n root (causal n root)))
        (node-children root)))
    (apply append
           (map (lambda (n) 
		  (if (null? (node-parents n)) (list n) (list)))
                nodes))))


;; Propagate evidence down from root nodes and up from terminal nodes.
;; only boundary nodes are allowed to be assigned values as evidence.
(define (propagate evidence)
  (let ((n (first evidence)) (v (second evidence)))
    (set-node-value n v)
    (cond ((null? (node-parents n))
           (for-each (lambda (m) 
		       (down m n (causal m n))) (node-children n)))
          ((null? (node-children n))
           (for-each (lambda (m) 
		       (up m n (diagnostic m n))) (node-parents n)))
          (else (display "Illegal evidence node!")))))


(define (sum function terms) (apply + (map function terms)))


(define (prod function terms) (apply * (map function terms)))


;; The description so far has focused on the details of propagating
;; evidence and computing causal and diagnostic support. Now the
;; description turns top down with a top-level function that computes 
;; the posterior distributions for all the nodes in a tree-structured
;; network, given evidence in the form of a list of pairs assigning
;; values to nodes.

(define (exact nodes evidence)
  (initialize-priors nodes)
  (for-each propagate evidence)
  (for-each posterior nodes)
  (for-each display-node nodes)
  #t)


;; 
;; Approximate inference using stochastic simulation
;; 

;; Now we provide the functions implementing the likelihood-weighting
;; algorithm. Since the likelihood-weighting algorithm applies to
;; arbitrary networks and in particular to networks with nodes that have
;; more than one parent, we will find it useful to define functions for
;; handling conditional probability distributions with more than one
;; conditioning variable.
;; 
;; Index takes the distribution of a node and a list of values. For our
;; purposes, the length of the list of values should be equal to one plus
;; the length of the parents of the node. Index returns the probability
;; given by the distribution for the case in which the values in the list
;; are assigned to the variables. The last value in the list is assigned
;; to the variable corresponding to the node and the earlier values in
;; the list are in the order determined by the parents of the node.
(define (index distribution values)
  (if (null? values)
    distribution
    (index (nth (first values) distribution) (rest values))))


;; Compute the probability that the variable corresponding to a node
;; takes on a particular value given the values of its parents. Note the
;; way that mapcar is used to build a list of the values of the parents
;; and then the specified value is added to the end of this list using
;; nconc.
(define (probability node value)
  (index (node-distribution node)
         (append (map node-value (node-parents node)) (list value))))


;; Estimate the posterior probabilities using likelihood weighting. 
;; Evidence is provided as a list of pairs assigning values to nodes.
(define (approximate nodes evidence iterations)
  (initialize-scores nodes)
  (do ((i 1 (+ 1 i)))
      ((> i iterations))
    (initialize-values nodes evidence)
    (simulate nodes)
    (update-scores nodes evidence))
  (set-evidence nodes evidence)
  (for-each estimate nodes)
  (for-each display-node nodes)
  #t)


;; The scores for each node are initialized to a pair of zeros.
(define (initialize-scores nodes)
  (for-each (lambda (n) (set-node-scores n (list 0 0))) nodes))


;; The value of each node is initialized to nil unless the node is 
;; assigned a value by the evidence and the node has no parents. 
;; Evidence nodes with parents are assigned values by stochastic 
;; simulation but the resulting assignments are weighted.
(define (initialize-values nodes evidence)
  (for-each
    (lambda (n)
      (let ((pair (assoc2 n evidence eq? (lambda (x) x))))
        (if (and pair (null? (node-parents n)))
          (set-node-value n (second pair))
          (set-node-value n (list)))))
    nodes))


;; Clean up the values of nodes after stochastic simulation. 
;; Only evidence nodes should have values after the clean up.
(define (set-evidence nodes evidence)
  (for-each
    (lambda (node)
      (let ((pair (assoc2 node evidence eq? (lambda (x) x))))
        (if pair
          (set-node-value node (second pair))
          (set-node-value node (list)))))
    nodes))


;; Update the scores of each node using likelihood weighting.
(define (update-scores nodes evidence)
  (let ((w (sample-weight evidence)))
    (for-each
      (lambda (n)
        (let ((s (node-scores n)))
          (if (= 0 (node-value n))
            (set-car! s (+ w (first s)))
            (set-car! (list-tail s 1) (+ w (second s))))))
      nodes)))


;; Compute the likelihood weighting assigned to the evidence 
;; given the sample.
(define (sample-weight evidence)
  (apply *
         (map (lambda (assignment) 
		(apply probability assignment)) 
	      evidence)))


;; Simulate by propagating forward through the network.
(define (simulate nodes)
  (cond ((null? nodes) #f)
        ((or (not (null? (node-value (first nodes))))
             (some2 (lambda (n) (null? (node-value n)))
                    (node-parents (first nodes))))
         (simulate (rest nodes)))
        (else
         (let ((number (random 1.)) (n (first nodes)))
           (if (< number (probability n 0))
             (set-node-value n 0)
             (set-node-value n 1))
           (simulate (append (rest nodes) (node-children n)))))))


;; Estimate the posterior probability for the case in which the node 
;; is assigned False. The probability for the case in which the node is
;; assigned True is just one minus the probability it is False.
(define (estimate n)
  (if (not (null? (node-value n)))
    (if (eq? (node-value n) 1)
      (set-node-probability n 0.)
      (set-node-probability n 1.))
    (set-node-probability
      n
      (/ (first (node-scores n))
         (+ (first (node-scores n)) (second (node-scores n)))))))


;;
;; Examples
;;

;; Simplified version of the buggy Lisp code example
(let ((h (make-node 'h))
      (b (make-node 'b))
      (l (make-node 'l))
      (nodes #f)
      (evidence #f))
  (set-node-children h (list l))
  (set-node-children b (list l))
  (set-node-parents l (list h b))
  (set-node-distribution h '(.7 .3))
  (set-node-distribution b '(.2 .8))
  (set-node-distribution l '(((.2 .8) (.3 .7)) ((.1 .9) (.8 .2))))
  (set! nodes (list h l b)) 
  (set! evidence (list (list h 0)))
  (set! buggy-approximate (lambda (n) (approximate nodes evidence n))))


;; Sonya example in the text
(let ((o (make-node 'o))
      (c (make-node 'c))
      (l (make-node 'l))
      (nodes #f)
      (evidence #f))
  (set-node-children o (list c l))
  (set-node-parents c (list o))
  (set-node-parents l (list o))
  (set-node-distribution o '(.6 .4))
  (set-node-distribution c '((.7 .3) (.2 .8)))
  (set-node-distribution l '((.9 .1) (.4 .6)))
  (set! nodes (list o c l)) 
  (set! evidence (list (list l 1)))
  (set! sonya-exact (lambda () (exact nodes evidence)))
  (set! sonya-approximate (lambda (n) (approximate nodes evidence n))))


;; Variation on the Sonya example.
(let ((o (make-node 'o))
      (c (make-node 'c))
      (l (make-node 'l))
      (q (make-node 'q))
      (nodes #f)
      (evidence #f))
  (set-node-children o (list c l))
  (set-node-parents c (list o))
  (set-node-parents l (list o))
  (set-node-children l (list q))
  (set-node-parents q (list l))
  (set-node-distribution o '(.6 .4))
  (set-node-distribution c '((.7 .3) (.2 .8)))
  (set-node-distribution l '((.9 .1) (.4 .6)))
  (set-node-distribution q '((.9 .1) (.4 .6)))
  (set! nodes (list o c l q)) 
  (set! evidence (list (list q 1)))
  (set! var-sonya-exact (lambda () (exact nodes evidence)))
  (set! var-sonya-approximate 
	(lambda (n) (approximate nodes evidence n))))


;; Somewhat more complex singly-connected network
(let ((a (make-node 'a))
      (b (make-node 'b))
      (c (make-node 'c))
      (d (make-node 'd))
      (e (make-node 'e))
      (f (make-node 'f))
      (g (make-node 'g))
      (nodes #f)
      (evidence #f))
  (set-node-children a (list c d))
  (set-node-children b (list d))
  (set-node-children d (list f g))
  (set-node-children e (list g))
  (set-node-parents c (list a))
  (set-node-parents d (list a b))
  (set-node-parents f (list d))
  (set-node-parents g (list d e))
  (set-node-distribution a '(.6 .4))
  (set-node-distribution b '(.6 .4))
  (set-node-distribution c '((.7 .3) (.2 .8)))
  (set-node-distribution d '(((.2 .8) (.3 .7)) ((.1 .9) (.8 .2))))
  (set-node-distribution e '(.6 .4))
  (set-node-distribution f '((.7 .3) (.4 .6)))
  (set-node-distribution g '(((.2 .8) (.3 .7)) ((.1 .9) (.8 .2))))
  (set! nodes (list a b c d e f g)) 
  (set! evidence (list (list f 1)))
  (set! singly-approximate 
	(lambda (n) (approximate nodes evidence n))))
 

;; Simple test
(define (test) (sonya-exact) (sonya-approximate 100))



;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

