;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information
;;******************************************
;; Solution to Polyhedral Labeling Exercise:
;;******************************************

(load "waltz")

;;*******************************************************************
;;  Waltz filtering is not guaranteed to yield a unique 
;;  interpretation. Write a function that takes
;;  the final result produced by waltz filtering
;;  and then performs a search for all of the possible solutions.
;;  Call this function search-solutions, which takes as an argument
;;  a diagram produced by waltz filtering and returns a list of all the 
;;  consistent interpretations.  This function must do both 
;;  searching and propagation.  That is, trying all labelings for
;;  one ambiguous vertex, and propagating for each choice.
;;  If there is no ambiguous vertex, then return the diagram.
;;  If there is one, make copies of the diagram trying each of
;;  the possible labelings.  Propagate constraints and append
;;  all the solutions together.
;;*******************************************************************

;; Apply relaxation and then find all unambiguous diagrams 
;; that are consistent with the waltz filtered vertex labels.
(defun filter-and-generate-solutions (diagram)
  (format t "~%The initial diagram is:")
  (show-DIAGRAM diagram)
  (mapc #'waltz (DIAGRAM-vertices diagram))
  (format t "~%After filtering the diagram is:")
  (show-DIAGRAM diagram)
  (let* ((solutions (if (impossible-diagram diagram)
                        nil
                        (search-solutions diagram)))
         (n (length solutions)))
    (unless (= n 1)
      (format t "~%There are ~r solutions:" n)
      (mapc #'show-DIAGRAM solutions)))
  (format t "~%Done!"))

;; An impossible diagram is one with an impossible vertex.
;; Given a diagram, returns T if there is at least one vertex
;; that has no labelings possible for it.
(defun impossible-diagram (diagram)
  (some #'impossible-vertex (DIAGRAM-vertices diagram)))

;; Try all labelings for one ambiguous vertex, and propagate.
(defun search-solutions (diagram)
  ;; If there is no ambiguous vertex, return the diagram.
  ;; If there is one, make copies of the diagram trying each of 
  ;; the possible labelings.  Propagate constraints and append
  ;; all the solutions together.
  (let ((v (find-if #'ambiguous-vertex
                    (DIAGRAM-vertices diagram))))
    (if (null v)
        (list diagram)
        (mapcan 
          #'(lambda (v-labeling)
              (let* ((new-diagram (copy-DIAGRAM diagram))
                     (new-v (find-vertex (VERTEX-name v) new-diagram)))
                (set-VERTEX-labelings new-v (list v-labeling))
                (if (waltz new-v)
                    (search-solutions new-diagram)
                    nil)))
          (VERTEX-labelings v)))))

;;; Given a vertex, returns T if there is more than one
;;; current possible labeling remaining for this vertex; 
;;; otherwise, returns NIL.
(defun ambiguous-vertex (vertex)
  (> (VERTEX-number-of-labelings vertex) 1))


;;***************
;; Test function:
;;***************

(defun test-filter-and-generate-solutions ()
  (filter-and-generate-solutions (copy-DIAGRAM cube))
  (filter-and-generate-solutions (ground (copy-DIAGRAM cube) 'g 'd)))


;; Copyright 1994, Brown University, Providence, RI
;; 
;; Permission to use and modify this software and its documentation for any
;; purpose other than its incorporation into a commercial product is hereby
;; granted without fee.  Permission to copy and distribute this software and its
;; documentation only for non-commercial use is also granted without fee,
;; provided, however, that the above copyright notice appear in all copies, that
;; both that copyright notice and this permission notice appear in supporting
;; documentation, that the name of Brown University not be used in advertising or
;; publicity pertaining to distribution of the software without specific, written
;; prior permission, and that the person doing the distribution notify Brown
;; University of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this software for
;; any purpose.  It is provided "as is" without express or implied warranty.
;; Brown University requests notification of any modifications to this software
;; or its documentation.
;; 
;; Send the following redistribution information:
;; 
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;; 
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;; 
;; 		or
;; 
;; 	brusd@cs.brown.edu
;; 
;; We will acknowledge all electronic notifications.
