;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;;****************************
;; Labeling Polyhedral Scenes:
;;****************************

;;;*********************
;;; Abstract Data Types:
;;;*********************

;; A vertex consists of a name, type (one of ELL, FORK, ARROW, or TEE), 
;; list of neighboring vertices, and list of labelings, where a labeling
;; is a list of labels and a label is one of +, -, LEFT, or RIGHT.
(define (make-vertex name type labelings) 
  (list name type (list) labelings))

(define (vertex-name v) (first v))

(define (vertex-type v) (second v))

(define (vertex-neighbors v) (third v))

(define (vertex-labelings v) (fourth v))

(define (vertex-number-of-labelings v) (length (vertex-labelings v)))

(define (set-vertex-name v name) (set-car! v name))

(define (set-vertex-type v type) (set-car! (list-tail v 1) type))

(define (set-vertex-neighbors v neighbors)
  (set-car! (list-tail v 2) neighbors))

(define (set-vertex-labelings v labelings)
  (set-car! (list-tail v 3) labelings))

(define (copy-vertex v) (copy-list v))


;; A diagram is just a list of vertices to encode the information from 
;; the line drawings. Lines are implicit in the list of vertices and the
;; information regarding neighboring vertices. The function copy-DIAGRAM
;; allows us to make multiple copies of the same diagram to generate
;; multiple interpretations of an ambiguous line drawing.
(define (make-diagram vertices) (list vertices))

(define (diagram-vertices d) (first d))

(define (copy-diagram d)
  (let ((new (make-diagram (map copy-vertex (diagram-vertices d)))))
    (for-each
      (lambda (vertex)
        (set-vertex-neighbors
          vertex
          (map (lambda (neighbor) 
		 (find-vertex (vertex-name neighbor) new))
               (vertex-neighbors vertex))))
      (diagram-vertices new))
    new))


;;************************************************************
;; Functions for contructing diagrams from vertex descriptors:
;;************************************************************

;; A vertex descriptor is a list consisting of a name and type for 
;; a vertex along with two additional vertex names for an ELL type
;; or three additional vertex names for ARROW, FORK, or TEE types.
;; For example, (A ARROW B C D) is a vertex descriptor for a vertex
;; named A of type ARROW connected to vertices named B C and D. 
;; As another example, (E ELL F G) is a vertex descriptor for a vertex
;; named E of type ELL connected to vertices named F and G.

;; Construct a new vertex given a vertex descriptor.
(define (construct-vertex vertex-descriptor)
  (make-vertex
    (first vertex-descriptor)
    (second vertex-descriptor)
    (possible-labelings (second vertex-descriptor))))


;; Construct a new diagram given a list of vertex descriptors.
(define (construct-diagram vertex-descriptors)
  (let ((diagram (make-diagram (map construct-vertex 
				    vertex-descriptors))))
    (for-each
      (lambda (d)
        (set-vertex-neighbors
          (find-vertex (first d) diagram)
          (map (lambda (n) (find-vertex n diagram)) (v-d-neighbors d))))
      vertex-descriptors)
    diagram))


;;; Given a vertex descriptor, returns a list of the names of
;;; neighboring vertices.
(define (v-d-neighbors vertex-descriptor) 
  (rest (rest vertex-descriptor)))


;; Find the vertex in the given diagram with the given name.
(define (find-vertex name diagram)
  (let ((result-of-find
          (member2 name (diagram-vertices diagram) eq? vertex-name)))
    (and result-of-find (car result-of-find))))


;; Indicate that the line between the two vertices is adjacent to 
;; the ground (floor) by labeling the line - thereby indicating 
;; that the line is concave.
(define (ground diagram vertex-a vertex-b)
  (let* ((v (find-vertex vertex-a diagram))
	 (u (find-vertex vertex-b diagram)) 
	 (i (index u (vertex-neighbors v))))
    (set-vertex-labelings
      v
      (apply append
             (map (lambda (l) 
		    (if (eq? (nth i l) '-) 
			(list l)
			(list)))
                  (vertex-labelings v))))
    diagram))


;; Compute the index of an item in a list. 
(define (index item list) (aux-index item list 0))

(define (aux-index item list index)
  (if (null? list)
    (display "Item not in list!")
    (if (eq? item (car list)) 
	index 
	(aux-index item (rest list) (+ 1 index)))))


;;************************************************
;; Functions for displaying vertices and diagrams:
;;************************************************

;; Display a vertex.
(define (show-vertex vertex)
  (begin
    (newline)
    (display "  ")
    (display (vertex-name vertex))
    (display "/")
    (display (vertex-number-of-labelings vertex))
    (display " ")
    (display (vertex-type vertex))
    (display ":"))
  (for-each
    (lambda (neighbor labels)
      (begin
        (display " ")
        (display (vertex-name vertex))
        (display (vertex-name neighbor))
        (display "=")
        (display labels)))
    (vertex-neighbors vertex)
    (matrix-transpose (vertex-labelings vertex))))


;; Turn a matrix on its side.
(define (matrix-transpose matrix)
  (if (null? matrix)
      (list)
      (let ((m (length (car matrix))))
	(do ((i 0 (+ i 1)) (result (list)))
	    ((= i m) (reverse result))
	  (set! result (cons (map (lambda (x) (nth i x)) 
				  matrix)
			     result))))))

;; Display a diagram.
(define (show-diagram diagram)
  (for-each show-vertex (diagram-vertices diagram))
  (begin
    (newline)
    (display "For ")
    (display (reduce *
                     (map vertex-number-of-labelings
                          (diagram-vertices diagram))))
    (display " interpretation(s).")))


;;***************************************
;; Functions to label or check labelings:
;;***************************************

;; There are four vertex types in the trihedral-vertex blocks world
;; used in the examples in this section, ELL, FORK, TEE, and ARROW. 
;; There are two labels for nonoccluding lines in the trihedral-vertex 
;; blocks world, + indicates a nonoccluding convex line and - indicates 
;; a nonoccluding concave line.
;; Occluding lines are labeled with respect to a vertex. 
;; RIGHT indicates an occluding line with the occluding matter to
;; the right of the line as you face outward from the vertex.
;; LEFT indicates an occluding line with the occluding matter to
;; the left of the line as you face outward from the vertex.
;; If a line is labeled RIGHT with respect to the vertex at one end 
;; of the line, then it is labeled LEFT with respect to the vertex 
;; at the other of the line. 

;; The function possible-labelings returns a list of the possible 
;; labelings in the junction catalog for a given vertex type. 
(define (possible-labelings vertex-type)
  (cond 
   ((eq? vertex-type 'ell)
    '((right left) (left right) (+ right) (left +) (- left) (right -)))
   ((eq? vertex-type 'fork)
    '((+ + +) (- - -) (left right -) (- left right) (right - left)))
   ((eq? vertex-type 'tee)
    '((right left +) (right left -) (right left left)(right left right)))
   ((eq? vertex-type 'arrow) '((left right +) (- - +) (+ + -)))
   (else (display "Unfamiliar vertex type!"))))


;; The function impossible-vertex returns #t if the vertex 
;; has no possible labeling left and #f otherwise.
(define (impossible-vertex vertex) (null? (vertex-labelings vertex)))


;;***************************
;; Waltz filtering algorithm:
;;***************************

;;*******************************************************************
;; 
;; In general, computing a consistent labeling for a polyhedral scene 
;; is computationally hard.
;; In some cases, however, we can eliminate a large number of 
;; impossible labelings, thereby reducing the amount of search 
;; required to find a consistent labeling.
;; The following algorithm, due to Waltz [1975] provides an efficient 
;; filter on the set of possible labelings for a polyhedral scene.
;; 
;; The algorithm works as follows. 
;; We keep a list of the possible list of labelings for each vertex
;; stored with the vertex.
;; For each vertex, we consider for each possible labeling if that
;; labeling is consistent with the possible labelings at neighboring
;; vertices.
;; A labeling is consistent with the possible labelings at neighboring
;; vertices if there is some way that we can assign the labels in the
;; labeling given the labelings at the neighboring vertices.
;; If the labeling is not consistent, then we eliminate the labeling 
;; from the possible labelings stored with the vertex.
;; If we eliminate at least one labeling, then the algorithm applies
;; itself recursively to the neighboring vertices.
;; If the algorithm ever finds a vertex that has no possible labelings,
;; then it halts, otherwise it continues to make recursive calls until 
;; no more labelings can be eliminated.
;; 
;; If there are n vertices and at most L labelings for each vertex
;; type, then the Waltz filtering algorithm will terminate in at most 
;; n X L steps.
;; 
;; It still may be necessary to perform some additional search to 
;; find a consistent labeling. In the exercises, you are asked to 
;; write code to perform this additional search.
;; 
;;*******************************************************************

;; Eliminate labelings in a diagram using the Waltz filtering algorithm.
(define (filter diagram)
  (begin (newline) (display "The initial diagram is:"))
  (show-diagram diagram)
  (for-each waltz (diagram-vertices diagram))
  (begin (newline) (display "After filtering the diagram is:"))
  (show-diagram diagram))


;; Reduce the labelings on a vertex by considering the labeling of its
;; neighbors. If we can reduce the number of labelings, then recurse on 
;; the neighboring vertices. Return #f only if you encounter a vertex
;; with no possible labelings. 
(define (waltz vertex)
  (let ((old-num (vertex-number-of-labelings vertex)))
    (set-vertex-labelings vertex (consistent-labelings vertex))
    (if (impossible-vertex vertex)
      #f
      (or (not (< (vertex-number-of-labelings vertex) old-num))
          (every2 waltz (vertex-neighbors vertex))))))


;; Return the set of labelings that are consistent with neighbors.
(define (consistent-labelings vertex)
  (let ((neighbor-labels
          (map (lambda (neighbor) (labels-for neighbor vertex))
               (vertex-neighbors vertex))))
    (apply append
           (map (lambda (labeling)
                  (if (every2true
                        (map member
                             (map reverse-label labeling)
                             neighbor-labels))
                    (list labeling)
                    (list)))
                (vertex-labelings vertex)))))


;; Return all the labels for the line going to vertex.
(define (labels-for vertex from)
  (let ((i (index from (vertex-neighbors vertex))))
    (map (lambda (labeling) (nth i labeling)) 
	 (vertex-labelings vertex))))


;; Account for the fact that one vertex's right is another's left.
(define (reverse-label label)
  (if (eq? label 'left) 
      'right 
      (if (eq? label 'right) 
	  'left 
	  label)))
 

;;***************
;; Test function:
;;***************

(load "diagrams")

(define (test)
  (filter (copy-diagram cube))
  (filter (ground (copy-diagram cube) 'g 'd))
  (filter (copy-diagram cube-on-plate))
  (filter (ground (copy-diagram cube-on-plate) 'k 'm))
  (filter (ground (copy-diagram tower) 'l 'k)))
  (filter (ground (copy-diagram arch2) 'x 'z))
  (filter (ground (copy-diagram arch3) 'x 'z)))


;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

