;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;; We start by defining the abstract data types.

;; A feature consists of an attribute and a value.
(defun make-FEATURE (attribute value) (list attribute value))
(defun FEATURE-attribute (feature) (first feature))
(defun FEATURE-value (feature) (second feature))

;; A constituent consists of a category and zero or more features.
(defun make-CONSTITUENT (category features) (cons category features))
(defun CONSTITUENT-category (constituent) (first constituent))
(defun CONSTITUENT-features (constituent) (rest constituent))

;; The left hand side of a rule is a constituent and the 
;; right hand side of a rule is a list of constituents.
(defun make-RULE (lhs rhs) (cons lhs (cons '-> rhs)))
(defun RULE-lhs (rule) (first rule)) 
(defun RULE-rhs (rule) (rest (rest rule)))

;; Variables are defined as in the logic chapter.
(defun make-VAR (var) (list '? var))
(defun is-VAR (x) (and (consp x) (eq (first x) '?)))
(defun VAR-name (var) (second var))

;; A binding consists of a variable and a value.
(defun make-BDG (variable value) (list variable value))
(defun BDG-variable (bdg) (first bdg))
(defun BDG-value (bdg) (second bdg))

;; A utility for making substitutions given a list of bindings.
(defun subst-bdgs (expr bdgs)
  (if (null bdgs) 
      expr
    (if (equal (first bdgs) '(match t))
	(subst-bdgs expr (rest bdgs))
      (subst-bdgs (subst (BDG-value (first bdgs))
			 (BDG-variable (first bdgs)) expr :test #'equal) 
		  (rest bdgs)))))

;; A state consists of a list of arcs and a list of words.
(defun make-STATE (arcs words) (list arcs words))
(defun STATE-arcs (state) (first state))
(defun STATE-words (state) (second state))

;; An arc consists of a list constituents already accounted
;; for, a list of constituents that need to be accounted for,
;; and an integer number for the next subconstituent. 
(defun make-ARC (seen seek number) (list seen seek number))
(defun ARC-seen (arc) (first arc))
(defun ARC-seek (arc) (second arc))
(defun ARC-number (arc) (third arc))

;; An arc is extended with a given constituent by creating a new arc in
;; which the first unaccounted-for consituent is now accounted for by the
;; given constituent and the rest of the unaccounted-for constituents
;; still need to be accounted for.  Substitutions are made where
;; appropriate and the new arc is given a subconstituent number
;; corresponding to one plus the number of the original arc's
;; subconstituent number.
(defun ARC-extend (arc constituent bdgs)
  (make-ARC (subst-bdgs (append (ARC-seen arc)
				(list (list (ARC-number arc) 
					    constituent)))
			bdgs)
	    (subst-bdgs (rest (ARC-seek arc)) bdgs)
	    (+ (ARC-number arc) 1)))

;; An arc is complete if it has no constituents left to account for.
(defun complete (arc) (null (ARC-seek arc)))


;; The nest three functions implement a special-purpose 
;; matching algorithm for constituents and patterns thereof.
(defun constituent-match (pattern constituent)
  (if (not (eq (CONSTITUENT-category pattern) 
	       (CONSTITUENT-category constituent))) 
      (list)
    (feature-match (CONSTITUENT-features pattern) 
		   (CONSTITUENT-features constituent))))

(defun feature-match (pfeatures cfeatures)
 (if (null pfeatures) 
     '((match t))
   (let ((bdgs (value-match 
		(FEATURE-value (first pfeatures))
		(FEATURE-value (assoc (FEATURE-attribute 
				       (first pfeatures))
				      cfeatures)))))
     (if (null bdgs)
	 (list)
       (let ((result (feature-match (subst-bdgs (rest pfeatures) bdgs)
				    (subst-bdgs cfeatures bdgs))))
	 (if (null result)
	     (list)
	   (append bdgs result)))))))

(defun value-match (pvalue value)
  (cond ((eq pvalue value) '((match t)))
	;; If either argument is a variable, create new bindings.
	((is-VAR pvalue) (list (make-BDG pvalue value)))
	((is-VAR value) (list (make-BDG value pvalue)))
	(t (list))))


;; The function parse-found is used as the goal 
;; function for a generic search routine. 
(defun parse-found (state)
 (and (null (STATE-words state)) 
      (eq (length (STATE-arcs state)) 1) 
      (complete (first (STATE-arcs state)))))

;; The function next-states is used to generate 
;; successor states for a generic search routine. 
(defun next-states (state)
  (let ((words (STATE-words state)) (arcs (STATE-arcs state)))
    (cond ((null (first arcs)) (list)) ;; If no arc, then no successors.
	  ;; If the first arc is complete, then
	  ;; use it to extend the second arc.
	  ((and (complete (first arcs)) (second arcs))
	   (build-parse-state (ARC-seen (first arcs)) 
			      (second arcs) 
			      (rest (rest arcs)) 
			      words))
	  ;; If the first symbol sought is a lexical category, check in
	  ;; the lexicon otherwise create a new arc for the first symbol.
	  ((lexical-category (CONSTITUENT-category 
			      (first (ARC-seek (first arcs)))))
	   (let ((lexicon-entries (lexicon-lookup (first words))))
	     (if (null lexicon-entries)
		 (list)
	       (mapcan #'(lambda (entry)
			   (build-parse-state entry
					      (first arcs) 
					      (rest arcs) 
					      (rest words)))
		       lexicon-entries))))
	  (t (make-states-for-constituent (first (ARC-seek (first arcs)))
					  arcs words)))))

;; The function build-parse-state uses a constituent to extend 
;; the top arc and returns a new parse state.
(defun build-parse-state (constituent arc rest-arcs words)
  (let ((bdgs (constituent-match (first (ARC-seek arc)) constituent)))
    (if (null bdgs)
	(list)
      (list (make-STATE (cons (ARC-extend arc constituent bdgs)
			      rest-arcs)
			words)))))

;; The function make-states-for-constituent finds possible rules from the
;; grammar for a constituent and constructs an arc for each one found.
(defun make-states-for-constituent (constituent arcs words)
  (mapcar #'(lambda (rule)
              (make-STATE (cons (make-ARC (RULE-lhs rule) 
					  (RULE-rhs rule) 1)
				arcs) 
			  words))
          (get-rules constituent)))

;; Finally, a parse function that uses the standard depth-first search
;; procedure defined in Chapter 4.
(defun parse (expr)
  (let ((result (dfs (make-states-for-constituent '(s) (list) expr)
		     #'parse-found 
		     #'next-states)))
    (and result (ARC-seen (first (STATE-arcs result))))))

;; Here is the depth-first search procedure repeated from Chapter 4.
(defun dfs (nodes goalp next)
  (cond ((null nodes) nil)
        ((funcall goalp (first nodes)) (first nodes))
        (t (dfs (append (funcall next (first nodes)) 
			(rest nodes))
                goalp
                next))))

;; All that is left is to define the functions that access the grammar
;; and the lexicon. These are fairly straightforward, and the only
;; complication arises in making a copy of each rule so that the
;; variables are uniquely defined each time a rule is used. Here is the
;; function for looking up words in the lexicon. It returns a list of
;; constituents that could have generated the specified word. It also
;; replaces any * in the rule with a new constant.
(let ((lexicon
       '((dog (n (agr 3s) (root dog)))
	 (dogs (n (agr 3p) (root dog)))
	 (pizza (n (agr 3s) (root pizza)))
	 (saw (v (agr (? x)) (vform past) (subcat np) (root see)))
	 (barks (v (agr 3s) (vform pres) (subcat none) (root bark)))
	 (the (art (agr 3s) (root the))))))
  (defun lexicon-lookup (word)
    (mapcar #'(lambda (e) (subst (newname 'v) '* (second e)))
	    (mapcan #'(lambda (x) 
			(if (eq (first x) word) 
			    (list x)
			  (list)))
		    lexicon))))

;; Here is the function to find rules in the grammar whose left-hand 
;; side matches a specified constant. Variables in the rule are either 
;; new copies, or are bound to values based on the match against the 
;; constituent.
(let ((grammar
      '(((s (agr (? a))) -> 
	 (np (agr (? a))) (vp (agr (? a))))
	((np (agr (? a))) -> 
	 (art (agr (? a))) (n (agr (? a))))
	((vp (agr (? a)) (vform (? v))) -> 
	 (v (agr (? a)) (vform (? v)) (subcat none)))
	((vp (agr (? a)) (vform (? v))) -> 
	 (v (agr (? a)) (vform (? v)) (subcat np)) (np)))))
  (defun get-rules (constituent)
    (let ((new-rules (list)))
      (mapc #'(lambda (x) 
		(let ((bdgs 
		       (constituent-match constituent (RULE-lhs x))))
		  (if (not (null bdgs))
		      (setq new-rules (cons (cons bdgs x) new-rules)))))
	    grammar)
      (reverse (mapcar #'gen-new-varnames new-rules)))))

;; We need a function to recognize legal lexical categories.
(let ((categories '(n v p art adj adv aux)))
  (defun lexical-category (x) (member x categories)))

;; The function gen-new-varnames in conjunction with gen-varbindings 
;; renames all variables in a rule to make sure they are unique.
(defun gen-new-varnames (rule-bdgs-pair)
  (subst (newname 'v) '*	
	 (subst-bdgs (rest rule-bdgs-pair) 
		     (gen-varbindings (rest rule-bdgs-pair) 
				      (first rule-bdgs-pair)))))

(defun gen-varbindings (expr bdgs)
 (cond ((atom expr) bdgs)
       ((is-VAR expr) 
	;; If already bound, then we don't need a new variable.
	(if (assoc expr bdgs :test #'equal)
	    bdgs
	  (cons (make-BDG expr (make-VAR (newname (VAR-name expr)))) 
		bdgs)))
	(t (gen-varbindings (rest expr) 
			    (gen-varbindings (first expr) bdgs)))))

(defun newname (x)
  (intern (symbol-name (gensym (symbol-name x)))))

;; Here is a simple test function to demonstrate the parse.
(defun test () (parse '(the dog saw the pizza)))

;; A variant parser with more interesting output.
(defun parse-and-trace (expr)
  (let ((result (dfs (make-states-for-constituent '(s) (list) expr)
		     #'(lambda (x) (print-state x) (parse-found x))
		     #'next-states)))
    (and result (ARC-seen (first (STATE-arcs result))))))

(defun print-state (state)
  (format t "~%")
  (mapc #'print-arc (STATE-arcs state))
  (format t "Input: ~S~%" (STATE-words state)))

(defun print-arc (arc)
  (format t "(Seen: ~S  Seeking: ~S)~%"
          (ARC-seen arc)
          (ARC-seek arc)))

(defun verbose-test () (parse-and-trace '(the dog saw the pizza)))


;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

