;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; (in-package 'user)

;; We start by defining the abstract data types.

;; A feature consists of an attribute and a value.
(define (make-feature attribute value) (list attribute value))

(define (feature-attribute feature) (first feature))

(define (feature-value feature) (second feature))


;; A constituent consists of a category and zero or more features.
(define (make-constituent category features) (cons category features))

(define (constituent-category constituent) (first constituent))

(define (constituent-features constituent) (rest constituent))


;; The left hand side of a rule is a constituent and the 
;; right hand side of a rule is a list of constituents.
(define (make-rule lhs rhs) (cons lhs (cons '-> rhs)))

(define (rule-lhs rule) (first rule))
 
(define (rule-rhs rule) (rest (rest rule)))


;; Variables are defined as in the logic chapter.
(define (make-var var) (list '? var))

(define (is-var x) (and (pair? x) (eq? (first x) '?)))

(define (var-name var) (second var))


;; A binding consists of a variable and a value.
(define (make-bdg variable value) (list variable value))

(define (bdg-variable bdg) (first bdg))

(define (bdg-value bdg) (second bdg))


;; A utility for making substitutions given a list of bindings.
(define (subst-bdgs expr bdgs)
  (if (null? bdgs)
    expr
    (if (equal? (first bdgs) '(match #t))
      (subst-bdgs expr (rest bdgs))
      (subst-bdgs
        (subst2 (bdg-value (first bdgs))
                (bdg-variable (first bdgs))
                expr
                equal?
                (lambda (x) x))
        (rest bdgs)))))


;; A state consists of a list of arcs and a list of words.
(define (make-state arcs words) (list arcs words))

(define (state-arcs state) (first state))

(define (state-words state) (second state))


;; An arc consists of a list constituents already accounted
;; for, a list of constituents that need to be accounted for,
;; and an integer number for the next subconstituent. 
(define (make-arc seen seek number) (list seen seek number))

(define (arc-seen arc) (first arc))

(define (arc-seek arc) (second arc))

(define (arc-number arc) (third arc))


;; An arc is extended with a given constituent by creating a new arc in
;; which the first unaccounted-for consituent is now accounted for by the
;; given constituent and the rest of the unaccounted-for constituents
;; still need to be accounted for.  Substitutions are made where
;; appropriate and the new arc is given a subconstituent number
;; corresponding to one plus the number of the original arc's
;; subconstituent number.
(define (arc-extend arc constituent bdgs)
  (make-arc
    (subst-bdgs
      (append (arc-seen arc) 
	      (list (list (arc-number arc) constituent)))
      bdgs)
    (subst-bdgs (rest (arc-seek arc)) bdgs)
    (+ (arc-number arc) 1)))


;; An arc is complete if it has no constituents left to account for.
(define (complete arc) (null? (arc-seek arc)))


;; The nest three functions implement a special-purpose 
;; matching algorithm for constituents and patterns thereof.
(define (constituent-match pattern constituent)
  (if (not (eq? (constituent-category pattern)
                (constituent-category constituent)))
    (list)
    (feature-match
      (constituent-features pattern)
      (constituent-features constituent))))


(define (feature-match pfeatures cfeatures)
  (if (null? pfeatures)
    '((match #t))
    (let ((bdgs (value-match
                  (feature-value (first pfeatures))
                  (feature-value
                    (assoc2 (feature-attribute (first pfeatures))
                            cfeatures
                            eq?
                            (lambda (x) x))))))
      (if (null? bdgs)
        (list)
        (let ((result (feature-match
                        (subst-bdgs (rest pfeatures) bdgs)
                        (subst-bdgs cfeatures bdgs))))
          (if (null? result) (list) (append bdgs result)))))))


(define (value-match pvalue value)
  (cond ((eq? pvalue value) '((match #t)))
        ((is-var pvalue) (list (make-bdg pvalue value)))
        ((is-var value) (list (make-bdg value pvalue)))
        (else (list))))


;; The function parse-found is used as the goal 
;; function for a generic search routine. 
(define (parse-found state)
  (and (null? (state-words state))
       (eq? (length (state-arcs state)) 1)
       (complete (first (state-arcs state)))))


;; The function next-states is used to generate 
;; successor states for a generic search routine. 
(define (next-states state)
  (let ((words (state-words state)) (arcs (state-arcs state)))
    (cond ((null? (first arcs)) (list))
          ((and (complete (first arcs)) (second arcs))
           (build-parse-state
             (arc-seen (first arcs))
             (second arcs)
             (rest (rest arcs))
             words))
          ((lexical-category
             (constituent-category (first (arc-seek (first arcs)))))
           (let ((lexicon-entries (lexicon-lookup (first words))))
             (if (null? lexicon-entries)
               (list)
               (apply append
                      (map (lambda (entry)
                             (build-parse-state
                               entry
                               (first arcs)
                               (rest arcs)
                               (rest words)))
                           lexicon-entries)))))
          (else
           (make-states-for-constituent
             (first (arc-seek (first arcs)))
             arcs
             words)))))


;; The function build-parse-state uses a constituent to extend 
;; the top arc and returns a new parse state.
(define (build-parse-state constituent arc rest-arcs words)
  (let ((bdgs (constituent-match (first (arc-seek arc)) constituent)))
    (if (null? bdgs)
      (list)
      (list (make-state
              (cons (arc-extend arc constituent bdgs) rest-arcs)
              words)))))


;; The function make-states-for-constituent finds possible rules from the
;; grammar for a constituent and constructs an arc for each one found.
(define (make-states-for-constituent constituent arcs words)
  (map (lambda (rule)
         (make-state
           (cons (make-arc (rule-lhs rule) (rule-rhs rule) 1) arcs)
           words))
       (get-rules constituent)))


;; Finally, a parse function that uses the standard depth-first search
;; procedure defined in Chapter 4.
(define (parse expr)
  (let ((result (dfs (make-states-for-constituent '(s) (list) expr)
                     parse-found
                     next-states)))
    (and result (arc-seen (first (state-arcs result))))))


;; Here is the depth-first search procedure repeated from Chapter 4.
(define (dfs nodes goalp next)
  (cond ((null? nodes) #f)
        ((goalp (first nodes)) (first nodes))
        (else (dfs (append (next (first nodes)) (rest nodes)) 
		   goalp 
		   next))))


;; All that is left is to define the functions that access the grammar
;; and the lexicon. These are fairly straightforward, and the only
;; complication arises in making a copy of each rule so that the
;; variables are uniquely defined each time a rule is used. Here is the
;; function for looking up words in the lexicon. It returns a list of
;; constituents that could have generated the specified word. It also
;; replaces any * in the rule with a new constant.
(let ((lexicon '((dog (n (agr 3s) (root dog)))
                 (dogs (n (agr 3p) (root dog)))
                 (pizza (n (agr 3s) (root pizza)))
                 (saw (v (agr (? x)) (vform past) 
			 (subcat np) (root see)))
                 (barks (v (agr 3s) (vform pres) 
			   (subcat none) (root bark)))
                 (the (art (agr 3s) (root the))))))
  (set! lexicon-lookup
    (lambda (word)
      (map (lambda (e) 
	     (subst2 (newname 'v) '* (second e) eq? (lambda (x) x)))
           (apply append
                  (map (lambda (x) 
			 (if (eq? (first x) word) (list x) (list)))
                       lexicon))))))


;; Here is the function to find rules in the grammar whose left-hand 
;; side matches a specified constant. Variables in the rule are either 
;; new copies, or are bound to values based on the match against the 
;; constituent.
(let ((grammar '(((s (agr (? a))) -> (np (agr (? a))) (vp (agr (? a))))
                 ((np (agr (? a))) -> (art (agr (? a))) (n (agr (? a))))
                 ((vp (agr (? a)) (vform (? v))) ->
                  (v (agr (? a)) (vform (? v)) (subcat none)))
                 ((vp (agr (? a)) (vform (? v))) ->
                  (v (agr (? a)) (vform (? v)) (subcat np)) (np)))))
  (set! get-rules
    (lambda (constituent)
      (let ((new-rules (list)))
        (for-each
          (lambda (x)
            (let ((bdgs (constituent-match constituent (rule-lhs x))))
              (if (not (null? bdgs))
                (set! new-rules (cons (cons bdgs x) new-rules)))))
          grammar)
        (reverse (map gen-new-varnames new-rules))))))


;; We need a function to recognize legal lexical categories.
(let ((categories '(n v p art adj adv aux)))
  (set! lexical-category
    (lambda (x) (member2 x categories eq? (lambda (x) x)))))


;; The function gen-new-varnames in conjunction with gen-varbindings 
;; renames all variables in a rule to make sure they are unique.
(define (gen-new-varnames rule-bdgs-pair)
  (subst2 (newname 'v)
          '*
          (subst-bdgs
            (rest rule-bdgs-pair)
            (gen-varbindings (rest rule-bdgs-pair) 
			     (first rule-bdgs-pair)))
          eq?
          (lambda (x) x)))


(define (gen-varbindings expr bdgs)
  (cond ((atom expr) bdgs)
        ((is-var expr)
         (if (assoc2 expr bdgs equal? (lambda (x) x))
           bdgs
           (cons (make-bdg expr (make-var (newname (var-name expr)))) 
		 bdgs)))
        (else
         (gen-varbindings (rest expr) 
			  (gen-varbindings (first expr) bdgs)))))


(let ((n 0))
    (set! newname 
	  (lambda (x)
	    (set! n (+ n 1))
	    (string->symbol 
	     (string-append (symbol->string x)
			    (number->string n))))))


;; Here is a simple test function to demonstrate the parse.
(define (test) (parse '(the dog saw the pizza)))


;; A variant parser with more interesting output.
(define (parse-and-trace expr)
  (let ((result (dfs (make-states-for-constituent '(s) (list) expr)
                     (lambda (x) (print-state x) (parse-found x))
                     next-states)))
    (and result (arc-seen (first (state-arcs result))))))


(define (print-state state)
  (begin (newline))
  (for-each print-arc (state-arcs state))
  (begin (display "Input: ") 
	 (display (state-words state)) 
	 (newline)))


(define (print-arc arc)
  (begin
    (display "(Seen: ")
    (display (arc-seen arc))
    (display "  Seeking: ")
    (display (arc-seek arc))
    (display ")")
    (newline)))


(define (verbose-test) (parse-and-trace '(the dog saw the pizza)))



;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

