(*

This following prcedures implement algorithms to convert boolean
formulas to one of two standard normal forms.  Mathematica's
LogicalExpand already implements conversion to disjunctive normal form
(DNF) so most of the functions in this file are concerned with testing
for and conversion to conjunctive normal form (CNF). We begin with
some predicates for testing formulas. An atomic formula is simply a
boolean variable; a literal is either atomic or a negated atom; a
clause is a literal or a disjunction of literals; and a formula is in
conjunctive normal form if it is either a clause or a conjunction of
clauses.

*)

AtomicQ[f_] := 
  Not[MemberQ[{And,Or,Not,Implies},Head[f]]]

LiteralQ[f_] := 
  Or[AtomicQ[f],
     And[SameQ[Head[f],Not],AtomicQ[f[[1]]]]]
         
ClauseQ[f_] :=
  Or[LiteralQ[f],
     And[SameQ[Head[f],Or],
         Apply[And,Map[LiteralQ,Level[f,{1}]]]]]
                       
CNFQ[f_] :=
  Or[ClauseQ[f],
     And[SameQ[Head[f],And],
         Apply[And,Map[ClauseQ,Level[f,{1}]]]]]

(* 

Reverse And and Or for purposes of pushing a negation into a
disjunction or conjunction.  Not[And[p,q]] is equivalent to
Or[Not[p],Not[q]] and similarly Not[Or[p,q]] is equivalent to
And[Not[p],Not[q]]. As a simple exercise, you should convince your
self that this is true.

*)
                  
ReverseOrAnd[f_] := 
  If[SameQ[f,And],Or,And]

(* 

The following procedure pushes all negation all the way in, i.e., so
that the formula has no negations except at the atomic level.
Mathematica routinely performs some simplifications so that
Not[Not[p]] is just p, And[p] and Or[p] are both p, And[] is True and
Or[] is False.

*)

NegativeIn[f_] :=
  If[LiteralQ[f],f,
     If[SameQ[Head[f],Not],
        Apply[ReverseOrAnd[Head[First[f]]],
              Map[(NegativeIn[Not[#]])&,Level[f,{2}]]],
        Apply[Head[f],Map[NegativeIn,Level[f,{1}]]]]]

(*

It will simplify matters if the formulas we have to deal with consist
of only Or, And, Not, and atomic formulas. We use the fact that 
Implies[p.q] is logically equivalent to Or[Not[p],q] to justify the
folling conversion procedure.

*)

EliminateImplications[f_] := 
  If[AtomicQ[f],f,
     If[SameQ[Head[f],Implies],
        Or[Not[EliminateImplications[f[[1]]]],
           EliminateImplications[f[[2]]]],
        Apply[Head[f],Map[EliminateImplications,Level[f,{1}]]]]]

(*

As mentioned LogicalExpand converts formulas to disjunctive normal form.

*)

DNF[f_] := 
  LogicalExpand[EliminateImplications[f]]

(* 

In the following, we'll make use of Mathematica's extensive symbolic
manipulation capabilities to simplify our task of implementing a
procedure for converting formulas to conjunctive normal form.  In particular.
we'll use Simplify to simplify nested conjunctions
and disjunctions, and we'll also use Distribute to apply the distributive
law for And and Or.

You are probably familiar with applying the distribution law for Plus
(+) and Times (*) in arithmetic formulas.  Given a formula like ((a +
b) * (c * d * e) which is a product of sums, we can convert it to a
sum of products (a * c * d * e) + (b * c * d * e) using the
distribution law. In Mathematica, this is done by invoking
Distribute[(a + b) * (c * d * e),Plus,Times].  Similarly, we can
convert a disjunction into a conjunction, For example, from
Distribute[Or[And[a,b],And[c,d,e]],And,Or] we get
And[Or[a,c],Or[a,d],Or[a,e],Or[b,c],Or[b,d],Or[b,e]]. You'll notice
that in applying the distribution law formulas can get larger; in
fact, in some cases they can get much larger, but there is no general,
tractable method for avoiding such expansions (combinatorial
explosions).

*)

CNF[f_] := 
  AuxCNF[Simplify[NegativeIn[EliminateImplications[f]]]]

AuxCNF[f_] :=
  If[CNFQ[f], f,
    If[MatchQ[Head[f],Or],
       AuxCNF[Distribute[f,And,Or]],
       CNF[Simplify[Apply[And,
                          Map[(Distribute[#,And,Or])&,
                              Level[f,{1}]]]]]]]

(*

Here are some examples of converting boolean formulas.

*)

Module[{f1 = Not[And[Not[1],Not[Or[2,3]]]],
	f2 = Or[Not[1],Not[And[Or[1,2],3]]],
	f3 = Implies[1,Not[And[Implies[Not[1],2],3]]],
	f4 = Not[Or[1,And[2,1],And[3,4]]],
	f5 = Or[And[1,Or[2,3]],And[4,5],Or[6,7]],
	f6 = Not[And[Or[1,And[2,3]],4,Or[5,And[6,7],8]]]},
  Scan[(Print["Original formula: ",FullForm[#]] ;
        Print["CNF formula: ",FullForm[CNF[#]]] ;
        Print["DNF formula: ",FullForm[DNF[#]]])&, 
       {f1,f2,f3,f4,f5,f6}]]
