;; Copyright 1994, Brown University, Providence, RI
;; See end of file for full copyright information

;; Scheme procedures for translating from Common Lisp to Scheme.  

;; The following translation routines handle most of the subset 
;; of Common Lisp described in "Artificial Intelligence: Theory 
;; and Practice", Benjamin/Cummings, 1995.  The routines assume 
;; that a distinction is made between NIL and () in Common Lisp.
;; Specifically, NIL in Common lisp is interpreted as boolean
;; false (#f), and () is interpreted as the empty list ('()).
;; The following implementation assumes only those features of 
;; Scheme described in the R4RS and IEEE Scheme standards.

;; Translation of code written in the subset of Common Lisp for
;; "Artificial Intelligence: Theory and Practice" is largely 
;; automatic; however, there are some constructs that require
;; tweaking. In particular, use () when you mean the empty list
;; and use nil when you mean boolean false. In Common Lisp, the
;; SETQ form returns the value of last variable set. In Scheme,
;; the result of SET! is undefined. The translator given below
;; substitutes SET! for SETQ. One  alternative is to substitute 
;; (BEGIN (SET! <var> <expr>) <var>) for (SETQ <var> <expr>), but 
;; this would have been clumsy and unnecessary in most cases. The
;; translation procedures cannot handle Common lisp read macros, 
;; e.g., "#'". In this same directory, you will find a Perl script 
;; that will preprocess a file with ".lisp" extension to produce 
;; a file with ".temp" extension with the read macros removed.

;; Here is a convenient loop for processing files.
(define main-loop
  (lambda () 
    (display "File: ")
    (let ((file (read)))
      (if (not (eq? file 'quit))
          (begin (reset-let-translator)
                 (translate (symbol->string file))
                 (main-loop))))))

;; Here is a read-eval-print loop for experimenting.
(define common-loop
  (lambda () 
    (display "> ")
    (let ((input (read)))
      (if (not (eq? input 'quit))
          (begin (display (eval (schemify input)))
                 (newline)
                 (common-loop))))))

;; Read in a file named "file.temp" and create a translation file 
;; named "file.scm". Specify the file name without the extension.
(define (translate file)
  (let ((input-port (open-input-file (string-append file ".temp")))
        (output-port (open-output-file (string-append file ".scm"))))
    (aux-translate input-port output-port)
    (close-input-port input-port)
    (close-output-port output-port)))

;; Translate line-by-line and expresion-by-expression. 
(define (aux-translate input-port output-port)
  (read-write-white-space input-port output-port)
  (cond ((end-of-file? input-port))
        ((comment? input-port)
         (read-write-line input-port output-port)
         (aux-translate input-port output-port))
        (else (pretty-print (schemify (read input-port)) 
                            output-port)
              (aux-translate input-port output-port))))

;; Read and write characters until a newline is encountered.
(define (read-write-line input-port output-port)
  (if (not (char=? (peek-char input-port) #\newline))
      (begin (write-char (read-char input-port) output-port)
             (read-write-line input-port output-port))))

;; Read and write the next chunk of white space. 
(define (read-write-white-space input-port output-port)
  (if (not (end-of-file? input-port))
      (if (char-whitespace? (peek-char input-port))
          (begin (write-char (read-char input-port) output-port)
                 (read-write-white-space input-port output-port)))))

;; Check to see if the next character is a comment character.
;; This translator doesn't handle comments embedded in lists.
(define (comment? input-port)
  (char=? (peek-char input-port) #\;))

;; Check to see if we have reached the end of the file. 
(define (end-of-file? input-port)
  (eof-object? (peek-char input-port)))

;; Convert an expression from Common Lisp to Scheme. 
(define (schemify expression)
  (cond ((symbol? expression)
         (schemify-symbol expression))
        ((string? expression) expression)
        ((number? expression) expression)
        ((null? expression) '(quote ()))
        ((pair? expression) 
         (schemify-pair expression))
        (else (display "Unexpected input: ")
              (display expression))))

;; Translate Common Lisp symbols to their Scheme equivalents.
(define (schemify-symbol expression)
  (cond ((eq? expression 't) #t)
        ((eq? expression 'nil) #f)
        ((get expression 'scheme-equivalent))
        (else expression)))

;; eq -> eq?
(put 'eq 'scheme-equivalent 'eq?)

;; equal -> equal?
(put 'equal 'scheme-equivalent 'equal?)

;; consp -> pair?
(put 'consp 'scheme-equivalent 'pair?)

;; null -> null?
(put 'null 'scheme-equivalent 'null?)

;; symbolp -> symbol?
(put 'symbolp 'scheme-equivalent 'symbol?)

;; listp -> list?
(put 'listp 'scheme-equivalent 'list?)

;; oddp -> odd?
(put 'oddp 'scheme-equivalent 'odd?)

;; evenp -> even?
(put 'evenp 'scheme-equivalent 'even?)

;; integerp -> integer?
(put 'integerp 'scheme-equivalent 'integer?)

;; numberp -> number?
(put 'numberp 'scheme-equivalent 'number?)

;; zerop -> zero?
(put 'zerop 'scheme-equivalent 'zero?)

;; stringp -> string?
(put 'stringp 'scheme-equivalent 'string?)

;; mapcar -> map
(put 'mapcar 'scheme-equivalent 'map)

;; mapc -> for-each
(put 'mapc 'scheme-equivalent 'for-each)

;; progn -> begin
(put 'progn 'scheme-equivalent 'begin)

;; princ -> display
(put 'princ 'scheme-equivalent 'display)

;; mod -> modulo
(put 'mod 'scheme-equivalent 'modulo)

;; error -> display
;; Warning! Error is not an essential procedure in IEEE Scheme.
;; (put 'error 'scheme-equivalent 'display)

;; terpri -> newline
(put 'terpri 'scheme-equivalent 'newline)

;; nconc -> append
;; Warning! This uses APPEND for NCONC. 
(put 'nconc 'scheme-equivalent 'append)

;; Translate Common Lisp special forms to Scheme special forms.
(define (schemify-pair expression)
  (if (and (symbol? (car expression))
           (get (car expression) 
                'special-form-translator))
      ((get (car expression) 'special-form-translator) expression)
    (map schemify expression)))

;; (function expression) -> expression
(put 'function 'special-form-translator 
     (lambda (expression) (schemify (cadr expression))))

;; (funcall expression . args) -> (expression . args)
(put 'funcall 'special-form-translator 
     (lambda (expression) (schemify (cdr expression))))

;; In some implementations of Scheme, definitions (using define) are 
;; only allowed at the top level. Here we translate global definitions 
;; (using defun) in the scope of let statements into set! statements.
(let ((level 0))
  (letrec ((inside-let? (lambda () (> level 0)))
           (enter-let (lambda () (set! level (+ level 1))))
           (exit-let (lambda () (set! level (- level 1)))))
   (set! reset-let-translator (lambda () (set! level 0)))
   (put 'defun 'special-form-translator 
        (lambda (expression)
          (if (not (inside-let?))
              ;; (defun name args . body) -> 
              ;; (define (name . args) . body)
              (cons 'define 
                    (cons (cons (cadr expression)
                                (caddr expression))
                          (schemify (cdddr expression))))
            ;; (defun name args . body) -> 
            ;; (set! name (lambda args . body))
            (list 'set! 
                  (cadr expression)
                  (cons 'lambda 
                        (cons (caddr expression)
                              (schemify (cdddr expression))))))))
   ;; Some implementations of Scheme don't allow single 
   ;; variables in a let statement, only pairs of variables 
   ;; and initial values. The following tranlates a single 
   ;; variable of the form var-form into the pair (var-form '()). 
   (set! let-translator 
         (lambda (expression let-type)
           (let ((result '()))
             (enter-let)
             (set! result 
                   (cons let-type
                         (cons (map (lambda (var-form)
                                      (if (pair? var-form) var-form
                                        (list var-form #f)))
                                    (map schemify (cadr expression)))
                               (map schemify (cddr expression)))))
            (exit-let)
            result)))
   (put 'let 'special-form-translator 
        (lambda (expression)
          (let-translator expression 'let)))
   (put 'let* 'special-form-translator 
        (lambda (expression)
          (let-translator expression 'let*)))))

;; (cond . clauses) -> various forms
(put 'cond 'special-form-translator 
     (lambda (expression)
       (cons 'cond
             (map (lambda (clause)
                    (if (eq? (car clause) 't)
                        (cons 'else
                              (schemify (cdr clause)))
                      (schemify clause)))
                  (cdr expression)))))

;; (setq . args) -> various forms
(put 'setq 'special-form-translator 
     (lambda (expression)
       (cond ((= 3 (length expression))
              (cons 'set! 
                    (map schemify (cdr expression))))
             ((odd? (length expression))
              (cons 'begin
                    (do ((result (list))
                         (args (cdr expression) (cddr args)))
                        ((null? args) (reverse result))
                        (set! result 
                              (cons (cons 'set! 
                                          (list (schemify (car args))
                                                (schemify (cadr args))))
                                    result)))))
             (else (display "Poorly formed SET! expression: ")
                   (display expression)))))

;; (setf . args) -> various forms
(put 'setf 'special-form-translator 
     (lambda (expression)
       (cond ((= 3 (length expression))
              (schemify-setf (cadr expression)
                             (caddr expression)))
             ((odd? (length expression))
              (cons 'begin
                    (do ((result (list))
                         (args (cdr expression) (cddr args)))
                        ((null? args) (reverse result))
                        (set! result
                              (cons (schemify-setf (car args) 
                                                   (cadr args))
                                    result)))))
             (else (display "Poorly formed SETF expression: ")
                   (display expression)))))

;; Check for an abbreviation of the form c[a|d]*r.
(define (caaadddr? chars)
  (and (> (length chars) 3)
       (member (car chars) '(#\c #\C))
       (member (car (last-pair chars)) '(#\r #\R))
       (every2 (lambda (char) (member char '(#\a #\d #\A #\D)))
               (cdr (reverse (cdr chars))))))

;; The procedure LAST-PAIR assumes a non empty list.
(define (last-pair list)
  (if (null? (cdr list)) list (last-pair (cdr list))))

(define (schemify-setf location expression)
  (cond ((and (pair? location)
	      (member (car location) '(car first)))
	 ;; (setf (car expression1) expression2) ==>
	 ;; (set-car! expression1 expression2) 
         `(set-car! 
           ,(schemify (cadr location)) 
           ,(schemify expression)))
        ((and (pair? location)
	      (member (car location) '(cdr rest)))
	 ;; (setf (cdr expression1) expression2) ==>
	 ;; (set-cdr! expression1 expression2) 
         `(set-cdr! 
           ,(schemify (cadr location)) 
           ,(schemify expression)))
        ((and (pair? location)
	      (eq? (car location) 'nth))
	 ;; (setf (nth k list) expression) ==>
	 ;; (set-car! (list-tail list k) expression)
         `(set-car! 
           (list-tail ,(schemify (caddr location))
		      ,(schemify (cadr location)))
	   ,(schemify expression)))
        ((and (pair? location)
              (symbol? (car location))
              (let ((chars (string->list 
                            (symbol->string (car location)))))
                (and (caaadddr? chars)
                     (if (member (cadr chars) '(#\a #\A))
                         `(set-car! (,(string->symbol 
                                       (list->string 
                                        (cons #\c (cddr chars))))
                                     ,(schemify (cadr location)))
                                    ,(schemify expression))
                       `(set-cdr! (,(string->symbol 
                                     (list->string 
                                      (cons #\c (cddr chars))))
                                   ,(schemify (cadr location)))
                                  ,(schemify expression)))))))
        ((member (car location) '(first second third fourth fifth
                                  sixth seventh eighth ninth tenth))
         (list 'set-car! 
               (list 'list-tail 
                     (schemify (cadr location))
                     (cadr (assoc (car location)
                                  '((second 1)
                                    (third 2)
                                    (fourth 3)
                                    (fifth 4)
                                    (sixth 5)
                                    (seventh 6)
                                    (eighth 7)
                                    (ninth 8)
                                    (tenth 9)))))
               (schemify expression)))
        (else (display "Unknown SETF form: ")
              (display expression))))

;; (load string) -> 
(put 'load 'special-form-translator
     (lambda (expression)
        (list 'load 
              (if (member #\. (string->list (cadr expression)))
                  (cadr expression) 
                (string-append (cadr expression) ".scm")))))

;; (dotimes (var countform [resultform]) . body) -> 
(put 'dotimes 'special-form-translator
     (lambda (expression)
       `(do ((,(caadr expression) 1 
              (+ 1 ,(schemify (caadr expression)))))
            ((> ,(schemify (caadr expression))
                ,(schemify (cadadr expression)))
             ,@(if (null? (cddadr expression))
                   (list) (schemify (cddadr expression))))
            ,@(schemify (cddr expression)))))

;; (dolist (var listform [resultform]) . body) -> 
(put 'dolist 'special-form-translator
     (lambda (expression)
       (if (null? (cddadr expression))
           `(for-each (lambda (,(caadr expression))
                        ,@(schemify (cddr expression)))
                      ,(schemify (cadadr expression)))
         `(begin (for-each (lambda (,(caadr expression))
                             ,@(schemify (cddr expression)))
                           ,(schemify (cadadr expression)))
                 ,@(schemify (cddadr expression))))))

;; (mapcan procedure . args) -> 
;; Warning! This uses APPEND instead of NCONC. 
(put 'mapcan 'special-form-translator
     (lambda (expression)
       `(apply append (map ,(schemify (cadr expression))
                           ,@(map schemify (cddr expression))))))

;; (some predicate . args) -> 
;; Warning! This hack works like the Common Lisp version for all
;; predicates with a single argument, but applies the predicate to 
;; all arguments in the case of predicates with multiple arguments. 
(put 'some 'special-form-translator
     (lambda (expression)
       (if (> (length (cddr expression)) 1)
           `(some2true (map ,(schemify (cadr expression))
                            ,@(map schemify (cddr expression))))
         (cons 'some2 (schemify (cdr expression))))))

(define some2true
  (lambda (args)
    (cond ((null? args) #f)
          ((null? (cdr args)) (car args))
          ((car args))
          (else (some2true (cdr args))))))

(define (some2 predicate arguments)
  (do ((args arguments (cdr args))
       (flag #f (predicate (car args))))
      ((or (null? args) flag) flag)))

;; (every predicate . args) -> 
;; Warning! Same as for translation of SOME.
(put 'every 'special-form-translator
     (lambda (expression)
       (if (> (length (cddr expression)) 1)
           `(every2true (map ,(schemify (cadr expression))
                             ,@(map schemify (cddr expression))))
         (cons 'every2 (schemify (cdr expression))))))

(define every2true
  (lambda (args)
    (cond ((null? args) #t)
          ((null? (cdr args)) (car args))
          ((car args) (every2true (cdr args)))
          (else #f))))

(define (every2 predicate arguments)
  (do ((args arguments (cdr args))
       (flag #t (predicate (car args))))
      ((or (null? args) (not flag)) flag)))

;; The following version of format ignores the destination argument 
;; and handles only a few directives without any sophistication. 
;; (format destination control-string . args) ->
(put 'format 'special-form-translator
     (lambda (expression)
        (cons 'begin
              (format-translator (string->list (caddr expression))
                                 (cdddr expression) (list)))))

(define (format-translator chars args results)
  (if (null? chars) 
      (reverse results)
    (if (char=? #\~ (car chars))
        (cond ((char=? #\% (cadr chars)) 
               (format-translator (cddr chars)
                                  args
                                  (cons '(newline)
                                        results)))
              (else (format-translator
                     (cddr (divide-control-chars (list) chars 
                                                 '(#\F #\D #\A #\S
						   #\f #\d #\a #\s)))
                     (cdr args)
                     (cons `(display ,(schemify (car args)))
                           results))))
      (let ((split (divide-control-chars (list) chars '(#\~))))
        (format-translator (cdr split)
                           args
                           (cons `(display ,(list->string (car split)))
                                 results))))))

(define (divide-control-chars seen unseen dividers)
  (if (or (null? unseen) 
          (member (car unseen) dividers))
      (cons (reverse seen) unseen)
    (divide-control-chars (cons (car unseen)
                                seen)
                          (cdr unseen)
                          dividers)))

;; Definitions:

;; first => car 
(define first car)
;; rest => cdr 
(define rest cdr)
;; nth => list-ref
(define (nth k list) (list-ref list k))
;; nthcdr => list-tail 
(define (nthcdr k list) (list-tail list k))
;; second => cadr 
(define second cadr)
;; third => caddr
(define third caddr)
;; fourth => list-ref 3
(define fourth (lambda (x) (list-ref x 3)))
;; fifth => list-ref 4
(define fifth (lambda (x) (list-ref x 4)))
;; sixth => list-ref 5
(define sixth (lambda (x) (list-ref x 5)))
;; seventh => list-ref 6 
(define seventh (lambda (x) (list-ref x 6)))
;; eighth => list-ref 7
(define eighth (lambda (x) (list-ref x 7)))
;; ninth => list-ref 8
(define ninth (lambda (x) (list-ref x 8)))
;; tenth => list-ref 9
(define tenth (lambda (x) (list-ref x 9)))
;; in-package => quote
(define in-package (lambda (x) (quote x)))
;; atom => not pair?
(define atom (lambda (x) (not (pair? x))))
;; copy => copy-tree
(define copy (lambda (expression) (copy-tree expression)))
;; copy-tree => copy-tree
(define (copy-tree tree)
  (cond ((pair? tree)
         (cons (copy-tree (car tree))
               (copy-tree (cdr tree))))
        (else tree)))
;; copy-list => copy-list
(define (copy-list list)
  (cond ((pair? list)
         (cons (car list)
               (copy-list (cdr list))))
        (else list)))

(define (reduce function list)
  (if (< (length list) 2)
      (display "REDUCE expects at least two items to combine!")
      (aux-reduce function (cddr list)
		  (function (car list) (cadr list)))))

(define (aux-reduce function list sofar)
  (if (null? list) sofar
      (aux-reduce function (cdr list)
		  (function (car list) sofar))))

;; Dealing with keywords and optional arguments, e.g., member, 
;; assoc, sort, merge. The following only handles lists and 
;; a subset of the keywords allowed in the Common Lisp functions.

;; Find a particular keyword argument or supply the default.
(define (get-keyword-arg keyword expression)
  (if (null? expression) 
      (get-default-keyword-arg keyword)
    (if (eq? keyword (car expression))
        (cadr expression)
      (get-keyword-arg keyword (cdr expression)))))

;; Return the default function for a given keyword.
(define (get-default-keyword-arg keyword)
  (let ((result (assq keyword '((:key (lambda (x) x)) 
				(:test eq?)))))
    (if (pair? result) (cadr result)
      (begin (display "Unknown keyword: ")
             (display keyword)))))

;; (member item list &key :test :key) -> 
(put 'member 'special-form-translator
     (lambda (expression)
       ;; Unfortunately, MEMBER in Scheme uses EQUAL? as a test
       ;; and MEMBER in Common Lisp uses EQ as a default test. 
       `(member2
	 ,(schemify (cadr expression))
	 ,(schemify (caddr expression))
	 ,(schemify (get-keyword-arg ':test expression))
	 ,(schemify (get-keyword-arg ':key expression)))))

(define (member2 item list test key)
  (if (null? list) #f
    (if (test item (key (car list))) list
      (member2 item (cdr list) test key))))

;; (adjoin item list &key :test) ->
(put 'adjoin 'special-form-translator
     (lambda (expression)
       `(adjoin2
	 ,(schemify (cadr expression))
         ,(schemify (caddr expression))
         ,(schemify (get-keyword-arg ':test expression)))))

(define (adjoin2 item items test)
  (if (member2 item items test (lambda (x) x)) items
    (cons item items)))

;; Returns true if every element of list1 matches some element of list2.
;; (subsetp list1 list2 &key :test) -> 
(put 'subsetp 'special-form-translator
     (lambda (expression)
       `(subsetp2
	 ,(schemify (cadr expression))
         ,(schemify (caddr expression))
         ,(schemify (get-keyword-arg ':test expression)))))

(define (subsetp2 list1 list2 test)
  (or (null? list1)
      (and (member2 (car list1) list2 test (lambda (x) x))
	   (subsetp2 (cdr list1) list2 test))))

;; Returns a list containing the union of the elements list1 and list2.
;; (union list1 list2 &key :test) ->
(put 'union 'special-form-translator
     (lambda (expression)
       `(union2
	 ,(schemify (cadr expression))
         ,(schemify (caddr expression))
         ,(schemify (get-keyword-arg ':test expression)))))

(define (union2 list1 list2 test)
  (cond ((null? list1) list2)
	((member2 (car list1) list2 test (lambda (x) x))
	 (union2 (cdr list1) list2 test))
	(else (cons (car list1)
		    (union2 (cdr list1) list2 test)))))

;; (set-difference list1 list2 &key :test) ->
;; Returns the list of elements of list1 that are not elements of list2.
(put 'set-difference 'special-form-translator
     (lambda (expression)
       `(set-difference2
	 ,(schemify (cadr expression))
         ,(schemify (caddr expression))
         ,(schemify (get-keyword-arg ':test expression)))))

(define (set-difference2 list1 list2 test)
  (cond ((null? list1) (list))
	((member2 (car list1) list2 test (lambda (x) x))
	 (set-difference2 (cdr list1) list2 test))
	(else (cons (car list1)
		    (set-difference2 (cdr list1) list2 test)))))

;; (find item list &key :test :key) -> 
;; Returns the first element in list that matches item.
(put 'find 'special-form-translator
     (lambda (expression)
       `(let ((result-of-find
	       (member2 
		,(schemify (cadr expression))
		,(schemify (caddr expression))
		,(schemify (get-keyword-arg ':test expression))
		,(schemify (get-keyword-arg ':key expression)))))
	  (if result-of-find (car result-of-find) #f))))

;; (assoc indicator a-list &key :test :key) ->
(put 'assoc 'special-form-translator 
     (lambda (expression)
       `(assoc2
	 ,(schemify (cadr expression))
	 ,(schemify (caddr expression))
	 ,(schemify (get-keyword-arg ':test expression))
	 ,(schemify (get-keyword-arg ':key expression)))))

(define (assoc2 indicator a-list test key)
  (if (null? a-list) #f
    (if (test indicator (key (car (car a-list))))
        (car a-list)
      (assoc2 indicator (cdr a-list) test key))))

;; (sort list predicate &key :key) -> 
(put 'sort 'special-form-translator
     (lambda (expression)
       `(sort2
	 ,(schemify (cadr expression))
	 ,(schemify (caddr expression))
	 ,(schemify (get-keyword-arg ':key expression)))))

;; Destructive merge sort function.
(define (sort2 items compare key)
  (let ((half (floor (/ (length items) 2))))
    (if (< half 1) items
      (let ((items1 items)
            (items2 (list-tail items half)))
        (set-cdr! (list-tail items (- half 1)) '())
        (merge2 (sort2 items1 compare key)
		(sort2 items2 compare key)
		compare 
		key)))))

;; (merge result-type list1 list2 predicate &key :key) ->
(put 'merge 'special-form-translator
     (lambda (expression)
      `(merge2
	,(schemify (list-ref expression 2))
	,(schemify (list-ref expression 3))
	,(schemify (list-ref expression 4))
	,(schemify (get-keyword-arg ':key expression)))))

;; Nondestructive merge function with key function.
(define (merge2 items1 items2 compare key)
  (cond ((null? items1) items2)
        ((null? items2) items1)
        ((compare (key (car items1)) (key (car items2)))
         (cons (car items1) 
               (merge2 (cdr items1) items2 compare key)))
        (else (cons (car items2) 
                    (merge2 items1 (cdr items2) compare key)))))

;; (substitute new-item old-item list &key :test :key) ->
(put 'substitute 'special-form-translator
     (lambda (expression)
       `(substitute2
	 ,(schemify (list-ref expression 1))
	 ,(schemify (list-ref expression 2))
	 ,(schemify (list-ref expression 3))
	 ,(schemify (get-keyword-arg ':test expression))
	 ,(schemify (get-keyword-arg ':key expression)))))

;; Nondestructive substitution function for lists.
(define (substitute2 new old list test key)
  (cond ((null? list) list)
        ((test (key (car list)) old)
         (cons new (substitute2 new old (cdr list) test key)))
        (else (cons (car list) 
                    (substitute2 new old (cdr list) test key)))))

;; (subst new-item old-item tree &key :test :key) ->
(put 'subst 'special-form-translator
     (lambda (expression)
       `(subst2 
	 ,(schemify (list-ref expression 1))
	 ,(schemify (list-ref expression 2))
	 ,(schemify (list-ref expression 3))
	 ,(schemify (get-keyword-arg ':test expression))
	 ,(schemify (get-keyword-arg ':key expression)))))

;; Nondestructive substitution function for trees.
;; Warning! This doesn't quite do what the Common Lisp function does 
;; due to the fact that CAR and CDR of NIL are errors in Scheme. 
(define (subst2 new old tree test key)
  (cond ((null? tree) tree)
	((test (key tree) old) new)
	((pair? tree)
	 (cons (subst2 new old (car tree) test key)
	       (subst2 new old (cdr tree) test key)))
	(else tree)))

;; (remove item list &key :test :key) ->
(put 'remove 'special-form-translator
     (lambda (expression)
       `(remove2 
	 ,(schemify (list-ref expression 1))
	 ,(schemify (list-ref expression 2))
	 ,(schemify (get-keyword-arg ':test expression))
	 ,(schemify (get-keyword-arg ':key expression)))))

;; Nondestructive remove function for lists.
(define (remove2 item list test key)
  (cond ((null? list) list)
        ((test (key (car list)) item)
         (remove2 item (cdr list) test key))
        (else (cons (car list) 
                    (remove2 item (cdr list) test key)))))

;; Here is a quick hack for implementations without a random number 
;; generator. This hack obviously won't work for anything that 
;; critically depends on having a source of pseudo-random numbers. 
;; See the code in random.scm for a slower but more generally 
;; useful method of generating pseudo-random numbers. 
;; The sequence of numbers was generated by Allegro Common Lisp.
;; In keeping with Common Lisp, the following generates a real or 
;; integer value depending on the type of its first argument.
(let ((seed 0) (modulus 0)
      (numbers 
       '(0.9705307999928068 0.1780474390379136 0.837710515148488 
         0.0335874266585664 0.4034739876153859 0.500259589540663 
         0.7206118600187068 0.6704864307355003 0.669966249341804 
         0.0126268219837725 0.7346085200159962 0.408728100806709 
         0.1918372075012401 0.3727742601151382 0.728807818958676 
         0.6207268783213029 0.4380411131616530 0.290546760000921 
         0.0190118719884231 0.8918786456969107 0.062713459806651
         0.2018956442225163 0.6459758932938453 0.773087816866130 
         0.1267715913697212 0.1635043708204753 0.633631590234069 
         0.2267228813102028 0.2301775799446902 0.463506921105646 
         0.4549280877465431 0.2048867943464115 0.520520840104336 
         0.9513634294091472 0.0043688065787707 0.009610825465481 
         0.3398529735526999 0.4869002767492354 0.996466249011118 
         0.3247639040723377 0.9990544565384778 0.274261740767672 
         0.9908245514136687 0.2700160181240039 0.071359168515983 
         0.3291004969512707 0.4117313610420116 0.190274726493643 
         0.0115651874593532 0.3262106683027996 0.611541044148352 
         0.8978469319861658 0.21695406528333966 0.59177278954724 
         0.9635427571871638 0.5293383962467257 0.184138951158128 
         0.4472358457146994 0.5733008347150838 0.481996786703242 
         0.4801675418041463 0.6648107151579772 0.840999256292613 
         0.3260588919471180 0.6652211635848713 0.465078456372164 
         0.2833983538598248 0.4462510443158079 0.497185079008907 
         0.7293048283125256 0.5897281976201374 0.851475000817126 
         0.0111126873625369 0.7767227896160623 0.847237834781524 
         0.8186262336377435 0.2783696970171567 0.881524065924557 
         0.4430562432423935 0.7140269782477073 0.762787849079700 
         0.6110334792194977 0.3111394241331802 0.060985896569463 
         0.9065126103821330 0.8067430912682970 0.173699024444157 
         0.2354333955727880 0.8347759560497136 0.594925234152458 
         0.7968580887508513 0.7629463839022634 0.451884512483759 
         0.3498719912397701 0.7456174063442741 0.111634399438478 
         0.6952463988672779 0.4111747623305234 0.622826627576972))) 
  (set! modulus (length numbers))
  ;; This is an abomination but it distinguishes "1" from "1.0".
  (letrec 
   ((float? 
     (lambda (x) 
       (and (number? x)
            (member #\. (string->list (number->string x)))
            #t)))
    (aux-random
     (lambda () 
       (set! seed (+ 1 seed))
       (list-ref numbers (modulo seed modulus)))))
   ;; Only the function random is available globally.
   (set! random
         (lambda (max)
           (cond ((not (and (number? max) (> max 0)))
                  (display "Random expects a number > 0."))
                 ((float? max) (* max (aux-random)))
                 (else (inexact->exact (floor (* max 
                                                 (aux-random))))))))))


;; Copyright 1994, Brown University, Providence, RI
;; Permission to use and modify this software and its documentation
;; for any purpose other than its incorporation into a commercial
;; product is hereby granted without fee.  Permission to copy and
;; distribute this software and its documentation only for
;; non-commercial use is also granted without fee, provided, however
;; that the above copyright notice appear in all copies, that both
;; that copyright notice and this permission notice appear in
;; supporting documentation, that the name of Brown University not
;; be used in advertising or publicity pertaining to distribution
;; of the software without specific, written prior permission, and
;; that the person doing the distribution notify Brown University
;; of such distributions outside of his or her organization. Brown
;; University makes no representations about the suitability of this
;; software for any purpose. It is provided "as is" without express
;; or implied warranty.  Brown University requests notification of
;; any modifications to this software or its documentation.
;;
;; Send the following redistribution information
;;
;; 	Name:
;; 	Organization:
;; 	Address (postal and/or electronic):
;;
;; To:
;; 	Software Librarian
;; 	Computer Science Department, Box 1910
;; 	Brown University
;; 	Providence, RI 02912
;;
;; 		or
;;
;; 	brusd@cs.brown.edu
;;
;; We will acknowledge all electronic notifications.

