(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    160829,       3779]*)
(*NotebookOutlinePosition[    161661,       3808]*)
(*  CellTagsIndexPosition[    161617,       3804]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Supplementary Notes on Machine Learning", "Title"],

Cell[CellGroupData[{

Cell["Building Decisions Trees from Data", "Section"],

Cell[TextData[{
  "Our task is to find a function that accounts for a set of training \
examples and hopefully generalizes to examples that we haven't seen yet.  \
Using the terminology in the textbook, a feature is an attribute (",
  StyleBox["e.g.,",
    FontSlant->"Italic"],
  " color) together with a value for that attribute (",
  StyleBox["e.g.,",
    FontSlant->"Italic"],
  " blue), a dimension is an attribute together with a set of possible values \
for that attribute, and a training example is a list of features \
corresponding to the input for the function we are trying to learn and a \
value for the output of the function given the input."
}], "Text"],

Cell[TextData[{
  "In the following implementation, each training example is a list of length \
one more than the number of dimensions that define the domain.  In the text, \
we talk about an example being a pair of the form ",
  StyleBox["(x,f(x))",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " where ",
  StyleBox["x",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is an n-dimensional vector.  In this implementation, an example is of the \
form ",
  StyleBox["e = {e[[1]],...,e[[n]],e[[n+1]]}",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " where ",
  StyleBox["f(e[[1]],...,e[[n]]) = e[[n+1]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ".  In the following, the training examples will always be stored as a list \
which is the value of the variable ",
  StyleBox["data",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell["\<\
data = {{1,2,3,2,1,2},
        {1,1,2,2,1,2},
        {1,1,3,2,1,2},
        {1,2,1,1,1,1}} ;\
\>", "Input"],

Cell[CellGroupData[{

Cell["MatrixForm[data]", "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", GridBox[{
            {"1", "2", "3", "2", "1", "2"},
            {"1", "1", "2", "2", "1", "2"},
            {"1", "1", "3", "2", "1", "2"},
            {"1", "2", "1", "1", "1", "1"}
            }], ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]],

Cell[TextData[{
  "In this implementation, we assume that our target functions map from a \
finite space of dimension ",
  StyleBox["dims",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " to a finite one-dimensional space.  The ",
  StyleBox["i",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "th attribute takes on values from ",
  StyleBox["{1,...,vals[[i]]}",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and the function f that we are trying to learn takes on values from ",
  StyleBox["{1,...,vals[[dims+1]]}",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ". The following variables are set to properly characterize the target \
function. The assumption is that the training examples are generated by the \
target function and hence if ",
  StyleBox["vals",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and ",
  StyleBox["dims",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " characterize the target function then they characterize the training \
examples."
}], "Text"],

Cell["vals = {2,2,3,2,2,2} ;", "Input"],

Cell["dims = Length[vals] - 1 ;", "Input"],

Cell[TextData[{
  "To distinguish different experiments, we introduce a new data structure ",
  StyleBox["Experiment",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " that will combine a data set and information about the dimensions of the \
target function. "
}], "Text"],

Cell[BoxData[
    \(\(exp1\  = \ Experiment[data, \ vals, \ dims]\ ; \)\)], "Input"],

Cell["\<\
We are going to build a decision tree starting at the root. Each \
node of the tree will be represented as a list consisting of four items.\
\>",
   "Text"],

Cell[TextData[{
  "The first item is an integer uniquely identifying the node in the decision \
tree.  The second item is an integer, either 0, indicating that the node is a \
terminial node and thus doesn't branch, or ",
  StyleBox["i",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " between 1 and ",
  StyleBox["dims",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " indicating that the node is an internal node that branches on the ith \
attribute.  The third item is a list of the integers corresponding to the \
children of the node; this will be the empty list in the case of an terminal \
node.  The fourth and final item is a list of training examples that end up \
at that node. This list is always a subset of the training examples that end \
up at the parent of the node."
}], "Text"],

Cell[TextData[{
  "If ",
  StyleBox["n",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is an internal node that branches on attribute ",
  StyleBox["i",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", then ",
  StyleBox["n",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " will have ",
  StyleBox["vals[[i]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " children and its ",
  StyleBox["j",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "th child will end up with the subset of ",
  StyleBox["n",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "'s training examples that have feature ",
  StyleBox["{i,j}",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " (i.e., the ith attribute has value ",
  StyleBox["j",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ").  There are other things that we could store at a node, ",
  StyleBox["e.g.,",
    FontSlant->"Italic"],
  " the path from the root to the node, the value ",
  StyleBox["f(x)",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " for a terminal node, the partition according to ",
  StyleBox["f(x)",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " for an internal node, but we can also easily reconstruct these from the \
items specified above.  We'll use the following variables to keep track of \
the nodes that we have in the tree."
}], "Text"],

Cell["idnum = 0 ;", "Input"],

Cell["nodes = {} ;", "Input"],

Cell[TextData[{
  StyleBox["BuildDecisionTree",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is a recursive procedure that takes a node and a list of attributes that \
haven't been used yet to discriminate the examples stored at the node. ",
  StyleBox["BuildDecisionTree",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " recursively constructs a decision tree."
}], "Text"],

Cell[TextData[StyleBox[
"BuildDecisionTree[n_List,b_List] := \n    (* Exit if all the examples have \
the same value. *)\n    If[UniformQ[n[[4]]],Null,\n       Module[{a,p,c,d},\n \
          (* Find an attribute to branch on. *)\n           a = \
FindAttribute[n[[4]],b] ;\n           (* Partition according to the \
attribute. *)\n           p = PartitionExamples[n[[4]],a] ; \n           (* \
Create the list of resulting children. *)\n           c = \
Table[{++idnum,0,{},p[[i]]},{i,1,vals[[a]]}] ;\n           (* Set the list of \
attributes for recursive calls. *)\n           d = Complement[b,{a}] ; \n     \
      (* Update the global list of nodes. *)\n           nodes = \
Join[nodes,c] ;\n           (* Update the current node to account for the \
branching. *)\n           nodes[[n[[1]]]] = \n               \
{n[[1]],a,Table[c[[i]][[1]],{i,1,vals[[a]]}],n[[4]]} ;\n           (* \
Recursively build the tree starting at each child. *)\n           \
Map[(BuildDecisionTree[#,d])&,c] ;\n           (* Nothing to return. *)\n     \
      Return[]]]"]], "Input"],

Cell[TextData[{
  StyleBox["This auxiliary version of "],
  StyleBox["BuildDecisionTree",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[
  " constructs the root node, initializes the global variables and calls the \
two-parameter version of "],
  StyleBox["BuildDecisionTree",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[
  " that does all of the real work. Note that technically the variables \
introduced in a "],
  StyleBox["Block",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[
  " are not global variables but, rather, they are dynamically scoped \
variables. In contrast, the variables introduced in a "],
  StyleBox["Module",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  StyleBox[
  " are statically (or lexically) scoped variables. In case you're not \
familiar with this distinction, statically scoped variables are preferable \
for understandability and good programming practice, dynamically scoped \
variables are preferred to global variables but not by much, and global \
variables are to be avoided in the main. "]
}], "Text"],

Cell[TextData[{
  "BuildDecisi",
  StyleBox["onTre",
    FontSize->14],
  "e[Experiment[d_List,v_List,m_Integer]] := \n  Block[{data = d,vals = v, \
dims = m, idnum = 1,nodes = {}},\n    Module[{n = {1,0,{},d}},\n      \
AppendTo[nodes,n] ;\n      BuildDecisionTree[n,Range[1,dims]] ;\n      \
Return[DecisionTree[nodes,idnum]]]]"
}], "Input"],

Cell["\<\
This predicate tests a list of examples to see if they all have the \
same value.\
\>", "Text"],

Cell["\<\
UniformQ[e_List] := 
    If[Or[e == {},Rest[e] == {}],True,
       Module[{v = e[[1]][[dims+1]]},
           Apply[And,Table[v == e[[i]][[dims+1]],
                           {i,2,Length[e]}]]]]\
\>", "Input"],

Cell["\<\
This procedure partitions a set of examples into subsets according \
to the values of a given attribute.\
\>", "Text"],

Cell["\<\
PartitionExamples[e_List,a_Integer] := 
    Module[{r = Table[{},{vals[[a]]}]},
        Do[r[[ e[[i]][[a]] ]] = 
               Append[r[[ e[[i]][[a]] ]],e[[i]]] ,
           {i,1,Length[e]}] ;
        Return[r]]\
\>", "Input"],

Cell["\<\
For the first part of this exercise, we just choose a random \
attribute to branch on.\
\>", "Text"],

Cell["\<\
FindAttribute[e_List,b_List] := \
b[[Random[Integer,{1,Length[b]}]]]\
\>", "Input"],

Cell[TextData[{
  "It will be convenient to display the decision trees that result from \
invoking ",
  StyleBox["BuildDecisionTree",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " on a given ",
  StyleBox["Experiment",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ". ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " provides a nice package for manipulating and displaying graphs. "
}], "Text"],

Cell["<< DiscreteMath`Combinatorica`", "Input"],

Cell["\<\
Here is some ugly code for displaying experiments and decision \
trees; you can safely ignore the following three procedure definitions.\
\>", 
  "Text"],

Cell["\<\
ShowData[Experiment[data_List,_List,dims_Integer]] := 
    Module[{m},
        m = Prepend[Table[Prepend[data[[i]],i],
                          {i,1,Length[data]}],
                    Join[{\"#\"},
                         Table[i,{i,1,dims}],
                         {\"f(x)\"}]] ;
        Print[MatrixForm[m]]]\
\>", "Input"],

Cell["\<\
ShowDecisionTree[DecisionTree[nodes_List,idnum_Integer]] := 
    Module[{g},
        g = EmptyGraph[idnum] ;
        Do[Scan[(g = AddEdge[g,{i,#},Directed])&,
                nodes[[i]][[3]]],
           {i,1,idnum}] ;
        ShowLabeledGraph[
            RootedEmbedding[g,1],
            Table[If[nodes[[i]][[2]] == 0,i,
                     {i, nodes[[i]][[2]]}],
                  {i,1,idnum}]] ;
        Return[]]\
\>", "Input"],

Cell["\<\
ShowDecisionNodes[DecisionTree[nodes_List,idnum_Integer]] :=
    Module[{m,data = nodes[[1]][[4]]},
        m = Prepend[
                Table[{i,nodes[[i]][[2]],nodes[[i]][[3]],
                       Table[Position[data,
                                 nodes[[i]][[4]][[j]]][[1]][[1]],
                             {j,1,Length[nodes[[i]][[4]]]}]},
                      {i,1,idnum}],
                {\"Id\",\"Attribute\",\"Children\",\"Examples\"}] ;
        Print[MatrixForm[m]]]\
\>", "Input"],

Cell["Ok, let's build a decision tree from our test data.", "Text"],

Cell["dt1 = BuildDecisionTree[exp1] ;", "Input"],

Cell[CellGroupData[{

Cell["ShowData[exp1] ;", "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", GridBox[{
            {"\<\"#\"\>", "1", "2", "3", "4", "5", "\<\"f(x)\"\>"},
            {"1", "1", "2", "3", "2", "1", "2"},
            {"2", "1", "1", "2", "2", "1", "2"},
            {"3", "1", "1", "3", "2", "1", "2"},
            {"4", "1", "2", "1", "1", "1", "1"}
            }], ")"}],
      (MatrixForm[ #]&)]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell["ShowDecisionTree[dt1] ;", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 -0.157895 2.10526 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.025 w
.5 .89474 Mdot
.04545 .10526 Mdot
.5 .10526 Mdot
.95455 .10526 Mdot
.5 Mabswid
[ ] 0 setdash
.5 .89474 m
.04545 .10526 L
s
.5 .89474 m
.5 .10526 L
s
.5 .89474 m
.95455 .10526 L
s
.5 .89474 m
.04545 .10526 L
s
.5 .89474 m
.5 .10526 L
s
.5 .89474 m
.95455 .10526 L
s
gsave
.47273 .83158 -79.7812 -14.0625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.062500 moveto
 558.000000 14.062500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.062500 lineto
closepath
clip newpath
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
63.000000 11.250000 moveto
(8) show
69.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
75.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(,) show
84.562500 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(3) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
90.562500 11.250000 moveto
(<) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Math2Mono
%%+ Courier
grestore
[(2)] .01818 .04211 0 1 Mshowa
[(3)] .47273 .04211 0 1 Mshowa
[(4)] .92727 .04211 0 1 Mshowa
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/0P10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdoojn_l00;34gTJ2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_[C^]:JVYZf][Jo^kP00
oiT0IP3o0020P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;0000?`00O`00_`0o000o
?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0
_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo
_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno
_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo
_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo
0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_mWEc:0Pl=WIfJj^[WMgMiW<09W<<iW<
IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`c
c<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co
<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI
0?nI<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<ool0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@0000D01P6o00@1`@040Ad0000500030@000080
00<10000_000104000020L8000<10000700000H000<10000a@000`40003000030@0001`000070003
0@000<@000<10000^`070Ad0000800030@000<<000<10000^`000`40000200030@0001`000090003
0@000<4000<10000_@001@40000101h000090081`0030Kl00P4200030@0001`0000:00030@000<40
00<10000_@001040004N00001@000`40000200030@000<4000<10000_P000`400@0N00001@001040
00020Kh000@100000P7000030@0101h0000600@1`0040L800P4N0000o`2a0000o`2a0000o`2a0000
o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000
o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a0000o`2a000040060Kh01P6n00H1
4P0000l0206l00P1_0080A40000?00T1^`090K/02@4@00003`090K/02@6k00T1400000h02P6j00X1
^P0:0A00000?00T1^`090K/02@4@00003`090K/02@6k00T1400000l0206l00P1_0080A40000@00H1
_P060Kh01P4B00004`001040007000030@000;h000@10001500001H000<10000_P000`40002n0003
0@0001D0000G00030@000;d000<10000_@000`40000F000060000`40002l00030@000;`000<10000
5`0001P000<10000_0000`40002l00030@0001L0000I00030@000;/000<10000^`000`40000H0000
6@000`40002k00030@000;/000<10000600001X000<10000^P000`40002j00030@0001T0000J0003
0@000;X000<10000^P000`40000I00006`000`40002i00030@000;T000<100006P0001`000<10000
^0000`40002h00030@0001/0000L00030@000;P000<10000^0000`40000K00007@000`40002g0003
0@000;L000<10000700001d000<10000]`000`40002g00030@0001`0000N00030@000;H000<10000
]P000`40000M00007`000`40002e00030@000;D000<100007P0001l000<10000]@000`40002e0003
0@0001h0000P00030@000;@000<10000]0000`40000O000080000`40002d00030@000;@000<10000
7`00024000<10000/`000`40002c00030@000200000Q00030@000;<000<10000/`000`40000P0000
8P000`40002b00030@000;8000<100008@0002<000<10000/@000`40002a00030@000280000S0003
0@000;4000<10000/@000`40000R000090000`40002`00030@000;0000<100008`0002@000<10000
/0000`40002`00030@0002<0000U00030@000:l000<10000[`000`40000T00009@000`40002_0003
0@000:l000<10000900002H000<10000[P000`40002^00030@0002D0000W00030@000:d000<10000
[@000`40000V00009`000`40002]00030@000:d000<100009P0002P000<10000[0000`40002/0003
0@0002L0000X00030@000:`000<10000[0000`40000W0000:@000`40002[00030@000:/000<10000
:00002T000<10000Z`000`40002[00030@0002P0000Z00030@000:X000<10000ZP000`40000Y0000
:`000`40002Y00030@000:T000<10000:P0002/000<10000Z@000`40002Y00030@0002X0000/0003
0@000:P000<10000Z0000`40000[0000;0000`40002X00030@000:P000<10000:`0002d000<10000
Y`000`40002W00030@0002`0000^00030@000:H000<10000YP000`40000]0000;P000`40002V0003
0@000:H000<10000;@0002l000<10000Y@000`40002U00030@0002h0000_00030@000:D000<10000
Y@000`40000^0000<0000`40002T00030@000:@000<10000;`00030000<10000Y0000`40002T0003
0@0002l0000a00030@000:<000<10000X`000`40000`0000<P000`40002R00030@000:8000<10000
<@00038000<10000XP000`40002R00030@000340000c00030@000:4000<10000X@000`40000b0000
<`000`40002Q00030@000:4000<10000<P0003@000<10000X0000`40002P00030@0003<0000d0003
0@000:0000<10000X0000`40000c0000=@000`40002O00030@0009l000<10000=00003H000<10000
WP000`40002N00030@0003D0000f00030@0009h000<10000WP000`40000e0000=`000`40002M0003
0@0009d000<10000=P0003L000<10000W@000`40002M00030@0003H0000h00030@0009`000<10000
W0000`40000g0000>0000`40002L00030@0009`000<10000=`0003T000<10000V`000`40002K0003
0@0003P0000j00030@0009X000<10000VP000`40000i0000>P000`40002J00030@0009X000<10000
>@0003/000<10000V@000`40002I00030@0003X0000k00030@0009T000<10000V@000`40000j0000
?0000`40002H00030@0009P000<10000>`0003d000<10000U`000`40002G00030@0003`0000m0003
0@0009L000<10000U`000`40000l0000?P000`40002F00030@0009H000<10000?@0003h000<10000
UP000`40002F00030@0003d0000o00030@0009D000<10000U@000`40000n0000?`000`40002E0003
0@0009D000<10000?P00040000<10000U0000`40002D00030@0003l0001100030@0009<000<10000
T`000`4000100000@@000`40002C00030@0009<000<10000@000048000<10000TP000`40002B0003
0@000440001200030@00098000<10000TP000`4000110000@`000`40002A00030@00094000<10000
@P0004<000<10000T@000`40002A00030@000480001400030@00090000<10000T0000`4000130000
A@000`40002?00030@0008l000<10000A00004D000<10000S`000`40002?00030@0004@000160003
0@0008h000<10000SP000`4000150000AP000`40002>00030@0008h000<10000A@0004L000<10000
S@000`40002=00030@0004H0001700030@0008d000<10000S@000`4000160000B0000`40002<0003
0@0008`000<10000A`0004T000<10000R`000`40002;00030@0004P0001900030@0008/000<10000
R`000`4000180000BP000`40002:00030@0008X000<10000B@0004X000<10000RP000`40002:0003
0@0004T0001;00030@0008T000<10000R@000`40001:0000B`000`40002900030@0008T000<10000
BP0004`000<10000R0000`40002800030@0004/0001=00030@0008L000<10000Q`000`40001<0000
C@000`40002700030@0008L000<10000C00004h000<10000QP000`40002600030@0004d0001>0003
0@0008H000<10000QP000`40001=0000C`000`40002500030@0008D000<10000CP00050000<10000
Q0000`40002400030@0004l0001@00030@0008@000<10000Q0000`40001?0000D@000`4000230003
0@0008<000<10000D000054000<10000P`000`40002300030@000500001B00030@00088000<10000
PP000`40001A0000DP000`40002200030@00088000<10000D@0005<000<10000P@000`4000210003
0@000580001D00030@00080000<10000P0000`40001C0000E0000`40002000030@00080000<10000
D`0005D000<10000O`000`40001o00030@0005@0001E00030@0007l000<10000O`000`40001D0000
EP000`40001n00030@0007h000<10000E@0005H000<10000OP000`40001n00030@0005D0001G0003
0@0007d000<10000O@000`40001F0000F0000`40001l00030@0007`000<10000E`0005P000<10000
O0000`40001l00030@0005L0001I00030@0007/000<10000N`000`40001H0000F@000`40001k0003
0@0007/000<10000F00005X000<10000NP000`40001j00030@0005T0001J00030@0007X000<10000
NP000`40001I0000F`000`40001i00030@0007T000<10000FP0005`000<10000N0000`40001h0003
0@0005/0001L00030@0007P000<10000N0000`40001K0000G@000`40001g00030@0007L000<10000
G00005d000<10000M`000`40001g00030@0005`0001N00030@0007H000<10000MP000`40001M0000
G`000`40001e00030@0007D000<10000GP0005l000<10000M@000`40001e00030@0005h0001P0003
0@0007@000<10000M0000`40001O0000H0000`40001d00030@0007@000<10000G`00064000<10000
L`000`40001c00030@000600001Q00030@0007<000<10000L`000`40001P0000HP000`40001b0003
0@00078000<10000H@0006<000<10000L@000`40001a00030@000680001S00030@00074000<10000
L@000`40001R0000I0000`40001`00030@00070000<10000H`0006@000<10000L0000`40001`0003
0@0006<0001U00030@0006l000<10000K`000`40001T0000I@000`40001_00030@0006l000<10000
I00006H000<10000KP000`40001^00030@0006D0001W00030@0006d000<10000K@000`40001V0000
I`000`40001]00030@0006d000<10000IP0006P000<10000K0000`40001/00030@0006L0001X0003
0@0006`000<10000K0000`40001W0000J@000`40001[00030@0006/000<10000J00006T000<10000
J`000`40001[00030@0006P0001Z00030@0006X000<10000JP000`40001Y0000J`000`40001Y0003
0@0006T000<10000JP0006/000<10000J@000`40001Y00030@0006X0001/00030@0006P000<10000
J0000`40001[0000K0000`40001X00030@0006P000<10000J`0006d000<10000I`000`40001W0003
0@0006`0001^00030@0006H000<10000IP000`40001]0000KP000`40001V00030@0006H000<10000
K@0006l000<10000I@000`40001U00030@0006h0001_00030@0006D000<10000I@000`40001^0000
L0000`40001T00030@0006@000<10000K`00070000<10000I0000`40001T00030@0006l0001a0003
0@0006<000<10000H`000`40001`0000LP000`40001R00030@00068000<10000L@00078000<10000
HP000`40001R00030@000740001c00030@00064000<10000H@000`40001b0000L`000`40001Q0003
0@00064000<10000LP0007@000<10000H0000`40001P00030@0007<0001d00030@00060000<10000
H0000`40001c0000M@000`40001O00030@0005l000<10000M00007H000<10000GP000`40001N0003
0@0007D0001f00030@0005h000<10000GP000`40001e0000M`000`40001M00030@0005d000<10000
MP0007L000<10000G@000`40001M00030@0007H0001h00030@0005`000<10000G0000`40001g0000
N0000`40001L00030@0005`000<10000M`0007T000<10000F`000`40001K00030@0007P0001j0003
0@0005X000<10000FP000`40001i0000NP000`40001J00030@0005X000<10000N@0007/000<10000
F@000`40001I00030@0007X0001k00030@0005T000<10000F@000`40001j0000O0000`40001H0003
0@0005P000<10000N`0007`000<10000F0000`40001H00030@0007/0001m00030@0005L000<10000
E`000`40001l0000OP000`40001F00030@0005H000<10000O@0007h000<10000EP000`40001F0003
0@0007d0001o00030@0005D000<10000E@000`40001n0000O`000`40001E00030@0005D000<10000
OP00080000<10000E0000`40001D00030@0007l0002100030@0005<000<10000D`000`4000200000
P@000`40001C00030@0005<000<10000P000088000<10000DP000`40001B00030@00084000220003
0@00058000<10000DP000`4000210000P`000`40001A00030@00054000<10000PP0008<000<10000
D@000`40001A00030@000880002400030@00050000<10000D0000`4000230000Q@000`40001?0003
0@0004l000<10000Q00008D000<10000C`000`40001?00030@0008@0002600030@0004h000<10000
CP000`4000250000QP000`40001>00030@0004h000<10000Q@0008L000<10000C@000`40001=0003
0@0008H0002700030@0004d000<10000C@000`4000260000R0000`40001<00030@0004`000<10000
Q`0008T000<10000B`000`40001;00030@0008P0002900030@0004/000<10000B`000`4000280000
RP000`40001:00030@0004X000<10000R@0008X000<10000BP000`40001:00030@0008T0002;0003
0@0004T000<10000B@000`40002:0000R`000`40001900030@0004T000<10000RP0008`000<10000
B0000`40001800030@0008/0002=00030@0004L000<10000A`000`40002<0000S@000`4000170003
0@0004L000<10000S00008h000<10000AP000`40001600030@0008d0002>00030@0004H000<10000
AP000`40002=0000S`000`40001500030@0004D000<10000SP00090000<10000A0000`4000140003
0@0008l0002@00030@0004@000<10000A0000`40002?0000T@000`40001300030@0004<000<10000
T000094000<10000@`000`40001300030@000900002B00030@00048000<10000@P000`40002A0000
TP000`40001200030@00048000<10000T@0009<000<10000@@000`40001100030@000980002D0003
0@00040000<10000@0000`40002C0000U0000`40001000030@00040000<10000T`0009D000<10000
?`000`40000o00030@0009@0002E00030@0003l000<10000?`000`40002D0000UP000`40000n0003
0@0003h000<10000U@0009H000<10000?P000`40000n00030@0009D0002G00030@0003d000<10000
?@000`40002F0000V0000`40000l00030@0003`000<10000U`0009P000<10000?0000`40000l0003
0@0009L0002I00030@0003/000<10000>`000`40002H0000V@000`40000k00030@0003/000<10000
V00009X000<10000>P000`40000j00030@0009T0002J00030@0003X000<10000>P000`40002I0000
V`000`40000i00030@0003T000<10000VP0009`000<10000>0000`40000h00030@0009/0002L0003
0@0003P000<10000>0000`40002K0000W@000`40000g00030@0003L000<10000W00009d000<10000
=`000`40000g00030@0009`0002N00030@0003H000<10000=P000`40002M0000W`000`40000e0003
0@0003D000<10000WP0009l000<10000=@000`40000e00030@0009h0002P00030@0003@000<10000
=0000`40002O0000X0000`40000d00030@0003@000<10000W`000:4000<10000<`000`40000c0003
0@000:00002Q00030@0003<000<10000<`000`40002P0000XP000`40000b00030@00038000<10000
X@000:<000<10000<@000`40000a00030@000:80002S00030@00034000<10000<@000`40002R0000
Y0000`40000`00030@00030000<10000X`000:@000<10000<0000`40000`00030@000:<0002U0003
0@0002l000<10000;`000`40002T0000Y@000`40000_00030@0002l000<10000Y0000:H000<10000
;P000`40000^00030@000:D0002W00030@0002d000<10000;@000`40002V0000Y`000`40000]0003
0@0002d000<10000YP000:P000<10000;0000`40000/00030@000:L0002X00030@0002`000<10000
;0000`40002W0000Z@000`40000[00030@0002/000<10000Z0000:T000<10000:`000`40000[0003
0@000:P0002Z00030@0002X000<10000:P000`40002Y0000Z`000`40000Y00030@0002T000<10000
ZP000:/000<10000:@000`40000Y00030@000:X0002/00030@0002P000<10000:0000`40002[0000
[0000`40000X00030@0002P000<10000Z`000:d000<100009`000`40000W00030@000:`0002]0003
0@0002L000<100009`000`40002/0000[P000`40000V00030@0002H000<10000[@000:l000<10000
9@000`40000U00030@000:h0002_00030@0002D000<100009@000`40002^0000/0000`40000T0003
0@0002@000<10000[`000;0000<1000090000`40000T00030@000:l0002a00030@0002<000<10000
8`000`40002`0000/P000`40000R00030@00028000<10000/@000;8000<100008P000`40000R0003
0@000;40002c00030@00024000<100008@000`40002b0000/`000`40000Q00030@00024000<10000
/P000;@000<1000080000`40000P00030@000;<0002d00030@00020000<1000080000`40002c0000
]@000`40000O00030@0001l000<10000]0000;H000<100007P000`40000N00030@000;D0002f0003
0@0001h000<100007P000`40002e0000]`000`40000M00030@0001d000<10000]P000;L000<10000
7@000`40000M00030@000;H0002h00030@0001`000<1000070000`40002g0000^0000`40000L0003
0@0001`000<10000]`000;T000<100006`000`40000K00030@000;P0002j00030@0001X000<10000
6P000`40002i0000^P000`40000J00030@0001X000<10000^@000;/000<100006@000`40000I0003
0@000;X0002k00030@0001T000<100006@000`40002j0000_0000`40000H00030@0001P000<10000
^`000;`000<1000060000`40000H00030@000;/0002m00030@0001L000<100005`000`40002l0000
_P000`40000F00030@0001H000<10000_@000;h000<100005P000`40000F00030@000;d0002o0003
0@0001D000<100005@000`40002n0000_`000`40000E00030@0001D000<10000_P000<0000<10000
50000`40000D00030@000;l0003100030@0001<000<100004`000`4000300000`@000`40000C0003
0@0001<000<10000`0000;L00P4900030@00018000<100000`020@d000<10000`@000;H000<10000
2@001@400001010000<100001@000`40000:00030@000<40002f00030@0000X000@1000140000`40
000500030@0000T000<10000`P000;H000<100000`080@800P4=00@11P000`40000900030@000<80
002f00030@0000H000<100000P001040004<00030@0000<11@000`40000800030@000<<0002f0003
0@0000H000<100000`0010400@4>00030@0100D000<100001`000`4000340000]@000`4000070003
0@0000<000<100003`000`400@0600030@0000H000<10000a0000;H000<100001P000`4000040003
0@0000h000<100401@000`40000600030@000<D0002f00030@0000H000<1000010000`40000>0081
1P000`40000600030@000<D0002f00030@0000H000<100001@000`40000<00<11P000`4000050003
0@000<H0002f00030@0000H000<100001@000`40000=00030@0100D000<100001@000`4000360000
]`020@H000<100001P000`40000<00030@0100<00P4700030@000<L0002l008100<00@0020000`40
000800030@0000<12`000`4000380000_P020@T000<100002@040@`000<10000b0000<X000<10000
2P000`40000:00030@000<T0003:00030@0000X000<100002P000`4000390000b`000`4000090003
0@0000T000<10000bP000</000<100002@000`40000900030@000<X0003<00030@0000P000<10000
20000`40003;0000c@000`40000700030@0000L000<10000c0000<d000<100001`000`4000070003
0@000<`0003>00030@0000H000<100001P000`40003=0000cP000`40000600030@0000H000<10000
c@000<l000<100001@000`40000500030@000<h0003@00030@0000@000<1000010000`40003?0000
d0000`40000400030@0000@000<10000c`000=4000<100000`000`40000300030@000=00003A0003
0@0000<000<100000`000`40003@0000dP000`40000200030@00008000<10000d@000=8000<10000
0P000`40000200030@000=40003C00050@0000400`000`40003B0000e0001040004200030@000=<0
003D00040@000@8000<10000d`000=@01P7F0000d`080MD0003C00T1e0000=<02@7D0000dP0:0M@0
003C00T1e0000=<02@7D0000d`080MD0003D00H1eP000=L000<10000eP000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000001\
\>"],
  ImageRangeCache->{{{0, 431}, {431, 0}} -> {-0.0500054, 0.0749979, 
  0.00382834, 0.00165315}}]
}, Open  ]],

Cell[TextData[{
  "Each nonterminal node is labeled with a pair ",
  StyleBox["{i,j}",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " where ",
  StyleBox["i",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is the integer identifier for the node and ",
  StyleBox["j",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is the attribute that the node branches on.  Terminal nodes are labeled \
by the integer identifer for the node."
}], "Text"],

Cell[CellGroupData[{

Cell["ShowDecisionNodes[dt1] ;", "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", GridBox[{
            {"\<\"Id\"\>", "\<\"Attribute\"\>", "\<\"Children\"\>", 
              "\<\"Examples\"\>"},
            {"1", "3", \({2, 3, 4}\), \({1, 2, 3, 4}\)},
            {"2", "0", \({}\), \({4}\)},
            {"3", "0", \({}\), \({2}\)},
            {"4", "0", \({}\), \({1, 3}\)}
            }], ")"}],
      (MatrixForm[ #]&)]], "Print"]
}, Open  ]],

Cell["\<\
Let's try another example, but this time a little less abstract, \
though we'll have to translate into the integer-based encoding that we used \
in our implementation.  In this example (borrowed from \"Artificial \
Intelligence, Third Edition\" by Patrick Winston), the function we're trying \
to learn is a classifier for whether or not people described in terms of four \
attributes sunburn or not.  The four attributes are: hair - one of blonde, \
brown or red; height - one of short, tall or average; weight - one of light, \
heavy or average; lotion - one of yes or no.  Here's the data from the \
example in Winston's book.\
\>", "Text"],

Cell[TextData[{
  "Make ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " stop warning me about variables that are spelled  similarly to other \
variables I've used."
}], "Text"],

Cell[BoxData[
    \(\(\(Off[General::spell1]\ ; \)\ \)\)], "Input"],

Cell[CellGroupData[{

Cell["\<\
sunburn1 = {{\"name\", \"hair\", \"height\", \"weight\", \
\"lotion\", \"sunburn\"},
            {\"sarah\", \"blonde\", \"average\", \"light\",   \"no\",  \
\"yes\"},
            {\"dana\",  \"blonde\", \"tall\",    \"average\", \"yes\", \
\"no\"},
            {\"alex\",  \"brown\",  \"short\",   \"average\", \"yes\", \
\"no\"},
            {\"annie\", \"blonde\", \"short\",   \"average\", \"no\",  \
\"yes\"},
            {\"emily\", \"red\",    \"average\", \"heavy\",   \"no\",  \
\"yes\"},
            {\"pete\",  \"brown\",  \"tall\",    \"heavy\",   \"no\",  \
\"no\"},
            {\"john\",  \"brown\",  \"average\", \"heavy\",   \"no\",  \
\"no\"},
            {\"katie\", \"blonde\", \"short\",   \"light\",   \"yes\", \
\"no\"}} ;
        
MatrixForm[sunburn1]     \
\>", "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", GridBox[{
            {"\<\"name\"\>", "\<\"hair\"\>", "\<\"height\"\>", 
              "\<\"weight\"\>", "\<\"lotion\"\>", "\<\"sunburn\"\>"},
            {"\<\"sarah\"\>", "\<\"blonde\"\>", "\<\"average\"\>", 
              "\<\"light\"\>", "\<\"no\"\>", "\<\"yes\"\>"},
            {"\<\"dana\"\>", "\<\"blonde\"\>", "\<\"tall\"\>", 
              "\<\"average\"\>", "\<\"yes\"\>", "\<\"no\"\>"},
            {"\<\"alex\"\>", "\<\"brown\"\>", "\<\"short\"\>", 
              "\<\"average\"\>", "\<\"yes\"\>", "\<\"no\"\>"},
            {"\<\"annie\"\>", "\<\"blonde\"\>", "\<\"short\"\>", 
              "\<\"average\"\>", "\<\"no\"\>", "\<\"yes\"\>"},
            {"\<\"emily\"\>", "\<\"red\"\>", "\<\"average\"\>", 
              "\<\"heavy\"\>", "\<\"no\"\>", "\<\"yes\"\>"},
            {"\<\"pete\"\>", "\<\"brown\"\>", "\<\"tall\"\>", 
              "\<\"heavy\"\>", "\<\"no\"\>", "\<\"no\"\>"},
            {"\<\"john\"\>", "\<\"brown\"\>", "\<\"average\"\>", 
              "\<\"heavy\"\>", "\<\"no\"\>", "\<\"no\"\>"},
            {"\<\"katie\"\>", "\<\"blonde\"\>", "\<\"short\"\>", 
              "\<\"light\"\>", "\<\"yes\"\>", "\<\"no\"\>"}
            }], ")"}],
      (MatrixForm[ #]&)]], "Output"]
}, Open  ]],

Cell["\<\
TranslateSunburn[s_String] := 
    Select[{{\"blonde\", 1}, {\"brown\", 2}, {\"red\", 3}, 
            {\"average\", 2}, {\"yes\", 1}, {\"no\", 2},  
            {\"short\", 1},{\"tall\", 3}, {\"light\", 1}, 
            {\"heavy\", 3},{\"hair\", 1}, {\"height\", 2},  
            {\"weight\", 3},{\"lotion\", 4}},
            (First[#] === s)&][[1]][[2]]\
\>", "Input"],

Cell["\<\
exp2 = Experiment[
           Table[Table[TranslateSunburn[sunburn1[[i]][[j]]],
                       {j,2,Length[sunburn1[[1]]]}],
                 {i,2,Length[sunburn1]}],
           {3,3,3,2,2},4] ;\
\>", "Input"],

Cell["dt2 = BuildDecisionTree[exp2] ;", "Input"],

Cell[CellGroupData[{

Cell["ShowData[exp2] ;", "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", GridBox[{
            {"\<\"#\"\>", "1", "2", "3", "4", "\<\"f(x)\"\>"},
            {"1", "1", "2", "1", "2", "1"},
            {"2", "1", "3", "2", "1", "2"},
            {"3", "2", "1", "2", "1", "2"},
            {"4", "1", "1", "2", "2", "1"},
            {"5", "3", "2", "3", "2", "1"},
            {"6", "2", "3", "3", "2", "2"},
            {"7", "2", "2", "3", "2", "2"},
            {"8", "1", "1", "1", "1", "2"}
            }], ")"}],
      (MatrixForm[ #]&)]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell["ShowDecisionTree[dt2] ;", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 -0.144016 1.78499 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.025 w
.58264 .91075 Mdot
.21074 .63692 Mdot
.58264 .63692 Mdot
.95455 .63692 Mdot
.08678 .36308 Mdot
.21074 .36308 Mdot
.33471 .36308 Mdot
.04545 .08925 Mdot
.08678 .08925 Mdot
.1281 .08925 Mdot
.45868 .36308 Mdot
.58264 .36308 Mdot
.70661 .36308 Mdot
.66529 .08925 Mdot
.70661 .08925 Mdot
.74793 .08925 Mdot
.5 Mabswid
[ ] 0 setdash
.58264 .91075 m
.21074 .63692 L
s
.58264 .91075 m
.58264 .63692 L
s
.58264 .91075 m
.95455 .63692 L
s
.21074 .63692 m
.08678 .36308 L
s
.21074 .63692 m
.21074 .36308 L
s
.21074 .63692 m
.33471 .36308 L
s
.58264 .63692 m
.45868 .36308 L
s
.58264 .63692 m
.58264 .36308 L
s
.58264 .63692 m
.70661 .36308 L
s
.08678 .36308 m
.04545 .08925 L
s
.08678 .36308 m
.08678 .08925 L
s
.08678 .36308 m
.1281 .08925 L
s
.70661 .36308 m
.66529 .08925 L
s
.70661 .36308 m
.70661 .08925 L
s
.70661 .36308 m
.74793 .08925 L
s
.58264 .91075 m
.21074 .63692 L
s
.58264 .91075 m
.58264 .63692 L
s
.58264 .91075 m
.95455 .63692 L
s
.21074 .63692 m
.08678 .36308 L
s
.21074 .63692 m
.21074 .36308 L
s
.21074 .63692 m
.33471 .36308 L
s
.58264 .63692 m
.45868 .36308 L
s
.58264 .63692 m
.58264 .36308 L
s
.58264 .63692 m
.70661 .36308 L
s
.08678 .36308 m
.04545 .08925 L
s
.08678 .36308 m
.08678 .08925 L
s
.08678 .36308 m
.1281 .08925 L
s
.70661 .36308 m
.66529 .08925 L
s
.70661 .36308 m
.70661 .08925 L
s
.70661 .36308 m
.74793 .08925 L
s
gsave
.55537 .8572 -79.7812 -14.0625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.062500 moveto
 558.000000 14.062500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.062500 lineto
closepath
clip newpath
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
63.000000 11.250000 moveto
(8) show
69.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
75.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(,) show
84.562500 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(2) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
90.562500 11.250000 moveto
(<) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Math2Mono
%%+ Courier
grestore
gsave
.18347 .58337 -79.7812 -14.0625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.062500 moveto
 558.000000 14.062500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.062500 lineto
closepath
clip newpath
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
63.000000 11.250000 moveto
(8) show
69.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(2) show
75.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(,) show
84.562500 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
90.562500 11.250000 moveto
(<) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Math2Mono
%%+ Courier
grestore
gsave
.55537 .58337 -79.7812 -14.0625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.062500 moveto
 558.000000 14.062500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.062500 lineto
closepath
clip newpath
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
63.000000 11.250000 moveto
(8) show
69.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(3) show
75.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(,) show
84.562500 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(3) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
90.562500 11.250000 moveto
(<) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Math2Mono
%%+ Courier
grestore
[(4)] .92727 .58337 0 1 Mshowa
gsave
.0595 .30953 -79.7812 -14.0625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.062500 moveto
 558.000000 14.062500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.062500 lineto
closepath
clip newpath
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
63.000000 11.250000 moveto
(8) show
69.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(5) show
75.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(,) show
84.562500 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(3) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
90.562500 11.250000 moveto
(<) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Math2Mono
%%+ Courier
grestore
[(6)] .18347 .30953 0 1 Mshowa
[(7)] .30744 .30953 0 1 Mshowa
[(8)] .01818 .0357 0 1 Mshowa
[(9)] .0595 .0357 0 1 Mshowa
[(10)] .10083 .0357 0 1 Mshowa
[(11)] .4314 .30953 0 1 Mshowa
[(12)] .55537 .30953 0 1 Mshowa
gsave
.67934 .30953 -82.7812 -14.0625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.062500 moveto
 558.000000 14.062500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.062500 lineto
closepath
clip newpath
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
63.000000 11.250000 moveto
(8) show
69.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(13) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(,) show
90.562500 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
96.562500 11.250000 moveto
(<) show
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Math2Mono
%%+ Courier
grestore
[(14)] .63802 .0357 0 1 Mshowa
[(15)] .67934 .0357 0 1 Mshowa
[(16)] .72066 .0357 0 1 Mshowa
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/0P10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdoojn_l00;34gTJ2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_[C^]:JVYZf][Jo^kP00
oiT0IP3o0020P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;0000?`00O`00_`0o000o
?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0
_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo
_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno
_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo
_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo
0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_mWEc:0Pl=WIfJj^[WMgMiW<09W<<iW<
IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`c
c<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co
<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI
0?nI<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<ool0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@0000H0104=00<12@070@D00P7I00L11@040@801`4300@110070@D0
0`5c00001@020@800P4?00812P000`40000500040@000M/000<1000020000`40000400030@0000<0
00<100000P000`40000400030@0000D000@10001L`0000D000<100000P000`40000>00030@0000P0
00<1000010000`40000200030@000=P000<1000020000`40000400030@0000T000<100000`000`40
000400030@00008000<10000L00000D000<100000P000`40000>00812@000`40000400030@000080
00<10000f0000`40000300L11@000`40000900030@0000<000<1000010000`40000200030@000700
000500810P020A4000<100001`000`40000400030@00008000<10000f0000`40000300030@000080
00<1000010000`40000900030@0000<000<1000010000`40000200030@000700000600040@000@h0
0`400`01000800030@0000@000<100000P000`40003H00030@0000@000D100000@0600030@0000T0
00<100000`000`40000400810`000`40001`00001P040@d00P4200812@000`40000400030@000080
00<10000f0000`40000400810P000`40000400030@0000@000D100000@0600030@0000@00P420003
0@000740000500030@00008000<100002P00104000020@T000<1000010000`40000200030@000=P0
00<100001@001040004600030@0000@01@4600030@0000@000<100400P5c00001@000`4000020003
0@0000X000<100000P000`40000700030@0000@000<100000P000`40003H00030@0000H000<10040
1P000`40000400030@0000P000<100001@000`40001d00001@020@800P4<00810P000`4000050081
00<00@001P001040007H008100<00@001`000`400@03008100<00@001@000`400005008100<00@00
1P020GD0000600@13P030@X00P480081f`020@T00P4500811P050@D00P4900<1LP000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@000?l0/@0001001P4<00H130060N401P4<00H130060F/0000?00P1
2P080@X0207O00P12P080@X0205Z00003`090@T02@4900T1gP090@T02@4900T1J@0000l02@4900T1
2@090Mh02@4900T12@090FT0000>00X1200:0@P02P7M00X1200:0@P02P5Y00003`090@T02@4900T1
gP090@T02@4900T1J@0000l02@4900T12@090Mh02@4900T12@090FT0000?00P12P080@X0207O00P1
2P080@X0205Z000040060@`01P4<00H1h@060@`01P4<00H1J`0001<00P4@00030@0000h00P7V0081
40000`40000>0081K@0001@000<100003P000`40000>00030@000>H000<100003P000`40000>0003
0@0006`0000D00030@0000h000<100003P000`40003V00030@0000h000<100003P000`40001/0000
50000`40000>00030@0000h000<10000iP000`40000>00030@0000h000<10000K00001@000<10000
3P000`40000>00030@000>H000<100003P000`40000>00030@0006`0000E00030@0000d000<10000
3@000`40003X00030@0000d000<100003@000`40001]00005@000`40000=00030@0000d000<10000
j0000`40000=00030@0000d000<10000K@0001D000<100003@000`40000=00030@000>P000<10000
3@000`40000=00030@0006d0000E00030@0000d000<100003@000`40003X00030@0000d000<10000
3@000`40001]00005@000`40000=00030@0000d000<10000j0000`40000=00030@0000d000<10000
K@0001D000<100003@000`40000=00030@000>P000<100003@000`40000=00030@0006d0000E0003
0@0000d000<100003@000`40003X00030@0000d000<100003@000`40001]00005P000`40000<0003
0@0000`000<10000jP000`40000<00030@0000`000<10000KP0001H000<1000030000`40000<0003
0@000>X000<1000030000`40000<00030@0006h0000F00030@0000`000<1000030000`40003Z0003
0@0000`000<1000030000`40001^00005P000`40000<00030@0000`000<10000jP000`40000<0003
0@0000`000<10000KP0001H000<1000030000`40000<00030@000>X000<1000030000`40000<0003
0@0006h0000F00030@0000`000<1000030000`40003Z00030@0000`000<1000030000`40001^0000
5`000`40000;00030@0000/000<10000k0000`40000;00030@0000/000<10000K`0001L000<10000
2`000`40000;00030@000>`000<100002`000`40000;00030@0006l0000G00030@0000/000<10000
2`000`40003/00030@0000/000<100002`000`40001_00005`000`40000;00030@0000/000<10000
k0000`40000;00030@0000/000<10000K`0001L000<100002`000`40000;00030@000>`000<10000
2`000`40000;00030@0006l0000G00030@0000/000<100002`000`40003/00030@0000/000<10000
2`000`40001_00005`000`40000;00030@0000/000<10000k0000`40000;00030@0000/000<10000
K`0001P000<100002P000`40000:00030@000>h000<100002P000`40000:00030@000700000H0003
0@0000X000<100002P000`40003^00030@0000X000<100002P000`40001`000060000`40000:0003
0@0000X000<10000kP000`40000:00030@0000X000<10000L00001P000<100002P000`40000:0003
0@000>h000<100002P000`40000:00030@000700000H00030@0000X000<100002P000`40003^0003
0@0000X000<100002P000`40001`000060000`40000:00030@0000X000<10000kP000`40000:0003
0@0000X000<10000L00001P000<100002P000`40000:00030@000>h000<100002P000`40000:0003
0@000700000I00030@0000T000<100002@000`40003`00030@0000T000<100002@000`40001a0000
6@000`40000900030@0000T000<10000l0000`40000900030@0000T000<10000L@0001T000<10000
2@000`40000900030@000?0000<100002@000`40000900030@000740000I00030@0000T000<10000
2@000`40003`00030@0000T000<100002@000`40001a00006@000`40000900030@0000T000<10000
l0000`40000900030@0000T000<10000L@0001T000<100002@000`40000900030@000?0000<10000
2@000`40000900030@000740000J00030@0000P000<1000020000`40003b00030@0000P000<10000
20000`40001b00006P000`40000800030@0000P000<10000lP000`40000800030@0000P000<10000
LP0001X000<1000020000`40000800030@000?8000<1000020000`40000800030@000780000J0003
0@0000P000<1000020000`40003b00030@0000P000<1000020000`40001b00006P000`4000080003
0@0000P000<10000lP000`40000800030@0000P000<10000LP0001X000<1000020000`4000080003
0@000?8000<1000020000`40000800030@000780000J00030@0000P000<1000020000`40003b0003
0@0000P000<1000020000`40001b00006`000`40000700030@0000L000<10000m0000`4000070003
0@0000L000<10000L`0001/000<100001`000`40000700030@000?@000<100001`000`4000070003
0@0007<0000K00030@0000L000<100001`000`40003d00030@0000L000<100001`000`40001c0000
6`000`40000700030@0000L000<10000m0000`40000700030@0000L000<10000L`0001/000<10000
1`000`40000700030@000?@000<100001`000`40000700030@0007<0000K00030@0000L000<10000
1`000`40003d00030@0000L000<100001`000`40001c000070000`40000600030@0000H000<10000
mP000`40000600030@0000H000<10000M00001`000<100001P000`40000600030@000?H000<10000
1P000`40000600030@0007@0000L00030@0000H000<100001P000`40003f00030@0000H000<10000
1P000`40001d000070000`40000600030@0000H000<10000mP000`40000600030@0000H000<10000
M00001`000<100001P000`40000600030@000?H000<100001P000`40000600030@0007@0000L0003
0@0000H000<100001P000`40003f00030@0000H000<100001P000`40001d000070000`4000060003
0@0000H000<10000mP000`40000600030@0000H000<10000M00001d000<100001@000`4000050003
0@000?P000<100001@000`40000500030@0007D0000M00030@0000D000<100001@000`40003h0003
0@0000D000<100001@000`40001e00007@000`40000500030@0000D000<10000n0000`4000050003
0@0000D000<10000M@0001d000<100001@000`40000500030@000?P000<100001@000`4000050003
0@0007D0000M00030@0000D000<100001@000`40003h00030@0000D000<100001@000`40001e0000
7@000`40000500030@0000D000<10000n0000`40000500030@0000D000<10000M@0001h000<10000
10000`40000400030@000?X000<1000010000`40000400030@0007H0000N00030@0000@000<10000
10000`40003j00030@0000@000<1000010000`40001f00007P000`40000400030@0000@000<10000
nP000`40000400030@0000@000<10000MP0001h000<1000010000`40000400030@000?X000<10000
10000`40000400030@0007H0000N00030@0000@000<1000010000`40003j00030@0000@000<10000
10000`40001f00007P000`40000400030@0000@000<10000nP000`40000400030@0000@000<10000
MP0001h000<1000010000`40000400030@000?X000<1000010000`40000400030@0007H0000O0003
0@0000<000<100000`000`40003l00030@0000<000<100000`000`40001g00007`000`4000030003
0@0000<000<10000o0000`40000300030@0000<000<10000M`0001l000<100000`000`4000030003
0@000?`000<100000`000`40000300030@0007L0000O00030@0000<000<100000`000`40003l0003
0@0000<000<100000`000`40001g00007`000`40000300030@0000<000<10000o0000`4000030003
0@0000<000<10000M`0001l000<100000`000`40000300030@000?`000<100000`000`4000030003
0@0007L0000P00030@00008000<100000P000`40003n00030@00008000<100000P000`40001h0000
80000`40000200030@00008000<10000oP000`40000200030@00008000<10000N000020000<10000
0P000`40000200030@000?h000<100000P000`40000200030@0007P0000500816@000`4000020003
0@0000800`7O00817@000`40000200030@00008000<100000`020G<0000400030@0000d000<10000
2@000`40000200030@00008000@10001g@000`40000F00030@0000@000<100000P000`4000020003
0@0000D000<10000L00000@000<100003@000`40000900030@00008000<100000P001040004P00<1
<`000`40000Z00L10P070BH01`4300H150000`40000F00030@0000@000<100000P000`4000020003
0@0000D000<10000L00000@000<1000010040@D00P4:00030@0000@10`001040004O00040@000C<0
00<10000;@000`40000600030@0002X000<1000010000`40000200030@00018000<100000`070@@0
104400811@000`40000200P11@000`40001`000010000`40000300030@00008000<100000`000`40
000900810P030@030040008000<1000070000`40000200030@00034000<10000;0000`4000060003
0@0002X000<100001@000`40000F00030@0000H000<100001000104000020@@000<1000010001@40
000100<000<100001P000`40001`000010000`40000900030@0000800P4:00090@0000400@0100<0
00<1000070000`40000200030@00034000<10000;0000`40000600030@0002X000<100001P000`40
000E00030@0000H000<100002@000`40000200811@001@40000100<000<100001P000`40001`0000
0`000`40000:00030@0000h000T100000@01004010000`40000K00030@00008000<10000<@000`40
000/00030@0000H000<10000:P000`40000700030@0001<000<100001`000`40000900030@0000T0
00D100000@0300030@0000L000<10000K`0000@000<100002@000`40000>00090@0000400@0100<0
00<1000070020@<000<10000<P000`40000[00030@0000H000<10000:P000`40000800030@0001<0
00<100001P000`40000900030@0000T000D100000@0300030@0000H000<10000L00000@000<10000
2@000`40000>00040@0000810P001@40000101h00P4200030@0003<000<10000:`000`4000060003
0@0002X000<1000020020A@000<100001P000`40000800030@0000X000D100000@0300030@0000H0
00<10000L00000@000<1000010001@400001014000<100000`4200050@0000407P000`400@020CD0
00<10000:`000`40000600030@0002X000<100002@000`40000B00030@0000H000<100001P030@`0
00D100000@0300030@0000H000<10000L00000@000<1000010050A4000T100000@0100400`000`40
000M00030@0003H000<10000:`000`40000600030@0002X000<1000010000`40000200030@000180
00<100001P000`40000900030@0000T000D100000@0300030@0000H000<10000L00000D00P440003
0@0001@000D1000100020@800P4P0081<`000`40000200030@0002L00P400`010004008100<00@00
:0020@03004000D000@100000P4E00811P000`40000900030@0000X000@100010P020@H00P5c0000
2`000`40000D00030@0000@19P030C001P4[00811`020B/00P4700@16`020@03004000D000@10000
0P4<00030@0000D1N`0000/01@4B00D100<00@00j@020@L0104=00040@000@800P5k00008P001040
004200030@000?l00`001040004200030@0007X0000R00040@000@8000<10000o`0300040@000@80
00<10000NP00028000@100010P000`40003o00<000@100010P000`40001j00008`001@400@010?l0
1`001@400@0107d0000S00050@010040o`0700050@010040O@0002<000D100400@3o00L000D10040
0@1m00008`001@400@010?l01`001@400@0107d0000S00050@010040o`0700050@010040O@0002<0
00D100400@3o00L000D100400@1m00008`001@400@010?l01`001@400@0107d0000T00<1o`0900<1
OP0002@00`7o00T00`5n000090030Ol02@030Gh0000T00<1o`0900<1OP0002@00`7o00T00`5n0000
8P060Bl01P4`00H1;`060C001P4_00H1O@000240204]00P1;P080Bd0204^00P1;@080G`0000Q00T1
;0090Bd02@4/00T1;@090B`02@5k00008@090B`02@4]00T1;0090Bd02@4/00T1N`0002002P4[00X1
;00:0B/02P4/00X1:`0:0G/0000Q00T1;0090Bd02@4/00T1;@090B`02@5k00008@090B`02@4]00T1
;0090Bd02@4/00T1N`000240204]00P1;P080Bd0204^00P1;@080G`0000R00H1;`060C001P4_00H1
<0060Bl01P5m00009@000`400@0b00030@00034000<10040=0000`400@0c00030@00030000<10040
O`0002P000<10000;`000`40000`00030@0003P000<10000<0000`40000_00030@000800000X0003
0@0002l000<10000<0000`40000h00030@00030000<10000;`000`4000200000:@000`40000^0003
0@0002l000<10000>P000`40000_00030@0002h000<10000P@0002T000<10000;P000`40000_0003
0@0003X000<10000;`000`40000^00030@000840000Y00030@0002h000<10000;P000`40000l0003
0@0002h000<10000;P000`4000210000:P000`40000]00030@0002h000<10000?0000`40000^0003
0@0002d000<10000PP0002X000<10000;@000`40000^00030@0003`000<10000;P000`40000]0003
0@000880000[00030@0002`000<10000;@000`40000n00030@0002d000<10000;0000`4000230000
:`000`40000/00030@0002d000<10000?P000`40000]00030@0002`000<10000P`0002`000<10000
:`000`40000/00030@00040000<10000;0000`40000[00030@0008@0000/00030@0002/000<10000
;0000`40001000030@0002`000<10000:`000`4000240000;@000`40000Z00030@0002/000<10000
@P000`40000[00030@0002X000<10000Q@0002d000<10000:P000`40000[00030@00048000<10000
:`000`40000Z00030@0008D0000^00030@0002T000<10000:P000`40001400030@0002X000<10000
:@000`4000260000;P000`40000Y00030@0002X000<10000A0000`40000Z00030@0002T000<10000
QP0002h000<10000:@000`40000Y00030@0004H000<10000:@000`40000Y00030@0008H0000_0003
0@0002P000<10000:@000`40001600030@0002T000<10000:0000`4000270000;`000`40000X0003
0@0002P000<10000B0000`40000X00030@0002P000<10000Q`00030000<100009`000`40000X0003
0@0004P000<10000:0000`40000W00030@0008P0000`00030@0002L000<10000:0000`4000180003
0@0002P000<100009`000`4000280000<@000`40000V00030@0002L000<10000BP000`40000W0003
0@0002H000<10000R@00034000<100009P000`40000W00030@0004X000<100009`000`40000V0003
0@0008T0000b00030@0002D000<100009P000`40001<00030@0002H000<100009@000`40002:0000
<P000`40000U00030@0002H000<10000C0000`40000V00030@0002D000<10000RP00038000<10000
9@000`40000U00030@0004h000<100009@000`40000U00030@0008X0000c00030@0002@000<10000
9@000`40001>00030@0002D000<1000090000`40002;0000<`000`40000T00030@0002@000<10000
D0000`40000T00030@0002@000<10000R`0003@000<100008`000`40000T00030@00050000<10000
90000`40000S00030@0008`0000d00030@0002<000<100008`000`40001B00030@0002<000<10000
8`000`40002<0000=@000`40000R00030@0002<000<10000DP000`40000S00030@00028000<10000
S@0003D000<100008P000`40000S00030@00058000<100008`000`40000R00030@0008d0000f0003
0@00024000<100008P000`40001D00030@00028000<100008@000`40002>0000=P000`40000Q0003
0@00028000<10000E0000`40000R00030@00024000<10000SP0003L000<1000080000`40000Q0003
0@0005H000<100008@000`40000P00030@0008l0000g00030@00020000<100008@000`40001F0003
0@00024000<1000080000`40002?0000=`000`40000P00030@00020000<10000F0000`40000P0003
0@00020000<10000S`0003P000<100007`000`40000P00030@0005P000<1000080000`40000O0003
0@000900000h00030@0001l000<100007`000`40001J00030@0001l000<100007`000`40002@0000
>@000`40000N00030@0001l000<10000FP000`40000O00030@0001h000<10000T@0003T000<10000
7P000`40000N00030@0005`000<100007P000`40000N00030@000940000j00030@0001d000<10000
7P000`40001L00030@0001h000<100007@000`40002B0000>P000`40000M00030@0001d000<10000
GP000`40000M00030@0001d000<10000TP0003/000<1000070000`40000M00030@0005h000<10000
7@000`40000L00030@0009<0000k00030@0001`000<100007@000`40001N00030@0001d000<10000
70000`40002C0000>`000`40000L00030@0001`000<10000H0000`40000L00030@0001`000<10000
T`0003`000<100006`000`40000L00030@00060000<1000070000`40000K00030@0009@0000l0003
0@0001/000<100006`000`40001R00030@0001/000<100006`000`40002D0000?@000`40000J0003
0@0001/000<10000HP000`40000K00030@0001X000<10000U@0003d000<100006P000`40000J0003
0@0006@000<100006P000`40000J00030@0009D0000n00030@0001T000<100006P000`40001T0003
0@0001X000<100006@000`40002F0000?P000`40000I00030@0001T000<10000IP000`40000I0003
0@0001T000<10000UP0003l000<1000060000`40000I00030@0006H000<100006@000`40000H0003
0@0009L0000o00030@0001P000<1000060000`40001X00030@0001P000<1000060000`40002G0000
@0000`40000G00030@0001P000<10000J0000`40000H00030@0001P000<10000U`00040000<10000
5`000`40000H00030@0006P000<1000060000`40000G00030@0009P0001000030@0001L000<10000
5`000`40001Z00030@0001L000<100005`000`40002H0000@@000`40000F00030@0001L000<10000
JP000`40000G00030@0001H000<10000V@00044000<100005P000`40000F00030@0006`000<10000
5P000`40000F00030@0009T0001200030@0001D000<100005P000`40001/00030@0001H000<10000
5@000`40002J0000@P000`40000E00030@0001D000<10000KP000`40000E00030@0001D000<10000
VP0004<000<1000050000`40000E00030@0006h000<100005@000`40000D00030@0009/000130003
0@0001@000<1000050000`40001`00030@0001@000<1000050000`40002K0000A0000`40000C0003
0@0001@000<10000L0000`40000D00030@0001<000<10000W00004@000<100004`000`40000C0003
0@00078000<100004`000`40000C00030@0009`0001400030@0001<000<100004`000`40001b0003
0@0001<000<100004`000`40002L0000A@000`40000B00030@0001<000<10000LP000`40000C0003
0@00018000<10000W@0004D000<100004P000`40000B00030@0007@000<100004P000`40000B0003
0@0009d0001600030@00014000<100004P000`40001d00030@00018000<100004@000`40002N0000
AP000`40000A00030@00014000<10000MP000`40000A00030@00014000<10000WP0004L000<10000
40000`40000A00030@0007H000<100004@000`40000@00030@0009l0001700030@00010000<10000
40000`40001h00030@00010000<1000040000`40002O0000B0000`40000?00030@00010000<10000
N0000`40000@00030@0000l000<10000X00004P000<100003`000`40000?00030@0007X000<10000
3`000`40000?00030@000:00001800030@0000l000<100003`000`40001j00030@0000l000<10000
3`000`40002P0000B@000`40000>00030@0000h000<10000O0000`40000>00030@0000h000<10000
X@0004T000<100003P000`40000>00030@0007`000<100003P000`40000>00030@000:40001:0003
0@0000d000<100003@000`40001n00030@0000d000<100003@000`40002R0000BP000`40000=0003
0@0000d000<10000OP000`40000=00030@0000d000<10000XP0004/000<1000030000`40000=0003
0@0007h000<100003@000`40000<00030@000:<0000j00813`000`40000<00030@0000@00P460003
0@0006l00P4?00030@0000`000<100000`020@L000<10000X`0003T000<100003@001040004=0003
0@0000H000<100000`000`40001^00030@0000d000<100403P000`40000500030@0000<000<10000
Y00003T000<100003@001040004=00030@0000H000<100000P000`40001_00030@0000d000@10001
3@000`40000500030@0000<000<10000P`040Ad0000i00030@0000@01P4300810P000`40000700L1
1@000`40000200030@0006l000<100001@040@@00P400`01000:00@11P000`40000200030@0008H0
00<10000700003T000<1000010000`40000200030@00008000@1000130000`40000600050@000040
LP000`40000400040@00008110001040004900030@0000<11@000`40000200030@0008H000<10000
700003T000<100001@000`400006008100<00@002`000`40000600050@000040LP000`4000090003
0@0000800P400`01000;00030@0100D000<100000P000`40002100L17@0003P000<100001`000`40
000900030@0000T000<100001`000`400@1b00030@0000X000<100001P000`40000900030@0100H0
00@10001Q0000`40000200030@0001`0000i00030@0000L000<1000020000`40000900030@0000H0
00@10001L`000`40000900030@0000H000<100002@000`400@0500050@000040Q@001@40000101h0
000i00030@0000P000<1000020000`40000800030@0000H000<10040M0000`40000800030@0000P0
00<1000020020@H000@10001QP020@8000<10000700003T000<1000020020@T000<1000020000`40
000600030@0107@000<100001P030@X000<100001`030@H000@10001Q`001040004N0000>@000`40
000900030@0000P000<100001`000`40000600030@0107@000<100002@000`40000700030@0000P0
00<100401@000`400@2900030@0101h0000j008110000`40000200030@0000P000<100001`000`40
0004008100<00@00M@020@T000<1000020000`40000700030@0100<00P4200030@0008L000<10040
7P00040000@100000P4;00030@0000<00P400`01000800030@0007X000@100000P4:00030@0000@0
00<100000`4600030@0008T00P4N0000@@040@`000<100001@020@P000<10000O0040@`000<10000
10040@L000<10000Z@00054000<100001P000`40000600030@0008`000<100001P000`4000060003
0@000:T0001B00030@0000D000<100001@000`40002>00030@0000D000<100001@000`40002Z0000
DP000`40000500030@0000D000<10000SP000`40000500030@0000D000<10000ZP0005<000<10000
10000`40000400030@00090000<1000010000`40000400030@000:/0001C00030@0000@000<10000
10000`40002@00030@0000@000<1000010000`40002[0000E0000`40000300030@0000<000<10000
TP000`40000300030@0000<000<10000[00005@000<100000`000`40000300030@00098000<10000
0`000`40000300030@000:`0001E00030@00008000<100000P000`40002D00030@00008000<10000
0P000`40002]0000E@000`40000200030@00008000<10000U0000`40000200030@00008000<10000
[@0005H000D100000@0400030@0009@000<100000P001@4000010;00001F00050@0000400`000`40
002F00050@0000400`000`40002^0000EP001@40000100<000<10000UP001@40000100<000<10000
[P0005L000@100010P000`40002H00040@000@8000<10000[`0005L000@100010P000`40002H0004
0@000@8000<10000[`0005P000D100400@2L00050@010040/P0005L01P6K00H1VP060A80001F00P1
V@080IP0204A0000EP090IP02@6G00T1400005H02@6H00T1U`090A00001E00X1U`0:0IH02P4@0000
EP090IP02@6G00T1400005H02@6H00T1U`090A00001F00T1V0080IP0204A0000E`060@800P6G00H1
U`020@03004100@14P0005X000<1000010000`40002G00030@0009H000<1000010000`40000B0000
HP000`40002F00030@0009D000<100006P0006<00P6F00030@0009<00P4M0000I@000`40002C0003
0@00098000<100007@0006H000<10000TP000`40002A00030@0001h0001W0081TP000`40002?0081
8@0006T000<10000S`000`40002>00030@000240001Z0081S`000`40002=00030@000280001/0003
0@0008`000<10000R`020BD0001]00030@0008/000<10000RP000`40000U0000KP020H/000<10000
R0020BP0001`00030@0008P000<10000Q`000`40000X0000L@000`40002700030@0008H000<10000
:@0007800P6700030@0008@00P4/0000M0000`40002400030@0008<000<10000;00007D000<10000
P`000`40002200030@0002d0001f0081P`000`4000200081<00007P000<10000P0000`40001o0003
0@000300001i0081P0000`40001n00030@000340001k00030@0007d000<10000O0020C@0001l0003
0@0007`000<10000N`000`40000d0000O@020G`000<10000NP000`40000e0000O`000`40001i0003
0@0007P00P4h0000P0000`40001h00030@0007L000<10000>00008400P5h00030@0007D00P4k0000
P`000`40001e00030@0007@000<10000>`0008@000<10000M0000`40001c00030@0003`000250081
M0000`40001a0081?`0008L000<10000L@000`40001`00030@0003l000280081L@000`40001_0003
0@000400002:00030@0006h000<10000K@020D<0002;00030@0006d000<10000K0000`4000130000
S0020Fd000<10000J`000`4000140000SP000`40001Z00030@0006T00P570000S`000`40001Y0003
0@0006P000<10000A`0009000P5Y00030@0006L000<10000B000098000<10000IP000`40001U0081
B`0009<000<10000I@000`40001T00030@0004/0002D0081I@000`40001S00030@0004`0002F0003
0@00068000<10000H@020Dl0002G0081HP000`40001P00030@0004l0002I00030@0005l000<10000
GP020E80002J00030@0005h000<10000G@000`40001B0000V`020Eh000<10000G0000`40001C0000
W@000`40001K00030@0005X00P5F0000WP000`40001J00030@0005T000<10000EP0009l00P5J0003
0@0005P000<10000E`000:4000<10000E`000`40001F0081FP000:8000<10000EP000`40001E0003
0@0005X0002S0081EP000`40001D00030@0005/0002U00030@0005<000<10000DP020Eh0002V0081
D`000`40001A00030@0005h0002X00030@00050000<10000D0000`40001O0000Z@000`40001?0003
0@0004h00P5R0000ZP020Dl000<10000C@000`40001R0000[0000`40001<00030@0004/00P5U0000
[@000`40001;00030@0004X000<10000I@000:h00P5;00030@0004T000<10000IP000;0000<10000
B0000`4000170081J@000;400P5800030@0004H000<10000J@000;<000<10000A@000`4000150003
0@0006X0002d00030@0004@000<10000@`020Fd0002e0081A0000`40001200030@0006d0002g0003
0@00044000<10000@@000`40001^0000^0000`40001000030@0003l00P5a0000^@020D0000<10000
?P000`40001a0000^`000`40000m00030@0003d000<10000LP000;`000<10000?0000`40000k0081
M@000;d00P4l00030@0003X000<10000M@000;l000<10000>@000`40000h0081N0000<000P4i0003
0@0003L000<10000N0000<8000<10000=P000`40000f00030@0007T0003300030@0003D000<10000
=0020G`000340081=@000`40000c00030@0007`0003600030@00038000<10000<P000`40001m0000
a`000`40000a00030@0003000P600000b0020C4000<10000;`000`4000200000bP000`40000^0003
0@0002h000<10000P@000</000<10000;@000`40000/0081Q0000<`00P4]00030@0002/000<10000
Q0000<h000<100002P020Ah000<100000`020BD000<10000Q@000<l00P4900030@0000d000<10000
3P000`40000500030@0002000P680000d@000`40000600030@0000d000<100003P000`4000050003
0@0001l000<10000R0000=8000<100001@000`40000300L10`020@`01P4500030@0001d00P6;0000
d`020@D000<100001P000`40000500030@0000X000H10001004500030@0001`000<10000R`000=D0
00<100000P000`40000600030@0000D00P4<00030@0100L000<100006`000`40002<0000eP001040
004900030@0001@00P4800030@0001P00P6?0000e`020@03004000L000<100005@000`4000050003
0@0001P000<10000S`000=T00P4800030@0001D00P4600030@0001L000<10000T0000=X000<10000
1P000`40000E00<11@000`40000E0081T`000=X00`4600030@0001D000<100401@000`40000D0003
0@0009<0003K00<11@000`40000B00060@000@010`020AH000<10000U0000=h010400`01000C0003
0@0000<16@020IL0003P00040@010AD0104I00030@0009L0003Q00030@0001L000<100005`000`40
002H0000hP020AL000<100005@020I/0003T00030@0001@000<1000050000`40002K0000i@000`40
000C00030@0001<000<10000W0000>H00P4C00030@0001400P6O0000j0000`40000@00030@000100
00<10000W`000>T000<100003`000`40000>0081XP000>X00P4?00030@0000d000<10000XP000>`0
00<1000030000`40000<00030@000:<0003]008130000`40000:0081YP000>l000<100002@000`40
000900030@000:H0003`00030@0000P000<1000020000`40002W0000l@020@P000<100001P020JX0
003c00030@0000D000<100001@000`40002Z0000m0000`40000400030@0000@000<10000Z`000?D0
0P400`010@040@800P6^0000m`090K00003g00T1/0000?L02@6`0000mP0:0K00003g00T1/0000?L0
2@6`0000m`080K40003h00H1/P000?/000<10000/P000?l0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000001\
\>"],
  ImageRangeCache->{{{151, 438}, {403, 116}} -> {-0.628757, 0.307114, 
  0.00383279, 0.00195203}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["ShowDecisionNodes[dt2] ;", "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", GridBox[{
            {"\<\"Id\"\>", "\<\"Attribute\"\>", "\<\"Children\"\>", 
              "\<\"Examples\"\>"},
            {"1", "2", \({2, 3, 4}\), \({1, 2, 3, 4, 5, 6, 7, 8}\)},
            {"2", "1", \({5, 6, 7}\), \({3, 4, 8}\)},
            {"3", "3", \({11, 12, 13}\), \({1, 5, 7}\)},
            {"4", "0", \({}\), \({2, 6}\)},
            {"5", "3", \({8, 9, 10}\), \({4, 8}\)},
            {"6", "0", \({}\), \({3}\)},
            {"7", "0", \({}\), \({}\)},
            {"8", "0", \({}\), \({8}\)},
            {"9", "0", \({}\), \({4}\)},
            {"10", "0", \({}\), \({}\)},
            {"11", "0", \({}\), \({1}\)},
            {"12", "0", \({}\), \({}\)},
            {"13", "1", \({14, 15, 16}\), \({5, 7}\)},
            {"14", "0", \({}\), \({}\)},
            {"15", "0", \({}\), \({7}\)},
            {"16", "0", \({}\), \({5}\)}
            }], ")"}],
      (MatrixForm[ #]&)]], "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Choosing Attributes to Produce Small Decision Trees", "Section"],

Cell["\<\
Now let's see if we can improve on our random method for selecting \
an attribute to branch on. The following method choose the \"best\" attribute \
where \"best\" is defined (as in the lecture notes) as the attribute that \
maximizes the information gain.\
\>", "Text"],

Cell["\<\
FindAttribute[e_List,b_List] := 
    Module[{u = Infinity, v, w},
        Do[v = EvaluateAttribute[e,b[[i]]] ;
           If[v < u,u = v ; w = b[[i]]],
           {i,1,Length[b]}] ;
        Return[w]]\
\>", "Input"],

Cell[TextData[{
  "The following is an implementation is based on the approach described in \
the textbook and in the lecture slides.  Let ",
  StyleBox["w",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " be the set of examples associated with a given node.  Each attribute ",
  StyleBox["z",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " induces a partition ",
  StyleBox["p",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " of ",
  StyleBox["w",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ".  The expected information content (or entropy) of the partition ",
  StyleBox["p",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " of ",
  StyleBox["w",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " induced by ",
  StyleBox["z",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is defined as the sum over the blocks of ",
  StyleBox["p",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " of the entropy of each block weighted by the fraction of examples in the \
block (",
  StyleBox["i.e.,",
    FontSlant->"Italic"],
  " the number of examples in the block divided by total number of examples). \
 We don't actually maximize the information gain as in the lecture notes.  \
The information gain associated with a set of examples w and an attribute ",
  StyleBox["z",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is defined as the entropy of ",
  StyleBox["w",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " minus the expected entropy of the partition ",
  StyleBox["p",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " of ",
  StyleBox["w",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " induced by ",
  StyleBox["z",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ".  Since in deciding how to partition a node with examples ",
  StyleBox["w",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  ", the term corresponding to the entropy of ",
  StyleBox["w",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is the same no matter which attribute we are considering, it is only \
necessary to minimize the expected entropy of the resulting partition. The \
following procedure implements this idea."
}], "Text"],

Cell["\<\
EvaluateAttribute[d_List,a_Integer] := 
    Module[{s,q,r,t,h = vals[[a]],k = vals[[dims+1]]},
        r = PartitionStatistics[d,a,h,k] ;
        t = Apply[Plus,r] ;
        Return[Sum[q = Take[r,{(i-1) k + 1, i k}] ;
                   s = Apply[Plus,q] ;
                   If[s == 0,0, s/t Entropy[q,s,k]],
                   {i,1,h}]]]\
\>", "Input"],

Cell["\<\
Entropy[q_List,s_Integer,k_Integer] :=
    Sum[If[q[[i]] == 0,0,
           - q[[i]]/s Log[2, q[[i]]/s]],
        {i,1,k}]  \
\>", "Input"],

Cell["\<\
For the case of a binary attribute, we can show how the entropy \
changes as a function of a single parameter.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[
      \(-\ p\)\ Log[2, p]\  - \ \((1 - p)\)\ Log[2, \ 1\  - \ p], {p, 0, 1}]
      \ ; \)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0147151 0.588604 [
[.21429 .00222 -9 -9 ]
[.21429 .00222 9 0 ]
[.40476 .00222 -9 -9 ]
[.40476 .00222 9 0 ]
[.59524 .00222 -9 -9 ]
[.59524 .00222 9 0 ]
[.78571 .00222 -9 -9 ]
[.78571 .00222 9 0 ]
[.97619 .00222 -3 -9 ]
[.97619 .00222 3 0 ]
[.01131 .13244 -18 -4.5 ]
[.01131 .13244 0 4.5 ]
[.01131 .25016 -18 -4.5 ]
[.01131 .25016 0 4.5 ]
[.01131 .36788 -18 -4.5 ]
[.01131 .36788 0 4.5 ]
[.01131 .4856 -18 -4.5 ]
[.01131 .4856 0 4.5 ]
[.01131 .60332 -6 -4.5 ]
[.01131 .60332 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .01472 m
.21429 .02097 L
s
[(0.2)] .21429 .00222 0 1 Mshowa
.40476 .01472 m
.40476 .02097 L
s
[(0.4)] .40476 .00222 0 1 Mshowa
.59524 .01472 m
.59524 .02097 L
s
[(0.6)] .59524 .00222 0 1 Mshowa
.78571 .01472 m
.78571 .02097 L
s
[(0.8)] .78571 .00222 0 1 Mshowa
.97619 .01472 m
.97619 .02097 L
s
[(1)] .97619 .00222 0 1 Mshowa
.125 Mabswid
.07143 .01472 m
.07143 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.16667 .01472 m
.16667 .01847 L
s
.2619 .01472 m
.2619 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.35714 .01472 m
.35714 .01847 L
s
.45238 .01472 m
.45238 .01847 L
s
.5 .01472 m
.5 .01847 L
s
.54762 .01472 m
.54762 .01847 L
s
.64286 .01472 m
.64286 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.7381 .01472 m
.7381 .01847 L
s
.83333 .01472 m
.83333 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.92857 .01472 m
.92857 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .13244 m
.03006 .13244 L
s
[(0.2)] .01131 .13244 1 0 Mshowa
.02381 .25016 m
.03006 .25016 L
s
[(0.4)] .01131 .25016 1 0 Mshowa
.02381 .36788 m
.03006 .36788 L
s
[(0.6)] .01131 .36788 1 0 Mshowa
.02381 .4856 m
.03006 .4856 L
s
[(0.8)] .01131 .4856 1 0 Mshowa
.02381 .60332 m
.03006 .60332 L
s
[(1)] .01131 .60332 1 0 Mshowa
.125 Mabswid
.02381 .04415 m
.02756 .04415 L
s
.02381 .07358 m
.02756 .07358 L
s
.02381 .10301 m
.02756 .10301 L
s
.02381 .16187 m
.02756 .16187 L
s
.02381 .1913 m
.02756 .1913 L
s
.02381 .22073 m
.02756 .22073 L
s
.02381 .27959 m
.02756 .27959 L
s
.02381 .30902 m
.02756 .30902 L
s
.02381 .33845 m
.02756 .33845 L
s
.02381 .39731 m
.02756 .39731 L
s
.02381 .42674 m
.02756 .42674 L
s
.02381 .45617 m
.02756 .45617 L
s
.02381 .51503 m
.02756 .51503 L
s
.02381 .54446 m
.02756 .54446 L
s
.02381 .57389 m
.02756 .57389 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .01472 m
.04262 .09713 L
.06244 .15886 L
.10458 .26128 L
.14264 .33428 L
.18314 .39819 L
.22214 .44917 L
.26358 .49387 L
.30351 .52882 L
.34193 .55563 L
.36137 .56681 L
.38279 .57733 L
.40334 .5857 L
.42214 .59192 L
.44155 .59691 L
.45226 .59904 L
.46243 .60067 L
.47202 .60185 L
.47738 .60236 L
.48224 .60273 L
.48679 .60299 L
.48935 .60311 L
.49171 .60319 L
.49377 .60325 L
.49495 .60327 L
.49604 .60329 L
.4972 .6033 L
.49825 .60331 L
.49947 .60332 L
.50063 .60332 L
.50176 .60331 L
.50299 .6033 L
.50415 .60329 L
.50521 .60327 L
.50768 .60321 L
.51028 .60312 L
.51518 .60289 L
.52053 .60253 L
.53008 .60162 L
.53895 .60047 L
.54948 .59873 L
.55941 .59669 L
.578 .59188 L
.59801 .5852 L
.61647 .57766 L
.65587 .55698 L
.69376 .53094 L
.7341 .49611 L
.77292 .45499 L
.8142 .40197 L
Mistroke
.85396 .34014 L
.8922 .26804 L
.91356 .22067 L
.9329 .17173 L
.95174 .11609 L
.96234 .07922 L
.97208 .03832 L
.97619 .01472 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012PP10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdoojn_l00;34gTJ2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_[C^]:JVYZf][Jo^kP00
oiT0IP3o0020P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;0000?`00O`00_`0o000o
?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0
_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo
_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno
_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo
_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo
0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_mWEc:0Pl=WIfJj^[WMgMiW<09W<<iW<
IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`c
c<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co
<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI
0?nI<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<ool0/@000?l0/@000?l0/@0006<00P4700<1
10060CP00P4700<11`030CP00P4700<11@040CP00P4700<11@040D401@490000HP001040004600<1
1@001@40000103L000@100011P030@P000<10000=P001040004600<110000`40000200030@0003@0
00@100011P030@@000<100000P000`40001000030@0000T0001Q00030@00008000<100000`030@H0
00<10000=`000`40000200030@0000<00`4400H1=P000`40000200030@0000<00`4400030@000080
00<10000<`000`40000200030@0000<00`4400030@00008000<10000@0000`4000090000H@000`40
000200030@0000d000<10000=P000`40000200030@0000X000D100000@0g00030@00008000<10000
2P000`40000200030@0003<000<100000P000`40000:00030@00008000<10000@0000`4000090000
H@000`40000200030@0000h000<10000=@000`40000200030@0000/000@10001=`000`4000020003
0@0000X00P4300030@0003<000<100000P000`40000;00@1@`000`4000090000H@000`4000020003
0@0000l000<10000=0000`40000200030@0000`000<10040=`000`40000200030@0000X000<10040
0P4f00030@00008000<100002P000`40000200030@00040000<100002@00064000<100000P000`40
000?00030@0003@000<100000P000`40000<00030@0103L000<100000P000`40000;00030@0003L0
00<100000P000`40000:00030@00008000<10000@0000`4000090000HP001040004=00030@000080
00<10000=@001040004@0081>0001040004>0081>@001040004=00030@00008000<10000?P030@/0
001S00813`040CT00P4B00030@0003L00P4A00<1=`020@l0105200812`00020000<10000o`2>0000
80000`40003o08h0000P00030@000?l0SP00020000<10000o`2>000080000`40003o08h0000P0003
0@000?l0SP0001H0o`6J0@40000P00030@00010000<100004@000`40000@00030@00010000<10000
4@000`40000@00030@00014000<1000040000`40000A00030@00010000<100004@000`40000@0003
0@00014000<1000040000`40000@00030@00014000<1000040000`40000A00030@00010000<10000
4@000`400008000080000`40000@00030@00014000<1000040000`40000@00030@00014000<10000
40000`40000A00030@00010000<100004@000`40000@00030@00014000<1000040000`40000A0003
0@00010000<1000040000`40000A00030@00010000<100004@000`40000@00030@00014000<10000
200002000P5;00030@0004/000<10000B`000`40001;00030@0004X00P4:000080020Ol0P`000`40
0009000080020Ol0P`000`400009000080020Ol0P`000`400009000080020Ol0P`000`4000090000
80000`400@3o084000<100002P00020000<10040o`2100030@0000X0000P00030@010?l0P@000`40
000:000080000`400@3o084000<100002P0002000`7o084000<100002P00020000@10001o`1o0003
0@0000/0000P00040@000Ol0O`000`40000;000080001040007o07l000<100002`00020000@10001
o`1o00030@0000/0000P00040@000Ol0OP000`40000<000080001@4000010?l0O@000`40000<0000
80001@4000010?l0O@000`40000<000080001@4000010?l0O@000`40000<000080001@4000010?l0
O@000`40000<000080001@4000010?l0O0000`40000=000080000`40000200030@000?l0N@000`40
000=000080020@<000<10000o`1i00030@0000d0000P00030@00008000<10000o`1i00030@0000d0
000P00030@00008000<10000o`1h00030@0000h0000P00030@00008000<10000o`1h00030@0000h0
000P00030@0000<000<10000o`1g00030@0000h0000P00030@0000<000<10000o`1g00030@0000h0
000P00030@0000<000<10000o`1f00030@0000l0000P00030@0000<000<10000o`1f00030@0000l0
000P00030@0000<000<10000o`1f00030@0000l0000P00030@0000@000<10000o`1e00030@0000l0
000P00030@0000@000<10000o`1d00030@000100000P00030@0000@000<10000o`1d00030@000100
000P00811@000`40003o07@000<100004000020000<100001@000`40003o078000<100004@000200
00<100001@000`40003o078000<100004@00020000<100001@000`40003o078000<100004@000200
00<100001P000`40003o074000<100004@00020000<100001P000`40003o070000<100004P000200
00<100001P000`40003o070000<100004P00020000<100001`000`40003o06l000<100004P000200
00<100001`000`40003o06h000<100004`00020000<100001`000`40003o06h000<100004`0000<0
0P4700<110060@L000<1000020000`40003o06d000<100004`00008000@100011P030@D000D10000
0@0700030@0000P000<10000o`1/00030@0001@000000`01000300030@0000<00`4600030@0000P0
0`4800030@000?l0K0000`40000D000000<00@000`000`40000=00030@0000L000<100002@000`40
003o06/000<1000050000003004000<000<100003P000`40000600030@0000T000<10000o`1Z0003
0@0001D000000`01000300030@0000l000<100001@000`40000900030@000?l0JP000`40000E0000
00<00@000`000`40000?00030@0000D000<100002P000`40003o06T000<100005@00008000@10001
3@000`40000200030@0000D000<100002P000`40003o06P000<100005P0000<00P4?00@120000`40
000:00030@000?l0J0000`40000F000080000`40000;00030@000?l0IP000`40000G000080000`40
000;00030@000?l0IP000`40000G000080000`40000;00030@000?l0IP000`40000G000080000`40
000<00030@000?l0I0000`40000H000080000`40000<00030@000?l0I0000`40000H000080020@d0
00<10000o`1T00030@0001P0000P00030@0000d000<10000o`1R00030@0001T0000P00030@0000d0
00<10000o`1R00030@0001T0000P00030@0000h000<10000o`1Q00030@0001T0000P00030@0000h0
00<10000o`1P00030@0001X0000P00030@0000h000<10000o`1P00030@0001X0000P00030@0000l0
00<10000o`1N00030@0001/0000P00030@0000l000<10000o`1N00030@0001/0000P00030@000100
00<10000o`1M00030@0001/0000P00030@00010000<10000o`1L00030@0001`0000P00030@000140
00<10000o`1K00030@0001`0000P00030@00014000<10000o`1J00030@0001d0000P00814P000`40
003o05X000<100007@00020000<100004P000`40003o05T000<100007@00020000<100004P000`40
003o05P000<100007P00020000<100004`000`40003o05L000<100007P00020000<100004`000`40
003o05H000<100007`00020000<1000050000`40003o05D000<100007`00020000<1000050000`40
003o05D000<100007`00020000<1000050000`40003o05@000<100008000020000<100005@000`40
003o05<000<100008000020000<100005@000`40003o058000<100008@00020000<100005P000`40
003o054000<100008@00020000<100005P000`40003o054000<100008@0002000P4H00030@000?l0
C`000`40000R000080000`40000G00030@000?l0C`000`40000R000080000`40000G00030@000?l0
CP000`40000S000080000`40000H00030@000?l0C@000`40000S000080000`40000H00030@000?l0
C0000`40000T000080000`40000I00030@000?l0B`000`40000T000080000`40000I00030@000?l0
B`000`40000T000080000`40000J00030@000?l0B@000`40000U000080000`40000J00030@000?l0
B@000`40000U000080000`40000J00030@000?l0B0000`40000V00000`020@L00`4700<11`000`40
000K00030@000?l0A`000`40000V00000P001040004600<120000`40000600030@0001/000<10000
o`1600030@0002L000000`01000300030@0000<00`4400H11`030A`000<10000o`1500030@0002L0
00000`01000300030@0000X000D100000@0800030@0001`000<10000o`1400030@0002P000000`01
000300030@0000/000@1000120000`40000M00030@000?l0@`000`40000X000000<00@000`000`40
000<00030@0100P000<100007@000`40003o048000<10000:@000003004000<000<1000030000`40
0@0800030@0001d000<10000o`1200030@0002T0000200040@000A000P4800030@0001h000<10000
o`1100030@0002T0000300814P000`40000600030@0001h000<10000o`1000030@0002X0000P0003
0@0001l000<10000o`0o00030@0002X0000P00030@0001l000<10000o`0n00030@0002/0000P0003
0@00020000<10000o`0m00030@0002/0000P00030@00020000<10000o`0l00030@0002`0000P0003
0@00024000<10000o`0k00030@0002`0000P00818P000`40003o03X000<10000;@00020000<10000
8P000`40003o03T000<10000;@00020000<100008P000`40003o03P000<10000;P00020000<10000
8`000`40003o03L000<10000;P00020000<100008`000`40003o03H000<10000;`00020000<10000
90000`40003o03D000<10000;`00020000<1000090000`40003o03@000<10000<000020000<10000
9@000`40003o03<000<10000<000020000<100009@000`40003o038000<10000<@00020000<10000
9P000`40003o034000<10000<@00020000<100009P000`40003o030000<10000<P00020000<10000
9`000`40003o02h000<10000<`0002000P4X00030@000?l0;P000`40000c000080000`40000X0003
0@000?l0;0000`40000d000080000`40000X00030@000?l0;0000`40000d000080000`40000Y0003
0@000?l0:P000`40000e000080000`40000Y00030@000?l0:P000`40000e000080000`40000Z0003
0@000?l0:0000`40000f000080000`40000Z00030@000?l0:0000`40000f000080000`40000[0003
0@000?l09P000`40000g000080000`40000[00030@000?l09P000`40000g000080000`40000/0003
0@000?l090000`40000h000080000`40000/00030@000?l090000`40000h000080000`40000]0003
0@000?l08P000`40000i000080000`40000]00030@000?l08P000`40000i000080020Bl000<10000
o`0P00030@0003X0000P00030@0002l000<10000o`0O00030@0003X0000P00030@0002l000<10000
o`0N00030@0003/0000P00030@00030000<10000o`0L00030@0003`0000P00030@00034000<10000
o`0K00030@0003`0000P00030@00034000<10000o`0J00030@0003d0000P00030@00038000<10000
o`0H00030@0003h0000P00030@0003<000<10000o`0G00030@0003h0000P00030@0003<000<10000
o`0F00030@0003l0000300811`030@D0104800030@0003@000<10000o`0E00030@0003l000020004
0@000@H00`4400030@00008000<100001@000`40000e00030@000?l04`000`400010000000<00@00
0`000`40000300<110000`40000200030@0000D000<10000=@000`40003o018000<10000@@000003
004000<000<100002P000`40000200030@0000D00`4f00030@000?l04@000`400011000000<00@00
0`000`40000:00810`000`40000500030@0003H000<10000o`0@00030@00048000000`0100030003
0@0000X000<100400P4800030@0003L000<10000o`0>00030@0004<000000`01000300030@0000/0
00<100002@000`40000h00030@000?l03@000`40001300000P001040004>00812P000`40000h0003
0@000?l030000`40001400000`020A400`4700030@0003T000<10000o`0:00030@0004D0000P0003
0@0003X000<10000o`0900030@0004D0000P00030@0003X000<10000o`0800030@0004H0000P0003
0@0003/000<10000o`0700030@0004H0000P00030@0003`000<10000o`0500030@0004L0000P0003
0@0003`000<10000o`0400030@0004P0000P00030@0003d000<10000o`0300030@0004P0000P0081
?`000`40003o004000<10000B@00020000<10000?P000`40003o00030@0004X0000P00030@0003l0
00<10000oP000`40001:000080000`40001000030@000?`000<10000B`00020000<10000@@000`40
003j00030@0004`0000P00030@00044000<10000n@000`40001=000080000`40001200030@000?P0
00<10000C@00020000<10000@`000`40003f00030@0004h0000P00030@0004@000<10000m0000`40
001?000080000`40001400030@000?<000<10000D000020000<10000A@000`40003a00030@000540
000P00030@0004H000<10000l0000`40001A000080020DP000<10000kP000`40001B000080000`40
001800030@000>`000<10000D`00020000<10000B0000`40003[00030@0005@0000P00030@0004T0
00<10000jP000`40001D000080000`40001:00030@000>P000<10000E@00020000<10000B`000`40
003V00030@0005H0000P00030@0004/000<10000i@000`40001G000080000`40001<00030@000><0
00<10000F000020000<10000C@000`40003R00030@0005P0000P00030@0004h000<10000h0000`40
001I000080000`40001>00030@000=l000<10000FP00020000<10000C`000`40003M00030@0005/0
000P0081D@000`40003L00030@0005/0000P00030@00054000<10000fP000`40001L000080000`40
001B00030@000=P000<10000G@00020000<10000D`000`40003F00030@0005h0000P00030@0005@0
00<10000e0000`40001O000080000`40001E00030@000=8000<10000H000020000<10000EP000`40
003@00030@000640000P00030@0005L000<10000cP000`40001R000080000`40001H00030@000<`0
00<10000H`0000<00P4700<11@040@P000<10000F@000`40003:00030@0006@0000200040@000@H0
0`4400030@00008000<100001@000`40001J00030@000<P000<10000I@000003004000<000<10000
0`030@@000<100000P000`40000500030@0005/000<10000aP000`40001V000000<00@000`000`40
000:00030@00008000<100001@030E`000<10000a0000`40001W000000<00@000`000`40000;00@1
20000`40001M00030@000<8000<10000J0000003004000<000<100002P000`40000200030@0000D0
00<10000GP000`40003000030@0006T000000`01000300030@0000X000<100000P000`4000050003
0@0005l00P6o00030@0006X0000200040@000@d000<100000P000`40000500030@00064000<10000
^P020Fd0000300813`040@P000<10000HP000`40002h00030@0006d0000P00030@0006<000<10000
]P000`40001^000080000`40001T00030@000;@000<10000K`00020000<10000I@000`40002b0003
0@000700000P00030@0006H000<10000[`020G<0000P00030@0006L00P6^00030@0007<0000P0003
0@0006T000<10000ZP000`40001d000080020F/000<10000Z0000`40001e000080000`40001[0003
0@000:D00P5h000080000`40001/00030@000:<000<10000N000020000<10000K@000`40002Q0003
0@0007T0000P00030@0006h000<10000W`000`40001j000080000`40001_0081WP000`40001k0000
80000`40001a00030@0009T00P5n000080000`40001b0081V0000`40001n000080000`40001d0003
0@0009@000<10000O`00020000<10000M@020I800P62000080000`40001g00030@0008h000<10000
PP00020000<10000N0000`40002<00030@0008<0000P0081NP020HX00P66000080000`40001k0003
0@0008H000<10000QP00020000<10000O0020H@00P69000080000`40001n00030@00080000<10000
R@00020000<10000O`020Gl000<10000RP00020000<10000P@000`40001j0081S@00020000<10000
PP020GP00P6?000080000`4000240081M0020I40000P00030@0008H000<10000K`020I<0000P0003
0@0008L00P5]0081U@00020000<10000R@020FT00P6G000080000`40002;0081I@020IT0000P0081
SP020F400P6K000080000`40002?0081G@020Id0000P00030@0009400P5H00<1W`00020000<10000
T`030E800`6R000080000`40002F0081CP020JD0000P00030@0009P00`5800<1Y`00020000<10000
V`030D800`6Z000080000`40002N00@1>`030Jd0000P00030@000:80104c00@1/00001@01@470003
0@000:H0104[00@1]00001H000<100001`000`40002Z00D18P040KP0000F00030@0000L000<10000
[`090@l02P6l00005P000`40000700<1^00?0LH0000F00030@0000L000<10000o`2>00005P000`40
000700030@000?l0SP0001H000<100001`000`40003o08h0000D00<12@000`40003o08h0000E0081
2@000`40003o08h0000P00030@000?l0SP000001\
\>"],
  ImageRangeCache->{{{151, 438}, {292.938, 116}} -> {-0.666115, 0.647357, 
  0.00386179, 0.00624851}}]
}, Open  ]],

Cell[TextData[{
  "We need statistics describing the result of partitioning a set of examples \
with a given attribute. Let ",
  StyleBox["z",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " be an attribute that can take on ",
  StyleBox["h = vals[[z]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " values and suppose that the function we are trying to learn can take on ",
  
  StyleBox["k = vals[[dims+1]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " (result) values.  The following procedure takes a set of ",
  StyleBox["n",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " examples and returns a list ",
  StyleBox["r",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " of ",
  StyleBox["(h * k)",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " numbers that sum to ",
  StyleBox["n",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and have the following property: the ",
  StyleBox["j",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "th element of ",
  StyleBox["Take[r,{(i-1) k + 1, i k}]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " is the number of examples that take on jth (output) value and whose \
attribute ",
  StyleBox["z",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " takes on the ",
  StyleBox["i",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "th value."
}], "Text"],

Cell["\<\
PartitionStatistics[d_List,a_Integer,h_Integer,k_Integer] :=
    Module[{r = Table[0,{h k}],v,c},
        Do[v = d[[i]][[a]]-1 ; 
           c = d[[i]][[dims+1]] ;
           r[[ (v k) + c ]] ++ ,
           {i,1,Length[d]}] ;
        Return[r]]\
\>", "Input"],

Cell["\<\
Let's use the new method of splitting to build a decision tree for \
the sunburn data.\
\>", "Text"],

Cell["dt3 = BuildDecisionTree[exp2] ;", "Input"],

Cell[CellGroupData[{

Cell["ShowDecisionTree[dt3] ;", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 -0.112245 1.83673 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.025 w
.55051 .90816 Mdot
.14646 .5 Mdot
.55051 .5 Mdot
.95455 .5 Mdot
.04545 .09184 Mdot
.24747 .09184 Mdot
.5 Mabswid
[ ] 0 setdash
.55051 .90816 m
.14646 .5 L
s
.55051 .90816 m
.55051 .5 L
s
.55051 .90816 m
.95455 .5 L
s
.14646 .5 m
.04545 .09184 L
s
.14646 .5 m
.24747 .09184 L
s
.55051 .90816 m
.14646 .5 L
s
.55051 .90816 m
.55051 .5 L
s
.55051 .90816 m
.95455 .5 L
s
.14646 .5 m
.04545 .09184 L
s
.14646 .5 m
.24747 .09184 L
s
gsave
.52323 .85306 -80 -18 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 22 translate 1 -1 scale
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
63.000000 15.000000 moveto
(8) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(,) show
84.562500 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
90.562500 15.000000 moveto
(<) show
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
gsave
.11919 .4449 -80 -18 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 22 translate 1 -1 scale
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
63.000000 15.000000 moveto
(8) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(2) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(,) show
84.562500 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(4) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
90.562500 15.000000 moveto
(<) show
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
[(3)] .52323 .4449 0 1 Mshowa
[(4)] .92727 .4449 0 1 Mshowa
[(5)] .01818 .03673 0 1 Mshowa
[(6)] .2202 .03673 0 1 Mshowa
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/0P10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdoojn_l00;34gTJ2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_[C^]:JVYZf][Jo^kP00
oiT0IP3o0020P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;0000?`00O`00_`0o000o
?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0
_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo
_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno
_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo
_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo
0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_mWEc:0Pl=WIfJj^[WMgMiW<09W<<iW<
IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`c
c<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co
<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI
0?nI<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<ool0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@0000D0105C00@1o`1A000010000`40000200030@0004l000<100000P000`40003o04h0
000900030@0004l000<100000P000`40003o04h0000900030@0004l000<100000P000`40003o04h0
000400030@00008000<10000C`020@<000<10000o`1>000010050E8000<100400P7o054000040003
0@0005D000<10000o`1B000010000`40001E0081o`1C000010050ED00`7o0500003o0;40003o0;40
003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40
003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40000A00@1
D`040Ol0A@0000l0205?00P1o`1300003`080Dl0207o04<0000>00X1C@0:0Ol0@P0000h02P5=00X1
o`1200003P0:0Dd02P7o0480000>00X1C@0:0Ol0@P0000l0205?00P1o`1300003`080Dl0207o04<0
000A00@1D`040Ol0A@0001@000<10000DP000`40003o04D0000E00030@00050000<10000o`160000
5@000`40001@00030@000?l0AP0001D000<10000D0000`40003o04H0000E00030@00050000<10000
o`1600005P000`40001>00030@000?l0A`0001H000<10000CP000`40003o04L0000F00030@0004h0
00<10000o`1700005P000`40001>00030@000?l0A`0001L000<10000C0000`40003o04P0000G0003
0@0004`000<10000o`1800005`000`40001<00030@000?l0B00001L000<10000C0000`40003o04P0
000H00030@0004X000<10000o`19000060000`40001:00030@000?l0B@0001P000<10000BP000`40
003o04T0000H00030@0004X000<10000o`1900006@000`40001900030@000?l0B@0001T000<10000
B0000`40003o04X0000I00030@0004P000<10000o`1:00006@000`40001800030@000?l0BP0001X0
00<10000A`000`40003o04X0000J00030@0004H000<10000o`1;00006P000`40001600030@000?l0
B`0001X000<10000AP000`40003o04/0000K00030@0004D000<10000o`1;00006`000`4000140003
0@000?l0C00001/000<10000A0000`40003o04`0000K00030@0004@000<10000o`1<000070000`40
001300030@000?l0C00001`000<10000@P000`40003o04d0000L00030@00048000<10000o`1=0000
70000`40001200030@000?l0C@0001d000<10000@@000`40003o04d0000M00030@00040000<10000
o`1>00007@000`40001000030@000?l0CP0001d000<10000@0000`40003o04h0000N00030@0003l0
00<10000o`1>00007P000`40000n00030@000?l0C`0001h000<10000?P000`40003o04l0000N0003
0@0003h000<10000o`1?00007`000`40000m00030@000?l0C`0001l000<10000?0000`40003o0500
000O00030@0003`000<10000o`1@00007`000`40000l00030@000?l0D000020000<10000>`000`40
003o0500000P00030@0003X000<10000o`1A000080000`40000j00030@000?l0D@00020000<10000
>P000`40003o0540000Q00030@0003T000<10000o`1A00008@000`40000h00030@000?l0DP000240
00<10000>0000`40003o0580000Q00030@0003P000<10000o`1B00008P000`40000g00030@000?l0
DP00028000<10000=P000`40003o05<0000R00030@0003H000<10000o`1C00008P000`40000f0003
0@000?l0D`0002<000<10000=@000`40003o05<0000S00030@0003@000<10000o`1D00008`000`40
000d00030@000?l0E00002<000<10000=0000`40003o05@0000T00030@0003<000<10000o`1D0000
90000`40000c00030@000?l0E00002@000<10000<P000`40003o05D0000T00030@00038000<10000
o`1E00009@000`40000a00030@000?l0E@0002D000<10000<@000`40003o05D0000U00030@000300
00<10000o`1F00009@000`40000`00030@000?l0EP0002H000<10000;`000`40003o05H0000V0003
0@0002l000<10000o`1F00009P000`40000^00030@000?l0E`0002H000<10000;P000`40003o05L0
000W00030@0002d000<10000o`1G00009`000`40000]00030@000?l0E`0002L000<10000;0000`40
003o05P0000W00030@0002`000<10000o`1H0000:0000`40000[00030@000?l0F00002P000<10000
:`000`40003o05P0000X00030@0002X000<10000o`1I0000:0000`40000Z00030@000?l0F@0002T0
00<10000:@000`40003o05T0000Y00030@0002T000<10000o`1I0000:@000`40000X00030@000?l0
FP0002T000<10000:0000`40003o05X0000Z00030@0002L000<10000o`1J0000:P000`40000W0003
0@000?l0FP0002X000<100009P000`40003o05/0000Z00030@0002H000<10000o`1K0000:`000`40
000U00030@000?l0F`0002/000<100009@000`40003o05/0000[00030@0002@000<10000o`1L0000
:`000`40000T00030@000?l0G00002`000<100008`000`40003o05`0000/00030@0002<000<10000
o`1L0000;0000`40000R00030@000?l0G@0002`000<100008P000`40003o05d0000]00030@000240
00<10000o`1M0000;@000`40000Q00030@000?l0G@0002d000<1000080000`40003o05h0000]0003
0@00020000<10000o`1N0000;P000`40000O00030@000?l0GP0002h000<100007`000`40003o05h0
000^00030@0001h000<10000o`1O0000;P000`40000N00030@000?l0G`0002l000<100007@000`40
003o05l0000_00030@0001d000<10000o`1O0000;`000`40000M00030@000?l0G`0002l000<10000
70000`40003o0600000`00030@0001/000<10000o`1P0000<0000`40000K00030@000?l0H0000300
00<100006`000`40003o0600000`00030@0001X000<10000o`1Q0000<@000`40000I00030@000?l0
H@00034000<100006@000`40003o0640000a00030@0001T000<10000o`1Q0000<@000`40000H0003
0@000?l0HP00038000<100005`000`40003o0680000b00030@0001L000<10000o`1R0000<P000`40
000G00030@000?l0HP00038000<100005P000`40003o06<0000c00030@0001D000<10000o`1S0000
<`000`40000E00030@000?l0H`0003<000<100005@000`40003o06<0000c00030@0001@000<10000
o`1T0000=0000`40000C00030@000?l0I00003@000<100004`000`40003o06@0000d00030@0001<0
00<10000o`1T00007P020A@000<100003P020@8000<10000o`1U00007@000`40000>00030@0000@0
00<100003`000`400@3o06L0000M00030@0000h00P4500030@0000l000<10040o`1W00007@000`40
000300H11@020@D000<100001P030@H000<10040o`1W00007@000`40000400050@0000401P020@@0
00<100001`000`4000050081o`1X00007@000`40000500030@0000L00P4500030@0000801P460081
o`1X000070000`40000700030@0000d000<100000P001@40000100P000<10000o`1V00007@000`40
000700030@0000`000<100000`00104000470081o`1X00007@000`40000800030@0000/000<10000
10000`400@0700030@0009D0106/00<17P0001d000<1000020000`40000<00030@0000<000<10040
1`000`40002D00030@00008000<10000ZP000`40000M00007@000`40000300030@00008000<10000
30000`40000400811`000`40002I00030@000:H01P4N00007P020@@0104?00030@0000D000<10000
0`030I/000<10000YP001@40000101l0000g00030@0000`000<10000U`030JX000@100017`0003P0
00<100002`000`40002J00030@000:P000<100407`0003P000<100002`000`40002J00030@000:P0
00<100407`0003P000<100002`000`40002E00030@00008000<10000Z@020Al0000h00030@0000X0
00<10000U`040Jd000<100007@0003T000<100002@000`40003o06T0000i00030@0000T000<10000
o`1Y0000>@000`40000900030@000?l0J@0003T000<1000020000`40003o06X0000j00030@0000L0
00<10000o`1Z0000>P000`40000700030@000?l0JP0003X000<100001`000`40003o06X0000j0003
0@0000L000<10000o`1Z0000>`000`40000500030@000?l0J`0003/000<100001@000`40003o06/0
000k00030@0000D000<10000o`1[0000>`000`40000500030@000?l0J`0003`000<100000`000`40
003o06`0000l00030@0000<000<10000o`1/0000?0000`40000300030@000?l0K00003`000<10000
0`000`40003o06`0000m00050@000040o`1_0000?@001@4000010?l0K`0003d000D100000@3o06l0
000m00050@000040o`1_0000?P000`400@3o0700000m00@1ZP040JX0104C0000>`080JH0206V00P1
4@0003/0206V00P1YP080A40000j00X1Y00:0J@02P4@0000>P0:0J@02P6T00X1400003X02P6T00X1
Y00:0A00000j00X1Y00:0J@02P4@0000>`080JH0206V00P14@0003/0206V00P1YP080A40000m00@1
0P000`40002U00@1Z0000`400@030A<0001400030@000:H000<10000YP000`40000G0000A@000`40
002U00030@000:D000<10000600004H000<10000Y0000`40002T00030@0001T0001700030@000:<0
00<10000X`000`40000J0000B0000`40002R00030@000:8000<100006`0004T000<10000X@000`40
002Q00030@0001`0001:00030@000:0000<10000X0000`40000M0000B`000`40002O00030@0009l0
00<100007P0004`000<10000WP000`40002N00030@0001l0001=00030@0009d000<10000W@000`40
000P0000CP000`40002L00030@0009`000<100008@0004l000<10000V`000`40002K00030@000280
001@00030@0009X000<10000VP000`40000S0000D@000`40002I00030@0009T000<1000090000580
00<10000V0000`40002H00030@0002D0001C00030@0009L000<10000U`000`40000V0000E0000`40
002F00030@0009H000<100009`0005D000<10000U@000`40002E00030@0002P0001F00030@0009@0
00<10000U0000`40000Y0000E`000`40002C00030@0009<000<10000:P0005P000<10000TP000`40
002B00030@0002/0001I00030@00094000<10000T@000`40000/0000FP000`40002@00030@000900
00<10000;@0005/000<10000S`000`40002?00030@0002h0001L00030@0008h000<10000SP000`40
000_0000G@000`40002=00030@0008d000<10000<00005h000<10000S0000`40002<00030@000340
001O00030@0008/000<10000R`000`40000b0000H0000`40002:00030@0008X000<10000<`000640
00<10000R@000`40002900030@0003@0001R00030@0008P000<10000R0000`40000e0000H`000`40
002700030@0008L000<10000=P0006@000<10000QP000`40002600030@0003L0001U00030@0008D0
00<10000Q@000`40000h0000IP000`40002400030@0008@000<10000>@0006L000<10000P`000`40
002300030@0003X0001X00030@00088000<10000PP000`40000k0000J@000`40002100030@000840
00<10000?00006X000<10000P0000`40002000030@0003d0001[00030@0007l000<10000O`000`40
000n0000K0000`40001n00030@0007h000<10000?`0006d000<10000O@000`40001m00030@000400
001^00030@0007`000<10000O0000`4000110000K`000`40001k00030@0007/000<10000@P000700
00<10000NP000`40001j00030@0004<0001a00030@0007T000<10000N@000`4000140000LP000`40
001h00030@0007P000<10000A@0007<000<10000M`000`40001g00030@0004H0001d00030@0007H0
00<10000MP000`4000170000M@000`40001e00030@0007D000<10000B00007H000<10000M0000`40
001d00030@0004T0001g00030@0007<000<10000L`000`40001:0000N0000`40001b00030@000780
00<10000B`0007T000<10000L@000`40001a00030@0004`0001j00030@00070000<10000L0000`40
001=0000N`000`40001_00030@0006l000<10000CP0007`000<10000KP000`40001^00030@0004l0
001m00030@0006d000<10000K@000`40001@0000OP000`40001/00030@0006`000<10000D@0007l0
00<10000J`000`40001[00030@000580002000030@0006X000<10000JP000`40001C0000P@000`40
001Y00030@0006T000<10000E000088000<10000J0000`40001X00030@0005D0002300030@0006L0
00<10000I`000`40001F0000Q0000`40001V00030@0006H000<10000E`0008D000<10000I@000`40
001U00030@0005P0002600030@0006@000<10000I0000`40001I0000Q`000`40001S00030@0006<0
00<10000FP0008P000<10000HP000`40001R00030@0005/0002900030@00064000<10000H@000`40
001L0000RP000`40001P00030@00060000<10000G@0008/000<10000G`000`40001O00030@0005h0
002<00030@0005h000<10000GP000`40001O0000S@000`40001M00030@0005d000<10000H00008h0
00<10000G0000`40001L00030@000640002?00030@0005/000<10000F`000`40001R0000T0000`40
001J00030@0005X000<10000H`00094000<10000F@000`40001I00030@0006@0002B00030@0005P0
00<10000F0000`40001U0000T`000`40001G00030@0005L000<10000IP0009@000<10000EP000`40
001F00030@0006L0002E00030@0005D000<10000E@000`40001X0000UP000`40001D00030@0005@0
00<10000J@0009H000<10000E0000`40001D00030@0006T0002G00030@0005<000<10000D`000`40
001Z0000V0000`40001B00030@00058000<10000J`0009T000<10000D@000`40001A00030@0006`0
002J00030@00050000<10000D0000`40001]0000V`000`40001?00030@0004l000<10000KP0009`0
00<10000CP000`40001>00030@0006l0002M00030@0004d000<10000C@000`40001`0000WP000`40
001<00030@0004`000<10000L@0009l000<10000B`000`40001;00030@000780002P00030@0004X0
00<10000BP000`40001c0000X@000`40001900030@0004T000<10000M0000:8000<10000B0000`40
001800030@0007D0002S00030@0004L000<10000A`000`40001f0000Y0000`40001600030@0004H0
00<10000M`000:D000<10000A@000`40001500030@0007P0002V00030@0004@000<10000A0000`40
001i0000Y`000`40001300030@0004<000<10000NP000:P000<10000@P000`40001200030@0007/0
002Y00030@00044000<10000@@000`40001l0000ZP000`40001000030@00040000<10000O@000:/0
00<10000?`000`40000o00030@0007h0002/00030@0003h000<10000?P000`40001o0000[@000`40
000m00030@0003d000<10000P0000:h000<10000?0000`40000l00030@000840002_00030@0003/0
00<10000>`000`4000220000/0000`40000j00030@0003X000<10000P`000;4000<10000>@000`40
000i00030@0008@0002b00030@0003P000<10000>0000`4000250000/`000`40000g00030@0003L0
00<10000QP000;@000<10000=P000`40000f00030@0008L0002e00030@0003D000<10000=@000`40
00280000]P000`40000d00030@0003@000<10000R@000;L000<10000<`000`40000c00030@0008X0
002h00030@00038000<10000<P000`40002;0000^@000`40000a00030@00034000<10000S0000;X0
00<10000<0000`40000`00030@0008d0002k00030@0002l000<10000;`000`40002>0000_0000`40
000^00030@0002h000<10000S`000;d000<10000;@000`40000]00030@000900002n00030@0002`0
00<10000;0000`40002A0000_`000`40000[00030@0002/000<10000TP000<0000<10000:P000`40
000Z00030@0009<0003100030@0002T000<10000:@000`40002D0000`P000`40000X00030@0002P0
00<10000U@000<<000<100001`020Ah000<100000`020B8000<10000UP000<@000<100001@000`40
000>00030@0000d000<100001@000`40000N00030@0009L0003500030@0000@000<100003P020@h0
00<100001@000`40000M00030@0009P0003600030@0000<000<1000010050@D00P4;00D11P000`40
000L00030@0009T0003700030@00008000<100001P000`400006008130020@L000<100006`000`40
002J0000b0001@40000100P000<100001P020@`00P4700030@0001X000<10000V`000<T000<10040
2@000`40000D008120000`40000H00030@0009`0003:00030@0100P000<1000050020@L000<10000
60000`40002M0000b`020@P000<1000050020@L000<100005`000`40002N0000c0000`4000060003
0@0001@00P4700030@0001H000<10000W`000<`00P4500<150040@L000<100005@000`40002P0000
c@020@D00P4E00<11@020AL000<10000X@000<l000<100006`000`40000K00030@000:80003@0003
0@0001X000<100006P000`40002S0000d@000`40000I00030@0001T000<10000Y0000=8000<10000
60000`40000H00030@000:D0003C00030@0001L000<100005`000`40002V0000e0000`40000F0003
0@0001H000<10000Y`000=D000<100005@000`40000E00030@000:P0003F00030@0001@000<10000
50000`40002Y0000e`000`40000C00030@0001<000<10000ZP000=P000<100004P000`40000B0003
0@000:/0003I00030@00014000<100004@000`40002/0000fP000`40000@00030@00010000<10000
[@000=/000<100003`000`40000?00030@000:h0003L00030@0000h000<100003P000`40002_0000
g@000`40000=00030@0000d000<10000/0000=h000<1000030000`40000<00030@000;40003O0003
0@0000/000<100002`000`40002b0000h0000`40000:00030@0000X000<10000/`000>4000<10000
2@000`40000900030@000;@0003R00030@0000P000<1000020000`40002e0000h`000`4000070003
0@0000L000<10000]P000>@000<100001P000`40000600030@000;L0003U00030@0000D000<10000
1@000`40002h0000iP000`40000400030@0000@000<10000^@000>L000<100000`000`4000030003
0@000;X0003X00030@00008000<100000P000`40002k0000j@000`400@030@8000<10000_0000>T0
206o0000j@080Kl0003X00X1_P000>P02P6n0000j00:0Kh0003X00X1_P000>T0206o0000j@080Kl0
003[00@1`@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000001\
\>"],
  ImageRangeCache->{{{151, 438}, {403, 116}} -> {-0.628757, 0.281166, 
  0.00383279, 0.00189704}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["ShowDecisionNodes[dt3] ;", "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", GridBox[{
            {\("Id"\), \("Attribute"\), \("Children"\), \("Examples"\)},
            {"1", "1", \({2, 3, 4}\), \({1, 2, 3, 4, 5, 6, 7, 8}\)},
            {"2", "4", \({5, 6}\), \({1, 2, 4, 8}\)},
            {"3", "0", \({}\), \({3, 6, 7}\)},
            {"4", "0", \({}\), \({5}\)},
            {"5", "0", \({}\), \({2, 8}\)},
            {"6", "0", \({}\), \({1, 4}\)}
            }], ")"}],
      (MatrixForm[ #]&)]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Demonstrating Attribute Evaluation and Selection", "Subsection"],

Cell[TextData[{
  "We provide a verbose variant of ",
  StyleBox["FindAttribute",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " that prints out what it's doing on a step-by-step basis. As an aside, the \
following  implementation illustrates one of the problems with dynamically \
scoped variables. In order to write this so it worked correctly, I had to \
remember that the procedure ",
  StyleBox["EvaluateAttribute",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " makes use of ",
  StyleBox["vals",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and ",
  StyleBox["dims",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " which are available only as dynamcially scoped variables. So, let this be \
a lesson to me. In general, the list of examples don't have to include the \
complete set of training examples in an experiment, but we need the ",
  StyleBox["Experiment",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " data structure to extract the dimensional information concerning the \
examples. "
}], "Text"],

Cell["\<\
FindAttributeVerbose[Experiment[_,v_List,m_Integer],e_List,b_List] \
:= 
  Block[{vals = v, dims = m},
    Module[{u = Infinity, z, w},
        Do[z = EvaluateAttribute[e,b[[i]]] ;
           Print[\"Attribute \",b[[i]],
                 \" has an expected entropy of \", N[z]] ;
           If[z < u,u = z ; w = b[[i]]],
           {i,1,Length[b]}] ;
        Return[w]]]\
\>", "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ShowData[exp2]\ ; \)\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", GridBox[{
            {\("#"\), "1", "2", "3", "4", \("f(x)"\)},
            {"1", "1", "2", "1", "2", "1"},
            {"2", "1", "3", "2", "1", "2"},
            {"3", "2", "1", "2", "1", "2"},
            {"4", "1", "1", "2", "2", "1"},
            {"5", "3", "2", "3", "2", "1"},
            {"6", "2", "3", "3", "2", "2"},
            {"7", "2", "2", "3", "2", "2"},
            {"8", "1", "1", "1", "1", "2"}
            }], ")"}],
      (MatrixForm[ #]&)]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(FindAttributeVerbose[exp2, exp2[\([1]\)], {1, 2, 3, 4}]\)], "Input"],

Cell[BoxData[
    InterpretationBox[
      \("Attribute "\[InvisibleSpace]1
        \[InvisibleSpace]" has an expected entropy of "\[InvisibleSpace]0.5`
        \),
      SequenceForm[ "Attribute ", 1, " has an expected entropy of ", 0.5],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("Attribute "\[InvisibleSpace]2
        \[InvisibleSpace]" has an expected entropy of \
"\[InvisibleSpace]0.688721875540867056`\),
      SequenceForm[ 
      "Attribute ", 2, " has an expected entropy of ", 0.68872187554086706],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("Attribute "\[InvisibleSpace]3
        \[InvisibleSpace]" has an expected entropy of \
"\[InvisibleSpace]0.938721875540867145`\),
      SequenceForm[ 
      "Attribute ", 3, " has an expected entropy of ", 0.93872187554086706],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("Attribute "\[InvisibleSpace]4
        \[InvisibleSpace]" has an expected entropy of \
"\[InvisibleSpace]0.606844121534167868`\),
      SequenceForm[ 
      "Attribute ", 4, " has an expected entropy of ", 0.60684412153416789],
      Editable->False]], "Print"],

Cell[BoxData[
    \(1\)], "Output"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Evaluating Decision Trees on New Examples", "Section"],

Cell[TextData[{
  "Having computed a decision tree we can now use it as a function and apply \
it to elements of our domain, examples that need not include the last \
component, namely ",
  StyleBox["f(x).",
    FontFamily->"Courier",
    FontWeight->"Bold"]
}], "Text"],

Cell["\<\
ApplyDecisionTree[example_List,dims_Integer,dt_DecisionTree] :=
    ApplyDecisionTree[example,dims,dt[[1]][[1]],dt]\
\>", "Input"],

Cell[TextData[{
  "If the node is terminal then its list of examples must all have the same \
value (",
  StyleBox["UniformQ",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "), otherwise use the attribute and the children (ordered by the ranked \
attribute values) to continue on the path from the root to the appropriate \
terminal node. Note that if you get to terminal node and it has no examples \
stored there then ",
  StyleBox["ApplyDecisionTree",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " returns ",
  StyleBox["0",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "; this means that the decision tree will always be wrong. A more \
reasonable thing would be to return a random number between ",
  StyleBox["1",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " and the ",
  StyleBox["vals[[dims +1]]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " but that would make it somewhat more difficult to evaluate the decision \
tree in our simple experiments."
}], "Text"],

Cell["\<\
ApplyDecisionTree[example_List,dims_Integer,node_List,dt_\
DecisionTree] := 
    If[node[[2]] == 0,
       If[node[[4]] == {},0,
          node[[4]] [[ 1 ]] [[ dims + 1 ]]], 
       ApplyDecisionTree[example,dims,
           dt[[1]] [[ node[[3]] [[ example[[ node[[2]] ]] ]] ]],dt]] \
\>", 
  "Input"],

Cell["\<\
Given a sample consisting of a set of examples we can use the \
average error measure to evaluate a given decision tree.  Note that with this \
measure smaller numers indicate better performance.\
\>", "Text"],

Cell["\<\
EvaluateDecisionTree[q_List,dt_DecisionTree] := 
    Module[{t = Length[q], s = 0, dims = Length[q[[1]]] - 1},
        Do[s += If[q[[i]][[ dims + 1]] == 
                        ApplyDecisionTree[q[[i]],dims,dt],0,1],
           {i,1,t}] ;
        Return[s/t]]\
\>", "Input"],

Cell[TextData[{
  "As expected ",
  StyleBox["dt3",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " produces no errors on its training data ",
  StyleBox["exp2",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(EvaluateDecisionTree[exp2[\([1]\)], dt3]\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Open  ]],

Cell["\<\
We may not get such good performance if our training examples are \
not representative. To explore this possibility, let's return to the example \
of determining who sunburns and who doesn't. In the following we generalize \
on the earlier data and assume that the only people who sunburn either have \
red hair or they are blonde and they don't use lotion. The following \
generates the set of all possible examples.\
\>", "Text"],

Cell["\<\
SunburnExamples[] := 
    Module[{sunburn,examples = {}},
        vals = {3,3,3,2,2} ;
        dims = Length[vals] - 1 ;
        Do[If[Or[hair == TranslateSunburn[\"red\"],
                 And[hair == TranslateSunburn[\"blonde\"],
                     lotion == TranslateSunburn[\"no\"]]],
                 sunburn = TranslateSunburn[\"yes\"],
                 sunburn = TranslateSunburn[\"no\"]] ;
           AppendTo[examples,{hair,height,weight,lotion,sunburn}],
           {hair,1,vals[[TranslateSunburn[\"hair\"]]]},
           {height,1,vals[[TranslateSunburn[\"height\"]]]},
           {weight,1,vals[[TranslateSunburn[\"weight\"]]]},
           {lotion,1,vals[[TranslateSunburn[\"lotion\"]]]}] ;
        Return[examples]]\
\>", "Input"],

Cell["sunburn2 = SunburnExamples[] ;", "Input"],

Cell["We then sample from the set of possible examples.", "Text"],

Cell["\<\
SunburnSample[n_Integer] := 
    Table[sunburn2[[ Random[Integer,{1,n}] ]], {n}]\
\>", "Input"],

Cell["\<\
First we'll try an experiment with a small set of examples and \
evaluate the resulting decision tree using the original set of sunburn \
examples.\
\>", "Text"],

Cell["exp3 = Experiment[SunburnSample[12],{3,3,3,2,2},4] ;", "Input"],

Cell[BoxData[
    \(\(dt4\  = \ BuildDecisionTree[exp3]\ ; \)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(EvaluateDecisionTree[exp2[\([1]\)], dt4]\)], "Input"],

Cell[BoxData[
    \(1\/4\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["ShowDecisionTree[dt4] ;", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 0.0833333 1.66667 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.025 w
.5 .91667 Mdot
.04545 .08333 Mdot
.95455 .08333 Mdot
.5 Mabswid
[ ] 0 setdash
.5 .91667 m
.04545 .08333 L
s
.5 .91667 m
.95455 .08333 L
s
.5 .91667 m
.04545 .08333 L
s
.5 .91667 m
.95455 .08333 L
s
gsave
.47273 .86667 -80 -18 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 22 translate 1 -1 scale
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
63.000000 15.000000 moveto
(8) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(,) show
84.562500 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(4) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
90.562500 15.000000 moveto
(<) show
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
[(2)] .01818 .03333 0 1 Mshowa
[(3)] .92727 .03333 0 1 Mshowa
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/0P10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdoojn_l00;34gTJ2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_[C^]:JVYZf][Jo^kP00
oiT0IP3o0020P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;0000?`00O`00_`0o000o
?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0
_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo
_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno
_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo
_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo
0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_mWEc:0Pl=WIfJj^[WMgMiW<09W<<iW<
IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`c
c<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co
<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI
0?nI<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<ool0/@000?l0/@000?l0/@000?l0/@000?l0
/@0000@01P7o08@0104O00001@001@4000010?l0P`000`40000200030@0001`0000600030@000?l0
R@000`40000L00001`000`40003o08P000<10000700000P000<10000o`2400<17`0000T000<10000
o`2600030@0001`0000900030@000?l0QP000`40000L000010000`40000200030@000?l0P@000`40
000200030@0001`0000500@1o`2500@17`000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000140107o08D0104C00003`080Ol0P@080A40000?00P1o`2100P14@0000h0
2P7o07l02P4@00003P0:0Ol0O`0:0A00000>00X1o`1o00X1400000h02P7o07l02P4@00003`080Ol0
P@080A40000?00P1o`2100P14@0001401@7o08@0104C00005P000`40003o080000<100005@0001L0
00<10000o`1n00030@0001H0000G00030@000?l0OP000`40000F000060000`40003o07`000<10000
5`0001P000<10000o`1l00030@0001L0000I00030@000?l0NP000`40000H00006@000`40003o07X0
00<10000600001X000<10000o`1h00030@0001T0000J00030@000?l0N0000`40000I00006`000`40
003o07H000<100006P0001/000<10000o`1f00030@0001X0000L00030@000?l0M0000`40000K0000
7@000`40003o078000<10000700001d000<10000o`1b00030@0001`0000N00030@000?l0L0000`40
000M00007P000`40003o070000<100007@0001l000<10000o`1^00030@0001h0000O00030@000?l0
KP000`40000N000080000`40003o06`000<100007`00020000<10000o`1/00030@0001l0000Q0003
0@000?l0JP000`40000P00008@000`40003o06X000<100008000028000<10000o`1X00030@000240
000S00030@000?l0IP000`40000R00008`000`40003o06H000<100008P0002@000<10000o`1T0003
0@0002<0000T00030@000?l0I0000`40000S00009@000`40003o068000<10000900002D000<10000
o`1R00030@0002@0000V00030@000?l0H0000`40000U00009P000`40003o060000<100009@0002L0
00<10000o`1N00030@0002H0000W00030@000?l0GP000`40000V0000:0000`40003o05`000<10000
9`0002T000<10000o`1J00030@0002P0000Y00030@000?l0FP000`40000X0000:P000`40003o05P0
00<10000:@0002X000<10000o`1H00030@0002T0000[00030@000?l0EP000`40000Z0000:`000`40
003o05H000<10000:P0002`000<10000o`1D00030@0002/0000/00030@000?l0E0000`40000[0000
;@000`40003o058000<10000;00002d000<10000o`1B00030@0002`0000^00030@000?l0D0000`40
000]0000;P000`40003o050000<10000;@0002l000<10000o`1>00030@0002h0000`00030@000?l0
C0000`40000_0000<0000`40003o04`000<10000;`00034000<10000o`1:00030@000300000a0003
0@000?l0BP000`40000`0000<P000`40003o04P000<10000<@00038000<10000o`1800030@000340
000c00030@000?l0AP000`40000b0000<`000`40003o04H000<10000<P0003@000<10000o`140003
0@0003<0000d00030@000?l0A0000`40000c0000=@000`40003o048000<10000=00003H000<10000
o`1000030@0003D0000f00030@000?l0@0000`40000e0000=`000`40003o03h000<10000=P0003L0
00<10000o`0n00030@0003H0000h00030@000?l0?0000`40000g0000>0000`40003o03`000<10000
=`0003T000<10000o`0j00030@0003P0000i00030@000?l0>P000`40000h0000>P000`40003o03P0
00<10000>@0003X000<10000o`0h00030@0003T0000k00030@000?l0=P000`40000j0000?0000`40
003o03@000<10000>`0003`000<10000o`0d00030@0003/0000m00030@000?l0<P000`40000l0000
?@000`40003o038000<10000?00003h000<10000o`0`00030@0003d0000n00030@000?l0<0000`40
000m0000?`000`40003o02h000<10000?P0003l000<10000o`0^00030@0003h0001000030@000?l0
;0000`40000o0000@0000`40003o02`000<10000?`00044000<10000o`0Z00030@00040000120003
0@000?l0:0000`4000110000@P000`40003o02P000<10000@@0004<000<10000o`0V00030@000480
001300030@000?l09P000`4000120000A0000`40003o02@000<10000@`0004@000<10000o`0T0003
0@0004<0001500030@000?l08P000`4000140000A@000`40003o028000<10000A00004H000<10000
o`0P00030@0004D0001600030@000?l080000`4000150000A`000`40003o01h000<10000AP0004P0
00<10000o`0L00030@0004L0001800030@000?l070000`4000170000B@000`40003o01X000<10000
B00004T000<10000o`0J00030@0004P0001:00030@000?l060000`4000190000BP000`40003o01P0
00<10000B@0004/000<10000o`0F00030@0004X0001;00030@000?l05P000`40001:0000C0000`40
003o01@000<10000B`0004`000<10000o`0D00030@0004/0001=00030@000?l04P000`40001<0000
CP000`40003o010000<10000C@0004h000<10000o`0@00030@0004d0001?00030@000?l03P000`40
001>0000C`000`40003o00h000<10000CP00050000<10000o`0<00030@0004l0001@00030@000?l0
30000`40001?0000D@000`40003o00X000<10000D000054000<10000o`0:00030@000500001B0003
0@000?l020000`40001A0000DP000`40003o00P000<10000D@0005<000<10000o`0600030@000580
001D00030@000?l010000`40001C0000E0000`40003o00@000<10000D`0005D000<10000o`020003
0@0005@0001E00030@000?l00P000`40001D0000EP000`40003o00030@0005D0001F00030@000?l0
00<10000E@0005L000<10000o@000`40001F0000E`000`40003m00030@0005H0001H00030@000?/0
00<10000E`0005P000<10000n`000`40001G0000F@000`40003i00030@0005P0001J00030@000?L0
00<10000F@0005X000<10000m`000`40001I0000F`000`40003e00030@0005X0001K00030@000?D0
00<10000FP0005`000<10000l`000`40001K0000G0000`40003c00030@0005/0001M00030@000?40
00<10000G00005d000<10000l@000`40001L0000GP000`40003_00030@0005d0001N00030@000>l0
00<10000G@0005l000<10000k@000`40001N0000G`000`40003]00030@0005h0001P00030@000>/0
00<10000G`00064000<10000j@000`40001P0000H@000`40003Y00030@000600001R00030@000>L0
00<10000H@00068000<10000i`000`40001Q0000H`000`40003U00030@000680001S00030@000>D0
00<10000HP0006@000<10000h`000`40001S0000I0000`40003S00030@0006<0001U00030@000>40
00<10000I00006D000<10000h@000`40001T0000IP000`40003O00030@0006D0001W00030@000=d0
00<10000IP0006L000<10000g@000`40001V0000J0000`40003K00030@0006L0001X00030@000=/0
00<10000I`0006T000<10000f@000`40001X0000J@000`40003I00030@0006P0001Z00030@000=L0
00<10000J@0006X000<10000e`000`40001Y0000J`000`40003E00030@0006X0001[00030@000=D0
00<10000JP0006`000<10000d`000`40001[0000K@000`40003A00030@0006`0001]00030@000=40
00<10000K00006h000<10000c`000`40001]0000KP000`40003?00030@0006d0001_00030@000<d0
00<10000KP0006l000<10000c@000`40001^0000L0000`40003;00030@0006l0001`00030@000</0
00<10000K`00074000<10000b@000`40001`0000L@000`40003900030@000700001b00030@000<L0
00<10000L@0007<000<10000a@000`40001b0000L`000`40003500030@000780001d00030@000<<0
00<10000L`0007@000<10000``000`40001c0000M@000`40003100030@0007@0001e00030@000<40
00<10000M00007H000<10000_`000`40001e0000MP000`40002o00030@0007D0001g00030@000;d0
00<10000MP0007L000<10000_@000`40001f0000N0000`40002k00030@0007L0001i00030@000;T0
00<10000N00007T000<10000^@000`40001h0000NP000`40002g00030@0007T0001j00030@000;L0
00<10000N@0007/000<10000]@000`40001j0000N`000`40002e00030@0007X0001l00030@000;<0
00<10000N`0007`000<10000/`000`40001k0000O@000`40002a00030@0007`0001m00030@000;40
00<10000O00007h000<10000[`000`40001m0000O`000`40002]00030@0007h0001o00030@000:d0
00<10000OP00080000<10000Z`000`40001o0000P0000`40002[00030@0007l0002100030@000:T0
00<10000P000084000<10000Z@000`4000200000PP000`40002W00030@000840002200030@000:L0
00<10000P@0008<000<10000Y@000`4000220000P`000`40002U00030@000880002400030@000:<0
00<10000P`0008D000<10000X@000`4000240000Q@000`40002Q00030@0008@0002600030@0009l0
00<10000Q@0008H000<10000W`000`4000250000Q`000`40002M00030@0008H0002700030@0009d0
00<10000QP0008P000<10000V`000`4000270000R0000`40002K00030@0008L0002900030@0009T0
00<10000R00008T000<10000V@000`4000280000RP000`40002G00030@0008T0002;00030@0009D0
00<10000RP0008/000<10000U@000`40002:0000S0000`40002C00030@0008/0002<00030@0009<0
00<10000R`0008d000<10000T@000`40002<0000S@000`40002A00030@0008`0002>00030@0008l0
00<10000S@0008h000<10000S`000`40002=0000S`000`40002=00030@0008h0002?00030@0008d0
00<10000SP00090000<10000R`000`40002?0000T0000`40002;00030@0008l0002A00030@0008T0
00<10000T000098000<10000Q`000`40002A0000TP000`40002700030@000940002C00030@0008D0
00<10000TP0009<000<10000Q@000`40002B0000U0000`40002300030@0009<0002D00030@0008<0
00<10000T`0009D000<10000P@000`40002D0000U@000`40002100030@0009@0002F00030@0007l0
00<10000U@0009H000<10000O`000`40002E0000U`000`40001m00030@0009H0002H00030@0007/0
00<10000U`0009P000<10000N`000`40002G0000V@000`40001i00030@0009P0002I00030@0007T0
00<10000V00009X000<10000M`000`40002I0000VP000`40001g00030@0009T0002K00030@0007D0
00<10000VP0009/000<10000M@000`40002J0000W0000`40001c00030@0009/0002L00030@0007<0
00<10000V`0009d000<10000L@000`40002L0000WP000`40001_00030@0009d0002N00030@0006l0
00<10000W@0009l000<10000K@000`40002N0000W`000`40001]00030@0009h0002P00030@0006/0
00<10000W`000:0000<10000J`000`40002O0000X@000`40001Y00030@000:00002Q00030@0006T0
00<10000X0000:8000<10000I`000`40002Q0000XP000`40001W00030@000:40002S00030@0006D0
00<10000XP000:@000<10000H`000`40002S0000Y0000`40001S00030@000:<0002U00030@000640
00<10000Y0000:D000<10000H@000`40002T0000YP000`40001O00030@000:D0002V00030@0005l0
00<10000Y@000:L000<10000G@000`40002V0000Y`000`40001M00030@000:H0002X00030@0005/0
00<10000Y`000:P000<10000F`000`40002W0000Z@000`40001I00030@000:P0002Z00030@0005L0
00<10000Z@000:X000<10000E`000`40002Y0000Z`000`40001E00030@000:X0002[00030@0005D0
00<10000ZP000:`000<10000D`000`40002[0000[0000`40001C00030@000:/0002]00030@000540
00<10000[0000:d000<10000D@000`40002/0000[P000`40001?00030@000:d0002^00030@0004l0
00<10000[@000:l000<10000C@000`40002^0000/0000`40001;00030@000:l0002`00030@0004/0
00<10000[`000;4000<10000B@000`40002`0000/@000`40001900030@000;00002b00030@0004L0
00<10000/@000;8000<10000A`000`40002a0000/`000`40001500030@000;80002c00030@0004D0
00<10000/P000;@000<10000@`000`40002c0000]0000`40001300030@000;<0002e00030@000440
00<10000]0000;H000<10000?`000`40002e0000]P000`40000o00030@000;D0002g00030@0003d0
00<10000]P000;L000<10000?@000`40002f0000^0000`40000k00030@000;L0002h00030@0003/0
00<10000]`000;T000<10000>@000`40002h0000^@000`40000i00030@000;P0002j00030@0003L0
00<10000^@000;X000<10000=`000`40002i0000^`000`40000e00030@000;X0002l00030@0003<0
00<10000^`000;`000<10000<`000`40002k0000_@000`40000a00030@000;`0002m00030@000340
00<10000_0000;h000<10000;`000`40002m0000_P000`40000_00030@000;d0002o00030@0002d0
00<10000_P000;l000<10000;@000`40002n0000`0000`40000[00030@000;l0003000030@0002/0
00<10000_`000;L00P4800030@0001T00P4>00030@000<00002f00030@0000P000<100000`000`40
000E00030@0000/000<10000`0000;H000<100002@000`40000200815P000`40000:00030@000<40
002f00030@0000@01@400`01000200813@030@H000<100002@000`4000320000]P000`4000060005
0@00004010020@d000<100001@000`40000900030@000<80002f00030@0000H000<100000P001040
00020@T01P4600030@0000P000<10000``000;D000<100001`000`40000200030@0000`000D10000
0@0800030@0000L000<10000``000;H000<100001P000`40000300030@0000`000@100011`000`40
000700030@000<@0002f00030@0000H000<100000`000`40000=00030@0100L000<100001`000`40
00340000]P000`40000600030@0000@000<1000030000`400@0700030@0000H000<10000a@000;H0
00<1000010030@H000<100003@020@L000<100001P000`4000350000]`020@D00P4700030@0000d0
00<100000`020@P000<10000aP000<L000<100007@000`4000360000b0000`40000K00030@000<L0
003900030@0001T000<10000b0000<T000<100006@000`4000380000bP000`40000G00030@000<T0
003:00030@0001L000<10000b@000</000<100005@000`40003:0000b`000`40000E00030@000<X0
003<00030@0001<000<10000b`000<`000<100004`000`40003;0000c@000`40000A00030@000<`0
003=00030@00014000<10000c0000<h000<100003`000`40003=0000c`000`40000=00030@000<h0
003?00030@0000d000<10000cP000=0000<100002`000`40003?0000d0000`40000;00030@000<l0
003A00030@0000T000<10000d0000=4000<100002@000`40003@0000dP000`40000700030@000=40
003B00030@0000L000<10000d@000=<000<100001@000`40003B0000d`000`40000500030@000=80
003D00D100<00@00e0000=<0207E0000d`080MD0003B00X1e0000=802P7D0000dP0:0M@0003B00X1
e0000=<0207E0000d`080MD0003E00@1e`000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000001\
\>"],
  ImageRangeCache->{{{108, 395}, {644.25, 357.25}} -> {-0.463947, 0.696867, 
  0.00383279, 0.00209061}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["ShowData[exp3] ;", "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", GridBox[{
            {\("#"\), "1", "2", "3", "4", \("f(x)"\)},
            {"1", "1", "2", "3", "2", "1"},
            {"2", "1", "2", "2", "2", "1"},
            {"3", "1", "1", "2", "1", "2"},
            {"4", "1", "1", "1", "2", "1"},
            {"5", "1", "2", "1", "2", "1"},
            {"6", "1", "1", "2", "2", "1"},
            {"7", "1", "1", "3", "1", "2"},
            {"8", "1", "2", "1", "1", "2"},
            {"9", "1", "1", "3", "2", "1"},
            {"10", "1", "1", "2", "1", "2"},
            {"11", "1", "1", "3", "1", "2"},
            {"12", "1", "1", "1", "2", "1"}
            }], ")"}],
      (MatrixForm[ #]&)]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell["ShowDecisionNodes[dt4] ;", "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", GridBox[{
            {\("Id"\), \("Attribute"\), \("Children"\), \("Examples"\)},
            {"1", "4", \({2, 3}\), \({1, 2, 3, 4, 5, 6, 7, 8, 9, 3, 7, 4}\)},
            {"2", "0", \({}\), \({3, 7, 8, 3, 7}\)},
            {"3", "0", \({}\), \({1, 2, 4, 5, 6, 9, 4}\)}
            }], ")"}],
      (MatrixForm[ #]&)]], "Print"]
}, Open  ]],

Cell["\<\
Now we'll try an experiment with a somewhat larger set of examples.\
\
\>", "Text"],

Cell["exp4 = Experiment[SunburnSample[32],{3,3,3,2,2},4] ;", "Input"],

Cell[BoxData[
    \(\(dt5\  = \ BuildDecisionTree[exp4]\ ; \)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(EvaluateDecisionTree[exp2[\([1]\)], dt5]\)], "Input"],

Cell[BoxData[
    \(1\/8\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["ShowDecisionTree[dt5] ;", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0617978 1.01124 0.0918367 1.83673 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.025 w
.38764 .90816 Mdot
.05056 .5 Mdot
.72472 .5 Mdot
.5 .09184 Mdot
.72472 .09184 Mdot
.94944 .09184 Mdot
.5 Mabswid
[ ] 0 setdash
.38764 .90816 m
.05056 .5 L
s
.38764 .90816 m
.72472 .5 L
s
.72472 .5 m
.5 .09184 L
s
.72472 .5 m
.72472 .09184 L
s
.72472 .5 m
.94944 .09184 L
s
.38764 .90816 m
.05056 .5 L
s
.38764 .90816 m
.72472 .5 L
s
.72472 .5 m
.5 .09184 L
s
.72472 .5 m
.72472 .09184 L
s
.72472 .5 m
.94944 .09184 L
s
gsave
.3573 .85306 -80 -18 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 22 translate 1 -1 scale
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
63.000000 15.000000 moveto
(8) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(,) show
84.562500 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(4) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
90.562500 15.000000 moveto
(<) show
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
[(2)] .02022 .4449 0 1 Mshowa
gsave
.69438 .4449 -80 -18 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 22 translate 1 -1 scale
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
63.000000 15.000000 moveto
(8) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(3) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(,) show
84.562500 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
90.562500 15.000000 moveto
(<) show
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
[(4)] .46966 .03673 0 1 Mshowa
[(5)] .69438 .03673 0 1 Mshowa
[(6)] .9191 .03673 0 1 Mshowa
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/0P10`4000010?ooo`000?ojl000P?oG
001[0?ooAm?Cdoojn_l00;34gTJ2]<c<c;Ng]fIVI_Cdm<30`2m?CfUYJKjn_[C^]:JVYZf][Jo^kP00
oiT0IP3o0020P80008200?oo0820P6IVc3<cIP2009VIod:JYb8R8_mV<`3oo`0g??mVc<c8`8b:Q;no
_de=CB:;8^KViW=cLf=SHd92@]kNgYBDU;fm_IVIVNGUiK8R8P00Rld0002;0000?`00O`00_`0o000o
?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0
_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo
_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno
_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo
_kmookno0;no?knoOknookoo0;oo?kooOkoo_kooool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo
0?mo?omoOomo_omooono0?no?onoOono_onooooo?oooOooo_mWEc:0Pl=WIfJj^[WMgMiW<09W<<iW<
IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`c
c<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co
<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI
0?nI<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<ool0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000<T00`5L00@1G@040B<0003:00030@0005X000<100000P000`40001I00030@000080
00<1000080000<H01P5P00030@0005T000<100000P000`40000P0000aP001@400001064000<10000
F@000`40000200030@000200003700040@000E`000<100000P000`40001I00810`000`40000P0000
b0000`400@1L00D1G0000`400@020B<0003800030@0105`000<10000G`000`40000T0000b@020E`0
00<10000G`020BD0003:00030@0005X01@5O00<18P000?l0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0
/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000?l0/@000=D0105M00@1G@040AD0003C00P1
F@080ET0204C0000d`080ET0205I00P14`000=802P5G00X1E`0:0A80003B00X1E`0:0EL02P4B0000
dP0:0EL02P5G00X14P000=802P5G00X1E`0:0A80003C00P1F@080ET0204C0000d`080ET0205I00P1
4`000=D01@5L00@1G@040AD0003J00030@0005/000<10000F`000`40000G0000f`000`40001J0003
0@0005X000<1000060000=/000<10000FP000`40001J00030@0001P0003L00030@0005T000<10000
F@000`40000I0000g0000`40001I00030@0005T000<100006@000=d000<10000F0000`40001H0003
0@0001X0003M00030@0005P000<10000F0000`40000J0000gP000`40001G00030@0005L000<10000
6`000=h000<10000E`000`40001G00030@0001/0003O00030@0005H000<10000EP000`40000L0000
h0000`40001E00030@0005D000<100007@000>0000<10000E@000`40001E00030@0001d0003Q0003
0@0005@000<10000E0000`40000N0000h@000`40001D00030@0005@000<100007P000>8000<10000
D`000`40001C00030@0001l0003R00030@0005<000<10000D`000`40000O0000h`000`40001B0003
0@00058000<1000080000><000<10000DP000`40001B00030@000200003T00030@00054000<10000
D@000`40000Q0000i@000`40001@00030@00050000<100008P000>D000<10000D0000`40001@0003
0@000280003V00030@0004l000<10000C`000`40000S0000iP000`40001?00030@0004l000<10000
8`000>L000<10000CP000`40001>00030@0002@0003W00030@0004h000<10000CP000`40000T0000
j0000`40001=00030@0004d000<100009@000>P000<10000C@000`40001=00030@0002D0003Y0003
0@0004`000<10000C0000`40000V0000j@000`40001<00030@0004`000<100009P000>X000<10000
B`000`40001;00030@0002L0003[00030@0004X000<10000BP000`40000X0000j`000`40001:0003
0@0004X000<10000:0000>`000<10000B@000`40001900030@0002T0003/00030@0004T000<10000
B@000`40000Y0000k@000`40001800030@0004P000<10000:P000>d000<10000B0000`4000180003
0@0002X0003^00030@0004L000<10000A`000`40000[0000kP000`40001700030@0004L000<10000
:`000>l000<10000AP000`40001600030@0002`0003`00030@0004D000<10000A@000`40000]0000
l0000`40001500030@0004D000<10000;@000?4000<10000A0000`40001400030@0002h0003a0003
0@0004@000<10000A0000`40000^0000lP000`40001300030@0004<000<10000;`000?8000<10000
@`000`40001300030@0002l0003c00030@00048000<10000@P000`40000`0000l`000`4000120003
0@00048000<10000<0000?@000<10000@@000`40001100030@000340003d00030@00044000<10000
@@000`40000a0000m@000`40001000030@00040000<10000<P000?H000<10000?`000`40000o0003
0@0003<0003f00030@0003l000<10000?`000`40000c0000m`000`40000n00030@0003h000<10000
=0000?L000<10000?P000`40000n00030@0003@0003h00030@0003d000<10000?@000`40000e0000
n0000`40000m00030@0003d000<10000=@000?T000<10000?0000`40000l00030@0003H0003i0003
0@0003`000<10000?0000`40000f0000nP000`40000k00030@0003/000<10000=`000?/000<10000
>P000`40000j00030@0003P0003k00030@0003X000<10000>P000`40000h0000o0000`40000i0003
0@0003T000<10000>@000?`000<10000>@000`40000i00030@0003T0003m00030@0003P000<10000
>0000`40000j0000o@000`40000h00030@0003P000<10000>P000?h000<10000=`000`40000g0003
0@0003/0003n00030@0003L000<10000=`000`40000k0000o`000`40000f00030@0003H000<10000
?0000?l00@000`40000e00030@0003D000<10000?@000?l00@000`40000e00030@0003D000<10000
?@000?l00P000`40000d00030@0003@000<10000?P000?l00P000`40000d00030@0003@000<10000
?P000?l00`000`40000c00030@0003<000<10000?`000?l00`000`40000c00030@0003<000<10000
?`000?l010000`40000b00030@00038000<10000@0000?l010000`40000b00030@00038000<10000
@0000?l01@000`40000a00030@00034000<10000@@000?l01@000`40000a00030@00034000<10000
@@000?l01P000`40000`00030@00030000<10000@P000?l01`000`40000_00030@0002l000<10000
@`000?l01`000`40000_00030@0002l000<10000@`000?l020000`40000^00030@0002h000<10000
A0000?l020000`40000^00030@0002h000<10000A0000?l02@000`40000]00030@0002d000<10000
A@000?l02@000`40000]00030@0002d000<10000A@000?l02P000`40000/00030@0002`000<10000
AP000?l02P000`40000/00030@0002`000<10000AP000?l02`000`40000[00030@0002/000<10000
A`000?l030000`40000Z00030@0002X000<10000B0000?l030000`40000Z00030@0002X000<10000
B0000?l03@000`40000Y00030@0002T000<10000B@000?l03@000`40000Y00030@0002T000<10000
B@000?l03P000`40000X00030@0002P000<10000BP000?l03P000`40000X00030@0002P000<10000
BP000?l03`000`40000W00030@0002L000<10000B`000?l03`000`40000W00030@0002L000<10000
B`000?l040000`40000V00030@0002H000<10000C0000?l040000`40000V00030@0002H000<10000
C0000?l04@000`40000U00030@0002D000<10000C@000?l04P000`40000T00030@0002@000<10000
CP000?l04P000`40000T00030@0002@000<10000CP000?l04`000`40000S00030@0002<000<10000
C`000?l04`000`40000S00030@0002<000<10000C`000?l050000`40000R00030@00028000<10000
D0000?l050000`40000R00030@00028000<10000D0000?l05@000`40000Q00030@00024000<10000
D@000?l05@000`40000Q00030@00024000<10000D@000?l05P000`40000P00030@00020000<10000
DP000?l05`000`40000O00030@0001l000<10000D`000?l05`000`40000O00030@0001l000<10000
D`000?l060000`40000N00030@0001h000<10000E0000?l060000`40000N00030@0001h000<10000
E0000?l06@000`40000M00030@0001d000<10000E@000?l06@000`40000M00030@0001d000<10000
E@000?l06P000`40000L00030@0001`000<10000EP000?l06P000`40000L00030@0001`000<10000
EP000?l06`000`40000K00030@0001/000<10000E`000?l070000`40000J00030@0001X000<10000
F0000?l070000`40000J00030@0001X000<10000F0000?l07@000`40000I00030@0001T000<10000
F@000?l07@000`40000I00030@0001T000<10000F@000?l07P000`40000H00030@0001P000<10000
FP000?l07P000`40000H00030@0001P000<10000FP000?l07`000`40000G00030@0001L000<10000
F`000?l07`000`40000G00030@0001L000<10000F`000?l080000`40000F00030@0001H000<10000
G0000?l080000`40000F00030@0001H000<10000G0000?l08@000`40000E00030@0001D000<10000
G@000?l05`020@T000<1000050000`400002008140000`40001N0000o`0F00030@0000T000<10000
0P000`40000?00030@0000@000<100003@000`40001N0000o`0F00030@0000X000@100000P4@0003
0@0000@000<1000030000`40001O0000o`0F00030@0000@0104200040@00008130050@H000<10000
30000`40001O0000o`0F00030@0000<000<100000P001040004300813@000`400@0600030@0000/0
00<10000H0000?l05P000`40000800040@000@<00P4=00030@0100H000<100002`000`40001P0000
o`0E00030@0000T000D100000@0A00030@0100L000<100002@000`40001Q0000o`0F00030@0000D0
0`4400030@0000l000<100401P000`40000:00030@000640000500H1o`0;00030@0000P000<10000
0P000`40000>00030@0100H000<100002@000`40001R00001P001@4000010?l02`000`4000080003
0@0000<000<100003@000`400@0600030@0000P000<10000H`0000L000<10000o`0<00030@0000<0
00<100000P000`40000300030@0000/00`400`01000500030@0000P000<10000H`0000P000<10000
o`0<008110040@L000<100002`020@03004000<00P4:00030@0006@0000900030@000?l070000`40
000>00030@0000h000<10000I00000X000<10000o`0L00030@0000d000<100003@000`40001U0000
2P000`40003o01`000<100003@000`40000=00030@0006D0000500030@00008000<10000o`0M0003
0@0000`000<1000030000`40001V00001P040Ol080000`40000<00030@0000`000<10000IP000?l0
:`000`40000;00030@0000/000<10000I`000?l0:`000`40000;00030@0000/000<10000I`000?l0
;0000`40000:00030@0000X000<10000J0000?l0;@000`40000900030@0000T000<10000J@000?l0
;@000`40000900030@0000T000<10000J@000?l0;P000`40000800030@0000P000<10000JP000?l0
;P000`40000800030@0000P000<10000JP000?l0;`000`40000700030@0000L000<10000J`000?l0
;`000`40000700030@0000L000<10000J`000?l0<0000`40000600030@0000H000<10000K0000?l0
<0000`40000600030@0000H000<10000K0000?l0<@000`40000500030@0000D000<10000K@000?l0
<P000`40000400030@0000@000<10000KP000?l0<P000`40000400030@0000@000<10000KP000?l0
<`000`40000300030@0000<000<10000K`000?l0<`000`40000300030@0000<000<10000K`000?l0
=0000`40000200030@00008000<10000L0000?l0=0000`40000200030@00008000<10000L0000?l0
=@001@40000100<000<10000L@000?l0=@001@40000100<000<10000L@000?l0=P00104000420003
0@000780000C00@1o`0P00D1M@000140207o01`0205d00004@080Ol070080G@0000@00X1o`0J00X1
L`0001002P7o01X02P5c0000400:0Ol06P0:0G<0000@00X1o`0J00X1L`000140207o01`0205d0000
4@080Ol070080G@0000C00@10P000`40003o01T000<100400`5f00006P000`40003o01L000<10000
NP0001X000<10000o`0G00030@0007X0000K00030@000?l05@000`40001k000070000`40003o01<0
00<10000O00001d000<10000o`0A00030@0007d0000N00030@000?l03`000`40001n00007`000`40
003o00h000<10000OP0001l000<10000o`0=00030@0007l0000P00030@000?l02`000`4000200000
8@000`40003o00T000<10000P@00028000<10000o`0700030@000880000S00030@000?l01@000`40
0023000090000`40003o00@000<10000P`0002@000<10000o`0300030@0008@0000U00030@000?l0
0@000`40002500009P000`40003n00030@0008H0000W00030@000?`000<10000Q`0002P000<10000
nP000`4000280000:0000`40003j00030@0008P0000Y00030@000?P000<10000R@0002X000<10000
mP000`40002:0000:`000`40003d00030@0008/0000/00030@000?8000<10000S00002d000<10000
l@000`40002<0000;@000`40003`00030@0008d0000^00030@000>h000<10000SP0002l000<10000
k0000`40002?0000<0000`40003Z00030@000900000a00030@000>P000<10000T@00038000<10000
i`000`40002A0000<P000`40003V00030@000980000c00030@000>@000<10000T`0003@000<10000
hP000`40002D0000=@000`40003P00030@0009D0000f00030@000=h000<10000UP0003L000<10000
g@000`40002F0000=`000`40003L00030@0009L0000h00030@000=X000<10000V00003T000<10000
f0000`40002I0000>P000`40003F00030@0009X0000k00030@000=D000<10000VP0003`000<10000
d`000`40002K0000?0000`40003B00030@0009`0000m00030@000=0000<10000W@0003h000<10000
cP000`40002N0000?`000`40003<00030@0009l0001000030@000</000<10000W`00044000<10000
b@000`40002P0000@@000`40003800030@000:40001200030@000<H000<10000XP0004<000<10000
a0000`40002S0000A0000`40003200030@000:@0001500030@000<4000<10000Y00004H000<10000
_`000`40002U0000AP000`40002n00030@000:H0001700030@000;`000<10000Y`0004P000<10000
^P000`40002X0000B@000`40002i00030@000:P0001:00030@000;L000<10000Z@0004/000<10000
]@000`40002Z0000B`000`40002d00030@000:/0001<00030@000;8000<10000[00004d000<10000
/0000`40002]0000CP000`40002_00030@000:d0001?00030@000:d000<10000[P0004l000<10000
[0000`40002_0000D0000`40002Z00030@000;00001A00030@000:P000<10000/@00058000<10000
YP000`40002b0000D`000`40002U00030@000;80001D00030@000:<000<10000/`0005@000<10000
XP000`40002d0000E@000`40002P00030@000;D0001F00030@0009h000<10000]P0005L000<10000
W@000`40002f0000F0000`40002K00030@000;L0001I00030@0009T000<10000^00005T000<10000
V0000`40002i0000FP000`40002F00030@000;X0001K00030@0009@000<10000^`0005`000<10000
T`000`40002k0000G@000`40002A00030@000;`0001N00030@0008l000<10000_@0005h000<10000
SP000`40002n0000G`000`40002<00030@000;l0001P00030@0008X000<10000`000064000<10000
R@000`4000300000HP000`40002700030@000<40001S00030@0008D000<10000`P0006<000<10000
Q0000`4000330000I0000`40002200030@000<@0001U00030@00080000<10000a@0006H000<10000
O`000`4000350000I`000`40001m00030@000<H0001X00030@0007/000<10000a`0006P000<10000
NP000`4000380000J@000`40001h00030@000<T0001Z00030@0007L000<10000b@0006/000<10000
M@000`40003:0000K0000`40001c00030@000</0001]00030@00074000<10000c00006d000<10000
L0000`40003=0000KP000`40001^00030@000<h0001_00030@0006d000<10000cP00070000<10000
J`000`40003?0000L@000`40001Y00030@000=00001a00030@0006P000<10000d@00078000<10000
IP000`40003B0000L`000`40001T00030@000=<0001d00030@0006<000<10000d`0007D000<10000
H@000`40003D0000MP000`40001O00030@000=D0001f00030@0005h000<10000eP0007L000<10000
G0000`40003G0000N0000`40001K00030@000=L0001i00030@0005T000<10000f00007X000<10000
E`000`40003I0000N`000`40001E00030@000=X0001k00030@0005@000<10000f`0007`000<10000
DP000`40003L0000O@000`40001A00030@000=`0001n00030@0004l000<10000g@0007l000<10000
C@000`40003N0000P0000`40001;00030@000=l0002000030@0004X000<10000h000084000<10000
B0000`40003Q0000PP000`40001700030@000>40002300030@0004D000<10000hP0008@00`4T0081
7@000`40003S0000Q0020@l000<100005@000`40000I00030@000>@0002400813`020AH000<10000
60000`40003U0000Q0000`400@0400D11@020@d00`4600030@0001P000<10000i@0008@000@10001
1@000`40000600813@000`40000500030@0001L000<10000iP0008@000D100000@0400030@0000H0
0P4900H11P000`40000F00030@000>L0002300030@0000<000D100000@0C00050@00004020000`40
000D00030@000>P0002400030@0000<000@1000150001040004700030@0001@000<10000j@0008@0
00<100000`001040004E00030@0100L000<100004`000`40003Z0000Q0000`40000400030@0101D0
00<100401`000`40000C00030@000>X0002400030@0000@00`4F00811`000`40000B00030@000>/0
002500811@020AL000<100000`020A@000<10000k00008h000<10000;`000`40003]0000S`000`40
000]00030@000>h0002?00030@0002`000<10000k`00090000<10000:`000`40003_0000T@000`40
000Y00030@000?00002B00030@0002L000<10000l@0009<000<100009@000`40003b0000U0000`40
000S00030@000?<0002D00030@0002<000<10000l`0009D000<100008@000`40003d0000UP000`40
000O00030@000?D0002G00030@0001d000<10000mP0009P000<100006`000`40003g0000V0000`40
000J00030@000?P0002I00030@0001T000<10000n00009X000<100005`000`40003i0000V`000`40
000E00030@000?X0002L00030@0001<000<10000n`0009d000<100004@000`40003l0000W@000`40
000@00030@000?d0002N00030@0000l000<10000o@0009l000<100003@000`40003n0000X0000`40
000;00030@000?l0002Q00030@0000T000<10000o`010000XP000`40000700030@000?l00P000:80
00<100001`000`40003o0080002S00030@0100<10P000`40003o00<0002S00P1o`060000X`080Ol0
1P000:802P7o00D0002R00X1o`050000XP0:0Ol01@000:802P7o00D0002S00P1o`060000X`080Ol0
1P000:D0107o00P0003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40
003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40
003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40003o0;40
003o0;40003o0;40003o0;40003o0;40003o0;40003o0;400000\
\>"],
  ImageRangeCache->{{{108, 395}, {403, 116}} -> {-0.311022, 0.170054, 
  0.00344563, 0.00189704}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["ShowData[exp4] ;", "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", GridBox[{
            {\("#"\), "1", "2", "3", "4", \("f(x)"\)},
            {"1", "1", "2", "1", "2", "1"},
            {"2", "2", "2", "1", "1", "2"},
            {"3", "1", "2", "1", "2", "1"},
            {"4", "2", "3", "1", "1", "2"},
            {"5", "2", "1", "3", "2", "2"},
            {"6", "1", "1", "2", "2", "1"},
            {"7", "2", "2", "3", "1", "2"},
            {"8", "1", "2", "2", "1", "2"},
            {"9", "1", "2", "3", "2", "1"},
            {"10", "1", "3", "1", "1", "2"},
            {"11", "1", "2", "2", "1", "2"},
            {"12", "1", "3", "1", "2", "1"},
            {"13", "1", "2", "2", "2", "1"},
            {"14", "2", "2", "2", "1", "2"},
            {"15", "2", "2", "2", "1", "2"},
            {"16", "1", "2", "1", "2", "1"},
            {"17", "1", "3", "1", "1", "2"},
            {"18", "2", "2", "2", "2", "2"},
            {"19", "1", "2", "3", "1", "2"},
            {"20", "2", "1", "3", "1", "2"},
            {"21", "2", "2", "3", "1", "2"},
            {"22", "1", "2", "2", "1", "2"},
            {"23", "2", "2", "2", "2", "2"},
            {"24", "2", "1", "1", "2", "2"},
            {"25", "2", "2", "3", "2", "2"},
            {"26", "1", "1", "2", "1", "2"},
            {"27", "2", "1", "2", "1", "2"},
            {"28", "1", "3", "1", "2", "1"},
            {"29", "1", "2", "3", "1", "2"},
            {"30", "1", "3", "1", "1", "2"},
            {"31", "2", "1", "2", "1", "2"},
            {"32", "1", "1", "3", "2", "1"}
            }], ")"}],
      (MatrixForm[ #]&)]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell["ShowDecisionNodes[dt5] ;", "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", GridBox[{
            {\("Id"\), \("Attribute"\), \("Children"\), \("Examples"\)},
           
             {"1", "4", \({2, 3}\), 
              \({1, 2, 1, 4, 5, 6, 7, 8, 9, 10, 8, 12, 13, 14, 14, 1, 10, 18, 
                19, 20, 7, 8, 18, 24, 25, 26, 27, 12, 19, 10, 27, 32}\)},
           
             {"2", "0", \({}\), 
              \({2, 4, 7, 8, 10, 8, 14, 14, 10, 19, 20, 7, 8, 26, 27, 19, 10, 
                27}\)},
           
             {"3", "1", \({4, 5, 6}\), 
              \({1, 1, 5, 6, 9, 12, 13, 1, 18, 18, 24, 25, 12, 32}\)},
            {"4", "0", \({}\), \({1, 1, 6, 9, 12, 13, 1, 12, 32}\)},
            {"5", "0", \({}\), \({5, 18, 18, 24, 25}\)},
            {"6", "0", \({}\), \({}\)}
            }], ")"}],
      (MatrixForm[ #]&)]], "Print"]
}, Open  ]]
}, Closed]]
},
FrontEndVersion->"X 3.0",
ScreenRectangle->{{0, 1280}, {0, 1024}},
WindowSize->{1015, 715},
WindowMargins->{{59, Automatic}, {Automatic, 27}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"Magnification"->1},
Magnification->1.5,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1709, 49, 56, 0, 84, "Title"],

Cell[CellGroupData[{
Cell[1790, 53, 53, 0, 81, "Section"],
Cell[1846, 55, 669, 13, 115, "Text"],
Cell[2518, 70, 888, 26, 144, "Text"],
Cell[3409, 98, 118, 5, 128, "Input"],

Cell[CellGroupData[{
Cell[3552, 107, 33, 0, 68, "Input"],
Cell[3588, 109, 289, 8, 139, "Output"]
}, Open  ]],
Cell[3892, 120, 1030, 30, 144, "Text"],
Cell[4925, 152, 39, 0, 68, "Input"],
Cell[4967, 154, 42, 0, 68, "Input"],
Cell[5012, 156, 280, 7, 66, "Text"],
Cell[5295, 165, 84, 1, 69, "Input"],
Cell[5382, 168, 166, 4, 40, "Text"],
Cell[5551, 174, 812, 17, 141, "Text"],
Cell[6366, 193, 1385, 47, 143, "Text"],
Cell[7754, 242, 28, 0, 68, "Input"],
Cell[7785, 244, 29, 0, 68, "Input"],
Cell[7817, 246, 393, 10, 67, "Text"],
Cell[8213, 258, 1071, 15, 468, "Input"],
Cell[9287, 275, 1104, 29, 169, "Text"],
Cell[10394, 306, 343, 8, 175, "Input"],
Cell[10740, 316, 105, 3, 40, "Text"],
Cell[10848, 321, 219, 6, 148, "Input"],
Cell[11070, 329, 128, 3, 40, "Text"],
Cell[11201, 334, 237, 7, 168, "Input"],
Cell[11441, 343, 110, 3, 40, "Text"],
Cell[11554, 348, 93, 3, 68, "Input"],
Cell[11650, 353, 422, 14, 66, "Text"],
Cell[12075, 369, 47, 0, 68, "Input"],
Cell[12125, 371, 163, 4, 65, "Text"],
Cell[12291, 377, 340, 9, 208, "Input"],
Cell[12634, 388, 444, 13, 288, "Input"],
Cell[13081, 403, 509, 11, 248, "Input"],
Cell[13593, 416, 67, 0, 40, "Text"],
Cell[13663, 418, 48, 0, 68, "Input"],

Cell[CellGroupData[{
Cell[13736, 422, 33, 0, 68, "Input"],
Cell[13772, 424, 376, 9, 128, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[14185, 438, 40, 0, 68, "Input"],
Cell[14228, 440, 18016, 324, 444, 2565, 129, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[32259, 767, 467, 15, 66, "Text"],

Cell[CellGroupData[{
Cell[32751, 786, 41, 0, 68, "Input"],
Cell[32795, 788, 399, 10, 127, "Print"]
}, Open  ]],
Cell[33209, 801, 653, 10, 140, "Text"],
Cell[33865, 813, 187, 6, 40, "Text"],
Cell[34055, 821, 67, 1, 69, "Input"],

Cell[CellGroupData[{
Cell[34147, 826, 805, 21, 268, "Input"],
Cell[34955, 849, 1261, 22, 259, "Output"]
}, Open  ]],
Cell[36231, 874, 381, 8, 188, "Input"],
Cell[36615, 884, 227, 6, 148, "Input"],
Cell[36845, 892, 48, 0, 68, "Input"],

Cell[CellGroupData[{
Cell[36918, 896, 33, 0, 68, "Input"],
Cell[36954, 898, 527, 13, 224, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[37518, 916, 40, 0, 68, "Input"],
Cell[37561, 918, 35143, 787, 444, 10367, 477, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[72741, 1710, 41, 0, 68, "Input"],
Cell[72785, 1712, 959, 22, 415, "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[73793, 1740, 70, 0, 49, "Section"],
Cell[73866, 1742, 280, 5, 65, "Text"],
Cell[74149, 1749, 225, 7, 168, "Input"],
Cell[74377, 1758, 2227, 71, 245, "Text"],
Cell[76607, 1831, 364, 9, 208, "Input"],
Cell[76974, 1842, 149, 5, 62, "Input"],
Cell[77126, 1849, 134, 3, 20, "Text"],

Cell[CellGroupData[{
Cell[77285, 1856, 130, 3, 40, "Input"],
Cell[77418, 1861, 17003, 426, 185, 3753, 258, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[94436, 2290, 1389, 49, 67, "Text"],
Cell[95828, 2341, 270, 8, 88, "Input"],
Cell[96101, 2351, 110, 3, 20, "Text"],
Cell[96214, 2356, 48, 0, 40, "Input"],

Cell[CellGroupData[{
Cell[96287, 2360, 40, 0, 40, "Input"],
Cell[96330, 2362, 18340, 362, 296, 3794, 178, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[114707, 2729, 41, 0, 40, "Input"],
Cell[114751, 2731, 484, 11, 98, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[115272, 2747, 70, 0, 56, "Subsection"],
Cell[115345, 2749, 1058, 28, 79, "Text"],
Cell[116406, 2779, 395, 11, 105, "Input"],

Cell[CellGroupData[{
Cell[116826, 2794, 55, 1, 40, "Input"],
Cell[116884, 2797, 519, 13, 117, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[117440, 2815, 88, 1, 40, "Input"],
Cell[117531, 2818, 273, 6, 20, "Print"],
Cell[117807, 2826, 306, 7, 20, "Print"],
Cell[118116, 2835, 306, 7, 20, "Print"],
Cell[118425, 2844, 306, 7, 20, "Print"],
Cell[118734, 2853, 35, 1, 36, "Output"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[118830, 2861, 60, 0, 49, "Section"],
Cell[118893, 2863, 270, 7, 66, "Text"],
Cell[119166, 2872, 140, 3, 88, "Input"],
Cell[119309, 2877, 1018, 28, 144, "Text"],
Cell[120330, 2907, 311, 9, 168, "Input"],
Cell[120644, 2918, 219, 4, 31, "Text"],
Cell[120866, 2924, 285, 7, 79, "Input"],
Cell[121154, 2933, 239, 10, 20, "Text"],

Cell[CellGroupData[{
Cell[121418, 2947, 73, 1, 40, "Input"],
Cell[121494, 2950, 35, 1, 68, "Output"]
}, Open  ]],
Cell[121544, 2954, 441, 7, 115, "Text"],
Cell[121988, 2963, 755, 16, 348, "Input"],
Cell[122746, 2981, 47, 0, 68, "Input"],
Cell[122796, 2983, 65, 0, 40, "Text"],
Cell[122864, 2985, 105, 3, 88, "Input"],
Cell[122972, 2990, 171, 4, 65, "Text"],
Cell[123146, 2996, 69, 0, 40, "Input"],
Cell[123218, 2998, 74, 1, 40, "Input"],

Cell[CellGroupData[{
Cell[123317, 3003, 73, 1, 40, "Input"],
Cell[123393, 3006, 38, 1, 44, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[123468, 3012, 40, 0, 40, "Input"],
Cell[123511, 3014, 14676, 264, 296, 2150, 105, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[138224, 3283, 33, 0, 40, "Input"],
Cell[138260, 3285, 698, 17, 166, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[138995, 3307, 41, 0, 40, "Input"],
Cell[139039, 3309, 375, 8, 58, "Print"]
}, Open  ]],
Cell[139429, 3320, 93, 3, 20, "Text"],
Cell[139525, 3325, 69, 0, 40, "Input"],
Cell[139597, 3327, 74, 1, 40, "Input"],

Cell[CellGroupData[{
Cell[139696, 3332, 73, 1, 40, "Input"],
Cell[139772, 3335, 38, 1, 44, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[139847, 3341, 40, 0, 40, "Input"],
Cell[139890, 3343, 18335, 362, 296, 3786, 178, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[158262, 3710, 33, 0, 40, "Input"],
Cell[158298, 3712, 1598, 37, 411, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[159933, 3754, 41, 0, 40, "Input"],
Cell[159977, 3756, 824, 19, 206, "Print"]
}, Open  ]]
}, Closed]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

