(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    335552,      17619]*)
(*NotebookOutlinePosition[    336411,      17649]*)
(*  CellTagsIndexPosition[    336367,      17645]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["Inference in Bayesian Networks", "Title"],

Cell[CellGroupData[{

Cell["Introduction", "Section"],

Cell[TextData[{
  "Refer to Chapter 8 in the text and to the tutorial article by Bruce \
D'Ambrosio (",
  StyleBox["Inference in Bayesian Networks",
    FontSlant->"Italic"],
  ", AI Magazine, Volume ",
  StyleBox["20",
    FontWeight->"Bold"],
  ", Number ",
  StyleBox["2",
    FontWeight->"Bold"],
  ", Summer 1999). In particular, a brief summary of the set of inference \
tasks outlined in D'Ambrosio's article would help here."
}], "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Answering Queries Exactly Using Marginalization", "Section"],

Cell[TextData[{
  "Include the appropriate invocations to load the ",
  StyleBox["BayesianNetworks",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " package."
}], "Text"],

Cell["\<\
(* PrependTo[$Path,\"Power HD:Tom:programming:Mathematica\"] ; *)\
\
\>", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["PrependTo[$Path,\"/u/tld/programming/Mathematica/\"] ;", "Input",
  InitializationCell->True],

Cell["Needs[\"Dynamics`Packages`BayesianNetworks`\"]", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Brute-force method for marginalizing out variables.", "Text",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm,
  FontFamily->"Times"],

Cell["\<\
Marginalize[e_List,n_BayesianNetwork] :=
  (* The list e assigns all the fixed variables. *)
  Module[{(* The list w of all non-fixed variables. *)
          w = Complement[Table[i,{i,1,NumVars[n]}],Map[#[[1]]&,e]]},
    (* Remap non-fixed variables. *)
    Do[Remap[w[[i]]] = i,{i,1,Length[w]}] ;
    (* Define lookup function for all variables. *) 
    Module[{s = 0,a},
      Do[a = IntegerDigits[j,2,Length[w]] ; 
         s += Product[Pr[{i,LookUp[i,a,e]},
                         Map[({#,LookUp[#,a,e]})&,Parents[i,n]],
                         n],
                      {i,1,NumVars[n]}],
         {j,0,2^(Length[w])-1}] ;
      Return[s]]]\
\>", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["\<\
Convert a Bayesian network into tabular format in preparation for \
use in bucket elimination.\
\>", "Text",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm,
  FontFamily->"Times"],

Cell["\<\
TabularFormat[e_List,n_BayesianNetwork] :=
  (* The list e assigns all the fixed variables. *)
  Module[{(* The list w of all non-evidence variables. *)
          w = Complement[Table[i,{i,1,NumVars[n]}],Map[#[[1]]&,e]],
          r = {},a,p,q,t},
    Do[p = Sort[Parents[i,n]] ;
       (* The subset q of w involved with Pr(i,Parents[i]) *)
       q = Intersection[Union[p,{i}],w] ;
       If[q == {}, 
          (* If q is empty, the tabular function is constant. *)
          a = {} ;
          t = Pr[{i,LookUp[i,a,e]},
                 Map[({#,LookUp[#,a,e]})&,p],
                 n] ;
          r = Append[r,t],
          (* Otherwise t ranges over the free variables. *)
          t = Table[0,{2^(Length[q])}] ;
          (* Remap the non-fixed variables. *)
          Do[Remap[q[[i]]] = i,{i,1,Length[q]}] ;
          (* Fill in the entries in the table. *)
          Do[a = IntegerDigits[j,2,Length[q]] ;
             t[[j+1]] =
               Pr[{i,LookUp[i,a,e]},
                  Map[({#,LookUp[#,a,e]})&,p],
                  n],
             {j,0,2^(Length[q])-1}] ;
          r = Append[r,{q,t}]],
       {i,1,NumVars[n]}] ;
     Return[r]]\
\>", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["\<\
Lookup the value for a given variable. A variable's value is either \
specified in the evidence or it is provided in the current assignment. \
\>", 
  "Text",
  PageWidth->WindowWidth,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm,
  FontFamily->"Times"],

Cell["\<\
LookUp[i_Integer,a_List,e_List] := 
  If[MemberQ[e,{i,_}],
     First[Select[e,(#[[1]] == i)&]][[2]],
     If[a[[Remap[i]]] == 0, False, True]]\
\>", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[TextData[{
  "Here is the basic bucket-elimination algorithm. The dynamically scoped",
  StyleBox[" c ",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "keeps track of the number of table lookups for comparison purposes."
}], "Text",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm,
  FontFamily->"Times"],

Cell["\<\
Module[{c = 0},
 BucketCount[] := c ;
 BucketReset[] := c = 0 ;
 BucketElimination[s_List] :=
  Module[{t = s, a, r, u, w, z,
          v = Apply[Union,Map[If[ListQ[#],First[#],{}]&,s]]},
    While[Not[Equal[v,{}]], 
      (* Select a variable z to sum out next. *)
      z = SelectGreedily[t,v] ;
      (* Remove z from the list of variables. *)
      v = Complement[v,{z}] ; 
      (* Find the list u of all functions that range over z. *)
      u = Select[t,And[ListQ[#],MemberQ[#[[1]],z]]&] ; 
      (* Remove u from t.*)
      t = Complement[t,u] ; 
      (* Find the list w of all variables involved in u. *)
      w = Apply[Union,Map[First,u]] ; 
      (* Note that Union sorts and eliminates duplicates. *)
      If[Length[w]==1, 
         t = Append[t,Apply[Plus,Apply[Times,Map[#[[2]]&,u]]]] ;
         c += 2 Length[u],
         (* Initialize and create the table used to store the result. *) 
         r = Table[0,{2^(Length[w]-1)}] ;   
         (* Remap the variable names in w. *)
         Do[Remap[w[[j]]] = j,{j,1,Length[w]}] ;
         (* Let indices range from 0 to 2^Length[w]-1. *)
         Do[a = IntegerDigits[i,2,Length[w]] ; 
            r[[1+FromDigits[Drop[a,{Remap[z]}],2]]] +=  
                Product[c += 1 ;
                        u[[j]][[2]]
                         [[1+FromDigits[Part[a,Map[Remap,u[[j]][[1]]]],2]]],
                        {j,1,Length[u]}],
            {i,0,2^Length[w]-1}] ;
         t = Append[t,{Complement[w,{z}],r}]]] ;
    Return[Apply[Times,t]]]]\
\>", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["\<\
In the following, the heuristic value of selecting a given variable \
is the the number of lookups required to marginalize out that variable using \
the tabular functions that it is involved in. The smaller this number the \
better selecting it is thought to be. \
\>", "Text"],

Cell["\<\
SelectGreedily[t_List,v_List] :=
  Module[{u, y, s, z , b = Infinity},
    Do[y = v[[j]] ;
       u = Select[t,And[ListQ[#],MemberQ[#[[1]],y]]&] ; 
       s = 2^(Sum[Length[u[[i]][[1]]],{i,1,Length[u]}]) * Length[u] ;
       If[s < b, b = s ; z = y],
       {j,1,Length[v]}] ; Print[\"Selecting \",z] ;
    Return[z]]\
\>", "Input"],

Cell["\<\
Here is a random selection routine to use for purposes of \
comparison. \
\>", "Text"],

Cell["\<\
SelectRandomly[t_List,v_List] :=
  v[[Random[Integer,{1,Length[v]}]]]    \
\>", "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Examples Illustrating the Marginalization Routines", "Section"],

Cell["\<\
To test the above routines we begin by creating a simple Bayesian \
network.\
\>", "Text"],

Cell["ResetBayesianNetworks[] ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["network1 = CreateBayesianNetwork[3] ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Assign the required probabilities.", "Text"],

Cell["Pr[1,{2,3}] = 0.1 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[1,{!2,3}] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[1,{2,!3}] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[1,{!2,!3}] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[2] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[3] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["\<\
Check to see if the distribution is complete and well formed.\
\>", 
  "Text"],

Cell["CheckDistribution[network1] ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[CellGroupData[{

Cell["ShowGraph[DependencyGraph[network1],Directed] ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.1617 1.10638 0.0486285 0.97257 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.025 w
.05532 .95137 Mdot
.05532 .04863 Mdot
.94468 .5 Mdot
.5 Mabswid
[ ] 0 setdash
.05532 .04863 m
.05532 .95137 L
s
.94468 .5 m
.05532 .95137 L
s
.05532 .95137 m
.02766 .90274 L
.08298 .90274 L
F
.05532 .95137 m
.0894 .906 L
.11706 .94811 L
F
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`4000010?ooo`000?ojl2@SS_oG
001[0?ooAm?Cdoojn[34g_XC@4J2]<c<c;Ng]fIVI_Cdm2m?CfUYJKjn_/30`;C^]:JVYZf][Jo^k/c8
`8b:Q;no_eICDe5Anb:;8^KViW=cL`00oiT0I_l003_j=020P80008200?oo0820P6IVc3<cIP2009VI
od:JYb8R8_mV<`00P03oo`3o000g??mVc6=SHd92@]kNgYBDU;fm_IZDU>GPhFaXJ;8R8P00Rld009VI
V@2;000i0?o<oom0o`1^0000?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o
?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo
_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo
_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0
_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOknookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo
?oooOooo_mWEc:0Pl=WIfJj^[WMgMiW<09W<<iW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc
0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI0?nI<ol08@000?l08@000?l08@000?l08@000?l0
8@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@0000h0107o00l0000=00H1
o`0>00003@060Ol03P0000d01P7o00h0000=00H1o`0>00003P040Ol03`0000l000<10000o`0?0000
3`000`40003o00l0000?00030@000?l03`0000l000<10000o`0?00003`000`40003o00l0000?0003
0@000?l03`0000l000<10000o`0?00003`000`40003o00l0000?00030@000?l03`0000l000<10000
o`0?00003`000`40003o00l0000?00030@000?l03`0000l000<10000o`0?00003`000`40003o00l0
000?00030@000?l03`0000l000<10000o`0?00003`000`40003o00l0000?00030@000?l03`0000l0
00<10000o`0?00003`000`40003o00l0000?00030@000?l03`0000l000<10000o`0?00003`000`40
003o00l0000?00030@000?l03`0000l000<10000o`0?00003`000`40003o00l0000?00030@000?l0
3`0000l000<10000o`0?00003`000`40003o00l0000?00030@000?l03`0000l000<10000o`0?0000
3`000`40003o00l0000?00030@000?l03`0000l000<10000o`0?00003`000`40003o00l0000?0003
0@000?l03`0000l000<10000o`0?00003`000`40003o00l0000?00030@000?l03`0000l000<10000
o`0?00003`000`40003o00l0000?00030@000?l03`0000l000<10000o`0?00003`000`40003o00l0
000?00030@000?l03`0000l000<10000o`0?00003`000`40003o00l0000?00030@000?l03`0000l0
00<10000o`0?00003`000`40003o00l0000?00030@000?l03`0000l000<10000o`0?00003`000`40
003o00l0000?00030@000?l03`0000l000<10000o`0?00003`000`40003o00l0000?00030@000?l0
3`0000l000<10000o`0?00003`000`40003o00l0000?00030@000?l03`0000l000<10000o`0?0000
3`000`40003o00l0000?00030@000?l03`0000l000<10000o`0?00003`000`40003o00l0000?0003
0@000?l03`0000l000<10000o`0?00003`000`40003o00l0000?00030@000?l03`0000l000<10000
o`0?00003`000`40003o00l0000?00030@000?l03`0000l000<10000o`0?00003`000`40003o00l0
000?00030@000?l03`0000l000<10000o`0?00003`000`40003o00l0000?00030@000?l03`0000l0
00<10000o`0?00003`000`40003o00l0000?00030@000?l03`0000l000<10000o`0?00003`000`40
003o00l0000?00030@000?l03`0000l000<10000o`0?00003`000`40003o00l0000?00030@000?l0
3`0000l000<10000o`0?00003`000`40003o00l0000?00030@000?l03`0000l000<10000o`0?0000
3`000`40003o00l0000?00030@000?l03`0000l000<10000o`0?00003`000`40003o00l0000?0003
0@000?l03`0000l000<10000o`0?00003`000`40003o00l0000?00030@000?l03`0000l000<10000
o`0?00003`000`40003o00l0000?00030@000?l03`0000l000<10000o`0?00003`000`40003o00l0
000?00030@000?l03`0000l000<10000o`0?00003`000`40003o00l0000?00030@000?l03`0000l0
00<10000o`0?00003`000`40003o00l0000?00030@000?l03`0000l000<10000o`0?00003`000`40
003o00l0000?00030@000?l03`0000l000<10000o`0?00003`000`40003o00l0000?00030@000?l0
3`0000l000<10000o`0?00003`000`40003o00l0000?00030@000?l03`0000l000<10000o`0?0000
3`000`40003o00l0000?00030@000?l03`0000l000<10000o0040@h0000?00030@000?/01P4=0000
3`000`40003k00H13@0000l000<10000n`060@d0000?00030@000?/01P4=00003`000`40003i0081
00<00@400P4>00003`000`40003g00815@0000l000<10000m@020AL0000?00030@000?<00P4I0000
3`000`40003a00816`0000l000<10000k`020Ad0000?00030@000>d00P4O00003`000`40003[0081
8@0000l000<10000j@020B<0000?00030@000>L00P4U00003`000`40003U00819`0000l000<10000
h`020BT0000?00030@000>400P4[00003`000`40003O0081;@0000l000<10000g@020Bl0000?0003
0@000=/00P4a00003`000`40003I0081<`0000l000<10000e`020CD0000?00030@000=D00P4g0000
3`000`40003C0081>@0000l000<10000d@020C/0000?00030@000<l00P4m00003`000`40003=0081
?`0000l000<10000b`020D40000?00030@000<T00P5300003`000`4000370081A@0000l000<10000
a@020DL0000?00030@000<<00P5900003`000`4000310081B`0000l000<10000_`020Dd0000?0003
0@000;d00P5?00003`000`40002l00030@0004l0000?00030@000;X00P5B00003`000`40002h0081
E00000l000<10000]P020EH0000?00030@000;@00P5H00003`000`40002b0081FP0000l000<10000
/0020E`0000?00030@000:h00P5N00003`000`40002/0081H00000l000<10000ZP020F80000?0003
0@000:P00P5T00003`000`40002V0081IP0000l000<10000Y0020FP0000?00030@000:800P5Z0000
3`000`40002P0081K00000l000<10000WP020Fh0000?00030@0009`00P5`00003`000`40002J0081
LP0000l000<10000V0020G@0000?00030@0009H00P5f00003`000`40002D0081N00000l000<10000
TP020GX0000?00030@0009000P5l00003`000`40002>0081OP0000l000<10000S0020H00000?0003
0@0008X00P6200003`000`4000280081Q00000l000<10000QP020HH0000?00030@0008@00P680000
3`000`4000220081RP0000l000<10000P0020H`0000?00030@0007h00P6>00003`000`40001m0003
0@0008h0000?00030@0007/00P6A00003`000`40001i0081T`0000l000<10000M`020ID0000?0003
0@0007D00P6G00003`000`40001c0081V@0000l000<10000L@020I/0000?00030@0006l00P6M0000
3`000`40001]0081W`0000l000<10000J`020J40000?00030@0006T00P6S00003`000`40001W0081
Y@0000l000<10000I@020JL0000?00030@0006<00P6Y00003`000`40001Q0081Z`0000l000<10000
G`020Jd0000?00030@0005d00P6_00003`000`40001K0081/@0000l000<10000F@020K<0000?0003
0@0005L00P6e00003`000`40001E0081]`0000l000<10000D`020KT0000?00030@0005400P6k0000
3`000`40001?0081_@0000l000<10000C@020Kl0000?00030@0004/00P7100003`000`4000190081
``0000l000<10000A`020LD0000?00030@0004D00P7700003`000`4000130081b@0000l000<10000
@@020L/0000?00030@0003l00P7=00003`000`40000m0081c`0000l000<10000?0000`40003?0000
3`000`40000j0081dP0000l000<10000>0020M@0000?00030@0003H00P7F00003`000`40000d0081
f00000l000<10000<P020MX0000?00030@0003000P7L00003`000`40000^0081gP0000l000<10000
;0020N00000?00030@0002X00P7R00003`000`40000X0081i00000l000<100009P020NH0000?0003
0@0002@00P7X00003`000`40000R0081jP0000l000<1000080020N`0000?00030@0001h00P7^0000
3`000`40000L0081l00000l000<100006P020O80000?00030@0001P00P7d0000200?0A400P7f0000
2@0=0@<000<100002P020OP0000900d10P030@T00P7j00002P0;0@8010470081o00000X02`400`01
0@040@@00P7n00002`090@801`400`010@3o0040000;00P10P090Ol00`0000`01`400`010@080Ol0
0`0000d01@400`010@0:0Ol00P0000d01@400`010@0:0Ol00P0000h04P7o0040000=01@1o`0000d0
507o00003@060Ol03P0000d01P7o00h0000>00@1o`0?0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000
o`0Q0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000\
\>"],
  ImageRangeCache->{{{108, 395}, {403, 116}} -> {-0.19398, 0.36558, 
  0.00314933, 0.00358263}}]
}, Open  ]],

Cell["\<\
Make sure that the procedure that converts a Bayesian network into \
a tabular format works correctly.\
\>", "Text"],

Cell[CellGroupData[{

Cell["TabularFormat[{},network1] ", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{"{", 
          RowBox[{\({1, 2, 3}\), ",", 
            RowBox[{"{", 
              RowBox[{"0.5`", ",", "0.5`", ",", "0.5`", ",", "0.9`", ",", 
                StyleBox["0.5`",
                  StyleBoxAutoDelete->True,
                  PrintPrecision->1], ",", 
                StyleBox["0.5`",
                  StyleBoxAutoDelete->True,
                  PrintPrecision->1], ",", 
                StyleBox["0.5`",
                  StyleBoxAutoDelete->True,
                  PrintPrecision->1], ",", 
                StyleBox["0.1`",
                  StyleBoxAutoDelete->True,
                  PrintPrecision->1]}], "}"}]}], "}"}], ",", 
        RowBox[{"{", 
          RowBox[{\({2}\), ",", 
            RowBox[{"{", 
              RowBox[{"0.5`", ",", 
                StyleBox["0.5`",
                  StyleBoxAutoDelete->True,
                  PrintPrecision->1]}], "}"}]}], "}"}], ",", 
        RowBox[{"{", 
          RowBox[{\({3}\), ",", 
            RowBox[{"{", 
              RowBox[{"0.5`", ",", 
                StyleBox["0.5`",
                  StyleBoxAutoDelete->True,
                  PrintPrecision->1]}], "}"}]}], "}"}]}], "}"}]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["TabularFormat[{{1,True},{3,False}},network1] ", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{"{", 
          RowBox[{\({2}\), ",", 
            RowBox[{"{", 
              RowBox[{
                StyleBox["0.5`",
                  StyleBoxAutoDelete->True,
                  PrintPrecision->1], ",", 
                StyleBox["0.5`",
                  StyleBoxAutoDelete->True,
                  PrintPrecision->1]}], "}"}]}], "}"}], ",", 
        RowBox[{"{", 
          RowBox[{\({2}\), ",", 
            RowBox[{"{", 
              RowBox[{"0.5`", ",", 
                StyleBox["0.5`",
                  StyleBoxAutoDelete->True,
                  PrintPrecision->1]}], "}"}]}], "}"}], ",", "0.5`"}], 
      "}"}]], "Output"]
}, Open  ]],

Cell["\<\
Test the bucket-elimination algorithm with a simple set of \
tables.\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
BucketElimination[{{{1,2,4},{1,2,3,4,5,6,7,8}},
                   {{1,3},{1,2,3,4}},
                   {{2,4},{1,2,3,4}}}]\
\>", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]3\),
      SequenceForm[ "Selecting ", 3],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]1\),
      SequenceForm[ "Selecting ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]2\),
      SequenceForm[ "Selecting ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]4\),
      SequenceForm[ "Selecting ", 4],
      Editable->False]], "Print"],

Cell[BoxData[
    \(580\)], "Output"]
}, Open  ]],

Cell["Now try it out on the Bayesian network defined above.", "Text"],

Cell[CellGroupData[{

Cell["BucketElimination[TabularFormat[{},network1]]", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]1\),
      SequenceForm[ "Selecting ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]2\),
      SequenceForm[ "Selecting ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]3\),
      SequenceForm[ "Selecting ", 3],
      Editable->False]], "Print"],

Cell[BoxData[
    \(1.`\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
BucketElimination[TabularFormat[{{1,True},{3,False}},network1]]\
\>\
", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]2\),
      SequenceForm[ "Selecting ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    \(0.25`\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
BucketElimination[TabularFormat[{{1,True},{2,True},{3,True}},\
network1]]\
\>", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[BoxData[
    \(0.025`\)], "Output"]
}, Open  ]],

Cell["\<\
Compare with the brute-force marginalizing scheme; obviously the \
two should agree.\
\>", "Text"],

Cell[CellGroupData[{

Cell["Marginalize[{},network1]", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[BoxData[
    \(1.`\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Marginalize[{{1,True},{3,False}},network1]", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[BoxData[
    \(0.25`\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Marginalize[{{1,True},{2,True},{3,True}},network1]", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[BoxData[
    \(0.025`\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "Here's another check on the correctness of our algorithms. The following \
should be ",
  StyleBox["Pr[1,{2,3}] = 0.1",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  " as specified above."
}], "Text"],

Cell[CellGroupData[{

Cell["\<\
Divide[Marginalize[{{1,True},{2,True},{3,True}},network1],
       Marginalize[{{2,True},{3,True}},network1]]\
\>", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[BoxData[
    \(0.1`\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "Build the network that D'Ambrosio uses in his tutorial. Let ",
  StyleBox["1 = ",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "sneeze, ",
  StyleBox["2 = ",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "cold, ",
  StyleBox["3 = ",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "allergy, ",
  StyleBox["4 = ",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "cat, and ",
  StyleBox["5 = ",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "scratches."
}], "Text"],

Cell["network2 = CreateBayesianNetwork[5] ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[1,{2,3}] = 0.9 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[1,{!2,3}] = 0.7 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[1,{2,!3}] = 0.8 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[1,{!2,!3}] = 0.01 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[2] = 0.05 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[3,{4}] = 0.75 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[3,{!4}] = 0.05 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[4] = 0.02 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[5,{4}] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[5,{!4}] = 0.05 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["CheckDistribution[network2] ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[CellGroupData[{

Cell["ShowGraph[DependencyGraph[network2],Directed] ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0221976 0.973522 0.0465153 0.930307 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.025 w
.60654 .95348 Mdot
.04868 .78027 Mdot
.04868 .21973 Mdot
.60654 .04652 Mdot
.95132 .5 Mdot
.5 Mabswid
[ ] 0 setdash
.04868 .78027 m
.60654 .95348 L
s
.04868 .21973 m
.60654 .95348 L
s
.60654 .04652 m
.04868 .21973 L
s
.60654 .04652 m
.95132 .5 L
s
.60654 .95348 m
.55273 .96123 L
.56777 .91699 L
F
.60654 .95348 m
.55824 .92952 L
.59762 .90218 L
F
.04868 .21973 m
.08745 .18324 L
.10249 .22748 L
F
.95132 .5 m
.90302 .47604 L
.9424 .4487 L
F
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`4000010?ooo`000?ojl2@SS_oG
001[0?ooAm?Cdoojn[34g_XC@4J2]<c<c;Ng]fIVI_Cdm2m?CfUYJKjn_/30`;C^]:JVYZf][Jo^k/c8
`8b:Q;no_eICDe5Anb:;8^KViW=cL`00oiT0I_l003_j=020P80008200?oo0820P6IVc3<cIP2009VI
od:JYb8R8_mV<`00P03oo`3o000g??mVc6=SHd92@]kNgYBDU;fm_IZDU>GPhFaXJ;8R8P00Rld009VI
V@2;000i0?o<oom0o`1^0000?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o
?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo
_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo
_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0
_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOknookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo
?oooOooo_mWEc:0Pl=WIfJj^[WMgMiW<09W<<iW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc
0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI0?nI<ol08@000?l08@000?l08@000?l08@000?l0
8@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000:d0105_0000[0060Fh0002/00H1
KP000:`01P5^0000ZP080Fh0002V00@10`040Fl0002S00<12P000`40001]0000X0030@h000<10000
K00009d00`4B00030@0006/0002J00<15P000`40001Z0000UP040AT000<10000JP0009<00`4N0003
0@0006T0002@00<18P000`40001X0000S@030BH000<10000I`0008T0104Y00030@0006L0002600<1
;P000`40001V0000P`030C8000<10000I@0008000`4f00030@0006@0001m00<1>@000`40001T0000
N@040Cd000<10000H`0007H00`5200030@000680001c00<1AP000`40001Q0000L0030DT000<10000
H@0006`0105=00030@000600001Y00<1DP000`40001O0000IP030EH000<10000GP0006<00`5J0003
0@0005d0001P00<1G@000`40001M0000G0040F4000<10000G00005T00`5V00030@0005/0001F00<1
JP000`40001J0000D`030Fd000<10000FP0005000`5a00030@0005T0001<00@1M@000`40001H0000
B@030GX000<10000E`0004H00`5m00030@0005L0001300<1P@000`40001F0000?`040HD000<10000
E@0003`00`6:00030@0005@0000i00<1S@000`40001D0000=P030I4000<10000D`0003<00`6E0003
0@000580000_00@1V@000`40001A000060020A800`6M00030@000540000G00<13`030J4000<10000
D00001H0104<00<1Y@000`40001?00005@060@L0106Y00030@0004h0000D00L110030Jh000<10000
C@0001<0204010010@6a00030@0004d0000B00X1]@000`40001<000030040@03004100T1]P000`40
001;00002`0A0KL000<10000BP0000/04P6f00030@0004X0000;0181]`000`40001900002`060@@0
206h00030@0004P0000<00@1aP000`40001700003`000`40003400030@0004L0000@00030@000<@0
00<10000AP00014000<10000a0000`40001500004P000`40003400030@0004@0000B00030@000<@0
00<10000A00001<000<10000a0000`400013000050000`40003400030@000480000E00030@000<@0
00<10000@@0001D000<10000a0000`40001100005P000`40003400030@000400000G00030@000<@0
00<10000?`0001P000<10000a0000`40000n00006@000`40003400030@0003d0000I00030@000<@0
00<10000?@0001X000<10000a0000`40000l00006`000`40003400030@0003/0000L00030@000<@0
00<10000>P0001`000<10000a0000`40000j00007@000`40003400030@0003T0000N00030@000<@0
00<10000>00001l000<10000a0000`40000g00007`000`40003400030@0003L0000P00030@000<@0
00<10000=P00024000<10000a0000`40000e00008P000`40003400030@0003@0000R00030@000<@0
00<10000=00002<000<10000a0000`40000c000090000`40003400030@000380000U00030@000<@0
00<10000<@0002H000<10000``000`40000a00009P000`40003400030@000300000W00030@000<@0
00<10000;`0002P000<10000a0000`40000^0000:@000`40003300030@0002h0000Y00030@000<@0
00<10000;@0002X000<10000a0000`40000/0000:`000`40003400030@0002/0000/00030@000<@0
00<10000:P0002`000<10000a0000`40000Z0000;@000`40003400030@0002T0000^00030@000<@0
00<10000:00002l000<10000a0000`40000W0000<0000`40003300030@0002L0000`00030@000<@0
00<100009P00034000<10000a0000`40000U0000<P000`40003400030@0002@0000c00030@000<<0
00<10000900003<000<10000a0000`40000S0000=0000`40003400030@000280000e00030@000<@0
00<100008@0003H000<10000``000`40000Q0000=P000`40003400030@000200000g00030@000<@0
00<100007`0003P000<10000a0000`40000N0000>@000`40003300030@0001h0000i00030@000<@0
00<100007@0003X000<10000a0000`40000L0000>`000`40003400030@0001/0000l00030@000<@0
00<100006P0003d000<10000``000`40000J0000?@000`40003400030@0001T0000n00030@000<@0
00<10000600003l000<10000a0000`40000G0000@0000`40003300030@0000@00P4A0000@0000`40
003400030@0000800`4A0000@@000`40003400030@0100@14@00048000<10000a0060A4000130003
0@000<801`4A0000@`000`40003000X1400004@000<10000_P0;0A00001500030@000;/03@4@0000
AP000`40002l00/1400004L000<10000_@090A00001700030@000;l0204?0000B0000`40003000P1
3@0004T000<10000`@070@`0001:00030@000<401P4<0000BP000`40003100H1300004/000<10000
`0060@`0001<00030@000<00104=0000C@000`40003@0000C@000`40003@0000CP000`40003?0000
C`000`40003>0000D0000`40003=0000D0000`40003=0000D@000`40003<0000DP000`40003;0000
D`000`40003:0000E0000`4000390000E0000`4000390000E@000`4000380000EP000`4000370000
E`000`4000360000E`000`4000360000F0000`4000350000F@000`4000340000FP000`4000330000
FP000`4000330000F`000`4000320000G0000`4000310000G@000`4000300000GP000`40002o0000
GP000`40002o0000G`000`40002n0000H0000`40002m0000H@000`40002l0000H@000`40002l0000
HP000`40002k0000H`000`40002j0000I0000`40002i0000I0000`40002i0000I@000`40002h0000
IP000`40002g0000I`000`40002f0000I`000`40002f0000J0000`40002e0000J@000`40002d0000
JP000`40002c0000J`000`40002b0000J`000`40002b0000K0000`40002a0000K@000`40002`0000
KP000`40002_0000KP000`40002_0000K`000`40002^0000L0000`40002]0000L@000`40002/0000
L@000`40002/0000LP000`40002[0000L`000`40002Z0000M0000`40002Y0000M@000`40002X0000
M@000`40002X0000MP000`40002W0000M`000`40002V0000N0000`40002U0000N0000`40002U0000
N@000`40002T0000NP000`40002S0000N`000`40002R0000N`000`40002R0000O0000`40002Q0000
O@000`40002P0000OP000`40002O0000OP000`40002O0000O`000`40002N0000P0000`40002M0000
P@000`40002L0000PP000`40002K0000PP000`40002K0000P`000`40002J0000Q0000`40002I0000
Q@000`40002H000030040GD000<10000V00000/01P5e00030@0009L0000;00H1MP000`40002F0000
2`060GL000<10000U@0000/01`5f00030@0009D0000<00@10P040G<000<10000U00001H00`5a0003
0@0009<0000I00<1K`000`40002B000070030Fd000<10000T@0001l00`5Z00030@000940000R00@1
I`000`40002@00009P030FD000<10000S`0002T00`5S00030@0008h0000/00<1H0000`40002>0000
;`040Ed000<10000S@0003<00`5K00030@0008`0000f00<1F@000`40002;0000>@030EH000<10000
R`0003`00`5D00030@0008X0000o00@1D@000`4000290000@`030Dl000<10000R00004H00`5<0003
0@0008P0001900<1BP000`4000270000C0040DL000<10000QP0005000`5500030@0008D0001C00<1
@`000`4000240000EP030D0000<10000Q00005T00`4n00030@0008<0001L00@1>`000`4000220000
H0030CT000<10000P@0006<00`4f00030@000840001V00<1=0000`4000200000J@030C8000<10000
O`0006`0104_00030@0007h0001`00<1;0000`40001n0000L`030BX000<10000O@0007H00`4X0003
0@0007`0001i00@19@000`40001k0000O@030B<000<10000NP0008000`4P00030@0000@00P5d0000
P`030Ah000<100000P030G@0002600<170000`400@040G@0002900@16@060G@0002=00<15@070G@0
002@00<13`0;0G<0002C00<1300;0G<0002F00@11P0=0G<0002J00<1100<0G<0002M00<100<00@40
2P5c0000X00>0G80002P0141K`000:004P5^0000W`0C0Fh0002O01<1KP0009l01`4600H1KP000:d0
105_0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000
o`0Q0000o`0Q0000\
\>"],
  ImageRangeCache->{{{108, 395}, {563, 276}} -> {-0.363749, 0.98372, 
  0.00357912, 0.00374538}}]
}, Open  ]],

Cell["\<\
In the following query we just recover the prior probability for \
cold.\
\>", "Text"],

Cell[CellGroupData[{

Cell["BucketElimination[TabularFormat[{{2,True}},network2]]", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]1\),
      SequenceForm[ "Selecting ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]5\),
      SequenceForm[ "Selecting ", 5],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]3\),
      SequenceForm[ "Selecting ", 3],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]4\),
      SequenceForm[ "Selecting ", 4],
      Editable->False]], "Print"],

Cell[BoxData[
    \(0.05`\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Marginalize[{{2,True}},network2]", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[BoxData[
    \(0.05`\)], "Output"]
}, Open  ]],

Cell["\<\
In this query we compute the prior probability that we have a cold \
and are having an allergy attack.\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
BucketElimination[TabularFormat[{{2,True},{3,True}},network2]]\
\>",
   "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]1\),
      SequenceForm[ "Selecting ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]5\),
      SequenceForm[ "Selecting ", 5],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]4\),
      SequenceForm[ "Selecting ", 4],
      Editable->False]], "Print"],

Cell[BoxData[
    \(0.00320000000000000017`\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Marginalize[{{2,True},{3,True}},network2]", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[BoxData[
    \(0.00320000000000000062`\)], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Answering Queries Approximately Using Random Sampling", "Section"],

Cell["\<\
RandomSampleWithEvidence[e_List,n_BayesianNetwork] := 
  Module[{v = TopologicalSort[DependencyGraph[n]],
          s = Table[Null,{NumVars[n]}]},
    Do[s[[e[[i]][[1]]]] = e[[i]][[2]],
       {i,Length[e]}] ;
    Do[If[SameQ[s[[v[[i]]]],Null],
          s[[v[[i]]]] = 
            If[Random[] > Pr[{v[[i]],True},
                             Map[({#,s[[#]]})&,
                                 Parents[v[[i]],n]],
                             n],
               False,True]],
       {i,NumVars[n]}] ;
    Return[s]]\
\>", "Input"],

Cell[CellGroupData[{

Cell["RandomSampleWithEvidence[{{2,True}},network1]", "Input"],

Cell[BoxData[
    \({False, True, True}\)], "Output"]
}, Open  ]],

Cell["\<\
LikelihoodWeighting[e_List,n_BayesianNetwork,m_Integer] := 
  Module[{t = 0,s,v,w},
    v = Table[0,{NumVars[n]}] ;
    Do[s = RandomSampleWithEvidence[e,n] ;
       w = Product[Pr[e[[i]],
                      Map[({#,s[[#]]})&,
                          Parents[e[[i]][[1]],n]],
                      n],
                   {i,Length[e]}] ;
       v += w Table[If[s[[i]],1,0],{i,Length[s]}] ;
       (* Do[If[i[[s]],v += w],{i,Length[s]}] *)
       t += w,  
       {m}] ;
    Return[v / t]]\
\>", "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(LikelihoodWeighting[{}, network1, 100]\)], "Input"],

Cell[BoxData[
    \({17\/50, 21\/50, 11\/20}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(LikelihoodWeighting[{{2, True}, {3, True}}, network1, 100]\)], "Input"],

Cell[BoxData[
    \({0.1`, 1.`, 1.`}\)], "Output"]
}, Open  ]],

Cell["\<\
GraphLikelihoodWeighting[e_List,n_BayesianNetwork,
                         m_Integer,j_Integer] := 
  Module[{t = 0, d = {}, v = 0, s, w},
    Do[s = RandomSampleWithEvidence[e,n] ;
       w = Product[Pr[e[[i]],
                      Map[({#,s[[#]]})&,
                          Parents[e[[i]][[1]],n]],
                      n],
                   {i,Length[e]}] ;
       v += w If[s[[j]],1,0] ;
       t += w ;  
       d = Append[d,v/t],
       {m}] ;
    ListPlot[d,PlotJoined -> True] ;
    Return[Null]] \
\>", "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(GraphLikelihoodWeighting[{{2, True}, {3, True}}, network1, 5000, 1]\)], 
  "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.000190476 -2.40462 27.8721 [
[.21429 .37009 -12 -9 ]
[.21429 .37009 12 0 ]
[.40476 .37009 -12 -9 ]
[.40476 .37009 12 0 ]
[.59524 .37009 -12 -9 ]
[.59524 .37009 12 0 ]
[.78571 .37009 -12 -9 ]
[.78571 .37009 12 0 ]
[.97619 .37009 -12 -9 ]
[.97619 .37009 12 0 ]
[.01131 .10387 -24 -4.5 ]
[.01131 .10387 0 4.5 ]
[.01131 .24323 -30 -4.5 ]
[.01131 .24323 0 4.5 ]
[.01131 .52195 -30 -4.5 ]
[.01131 .52195 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .38259 m
.21429 .38884 L
s
[(1000)] .21429 .37009 0 1 Mshowa
.40476 .38259 m
.40476 .38884 L
s
[(2000)] .40476 .37009 0 1 Mshowa
.59524 .38259 m
.59524 .38884 L
s
[(3000)] .59524 .37009 0 1 Mshowa
.78571 .38259 m
.78571 .38884 L
s
[(4000)] .78571 .37009 0 1 Mshowa
.97619 .38259 m
.97619 .38884 L
s
[(5000)] .97619 .37009 0 1 Mshowa
.125 Mabswid
.0619 .38259 m
.0619 .38634 L
s
.1 .38259 m
.1 .38634 L
s
.1381 .38259 m
.1381 .38634 L
s
.17619 .38259 m
.17619 .38634 L
s
.25238 .38259 m
.25238 .38634 L
s
.29048 .38259 m
.29048 .38634 L
s
.32857 .38259 m
.32857 .38634 L
s
.36667 .38259 m
.36667 .38634 L
s
.44286 .38259 m
.44286 .38634 L
s
.48095 .38259 m
.48095 .38634 L
s
.51905 .38259 m
.51905 .38634 L
s
.55714 .38259 m
.55714 .38634 L
s
.63333 .38259 m
.63333 .38634 L
s
.67143 .38259 m
.67143 .38634 L
s
.70952 .38259 m
.70952 .38634 L
s
.74762 .38259 m
.74762 .38634 L
s
.82381 .38259 m
.82381 .38634 L
s
.8619 .38259 m
.8619 .38634 L
s
.9 .38259 m
.9 .38634 L
s
.9381 .38259 m
.9381 .38634 L
s
.25 Mabswid
0 .38259 m
1 .38259 L
s
.02381 .10387 m
.03006 .10387 L
s
[(0.09)] .01131 .10387 1 0 Mshowa
.02381 .24323 m
.03006 .24323 L
s
[(0.095)] .01131 .24323 1 0 Mshowa
.02381 .52195 m
.03006 .52195 L
s
[(0.105)] .01131 .52195 1 0 Mshowa
.125 Mabswid
.02381 .13174 m
.02756 .13174 L
s
.02381 .15962 m
.02756 .15962 L
s
.02381 .18749 m
.02756 .18749 L
s
.02381 .21536 m
.02756 .21536 L
s
.02381 .2711 m
.02756 .2711 L
s
.02381 .29898 m
.02756 .29898 L
s
.02381 .32685 m
.02756 .32685 L
s
.02381 .35472 m
.02756 .35472 L
s
.02381 .41046 m
.02756 .41046 L
s
.02381 .43834 m
.02756 .43834 L
s
.02381 .46621 m
.02756 .46621 L
s
.02381 .49408 m
.02756 .49408 L
s
.02381 .076 m
.02756 .076 L
s
.02381 .04813 m
.02756 .04813 L
s
.02381 .02025 m
.02756 .02025 L
s
.02381 .54983 m
.02756 .54983 L
s
.02381 .5777 m
.02756 .5777 L
s
.02381 .60557 m
.02756 .60557 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.03027 0 m
.03029 .05469 L
s
.03029 .05469 m
.03043 0 L
s
.03048 0 m
.03065 .61803 L
s
.03084 .61803 m
.03086 .60858 L
.03105 .52929 L
.03124 .45406 L
.03143 .38259 L
.03162 .31461 L
.03181 .24987 L
.032 .18814 L
.03219 .12921 L
.03238 .0729 L
.03257 .01904 L
s
.03257 .01904 m
.03264 0 L
s
.03319 0 m
.03333 .38259 L
.03352 .32794 L
.03371 .27539 L
.0339 .22483 L
s
.0339 .22483 m
.03406 .61803 L
s
.03435 .61803 m
.03448 .58168 L
.03467 .52929 L
.03486 .4787 L
s
.03486 .4787 m
.03492 .61803 L
s
.03611 .61803 m
.03619 .59699 L
.03638 .55151 L
.03657 .50739 L
.03676 .46457 L
.03695 .42299 L
.03714 .38259 L
.03733 .34334 L
.03752 .30517 L
.03771 .26805 L
.0379 .23193 L
.0381 .19678 L
.03829 .16255 L
.03848 .12921 L
.03867 .09672 L
.03886 .06506 L
.03905 .03419 L
.03924 .34818 L
.03943 .31461 L
.03962 .28185 L
.03981 .58168 L
.04 .54655 L
.04019 .51223 L
.04038 .4787 L
.04057 .44594 L
.04076 .41391 L
.04095 .38259 L
.04114 .35196 L
.04133 .322 L
.04152 .59238 L
.04171 .5605 L
.0419 .52929 L
.0421 .49873 L
.04229 .46879 L
.04248 .43947 L
.04267 .41075 L
s
.04267 .41075 m
.04282 .61803 L
s
.0484 .61803 m
.04857 .59699 L
.04876 .57408 L
.04895 .55151 L
.04914 .52929 L
.04933 .50739 L
s
.04933 .50739 m
.04945 .61803 L
s
.06245 .61803 m
.06248 .616 L
.06267 .6012 L
.06286 .58653 L
.06305 .57201 L
s
.06305 .57201 m
.06312 .61803 L
s
.06421 .61803 m
.06438 .60505 L
.06457 .59098 L
.06476 .57705 L
.06495 .56325 L
.06514 .54957 L
.06533 .53602 L
.06552 .52259 L
.06571 .50928 L
.0659 .4961 L
.0661 .48303 L
.06629 .47008 L
.06648 .45725 L
.06667 .44453 L
.06686 .43192 L
.06705 .41943 L
.06724 .40704 L
.06743 .39476 L
.06762 .38259 L
.06781 .37053 L
.068 .35856 L
.06819 .34671 L
.06838 .33495 L
.06857 .32329 L
.06876 .31173 L
.06895 .30027 L
.06914 .40601 L
.06933 .39425 L
.06952 .38259 L
.06971 .37103 L
.0699 .35956 L
.0701 .34818 L
.07029 .3369 L
.07048 .32571 L
.07067 .31461 L
.07086 .3036 L
.07105 .29268 L
.07124 .28185 L
.07143 .2711 L
.07162 .26044 L
.07181 .24987 L
.072 .23938 L
.07219 .22897 L
.07238 .21864 L
.07257 .20839 L
.07276 .19822 L
.07295 .18814 L
.07314 .17813 L
.07333 .16819 L
.07352 .26512 L
Mistroke
.07371 .25493 L
.0739 .24482 L
.0741 .23479 L
.07429 .22483 L
.07448 .21494 L
.07467 .30952 L
.07486 .29939 L
.07505 .28934 L
.07524 .27936 L
.07543 .26946 L
.07562 .25963 L
.07581 .24987 L
.076 .24018 L
.07619 .23056 L
.07638 .22101 L
.07657 .21154 L
.07676 .20213 L
.07695 .19278 L
.07714 .18351 L
.07733 .1743 L
.07752 .16515 L
.07771 .15607 L
.0779 .14705 L
.0781 .1381 L
.07829 .12921 L
.07848 .12038 L
.07867 .20839 L
.07886 .29579 L
.07905 .28648 L
.07924 .27723 L
.07943 .26805 L
.07962 .25893 L
.07981 .24987 L
.08 .24087 L
.08019 .23193 L
.08038 .22306 L
.08057 .21424 L
.08076 .20548 L
.08095 .19678 L
.08114 .18814 L
.08133 .17955 L
.08152 .17102 L
.08171 .16255 L
.0819 .15413 L
.0821 .14577 L
.08229 .13746 L
.08248 .12921 L
.08267 .12101 L
.08286 .11286 L
.08305 .10477 L
Mistroke
.08324 .09672 L
.08343 .08873 L
.08362 .16956 L
.08381 .16139 L
.084 .15326 L
.08419 .1452 L
.08438 .13718 L
.08457 .12921 L
.08476 .12129 L
.08495 .11342 L
.08514 .19216 L
.08533 .18412 L
.08552 .17613 L
.08571 .16819 L
.0859 .1603 L
.0861 .15246 L
.08629 .14466 L
.08648 .13691 L
.08667 .12921 L
.08686 .12155 L
.08705 .11395 L
.08724 .10638 L
.08743 .09886 L
.08762 .09139 L
.08781 .08396 L
.088 .15928 L
.08819 .1517 L
.08838 .14416 L
.08857 .13666 L
.08876 .12921 L
.08895 .1218 L
.08914 .11444 L
.08933 .10711 L
.08952 .09983 L
.08971 .09259 L
.0899 .0854 L
.0901 .07824 L
.09029 .07113 L
.09048 .06405 L
.09067 .05702 L
.09086 .05003 L
.09105 .04307 L
.09124 .11489 L
.09143 .1078 L
.09162 .10074 L
.09181 .09372 L
.092 .08674 L
.09219 .0798 L
.09238 .0729 L
.09257 .06604 L
Mistroke
.09276 .05921 L
.09295 .05243 L
.09314 .04568 L
.09333 .03896 L
.09352 .03229 L
.09371 .10159 L
.0939 .09478 L
.0941 .08801 L
.09429 .08127 L
.09448 .07457 L
.09467 .06791 L
.09486 .06128 L
.09505 .05469 L
.09524 .04813 L
.09543 .0416 L
.09562 .03512 L
.09581 .02866 L
.096 .02224 L
.09619 .01585 L
.09638 .0095 L
.09657 .00318 L
Mfstroke
.09657 .00318 m
.09667 0 L
s
.09886 0 m
.09886 .00059 L
s
.09886 .00059 m
.09888 0 L
s
.10282 0 m
.10286 .0132 L
.10305 .07439 L
.10324 .06845 L
.10343 .06253 L
.10362 .05664 L
.10381 .05078 L
.104 .04495 L
.10419 .03914 L
.10438 .03337 L
.10457 .02762 L
.10476 .02189 L
.10495 .0162 L
.10514 .01053 L
.10533 .00489 L
.10552 .06424 L
.10571 .0585 L
.1059 .05278 L
.1061 .04709 L
.10629 .04143 L
.10648 .0358 L
.10667 .03019 L
.10686 .0246 L
.10705 .01904 L
.10724 .01351 L
.10743 .008 L
.10762 .00252 L
s
.10762 .00252 m
.10771 0 L
s
.10782 0 m
.108 .05469 L
.10819 .04913 L
.10838 .04361 L
.10857 .03811 L
.10876 .03263 L
.10895 .02718 L
.10914 .02175 L
.10933 .07842 L
.10952 .0729 L
.10971 .06741 L
.1099 .06194 L
.1101 .05649 L
.11029 .05107 L
.11048 .04568 L
.11067 .0403 L
.11086 .03495 L
.11105 .09048 L
.11124 .08505 L
.11143 .07963 L
.11162 .07425 L
.11181 .06888 L
.112 .06354 L
.11219 .05822 L
.11238 .05292 L
.11257 .04765 L
.11276 .0424 L
.11295 .03717 L
.11314 .09139 L
.11333 .08608 L
.11352 .08079 L
.11371 .07553 L
.1139 .07028 L
.1141 .06506 L
.11429 .05986 L
.11448 .05469 L
.11467 .04953 L
.11486 .0444 L
.11505 .03928 L
.11524 .03419 L
.11543 .02912 L
.11562 .02407 L
.11581 .01904 L
.116 .01404 L
.11619 .00905 L
.11638 .00408 L
s
.11638 .00408 m
.11654 0 L
s
.11803 0 m
.1181 .01659 L
.11829 .01171 L
.11848 .00685 L
.11867 .00201 L
s
.11867 .00201 m
.11875 0 L
s
.11908 0 m
.11924 .04323 L
.11943 .03836 L
.11962 .0335 L
.11981 .08396 L
.12 .07903 L
.12019 .07413 L
.12038 .06924 L
.12057 .06437 L
.12076 .05952 L
.12095 .10934 L
.12114 .10442 L
.12133 .09952 L
.12152 .14897 L
.12171 .144 L
.1219 .13905 L
.1221 .13412 L
.12229 .18312 L
.12248 .17813 L
.12267 .17315 L
.12286 .16819 L
.12305 .21675 L
.12324 .21173 L
.12343 .20673 L
.12362 .20174 L
.12381 .24987 L
.124 .24482 L
.12419 .23979 L
.12438 .23479 L
.12457 .2298 L
.12476 .22483 L
.12495 .21987 L
.12514 .21494 L
.12533 .21003 L
.12552 .20513 L
.12571 .20025 L
.1259 .19539 L
.1261 .19055 L
.12629 .18573 L
.12648 .18092 L
.12667 .17613 L
.12686 .17136 L
.12705 .16661 L
.12724 .16187 L
.12743 .15716 L
.12762 .15246 L
.12781 .14777 L
.128 .14311 L
.12819 .13846 L
.12838 .13382 L
Mistroke
.12857 .12921 L
.12876 .12461 L
.12895 .12003 L
.12914 .11546 L
.12933 .11091 L
.12952 .10638 L
.12971 .10187 L
.1299 .09737 L
.1301 .14283 L
.13029 .13828 L
.13048 .13373 L
.13067 .12921 L
.13086 .1743 L
.13105 .16971 L
.13124 .16515 L
.13143 .1606 L
.13162 .15607 L
.13181 .15155 L
.132 .14705 L
.13219 .14257 L
.13238 .1381 L
.13257 .13365 L
.13276 .12921 L
.13295 .12479 L
.13314 .12038 L
.13333 .11599 L
.13352 .11161 L
.13371 .10725 L
.1339 .10291 L
.1341 .09858 L
.13429 .09426 L
.13448 .08996 L
.13467 .08567 L
.13486 .0814 L
.13505 .07714 L
.13524 .0729 L
.13543 .06867 L
.13562 .06446 L
.13581 .06026 L
.136 .05608 L
.13619 .05191 L
.13638 .04775 L
.13657 .04361 L
.13676 .03948 L
.13695 .03536 L
.13714 .03126 L
.13733 .02718 L
.13752 .06979 L
.13771 .06565 L
.1379 .10806 L
Mistroke
.1381 .10387 L
.13829 .0997 L
.13848 .09554 L
.13867 .09139 L
.13886 .08726 L
.13905 .08314 L
.13924 .07903 L
.13943 .07494 L
.13962 .07086 L
.13981 .0668 L
.14 .06275 L
.14019 .05871 L
.14038 .05469 L
.14057 .05067 L
.14076 .04667 L
.14095 .04269 L
.14114 .03872 L
.14133 .03476 L
.14152 .03081 L
.14171 .02687 L
.1419 .02295 L
.1421 .01904 L
.14229 .01515 L
.14248 .01126 L
.14267 .00739 L
.14286 .04813 L
.14305 .04421 L
.14324 .0403 L
.14343 .03641 L
.14362 .03253 L
.14381 .02866 L
.144 .0248 L
.14419 .02096 L
.14438 .01713 L
.14457 .01331 L
.14476 .0095 L
.14495 .00571 L
.14514 .04568 L
.14533 .04184 L
.14552 .08163 L
.14571 .12129 L
.1459 .11735 L
.1461 .11342 L
.14629 .10951 L
.14648 .1056 L
.14667 .10171 L
.14686 .09783 L
.14705 .09396 L
.14724 .09011 L
.14743 .08626 L
Mistroke
.14762 .08243 L
.14781 .12143 L
.148 .11755 L
.14819 .11369 L
.14838 .15246 L
.14857 .14855 L
.14876 .14466 L
.14895 .14078 L
.14914 .13691 L
.14933 .13305 L
.14952 .12921 L
.14971 .12538 L
.1499 .12155 L
.1501 .15978 L
.15029 .15592 L
.15048 .15207 L
.15067 .14823 L
.15086 .1444 L
.15105 .14059 L
.15124 .13678 L
.15143 .13299 L
.15162 .12921 L
.15181 .12544 L
.152 .12168 L
.15219 .11793 L
.15238 .11419 L
.15257 .11047 L
.15276 .10675 L
.15295 .10305 L
.15314 .09936 L
.15333 .13666 L
.15352 .17386 L
.15371 .17008 L
.1539 .20712 L
.1541 .2033 L
.15429 .19949 L
.15448 .19569 L
.15467 .19191 L
.15486 .22865 L
.15505 .22483 L
.15524 .22101 L
.15543 .25755 L
.15562 .2537 L
.15581 .24987 L
.156 .24604 L
.15619 .24223 L
.15638 .23843 L
.15657 .23463 L
.15676 .23085 L
.15695 .22708 L
Mistroke
.15714 .22332 L
.15733 .21957 L
.15752 .21584 L
.15771 .21211 L
.1579 .20839 L
.1581 .20469 L
.15829 .20099 L
.15848 .1973 L
.15867 .233 L
.15886 .22928 L
.15905 .22557 L
.15924 .22187 L
.15943 .21818 L
.15962 .2145 L
.15981 .21083 L
.16 .20717 L
.16019 .20353 L
.16038 .19989 L
.16057 .19626 L
.16076 .19264 L
.16095 .18904 L
.16114 .18544 L
.16133 .18185 L
.16152 .21682 L
.16171 .2132 L
.1619 .20959 L
.1621 .20599 L
.16229 .2024 L
.16248 .19882 L
.16267 .19525 L
.16286 .19169 L
.16305 .18814 L
.16324 .18459 L
.16343 .18106 L
.16362 .17754 L
.16381 .17403 L
.164 .17052 L
.16419 .16703 L
.16438 .16354 L
.16457 .16007 L
.16476 .1566 L
.16495 .15315 L
.16514 .1497 L
.16533 .14626 L
.16552 .14283 L
.16571 .13941 L
.1659 .136 L
.1661 .1326 L
.16629 .12921 L
.16648 .12583 L
Mistroke
.16667 .12245 L
.16686 .1562 L
.16705 .1528 L
.16724 .1494 L
.16743 .14601 L
.16762 .14263 L
.16781 .13926 L
.168 .17272 L
.16819 .16932 L
.16838 .16593 L
.16857 .19922 L
.16876 .1958 L
.16895 .19239 L
.16914 .18899 L
.16933 .18559 L
.16952 .1822 L
.16971 .17883 L
.1699 .17546 L
.1701 .1721 L
.17029 .16875 L
.17048 .16541 L
.17067 .16207 L
.17086 .15875 L
.17105 .15543 L
.17124 .18814 L
.17143 .18479 L
.17162 .18145 L
.17181 .17813 L
.172 .17481 L
.17219 .17149 L
.17238 .16819 L
.17257 .1649 L
.17276 .16161 L
.17295 .19393 L
.17314 .19062 L
.17333 .18731 L
.17352 .18401 L
.17371 .18072 L
.1739 .17744 L
.1741 .17417 L
.17429 .17091 L
.17448 .16765 L
.17467 .1644 L
.17486 .16116 L
.17505 .15793 L
.17524 .18977 L
.17543 .18651 L
.17562 .18326 L
.17581 .18001 L
.176 .17678 L
Mistroke
.17619 .17355 L
.17638 .17033 L
.17657 .16712 L
.17676 .16392 L
.17695 .16072 L
.17714 .15754 L
.17733 .15436 L
.17752 .15119 L
.17771 .14803 L
.1779 .14487 L
.1781 .14172 L
.17829 .13858 L
.17848 .13545 L
.17867 .13233 L
.17886 .12921 L
.17905 .1261 L
.17924 .123 L
.17943 .11991 L
.17962 .11682 L
.17981 .11374 L
.18 .11067 L
.18019 .10761 L
.18038 .10455 L
.18057 .1015 L
.18076 .09846 L
.18095 .09543 L
.18114 .0924 L
.18133 .08938 L
.18152 .08637 L
.18171 .11698 L
.1819 .11395 L
.1821 .14446 L
.18229 .14139 L
.18248 .13833 L
.18267 .16871 L
.18286 .16562 L
.18305 .16255 L
.18324 .15948 L
.18343 .15642 L
.18362 .15337 L
.18381 .15032 L
.184 .14729 L
.18419 .14426 L
.18438 .14123 L
.18457 .13822 L
.18476 .13521 L
.18495 .16515 L
.18514 .16212 L
.18533 .15909 L
.18552 .15607 L
Mistroke
.18571 .15306 L
.1859 .15005 L
.1861 .14705 L
.18629 .14406 L
.18648 .14108 L
.18667 .1707 L
.18686 .16769 L
.18705 .16469 L
.18724 .19418 L
.18743 .19115 L
.18762 .18814 L
.18781 .18512 L
.188 .18212 L
.18819 .17912 L
.18838 .17613 L
.18857 .17315 L
.18876 .17017 L
.18895 .1672 L
.18914 .16424 L
.18933 .16128 L
.18952 .15833 L
.18971 .18739 L
.1899 .18442 L
.1901 .21338 L
.19029 .21038 L
.19048 .2074 L
.19067 .23623 L
.19086 .23322 L
.19105 .23022 L
.19124 .22722 L
.19143 .2559 L
.19162 .25288 L
.19181 .24987 L
.192 .27843 L
.19219 .30692 L
.19238 .30386 L
.19257 .3008 L
.19276 .29775 L
.19295 .29471 L
.19314 .29167 L
.19333 .28864 L
.19352 .28562 L
.19371 .31385 L
.1939 .31081 L
.1941 .30777 L
.19429 .30474 L
.19448 .30171 L
.19467 .2987 L
.19486 .29569 L
.19505 .29268 L
Mistroke
.19524 .28969 L
.19543 .2867 L
.19562 .28371 L
.19581 .28073 L
.196 .27776 L
.19619 .3056 L
.19638 .33337 L
.19657 .33035 L
.19676 .32734 L
.19695 .32433 L
.19714 .32134 L
.19733 .31834 L
.19752 .31536 L
.19771 .31238 L
.1979 .30941 L
.1981 .30644 L
.19829 .30348 L
.19848 .30053 L
.19867 .29758 L
.19886 .29464 L
.19905 .29171 L
.19924 .28878 L
.19943 .28586 L
.19962 .28294 L
.19981 .28003 L
.2 .27713 L
.20019 .27423 L
.20038 .27134 L
.20057 .26846 L
.20076 .26558 L
.20095 .26271 L
.20114 .25985 L
.20133 .25699 L
.20152 .25414 L
.20171 .25129 L
.2019 .24845 L
.2021 .24561 L
.20229 .24279 L
.20248 .23996 L
.20267 .23715 L
.20286 .23434 L
.20305 .23153 L
.20324 .25832 L
.20343 .2555 L
.20362 .25268 L
.20381 .24987 L
.204 .24706 L
.20419 .24426 L
.20438 .24147 L
.20457 .23868 L
Mistroke
.20476 .2359 L
.20495 .23312 L
.20514 .25963 L
.20533 .25683 L
.20552 .25404 L
.20571 .25126 L
.2059 .24848 L
.2061 .24571 L
.20629 .24294 L
.20648 .24018 L
.20667 .23743 L
.20686 .23468 L
.20705 .23193 L
.20724 .22919 L
.20743 .22646 L
.20762 .22374 L
.20781 .22101 L
.208 .2183 L
.20819 .21559 L
.20838 .21289 L
.20857 .21019 L
.20876 .20749 L
.20895 .20481 L
.20914 .20213 L
.20933 .22807 L
.20952 .22537 L
.20971 .22267 L
.2099 .21998 L
.2101 .2173 L
.21029 .21462 L
.21048 .21195 L
.21067 .20928 L
.21086 .20662 L
.21105 .20396 L
.21124 .20131 L
.21143 .19866 L
.21162 .22429 L
.21181 .24987 L
.212 .24718 L
.21219 .2445 L
.21238 .24182 L
.21257 .23915 L
.21276 .23649 L
.21295 .23383 L
.21314 .23117 L
.21333 .25654 L
.21352 .25387 L
.21371 .2512 L
.2139 .24854 L
.2141 .24588 L
Mistroke
.21429 .24323 L
.21448 .24059 L
.21467 .23795 L
.21486 .23531 L
.21505 .23268 L
.21524 .23006 L
.21543 .22744 L
.21562 .22483 L
.21581 .22222 L
.216 .21961 L
.21619 .21702 L
.21638 .21442 L
.21657 .21183 L
.21676 .20925 L
.21695 .23416 L
.21714 .23156 L
.21733 .22897 L
.21752 .22638 L
.21771 .22379 L
.2179 .22121 L
.2181 .21864 L
.21829 .21607 L
.21848 .21351 L
.21867 .21095 L
.21886 .23561 L
.21905 .23303 L
.21924 .23046 L
.21943 .2279 L
.21962 .22534 L
.21981 .22278 L
.22 .22023 L
.22019 .21768 L
.22038 .21514 L
.22057 .21261 L
.22076 .21008 L
.22095 .20755 L
.22114 .20503 L
.22133 .20251 L
.22152 .2 L
.22171 .19749 L
.2219 .19499 L
.2221 .19249 L
.22229 .19 L
.22248 .18751 L
.22267 .18503 L
.22286 .18255 L
.22305 .18008 L
.22324 .17761 L
.22343 .17515 L
.22362 .17269 L
Mistroke
.22381 .17023 L
.224 .16778 L
.22419 .19183 L
.22438 .18937 L
.22457 .18691 L
.22476 .18445 L
.22495 .182 L
.22514 .17955 L
.22533 .17711 L
.22552 .17467 L
.22571 .17224 L
.2259 .19608 L
.2261 .19363 L
.22629 .19118 L
.22648 .18875 L
.22667 .18631 L
.22686 .21003 L
.22705 .20758 L
.22724 .23123 L
.22743 .22876 L
.22762 .2263 L
.22781 .22384 L
.228 .22139 L
.22819 .21894 L
.22838 .2165 L
.22857 .21406 L
.22876 .21163 L
.22895 .2092 L
.22914 .20678 L
.22933 .20436 L
.22952 .20194 L
.22971 .22531 L
.2299 .22288 L
.2301 .22046 L
.23029 .21803 L
.23048 .24131 L
.23067 .23887 L
.23086 .23644 L
.23105 .23401 L
.23124 .23159 L
.23143 .22917 L
.23162 .22675 L
.23181 .22434 L
.232 .22194 L
.23219 .21954 L
.23238 .21714 L
.23257 .24018 L
.23276 .23777 L
.23295 .23536 L
.23314 .23296 L
Mistroke
.23333 .23056 L
.23352 .25348 L
.23371 .25107 L
.2339 .24866 L
.2341 .24626 L
.23429 .24386 L
.23448 .24147 L
.23467 .23908 L
.23486 .23669 L
.23505 .23431 L
.23524 .23193 L
.23543 .22956 L
.23562 .25226 L
.23581 .24987 L
.236 .24749 L
.23619 .24511 L
.23638 .24273 L
.23657 .24036 L
.23676 .238 L
.23695 .23563 L
.23714 .23328 L
.23733 .23092 L
.23752 .25342 L
.23771 .25105 L
.2379 .24869 L
.2381 .24633 L
.23829 .24397 L
.23848 .24162 L
.23867 .23928 L
.23886 .23694 L
.23905 .25926 L
.23924 .25691 L
.23943 .25456 L
.23962 .25221 L
.23981 .24987 L
.24 .24753 L
.24019 .24519 L
.24038 .24286 L
.24057 .26503 L
.24076 .28716 L
.24095 .2848 L
.24114 .28244 L
.24133 .28009 L
.24152 .27774 L
.24171 .27539 L
.2419 .27305 L
.2421 .29504 L
.24229 .29268 L
.24248 .29033 L
.24267 .31225 L
Mistroke
.24286 .30988 L
.24305 .30752 L
.24324 .30517 L
.24343 .30282 L
.24362 .30047 L
.24381 .29813 L
.244 .29579 L
.24419 .29346 L
.24438 .29113 L
.24457 .2888 L
.24476 .28648 L
.24495 .30817 L
.24514 .30584 L
.24533 .30351 L
.24552 .30118 L
.24571 .29886 L
.2459 .29654 L
.2461 .29422 L
.24629 .29191 L
.24648 .28961 L
.24667 .2873 L
.24686 .285 L
.24705 .30649 L
.24724 .30418 L
.24743 .30187 L
.24762 .29957 L
.24781 .29727 L
.248 .29497 L
.24819 .29268 L
.24838 .29039 L
.24857 .28811 L
.24876 .28583 L
.24895 .28355 L
.24914 .30484 L
.24933 .30255 L
.24952 .30027 L
.24971 .29799 L
.2499 .31919 L
.2501 .3169 L
.25029 .31461 L
.25048 .31233 L
.25067 .31005 L
.25086 .30777 L
.25105 .30549 L
.25124 .30322 L
.25143 .30096 L
.25162 .2987 L
.25181 .29644 L
.252 .29418 L
.25219 .29193 L
Mistroke
.25238 .28969 L
.25257 .31065 L
.25276 .30839 L
.25295 .30614 L
.25314 .30388 L
.25333 .30164 L
.25352 .29939 L
.25371 .29715 L
.2539 .29492 L
.2541 .29268 L
.25429 .29045 L
.25448 .28823 L
.25467 .309 L
.25486 .30677 L
.25505 .30453 L
.25524 .3023 L
.25543 .30008 L
.25562 .29785 L
.25581 .29564 L
.256 .31628 L
.25619 .31405 L
.25638 .31183 L
.25657 .3096 L
.25676 .30739 L
.25695 .30517 L
.25714 .30296 L
.25733 .30075 L
.25752 .29854 L
.25771 .29634 L
.2579 .29415 L
.2581 .29195 L
.25829 .28976 L
.25848 .28757 L
.25867 .28539 L
.25886 .28321 L
.25905 .28103 L
.25924 .27886 L
.25943 .29922 L
.25962 .29704 L
.25981 .29486 L
.26 .29268 L
.26019 .29051 L
.26038 .28834 L
.26057 .28617 L
.26076 .28401 L
.26095 .28185 L
.26114 .27969 L
.26133 .27754 L
.26152 .27539 L
.26171 .27325 L
Mistroke
.2619 .2711 L
.2621 .26897 L
.26229 .26683 L
.26248 .2647 L
.26267 .26257 L
.26286 .26044 L
.26305 .25832 L
.26324 .2562 L
.26343 .27624 L
.26362 .27411 L
.26381 .27199 L
.264 .26987 L
.26419 .26775 L
.26438 .26563 L
.26457 .28557 L
.26476 .28344 L
.26495 .28132 L
.26514 .2792 L
.26533 .27708 L
.26552 .27497 L
.26571 .27286 L
.2659 .27075 L
.2661 .26865 L
.26629 .26655 L
.26648 .26445 L
.26667 .26236 L
.26686 .26027 L
.26705 .25818 L
.26724 .2561 L
.26743 .25402 L
.26762 .25194 L
.26781 .24987 L
.268 .2478 L
.26819 .24573 L
.26838 .24367 L
.26857 .24161 L
.26876 .23955 L
.26895 .23749 L
.26914 .23544 L
.26933 .23339 L
.26952 .23135 L
.26971 .22931 L
.2699 .22727 L
.2701 .22523 L
.27029 .2232 L
.27048 .22117 L
.27067 .21914 L
.27086 .21712 L
.27105 .23658 L
.27124 .23454 L
Mistroke
.27143 .23251 L
.27162 .23048 L
.27181 .22846 L
.272 .22644 L
.27219 .22442 L
.27238 .22241 L
.27257 .2204 L
.27276 .23971 L
.27295 .23769 L
.27314 .23567 L
.27333 .23366 L
.27352 .23165 L
.27371 .22964 L
.2739 .22763 L
.2741 .22563 L
.27429 .24482 L
.27448 .24281 L
.27467 .2408 L
.27486 .23879 L
.27505 .23679 L
.27524 .2559 L
.27543 .27499 L
.27562 .27296 L
.27581 .27094 L
.276 .26891 L
.27619 .2669 L
.27638 .26488 L
.27657 .26287 L
.27676 .26086 L
.27695 .25886 L
.27714 .25685 L
.27733 .25485 L
.27752 .27378 L
.27771 .27177 L
.2779 .26977 L
.2781 .26776 L
.27829 .26576 L
.27848 .26377 L
.27867 .26177 L
.27886 .25978 L
.27905 .25779 L
.27924 .25581 L
.27943 .25382 L
.27962 .2726 L
.27981 .27061 L
.28 .26862 L
.28019 .26663 L
.28038 .26465 L
.28057 .26267 L
.28076 .26069 L
Mistroke
.28095 .25872 L
.28114 .25674 L
.28133 .25478 L
.28152 .25281 L
.28171 .25085 L
.2819 .24889 L
.2821 .24693 L
.28229 .24498 L
.28248 .24303 L
.28267 .24108 L
.28286 .23913 L
.28305 .23719 L
.28324 .23525 L
.28343 .25376 L
.28362 .25181 L
.28381 .24987 L
.284 .24792 L
.28419 .24598 L
.28438 .26442 L
.28457 .26247 L
.28476 .26052 L
.28495 .25858 L
.28514 .25664 L
.28533 .275 L
.28552 .29334 L
.28571 .29137 L
.2859 .28942 L
.2861 .28746 L
.28629 .28551 L
.28648 .28355 L
.28667 .28161 L
.28686 .27966 L
.28705 .27772 L
.28724 .27578 L
.28743 .29398 L
.28762 .29203 L
.28781 .29009 L
.288 .28814 L
.28819 .2862 L
.28838 .28427 L
.28857 .28233 L
.28876 .2804 L
.28895 .27847 L
.28914 .27655 L
.28933 .27462 L
.28952 .2727 L
.28971 .27078 L
.2899 .26887 L
.2901 .26696 L
.29029 .26505 L
Mistroke
.29048 .26314 L
.29067 .26124 L
.29086 .25933 L
.29105 .25744 L
.29124 .25554 L
.29143 .25365 L
.29162 .25176 L
.29181 .24987 L
.292 .24798 L
.29219 .2461 L
.29238 .26399 L
.29257 .2621 L
.29276 .26021 L
.29295 .25832 L
.29314 .25644 L
.29333 .25456 L
.29352 .27236 L
.29371 .27047 L
.2939 .26859 L
.2941 .2667 L
.29429 .26482 L
.29448 .26294 L
.29467 .26107 L
.29486 .2592 L
.29505 .25732 L
.29524 .25546 L
.29543 .25359 L
.29562 .25173 L
.29581 .26939 L
.296 .26752 L
.29619 .26565 L
.29638 .26378 L
.29657 .26192 L
.29676 .26006 L
.29695 .2582 L
.29714 .25634 L
.29733 .25449 L
.29752 .27203 L
.29771 .27017 L
.2979 .28768 L
.2981 .30517 L
.29829 .30329 L
.29848 .30141 L
.29867 .29954 L
.29886 .29766 L
.29905 .29579 L
.29924 .29393 L
.29943 .29206 L
.29962 .30945 L
.29981 .30757 L
Mistroke
.3 .3057 L
.30019 .30384 L
.30038 .32117 L
.30057 .31929 L
.30076 .31742 L
.30095 .31555 L
.30114 .31368 L
.30133 .31181 L
.30152 .30995 L
.30171 .30809 L
.3019 .30623 L
.3021 .30438 L
.30229 .30252 L
.30248 .30067 L
.30267 .29882 L
.30286 .29698 L
.30305 .29514 L
.30324 .2933 L
.30343 .29146 L
.30362 .28962 L
.30381 .30675 L
.304 .30491 L
.30419 .30307 L
.30438 .30123 L
.30457 .29939 L
.30476 .29756 L
.30495 .29573 L
.30514 .2939 L
.30533 .29207 L
.30552 .29025 L
.30571 .28843 L
.3059 .28661 L
.3061 .3036 L
.30629 .30178 L
.30648 .29995 L
.30667 .3169 L
.30686 .31507 L
.30705 .33198 L
.30724 .33014 L
.30743 .32831 L
.30762 .32647 L
.30781 .32464 L
.308 .32281 L
.30819 .32099 L
.30838 .33782 L
.30857 .33598 L
.30876 .35278 L
.30895 .35094 L
.30914 .3491 L
.30933 .36586 L
Mistroke
.30952 .36401 L
.30971 .36217 L
.3099 .37888 L
.3101 .37703 L
.31029 .37518 L
.31048 .37333 L
.31067 .39 L
.31086 .38814 L
.31105 .38629 L
.31124 .38444 L
.31143 .38259 L
.31162 .38075 L
.31181 .37891 L
.312 .37707 L
.31219 .37523 L
.31238 .37339 L
.31257 .37156 L
.31276 .36973 L
.31295 .3679 L
.31314 .38443 L
.31333 .38259 L
.31352 .38076 L
.31371 .37893 L
.3139 .3771 L
.3141 .37528 L
.31429 .37345 L
.31448 .37163 L
.31467 .36982 L
.31486 .368 L
.31505 .36619 L
.31524 .36438 L
.31543 .36257 L
.31562 .36076 L
.31581 .35896 L
.316 .35716 L
.31619 .35536 L
.31638 .35356 L
.31657 .35176 L
.31676 .34997 L
.31695 .34818 L
.31714 .34639 L
.31733 .34461 L
.31752 .34283 L
.31771 .34105 L
.3179 .33927 L
.3181 .33749 L
.31829 .33572 L
.31848 .33395 L
.31867 .33218 L
.31886 .33041 L
Mistroke
.31905 .32865 L
.31924 .32688 L
.31943 .32512 L
.31962 .34131 L
.31981 .33955 L
.32 .33778 L
.32019 .33602 L
.32038 .33426 L
.32057 .3325 L
.32076 .33075 L
.32095 .32899 L
.32114 .32724 L
.32133 .32549 L
.32152 .34158 L
.32171 .33982 L
.3219 .33807 L
.3221 .33632 L
.32229 .33457 L
.32248 .33282 L
.32267 .33108 L
.32286 .32933 L
.32305 .32759 L
.32324 .34359 L
.32343 .34184 L
.32362 .34009 L
.32381 .35605 L
.324 .3543 L
.32419 .35255 L
.32438 .3508 L
.32457 .36671 L
.32476 .36495 L
.32495 .3632 L
.32514 .36145 L
.32533 .3597 L
.32552 .35796 L
.32571 .3738 L
.3259 .37205 L
.3261 .3703 L
.32629 .36855 L
.32648 .36681 L
.32667 .36506 L
.32686 .36332 L
.32705 .36158 L
.32724 .35985 L
.32743 .35811 L
.32762 .35638 L
.32781 .35465 L
.328 .37038 L
.32819 .38608 L
.32838 .38434 L
Mistroke
.32857 .38259 L
.32876 .38085 L
.32895 .37911 L
.32914 .37738 L
.32933 .37564 L
.32952 .37391 L
.32971 .37218 L
.3299 .37045 L
.3301 .36873 L
.33029 .367 L
.33048 .38259 L
.33067 .38086 L
.33086 .37913 L
.33105 .37741 L
.33124 .37568 L
.33143 .37396 L
.33162 .37224 L
.33181 .38776 L
.332 .38604 L
.33219 .38431 L
.33238 .38259 L
.33257 .39807 L
.33276 .39634 L
.33295 .39461 L
.33314 .39289 L
.33333 .39117 L
.33352 .38945 L
.33371 .38773 L
.3339 .38602 L
.3341 .3843 L
.33429 .39969 L
.33448 .39797 L
.33467 .39626 L
.33486 .39454 L
.33505 .40988 L
.33524 .40816 L
.33543 .40644 L
.33562 .40473 L
.33581 .42003 L
.336 .4183 L
.33619 .41658 L
.33638 .41486 L
.33657 .41315 L
.33676 .41143 L
.33695 .40972 L
.33714 .40801 L
.33733 .4063 L
.33752 .40459 L
.33771 .40289 L
.3379 .40119 L
Mistroke
.3381 .39948 L
.33829 .39779 L
.33848 .39609 L
.33867 .3944 L
.33886 .3927 L
.33905 .39101 L
.33924 .38932 L
.33943 .38764 L
.33962 .38595 L
.33981 .38427 L
.34 .38259 L
.34019 .38091 L
.34038 .39601 L
.34057 .39432 L
.34076 .39264 L
.34095 .4077 L
.34114 .40601 L
.34133 .40433 L
.34152 .40264 L
.34171 .40096 L
.3419 .39928 L
.3421 .3976 L
.34229 .4126 L
.34248 .41091 L
.34267 .40923 L
.34286 .40755 L
.34305 .40587 L
.34324 .4042 L
.34343 .40252 L
.34362 .40085 L
.34381 .39918 L
.344 .39752 L
.34419 .39585 L
.34438 .39419 L
.34457 .40907 L
.34476 .42395 L
.34495 .42227 L
.34514 .42059 L
.34533 .41892 L
.34552 .43375 L
.34571 .43207 L
.3459 .43039 L
.3461 .42872 L
.34629 .42704 L
.34648 .42537 L
.34667 .4237 L
.34686 .42203 L
.34705 .42037 L
.34724 .43512 L
.34743 .43345 L
Mistroke
.34762 .43178 L
.34781 .43011 L
.348 .44482 L
.34819 .44315 L
.34838 .44148 L
.34857 .43981 L
.34876 .43814 L
.34895 .43648 L
.34914 .43481 L
.34933 .43315 L
.34952 .43149 L
.34971 .42983 L
.3499 .42818 L
.3501 .42652 L
.35029 .42487 L
.35048 .43947 L
.35067 .43782 L
.35086 .43616 L
.35105 .43451 L
.35124 .43286 L
.35143 .43121 L
.35162 .42956 L
.35181 .42791 L
.352 .44245 L
.35219 .44079 L
.35238 .43914 L
.35257 .4375 L
.35276 .43585 L
.35295 .43421 L
.35314 .43257 L
.35333 .43093 L
.35352 .42929 L
.35371 .42765 L
.3539 .42602 L
.3541 .42438 L
.35429 .42275 L
.35448 .42113 L
.35467 .4195 L
.35486 .43391 L
.35505 .43228 L
.35524 .43065 L
.35543 .42902 L
.35562 .42739 L
.35581 .42577 L
.356 .42414 L
.35619 .42252 L
.35638 .4209 L
.35657 .41929 L
.35676 .41767 L
.35695 .43199 L
Mistroke
.35714 .43037 L
.35733 .42875 L
.35752 .42714 L
.35771 .42552 L
.3579 .42391 L
.3581 .4223 L
.35829 .42069 L
.35848 .41908 L
.35867 .41747 L
.35886 .41587 L
.35905 .41427 L
.35924 .41266 L
.35943 .41107 L
.35962 .40947 L
.35981 .40787 L
.36 .40628 L
.36019 .40469 L
.36038 .4031 L
.36057 .41727 L
.36076 .41568 L
.36095 .41409 L
.36114 .41249 L
.36133 .41091 L
.36152 .40932 L
.36171 .40773 L
.3619 .40615 L
.3621 .40456 L
.36229 .40298 L
.36248 .4014 L
.36267 .39983 L
.36286 .39825 L
.36305 .39668 L
.36324 .39511 L
.36343 .39353 L
.36362 .39197 L
.36381 .3904 L
.364 .40444 L
.36419 .40287 L
.36438 .4013 L
.36457 .39973 L
.36476 .39816 L
.36495 .3966 L
.36514 .41059 L
.36533 .40902 L
.36552 .40745 L
.36571 .40588 L
.3659 .40432 L
.3661 .40276 L
.36629 .40119 L
.36648 .39963 L
Mistroke
.36667 .39808 L
.36686 .412 L
.36705 .4259 L
.36724 .42433 L
.36743 .42276 L
.36762 .4212 L
.36781 .41963 L
.368 .41807 L
.36819 .41651 L
.36838 .41495 L
.36857 .41339 L
.36876 .41183 L
.36895 .41028 L
.36914 .40873 L
.36933 .40718 L
.36952 .40563 L
.36971 .41943 L
.3699 .43321 L
.3701 .43165 L
.37029 .43009 L
.37048 .42854 L
.37067 .42698 L
.37086 .42543 L
.37105 .42387 L
.37124 .42232 L
.37143 .42077 L
.37162 .41923 L
.37181 .41768 L
.372 .41614 L
.37219 .41459 L
.37238 .41305 L
.37257 .41151 L
.37276 .40998 L
.37295 .42365 L
.37314 .42211 L
.37333 .42057 L
.37352 .41903 L
.37371 .41749 L
.3739 .41595 L
.3741 .41442 L
.37429 .41289 L
.37448 .41136 L
.37467 .40983 L
.37486 .4083 L
.37505 .40678 L
.37524 .40525 L
.37543 .40373 L
.37562 .40221 L
.37581 .41577 L
.376 .41425 L
Mistroke
.37619 .41272 L
.37638 .4112 L
.37657 .40968 L
.37676 .40816 L
.37695 .40665 L
.37714 .40513 L
.37733 .40362 L
.37752 .4021 L
.37771 .40059 L
.3779 .39908 L
.3781 .39758 L
.37829 .39607 L
.37848 .39457 L
.37867 .39307 L
.37886 .39156 L
.37905 .39006 L
.37924 .38857 L
.37943 .38707 L
.37962 .38558 L
.37981 .399 L
.38 .3975 L
.38019 .396 L
.38038 .3945 L
.38057 .39301 L
.38076 .40639 L
.38095 .40489 L
.38114 .40339 L
.38133 .4019 L
.38152 .4004 L
.38171 .39891 L
.3819 .39742 L
.3821 .39593 L
.38229 .39444 L
.38248 .39295 L
.38267 .39147 L
.38286 .38999 L
.38305 .3885 L
.38324 .38702 L
.38343 .38555 L
.38362 .38407 L
.38381 .38259 L
.384 .39586 L
.38419 .39438 L
.38438 .3929 L
.38457 .39142 L
.38476 .38995 L
.38495 .38847 L
.38514 .387 L
.38533 .38553 L
.38552 .38406 L
Mistroke
.38571 .39726 L
.3859 .39579 L
.3861 .39432 L
.38629 .40749 L
.38648 .40601 L
.38667 .40454 L
.38686 .40307 L
.38705 .40159 L
.38724 .40012 L
.38743 .39865 L
.38762 .39719 L
.38781 .4103 L
.388 .40883 L
.38819 .40736 L
.38838 .40589 L
.38857 .41898 L
.38876 .41751 L
.38895 .43057 L
.38914 .42909 L
.38933 .42762 L
.38952 .42614 L
.38971 .42467 L
.3899 .4232 L
.3901 .42173 L
.39029 .42026 L
.39048 .41879 L
.39067 .4318 L
.39086 .43032 L
.39105 .42885 L
.39124 .42738 L
.39143 .42592 L
.39162 .42445 L
.39181 .42299 L
.392 .42152 L
.39219 .42006 L
.39238 .4186 L
.39257 .41714 L
.39276 .41569 L
.39295 .41423 L
.39314 .41278 L
.39333 .41133 L
.39352 .40988 L
.39371 .40843 L
.3939 .40698 L
.3941 .40553 L
.39429 .40409 L
.39448 .41697 L
.39467 .41552 L
.39486 .41407 L
.39505 .41262 L
Mistroke
.39524 .41118 L
.39543 .40974 L
.39562 .40829 L
.39581 .40685 L
.396 .40542 L
.39619 .40398 L
.39638 .40254 L
.39657 .40111 L
.39676 .41391 L
.39695 .41247 L
.39714 .41103 L
.39733 .4096 L
.39752 .40816 L
.39771 .42093 L
.3979 .41949 L
.3981 .41805 L
.39829 .41662 L
.39848 .41518 L
.39867 .41375 L
.39886 .41232 L
.39905 .41089 L
.39924 .40946 L
.39943 .40803 L
.39962 .40661 L
.39981 .40518 L
.4 .40376 L
.40019 .40234 L
.40038 .41502 L
.40057 .41359 L
.40076 .41217 L
.40095 .41075 L
.40114 .40932 L
.40133 .40791 L
.40152 .40649 L
.40171 .40507 L
.4019 .40365 L
.4021 .40224 L
.40229 .40083 L
.40248 .41344 L
.40267 .41202 L
.40286 .4106 L
.40305 .40919 L
.40324 .40778 L
.40343 .40637 L
.40362 .40496 L
.40381 .40355 L
.404 .41611 L
.40419 .41469 L
.40438 .41328 L
.40457 .41187 L
Mistroke
.40476 .41046 L
.40495 .40906 L
.40514 .40765 L
.40533 .40625 L
.40552 .40485 L
.40571 .40344 L
.4059 .40204 L
.4061 .40065 L
.40629 .39925 L
.40648 .39785 L
.40667 .39646 L
.40686 .39507 L
.40705 .39367 L
.40724 .39228 L
.40743 .3909 L
.40762 .38951 L
.40781 .38812 L
.408 .38674 L
.40819 .38535 L
.40838 .38397 L
.40857 .38259 L
.40876 .38121 L
.40895 .37984 L
.40914 .37846 L
.40933 .37708 L
.40952 .37571 L
.40971 .37434 L
.4099 .37297 L
.4101 .3716 L
.41029 .37023 L
.41048 .36886 L
.41067 .3675 L
.41086 .36613 L
.41105 .37848 L
.41124 .37711 L
.41143 .37574 L
.41162 .37438 L
.41181 .37301 L
.412 .37165 L
.41219 .37029 L
.41238 .36893 L
.41257 .36757 L
.41276 .36621 L
.41295 .3785 L
.41314 .39077 L
.41333 .38941 L
.41352 .38804 L
.41371 .38668 L
.4139 .38531 L
.4141 .38395 L
Mistroke
.41429 .38259 L
.41448 .39482 L
.41467 .39346 L
.41486 .3921 L
.41505 .39073 L
.41524 .38937 L
.41543 .38802 L
.41562 .38666 L
.41581 .3853 L
.416 .38395 L
.41619 .39612 L
.41638 .39476 L
.41657 .39341 L
.41676 .39205 L
.41695 .39069 L
.41714 .38934 L
.41733 .40148 L
.41752 .40012 L
.41771 .39877 L
.4179 .41088 L
.4181 .40952 L
.41829 .40816 L
.41848 .40681 L
.41867 .40545 L
.41886 .41753 L
.41905 .41617 L
.41924 .41481 L
.41943 .41346 L
.41962 .4121 L
.41981 .41075 L
.42 .40939 L
.42019 .40804 L
.42038 .40669 L
.42057 .40534 L
.42076 .40399 L
.42095 .40264 L
.42114 .4013 L
.42133 .41331 L
.42152 .41196 L
.42171 .42395 L
.4219 .4226 L
.4221 .42125 L
.42229 .4199 L
.42248 .43186 L
.42267 .43051 L
.42286 .42916 L
.42305 .4278 L
.42324 .42645 L
.42343 .4251 L
.42362 .42376 L
Mistroke
.42381 .42241 L
.424 .42106 L
.42419 .41972 L
.42438 .43163 L
.42457 .43028 L
.42476 .44218 L
.42495 .44082 L
.42514 .43947 L
.42533 .43813 L
.42552 .43678 L
.42571 .43543 L
.4259 .44729 L
.4261 .44594 L
.42629 .44459 L
.42648 .44324 L
.42667 .44189 L
.42686 .45372 L
.42705 .45237 L
.42724 .45102 L
.42743 .44967 L
.42762 .44833 L
.42781 .44698 L
.428 .44564 L
.42819 .4443 L
.42838 .44296 L
.42857 .44162 L
.42876 .44028 L
.42895 .43894 L
.42914 .4376 L
.42933 .43627 L
.42952 .43493 L
.42971 .4336 L
.4299 .43227 L
.4301 .43094 L
.43029 .42961 L
.43048 .42828 L
.43067 .42696 L
.43086 .42563 L
.43105 .42431 L
.43124 .42299 L
.43143 .43469 L
.43162 .43336 L
.43181 .43204 L
.432 .43072 L
.43219 .42939 L
.43238 .44107 L
.43257 .43974 L
.43276 .43841 L
.43295 .43709 L
.43314 .43577 L
Mistroke
.43333 .43445 L
.43352 .43313 L
.43371 .43181 L
.4339 .43049 L
.4341 .42918 L
.43429 .42786 L
.43448 .42655 L
.43467 .42523 L
.43486 .42392 L
.43505 .42261 L
.43524 .4213 L
.43543 .42 L
.43562 .41869 L
.43581 .43027 L
.436 .42896 L
.43619 .42765 L
.43638 .42634 L
.43657 .42504 L
.43676 .42373 L
.43695 .42243 L
.43714 .42113 L
.43733 .41982 L
.43752 .41852 L
.43771 .41722 L
.4379 .41593 L
.4381 .41463 L
.43829 .41333 L
.43848 .41204 L
.43867 .41075 L
.43886 .40945 L
.43905 .40816 L
.43924 .40687 L
.43943 .40559 L
.43962 .4043 L
.43981 .40301 L
.44 .41448 L
.44019 .41319 L
.44038 .4119 L
.44057 .41062 L
.44076 .40933 L
.44095 .40805 L
.44114 .40676 L
.44133 .40548 L
.44152 .4042 L
.44171 .40292 L
.4419 .40164 L
.4421 .40036 L
.44229 .39908 L
.44248 .39781 L
.44267 .39653 L
Mistroke
.44286 .39526 L
.44305 .39399 L
.44324 .39272 L
.44343 .39145 L
.44362 .39018 L
.44381 .40155 L
.444 .40028 L
.44419 .39901 L
.44438 .39774 L
.44457 .40909 L
.44476 .40782 L
.44495 .40654 L
.44514 .40527 L
.44533 .404 L
.44552 .40273 L
.44571 .40147 L
.4459 .4002 L
.4461 .39894 L
.44629 .39767 L
.44648 .39641 L
.44667 .39515 L
.44686 .40644 L
.44705 .41771 L
.44724 .42898 L
.44743 .42771 L
.44762 .42644 L
.44781 .42516 L
.448 .42389 L
.44819 .42262 L
.44838 .43386 L
.44857 .43259 L
.44876 .43132 L
.44895 .43005 L
.44914 .44126 L
.44933 .43998 L
.44952 .43871 L
.44971 .43744 L
.4499 .43617 L
.4501 .4349 L
.45029 .43363 L
.45048 .43236 L
.45067 .44354 L
.45086 .44227 L
.45105 .441 L
.45124 .43973 L
.45143 .45088 L
.45162 .4496 L
.45181 .44833 L
.452 .44707 L
.45219 .4458 L
Mistroke
.45238 .44453 L
.45257 .44326 L
.45276 .442 L
.45295 .44074 L
.45314 .43947 L
.45333 .43821 L
.45352 .43695 L
.45371 .43569 L
.4539 .43444 L
.4541 .43318 L
.45429 .43192 L
.45448 .43067 L
.45467 .42942 L
.45486 .42816 L
.45505 .42691 L
.45524 .42566 L
.45543 .42441 L
.45562 .42317 L
.45581 .42192 L
.456 .42067 L
.45619 .41943 L
.45638 .41818 L
.45657 .41694 L
.45676 .4157 L
.45695 .41446 L
.45714 .41322 L
.45733 .41198 L
.45752 .41075 L
.45771 .40951 L
.4579 .40828 L
.4581 .40704 L
.45829 .40581 L
.45848 .40458 L
.45867 .40335 L
.45886 .40212 L
.45905 .40089 L
.45924 .39966 L
.45943 .39844 L
.45962 .39721 L
.45981 .39599 L
.46 .39476 L
.46019 .39354 L
.46038 .39232 L
.46057 .40326 L
.46076 .40203 L
.46095 .40081 L
.46114 .39959 L
.46133 .39837 L
.46152 .39715 L
.46171 .39593 L
Mistroke
.4619 .39471 L
.4621 .39349 L
.46229 .39228 L
.46248 .39106 L
.46267 .38985 L
.46286 .38864 L
.46305 .38743 L
.46324 .38622 L
.46343 .38501 L
.46362 .3838 L
.46381 .38259 L
.464 .38139 L
.46419 .38018 L
.46438 .37898 L
.46457 .37777 L
.46476 .37657 L
.46495 .37537 L
.46514 .37417 L
.46533 .37297 L
.46552 .37178 L
.46571 .37058 L
.4659 .36938 L
.4661 .36819 L
.46629 .36699 L
.46648 .3658 L
.46667 .36461 L
.46686 .36342 L
.46705 .36223 L
.46724 .36104 L
.46743 .35985 L
.46762 .35867 L
.46781 .35748 L
.468 .3563 L
.46819 .35511 L
.46838 .36587 L
.46857 .36469 L
.46876 .3635 L
.46895 .36232 L
.46914 .36113 L
.46933 .35995 L
.46952 .35877 L
.46971 .35759 L
.4699 .35641 L
.4701 .35523 L
.47029 .35405 L
.47048 .36476 L
.47067 .36358 L
.47086 .3624 L
.47105 .36123 L
.47124 .36005 L
Mistroke
.47143 .35887 L
.47162 .3577 L
.47181 .35652 L
.472 .35535 L
.47219 .35418 L
.47238 .353 L
.47257 .35183 L
.47276 .35066 L
.47295 .3495 L
.47314 .34833 L
.47333 .34716 L
.47352 .346 L
.47371 .34483 L
.4739 .34367 L
.4741 .34251 L
.47429 .34134 L
.47448 .34018 L
.47467 .33902 L
.47486 .34964 L
.47505 .34847 L
.47524 .34731 L
.47543 .34615 L
.47562 .34499 L
.47581 .34383 L
.476 .34267 L
.47619 .34152 L
.47638 .34036 L
.47657 .33921 L
.47676 .33805 L
.47695 .34862 L
.47714 .34746 L
.47733 .3463 L
.47752 .34515 L
.47771 .34399 L
.4779 .34284 L
.4781 .35338 L
.47829 .3639 L
.47848 .36274 L
.47867 .36158 L
.47886 .36043 L
.47905 .35927 L
.47924 .35811 L
.47943 .35696 L
.47962 .3558 L
.47981 .35465 L
.48 .3535 L
.48019 .35235 L
.48038 .3512 L
.48057 .35005 L
.48076 .3489 L
Mistroke
.48095 .34775 L
.48114 .34661 L
.48133 .34546 L
.48152 .35592 L
.48171 .35477 L
.4819 .35362 L
.4821 .35247 L
.48229 .35133 L
.48248 .35018 L
.48267 .34904 L
.48286 .3479 L
.48305 .34676 L
.48324 .34561 L
.48343 .34447 L
.48362 .34334 L
.48381 .3422 L
.484 .34106 L
.48419 .33993 L
.48438 .33879 L
.48457 .33766 L
.48476 .33652 L
.48495 .33539 L
.48514 .33426 L
.48533 .33313 L
.48552 .332 L
.48571 .33087 L
.4859 .32974 L
.4861 .32862 L
.48629 .32749 L
.48648 .32637 L
.48667 .32524 L
.48686 .32412 L
.48705 .323 L
.48724 .32188 L
.48743 .32076 L
.48762 .31964 L
.48781 .32996 L
.488 .32884 L
.48819 .32772 L
.48838 .3266 L
.48857 .32548 L
.48876 .32436 L
.48895 .32324 L
.48914 .32212 L
.48933 .32101 L
.48952 .31989 L
.48971 .31878 L
.4899 .31767 L
.4901 .31656 L
.49029 .31544 L
Mistroke
.49048 .31433 L
.49067 .31322 L
.49086 .31212 L
.49105 .31101 L
.49124 .3099 L
.49143 .3088 L
.49162 .30769 L
.49181 .30659 L
.492 .30548 L
.49219 .30438 L
.49238 .31461 L
.49257 .31351 L
.49276 .3124 L
.49295 .32262 L
.49314 .32151 L
.49333 .3204 L
.49352 .3306 L
.49371 .32949 L
.4939 .32838 L
.4941 .32728 L
.49429 .32617 L
.49448 .32507 L
.49467 .32396 L
.49486 .33413 L
.49505 .33302 L
.49524 .33192 L
.49543 .33081 L
.49562 .32971 L
.49581 .3286 L
.496 .3275 L
.49619 .3264 L
.49638 .3253 L
.49657 .3242 L
.49676 .3231 L
.49695 .322 L
.49714 .3209 L
.49733 .31981 L
.49752 .31871 L
.49771 .31762 L
.4979 .31652 L
.4981 .31543 L
.49829 .31434 L
.49848 .31325 L
.49867 .31216 L
.49886 .31107 L
.49905 .30998 L
.49924 .30889 L
.49943 .30781 L
.49962 .30672 L
.49981 .30563 L
Mistroke
.5 .30455 L
.50019 .30347 L
.50038 .31352 L
.50057 .31244 L
.50076 .31135 L
.50095 .31027 L
.50114 .30919 L
.50133 .3081 L
.50152 .30702 L
.50171 .30594 L
.5019 .30486 L
.5021 .30378 L
.50229 .3027 L
.50248 .30163 L
.50267 .30055 L
.50286 .29947 L
.50305 .2984 L
.50324 .3084 L
.50343 .30732 L
.50362 .31731 L
.50381 .32729 L
.504 .32621 L
.50419 .32512 L
.50438 .32404 L
.50457 .32296 L
.50476 .32188 L
.50495 .3208 L
.50514 .31972 L
.50533 .31865 L
.50552 .31757 L
.50571 .31649 L
.5059 .31542 L
.5061 .31434 L
.50629 .31327 L
.50648 .3122 L
.50667 .32212 L
.50686 .32105 L
.50705 .31997 L
.50724 .3189 L
.50743 .31782 L
.50762 .31675 L
.50781 .31568 L
.508 .32558 L
.50819 .3245 L
.50838 .32343 L
.50857 .32236 L
.50876 .32129 L
.50895 .32022 L
.50914 .31915 L
.50933 .31808 L
Mistroke
.50952 .31701 L
.50971 .31594 L
.5099 .31488 L
.5101 .31381 L
.51029 .31275 L
.51048 .31168 L
.51067 .31062 L
.51086 .30956 L
.51105 .3085 L
.51124 .30744 L
.51143 .30638 L
.51162 .30532 L
.51181 .30426 L
.512 .30321 L
.51219 .30215 L
.51238 .3011 L
.51257 .30004 L
.51276 .29899 L
.51295 .29793 L
.51314 .30773 L
.51333 .30668 L
.51352 .31646 L
.51371 .32624 L
.5139 .32518 L
.5141 .33495 L
.51429 .33388 L
.51448 .33282 L
.51467 .33176 L
.51486 .3307 L
.51505 .32964 L
.51524 .32858 L
.51543 .32752 L
.51562 .32646 L
.51581 .33619 L
.516 .33513 L
.51619 .33407 L
.51638 .33301 L
.51657 .33196 L
.51676 .34167 L
.51695 .34061 L
.51714 .33955 L
.51733 .33849 L
.51752 .33743 L
.51771 .33637 L
.5179 .33532 L
.5181 .33426 L
.51829 .3332 L
.51848 .33215 L
.51867 .3311 L
.51886 .34077 L
Mistroke
.51905 .33971 L
.51924 .33866 L
.51943 .3376 L
.51962 .33655 L
.51981 .3355 L
.52 .33444 L
.52019 .33339 L
.52038 .33234 L
.52057 .33129 L
.52076 .33025 L
.52095 .3292 L
.52114 .32815 L
.52133 .3271 L
.52152 .32606 L
.52171 .32501 L
.5219 .32397 L
.5221 .32293 L
.52229 .32189 L
.52248 .32084 L
.52267 .33045 L
.52286 .3294 L
.52305 .32836 L
.52324 .32732 L
.52343 .3369 L
.52362 .33586 L
.52381 .33481 L
.524 .33377 L
.52419 .33273 L
.52438 .33168 L
.52457 .33064 L
.52476 .3402 L
.52495 .33916 L
.52514 .33812 L
.52533 .33707 L
.52552 .33603 L
.52571 .33499 L
.5259 .33395 L
.5261 .33292 L
.52629 .33188 L
.52648 .3414 L
.52667 .34036 L
.52686 .34988 L
.52705 .34883 L
.52724 .34779 L
.52743 .34675 L
.52762 .34571 L
.52781 .34467 L
.528 .34363 L
.52819 .34259 L
.52838 .34156 L
Mistroke
.52857 .34052 L
.52876 .35 L
.52895 .34896 L
.52914 .34792 L
.52933 .34689 L
.52952 .34585 L
.52971 .34481 L
.5299 .34378 L
.5301 .35323 L
.53029 .35219 L
.53048 .35116 L
.53067 .35012 L
.53086 .34909 L
.53105 .34805 L
.53124 .34702 L
.53143 .34599 L
.53162 .34496 L
.53181 .34392 L
.532 .34289 L
.53219 .34187 L
.53238 .34084 L
.53257 .33981 L
.53276 .33878 L
.53295 .33776 L
.53314 .33673 L
.53333 .3357 L
.53352 .33468 L
.53371 .33366 L
.5339 .33263 L
.5341 .33161 L
.53429 .33059 L
.53448 .32957 L
.53467 .32855 L
.53486 .32753 L
.53505 .32652 L
.53524 .3255 L
.53543 .32448 L
.53562 .32347 L
.53581 .32245 L
.536 .32144 L
.53619 .32042 L
.53638 .31941 L
.53657 .3184 L
.53676 .31739 L
.53695 .31638 L
.53714 .31537 L
.53733 .31436 L
.53752 .31335 L
.53771 .31234 L
.5379 .31134 L
Mistroke
.5381 .31033 L
.53829 .30933 L
.53848 .30832 L
.53867 .30732 L
.53886 .30632 L
.53905 .30531 L
.53924 .31461 L
.53943 .31361 L
.53962 .3126 L
.53981 .3116 L
.54 .3106 L
.54019 .3096 L
.54038 .3086 L
.54057 .3076 L
.54076 .3066 L
.54095 .3056 L
.54114 .31486 L
.54133 .31386 L
.54152 .31286 L
.54171 .31186 L
.5419 .31086 L
.5421 .30986 L
.54229 .30887 L
.54248 .30787 L
.54267 .30688 L
.54286 .30588 L
.54305 .30489 L
.54324 .30389 L
.54343 .31312 L
.54362 .31212 L
.54381 .31113 L
.544 .31013 L
.54419 .30914 L
.54438 .30814 L
.54457 .30715 L
.54476 .30616 L
.54495 .30517 L
.54514 .30418 L
.54533 .31337 L
.54552 .31238 L
.54571 .31139 L
.5459 .3104 L
.5461 .30941 L
.54629 .30842 L
.54648 .30743 L
.54667 .30644 L
.54686 .30545 L
.54705 .31461 L
.54724 .31362 L
.54743 .31263 L
Mistroke
.54762 .31165 L
.54781 .31066 L
.548 .30967 L
.54819 .30869 L
.54838 .3077 L
.54857 .30672 L
.54876 .30573 L
.54895 .30475 L
.54914 .30377 L
.54933 .30278 L
.54952 .3018 L
.54971 .30082 L
.5499 .29984 L
.5501 .29887 L
.55029 .29789 L
.55048 .30699 L
.55067 .30601 L
.55086 .30503 L
.55105 .31412 L
.55124 .31314 L
.55143 .31216 L
.55162 .31118 L
.55181 .3102 L
.552 .30922 L
.55219 .30824 L
.55238 .30726 L
.55257 .30629 L
.55276 .30531 L
.55295 .30433 L
.55314 .30336 L
.55333 .30238 L
.55352 .30141 L
.55371 .30044 L
.5539 .29947 L
.5541 .2985 L
.55429 .29752 L
.55448 .29656 L
.55467 .29559 L
.55486 .29462 L
.55505 .29365 L
.55524 .29268 L
.55543 .29172 L
.55562 .29075 L
.55581 .28979 L
.556 .28882 L
.55619 .29783 L
.55638 .29686 L
.55657 .2959 L
.55676 .30489 L
.55695 .30393 L
Mistroke
.55714 .30296 L
.55733 .30199 L
.55752 .30103 L
.55771 .30006 L
.5579 .2991 L
.5581 .29813 L
.55829 .29717 L
.55848 .29621 L
.55867 .29524 L
.55886 .29428 L
.55905 .29332 L
.55924 .30228 L
.55943 .30132 L
.55962 .30035 L
.55981 .29939 L
.56 .29843 L
.56019 .29747 L
.56038 .29651 L
.56057 .29555 L
.56076 .2946 L
.56095 .29364 L
.56114 .29268 L
.56133 .29173 L
.56152 .29077 L
.56171 .29969 L
.5619 .29873 L
.5621 .29777 L
.56229 .29682 L
.56248 .29586 L
.56267 .29491 L
.56286 .29395 L
.56305 .293 L
.56324 .29205 L
.56343 .30093 L
.56362 .29998 L
.56381 .29903 L
.564 .29807 L
.56419 .29712 L
.56438 .30599 L
.56457 .30503 L
.56476 .30408 L
.56495 .30313 L
.56514 .30217 L
.56533 .30122 L
.56552 .30027 L
.56571 .29932 L
.5659 .29837 L
.5661 .29742 L
.56629 .29647 L
.56648 .29552 L
Mistroke
.56667 .29458 L
.56686 .29363 L
.56705 .29268 L
.56724 .29174 L
.56743 .29079 L
.56762 .28985 L
.56781 .2889 L
.568 .28796 L
.56819 .28702 L
.56838 .28608 L
.56857 .28514 L
.56876 .2842 L
.56895 .28326 L
.56914 .28232 L
.56933 .28138 L
.56952 .28044 L
.56971 .27951 L
.5699 .27857 L
.5701 .27763 L
.57029 .2767 L
.57048 .27577 L
.57067 .27483 L
.57086 .2739 L
.57105 .27297 L
.57124 .27203 L
.57143 .2711 L
.57162 .27017 L
.57181 .26924 L
.572 .26831 L
.57219 .26739 L
.57238 .26646 L
.57257 .26553 L
.57276 .26461 L
.57295 .26368 L
.57314 .26275 L
.57333 .27149 L
.57352 .27056 L
.57371 .26964 L
.5739 .26871 L
.5741 .26779 L
.57429 .2765 L
.57448 .27558 L
.57467 .27465 L
.57486 .27372 L
.57505 .2728 L
.57524 .27187 L
.57543 .28057 L
.57562 .28927 L
.57581 .28834 L
.576 .28741 L
Mistroke
.57619 .28648 L
.57638 .28555 L
.57657 .28463 L
.57676 .2837 L
.57695 .28277 L
.57714 .28185 L
.57733 .28093 L
.57752 .28 L
.57771 .27908 L
.5779 .27816 L
.5781 .28681 L
.57829 .28589 L
.57848 .28496 L
.57867 .29361 L
.57886 .29268 L
.57905 .29176 L
.57924 .29083 L
.57943 .28991 L
.57962 .28898 L
.57981 .28806 L
.58 .28714 L
.58019 .28622 L
.58038 .2853 L
.58057 .28438 L
.58076 .29299 L
.58095 .29207 L
.58114 .29115 L
.58133 .29023 L
.58152 .2893 L
.58171 .28838 L
.5819 .28747 L
.5821 .29606 L
.58229 .29514 L
.58248 .29422 L
.58267 .2933 L
.58286 .29238 L
.58305 .29146 L
.58324 .30003 L
.58343 .29911 L
.58362 .29819 L
.58381 .29727 L
.584 .29635 L
.58419 .29543 L
.58438 .30399 L
.58457 .30307 L
.58476 .30215 L
.58495 .30123 L
.58514 .30031 L
.58533 .29939 L
.58552 .29848 L
Mistroke
.58571 .29756 L
.5859 .29664 L
.5861 .29573 L
.58629 .30425 L
.58648 .31277 L
.58667 .32128 L
.58686 .32036 L
.58705 .31944 L
.58724 .31852 L
.58743 .32702 L
.58762 .32609 L
.58781 .32517 L
.588 .32425 L
.58819 .32333 L
.58838 .32241 L
.58857 .33089 L
.58876 .32997 L
.58895 .32905 L
.58914 .32813 L
.58933 .3272 L
.58952 .32629 L
.58971 .32537 L
.5899 .32445 L
.5901 .32353 L
.59029 .32261 L
.59048 .3217 L
.59067 .32078 L
.59086 .31986 L
.59105 .31895 L
.59124 .31803 L
.59143 .31712 L
.59162 .31621 L
.59181 .3153 L
.592 .31438 L
.59219 .31347 L
.59238 .31256 L
.59257 .31165 L
.59276 .31074 L
.59295 .30983 L
.59314 .30893 L
.59333 .30802 L
.59352 .31643 L
.59371 .31552 L
.5939 .32392 L
.5941 .33232 L
.59429 .33141 L
.59448 .3305 L
.59467 .32958 L
.59486 .32867 L
.59505 .32776 L
Mistroke
.59524 .32685 L
.59543 .32594 L
.59562 .32503 L
.59581 .32412 L
.596 .32321 L
.59619 .3223 L
.59638 .3214 L
.59657 .32049 L
.59676 .31958 L
.59695 .31868 L
.59714 .31777 L
.59733 .31687 L
.59752 .32522 L
.59771 .32431 L
.5979 .33266 L
.5981 .33175 L
.59829 .33084 L
.59848 .32993 L
.59867 .32903 L
.59886 .32812 L
.59905 .32722 L
.59924 .32631 L
.59943 .32541 L
.59962 .32451 L
.59981 .3236 L
.6 .3227 L
.60019 .3218 L
.60038 .3209 L
.60057 .32 L
.60076 .3191 L
.60095 .3182 L
.60114 .3173 L
.60133 .31641 L
.60152 .31551 L
.60171 .31461 L
.6019 .31372 L
.6021 .31282 L
.60229 .31193 L
.60248 .31103 L
.60267 .31014 L
.60286 .31841 L
.60305 .31752 L
.60324 .31662 L
.60343 .31573 L
.60362 .31484 L
.60381 .31394 L
.604 .31305 L
.60419 .31216 L
.60438 .31127 L
.60457 .31038 L
Mistroke
.60476 .30949 L
.60495 .31773 L
.60514 .31684 L
.60533 .31595 L
.60552 .31506 L
.60571 .31417 L
.6059 .31328 L
.6061 .31239 L
.60629 .3115 L
.60648 .31061 L
.60667 .31883 L
.60686 .31794 L
.60705 .32616 L
.60724 .32526 L
.60743 .32437 L
.60762 .32348 L
.60781 .32259 L
.608 .3217 L
.60819 .32082 L
.60838 .31993 L
.60857 .31904 L
.60876 .31815 L
.60895 .31727 L
.60914 .31638 L
.60933 .3155 L
.60952 .31461 L
.60971 .32279 L
.6099 .33096 L
.6101 .33913 L
.61029 .33824 L
.61048 .33735 L
.61067 .33646 L
.61086 .33557 L
.61105 .34372 L
.61124 .34283 L
.61143 .34194 L
.61162 .34105 L
.61181 .34016 L
.612 .33927 L
.61219 .33838 L
.61238 .34651 L
.61257 .34562 L
.61276 .34473 L
.61295 .34384 L
.61314 .34296 L
.61333 .34207 L
.61352 .34118 L
.61371 .34029 L
.6139 .33941 L
.6141 .33852 L
Mistroke
.61429 .33764 L
.61448 .33675 L
.61467 .33587 L
.61486 .33499 L
.61505 .34308 L
.61524 .3422 L
.61543 .34131 L
.61562 .34043 L
.61581 .33955 L
.616 .33866 L
.61619 .33778 L
.61638 .3369 L
.61657 .33602 L
.61676 .33514 L
.61695 .33426 L
.61714 .33338 L
.61733 .3325 L
.61752 .33162 L
.61771 .33075 L
.6179 .32987 L
.6181 .33793 L
.61829 .33705 L
.61848 .33617 L
.61867 .34422 L
.61886 .34334 L
.61905 .35138 L
.61924 .35049 L
.61943 .34961 L
.61962 .34873 L
.61981 .34785 L
.62 .34697 L
.62019 .34609 L
.62038 .34522 L
.62057 .34434 L
.62076 .34346 L
.62095 .34258 L
.62114 .3506 L
.62133 .3586 L
.62152 .35772 L
.62171 .35684 L
.6219 .35596 L
.6221 .35508 L
.62229 .35421 L
.62248 .35333 L
.62267 .36132 L
.62286 .36044 L
.62305 .36842 L
.62324 .36754 L
.62343 .36666 L
.62362 .36578 L
Mistroke
.62381 .3649 L
.624 .36402 L
.62419 .36314 L
.62438 .36226 L
.62457 .36138 L
.62476 .36051 L
.62495 .35963 L
.62514 .35876 L
.62533 .35788 L
.62552 .35701 L
.62571 .35613 L
.6259 .35526 L
.6261 .35439 L
.62629 .35351 L
.62648 .35264 L
.62667 .35177 L
.62686 .3509 L
.62705 .35883 L
.62724 .35796 L
.62743 .35709 L
.62762 .35622 L
.62781 .35534 L
.628 .35447 L
.62819 .3536 L
.62838 .35274 L
.62857 .35187 L
.62876 .351 L
.62895 .35013 L
.62914 .34927 L
.62933 .3484 L
.62952 .34753 L
.62971 .34667 L
.6299 .3458 L
.6301 .34494 L
.63029 .34408 L
.63048 .34321 L
.63067 .34235 L
.63086 .34149 L
.63105 .34063 L
.63124 .33977 L
.63143 .33891 L
.63162 .34678 L
.63181 .34592 L
.632 .34506 L
.63219 .3442 L
.63238 .34334 L
.63257 .34248 L
.63276 .34162 L
.63295 .34076 L
.63314 .3399 L
Mistroke
.63333 .33904 L
.63352 .33819 L
.63371 .33733 L
.6339 .33647 L
.6341 .33562 L
.63429 .33476 L
.63448 .33391 L
.63467 .33305 L
.63486 .3322 L
.63505 .33135 L
.63524 .3305 L
.63543 .32964 L
.63562 .32879 L
.63581 .33662 L
.636 .33576 L
.63619 .33491 L
.63638 .33406 L
.63657 .33321 L
.63676 .34102 L
.63695 .34017 L
.63714 .33931 L
.63733 .33846 L
.63752 .33761 L
.63771 .33676 L
.6379 .33591 L
.6381 .33506 L
.63829 .33421 L
.63848 .33336 L
.63867 .33251 L
.63886 .33166 L
.63905 .33082 L
.63924 .32997 L
.63943 .32912 L
.63962 .32828 L
.63981 .32743 L
.64 .32659 L
.64019 .32575 L
.64038 .3249 L
.64057 .32406 L
.64076 .32322 L
.64095 .32237 L
.64114 .33013 L
.64133 .32929 L
.64152 .33704 L
.64171 .3362 L
.6419 .33535 L
.6421 .33451 L
.64229 .33366 L
.64248 .33282 L
.64267 .33198 L
Mistroke
.64286 .33114 L
.64305 .33029 L
.64324 .32945 L
.64343 .32861 L
.64362 .32777 L
.64381 .32693 L
.644 .32609 L
.64419 .32526 L
.64438 .32442 L
.64457 .32358 L
.64476 .32274 L
.64495 .32191 L
.64514 .32107 L
.64533 .32024 L
.64552 .3194 L
.64571 .31857 L
.6459 .31773 L
.6461 .3169 L
.64629 .31607 L
.64648 .31524 L
.64667 .3144 L
.64686 .31357 L
.64705 .31274 L
.64724 .31191 L
.64743 .31108 L
.64762 .31025 L
.64781 .30942 L
.648 .3171 L
.64819 .31627 L
.64838 .31544 L
.64857 .31461 L
.64876 .31378 L
.64895 .31295 L
.64914 .31213 L
.64933 .3113 L
.64952 .31047 L
.64971 .30965 L
.6499 .30882 L
.6501 .308 L
.65029 .30717 L
.65048 .31482 L
.65067 .31399 L
.65086 .32163 L
.65105 .3208 L
.65124 .31998 L
.65143 .31915 L
.65162 .31832 L
.65181 .3175 L
.652 .31667 L
.65219 .31585 L
Mistroke
.65238 .31502 L
.65257 .3142 L
.65276 .31338 L
.65295 .31255 L
.65314 .31173 L
.65333 .31091 L
.65352 .31009 L
.65371 .3177 L
.6539 .31687 L
.6541 .31605 L
.65429 .32365 L
.65448 .32282 L
.65467 .322 L
.65486 .32118 L
.65505 .32036 L
.65524 .31953 L
.65543 .31871 L
.65562 .31789 L
.65581 .31707 L
.656 .31625 L
.65619 .31543 L
.65638 .323 L
.65657 .32218 L
.65676 .32136 L
.65695 .32054 L
.65714 .31972 L
.65733 .3189 L
.65752 .31809 L
.65771 .31727 L
.6579 .31645 L
.6581 .324 L
.65829 .32318 L
.65848 .32236 L
.65867 .32155 L
.65886 .32073 L
.65905 .31991 L
.65924 .31909 L
.65943 .31828 L
.65962 .31746 L
.65981 .31665 L
.66 .31583 L
.66019 .31502 L
.66038 .3142 L
.66057 .31339 L
.66076 .31258 L
.66095 .31177 L
.66114 .31095 L
.66133 .31014 L
.66152 .30933 L
.66171 .30852 L
Mistroke
.6619 .30771 L
.6621 .31522 L
.66229 .31441 L
.66248 .3136 L
.66267 .31279 L
.66286 .31198 L
.66305 .31117 L
.66324 .31036 L
.66343 .30955 L
.66362 .30874 L
.66381 .30794 L
.664 .30713 L
.66419 .30632 L
.66438 .3138 L
.66457 .313 L
.66476 .31219 L
.66495 .31138 L
.66514 .31057 L
.66533 .30977 L
.66552 .30896 L
.66571 .30816 L
.6659 .30735 L
.6661 .31481 L
.66629 .31401 L
.66648 .3132 L
.66667 .3124 L
.66686 .31159 L
.66705 .31079 L
.66724 .30998 L
.66743 .30918 L
.66762 .30838 L
.66781 .30757 L
.668 .30677 L
.66819 .30597 L
.66838 .30517 L
.66857 .30437 L
.66876 .30357 L
.66895 .30277 L
.66914 .30197 L
.66933 .30117 L
.66952 .30037 L
.66971 .29958 L
.6699 .29878 L
.6701 .29798 L
.67029 .29719 L
.67048 .29639 L
.67067 .29559 L
.67086 .2948 L
.67105 .30221 L
.67124 .30141 L
Mistroke
.67143 .30062 L
.67162 .29982 L
.67181 .29903 L
.672 .29823 L
.67219 .29744 L
.67238 .29664 L
.67257 .29585 L
.67276 .29506 L
.67295 .29427 L
.67314 .29347 L
.67333 .29268 L
.67352 .29189 L
.67371 .2911 L
.6739 .29031 L
.6741 .28952 L
.67429 .28873 L
.67448 .28794 L
.67467 .28716 L
.67486 .28637 L
.67505 .28558 L
.67524 .2848 L
.67543 .28401 L
.67562 .28322 L
.67581 .28244 L
.676 .28165 L
.67619 .28087 L
.67638 .28009 L
.67657 .2793 L
.67676 .27852 L
.67695 .27774 L
.67714 .27695 L
.67733 .27617 L
.67752 .27539 L
.67771 .27461 L
.6779 .27383 L
.6781 .27305 L
.67829 .27227 L
.67848 .2796 L
.67867 .27882 L
.67886 .27804 L
.67905 .27726 L
.67924 .27648 L
.67943 .2757 L
.67962 .27492 L
.67981 .27415 L
.68 .27337 L
.68019 .27259 L
.68038 .27182 L
.68057 .27104 L
.68076 .27026 L
Mistroke
.68095 .26949 L
.68114 .26871 L
.68133 .27601 L
.68152 .27524 L
.68171 .27446 L
.6819 .27369 L
.6821 .27291 L
.68229 .27214 L
.68248 .27942 L
.68267 .27865 L
.68286 .27787 L
.68305 .2771 L
.68324 .27632 L
.68343 .27555 L
.68362 .27477 L
.68381 .274 L
.684 .27323 L
.68419 .27245 L
.68438 .27168 L
.68457 .27091 L
.68476 .27014 L
.68495 .26937 L
.68514 .2686 L
.68533 .26783 L
.68552 .26706 L
.68571 .26629 L
.6859 .26552 L
.6861 .26476 L
.68629 .26399 L
.68648 .26322 L
.68667 .26245 L
.68686 .26169 L
.68705 .26092 L
.68724 .26016 L
.68743 .25939 L
.68762 .25863 L
.68781 .25786 L
.688 .2571 L
.68819 .25634 L
.68838 .26356 L
.68857 .27078 L
.68876 .27002 L
.68895 .26925 L
.68914 .26849 L
.68933 .26772 L
.68952 .27493 L
.68971 .27417 L
.6899 .2734 L
.6901 .27263 L
.69029 .27187 L
Mistroke
.69048 .2711 L
.69067 .27034 L
.69086 .27753 L
.69105 .27677 L
.69124 .28396 L
.69143 .28319 L
.69162 .28242 L
.69181 .28166 L
.692 .28089 L
.69219 .28013 L
.69238 .27936 L
.69257 .2786 L
.69276 .27783 L
.69295 .27707 L
.69314 .27631 L
.69333 .27554 L
.69352 .28271 L
.69371 .28195 L
.6939 .28118 L
.6941 .28042 L
.69429 .28757 L
.69448 .28681 L
.69467 .28605 L
.69486 .28528 L
.69505 .28452 L
.69524 .28376 L
.69543 .28299 L
.69562 .28223 L
.69581 .28147 L
.696 .28071 L
.69619 .27995 L
.69638 .27919 L
.69657 .27843 L
.69676 .27767 L
.69695 .27691 L
.69714 .27615 L
.69733 .27539 L
.69752 .27463 L
.69771 .28175 L
.6979 .281 L
.6981 .28024 L
.69829 .27948 L
.69848 .27872 L
.69867 .27796 L
.69886 .27721 L
.69905 .27645 L
.69924 .27569 L
.69943 .27494 L
.69962 .27418 L
.69981 .27343 L
Mistroke
.7 .27267 L
.70019 .27192 L
.70038 .27117 L
.70057 .27041 L
.70076 .2775 L
.70095 .27675 L
.70114 .27599 L
.70133 .27524 L
.70152 .27449 L
.70171 .27374 L
.7019 .27298 L
.7021 .27223 L
.70229 .27148 L
.70248 .27073 L
.70267 .26998 L
.70286 .26923 L
.70305 .26848 L
.70324 .26773 L
.70343 .26698 L
.70362 .26623 L
.70381 .26548 L
.704 .26473 L
.70419 .26399 L
.70438 .26324 L
.70457 .27029 L
.70476 .26954 L
.70495 .2688 L
.70514 .26805 L
.70533 .2673 L
.70552 .26656 L
.70571 .26581 L
.7059 .27285 L
.7061 .2721 L
.70629 .27135 L
.70648 .27838 L
.70667 .27763 L
.70686 .27689 L
.70705 .27614 L
.70724 .27539 L
.70743 .27465 L
.70762 .2739 L
.70781 .27315 L
.708 .28017 L
.70819 .27942 L
.70838 .27867 L
.70857 .27793 L
.70876 .27718 L
.70895 .27644 L
.70914 .27569 L
.70933 .27495 L
Mistroke
.70952 .2742 L
.70971 .27346 L
.7099 .27271 L
.7101 .27197 L
.71029 .27123 L
.71048 .27049 L
.71067 .26974 L
.71086 .269 L
.71105 .26826 L
.71124 .26752 L
.71143 .26678 L
.71162 .26604 L
.71181 .2653 L
.712 .26456 L
.71219 .27154 L
.71238 .2708 L
.71257 .27006 L
.71276 .26932 L
.71295 .26858 L
.71314 .26784 L
.71333 .2671 L
.71352 .26636 L
.71371 .26562 L
.7139 .26489 L
.7141 .26415 L
.71429 .2711 L
.71448 .27037 L
.71467 .26963 L
.71486 .26889 L
.71505 .26815 L
.71524 .26742 L
.71543 .26668 L
.71562 .26595 L
.71581 .26521 L
.716 .26448 L
.71619 .26374 L
.71638 .26301 L
.71657 .26228 L
.71676 .26154 L
.71695 .26081 L
.71714 .26008 L
.71733 .25935 L
.71752 .25861 L
.71771 .25788 L
.7179 .25715 L
.7181 .25642 L
.71829 .25569 L
.71848 .25496 L
.71867 .25423 L
.71886 .25351 L
Mistroke
.71905 .25278 L
.71924 .25205 L
.71943 .25132 L
.71962 .25059 L
.71981 .2575 L
.72 .26439 L
.72019 .26366 L
.72038 .26293 L
.72057 .2622 L
.72076 .26148 L
.72095 .26075 L
.72114 .26002 L
.72133 .25929 L
.72152 .25856 L
.72171 .25784 L
.7219 .25711 L
.7221 .25638 L
.72229 .25566 L
.72248 .25493 L
.72267 .25421 L
.72286 .25348 L
.72305 .25276 L
.72324 .25963 L
.72343 .2589 L
.72362 .25818 L
.72381 .25745 L
.724 .25673 L
.72419 .256 L
.72438 .25528 L
.72457 .25456 L
.72476 .26141 L
.72495 .26069 L
.72514 .25996 L
.72533 .25924 L
.72552 .25851 L
.72571 .25779 L
.7259 .25707 L
.7261 .25635 L
.72629 .26318 L
.72648 .26246 L
.72667 .26174 L
.72686 .26102 L
.72705 .26029 L
.72724 .25957 L
.72743 .25885 L
.72762 .25813 L
.72781 .25741 L
.728 .26423 L
.72819 .26351 L
.72838 .26279 L
Mistroke
.72857 .26206 L
.72876 .26134 L
.72895 .26062 L
.72914 .26743 L
.72933 .26671 L
.72952 .26599 L
.72971 .26527 L
.7299 .26455 L
.7301 .26383 L
.73029 .26311 L
.73048 .26239 L
.73067 .26167 L
.73086 .26095 L
.73105 .26023 L
.73124 .25952 L
.73143 .2588 L
.73162 .25808 L
.73181 .25737 L
.732 .25665 L
.73219 .25594 L
.73238 .25522 L
.73257 .25451 L
.73276 .26128 L
.73295 .26056 L
.73314 .25985 L
.73333 .25913 L
.73352 .25842 L
.73371 .2577 L
.7339 .25699 L
.7341 .25627 L
.73429 .25556 L
.73448 .25485 L
.73467 .25414 L
.73486 .25342 L
.73505 .25271 L
.73524 .252 L
.73543 .25129 L
.73562 .25058 L
.73581 .24987 L
.736 .24916 L
.73619 .24845 L
.73638 .24774 L
.73657 .24703 L
.73676 .24632 L
.73695 .25306 L
.73714 .25235 L
.73733 .25164 L
.73752 .25093 L
.73771 .25022 L
.7379 .24951 L
Mistroke
.7381 .24881 L
.73829 .2481 L
.73848 .24739 L
.73867 .24669 L
.73886 .24598 L
.73905 .2527 L
.73924 .25199 L
.73943 .25128 L
.73962 .25057 L
.73981 .24987 L
.74 .24916 L
.74019 .24846 L
.74038 .25516 L
.74057 .25445 L
.74076 .25375 L
.74095 .25304 L
.74114 .25234 L
.74133 .25163 L
.74152 .25092 L
.74171 .25022 L
.7419 .24952 L
.7421 .24881 L
.74229 .24811 L
.74248 .24741 L
.74267 .2467 L
.74286 .246 L
.74305 .2453 L
.74324 .25198 L
.74343 .25865 L
.74362 .25795 L
.74381 .25724 L
.744 .25654 L
.74419 .25583 L
.74438 .25513 L
.74457 .25443 L
.74476 .25373 L
.74495 .25302 L
.74514 .25232 L
.74533 .25162 L
.74552 .25092 L
.74571 .25022 L
.7459 .24952 L
.7461 .24882 L
.74629 .24812 L
.74648 .24742 L
.74667 .24672 L
.74686 .24602 L
.74705 .24532 L
.74724 .24463 L
.74743 .24393 L
Mistroke
.74762 .24323 L
.74781 .24254 L
.748 .24184 L
.74819 .24114 L
.74838 .24777 L
.74857 .24708 L
.74876 .24638 L
.74895 .24568 L
.74914 .24499 L
.74933 .25161 L
.74952 .25091 L
.74971 .25022 L
.7499 .24952 L
.7501 .24882 L
.75029 .24813 L
.75048 .24743 L
.75067 .24674 L
.75086 .24604 L
.75105 .24535 L
.75124 .24466 L
.75143 .24396 L
.75162 .24327 L
.75181 .24258 L
.752 .24917 L
.75219 .24848 L
.75238 .24779 L
.75257 .24709 L
.75276 .2464 L
.75295 .24571 L
.75314 .24502 L
.75333 .24432 L
.75352 .24363 L
.75371 .24294 L
.7539 .24225 L
.7541 .24156 L
.75429 .24087 L
.75448 .24745 L
.75467 .24675 L
.75486 .24606 L
.75505 .24537 L
.75524 .24468 L
.75543 .24399 L
.75562 .2433 L
.75581 .24262 L
.756 .24193 L
.75619 .24124 L
.75638 .24055 L
.75657 .23986 L
.75676 .23918 L
.75695 .23849 L
Mistroke
.75714 .2378 L
.75733 .23712 L
.75752 .23643 L
.75771 .23574 L
.7579 .23506 L
.7581 .23437 L
.75829 .23369 L
.75848 .23301 L
.75867 .23232 L
.75886 .23164 L
.75905 .23096 L
.75924 .23027 L
.75943 .22959 L
.75962 .23612 L
.75981 .24265 L
.76 .24197 L
.76019 .24129 L
.76038 .2406 L
.76057 .23992 L
.76076 .24644 L
.76095 .24575 L
.76114 .24507 L
.76133 .24438 L
.76152 .2437 L
.76171 .24302 L
.7619 .24233 L
.7621 .24165 L
.76229 .24097 L
.76248 .24029 L
.76267 .2396 L
.76286 .23892 L
.76305 .23824 L
.76324 .23756 L
.76343 .23688 L
.76362 .2362 L
.76381 .23552 L
.764 .23484 L
.76419 .23416 L
.76438 .23348 L
.76457 .2328 L
.76476 .23213 L
.76495 .23145 L
.76514 .23077 L
.76533 .23009 L
.76552 .23658 L
.76571 .2359 L
.7659 .23522 L
.7661 .23454 L
.76629 .23386 L
.76648 .23319 L
Mistroke
.76667 .23251 L
.76686 .23184 L
.76705 .23116 L
.76724 .23763 L
.76743 .23695 L
.76762 .23627 L
.76781 .24273 L
.768 .24205 L
.76819 .24138 L
.76838 .2407 L
.76857 .24002 L
.76876 .24647 L
.76895 .2458 L
.76914 .24512 L
.76933 .24444 L
.76952 .24377 L
.76971 .24309 L
.7699 .24241 L
.7701 .24174 L
.77029 .24106 L
.77048 .2475 L
.77067 .24682 L
.77086 .24615 L
.77105 .24547 L
.77124 .24479 L
.77143 .24412 L
.77162 .24344 L
.77181 .24277 L
.772 .2421 L
.77219 .24852 L
.77238 .24784 L
.77257 .24717 L
.77276 .25358 L
.77295 .25291 L
.77314 .25223 L
.77333 .25864 L
.77352 .25796 L
.77371 .25728 L
.7739 .25661 L
.7741 .25593 L
.77429 .25526 L
.77448 .25458 L
.77467 .26098 L
.77486 .2603 L
.77505 .25963 L
.77524 .25895 L
.77543 .25828 L
.77562 .2576 L
.77581 .26399 L
.776 .26331 L
Mistroke
.77619 .26264 L
.77638 .26196 L
.77657 .26129 L
.77676 .26061 L
.77695 .25994 L
.77714 .25926 L
.77733 .25859 L
.77752 .25792 L
.77771 .25725 L
.7779 .25657 L
.7781 .2559 L
.77829 .25523 L
.77848 .25456 L
.77867 .25389 L
.77886 .25322 L
.77905 .25255 L
.77924 .25188 L
.77943 .25121 L
.77962 .25054 L
.77981 .24987 L
.78 .2492 L
.78019 .25555 L
.78038 .25488 L
.78057 .25421 L
.78076 .25354 L
.78095 .25989 L
.78114 .25921 L
.78133 .25855 L
.78152 .26488 L
.78171 .26421 L
.7819 .26354 L
.7821 .26287 L
.78229 .2622 L
.78248 .26153 L
.78267 .26086 L
.78286 .26019 L
.78305 .25952 L
.78324 .25886 L
.78343 .25819 L
.78362 .25752 L
.78381 .25685 L
.784 .25619 L
.78419 .25552 L
.78438 .25485 L
.78457 .25419 L
.78476 .25352 L
.78495 .25286 L
.78514 .25219 L
.78533 .25153 L
.78552 .25086 L
Mistroke
.78571 .25717 L
.7859 .26347 L
.7861 .2628 L
.78629 .26214 L
.78648 .26147 L
.78667 .2608 L
.78686 .26014 L
.78705 .25947 L
.78724 .25881 L
.78743 .25814 L
.78762 .25748 L
.78781 .25682 L
.788 .25615 L
.78819 .26244 L
.78838 .26177 L
.78857 .26111 L
.78876 .26044 L
.78895 .25978 L
.78914 .25912 L
.78933 .25845 L
.78952 .25779 L
.78971 .25713 L
.7899 .25647 L
.7901 .25581 L
.79029 .25515 L
.79048 .25448 L
.79067 .25382 L
.79086 .25316 L
.79105 .2525 L
.79124 .25184 L
.79143 .25119 L
.79162 .25053 L
.79181 .24987 L
.792 .24921 L
.79219 .25546 L
.79238 .2548 L
.79257 .25414 L
.79276 .25348 L
.79295 .25283 L
.79314 .25217 L
.79333 .25151 L
.79352 .25085 L
.79371 .2502 L
.7939 .24954 L
.7941 .24888 L
.79429 .24823 L
.79448 .24757 L
.79467 .24692 L
.79486 .24626 L
.79505 .24561 L
Mistroke
.79524 .24495 L
.79543 .2443 L
.79562 .24364 L
.79581 .24299 L
.796 .24234 L
.79619 .24169 L
.79638 .24103 L
.79657 .24038 L
.79676 .23973 L
.79695 .23908 L
.79714 .23843 L
.79733 .23778 L
.79752 .23712 L
.79771 .23647 L
.7979 .23582 L
.7981 .23518 L
.79829 .23453 L
.79848 .23388 L
.79867 .23323 L
.79886 .23258 L
.79905 .23193 L
.79924 .23813 L
.79943 .23748 L
.79962 .23683 L
.79981 .23619 L
.8 .23554 L
.80019 .23489 L
.80038 .23424 L
.80057 .23359 L
.80076 .23295 L
.80095 .2323 L
.80114 .23166 L
.80133 .23784 L
.80152 .23719 L
.80171 .23654 L
.8019 .2359 L
.8021 .23525 L
.80229 .2346 L
.80248 .23396 L
.80267 .23331 L
.80286 .23267 L
.80305 .23202 L
.80324 .23138 L
.80343 .23755 L
.80362 .2369 L
.80381 .23626 L
.804 .23561 L
.80419 .23497 L
.80438 .23432 L
.80457 .23368 L
Mistroke
.80476 .23303 L
.80495 .23239 L
.80514 .23175 L
.80533 .23111 L
.80552 .23046 L
.80571 .22982 L
.8059 .22918 L
.8061 .23533 L
.80629 .23468 L
.80648 .23404 L
.80667 .2334 L
.80686 .23276 L
.80705 .23212 L
.80724 .23147 L
.80743 .23083 L
.80762 .23019 L
.80781 .22955 L
.808 .22891 L
.80819 .22827 L
.80838 .2344 L
.80857 .23376 L
.80876 .23312 L
.80895 .23248 L
.80914 .23184 L
.80933 .23796 L
.80952 .23732 L
.80971 .23668 L
.8099 .23604 L
.8101 .2354 L
.81029 .24151 L
.81048 .24087 L
.81067 .24023 L
.81086 .23959 L
.81105 .23895 L
.81124 .23831 L
.81143 .23767 L
.81162 .23703 L
.81181 .23639 L
.812 .23576 L
.81219 .23512 L
.81238 .23448 L
.81257 .23384 L
.81276 .23321 L
.81295 .23257 L
.81314 .23866 L
.81333 .23802 L
.81352 .24411 L
.81371 .24347 L
.8139 .24955 L
.8141 .24891 L
Mistroke
.81429 .24827 L
.81448 .24763 L
.81467 .24699 L
.81486 .24635 L
.81505 .24571 L
.81524 .24508 L
.81543 .24444 L
.81562 .2438 L
.81581 .24316 L
.816 .24253 L
.81619 .24189 L
.81638 .24126 L
.81657 .24062 L
.81676 .23998 L
.81695 .23935 L
.81714 .23871 L
.81733 .23808 L
.81752 .23745 L
.81771 .23681 L
.8179 .24286 L
.8181 .24223 L
.81829 .24828 L
.81848 .24764 L
.81867 .24701 L
.81886 .24637 L
.81905 .24574 L
.81924 .2451 L
.81943 .24447 L
.81962 .24383 L
.81981 .2432 L
.82 .24257 L
.82019 .24193 L
.82038 .2413 L
.82057 .24067 L
.82076 .24003 L
.82095 .2394 L
.82114 .24543 L
.82133 .2448 L
.82152 .24416 L
.82171 .25018 L
.8219 .24955 L
.8221 .25557 L
.82229 .26158 L
.82248 .26759 L
.82267 .26696 L
.82286 .26632 L
.82305 .26568 L
.82324 .26505 L
.82343 .26441 L
.82362 .26378 L
Mistroke
.82381 .26978 L
.824 .26914 L
.82419 .2685 L
.82438 .26787 L
.82457 .26723 L
.82476 .2666 L
.82495 .26596 L
.82514 .26533 L
.82533 .26469 L
.82552 .26406 L
.82571 .27004 L
.8259 .27603 L
.8261 .27539 L
.82629 .27476 L
.82648 .27412 L
.82667 .27348 L
.82686 .27285 L
.82705 .27221 L
.82724 .27158 L
.82743 .27755 L
.82762 .27692 L
.82781 .27628 L
.828 .27565 L
.82819 .27501 L
.82838 .28098 L
.82857 .28034 L
.82876 .2797 L
.82895 .27907 L
.82914 .27843 L
.82933 .2778 L
.82952 .28376 L
.82971 .28312 L
.8299 .28248 L
.8301 .28185 L
.83029 .2878 L
.83048 .28716 L
.83067 .28653 L
.83086 .28589 L
.83105 .28526 L
.83124 .28462 L
.83143 .28399 L
.83162 .28993 L
.83181 .28929 L
.832 .28866 L
.83219 .28802 L
.83238 .28739 L
.83257 .28675 L
.83276 .28612 L
.83295 .29205 L
.83314 .29141 L
Mistroke
.83333 .29078 L
.83352 .29014 L
.83371 .28951 L
.8339 .28888 L
.8341 .28824 L
.83429 .28761 L
.83448 .28698 L
.83467 .28635 L
.83486 .28571 L
.83505 .28508 L
.83524 .28445 L
.83543 .28382 L
.83562 .28973 L
.83581 .2891 L
.836 .28847 L
.83619 .29437 L
.83638 .29374 L
.83657 .2931 L
.83676 .29247 L
.83695 .29184 L
.83714 .29774 L
.83733 .2971 L
.83752 .29647 L
.83771 .29584 L
.8379 .29521 L
.8381 .29458 L
.83829 .29394 L
.83848 .29331 L
.83867 .29268 L
.83886 .29205 L
.83905 .29793 L
.83924 .2973 L
.83943 .29667 L
.83962 .29604 L
.83981 .30192 L
.84 .30129 L
.84019 .30065 L
.84038 .30002 L
.84057 .29939 L
.84076 .29876 L
.84095 .29813 L
.84114 .2975 L
.84133 .29687 L
.84152 .29624 L
.84171 .29561 L
.8419 .29499 L
.8421 .29436 L
.84229 .29373 L
.84248 .2931 L
.84267 .29247 L
Mistroke
.84286 .29185 L
.84305 .29122 L
.84324 .29059 L
.84343 .28997 L
.84362 .28934 L
.84381 .28871 L
.844 .28809 L
.84419 .28746 L
.84438 .28684 L
.84457 .28621 L
.84476 .28559 L
.84495 .28497 L
.84514 .28434 L
.84533 .28372 L
.84552 .2831 L
.84571 .28893 L
.8459 .28831 L
.8461 .28768 L
.84629 .29352 L
.84648 .29289 L
.84667 .29227 L
.84686 .29164 L
.84705 .29747 L
.84724 .29684 L
.84743 .29622 L
.84762 .29559 L
.84781 .29497 L
.848 .29434 L
.84819 .29372 L
.84838 .2931 L
.84857 .29247 L
.84876 .29185 L
.84895 .29123 L
.84914 .29061 L
.84933 .28999 L
.84952 .28936 L
.84971 .28874 L
.8499 .28812 L
.8501 .2875 L
.85029 .28688 L
.85048 .28626 L
.85067 .28564 L
.85086 .28502 L
.85105 .2844 L
.85124 .28378 L
.85143 .28316 L
.85162 .28896 L
.85181 .29475 L
.852 .29413 L
.85219 .29992 L
Mistroke
.85238 .2993 L
.85257 .29868 L
.85276 .29805 L
.85295 .29743 L
.85314 .29681 L
.85333 .30259 L
.85352 .30197 L
.85371 .30135 L
.8539 .30073 L
.8541 .30011 L
.85429 .30588 L
.85448 .30526 L
.85467 .30464 L
.85486 .30402 L
.85505 .3034 L
.85524 .30278 L
.85543 .30216 L
.85562 .30792 L
.85581 .3073 L
.856 .30668 L
.85619 .30606 L
.85638 .30544 L
.85657 .30482 L
.85676 .3042 L
.85695 .30358 L
.85714 .30296 L
.85733 .30234 L
.85752 .30172 L
.85771 .3011 L
.8579 .30048 L
.8581 .29987 L
.85829 .29925 L
.85848 .29863 L
.85867 .29802 L
.85886 .2974 L
.85905 .29678 L
.85924 .29617 L
.85943 .3019 L
.85962 .30129 L
.85981 .30067 L
.86 .30006 L
.86019 .29944 L
.86038 .29882 L
.86057 .29821 L
.86076 .29759 L
.86095 .29698 L
.86114 .3027 L
.86133 .30209 L
.86152 .30147 L
.86171 .30086 L
Mistroke
.8619 .30024 L
.8621 .29963 L
.86229 .29901 L
.86248 .2984 L
.86267 .29779 L
.86286 .29717 L
.86305 .29656 L
.86324 .29595 L
.86343 .29533 L
.86362 .29472 L
.86381 .29411 L
.864 .2935 L
.86419 .29289 L
.86438 .29227 L
.86457 .29166 L
.86476 .29105 L
.86495 .29044 L
.86514 .28983 L
.86533 .28922 L
.86552 .28861 L
.86571 .29431 L
.8659 .2937 L
.8661 .29939 L
.86629 .29878 L
.86648 .29817 L
.86667 .29756 L
.86686 .29695 L
.86705 .29634 L
.86724 .29573 L
.86743 .30141 L
.86762 .30709 L
.86781 .30648 L
.868 .30587 L
.86819 .31154 L
.86838 .31093 L
.86857 .31032 L
.86876 .30971 L
.86895 .3091 L
.86914 .30848 L
.86933 .30787 L
.86952 .30726 L
.86971 .30665 L
.8699 .31232 L
.8701 .3117 L
.87029 .31109 L
.87048 .31048 L
.87067 .31614 L
.87086 .3218 L
.87105 .32118 L
.87124 .32057 L
Mistroke
.87143 .31996 L
.87162 .31935 L
.87181 .31873 L
.872 .31812 L
.87219 .31751 L
.87238 .3169 L
.87257 .31629 L
.87276 .32193 L
.87295 .32132 L
.87314 .32071 L
.87333 .3201 L
.87352 .31949 L
.87371 .31888 L
.8739 .31827 L
.8741 .31766 L
.87429 .31705 L
.87448 .31644 L
.87467 .31583 L
.87486 .31522 L
.87505 .31461 L
.87524 .314 L
.87543 .3134 L
.87562 .31279 L
.87581 .31218 L
.876 .31157 L
.87619 .31097 L
.87638 .31036 L
.87657 .30975 L
.87676 .31537 L
.87695 .31476 L
.87714 .31416 L
.87733 .31977 L
.87752 .31916 L
.87771 .31855 L
.8779 .31795 L
.8781 .31734 L
.87829 .31673 L
.87848 .31613 L
.87867 .31552 L
.87886 .31491 L
.87905 .31431 L
.87924 .3137 L
.87943 .3131 L
.87962 .31249 L
.87981 .31189 L
.88 .31128 L
.88019 .31068 L
.88038 .31008 L
.88057 .30947 L
.88076 .30887 L
Mistroke
.88095 .30827 L
.88114 .30766 L
.88133 .30706 L
.88152 .31265 L
.88171 .31205 L
.8819 .31144 L
.8821 .31084 L
.88229 .31024 L
.88248 .30964 L
.88267 .30903 L
.88286 .30843 L
.88305 .30783 L
.88324 .30723 L
.88343 .30663 L
.88362 .30603 L
.88381 .30543 L
.884 .30483 L
.88419 .30423 L
.88438 .30363 L
.88457 .3092 L
.88476 .3086 L
.88495 .31416 L
.88514 .31356 L
.88533 .31296 L
.88552 .31236 L
.88571 .31176 L
.8859 .31116 L
.8861 .31056 L
.88629 .31611 L
.88648 .31551 L
.88667 .31491 L
.88686 .31431 L
.88705 .31371 L
.88724 .31311 L
.88743 .31251 L
.88762 .31191 L
.88781 .31131 L
.888 .31072 L
.88819 .31012 L
.88838 .30952 L
.88857 .30892 L
.88876 .30832 L
.88895 .30773 L
.88914 .30713 L
.88933 .30653 L
.88952 .30594 L
.88971 .30534 L
.8899 .31087 L
.8901 .31028 L
.89029 .30968 L
Mistroke
.89048 .30908 L
.89067 .30849 L
.89086 .30789 L
.89105 .3073 L
.89124 .31282 L
.89143 .31222 L
.89162 .31163 L
.89181 .31103 L
.892 .31044 L
.89219 .30984 L
.89238 .30924 L
.89257 .31476 L
.89276 .31416 L
.89295 .31357 L
.89314 .31297 L
.89333 .31238 L
.89352 .31178 L
.89371 .31119 L
.8939 .3167 L
.8941 .3161 L
.89429 .3155 L
.89448 .31491 L
.89467 .31431 L
.89486 .31372 L
.89505 .31313 L
.89524 .31253 L
.89543 .31194 L
.89562 .31134 L
.89581 .31075 L
.896 .31016 L
.89619 .30957 L
.89638 .30897 L
.89657 .30838 L
.89676 .30779 L
.89695 .3072 L
.89714 .30661 L
.89733 .30601 L
.89752 .3115 L
.89771 .31091 L
.8979 .31639 L
.8981 .3158 L
.89829 .3152 L
.89848 .31461 L
.89867 .31402 L
.89886 .31343 L
.89905 .31284 L
.89924 .31225 L
.89943 .31165 L
.89962 .31106 L
.89981 .31047 L
Mistroke
.9 .31594 L
.90019 .31535 L
.90038 .32082 L
.90057 .32022 L
.90076 .31963 L
.90095 .31904 L
.90114 .31845 L
.90133 .31786 L
.90152 .31727 L
.90171 .31668 L
.9019 .31609 L
.9021 .32154 L
.90229 .32095 L
.90248 .3264 L
.90267 .32581 L
.90286 .32522 L
.90305 .32463 L
.90324 .32404 L
.90343 .32948 L
.90362 .32889 L
.90381 .3283 L
.904 .3277 L
.90419 .32711 L
.90438 .32652 L
.90457 .33196 L
.90476 .33137 L
.90495 .33078 L
.90514 .33019 L
.90533 .32959 L
.90552 .329 L
.90571 .32841 L
.9059 .32782 L
.9061 .32723 L
.90629 .32664 L
.90648 .32605 L
.90667 .32547 L
.90686 .32488 L
.90705 .32429 L
.90724 .3237 L
.90743 .32311 L
.90762 .32252 L
.90781 .32794 L
.908 .32735 L
.90819 .32676 L
.90838 .32618 L
.90857 .32559 L
.90876 .325 L
.90895 .33041 L
.90914 .32982 L
.90933 .32923 L
Mistroke
.90952 .33464 L
.90971 .33405 L
.9099 .33945 L
.9101 .33886 L
.91029 .34426 L
.91048 .34367 L
.91067 .34308 L
.91086 .34249 L
.91105 .3419 L
.91124 .34131 L
.91143 .34072 L
.91162 .34014 L
.91181 .33955 L
.912 .33896 L
.91219 .33837 L
.91238 .33778 L
.91257 .33719 L
.91276 .33661 L
.91295 .33602 L
.91314 .33543 L
.91333 .33485 L
.91352 .33426 L
.91371 .33964 L
.9139 .33905 L
.9141 .33846 L
.91429 .33788 L
.91448 .34325 L
.91467 .34862 L
.91486 .34804 L
.91505 .34745 L
.91524 .34686 L
.91543 .34627 L
.91562 .34568 L
.91581 .3451 L
.916 .34451 L
.91619 .34987 L
.91638 .34928 L
.91657 .3487 L
.91676 .34811 L
.91695 .34752 L
.91714 .34694 L
.91733 .34635 L
.91752 .34576 L
.91771 .34518 L
.9179 .35053 L
.9181 .34994 L
.91829 .34935 L
.91848 .34877 L
.91867 .34818 L
.91886 .3476 L
Mistroke
.91905 .34701 L
.91924 .34643 L
.91943 .34584 L
.91962 .34526 L
.91981 .34467 L
.92 .34409 L
.92019 .3435 L
.92038 .34292 L
.92057 .34234 L
.92076 .34175 L
.92095 .34117 L
.92114 .34059 L
.92133 .34 L
.92152 .33942 L
.92171 .33884 L
.9219 .33826 L
.9221 .33768 L
.92229 .33709 L
.92248 .33651 L
.92267 .34184 L
.92286 .34126 L
.92305 .34068 L
.92324 .34009 L
.92343 .33951 L
.92362 .33893 L
.92381 .33835 L
.924 .33777 L
.92419 .33719 L
.92438 .33661 L
.92457 .33603 L
.92476 .33545 L
.92495 .33487 L
.92514 .34018 L
.92533 .3396 L
.92552 .33902 L
.92571 .33844 L
.9259 .33787 L
.9261 .33729 L
.92629 .33671 L
.92648 .33613 L
.92667 .33555 L
.92686 .33497 L
.92705 .3344 L
.92724 .33382 L
.92743 .33324 L
.92762 .33266 L
.92781 .33209 L
.928 .33151 L
.92819 .33093 L
.92838 .33036 L
Mistroke
.92857 .32978 L
.92876 .32921 L
.92895 .32863 L
.92914 .32806 L
.92933 .33334 L
.92952 .33863 L
.92971 .33805 L
.9299 .33748 L
.9301 .3369 L
.93029 .33632 L
.93048 .33575 L
.93067 .33517 L
.93086 .3346 L
.93105 .33402 L
.93124 .33345 L
.93143 .33287 L
.93162 .3323 L
.93181 .33757 L
.932 .337 L
.93219 .33642 L
.93238 .33585 L
.93257 .33527 L
.93276 .3347 L
.93295 .33412 L
.93314 .33355 L
.93333 .33881 L
.93352 .33824 L
.93371 .3435 L
.9339 .34293 L
.9341 .34235 L
.93429 .34178 L
.93448 .3412 L
.93467 .34063 L
.93486 .34005 L
.93505 .33948 L
.93524 .33891 L
.93543 .33833 L
.93562 .34358 L
.93581 .34301 L
.936 .34243 L
.93619 .34186 L
.93638 .34711 L
.93657 .34653 L
.93676 .34596 L
.93695 .34538 L
.93714 .34481 L
.93733 .35005 L
.93752 .34947 L
.93771 .3489 L
.9379 .34833 L
Mistroke
.9381 .34775 L
.93829 .34718 L
.93848 .34661 L
.93867 .34603 L
.93886 .34546 L
.93905 .34489 L
.93924 .35012 L
.93943 .34954 L
.93962 .34897 L
.93981 .35419 L
.94 .35362 L
.94019 .35305 L
.94038 .35827 L
.94057 .35769 L
.94076 .35712 L
.94095 .35654 L
.94114 .35597 L
.94133 .3554 L
.94152 .35482 L
.94171 .35425 L
.9419 .35368 L
.9421 .35889 L
.94229 .35832 L
.94248 .35774 L
.94267 .35717 L
.94286 .3566 L
.94305 .35603 L
.94324 .35545 L
.94343 .35488 L
.94362 .35431 L
.94381 .35951 L
.944 .35894 L
.94419 .35837 L
.94438 .36356 L
.94457 .36299 L
.94476 .36242 L
.94495 .36184 L
.94514 .36127 L
.94533 .3607 L
.94552 .36013 L
.94571 .35956 L
.9459 .35899 L
.9461 .35842 L
.94629 .35785 L
.94648 .35728 L
.94667 .35671 L
.94686 .35614 L
.94705 .35557 L
.94724 .36075 L
.94743 .36018 L
Mistroke
.94762 .35961 L
.94781 .35904 L
.948 .35847 L
.94819 .3579 L
.94838 .36307 L
.94857 .3625 L
.94876 .36193 L
.94895 .36136 L
.94914 .36079 L
.94933 .36596 L
.94952 .36539 L
.94971 .36482 L
.9499 .36425 L
.9501 .36941 L
.95029 .36884 L
.95048 .36827 L
.95067 .3677 L
.95086 .37286 L
.95105 .37229 L
.95124 .37172 L
.95143 .37115 L
.95162 .37058 L
.95181 .37001 L
.952 .36944 L
.95219 .36887 L
.95238 .3683 L
.95257 .37345 L
.95276 .37288 L
.95295 .37231 L
.95314 .37745 L
.95333 .37688 L
.95352 .37631 L
.95371 .37574 L
.9539 .38088 L
.9541 .38031 L
.95429 .37974 L
.95448 .38487 L
.95467 .3843 L
.95486 .38373 L
.95505 .38316 L
.95524 .38259 L
.95543 .38202 L
.95562 .38145 L
.95581 .38088 L
.956 .38031 L
.95619 .37975 L
.95638 .37918 L
.95657 .37861 L
.95676 .37804 L
.95695 .37747 L
Mistroke
.95714 .3769 L
.95733 .37634 L
.95752 .37577 L
.95771 .3752 L
.9579 .37464 L
.9581 .37407 L
.95829 .3735 L
.95848 .37294 L
.95867 .37237 L
.95886 .3718 L
.95905 .37124 L
.95924 .37067 L
.95943 .37011 L
.95962 .37522 L
.95981 .37465 L
.96 .37409 L
.96019 .37352 L
.96038 .37296 L
.96057 .37239 L
.96076 .37183 L
.96095 .37126 L
.96114 .3707 L
.96133 .37013 L
.96152 .36957 L
.96171 .36901 L
.9619 .3741 L
.9621 .37354 L
.96229 .37298 L
.96248 .37241 L
.96267 .37185 L
.96286 .37129 L
.96305 .37072 L
.96324 .37016 L
.96343 .3696 L
.96362 .36903 L
.96381 .37412 L
.964 .37356 L
.96419 .373 L
.96438 .37243 L
.96457 .37187 L
.96476 .37131 L
.96495 .37075 L
.96514 .37018 L
.96533 .36962 L
.96552 .36906 L
.96571 .3685 L
.9659 .36794 L
.9661 .36738 L
.96629 .36682 L
.96648 .36626 L
Mistroke
.96667 .3657 L
.96686 .36514 L
.96705 .36458 L
.96724 .36402 L
.96743 .36346 L
.96762 .36853 L
.96781 .37359 L
.968 .37303 L
.96819 .37247 L
.96838 .37753 L
.96857 .37697 L
.96876 .37641 L
.96895 .37585 L
.96914 .37529 L
.96933 .37473 L
.96952 .37417 L
.96971 .37361 L
.9699 .37866 L
.9701 .3781 L
.97029 .37754 L
.97048 .37698 L
.97067 .38203 L
.97086 .38147 L
.97105 .38091 L
.97124 .38035 L
.97143 .37979 L
.97162 .38483 L
.97181 .38427 L
.972 .38371 L
.97219 .38315 L
.97238 .38819 L
.97257 .38763 L
.97276 .38707 L
.97295 .38651 L
.97314 .38595 L
.97333 .38539 L
.97352 .38483 L
.97371 .38427 L
.9739 .38371 L
.9741 .38315 L
.97429 .38818 L
.97448 .38762 L
.97467 .38706 L
.97486 .3865 L
.97505 .38594 L
.97524 .38538 L
.97543 .38482 L
.97562 .38427 L
.97581 .38371 L
.976 .38873 L
Mistroke
.97619 .38817 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`000?ojl2@SS_oG
001[0?ooAm?Cdoojn[34g_XC@4J2]<c<c;Ng]fIVI_Cdm2m?CfUYJKjn_/30`;C^]:JVYZf][Jo^k/c8
`8b:Q;no_eICDe5Anb:;8^KViW=cL`00oiT0I_l003_j=020P80008200?oo0820P6IVc3<cIP2009VI
od:JYb8R8_mV<`00P03oo`3o000g??mVc6=SHd92@]kNgYBDU;fm_IZDU>GPhFaXJ;8R8P00Rld009VI
V@2;000i0?o<oom0o`1^0000?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o
?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo
_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo
_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0
_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOknookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo
?oooOooo_mWEc:0Pl=WIfJj^[WMgMiW<09W<<iW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc
0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI0?nI<ol08@000?l08@000?l08@000?l08@000?l0
8@000?l08@000?l08@000?l08@000?l08@000?l08@0002400`4?00@100<00@40iP0002400`4?0005
0@010@000P4500030@000=h0000Q00<13`001@400@4000811@000`40003N00008@030@l000D10041
00020@D000<10000gP0002400`4?00050@010@000P4500030@000=h0000Q00<13`001@400@400081
1@000`40003N00008@030@l000D1004100020@@00P7P00008@030@l000D1004100020@@00P7P0000
8@030@03004000/00P400`010@030@<00`7P00008@030@03004000/00P400`010@030@<00`7P0000
8@030@03004000/00P400`010@030@<0107O00008@030@03004000/00P400`010@030@<0107O0000
8@030@03004000/00P400`010@030@<0107O00008@050@`00P400`010@030@<00P400`01003N0000
8@050@`00P400`010@030@<00P400`01003N00008@050@`00P400`010@030@<00P400`01003N0000
8@050@`00P400`010@030@<00P400`01003N00008@050@`00P400`010@030@<00P400`01003N0000
8@050@`00P400`010@030@<00P400`01003N00008@050@`00P4200@10`020@0300400=h0000Q00D1
30020@<000<100400`020@0300400=h0000Q00D12`030@<000<100400`020@03004000L000<10000
e00002401@4:00@10`000`400@03008100<00@001`000`40003D00008@050@X0104300030@0100<0
0P400`01000700030@000=@0000Q00D12P030@H000<100000`400`01000700030@000=@000080081
10000`40000300810`030@D01@4:00<11P000`4000030@0400400@D00P7F00001`00104000480004
0@000@D000<100000P050@X00`4600030@0000<100@00@411@020MH0000700040@000@P000@10001
1@000`40000200D12P030@H000<100000`4010010@450081eP0000L000@1000120001040004300<1
10050@X00`4600030@0100810P030@D00P7F00001`001040004800040@000@8000@1000110050@X0
0`4600040@010@<00`450081eP0000L000@1000120001040004200040@000@@01@4900<11`001040
0@4300<11@020MH0000800812P020@@00P4500D12@030@L000@100410`030@8000<100000P7F0000
8@050@T00`4700040@010@<00`4200030@000081eP0002401@4900<11`0010400@4300<10P000`40
00020MH0000Q00D120040@L0104300<10P000`4000030MD0000Q00D120040@L0104300<10P000`40
00030MD0000Q00D120040@L0104300<10P000`4000030MD0000Q00D120040@L0104300<10P000`40
00030MD0000Q00D120040@P00`4300<10P000`4000030MD0000Q00D120040@P00`44008100D00@41
00030MD0000Q00D120040@P00`44008100<00@401P7D00008@050@P00`4900<11@000`400@070M@0
000Q00D11`040@T00`4500030@0100<100@00@4120000`40003900008@050@L0104900<11@000`40
0@030@800P470081b`0002401@4700@12@030@D000<100400`4200811`020L/0000Q00D11`040@T0
0P4600030@0100<10P020@L00P7;00008@050@L0104900811P000`400@030@800P470081b`000240
1@4700@12@020@H000<100400`4200811`020L/0000Q00D11`040@T00P4600H10P020@L00P7;0000
8@050@L00`4:00811P040@04004000811`020L/0000Q00D11`030@X00P4600@11@000`4000020003
0@000081b`0002401@4700<12P020@H00`4600030@00008000<100000`7:00008@050@L00`4:0081
1P030@H000<100000P000`4000030LX0000Q00D11`030@X000<100001@030@H000<100000P000`40
0@030LX0000Q00D11`030@X000<100001@030@H000@100000P400`010@020LX0000Q00D11`030@X0
00<100001@030@H000@10000204800030@000;h0000Q00D11`030@X000<100001@020@L000@10000
104010010@480081_`0002401@4700<12P000`40000500811`00104000040@0400410@P00P6o0000
8@050@L00`4:00030@0000D00P4700030@0000D100<00@400P470081NP000`400@0800<1=`000240
1@4700<12P000`40000500811`000`4000050@800`4700810P000`40001e00<11`050CH000020081
10000`40000300810`030@<00`4500D11`030@X000<100001@000`40000600030@0000@10`030@L0
0P400`010@1g00@11P070C@000001@01000100P000@100011@000`40000300030@0000801@4700<1
2P000`40000500030@0000H000<10000104300@11P020@03004107@01@400`010@0500030@0000@1
=00000050040004020001040004500030@0000<000<100000P050@L00`4:00030@0000D000<10000
1P000`4000040@@00`46008100D00@400@1`00D100<00@000P4400810`030C@000001@01000100P0
00@100010`030@800`4500D11`030@X000<100001@000`40000600030@0000@110030@H000<10040
0P400`010@1Z00040@0000H110001P400@400@@00`4d000000D00@000@0800040@000@8000@10001
0P000`40000500D11`030A8000<100001P000`4000020@03004000<00`4600030@01008100<00@40
JP030@03004100811P060@H000<10000<P0000050040004020001040004200040@000@8000<10000
1@050@L00`4B00030@0000H000<100000P400`010004008110020@801P5R00030@0000D01@400`01
0@0600D11`000`40000b00000P020@X00P4400810`040@@01@4700<16`001040004900030@000080
0P4200H1A@000`40000J00050@0000400`050@T01@4700030@000380000Q00D11`030A/000@10001
2@000`40000200810P060DD000<100006P00104000050@03004100/00P490081<`0002401@4700<1
6`001040004900030@0000800P4200H1A@000`40000H00@100<00@400`4J0081<`0002401@4700<1
6`000`400@0:00030@0000800P4200H1A@020AT00P400`0100040A`00P4b00008@050@L00`4K00<1
2P00104000030@@00P400`01001300<160030@03004100@170020C80000Q00D11`030A/00`4:0004
0@0000<110020@03004004<000<1004060030@0600410@017P001`4000010040;00002401@4700<1
6`030@X00`4010010@44008100@00@01@P0010400@4G00040@0000818@030@0400400B`0000Q00D1
1`030A/00`4:00H11@020@0400400Cl00P400`0100020AL000<100000P000`40000O00<100@00@41
0P000`40000W00008@050@L00`4K00<12P060@D00P4010010@4o00@10P020AD00P4W00H10P000`40
000W00008@050@L00`4K00<12P060@P00`4a00040@000@P01@4300815@000`40000W008100@00@40
0`4X00008@050@H00`4L00<12P060@P00`4a00040@000@P01@4400030@0001800P4Z00030@0000D1
:00002401@4600<170030@X01@4900<1;`030@03004100D01P400`01000300030@0001400P4_0004
0@000@<0104Q00008@050@H00`4M00812P050@X00P4^00@100<00@401@050@P00P4200812@030@03
004102l000<100002@4P00008@050@H00P4N00812P001@400@0100X00P4^00L11@050@P00P401001
0@4900H1<P090B00000Q00D11P020Ah00P4<00030@0100X00`4]00<100@00@411@020@/01P4900H1
<`030@03004000<17`0002401@4600817P020@`000<100002P000`400@0/00@100H00@4100430003
0@0000X010400`01000600H1=@020@D00`4N00008@050@H00P4N00816`001040004Y00@100<00@40
0`42008130040@03004000H00P420081?@020Ah0000Q00D11P020Ah00P4K00030@000081:0040@03
004000<10P000`40000;00@100D00@000@0300<1@@030@<00P4H00008@050@H000<100007@000`40
000J00030@010081:0000`400@0400@100<00@0030040@03004000810`030D4000@100410P030AL0
000Q00D11P000`40000j00030@0100819P000`400@0600@100<00@004P030@801053008100<00@40
0P4G00008@050@H000<10000>P000`400@020BD0104700030@0100814`030@800P5500L15`000240
1@4600030@0003X000D100400@0U00<12P030A<00`4200030@0004@00`4300815P0002401@460003
0@0001L01043008110020@@00P4>00050@01004020040@<00P44008110020@P00`4:00<11@020@@0
0P4400<100<00@000P4M00030@0000800P44008110020AL00P4300<110020@@00P4400810`000240
1@4600030@0001T000D100000@0200040@000@8000@100010P000`40000;00050@01004020000`40
000300040@000@8000@100010P001040004600@12`000`40000300040@000@8000@100010P001@40
004000<100<00@006`001@400001008000@100010P001040004200030@0001D000<100000P030@80
00@100010P001040004200040@000@80000Q00D11P000`40000I00050@0000400P00104000420004
0@000@8000<100002`0010400@020@P000<100000P001040004200040@000@8000@100011@020@03
004101@000@100010P001040004200030@01008100<00@0060050@8000@100010P00104000420004
0@000A`00`4200040@000@8000@100010P001040004200008@050@H000<100006@001@4000010080
00@100010P001040004200030@0000/000@100400P4900050@0000400P001040004200040@000@80
00<100000`020@03004001800P4300040@000@8000D1000100030@03004001P000@100010`001040
004200040@000@8000@100016`050@05004000400P001040004200040@000@80000Q00D11P000`40
000I00050@0000400P001040004200040@000@8000<100002`0010400@020@X000@100010P001040
004200040@000@8000<100000`020@03004001@000@100010P001040004200040@000@8000<10000
60000`400@0300040@000@8000@100010P001040004K00030@01008100<00@000P4200040@000@80
00@100010P0002401@4600030@0001L00`4300040@000@8000@100010P001040004=00<100<00@40
1`001040004200040@000@800P401@010001008000<100000`000`40000C00040@000@8000@10001
0P001040004200040@000A/00P4300040@000@8000@100010P001040004K00060@00004010420004
0@000@8000@100010P0002401@4600030@0001T000<100000P020@@00P4400813P020@800`470081
10020@@00P4400811P000`40000D008110020@@00P4400817@000`400002008110020@@00P4L00X1
0`020@@00P4300008@050@H000<10000>P020@<00P4C00030@0000/000<10000N@030@03004000d0
000Q00D11P000`40000j00810`020@03004000l00P4<0081N`020@8000<10000300001/0o`460000
8@050@@000<100401`000`40000700030@0000H000<100001`000`40000700030@0000H000<10000
0`00104000020@03004000D000<100000`43008100<00@401@000`40000200030@00008000<10000
1P000`40000700030@0000L000<100001P000`40000700030@0000L000<100001P000`4000070003
0@0000L000<100001P000`40000700030@0000L000<100001P020@`0000Q00D11P000`40000k0003
0@0000800`4500040@0000<1100010400@4500040@000081S`0002401@4600030@00040010440004
0@0000<1100010400@4500810P020Hl0000Q00D11P000`40001000@110020@800`400`0100020@80
0P4400810P020Hl0000Q00D11P000`40001100<10`030@0600410040104200810`030@800P6?0000
8@050@H000<10000@@030@<00`401P010@0100@10P020@<00`4200030@0008h0000Q00D11P000`40
001100030@0100<01@4200D10`020@<00`4200030@0008h0000Q00D11P000`40001100030@0100<0
00<100400P4200D10`020@<000<100400P000`40002>00008@050@H000<10000@@001@400@400081
00@00@410P020@03004000<00P42008100<00@000P6@00008@050@H000<10000@`050@03004100<0
0P4700030@01008100@00@01T@0002401@4600030@0004<0104200810`020@L000@100410P000`40
0@2A00008@050@H000<10000@`040@<000@100000P4700040@010@800`6A00008@050@H000<10000
@`040@<000<100002P040@800`6A00008@050@H000<10000A0020A4010430081T@0002401@460003
0@0004@00P4A00<110020I40000Q00D11P000`40001400030@00014000<100000`000`40002@0000
8@050@H000<10000F0000`40000300030@000900000Q00D11P000`40001H00030@0009H0000Q00D1
1P000`40003a00008@050@H000<10000l@0002401@4500030@000?80000Q00D11@000`40003b0000
8@050@D000<10000lP0002401@4500030@000?80000Q00D11@000`40003b00008@050@D000<10000
lP0002401@4500030@000?80000Q00D11@000`40003b00008@050@D000<10000lP0002401@450003
0@000?80000Q00D100<00@000P000`40003b00008@050@030040008000<10000lP0002401`430003
0@000?800002008110000`40000200@10`020@<00`4500L10`000`40003b000000D00@000@0:0005
0@0000400P000`40000300030@0000801`4300030@000?8000001@01000100X000D100000@020003
0@0000<000<100000P070@<000<10000lP000005004000402P001@400001008000<100000`4500L1
0`000`40003b000000D00@000@0:00050@0000400P001040004700L10`000`40003b000000D00@00
0@0800<10`001040004200030@0000D01`4300030@000?80000200812`000`40000200810`040@@0
1`4300030@000?80000Q00L10`000`40003b00008@070@<000<10000lP0002401`4300030@000?80
000Q00L10`000`40003b00008@070@<000<10000lP0002401`4300030@000?80000Q00L10`000`40
003b00008@070@<000<10000lP0002401`4300030@000?80000Q00L10`000`40003b00008@070@<0
00<10000lP0002401`4300030@000?80000Q00L10`000`40003b00008@040@03004100<000<10000
lP00024010400`010@0300030@000?80000Q00@100<00@400`000`40003b00008@040@03004100<0
00<10000lP00024010400`010@020081m000024010400`010@020081m0000?l08@000?l08@000?l0
8@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000001\
\>"],
  ImageRangeCache->{{{108, 395}, {450.312, 273.375}} -> {-2918.97, 0.123487, 
  20.6762, 0.0001413}}]
}, Open  ]],

Cell[TextData[{
  "Compute the exact probability that you have a cold given that you've \
sneezed and observed cat scratches on the furniture: ",
  StyleBox["Pr[2,{1,5}]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell["\<\
Divide[Marginalize[{{2,True},{1,True},{5,True}},network2],
       Marginalize[{{1,True},{5,True}},network2]]\
\>", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[BoxData[
    \(0.253858977769233096`\)], "Output"]
}, Open  ]],

Cell[TextData[{
  "Estimate the probability that you have a cold given that you've sneezed \
and observed cat scratches on the furniture: ",
  StyleBox["Pr[2,{1,5}]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(GraphLikelihoodWeighting[{{1, True}, {5, True}}, network2, 5000, 2]\)], 
  "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.000190476 -1.32646 6.27921 [
[.21429 .5448 -12 -9 ]
[.21429 .5448 12 0 ]
[.40476 .5448 -12 -9 ]
[.40476 .5448 12 0 ]
[.59524 .5448 -12 -9 ]
[.59524 .5448 12 0 ]
[.78571 .5448 -12 -9 ]
[.78571 .5448 12 0 ]
[.97619 .5448 -12 -9 ]
[.97619 .5448 12 0 ]
[.01131 .05496 -24 -4.5 ]
[.01131 .05496 0 4.5 ]
[.01131 .18055 -24 -4.5 ]
[.01131 .18055 0 4.5 ]
[.01131 .30613 -24 -4.5 ]
[.01131 .30613 0 4.5 ]
[.01131 .43172 -24 -4.5 ]
[.01131 .43172 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .5573 m
.21429 .56355 L
s
[(1000)] .21429 .5448 0 1 Mshowa
.40476 .5573 m
.40476 .56355 L
s
[(2000)] .40476 .5448 0 1 Mshowa
.59524 .5573 m
.59524 .56355 L
s
[(3000)] .59524 .5448 0 1 Mshowa
.78571 .5573 m
.78571 .56355 L
s
[(4000)] .78571 .5448 0 1 Mshowa
.97619 .5573 m
.97619 .56355 L
s
[(5000)] .97619 .5448 0 1 Mshowa
.125 Mabswid
.0619 .5573 m
.0619 .56105 L
s
.1 .5573 m
.1 .56105 L
s
.1381 .5573 m
.1381 .56105 L
s
.17619 .5573 m
.17619 .56105 L
s
.25238 .5573 m
.25238 .56105 L
s
.29048 .5573 m
.29048 .56105 L
s
.32857 .5573 m
.32857 .56105 L
s
.36667 .5573 m
.36667 .56105 L
s
.44286 .5573 m
.44286 .56105 L
s
.48095 .5573 m
.48095 .56105 L
s
.51905 .5573 m
.51905 .56105 L
s
.55714 .5573 m
.55714 .56105 L
s
.63333 .5573 m
.63333 .56105 L
s
.67143 .5573 m
.67143 .56105 L
s
.70952 .5573 m
.70952 .56105 L
s
.74762 .5573 m
.74762 .56105 L
s
.82381 .5573 m
.82381 .56105 L
s
.8619 .5573 m
.8619 .56105 L
s
.9 .5573 m
.9 .56105 L
s
.9381 .5573 m
.9381 .56105 L
s
.25 Mabswid
0 .5573 m
1 .5573 L
s
.02381 .05496 m
.03006 .05496 L
s
[(0.22)] .01131 .05496 1 0 Mshowa
.02381 .18055 m
.03006 .18055 L
s
[(0.24)] .01131 .18055 1 0 Mshowa
.02381 .30613 m
.03006 .30613 L
s
[(0.26)] .01131 .30613 1 0 Mshowa
.02381 .43172 m
.03006 .43172 L
s
[(0.28)] .01131 .43172 1 0 Mshowa
.125 Mabswid
.02381 .08636 m
.02756 .08636 L
s
.02381 .11776 m
.02756 .11776 L
s
.02381 .14915 m
.02756 .14915 L
s
.02381 .21194 m
.02756 .21194 L
s
.02381 .24334 m
.02756 .24334 L
s
.02381 .27474 m
.02756 .27474 L
s
.02381 .33753 m
.02756 .33753 L
s
.02381 .36893 m
.02756 .36893 L
s
.02381 .40032 m
.02756 .40032 L
s
.02381 .46311 m
.02756 .46311 L
s
.02381 .49451 m
.02756 .49451 L
s
.02381 .52591 m
.02756 .52591 L
s
.02381 .02357 m
.02756 .02357 L
s
.02381 .5887 m
.02756 .5887 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.03057 0 m
.03067 .20974 L
.03086 .20817 L
.03105 .20661 L
.03124 .20505 L
.03143 .2035 L
.03162 .20195 L
.03181 .2004 L
.032 .19885 L
.03219 .19731 L
.03238 .19577 L
.03257 .19424 L
.03276 .09391 L
.03295 .09257 L
.03314 .09123 L
.03333 .0899 L
.03352 .08857 L
.03371 .08724 L
.0339 .08592 L
.0341 .0846 L
.03429 .08328 L
.03448 .08196 L
.03467 .4208 L
.03486 .41928 L
.03505 .41776 L
.03524 .41625 L
.03543 .41474 L
.03562 .41323 L
.03581 .41173 L
.036 .41023 L
.03619 .40873 L
.03638 .40723 L
.03657 .40573 L
.03676 .40424 L
.03695 .40275 L
.03714 .40127 L
.03733 .39978 L
.03752 .3983 L
.03771 .39682 L
.0379 .39534 L
.0381 .39387 L
.03829 .3924 L
.03848 .39093 L
.03867 .38946 L
.03886 .388 L
s
.03886 .388 m
.03901 .61803 L
s
.03953 .61803 m
.03962 .57058 L
.03981 .56915 L
.04 .56772 L
.04019 .56629 L
.04038 .56487 L
.04057 .56344 L
.04076 .56202 L
.04095 .5606 L
.04114 .55919 L
.04133 .55777 L
.04152 .55636 L
.04171 .55495 L
.0419 .55354 L
.0421 .55214 L
s
.0421 .55214 m
.04215 .61803 L
s
.09132 .61803 m
.09143 .49442 L
.09162 .49414 L
.09181 .49387 L
.092 .4936 L
.09219 .49333 L
.09238 .49306 L
.09257 .49278 L
.09276 .49251 L
.09295 .49224 L
.09314 .49197 L
.09333 .4917 L
.09352 .49143 L
.09371 .49116 L
.0939 .47237 L
.0941 .4721 L
.09429 .47184 L
.09448 .47157 L
.09467 .47131 L
.09486 .47104 L
.09505 .47077 L
.09524 .47051 L
.09543 .47024 L
.09562 .46998 L
.09581 .46971 L
.096 .46945 L
.09619 .46919 L
.09638 .46892 L
.09657 .46866 L
.09676 .46839 L
.09695 .46813 L
.09714 .46786 L
.09733 .4676 L
.09752 .46733 L
.09771 .46707 L
.0979 .46681 L
.0981 .46654 L
.09829 .51875 L
.09848 .50015 L
.09867 .49989 L
.09886 .49963 L
.09905 .49936 L
.09924 .4991 L
.09943 .49884 L
.09962 .49858 L
.09981 .49831 L
.1 .49805 L
.10019 .49779 L
.10038 .49753 L
.10057 .49726 L
Mistroke
.10076 .497 L
.10095 .49674 L
.10114 .49648 L
.10133 .49621 L
.10152 .49595 L
.10171 .49569 L
.1019 .49543 L
.1021 .49517 L
.10229 .49491 L
.10248 .47678 L
.10267 .47652 L
.10286 .47627 L
.10305 .52659 L
.10324 .5758 L
.10343 .57554 L
.10362 .57527 L
.10381 .57501 L
.104 .57474 L
.10419 .57448 L
.10438 .57422 L
.10457 .55592 L
.10476 .55566 L
.10495 .5554 L
.10514 .55514 L
.10533 .55489 L
.10552 .55463 L
.10571 .55437 L
.1059 .55411 L
.1061 .55385 L
.10629 .55359 L
.10648 .55334 L
.10667 .55308 L
.10686 .55282 L
.10705 .53493 L
.10724 .53468 L
.10743 .53442 L
.10762 .53417 L
.10781 .53392 L
.108 .53367 L
.10819 .53341 L
.10838 .53316 L
.10857 .53291 L
.10876 .53266 L
.10895 .5324 L
.10914 .53215 L
.10933 .5319 L
.10952 .5794 L
.10971 .57914 L
.1099 .57888 L
.1101 .57863 L
Mistroke
.11029 .57837 L
.11048 .57812 L
.11067 .57786 L
.11086 .57761 L
.11105 .57735 L
.11124 .55964 L
.11143 .55939 L
.11162 .55914 L
.11181 .55889 L
.112 .55864 L
.11219 .55839 L
.11238 .55814 L
.11257 .55789 L
.11276 .55764 L
.11295 .55739 L
.11314 .55714 L
.11333 .55689 L
.11352 .55664 L
.11371 .55639 L
.1139 .55614 L
.1141 .60231 L
.11429 .60205 L
.11448 .6018 L
.11467 .60155 L
.11486 .60129 L
.11505 .60104 L
.11524 .60079 L
.11543 .60054 L
.11562 .60028 L
.11581 .60003 L
.116 .59978 L
.11619 .59953 L
.11638 .59927 L
.11657 .59902 L
.11676 .59877 L
.11695 .59852 L
.11714 .59826 L
.11733 .59801 L
.11752 .59776 L
.11771 .59751 L
.1179 .59726 L
.1181 .597 L
.11829 .59675 L
.11848 .57929 L
.11867 .57904 L
.11886 .5788 L
.11905 .57855 L
.11924 .5783 L
.11943 .57806 L
.11962 .57781 L
Mistroke
.11981 .57756 L
.12 .57732 L
.12019 .57707 L
.12038 .57682 L
.12057 .57658 L
.12076 .57633 L
.12095 .57608 L
.12114 .57584 L
.12133 .57559 L
.12152 .57534 L
.12171 .5751 L
.1219 .57485 L
.1221 .57461 L
.12229 .57436 L
Mfstroke
.12229 .57436 m
.12247 .61803 L
s
.12268 .61803 m
.12286 .60168 L
.12305 .58471 L
.12324 .58447 L
.12343 .58423 L
.12362 .58398 L
.12381 .58374 L
.124 .5835 L
.12419 .58326 L
.12438 .58302 L
.12457 .58278 L
.12476 .56614 L
.12495 .5659 L
.12514 .56567 L
.12533 .56543 L
.12552 .5652 L
.12571 .56496 L
.1259 .56473 L
.1261 .56449 L
.12629 .56426 L
.12648 .56402 L
.12667 .56378 L
.12686 .56355 L
.12705 .56331 L
.12724 .56308 L
.12743 .56284 L
.12762 .56261 L
.12781 .54631 L
.128 .54608 L
.12819 .54585 L
.12838 .54562 L
.12857 .54539 L
.12876 .54516 L
.12895 .54493 L
.12914 .5447 L
.12933 .54447 L
.12952 .54424 L
.12971 .54401 L
.1299 .54378 L
.1301 .54355 L
.13029 .54332 L
.13048 .54309 L
.13067 .54286 L
.13086 .54263 L
.13105 .5424 L
.13124 .54217 L
.13143 .54194 L
.13162 .54171 L
.13181 .54148 L
.132 .54125 L
Mistroke
.13219 .54102 L
.13238 .58397 L
.13257 .58374 L
.13276 .58351 L
.13295 .56738 L
.13314 .56715 L
.13333 .55129 L
.13352 .55107 L
.13371 .55084 L
.1339 .55062 L
.1341 .55039 L
.13429 .55017 L
.13448 .54995 L
.13467 .54972 L
.13486 .5495 L
.13505 .54927 L
.13524 .59097 L
.13543 .59074 L
.13562 .59051 L
.13581 .57477 L
.136 .57454 L
.13619 .57432 L
.13638 .5741 L
.13657 .55862 L
.13676 .5584 L
.13695 .55818 L
.13714 .55796 L
.13733 .55774 L
.13752 .55752 L
.13771 .5573 L
.1379 .55708 L
.1381 .55686 L
.13829 .55664 L
.13848 .55643 L
.13867 .55621 L
.13886 .55599 L
.13905 .55577 L
.13924 .55555 L
.13943 .55533 L
.13962 .55511 L
.13981 .5549 L
.14 .55468 L
.14019 .55446 L
.14038 .55424 L
.14057 .55402 L
.14076 .5538 L
.14095 .59427 L
.14114 .59405 L
.14133 .59383 L
.14152 .59361 L
Mistroke
.14171 .59339 L
.1419 .59317 L
.1421 .59295 L
.14229 .59273 L
.14248 .59251 L
.14267 .59229 L
.14286 .59207 L
.14305 .59185 L
.14324 .59163 L
.14343 .59141 L
.14362 .59118 L
.14381 .5759 L
.144 .57568 L
.14419 .57546 L
.14438 .57525 L
.14457 .57503 L
.14476 .57481 L
.14495 .5746 L
.14514 .57438 L
.14533 .57416 L
.14552 .57395 L
.14571 .57373 L
.1459 .43357 L
.1461 .43338 L
.14629 .4332 L
.14648 .47097 L
.14667 .45792 L
.14686 .45774 L
.14705 .45755 L
.14724 .45737 L
.14743 .45718 L
.14762 .457 L
.14781 .45681 L
.148 .45663 L
.14819 .45644 L
.14838 .45626 L
.14857 .45607 L
.14876 .45589 L
.14895 .45571 L
.14914 .45552 L
.14933 .45534 L
.14952 .45515 L
.14971 .45497 L
.1499 .49192 L
.1501 .49173 L
.15029 .49154 L
.15048 .49136 L
.15067 .49117 L
.15086 .36932 L
.15105 .36916 L
Mistroke
.15124 .369 L
.15143 .36884 L
.15162 .36867 L
.15181 .36851 L
.152 .36835 L
.15219 .36819 L
.15238 .36802 L
.15257 .36786 L
.15276 .3677 L
.15295 .36754 L
.15314 .36738 L
.15333 .36721 L
.15352 .36705 L
.15371 .36689 L
.1539 .36673 L
.1541 .36657 L
.15429 .3664 L
.15448 .36624 L
.15467 .36608 L
.15486 .36592 L
.15505 .36576 L
.15524 .3656 L
.15543 .36543 L
.15562 .36527 L
.15581 .36511 L
.156 .3999 L
.15619 .39974 L
.15638 .39958 L
.15657 .39941 L
.15676 .39925 L
.15695 .39909 L
.15714 .39892 L
.15733 .39876 L
.15752 .3986 L
.15771 .39843 L
.1579 .39827 L
.1581 .39811 L
.15829 .39794 L
.15848 .39778 L
.15867 .39762 L
.15886 .43186 L
.15905 .43169 L
.15924 .42021 L
.15943 .42004 L
.15962 .41988 L
.15981 .41972 L
.16 .41955 L
.16019 .41939 L
.16038 .41923 L
.16057 .41907 L
Mistroke
.16076 .4189 L
.16095 .41874 L
.16114 .41858 L
.16133 .31164 L
.16152 .3115 L
.16171 .31135 L
.1619 .31121 L
.1621 .31107 L
.16229 .31092 L
.16248 .30096 L
.16267 .30081 L
.16286 .30067 L
.16305 .30053 L
.16324 .30039 L
.16343 .30025 L
.16362 .30011 L
.16381 .29997 L
.164 .29982 L
.16419 .29968 L
.16438 .29954 L
.16457 .33167 L
.16476 .33152 L
.16495 .33138 L
.16514 .33124 L
.16533 .33109 L
.16552 .33095 L
.16571 .33081 L
.1659 .33067 L
.1661 .33052 L
.16629 .33038 L
.16648 .33024 L
.16667 .33009 L
.16686 .32016 L
.16705 .32002 L
.16724 .31987 L
.16743 .31973 L
.16762 .31959 L
.16781 .31945 L
.168 .31931 L
.16819 .31917 L
.16838 .31903 L
.16857 .31889 L
.16876 .31875 L
.16895 .35027 L
.16914 .35013 L
.16933 .34999 L
.16952 .34985 L
.16971 .3497 L
.1699 .34956 L
.1701 .38065 L
Mistroke
.17029 .38051 L
.17048 .38036 L
.17067 .38022 L
.17086 .38008 L
.17105 .28488 L
.17124 .28475 L
.17143 .28462 L
.17162 .28449 L
.17181 .28436 L
.172 .31393 L
.17219 .3138 L
.17238 .31367 L
.17257 .31354 L
.17276 .31341 L
.17295 .31328 L
.17314 .31315 L
.17333 .31302 L
.17352 .31289 L
.17371 .31276 L
.1739 .34194 L
.1741 .34181 L
.17429 .37062 L
.17448 .37049 L
.17467 .37036 L
.17486 .36113 L
.17505 .361 L
.17524 .36087 L
.17543 .36074 L
.17562 .36061 L
.17581 .36048 L
.176 .36035 L
.17619 .35123 L
.17638 .3511 L
.17657 .35097 L
.17676 .35084 L
.17695 .37909 L
.17714 .37896 L
.17733 .37882 L
.17752 .37869 L
.17771 .40659 L
.1779 .40646 L
.1781 .40633 L
.17829 .4062 L
.17848 .40606 L
.17867 .40593 L
.17886 .39674 L
.17905 .38764 L
.17924 .38751 L
.17943 .38738 L
.17962 .38726 L
Mistroke
.17981 .38713 L
.18 .387 L
.18019 .38687 L
.18038 .38674 L
.18057 .38661 L
.18076 .38648 L
.18095 .38635 L
.18114 .38622 L
.18133 .38609 L
.18152 .41345 L
.18171 .41332 L
.1819 .40425 L
.1821 .40412 L
.18229 .40399 L
.18248 .39502 L
.18267 .39489 L
.18286 .38601 L
.18305 .38588 L
.18324 .38576 L
.18343 .38563 L
.18362 .3855 L
.18381 .38538 L
.184 .41214 L
.18419 .41201 L
.18438 .41189 L
.18457 .41176 L
.18476 .41163 L
.18495 .40277 L
.18514 .40264 L
.18533 .40252 L
.18552 .40239 L
.18571 .40227 L
.1859 .40214 L
.1861 .40201 L
.18629 .40189 L
.18648 .40176 L
.18667 .40164 L
.18686 .40151 L
.18705 .40138 L
.18724 .40126 L
.18743 .40113 L
.18762 .40101 L
.18781 .40088 L
.188 .40076 L
.18819 .40063 L
.18838 .4005 L
.18857 .40038 L
.18876 .40025 L
.18895 .42659 L
.18914 .42647 L
Mistroke
.18933 .42634 L
.18952 .42621 L
.18971 .42609 L
.1899 .42596 L
.1901 .42583 L
.19029 .42571 L
.19048 .42558 L
.19067 .42545 L
.19086 .42533 L
.19105 .34097 L
.19124 .34085 L
.19143 .34074 L
.19162 .36911 L
.19181 .369 L
.192 .36888 L
.19219 .36877 L
.19238 .36865 L
.19257 .36853 L
.19276 .36842 L
.19295 .3683 L
.19314 .36819 L
.19333 .36807 L
.19352 .36796 L
.19371 .35989 L
.1939 .35978 L
.1941 .35966 L
.19429 .35955 L
.19448 .35943 L
.19467 .35932 L
.19486 .38416 L
.19505 .38404 L
.19524 .38393 L
.19543 .38381 L
.19562 .3837 L
.19581 .38358 L
.196 .30634 L
.19619 .30624 L
.19638 .30613 L
.19657 .30603 L
.19676 .30592 L
.19695 .30582 L
.19714 .30571 L
.19733 .30561 L
.19752 .3055 L
.19771 .3054 L
.1979 .30529 L
.1981 .30519 L
.19829 .30508 L
.19848 .30498 L
.19867 .30487 L
Mistroke
.19886 .30477 L
.19905 .30466 L
.19924 .30455 L
.19943 .30445 L
.19962 .30434 L
.19981 .30424 L
.2 .30413 L
.20019 .30403 L
.20038 .30392 L
.20057 .30382 L
.20076 .30371 L
.20095 .30361 L
.20114 .3035 L
.20133 .3034 L
.20152 .30329 L
.20171 .30319 L
.2019 .30308 L
.2021 .29577 L
.20229 .31953 L
.20248 .31943 L
.20267 .31932 L
.20286 .31922 L
.20305 .31911 L
.20324 .31901 L
.20343 .3189 L
.20362 .31159 L
.20381 .31149 L
.204 .31139 L
.20419 .33483 L
.20438 .33473 L
.20457 .33462 L
.20476 .33452 L
.20495 .33441 L
.20514 .35762 L
.20533 .35751 L
.20552 .35741 L
.20571 .3573 L
.2059 .35719 L
.2061 .35709 L
.20629 .35698 L
.20648 .35688 L
.20667 .35677 L
.20686 .35667 L
.20705 .35656 L
.20724 .35646 L
.20743 .35635 L
.20762 .35624 L
.20781 .35614 L
.208 .35603 L
.20819 .35593 L
Mistroke
.20838 .35582 L
.20857 .35572 L
.20876 .34836 L
.20895 .34826 L
.20914 .34815 L
.20933 .34805 L
.20952 .34794 L
.20971 .34784 L
.2099 .34773 L
.2101 .34763 L
.21029 .34752 L
.21048 .34742 L
.21067 .34014 L
.21086 .34003 L
.21105 .33993 L
.21124 .33983 L
.21143 .33972 L
.21162 .33962 L
.21181 .33952 L
.212 .33941 L
.21219 .33931 L
.21238 .3392 L
.21257 .3391 L
.21276 .339 L
.21295 .33889 L
.21314 .33879 L
.21333 .33158 L
.21352 .33148 L
.21371 .33138 L
.2139 .33127 L
.2141 .33117 L
.21429 .33107 L
.21448 .33097 L
.21467 .33086 L
.21486 .33076 L
.21505 .33066 L
.21524 .33056 L
.21543 .33046 L
.21562 .33035 L
.21581 .33025 L
.216 .33015 L
.21619 .33005 L
.21638 .35278 L
.21657 .35268 L
.21676 .35257 L
.21695 .35247 L
.21714 .35237 L
.21733 .35227 L
.21752 .35216 L
.21771 .35206 L
Mistroke
.2179 .35196 L
.2181 .34477 L
.21829 .34467 L
.21848 .34456 L
.21867 .34446 L
.21886 .34436 L
.21905 .34426 L
.21924 .34415 L
.21943 .34405 L
.21962 .34395 L
.21981 .34385 L
.22 .34375 L
.22019 .36618 L
.22038 .36608 L
.22057 .36598 L
.22076 .29684 L
.22095 .29675 L
.22114 .29014 L
.22133 .28359 L
.22152 .2835 L
.22171 .30502 L
.2219 .30493 L
.2221 .30483 L
.22229 .30474 L
.22248 .30464 L
.22267 .30455 L
.22286 .30446 L
.22305 .30436 L
.22324 .30427 L
.22343 .30417 L
.22362 .30408 L
.22381 .30399 L
.224 .30389 L
.22419 .3038 L
.22438 .3037 L
.22457 .30361 L
.22476 .30352 L
.22495 .30342 L
.22514 .30333 L
.22533 .30324 L
.22552 .30314 L
.22571 .30305 L
.2259 .30295 L
.2261 .30286 L
.22629 .30277 L
.22648 .30267 L
.22667 .30258 L
.22686 .30249 L
.22705 .32379 L
.22724 .32369 L
Mistroke
.22743 .3236 L
.22762 .3235 L
.22781 .32341 L
.228 .32331 L
.22819 .32322 L
.22838 .32312 L
.22857 .32303 L
.22876 .32294 L
.22895 .32284 L
.22914 .32275 L
.22933 .32265 L
.22952 .32256 L
.22971 .32246 L
.2299 .32237 L
.2301 .32228 L
.23029 .32218 L
.23048 .32209 L
.23067 .32199 L
.23086 .3219 L
.23105 .3218 L
.23124 .32171 L
.23143 .32162 L
.23162 .3427 L
.23181 .34261 L
.232 .34251 L
.23219 .34242 L
.23238 .34232 L
.23257 .34223 L
.23276 .34213 L
.23295 .34204 L
.23314 .34194 L
.23333 .34185 L
.23352 .34176 L
.23371 .34166 L
.2339 .34157 L
.2341 .34147 L
.23429 .33486 L
.23448 .33476 L
.23467 .33467 L
.23486 .32811 L
.23505 .32801 L
.23524 .32792 L
.23543 .32783 L
.23562 .32773 L
.23581 .32764 L
.236 .32755 L
.23619 .32745 L
.23638 .34823 L
.23657 .34814 L
.23676 .34804 L
Mistroke
.23695 .34795 L
.23714 .34785 L
.23733 .34776 L
.23752 .34767 L
.23771 .34757 L
.2379 .34748 L
.2381 .34738 L
.23829 .34729 L
.23848 .3472 L
.23867 .3471 L
.23886 .34701 L
.23905 .34691 L
.23924 .34682 L
.23943 .34027 L
.23962 .34018 L
.23981 .34009 L
.24 .34 L
.24019 .3399 L
.24038 .33341 L
.24057 .33332 L
.24076 .33323 L
.24095 .32679 L
.24114 .32669 L
.24133 .3266 L
.24152 .32651 L
.24171 .32642 L
.2419 .32633 L
.2421 .32624 L
.24229 .32615 L
.24248 .32605 L
.24267 .32596 L
.24286 .32587 L
.24305 .32578 L
.24324 .32569 L
.24343 .3461 L
.24362 .34601 L
.24381 .34592 L
.244 .34582 L
.24419 .36606 L
.24438 .36596 L
.24457 .36587 L
.24476 .36578 L
.24495 .36569 L
.24514 .35921 L
.24533 .35912 L
.24552 .35903 L
.24571 .35893 L
.2459 .35884 L
.2461 .35875 L
.24629 .35866 L
Mistroke
.24648 .35856 L
.24667 .37856 L
.24686 .37847 L
.24705 .372 L
.24724 .37191 L
.24743 .37182 L
.24762 .37173 L
.24781 .37163 L
.248 .37154 L
.24819 .37062 L
.24838 .37053 L
.24857 .37044 L
.24876 .37035 L
.24895 .37025 L
.24914 .37016 L
.24933 .37007 L
.24952 .30812 L
.24971 .30803 L
.2499 .30795 L
.2501 .32725 L
.25029 .32716 L
.25048 .32708 L
.25067 .32699 L
.25086 .32691 L
.25105 .32682 L
.25124 .32673 L
.25143 .32665 L
.25162 .32656 L
.25181 .32058 L
.252 .32049 L
.25219 .32041 L
.25238 .32032 L
.25257 .32024 L
.25276 .32015 L
.25295 .32007 L
.25314 .31998 L
.25333 .3199 L
.25352 .31981 L
.25371 .31972 L
.2539 .31964 L
.2541 .31955 L
.25429 .31947 L
.25448 .31938 L
.25467 .3193 L
.25486 .31921 L
.25505 .26174 L
.25524 .26167 L
.25543 .26159 L
.25562 .26151 L
.25581 .26143 L
Mistroke
.256 .26135 L
.25619 .26127 L
.25638 .26119 L
.25657 .26111 L
.25676 .26103 L
.25695 .26095 L
.25714 .26087 L
.25733 .26079 L
.25752 .26072 L
.25771 .26064 L
.2579 .26056 L
.2581 .26048 L
.25829 .2604 L
.25848 .26032 L
.25867 .2548 L
.25886 .25472 L
.25905 .25465 L
.25924 .27324 L
.25943 .27316 L
.25962 .27308 L
.25981 .273 L
.26 .27292 L
.26019 .27284 L
.26038 .27276 L
.26057 .29121 L
.26076 .29113 L
.26095 .29105 L
.26114 .29097 L
.26133 .29089 L
.26152 .29081 L
.26171 .28525 L
.2619 .28517 L
.2621 .28509 L
.26229 .28501 L
.26248 .28494 L
.26267 .30319 L
.26286 .30311 L
.26305 .30303 L
.26324 .30295 L
.26343 .30287 L
.26362 .24898 L
.26381 .2489 L
.264 .26662 L
.26419 .26654 L
.26438 .26647 L
.26457 .26639 L
.26476 .26632 L
.26495 .26624 L
.26514 .26617 L
.26533 .26609 L
Mistroke
.26552 .26602 L
.26571 .26594 L
.2659 .26587 L
.2661 .26579 L
.26629 .26572 L
.26648 .26564 L
.26667 .26557 L
.26686 .26549 L
.26705 .26542 L
.26724 .26534 L
.26743 .26527 L
.26762 .26519 L
.26781 .26512 L
.268 .2599 L
.26819 .25982 L
.26838 .25463 L
.26857 .25456 L
.26876 .27205 L
.26895 .27198 L
.26914 .2719 L
.26933 .27183 L
.26952 .27176 L
.26971 .27168 L
.2699 .27161 L
.2701 .27153 L
.27029 .27146 L
.27048 .27138 L
.27067 .27131 L
.27086 .27124 L
.27105 .27116 L
.27124 .26598 L
.27143 .2659 L
.27162 .26583 L
.27181 .26575 L
.272 .26568 L
.27219 .26561 L
.27238 .26553 L
.27257 .26546 L
.27276 .26539 L
.27295 .26531 L
.27314 .26524 L
.27333 .26516 L
.27352 .26509 L
.27371 .26502 L
.2739 .26494 L
.2741 .26487 L
.27429 .2648 L
.27448 .26472 L
.27467 .25958 L
.27486 .25951 L
Mistroke
.27505 .25943 L
.27524 .25936 L
.27543 .25929 L
.27562 .25921 L
.27581 .25914 L
.276 .25907 L
.27619 .27634 L
.27638 .22634 L
.27657 .22627 L
.27676 .22621 L
.27695 .22614 L
.27714 .22607 L
.27733 .226 L
.27752 .22593 L
.27771 .22586 L
.2779 .22579 L
.2781 .22572 L
.27829 .22565 L
.27848 .22558 L
.27867 .22076 L
.27886 .22069 L
.27905 .22062 L
.27924 .22055 L
.27943 .22049 L
.27962 .22042 L
.27981 .22035 L
.28 .21556 L
.28019 .21549 L
.28038 .21542 L
.28057 .21535 L
.28076 .21528 L
.28095 .21052 L
.28114 .21046 L
.28133 .21039 L
.28152 .21032 L
.28171 .21025 L
.2819 .21019 L
.2821 .21012 L
.28229 .21005 L
.28248 .20998 L
.28267 .20991 L
.28286 .20985 L
.28305 .20978 L
.28324 .20971 L
.28343 .20964 L
.28362 .20958 L
.28381 .20485 L
.284 .20016 L
.28419 .20009 L
.28438 .20002 L
Mistroke
.28457 .19996 L
.28476 .19989 L
.28495 .19982 L
.28514 .19975 L
.28533 .19969 L
.28552 .21628 L
.28571 .21621 L
.2859 .21615 L
.2861 .21608 L
.28629 .21601 L
.28648 .1703 L
.28667 .17023 L
.28686 .17017 L
.28705 .17011 L
.28724 .17004 L
.28743 .16562 L
.28762 .16556 L
.28781 .1655 L
.288 .16543 L
.28819 .16537 L
.28838 .16531 L
.28857 .16524 L
.28876 .16518 L
.28895 .16512 L
.28914 .16506 L
.28933 .16499 L
.28952 .16493 L
.28971 .16487 L
.2899 .1648 L
.2901 .16474 L
.29029 .16468 L
.29048 .16462 L
.29067 .16455 L
.29086 .16449 L
.29105 .16443 L
.29124 .16436 L
.29143 .18046 L
.29162 .19646 L
.29181 .19639 L
.292 .19633 L
.29219 .19627 L
.29238 .1962 L
.29257 .19614 L
.29276 .19607 L
.29295 .19601 L
.29314 .19595 L
.29333 .19588 L
.29352 .19582 L
.29371 .19576 L
.2939 .19569 L
Mistroke
.2941 .19563 L
.29429 .19556 L
.29448 .1955 L
.29467 .19544 L
.29486 .19537 L
.29505 .19531 L
.29524 .19525 L
.29543 .19518 L
.29562 .19512 L
.29581 .19505 L
.296 .19499 L
.29619 .19493 L
.29638 .19486 L
.29657 .1948 L
.29676 .19474 L
.29695 .19467 L
.29714 .19404 L
.29733 .19397 L
.29752 .19391 L
.29771 .19385 L
.2979 .19378 L
.2981 .19372 L
.29829 .19365 L
.29848 .19359 L
.29867 .19353 L
.29886 .19346 L
.29905 .20933 L
.29924 .20926 L
.29943 .2092 L
.29962 .20914 L
.29981 .20907 L
.3 .20901 L
.30019 .38043 L
.30038 .38037 L
.30057 .3803 L
.30076 .38023 L
.30095 .38016 L
.30114 .38009 L
.30133 .38002 L
.30152 .37996 L
.30171 .37989 L
.3019 .37982 L
.3021 .37975 L
.30229 .37968 L
.30248 .37961 L
.30267 .37954 L
.30286 .37948 L
.30305 .37941 L
.30324 .37463 L
.30343 .37456 L
Mistroke
.30362 .37449 L
.30381 .36974 L
.304 .36502 L
.30419 .36495 L
.30438 .36488 L
.30457 .36482 L
.30476 .36475 L
.30495 .36468 L
.30514 .36461 L
.30533 .36455 L
.30552 .35985 L
.30571 .35978 L
.3059 .35972 L
.3061 .35965 L
.30629 .35958 L
.30648 .35952 L
.30667 .35945 L
.30686 .35938 L
.30705 .35932 L
.30724 .35925 L
.30743 .35918 L
.30762 .35912 L
.30781 .35905 L
.308 .35898 L
.30819 .35892 L
.30838 .35885 L
.30857 .35878 L
.30876 .35871 L
.30895 .35865 L
.30914 .35858 L
.30933 .35851 L
.30952 .35845 L
.30971 .35838 L
.3099 .3729 L
.3101 .37283 L
.31029 .37277 L
.31048 .3727 L
.31067 .37263 L
.31086 .37257 L
.31105 .3725 L
.31124 .37243 L
.31143 .37237 L
.31162 .3723 L
.31181 .37223 L
.312 .37216 L
.31219 .3721 L
.31238 .37203 L
.31257 .37196 L
.31276 .3719 L
.31295 .37183 L
Mistroke
.31314 .37176 L
.31333 .37169 L
.31352 .37163 L
.31371 .37156 L
.3139 .37149 L
.3141 .37143 L
.31429 .37136 L
.31448 .37129 L
.31467 .37123 L
.31486 .37116 L
.31505 .37109 L
.31524 .36641 L
.31543 .36635 L
.31562 .36628 L
.31581 .36621 L
.316 .36615 L
.31619 .36608 L
.31638 .36601 L
.31657 .36595 L
.31676 .36588 L
.31695 .36581 L
.31714 .36575 L
.31733 .36568 L
.31752 .36562 L
.31771 .36555 L
.3179 .36548 L
.3181 .36542 L
.31829 .36077 L
.31848 .3607 L
.31867 .36064 L
.31886 .36057 L
.31905 .36051 L
.31924 .36044 L
.31943 .36037 L
.31962 .36031 L
.31981 .36024 L
.32 .36017 L
.32019 .36011 L
.32038 .36004 L
.32057 .35998 L
.32076 .35991 L
.32095 .35984 L
.32114 .35978 L
.32133 .35971 L
.32152 .35965 L
.32171 .35898 L
.3219 .35892 L
.3221 .35885 L
.32229 .35879 L
.32248 .35872 L
Mistroke
.32267 .35865 L
.32286 .35859 L
.32305 .37294 L
.32324 .36831 L
.32343 .36824 L
.32362 .36818 L
.32381 .36811 L
.324 .36805 L
.32419 .36798 L
.32438 .36791 L
.32457 .36785 L
.32476 .36778 L
.32495 .36772 L
.32514 .36765 L
.32533 .36758 L
.32552 .36752 L
.32571 .36745 L
.3259 .36285 L
.3261 .36278 L
.32629 .36272 L
.32648 .36265 L
.32667 .36259 L
.32686 .36252 L
.32705 .36245 L
.32724 .36239 L
.32743 .36232 L
.32762 .36226 L
.32781 .36219 L
.328 .36213 L
.32819 .36206 L
.32838 .37627 L
.32857 .3762 L
.32876 .37614 L
.32895 .37607 L
.32914 .376 L
.32933 .37594 L
.32952 .39006 L
.32971 .38999 L
.3299 .38992 L
.3301 .3853 L
.33029 .38524 L
.33048 .38517 L
.33067 .38511 L
.33086 .38504 L
.33105 .38498 L
.33124 .38491 L
.33143 .38484 L
.33162 .38478 L
.33181 .38471 L
.332 .38465 L
Mistroke
.33219 .38458 L
.33238 .38451 L
.33257 .38445 L
.33276 .38438 L
.33295 .38432 L
.33314 .38425 L
.33333 .38419 L
.33352 .38412 L
.33371 .38405 L
.3339 .38399 L
.3341 .38392 L
.33429 .37934 L
.33448 .37927 L
.33467 .37921 L
.33486 .37914 L
.33505 .37907 L
.33524 .37901 L
.33543 .37894 L
.33562 .37888 L
.33581 .37881 L
.336 .37875 L
.33619 .33422 L
.33638 .33416 L
.33657 .3341 L
.33676 .33404 L
.33695 .34778 L
.33714 .34771 L
.33733 .34765 L
.33752 .30518 L
.33771 .30512 L
.3379 .30506 L
.3381 .305 L
.33829 .30495 L
.33848 .30489 L
.33867 .30483 L
.33886 .30477 L
.33905 .30471 L
.33924 .30465 L
.33943 .30459 L
.33962 .26425 L
.33981 .26419 L
.34 .26414 L
.34019 .26408 L
.34038 .26403 L
.34057 .26397 L
.34076 .26391 L
.34095 .26386 L
.34114 .2638 L
.34133 .26375 L
.34152 .26369 L
Mistroke
.34171 .26363 L
.3419 .26358 L
.3421 .26352 L
.34229 .2596 L
.34248 .25954 L
.34267 .25949 L
.34286 .22132 L
.34305 .22127 L
.34324 .22122 L
.34343 .22116 L
.34362 .22111 L
.34381 .22106 L
.344 .221 L
.34419 .22095 L
.34438 .2209 L
.34457 .22084 L
.34476 .22079 L
.34495 .21708 L
.34514 .21702 L
.34533 .21697 L
.34552 .21692 L
.34571 .21687 L
.3459 .21681 L
.3461 .21676 L
.34629 .22971 L
.34648 .22966 L
.34667 .2296 L
.34686 .22955 L
.34705 .2295 L
.34724 .22945 L
.34743 .22939 L
.34762 .22934 L
.34781 .22929 L
.348 .22923 L
.34819 .22918 L
.34838 .22913 L
.34857 .22907 L
.34876 .22902 L
.34895 .22897 L
.34914 .22891 L
.34933 .22886 L
.34952 .24174 L
.34971 .24168 L
.3499 .24163 L
.3501 .24158 L
.35029 .24152 L
.35048 .24147 L
.35067 .24142 L
.35086 .24136 L
.35105 .24131 L
Mistroke
.35124 .24126 L
.35143 .2412 L
.35162 .24115 L
.35181 .2411 L
.352 .2539 L
.35219 .25384 L
.35238 .25379 L
.35257 .25374 L
.35276 .25368 L
.35295 .25363 L
.35314 .25358 L
.35333 .25352 L
.35352 .25347 L
.35371 .25342 L
.3539 .24967 L
.3541 .24962 L
.35429 .24956 L
.35448 .26227 L
.35467 .25852 L
.35486 .25847 L
.35505 .25842 L
.35524 .25836 L
.35543 .25831 L
.35562 .25826 L
.35581 .2582 L
.356 .25767 L
.35619 .25394 L
.35638 .25389 L
.35657 .25384 L
.35676 .25378 L
.35695 .25373 L
.35714 .25002 L
.35733 .24997 L
.35752 .24992 L
.35771 .24987 L
.3579 .24981 L
.3581 .24976 L
.35829 .24971 L
.35848 .24965 L
.35867 .2496 L
.35886 .24955 L
.35905 .2495 L
.35924 .24944 L
.35943 .24939 L
.35962 .24934 L
.35981 .24928 L
.36 .2456 L
.36019 .24554 L
.36038 .24549 L
.36057 .24544 L
Mistroke
.36076 .24539 L
.36095 .24533 L
.36114 .24528 L
.36133 .24523 L
.36152 .24518 L
.36171 .24512 L
.3619 .24507 L
.3621 .24502 L
.36229 .24497 L
.36248 .24491 L
.36267 .24486 L
.36286 .24481 L
.36305 .24476 L
.36324 .2447 L
.36343 .24465 L
.36362 .2446 L
.36381 .24455 L
.364 .24449 L
.36419 .24444 L
.36438 .24439 L
.36457 .24434 L
.36476 .24428 L
.36495 .24423 L
.36514 .24057 L
.36533 .24052 L
.36552 .24046 L
.36571 .24041 L
.3659 .24036 L
.3661 .24031 L
.36629 .24026 L
.36648 .2402 L
.36667 .24015 L
.36686 .2401 L
.36705 .24005 L
.36724 .23999 L
.36743 .25252 L
.36762 .25247 L
.36781 .25241 L
.368 .25236 L
.36819 .26482 L
.36838 .26477 L
.36857 .26471 L
.36876 .26466 L
.36895 .26461 L
.36914 .26455 L
.36933 .2645 L
.36952 .26445 L
.36971 .2644 L
.3699 .26434 L
.3701 .26429 L
Mistroke
.37029 .26061 L
.37048 .26056 L
.37067 .26051 L
.37086 .26045 L
.37105 .2604 L
.37124 .26035 L
.37143 .2603 L
.37162 .26025 L
.37181 .26019 L
.372 .26014 L
.37219 .26009 L
.37238 .22424 L
.37257 .22419 L
.37276 .22414 L
.37295 .22409 L
.37314 .22404 L
.37333 .22054 L
.37352 .22049 L
.37371 .22044 L
.3739 .22039 L
.3741 .22034 L
.37429 .22029 L
.37448 .22024 L
.37467 .22019 L
.37486 .22014 L
.37505 .22009 L
.37524 .22004 L
.37543 .21999 L
.37562 .21994 L
.37581 .21989 L
.376 .21984 L
.37619 .21979 L
.37638 .21974 L
.37657 .21969 L
.37676 .21964 L
.37695 .21959 L
.37714 .21954 L
.37733 .21949 L
.37752 .21945 L
.37771 .21597 L
.3779 .21592 L
.3781 .21587 L
.37829 .21582 L
.37848 .21577 L
.37867 .21572 L
.37886 .21567 L
.37905 .21562 L
.37924 .21557 L
.37943 .21552 L
.37962 .21548 L
Mistroke
.37981 .21543 L
.38 .21538 L
.38019 .21533 L
.38038 .21528 L
.38057 .21523 L
.38076 .21177 L
.38095 .21172 L
.38114 .22384 L
.38133 .23589 L
.38152 .23584 L
.38171 .23579 L
.3819 .23574 L
.3821 .23569 L
.38229 .23564 L
.38248 .23559 L
.38267 .23554 L
.38286 .23549 L
.38305 .23544 L
.38324 .23539 L
.38343 .23534 L
.38362 .23529 L
.38381 .23524 L
.384 .23519 L
.38419 .23514 L
.38438 .23509 L
.38457 .23504 L
.38476 .23499 L
.38495 .23494 L
.38514 .23489 L
.38533 .23484 L
.38552 .2348 L
.38571 .23475 L
.3859 .23127 L
.3861 .23122 L
.38629 .23118 L
.38648 .23113 L
.38667 .23108 L
.38686 .22762 L
.38705 .22757 L
.38724 .22752 L
.38743 .22747 L
.38762 .22742 L
.38781 .22737 L
.388 .22733 L
.38819 .22728 L
.38838 .22723 L
.38857 .22718 L
.38876 .22713 L
.38895 .22708 L
.38914 .22703 L
Mistroke
.38933 .22698 L
.38952 .22693 L
.38971 .22688 L
.3899 .22683 L
.3901 .22678 L
.39029 .22674 L
.39048 .22669 L
.39067 .22664 L
.39086 .22659 L
.39105 .22315 L
.39124 .2231 L
.39143 .22305 L
.39162 .22301 L
.39181 .22296 L
.392 .22291 L
.39219 .22286 L
.39238 .22237 L
.39257 .22232 L
.39276 .22227 L
.39295 .22222 L
.39314 .22217 L
.39333 .22212 L
.39352 .22208 L
.39371 .22203 L
.3939 .22198 L
.3941 .22193 L
.39429 .22188 L
.39448 .22183 L
.39467 .21842 L
.39486 .21837 L
.39505 .21832 L
.39524 .21827 L
.39543 .21822 L
.39562 .21774 L
.39581 .21769 L
.396 .21764 L
.39619 .34812 L
.39638 .34807 L
.39657 .34802 L
.39676 .34797 L
.39695 .34792 L
.39714 .34787 L
.39733 .34782 L
.39752 .34776 L
.39771 .34771 L
.3979 .34766 L
.3981 .34761 L
.39829 .34756 L
.39848 .34751 L
.39867 .34746 L
Mistroke
.39886 .31235 L
.39905 .3123 L
.39924 .31225 L
.39943 .3122 L
.39962 .31215 L
.39981 .3121 L
.4 .31205 L
.40019 .312 L
.40038 .3231 L
.40057 .33414 L
.40076 .33409 L
.40095 .33404 L
.40114 .33399 L
.40133 .33394 L
.40152 .33389 L
.40171 .33384 L
.4019 .33379 L
.4021 .33374 L
.40229 .3337 L
.40248 .33365 L
.40267 .3336 L
.40286 .33355 L
.40305 .3335 L
.40324 .33345 L
.40343 .3334 L
.40362 .32994 L
.40381 .32989 L
.404 .32984 L
.40419 .3298 L
.40438 .32975 L
.40457 .3297 L
.40476 .32965 L
.40495 .3296 L
.40514 .32955 L
.40533 .3295 L
.40552 .32945 L
.40571 .3294 L
.4059 .32935 L
.4061 .3293 L
.40629 .32925 L
.40648 .3292 L
.40667 .32916 L
.40686 .32911 L
.40705 .32906 L
.40724 .32901 L
.40743 .32896 L
.40762 .32891 L
.40781 .32886 L
.408 .32881 L
.40819 .32876 L
Mistroke
.40838 .29503 L
.40857 .29498 L
.40876 .29494 L
.40895 .29489 L
.40914 .29484 L
.40933 .29479 L
.40952 .29475 L
.40971 .2947 L
.4099 .29465 L
.4101 .2946 L
.41029 .29456 L
.41048 .29451 L
.41067 .29446 L
.41086 .29442 L
.41105 .29437 L
.41124 .29432 L
.41143 .29427 L
.41162 .29423 L
.41181 .29418 L
.412 .29413 L
.41219 .29409 L
.41238 .29079 L
.41257 .29075 L
.41276 .2907 L
.41295 .29065 L
.41314 .2906 L
.41333 .30141 L
.41352 .30136 L
.41371 .30131 L
.4139 .30127 L
.4141 .30122 L
.41429 .30117 L
.41448 .30112 L
.41467 .30108 L
.41486 .30103 L
.41505 .30098 L
.41524 .30094 L
.41543 .30089 L
.41562 .30084 L
.41581 .3008 L
.416 .30075 L
.41619 .3007 L
.41638 .26838 L
.41657 .26833 L
.41676 .26829 L
.41695 .26824 L
.41714 .2682 L
.41733 .26815 L
.41752 .26811 L
.41771 .26806 L
Mistroke
.4179 .26802 L
.4181 .26797 L
.41829 .26792 L
.41848 .26788 L
.41867 .26783 L
.41886 .26779 L
.41905 .26774 L
.41924 .2677 L
.41943 .26454 L
.41962 .26449 L
.41981 .26445 L
.42 .2644 L
.42019 .26436 L
.42038 .26431 L
.42057 .26427 L
.42076 .26422 L
.42095 .26418 L
.42114 .26413 L
.42133 .26409 L
.42152 .26404 L
.42171 .264 L
.4219 .26085 L
.4221 .26081 L
.42229 .26077 L
.42248 .26072 L
.42267 .26068 L
.42286 .26063 L
.42305 .26059 L
.42324 .26054 L
.42343 .2605 L
.42362 .26045 L
.42381 .26041 L
.424 .26036 L
.42419 .26032 L
.42438 .26027 L
.42457 .26023 L
.42476 .2571 L
.42495 .25398 L
.42514 .25394 L
.42533 .25349 L
.42552 .25039 L
.42571 .25034 L
.4259 .2503 L
.4261 .25026 L
.42629 .25021 L
.42648 .25017 L
.42667 .25012 L
.42686 .25008 L
.42705 .21972 L
.42724 .1905 L
Mistroke
.42743 .19046 L
.42762 .19042 L
.42781 .19038 L
.428 .20064 L
.42819 .2006 L
.42838 .20056 L
.42857 .20052 L
.42876 .19764 L
.42895 .1976 L
.42914 .19756 L
.42933 .16942 L
.42952 .16938 L
.42971 .16934 L
.4299 .1693 L
.4301 .16926 L
.43029 .16922 L
.43048 .16918 L
.43067 .16914 L
.43086 .1691 L
.43105 .16906 L
.43124 .16902 L
.43143 .16898 L
.43162 .16894 L
.43181 .1689 L
.432 .16886 L
.43219 .16883 L
.43238 .16879 L
.43257 .16875 L
.43276 .16871 L
.43295 .16867 L
.43314 .16863 L
.43333 .16859 L
.43352 .16583 L
.43371 .16579 L
.4339 .16576 L
.4341 .16536 L
.43429 .16532 L
.43448 .16528 L
.43467 .16525 L
.43486 .16521 L
.43505 .16517 L
.43524 .16513 L
.43543 .16509 L
.43562 .16505 L
.43581 .16501 L
.436 .16497 L
.43619 .16493 L
.43638 .16489 L
.43657 .16485 L
.43676 .16481 L
Mistroke
.43695 .16477 L
.43714 .16474 L
.43733 .1647 L
.43752 .13774 L
.43771 .1377 L
.4379 .14763 L
.4381 .14759 L
.43829 .14755 L
.43848 .14752 L
.43867 .14748 L
.43886 .14744 L
.43905 .1474 L
.43924 .14736 L
.43943 .14733 L
.43962 .14729 L
.43981 .14725 L
.44 .14721 L
.44019 .14717 L
.44038 .14714 L
.44057 .1471 L
.44076 .14706 L
.44095 .14702 L
.44114 .12092 L
.44133 .12089 L
.44152 .12085 L
.44171 .12081 L
.4419 .12078 L
.4421 .12074 L
.44229 .12071 L
.44248 .13046 L
.44267 .13043 L
.44286 .13039 L
.44305 .13035 L
.44324 .13032 L
.44343 .13028 L
.44362 .13024 L
.44381 .13021 L
.444 .13017 L
.44419 .13013 L
.44438 .1301 L
.44457 .13006 L
.44476 .10477 L
.44495 .10474 L
.44514 .11434 L
.44533 .1143 L
.44552 .11427 L
.44571 .08973 L
.4459 .08969 L
.4461 .08966 L
.44629 .08963 L
Mistroke
.44648 .08722 L
.44667 .08718 L
.44686 .08715 L
.44705 .08712 L
.44724 .08708 L
.44743 .08705 L
.44762 .08701 L
.44781 .08698 L
.448 .08694 L
.44819 .08691 L
.44838 .08688 L
.44857 .08684 L
.44876 .09627 L
.44895 .10567 L
.44914 .10564 L
.44933 .11616 L
.44952 .11613 L
.44971 .11609 L
.4499 .11606 L
.4501 .11602 L
.45029 .11599 L
.45048 .11596 L
.45067 .11592 L
.45086 .11589 L
.45105 .11585 L
.45124 .11582 L
.45143 .12513 L
.45162 .1251 L
.45181 .12506 L
.452 .12503 L
.45219 .12499 L
.45238 .12496 L
.45257 .12492 L
.45276 .12489 L
.45295 .12245 L
.45314 .12241 L
.45333 .13168 L
.45352 .1076 L
.45371 .10757 L
.4539 .10753 L
.4541 .1075 L
.45429 .10747 L
.45448 .10743 L
.45467 .1074 L
.45486 .10736 L
.45505 .10733 L
.45524 .10497 L
.45543 .10493 L
.45562 .1049 L
.45581 .10487 L
Mistroke
.456 .10483 L
.45619 .1048 L
.45638 .10476 L
.45657 .10473 L
.45676 .1047 L
.45695 .10466 L
.45714 .10231 L
.45733 .10228 L
.45752 .10224 L
.45771 .10221 L
.4579 .10217 L
.4581 .10214 L
.45829 .11124 L
.45848 .11121 L
.45867 .11118 L
.45886 .11114 L
.45905 .11111 L
.45924 .11108 L
.45943 .11104 L
.45962 .11101 L
.45981 .11097 L
.46 .11094 L
.46019 .11091 L
.46038 .11087 L
.46057 .11084 L
.46076 .11081 L
.46095 .10845 L
.46114 .10842 L
.46133 .10838 L
.46152 .10835 L
.46171 .11741 L
.4619 .11737 L
.4621 .11734 L
.46229 .1173 L
.46248 .11727 L
.46267 .11724 L
.46286 .1172 L
.46305 .11717 L
.46324 .11481 L
.46343 .11478 L
.46362 .11475 L
.46381 .1124 L
.464 .11236 L
.46419 .11233 L
.46438 .1123 L
.46457 .11226 L
.46476 .11223 L
.46495 .1122 L
.46514 .11216 L
.46533 .11213 L
Mistroke
.46552 .11209 L
.46571 .11206 L
.4659 .10972 L
.4661 .10939 L
.46629 .10935 L
.46648 .10932 L
.46667 .10929 L
.46686 .10925 L
.46705 .10922 L
.46724 .10919 L
.46743 .10915 L
.46762 .10912 L
.46781 .10909 L
.468 .10905 L
.46819 .10902 L
.46838 .08605 L
.46857 .08602 L
.46876 .08376 L
.46895 .09263 L
.46914 .0926 L
.46933 .09257 L
.46952 .09254 L
.46971 .0925 L
.4699 .09247 L
.4701 .09021 L
.47029 .09018 L
.47048 .08793 L
.47067 .09785 L
.47086 .09782 L
.47105 .09779 L
.47124 .09775 L
.47143 .09772 L
.47162 .09769 L
.47181 .09766 L
.472 .09762 L
.47219 .0973 L
.47238 .09727 L
.47257 .09724 L
.47276 .09721 L
.47295 .09717 L
.47314 .09714 L
.47333 .09711 L
.47352 .09708 L
.47371 .09704 L
.4739 .09701 L
.4741 .09476 L
.47429 .09473 L
.47448 .09469 L
.47467 .09466 L
.47486 .09463 L
Mistroke
.47505 .0946 L
.47524 .09457 L
.47543 .09453 L
.47562 .0945 L
.47581 .09447 L
.476 .09444 L
.47619 .09441 L
.47638 .09437 L
.47657 .09213 L
.47676 .0921 L
.47695 .09206 L
.47714 .09203 L
.47733 .092 L
.47752 .09197 L
.47771 .09194 L
.4779 .0919 L
.4781 .09187 L
.47829 .09184 L
.47848 .1006 L
.47867 .10057 L
.47886 .09832 L
.47905 .09829 L
.47924 .09826 L
.47943 .09822 L
.47962 .09819 L
.47981 .09816 L
.48 .09813 L
.48019 .0981 L
.48038 .09586 L
.48057 .09583 L
.48076 .09579 L
.48095 .09576 L
.48114 .09573 L
.48133 .0957 L
.48152 .09567 L
.48171 .09563 L
.4819 .0956 L
.4821 .09557 L
.48229 .09554 L
.48248 .09551 L
.48267 .09548 L
.48286 .09324 L
.48305 .09321 L
.48324 .09318 L
.48343 .09315 L
.48362 .09312 L
.48381 .09309 L
.484 .09305 L
.48419 .09302 L
.48438 .09299 L
Mistroke
.48457 .09296 L
.48476 .09293 L
.48495 .09289 L
.48514 .09286 L
.48533 .10156 L
.48552 .10153 L
.48571 .10149 L
.4859 .10146 L
.4861 .10143 L
.48629 .1014 L
.48648 .10137 L
.48667 .10133 L
.48686 .1013 L
.48705 .10127 L
.48724 .10124 L
.48743 .10121 L
.48762 .10117 L
.48781 .09894 L
.488 .09891 L
.48819 .09888 L
.48838 .09885 L
.48857 .09882 L
.48876 .09878 L
.48895 .10744 L
.48914 .1052 L
.48933 .10517 L
.48952 .10514 L
.48971 .10511 L
.4899 .10508 L
.4901 .10504 L
.49029 .10501 L
.49048 .10498 L
.49067 .10495 L
.49086 .10492 L
.49105 .10488 L
.49124 .10485 L
.49143 .10482 L
.49162 .10479 L
.49181 .10476 L
.492 .10473 L
.49219 .10469 L
.49238 .10466 L
.49257 .10463 L
.49276 .1046 L
.49295 .10457 L
.49314 .10453 L
.49333 .1045 L
.49352 .10447 L
.49371 .10444 L
.4939 .10441 L
Mistroke
.4941 .10438 L
.49429 .11298 L
.49448 .11295 L
.49467 .11292 L
.49486 .11289 L
.49505 .11286 L
.49524 .12143 L
.49543 .1214 L
.49562 .12137 L
.49581 .12134 L
.496 .1213 L
.49619 .12127 L
.49638 .12124 L
.49657 .12121 L
.49676 .12117 L
.49695 .12114 L
.49714 .12111 L
.49733 .12108 L
.49752 .12105 L
.49771 .12101 L
.4979 .12098 L
.4981 .12095 L
.49829 .11871 L
.49848 .11868 L
.49867 .11865 L
.49886 .11861 L
.49905 .11858 L
.49924 .11855 L
.49943 .12708 L
.49962 .12705 L
.49981 .12702 L
.5 .12699 L
.50019 .12696 L
.50038 .12472 L
.50057 .12468 L
.50076 .12465 L
.50095 .12462 L
.50114 .12459 L
.50133 .12456 L
.50152 .12452 L
.50171 .12449 L
.5019 .12417 L
.5021 .12414 L
.50229 .12411 L
.50248 .12408 L
.50267 .12404 L
.50286 .12401 L
.50305 .12398 L
.50324 .12175 L
.50343 .12172 L
Mistroke
.50362 .12168 L
.50381 .12165 L
.504 .12162 L
.50419 .12159 L
.50438 .12156 L
.50457 .12153 L
.50476 .12149 L
.50495 .12146 L
.50514 .12143 L
.50533 .1214 L
.50552 .12137 L
.50571 .12133 L
.5059 .1213 L
.5061 .12127 L
.50629 .12124 L
.50648 .12121 L
.50667 .12118 L
.50686 .12114 L
.50705 .12111 L
.50724 .12959 L
.50743 .12736 L
.50762 .1358 L
.50781 .14421 L
.508 .14418 L
.50819 .14415 L
.50838 .14411 L
.50857 .14408 L
.50876 .14405 L
.50895 .15243 L
.50914 .1524 L
.50933 .15237 L
.50952 .15234 L
.50971 .15008 L
.5099 .15005 L
.5101 .15002 L
.51029 .15836 L
.51048 .15833 L
.51067 .1583 L
.51086 .15827 L
.51105 .16658 L
.51124 .16655 L
.51143 .16651 L
.51162 .16648 L
.51181 .16645 L
.512 .16642 L
.51219 .16639 L
.51238 .16635 L
.51257 .16632 L
.51276 .16629 L
.51295 .16626 L
Mistroke
.51314 .16622 L
.51333 .16619 L
.51352 .16616 L
.51371 .17444 L
.5139 .17441 L
.5141 .17438 L
.51429 .17434 L
.51448 .17431 L
.51467 .17428 L
.51486 .17425 L
.51505 .17421 L
.51524 .17418 L
.51543 .17415 L
.51562 .17412 L
.51581 .17409 L
.516 .17405 L
.51619 .17402 L
.51638 .17399 L
.51657 .17396 L
.51676 .17392 L
.51695 .17389 L
.51714 .17386 L
.51733 .17383 L
.51752 .15146 L
.51771 .15142 L
.5179 .15139 L
.5181 .15136 L
.51829 .15133 L
.51848 .1513 L
.51867 .15127 L
.51886 .15124 L
.51905 .1512 L
.51924 .15117 L
.51943 .15114 L
.51962 .14894 L
.51981 .14891 L
.52 .14888 L
.52019 .14885 L
.52038 .14882 L
.52057 .14879 L
.52076 .14875 L
.52095 .14872 L
.52114 .14869 L
.52133 .14866 L
.52152 .14863 L
.52171 .1486 L
.5219 .14857 L
.5221 .14637 L
.52229 .14634 L
.52248 .14416 L
Mistroke
.52267 .12266 L
.52286 .12263 L
.52305 .1226 L
.52324 .12257 L
.52343 .12254 L
.52362 .12251 L
.52381 .12248 L
.524 .13053 L
.52419 .21934 L
.52438 .21931 L
.52457 .21928 L
.52476 .21925 L
.52495 .21922 L
.52514 .21919 L
.52533 .21915 L
.52552 .21912 L
.52571 .21909 L
.5259 .21906 L
.5261 .21903 L
.52629 .22676 L
.52648 .22455 L
.52667 .22451 L
.52686 .22448 L
.52705 .22445 L
.52724 .22442 L
.52743 .22439 L
.52762 .22436 L
.52781 .22432 L
.528 .22429 L
.52819 .22426 L
.52838 .22423 L
.52857 .2242 L
.52876 .22417 L
.52895 .22413 L
.52914 .22192 L
.52933 .22189 L
.52952 .22186 L
.52971 .22956 L
.5299 .22952 L
.5301 .22949 L
.53029 .22946 L
.53048 .22943 L
.53067 .2371 L
.53086 .23707 L
.53105 .24471 L
.53124 .24468 L
.53143 .2523 L
.53162 .25226 L
.53181 .25223 L
.532 .2522 L
Mistroke
.53219 .25217 L
.53238 .24994 L
.53257 .24771 L
.53276 .24768 L
.53295 .25526 L
.53314 .26281 L
.53333 .26278 L
.53352 .24073 L
.53371 .2407 L
.5339 .24066 L
.5341 .32326 L
.53429 .32323 L
.53448 .32319 L
.53467 .32316 L
.53486 .32313 L
.53505 .3231 L
.53524 .32307 L
.53543 .32303 L
.53562 .323 L
.53581 .32297 L
.536 .32294 L
.53619 .32291 L
.53638 .32287 L
.53657 .32284 L
.53676 .32281 L
.53695 .32278 L
.53714 .32275 L
.53733 .32271 L
.53752 .32268 L
.53771 .32988 L
.5379 .32985 L
.5381 .32982 L
.53829 .32978 L
.53848 .32975 L
.53867 .32972 L
.53886 .32969 L
.53905 .33687 L
.53924 .33683 L
.53943 .3368 L
.53962 .33677 L
.53981 .33674 L
.54 .3367 L
.54019 .33667 L
.54038 .33664 L
.54057 .33661 L
.54076 .33657 L
.54095 .33654 L
.54114 .33651 L
.54133 .33648 L
.54152 .33645 L
Mistroke
.54171 .33641 L
.5419 .33638 L
.5421 .33635 L
.54229 .33632 L
.54248 .33628 L
.54267 .33625 L
.54286 .31398 L
.54305 .32107 L
.54324 .32104 L
.54343 .321 L
.54362 .32097 L
.54381 .32094 L
.544 .32091 L
.54419 .31871 L
.54438 .31868 L
.54457 .31865 L
.54476 .31862 L
.54495 .31858 L
.54514 .31855 L
.54533 .31852 L
.54552 .29684 L
.54571 .29681 L
.5459 .29678 L
.5461 .30378 L
.54629 .30375 L
.54648 .30372 L
.54667 .30369 L
.54686 .30366 L
.54705 .30363 L
.54724 .3036 L
.54743 .31058 L
.54762 .31055 L
.54781 .31052 L
.548 .31049 L
.54819 .31046 L
.54838 .31043 L
.54857 .31039 L
.54876 .31735 L
.54895 .31732 L
.54914 .31729 L
.54933 .31726 L
.54952 .31511 L
.54971 .32204 L
.5499 .32201 L
.5501 .32198 L
.55029 .32195 L
.55048 .32192 L
.55067 .32189 L
.55086 .32186 L
.55105 .32183 L
Mistroke
.55124 .32179 L
.55143 .32176 L
.55162 .32173 L
.55181 .3217 L
.552 .32167 L
.55219 .32164 L
.55238 .32161 L
.55257 .32158 L
.55276 .32155 L
.55295 .32152 L
.55314 .32149 L
.55333 .32146 L
.55352 .32143 L
.55371 .32139 L
.5539 .32136 L
.5541 .32133 L
.55429 .30007 L
.55448 .30004 L
.55467 .30001 L
.55486 .29791 L
.55505 .29788 L
.55524 .29785 L
.55543 .29782 L
.55562 .30467 L
.55581 .28394 L
.556 .28391 L
.55619 .28388 L
.55638 .28386 L
.55657 .28383 L
.55676 .2838 L
.55695 .28377 L
.55714 .28374 L
.55733 .28371 L
.55752 .29048 L
.55771 .28843 L
.5579 .2884 L
.5581 .28837 L
.55829 .28834 L
.55848 .28831 L
.55867 .28627 L
.55886 .28624 L
.55905 .28621 L
.55924 .28618 L
.55943 .28615 L
.55962 .28612 L
.55981 .28609 L
.56 .28607 L
.56019 .28604 L
.56038 .284 L
.56057 .28397 L
Mistroke
.56076 .28394 L
.56095 .28391 L
.56114 .28388 L
.56133 .28385 L
.56152 .28382 L
.56171 .28179 L
.5619 .28176 L
.5621 .28174 L
.56229 .28846 L
.56248 .28843 L
.56267 .2884 L
.56286 .36275 L
.56305 .36272 L
.56324 .36269 L
.56343 .36266 L
.56362 .36263 L
.56381 .3626 L
.564 .36257 L
.56419 .36254 L
.56438 .36251 L
.56457 .36248 L
.56476 .36245 L
.56495 .36242 L
.56514 .36239 L
.56533 .36236 L
.56552 .36233 L
.56571 .3623 L
.5659 .36227 L
.5661 .36224 L
.56629 .36221 L
.56648 .36218 L
.56667 .36215 L
.56686 .36212 L
.56705 .36209 L
.56724 .36206 L
.56743 .36203 L
.56762 .362 L
.56781 .36197 L
.568 .36194 L
.56819 .36164 L
.56838 .36161 L
.56857 .35952 L
.56876 .35949 L
.56895 .35946 L
.56914 .35943 L
.56933 .3594 L
.56952 .35937 L
.56971 .36586 L
.5699 .36583 L
.5701 .3658 L
Mistroke
.57029 .36577 L
.57048 .36368 L
.57067 .36365 L
.57086 .36362 L
.57105 .37008 L
.57124 .37005 L
.57143 .37002 L
.57162 .37 L
.57181 .36997 L
.572 .36994 L
.57219 .36991 L
.57238 .36988 L
.57257 .36985 L
.57276 .36982 L
.57295 .36979 L
.57314 .36976 L
.57333 .36973 L
.57352 .3697 L
.57371 .36967 L
.5739 .36964 L
.5741 .36961 L
.57429 .37605 L
.57448 .35536 L
.57467 .35533 L
.57486 .3553 L
.57505 .35528 L
.57524 .35525 L
.57543 .35522 L
.57562 .36159 L
.57581 .36156 L
.576 .36153 L
.57619 .3615 L
.57638 .36147 L
.57657 .35943 L
.57676 .3594 L
.57695 .36575 L
.57714 .36572 L
.57733 .36569 L
.57752 .36566 L
.57771 .36564 L
.5779 .36561 L
.5781 .36558 L
.57829 .36555 L
.57848 .36552 L
.57867 .36549 L
.57886 .36546 L
.57905 .36543 L
.57924 .3654 L
.57943 .36537 L
.57962 .36534 L
Mistroke
.57981 .36531 L
.58 .36528 L
.58019 .36526 L
.58038 .36523 L
.58057 .3652 L
.58076 .36517 L
.58095 .36312 L
.58114 .36309 L
.58133 .36307 L
.58152 .36304 L
.58171 .36301 L
.5819 .36097 L
.5821 .36094 L
.58229 .36091 L
.58248 .36723 L
.58267 .3672 L
.58286 .36718 L
.58305 .37348 L
.58324 .37345 L
.58343 .37342 L
.58362 .37339 L
.58381 .37336 L
.584 .37334 L
.58419 .37331 L
.58438 .37328 L
.58457 .37325 L
.58476 .37322 L
.58495 .37319 L
.58514 .37316 L
.58533 .37313 L
.58552 .3731 L
.58571 .37307 L
.5859 .35287 L
.5861 .35284 L
.58629 .35281 L
.58648 .35279 L
.58667 .35276 L
.58686 .35273 L
.58705 .3527 L
.58724 .35267 L
.58743 .35068 L
.58762 .35065 L
.58781 .35062 L
.588 .34863 L
.58819 .3486 L
.58838 .32898 L
.58857 .32895 L
.58876 .33513 L
.58895 .33511 L
.58914 .33316 L
Mistroke
.58933 .33933 L
.58952 .3393 L
.58971 .33927 L
.5899 .33924 L
.5901 .3373 L
.59029 .33727 L
.59048 .33534 L
.59067 .33531 L
.59086 .33528 L
.59105 .33525 L
.59124 .33522 L
.59143 .3352 L
.59162 .33517 L
.59181 .33514 L
.592 .33511 L
.59219 .31599 L
.59238 .31597 L
.59257 .31594 L
.59276 .31591 L
.59295 .31588 L
.59314 .31586 L
.59333 .31583 L
.59352 .31394 L
.59371 .32003 L
.5939 .32 L
.5941 .31997 L
.59429 .31809 L
.59448 .31806 L
.59467 .31803 L
.59486 .31801 L
.59505 .32407 L
.59524 .32404 L
.59543 .32402 L
.59562 .32399 L
.59581 .32396 L
.596 .32394 L
.59619 .32391 L
.59638 .32388 L
.59657 .32385 L
.59676 .32383 L
.59695 .3238 L
.59714 .32377 L
.59733 .3235 L
.59752 .32348 L
.59771 .32159 L
.5979 .32156 L
.5981 .32154 L
.59829 .32758 L
.59848 .32755 L
.59867 .32753 L
Mistroke
.59886 .3275 L
.59905 .32561 L
.59924 .32559 L
.59943 .32556 L
.59962 .32553 L
.59981 .3255 L
.6 .32548 L
.60019 .32545 L
.60038 .32357 L
.60057 .32959 L
.60076 .32956 L
.60095 .32953 L
.60114 .32951 L
.60133 .32948 L
.60152 .32945 L
.60171 .32943 L
.6019 .32916 L
.6021 .32913 L
.60229 .3291 L
.60248 .32908 L
.60267 .32905 L
.60286 .32902 L
.60305 .329 L
.60324 .32897 L
.60343 .32894 L
.60362 .32891 L
.60381 .32889 L
.604 .32886 L
.60419 .32883 L
.60438 .32695 L
.60457 .32693 L
.60476 .3269 L
.60495 .32687 L
.60514 .32685 L
.60533 .32682 L
.60552 .32679 L
.60571 .32492 L
.6059 .32489 L
.6061 .32486 L
.60629 .32484 L
.60648 .32481 L
.60667 .32478 L
.60686 .32476 L
.60705 .32473 L
.60724 .32286 L
.60743 .32283 L
.60762 .32281 L
.60781 .32278 L
.608 .32275 L
.60819 .32273 L
Mistroke
.60838 .3227 L
.60857 .32869 L
.60876 .32866 L
.60895 .32863 L
.60914 .32861 L
.60933 .32858 L
.60952 .32855 L
.60971 .32853 L
.6099 .3285 L
.6101 .32847 L
.61029 .32845 L
.61048 .32842 L
.61067 .32839 L
.61086 .32837 L
.61105 .32834 L
.61124 .32831 L
.61143 .32829 L
.61162 .32826 L
.61181 .32823 L
.612 .3282 L
.61219 .32818 L
.61238 .32815 L
.61257 .32628 L
.61276 .32626 L
.61295 .32623 L
.61314 .3262 L
.61333 .32618 L
.61352 .32615 L
.61371 .32612 L
.6139 .3261 L
.6141 .32607 L
.61429 .32604 L
.61448 .32602 L
.61467 .32599 L
.61486 .32596 L
.61505 .32594 L
.61524 .32591 L
.61543 .32588 L
.61562 .32586 L
.61581 .32583 L
.616 .3258 L
.61619 .32578 L
.61638 .32575 L
.61657 .32572 L
.61676 .3257 L
.61695 .32567 L
.61714 .32381 L
.61733 .32378 L
.61752 .32375 L
.61771 .32373 L
Mistroke
.6179 .3237 L
.6181 .32184 L
.61829 .32181 L
.61848 .32179 L
.61867 .32176 L
.61886 .32173 L
.61905 .32171 L
.61924 .32168 L
.61943 .32764 L
.61962 .32761 L
.61981 .33355 L
.62 .33352 L
.62019 .33349 L
.62038 .33347 L
.62057 .33344 L
.62076 .33341 L
.62095 .33339 L
.62114 .33336 L
.62133 .33333 L
.62152 .33331 L
.62171 .33923 L
.6219 .3392 L
.6221 .33734 L
.62229 .33731 L
.62248 .33728 L
.62267 .33726 L
.62286 .33723 L
.62305 .33721 L
.62324 .33718 L
.62343 .33715 L
.62362 .33713 L
.62381 .3371 L
.624 .33707 L
.62419 .33705 L
.62438 .33702 L
.62457 .33699 L
.62476 .33697 L
.62495 .33694 L
.62514 .33691 L
.62533 .33689 L
.62552 .33686 L
.62571 .335 L
.6259 .33314 L
.6261 .33311 L
.62629 .33309 L
.62648 .33306 L
.62667 .33303 L
.62686 .33301 L
.62705 .33298 L
.62724 .33888 L
Mistroke
.62743 .33885 L
.62762 .33882 L
.62781 .3388 L
.628 .33694 L
.62819 .33691 L
.62838 .33689 L
.62857 .33686 L
.62876 .33683 L
.62895 .33681 L
.62914 .33678 L
.62933 .34265 L
.62952 .34263 L
.62971 .3426 L
.6299 .34257 L
.6301 .34255 L
.63029 .34252 L
.63048 .34249 L
.63067 .34247 L
.63086 .34244 L
.63105 .34242 L
.63124 .34239 L
.63143 .34236 L
.63162 .34234 L
.63181 .34231 L
.632 .34228 L
.63219 .34226 L
.63238 .34223 L
.63257 .3422 L
.63276 .34218 L
.63295 .34215 L
.63314 .34212 L
.63333 .3421 L
.63352 .34207 L
.63371 .34204 L
.6339 .34202 L
.6341 .34199 L
.63429 .34196 L
.63448 .34194 L
.63467 .34191 L
.63486 .34777 L
.63505 .34774 L
.63524 .34772 L
.63543 .34586 L
.63562 .34583 L
.63581 .3458 L
.636 .34578 L
.63619 .34575 L
.63638 .32739 L
.63657 .32557 L
.63676 .32555 L
Mistroke
.63695 .32552 L
.63714 .3255 L
.63733 .32547 L
.63752 .32544 L
.63771 .32542 L
.6379 .32539 L
.6381 .32537 L
.63829 .33116 L
.63848 .33113 L
.63867 .33111 L
.63886 .33108 L
.63905 .33106 L
.63924 .33103 L
.63943 .331 L
.63962 .33098 L
.63981 .33095 L
.64 .33093 L
.64019 .3309 L
.64038 .32909 L
.64057 .32728 L
.64076 .32725 L
.64095 .32723 L
.64114 .3272 L
.64133 .32717 L
.64152 .32715 L
.64171 .32712 L
.6419 .3271 L
.6421 .32707 L
.64229 .32705 L
.64248 .33281 L
.64267 .33279 L
.64286 .33276 L
.64305 .33274 L
.64324 .33271 L
.64343 .33269 L
.64362 .33266 L
.64381 .33841 L
.644 .33839 L
.64419 .33836 L
.64438 .33834 L
.64457 .33831 L
.64476 .34405 L
.64495 .34402 L
.64514 .344 L
.64533 .34397 L
.64552 .34395 L
.64571 .34392 L
.6459 .34389 L
.6461 .34387 L
.64629 .34384 L
Mistroke
.64648 .34382 L
.64667 .34379 L
.64686 .34952 L
.64705 .34949 L
.64724 .34946 L
.64743 .34944 L
.64762 .34941 L
.64781 .34939 L
.648 .34936 L
.64819 .34933 L
.64838 .34931 L
.64857 .34928 L
.64876 .34926 L
.64895 .34923 L
.64914 .3492 L
.64933 .34918 L
.64952 .34915 L
.64971 .34913 L
.6499 .3491 L
.6501 .34907 L
.65029 .34905 L
.65048 .34902 L
.65067 .349 L
.65086 .34897 L
.65105 .34894 L
.65124 .34892 L
.65143 .34889 L
.65162 .34887 L
.65181 .34884 L
.652 .34881 L
.65219 .34879 L
.65238 .34876 L
.65257 .34874 L
.65276 .34871 L
.65295 .34689 L
.65314 .34687 L
.65333 .34684 L
.65352 .34681 L
.65371 .34679 L
.6539 .34676 L
.6541 .34674 L
.65429 .34671 L
.65448 .34668 L
.65467 .34666 L
.65486 .34663 L
.65505 .34661 L
.65524 .34658 L
.65543 .34655 L
.65562 .34653 L
.65581 .32857 L
Mistroke
.656 .32855 L
.65619 .32852 L
.65638 .3285 L
.65657 .32847 L
.65676 .32845 L
.65695 .32842 L
.65714 .32839 L
.65733 .33406 L
.65752 .33403 L
.65771 .33401 L
.6579 .33398 L
.6581 .33396 L
.65829 .33393 L
.65848 .31632 L
.65867 .3163 L
.65886 .31627 L
.65905 .31625 L
.65924 .31623 L
.65943 .3162 L
.65962 .31618 L
.65981 .31615 L
.66 .31441 L
.66019 .31439 L
.66038 .31436 L
.66057 .31434 L
.66076 .31431 L
.66095 .31429 L
.66114 .31426 L
.66133 .31424 L
.66152 .31421 L
.66171 .31419 L
.6619 .31245 L
.6621 .31243 L
.66229 .3124 L
.66248 .31238 L
.66267 .31235 L
.66286 .31233 L
.66305 .3123 L
.66324 .31228 L
.66343 .31225 L
.66362 .31223 L
.66381 .3122 L
.664 .31218 L
.66419 .31778 L
.66438 .31605 L
.66457 .31602 L
.66476 .316 L
.66495 .31597 L
.66514 .31595 L
.66533 .31592 L
Mistroke
.66552 .3159 L
.66571 .31587 L
.6659 .31585 L
.6661 .31582 L
.66629 .3158 L
.66648 .31577 L
.66667 .31575 L
.66686 .31573 L
.66705 .3157 L
.66724 .31568 L
.66743 .31565 L
.66762 .31392 L
.66781 .3139 L
.668 .31387 L
.66819 .31385 L
.66838 .31382 L
.66857 .3138 L
.66876 .31377 L
.66895 .31375 L
.66914 .31372 L
.66933 .3137 L
.66952 .31367 L
.66971 .31365 L
.6699 .31362 L
.6701 .3136 L
.67029 .31357 L
.67048 .31915 L
.67067 .31913 L
.67086 .31911 L
.67105 .31908 L
.67124 .31906 L
.67143 .31903 L
.67162 .31901 L
.67181 .31898 L
.672 .31896 L
.67219 .31893 L
.67238 .31891 L
.67257 .31888 L
.67276 .31886 L
.67295 .32442 L
.67314 .3244 L
.67333 .32437 L
.67352 .32264 L
.67371 .32262 L
.6739 .32817 L
.6741 .32814 L
.67429 .32812 L
.67448 .32809 L
.67467 .32807 L
.67486 .3336 L
Mistroke
.67505 .33358 L
.67524 .33355 L
.67543 .33353 L
.67562 .3335 L
.67581 .33348 L
.676 .33345 L
.67619 .33343 L
.67638 .3334 L
.67657 .33338 L
.67676 .33335 L
.67695 .33333 L
.67714 .3333 L
.67733 .33328 L
.67752 .3388 L
.67771 .33878 L
.6779 .33875 L
.6781 .33873 L
.67829 .3387 L
.67848 .34421 L
.67867 .34418 L
.67886 .34416 L
.67905 .34413 L
.67924 .34411 L
.67943 .34408 L
.67962 .34406 L
.67981 .34403 L
.68 .34401 L
.68019 .34398 L
.68038 .34396 L
.68057 .34393 L
.68076 .34391 L
.68095 .34389 L
.68114 .34386 L
.68133 .34384 L
.68152 .32657 L
.68171 .32655 L
.6819 .32652 L
.6821 .3265 L
.68229 .32647 L
.68248 .32645 L
.68267 .32474 L
.68286 .32472 L
.68305 .3247 L
.68324 .32467 L
.68343 .32465 L
.68362 .32462 L
.68381 .30775 L
.684 .30773 L
.68419 .3077 L
.68438 .30768 L
Mistroke
.68457 .30766 L
.68476 .30763 L
.68495 .30596 L
.68514 .30594 L
.68533 .30592 L
.68552 .30589 L
.68571 .30587 L
.6859 .31128 L
.6861 .31126 L
.68629 .31124 L
.68648 .31121 L
.68667 .31119 L
.68686 .31116 L
.68705 .31114 L
.68724 .31112 L
.68743 .31109 L
.68762 .31649 L
.68781 .31647 L
.688 .31645 L
.68819 .31642 L
.68838 .3164 L
.68857 .31637 L
.68876 .32176 L
.68895 .32174 L
.68914 .32171 L
.68933 .32169 L
.68952 .32167 L
.68971 .32164 L
.6899 .32162 L
.6901 .32159 L
.69029 .32157 L
.69048 .32155 L
.69067 .31987 L
.69086 .31985 L
.69105 .31982 L
.69124 .3198 L
.69143 .31978 L
.69162 .31975 L
.69181 .31973 L
.692 .31971 L
.69219 .31804 L
.69238 .31801 L
.69257 .31799 L
.69276 .32335 L
.69295 .32333 L
.69314 .32331 L
.69333 .32866 L
.69352 .32864 L
.69371 .32861 L
.6939 .32859 L
Mistroke
.6941 .32857 L
.69429 .32854 L
.69448 .32852 L
.69467 .32849 L
.69486 .32847 L
.69505 .32845 L
.69524 .32842 L
.69543 .3284 L
.69562 .32837 L
.69581 .32835 L
.696 .32833 L
.69619 .3283 L
.69638 .33365 L
.69657 .33362 L
.69676 .3336 L
.69695 .33357 L
.69714 .33355 L
.69733 .33353 L
.69752 .3335 L
.69771 .33348 L
.6979 .33345 L
.6981 .33343 L
.69829 .33341 L
.69848 .33338 L
.69867 .33336 L
.69886 .33333 L
.69905 .33331 L
.69924 .33329 L
.69943 .31668 L
.69962 .32198 L
.69981 .32195 L
.7 .32193 L
.70019 .32028 L
.70038 .32026 L
.70057 .32024 L
.70076 .32021 L
.70095 .32019 L
.70114 .32017 L
.70133 .31852 L
.70152 .3185 L
.70171 .31848 L
.7019 .31845 L
.7021 .31843 L
.70229 .31841 L
.70248 .32368 L
.70267 .32894 L
.70286 .32892 L
.70305 .3289 L
.70324 .32887 L
.70343 .32885 L
Mistroke
.70362 .32883 L
.70381 .3288 L
.704 .33405 L
.70419 .33403 L
.70438 .33401 L
.70457 .33398 L
.70476 .33396 L
.70495 .33394 L
.70514 .33391 L
.70533 .33389 L
.70552 .33387 L
.70571 .33384 L
.7059 .33382 L
.7061 .3338 L
.70629 .33903 L
.70648 .33901 L
.70667 .33899 L
.70686 .33896 L
.70705 .33894 L
.70724 .33892 L
.70743 .33889 L
.70762 .33887 L
.70781 .33884 L
.708 .33882 L
.70819 .3388 L
.70838 .33715 L
.70857 .33712 L
.70876 .3371 L
.70895 .33545 L
.70914 .33543 L
.70933 .33541 L
.70952 .33538 L
.70971 .33536 L
.7099 .33533 L
.7101 .34055 L
.71029 .34053 L
.71048 .34051 L
.71067 .34048 L
.71086 .34046 L
.71105 .34044 L
.71124 .34041 L
.71143 .34039 L
.71162 .33874 L
.71181 .33872 L
.712 .3387 L
.71219 .33867 L
.71238 .33865 L
.71257 .33862 L
.71276 .3386 L
.71295 .33858 L
Mistroke
.71314 .33855 L
.71333 .33853 L
.71352 .33851 L
.71371 .33848 L
.7139 .33846 L
.7141 .33844 L
.71429 .33841 L
.71448 .33839 L
.71467 .33837 L
.71486 .33834 L
.71505 .3367 L
.71524 .33668 L
.71543 .33665 L
.71562 .33663 L
.71581 .33661 L
.716 .33658 L
.71619 .33656 L
.71638 .33654 L
.71657 .3349 L
.71676 .33487 L
.71695 .31866 L
.71714 .31864 L
.71733 .31861 L
.71752 .31859 L
.71771 .31857 L
.7179 .32373 L
.7181 .32371 L
.71829 .32368 L
.71848 .32366 L
.71867 .32364 L
.71886 .32362 L
.71905 .32359 L
.71924 .32874 L
.71943 .32872 L
.71962 .3287 L
.71981 .32867 L
.72 .32865 L
.72019 .32863 L
.72038 .32861 L
.72057 .32858 L
.72076 .32697 L
.72095 .32695 L
.72114 .32693 L
.72133 .3269 L
.72152 .33268 L
.72171 .33266 L
.7219 .33264 L
.7221 .33261 L
.72229 .33259 L
.72248 .33257 L
Mistroke
.72267 .33254 L
.72286 .33252 L
.72305 .3325 L
.72324 .33247 L
.72343 .3376 L
.72362 .33757 L
.72381 .33755 L
.724 .33753 L
.72419 .34264 L
.72438 .34774 L
.72457 .34772 L
.72476 .34769 L
.72495 .34767 L
.72514 .34765 L
.72533 .34762 L
.72552 .33157 L
.72571 .33155 L
.7259 .33153 L
.7261 .33151 L
.72629 .33148 L
.72648 .33146 L
.72667 .33144 L
.72686 .33142 L
.72705 .33139 L
.72724 .33137 L
.72743 .33135 L
.72762 .33133 L
.72781 .3313 L
.728 .33128 L
.72819 .33126 L
.72838 .33123 L
.72857 .33629 L
.72876 .33627 L
.72895 .33625 L
.72914 .33622 L
.72933 .3362 L
.72952 .33618 L
.72971 .33615 L
.7299 .33613 L
.7301 .33611 L
.73029 .33452 L
.73048 .3345 L
.73067 .33447 L
.73086 .33445 L
.73105 .33443 L
.73124 .33441 L
.73143 .33438 L
.73162 .33436 L
.73181 .33434 L
.732 .33431 L
Mistroke
.73219 .33429 L
.73238 .33933 L
.73257 .33931 L
.73276 .33929 L
.73295 .33927 L
.73314 .33924 L
.73333 .33922 L
.73352 .3392 L
.73371 .34486 L
.7339 .34483 L
.7341 .34481 L
.73429 .34479 L
.73448 .34477 L
.73467 .34474 L
.73486 .34472 L
.73505 .3447 L
.73524 .34467 L
.73543 .34465 L
.73562 .34463 L
.73581 .34461 L
.736 .34458 L
.73619 .34456 L
.73638 .34454 L
.73657 .34295 L
.73676 .34292 L
.73695 .3429 L
.73714 .34288 L
.73733 .34285 L
.73752 .34283 L
.73771 .34281 L
.7379 .34279 L
.7381 .34276 L
.73829 .34274 L
.73848 .34272 L
.73867 .3427 L
.73886 .34267 L
.73905 .34265 L
.73924 .34263 L
.73943 .3426 L
.73962 .34258 L
.73981 .34256 L
.74 .34254 L
.74019 .34251 L
.74038 .34249 L
.74057 .34247 L
.74076 .34245 L
.74095 .34242 L
.74114 .3424 L
.74133 .34238 L
.74152 .34235 L
Mistroke
.74171 .34233 L
.7419 .34231 L
.7421 .34229 L
.74229 .34226 L
.74248 .34224 L
.74267 .34222 L
.74286 .3422 L
.74305 .34721 L
.74324 .34719 L
.74343 .34716 L
.74362 .34714 L
.74381 .35214 L
.744 .35212 L
.74419 .3521 L
.74438 .35208 L
.74457 .35205 L
.74476 .35203 L
.74495 .35201 L
.74514 .35198 L
.74533 .35196 L
.74552 .35194 L
.74571 .35192 L
.7459 .35189 L
.7461 .35187 L
.74629 .35185 L
.74648 .35182 L
.74667 .3518 L
.74686 .35178 L
.74705 .35176 L
.74724 .35173 L
.74743 .35171 L
.74762 .35169 L
.74781 .35009 L
.748 .35007 L
.74819 .35005 L
.74838 .33429 L
.74857 .33426 L
.74876 .33424 L
.74895 .33422 L
.74914 .3342 L
.74933 .33417 L
.74952 .33415 L
.74971 .33413 L
.7499 .33411 L
.7501 .33409 L
.75029 .33406 L
.75048 .33404 L
.75067 .33402 L
.75086 .334 L
.75105 .33397 L
Mistroke
.75124 .33395 L
.75143 .33393 L
.75162 .33889 L
.75181 .33886 L
.752 .33884 L
.75219 .33882 L
.75238 .3388 L
.75257 .33877 L
.75276 .33875 L
.75295 .33873 L
.75314 .33871 L
.75333 .33868 L
.75352 .33866 L
.75371 .33864 L
.7539 .33862 L
.7541 .34356 L
.75429 .34354 L
.75448 .34352 L
.75467 .3435 L
.75486 .34347 L
.75505 .34345 L
.75524 .34343 L
.75543 .34186 L
.75562 .3468 L
.75581 .34677 L
.756 .34675 L
.75619 .34673 L
.75638 .34671 L
.75657 .35163 L
.75676 .35161 L
.75695 .35158 L
.75714 .35156 L
.75733 .35154 L
.75752 .35152 L
.75771 .35149 L
.7579 .35147 L
.7581 .35145 L
.75829 .35143 L
.75848 .3514 L
.75867 .35138 L
.75886 .35136 L
.75905 .35134 L
.75924 .35131 L
.75943 .35129 L
.75962 .34973 L
.75981 .3497 L
.76 .34968 L
.76019 .34966 L
.76038 .34964 L
.76057 .34961 L
Mistroke
.76076 .34959 L
.76095 .34957 L
.76114 .34955 L
.76133 .34952 L
.76152 .3495 L
.76171 .34794 L
.7619 .34638 L
.7621 .34636 L
.76229 .34633 L
.76248 .34631 L
.76267 .34475 L
.76286 .34473 L
.76305 .34471 L
.76324 .34469 L
.76343 .34467 L
.76362 .34464 L
.76381 .34462 L
.764 .3446 L
.76419 .34458 L
.76438 .34455 L
.76457 .34453 L
.76476 .34451 L
.76495 .34449 L
.76514 .34447 L
.76533 .34291 L
.76552 .34289 L
.76571 .34287 L
.7659 .34284 L
.7661 .34129 L
.76629 .34127 L
.76648 .34125 L
.76667 .34123 L
.76686 .34121 L
.76705 .34118 L
.76724 .34116 L
.76743 .34114 L
.76762 .34112 L
.76781 .34601 L
.768 .34599 L
.76819 .34596 L
.76838 .34441 L
.76857 .34439 L
.76876 .34437 L
.76895 .34435 L
.76914 .34432 L
.76933 .32898 L
.76952 .32895 L
.76971 .32893 L
.7699 .32891 L
.7701 .32889 L
Mistroke
.77029 .32887 L
.77048 .32885 L
.77067 .32882 L
.77086 .3288 L
.77105 .32878 L
.77124 .32876 L
.77143 .32874 L
.77162 .32872 L
.77181 .32869 L
.772 .32867 L
.77219 .33413 L
.77238 .3341 L
.77257 .33408 L
.77276 .33406 L
.77295 .33254 L
.77314 .33102 L
.77333 .331 L
.77352 .31596 L
.77371 .31594 L
.7739 .31592 L
.7741 .3159 L
.77429 .31441 L
.77448 .31439 L
.77467 .31436 L
.77486 .31434 L
.77505 .31432 L
.77524 .3143 L
.77543 .31428 L
.77562 .29954 L
.77581 .29952 L
.776 .2995 L
.77619 .29948 L
.77638 .29946 L
.77657 .29944 L
.77676 .29942 L
.77695 .2994 L
.77714 .29938 L
.77733 .29935 L
.77752 .29933 L
.77771 .29931 L
.7779 .29929 L
.7781 .29927 L
.77829 .29781 L
.77848 .29779 L
.77867 .30256 L
.77886 .28809 L
.77905 .28807 L
.77924 .28805 L
.77943 .28803 L
.77962 .28801 L
Mistroke
.77981 .29274 L
.78 .29272 L
.78019 .2927 L
.78038 .29268 L
.78057 .29265 L
.78076 .29263 L
.78095 .29261 L
.78114 .29259 L
.78133 .29257 L
.78152 .29255 L
.78171 .29727 L
.7819 .29725 L
.7821 .29723 L
.78229 .29721 L
.78248 .29719 L
.78267 .29717 L
.78286 .29715 L
.78305 .29713 L
.78324 .29711 L
.78343 .29709 L
.78362 .29707 L
.78381 .29705 L
.784 .29703 L
.78419 .29701 L
.78438 .29557 L
.78457 .29555 L
.78476 .29553 L
.78495 .29551 L
.78514 .29549 L
.78533 .29547 L
.78552 .29545 L
.78571 .28121 L
.7859 .28119 L
.7861 .27978 L
.78629 .27976 L
.78648 .27974 L
.78667 .27972 L
.78686 .2797 L
.78705 .27968 L
.78724 .27966 L
.78743 .27964 L
.78762 .27962 L
.78781 .2796 L
.788 .27958 L
.78819 .27956 L
.78838 .27954 L
.78857 .27952 L
.78876 .2795 L
.78895 .27948 L
.78914 .27946 L
Mistroke
.78933 .27944 L
.78952 .27942 L
.78971 .2794 L
.7899 .27799 L
.7901 .27797 L
.79029 .27795 L
.79048 .27793 L
.79067 .27791 L
.79086 .27789 L
.79105 .27787 L
.79124 .27785 L
.79143 .27783 L
.79162 .32989 L
.79181 .32987 L
.792 .32985 L
.79219 .32983 L
.79238 .32981 L
.79257 .32979 L
.79276 .32835 L
.79295 .32833 L
.79314 .3329 L
.79333 .33288 L
.79352 .33286 L
.79371 .33284 L
.7939 .33282 L
.7941 .3328 L
.79429 .33278 L
.79448 .33276 L
.79467 .33274 L
.79486 .33272 L
.79505 .3327 L
.79524 .33267 L
.79543 .33265 L
.79562 .33263 L
.79581 .33261 L
.796 .33259 L
.79619 .33257 L
.79638 .33255 L
.79657 .33253 L
.79676 .33251 L
.79695 .33249 L
.79714 .33106 L
.79733 .33104 L
.79752 .33102 L
.79771 .331 L
.7979 .33098 L
.7981 .32955 L
.79829 .32953 L
.79848 .31536 L
.79867 .31534 L
Mistroke
.79886 .31532 L
.79905 .3153 L
.79924 .31528 L
.79943 .31526 L
.79962 .31524 L
.79981 .31522 L
.8 .3152 L
.80019 .31518 L
.80038 .31516 L
.80057 .31514 L
.80076 .31512 L
.80095 .3151 L
.80114 .31508 L
.80133 .31506 L
.80152 .31504 L
.80171 .31502 L
.8019 .32011 L
.8021 .32463 L
.80229 .32461 L
.80248 .32459 L
.80267 .32457 L
.80286 .32455 L
.80305 .32453 L
.80324 .32312 L
.80343 .3231 L
.80362 .32308 L
.80381 .32306 L
.804 .32304 L
.80419 .32302 L
.80438 .323 L
.80457 .32298 L
.80476 .32296 L
.80495 .32294 L
.80514 .32292 L
.80533 .3229 L
.80552 .32741 L
.80571 .32739 L
.8059 .32737 L
.8061 .32735 L
.80629 .32733 L
.80648 .32731 L
.80667 .32729 L
.80686 .32727 L
.80705 .32725 L
.80724 .32723 L
.80743 .32721 L
.80762 .32719 L
.80781 .32717 L
.808 .32715 L
.80819 .32713 L
Mistroke
.80838 .32711 L
.80857 .32709 L
.80876 .32707 L
.80895 .32705 L
.80914 .32703 L
.80933 .32701 L
.80952 .32699 L
.80971 .32697 L
.8099 .32695 L
.8101 .32693 L
.81029 .32691 L
.81048 .32689 L
.81067 .32687 L
.81086 .37144 L
.81105 .37142 L
.81124 .3714 L
.81143 .37138 L
.81162 .37136 L
.81181 .37134 L
.812 .37132 L
.81219 .3713 L
.81238 .37128 L
.81257 .36985 L
.81276 .36983 L
.81295 .36981 L
.81314 .36979 L
.81333 .37419 L
.81352 .37417 L
.81371 .37415 L
.8139 .37855 L
.8141 .37853 L
.81429 .37851 L
.81448 .37849 L
.81467 .38288 L
.81486 .38286 L
.81505 .38284 L
.81524 .38282 L
.81543 .3828 L
.81562 .38278 L
.81581 .38276 L
.816 .38274 L
.81619 .38272 L
.81638 .38269 L
.81657 .38267 L
.81676 .38265 L
.81695 .38263 L
.81714 .38261 L
.81733 .38259 L
.81752 .38257 L
.81771 .38255 L
Mistroke
.8179 .38253 L
.8181 .3811 L
.81829 .38108 L
.81848 .38106 L
.81867 .38104 L
.81886 .38102 L
.81905 .38099 L
.81924 .38537 L
.81943 .38535 L
.81962 .38533 L
.81981 .38531 L
.82 .38529 L
.82019 .38527 L
.82038 .38525 L
.82057 .38962 L
.82076 .38819 L
.82095 .38817 L
.82114 .38814 L
.82133 .38671 L
.82152 .38669 L
.82171 .38667 L
.8219 .38524 L
.8221 .38522 L
.82229 .3852 L
.82248 .38518 L
.82267 .38516 L
.82286 .38514 L
.82305 .38512 L
.82324 .3851 L
.82343 .38508 L
.82362 .38506 L
.82381 .38504 L
.824 .38361 L
.82419 .38359 L
.82438 .38357 L
.82457 .38355 L
.82476 .38353 L
.82495 .38351 L
.82514 .38208 L
.82533 .38206 L
.82552 .38204 L
.82571 .38202 L
.8259 .38637 L
.8261 .38635 L
.82629 .38633 L
.82648 .3849 L
.82667 .38488 L
.82686 .38486 L
.82705 .38484 L
.82724 .38482 L
Mistroke
.82743 .3848 L
.82762 .38914 L
.82781 .38912 L
.828 .3891 L
.82819 .38908 L
.82838 .38906 L
.82857 .38904 L
.82876 .38902 L
.82895 .389 L
.82914 .38898 L
.82933 .39331 L
.82952 .39329 L
.82971 .39327 L
.8299 .39325 L
.8301 .39323 L
.83029 .39321 L
.83048 .39319 L
.83067 .39317 L
.83086 .39315 L
.83105 .39172 L
.83124 .3917 L
.83143 .39168 L
.83162 .39166 L
.83181 .39164 L
.832 .39162 L
.83219 .3916 L
.83238 .39158 L
.83257 .39156 L
.83276 .39154 L
.83295 .39152 L
.83314 .3915 L
.83333 .39147 L
.83352 .39145 L
.83371 .39143 L
.8339 .39575 L
.8341 .39573 L
.83429 .39571 L
.83448 .39551 L
.83467 .39549 L
.83486 .39547 L
.83505 .39545 L
.83524 .39543 L
.83543 .394 L
.83562 .39398 L
.83581 .39396 L
.836 .39394 L
.83619 .39252 L
.83638 .3925 L
.83657 .39248 L
.83676 .39246 L
Mistroke
.83695 .39244 L
.83714 .39242 L
.83733 .3924 L
.83752 .3967 L
.83771 .39668 L
.8379 .39666 L
.8381 .39664 L
.83829 .39662 L
.83848 .3966 L
.83867 .39658 L
.83886 .39656 L
.83905 .39654 L
.83924 .39652 L
.83943 .3965 L
.83962 .39508 L
.83981 .39506 L
.84 .39504 L
.84019 .39502 L
.84038 .395 L
.84057 .39929 L
.84076 .39927 L
.84095 .39925 L
.84114 .39783 L
.84133 .39781 L
.84152 .38372 L
.84171 .38232 L
.8419 .3823 L
.8421 .38656 L
.84229 .38654 L
.84248 .38652 L
.84267 .3865 L
.84286 .37264 L
.84305 .37262 L
.84324 .3726 L
.84343 .37258 L
.84362 .37256 L
.84381 .37254 L
.844 .37253 L
.84419 .37251 L
.84438 .37249 L
.84457 .37247 L
.84476 .37245 L
.84495 .37243 L
.84514 .37241 L
.84533 .37239 L
.84552 .37237 L
.84571 .37235 L
.8459 .37233 L
.8461 .37231 L
.84629 .37229 L
Mistroke
.84648 .37227 L
.84667 .37225 L
.84686 .37223 L
.84705 .37646 L
.84724 .37644 L
.84743 .37642 L
.84762 .3764 L
.84781 .37638 L
.848 .37636 L
.84819 .37634 L
.84838 .37632 L
.84857 .3763 L
.84876 .37628 L
.84895 .37626 L
.84914 .38049 L
.84933 .38047 L
.84952 .38045 L
.84971 .37907 L
.8499 .37905 L
.8501 .37903 L
.85029 .37901 L
.85048 .37899 L
.85067 .37897 L
.85086 .37895 L
.85105 .37893 L
.85124 .37891 L
.85143 .37889 L
.85162 .37887 L
.85181 .37885 L
.852 .37884 L
.85219 .37882 L
.85238 .38303 L
.85257 .38301 L
.85276 .38299 L
.85295 .38297 L
.85314 .3693 L
.85333 .36929 L
.85352 .36927 L
.85371 .36925 L
.8539 .36923 L
.8541 .36921 L
.85429 .36919 L
.85448 .36917 L
.85467 .36915 L
.85486 .36913 L
.85505 .36911 L
.85524 .36909 L
.85543 .36907 L
.85562 .36905 L
.85581 .36903 L
Mistroke
.856 .36901 L
.85619 .36899 L
.85638 .36898 L
.85657 .36762 L
.85676 .3676 L
.85695 .36758 L
.85714 .36756 L
.85733 .36754 L
.85752 .37172 L
.85771 .37171 L
.8579 .37169 L
.8581 .37167 L
.85829 .37165 L
.85848 .37163 L
.85867 .37161 L
.85886 .37159 L
.85905 .37157 L
.85924 .37155 L
.85943 .37153 L
.85962 .37151 L
.85981 .37149 L
.86 .37147 L
.86019 .37145 L
.86038 .37143 L
.86057 .37142 L
.86076 .3714 L
.86095 .37138 L
.86114 .37555 L
.86133 .37553 L
.86152 .37551 L
.86171 .37549 L
.8619 .37547 L
.8621 .37545 L
.86229 .37543 L
.86248 .3796 L
.86267 .37958 L
.86286 .37822 L
.86305 .3782 L
.86324 .37818 L
.86343 .37816 L
.86362 .37814 L
.86381 .37812 L
.864 .3781 L
.86419 .37809 L
.86438 .37807 L
.86457 .37805 L
.86476 .37803 L
.86495 .37667 L
.86514 .37665 L
.86533 .37663 L
Mistroke
.86552 .37528 L
.86571 .37526 L
.8659 .37391 L
.8661 .37389 L
.86629 .37387 L
.86648 .37385 L
.86667 .378 L
.86686 .37798 L
.86705 .37796 L
.86724 .37794 L
.86743 .37793 L
.86762 .37791 L
.86781 .37789 L
.868 .38203 L
.86819 .38201 L
.86838 .38199 L
.86857 .38197 L
.86876 .38195 L
.86895 .38193 L
.86914 .38191 L
.86933 .38189 L
.86952 .38187 L
.86971 .38185 L
.8699 .38183 L
.8701 .38181 L
.87029 .3818 L
.87048 .38178 L
.87067 .38176 L
.87086 .36834 L
.87105 .36832 L
.87124 .3683 L
.87143 .36828 L
.87162 .36826 L
.87181 .36824 L
.872 .36822 L
.87219 .3682 L
.87238 .36818 L
.87257 .36817 L
.87276 .36815 L
.87295 .36813 L
.87314 .36811 L
.87333 .36809 L
.87352 .36807 L
.87371 .36805 L
.8739 .36803 L
.8741 .37214 L
.87429 .37625 L
.87448 .37623 L
.87467 .37621 L
.87486 .37619 L
Mistroke
.87505 .37617 L
.87524 .37615 L
.87543 .37613 L
.87562 .37612 L
.87581 .3761 L
.876 .37608 L
.87619 .37606 L
.87638 .37604 L
.87657 .37602 L
.87676 .376 L
.87695 .37598 L
.87714 .37596 L
.87733 .37594 L
.87752 .37592 L
.87771 .37591 L
.8779 .37589 L
.8781 .37587 L
.87829 .37585 L
.87848 .37583 L
.87867 .37581 L
.87886 .37579 L
.87905 .37577 L
.87924 .37575 L
.87943 .37573 L
.87962 .37572 L
.87981 .3757 L
.88 .37568 L
.88019 .36244 L
.88038 .36242 L
.88057 .3624 L
.88076 .36238 L
.88095 .36236 L
.88114 .36235 L
.88133 .36233 L
.88152 .36231 L
.88171 .36229 L
.8819 .36227 L
.8821 .36225 L
.88229 .36223 L
.88248 .36221 L
.88267 .3622 L
.88286 .36218 L
.88305 .36216 L
.88324 .36214 L
.88343 .36212 L
.88362 .3621 L
.88381 .36208 L
.884 .36206 L
.88419 .36205 L
.88438 .36203 L
Mistroke
.88457 .36201 L
.88476 .36199 L
.88495 .36068 L
.88514 .36066 L
.88533 .36064 L
.88552 .36471 L
.88571 .36469 L
.8859 .36468 L
.8861 .36336 L
.88629 .36335 L
.88648 .36333 L
.88667 .36331 L
.88686 .36329 L
.88705 .36327 L
.88724 .36196 L
.88743 .36194 L
.88762 .36192 L
.88781 .36191 L
.888 .36189 L
.88819 .36187 L
.88838 .36185 L
.88857 .36183 L
.88876 .36181 L
.88895 .36179 L
.88914 .36177 L
.88933 .36176 L
.88952 .36174 L
.88971 .36172 L
.8899 .3617 L
.8901 .36168 L
.89029 .36166 L
.89048 .36164 L
.89067 .36163 L
.89086 .36161 L
.89105 .36159 L
.89124 .36157 L
.89143 .36155 L
.89162 .36153 L
.89181 .36151 L
.892 .36149 L
.89219 .36148 L
.89238 .36146 L
.89257 .36144 L
.89276 .36142 L
.89295 .3614 L
.89314 .36138 L
.89333 .36136 L
.89352 .36135 L
.89371 .36133 L
.8939 .36131 L
Mistroke
.8941 .36129 L
.89429 .36127 L
.89448 .36125 L
.89467 .36123 L
.89486 .36121 L
.89505 .3612 L
.89524 .36118 L
.89543 .36116 L
.89562 .36114 L
.89581 .36112 L
.896 .35981 L
.89619 .3598 L
.89638 .35978 L
.89657 .35976 L
.89676 .35974 L
.89695 .35972 L
.89714 .3597 L
.89733 .35968 L
.89752 .35967 L
.89771 .35965 L
.8979 .35963 L
.8981 .35961 L
.89829 .35959 L
.89848 .36365 L
.89867 .3677 L
.89886 .36768 L
.89905 .36766 L
.89924 .36764 L
.89943 .36763 L
.89962 .36761 L
.89981 .36759 L
.9 .36757 L
.90019 .36755 L
.90038 .36753 L
.90057 .36751 L
.90076 .3675 L
.90095 .36748 L
.90114 .36746 L
.90133 .36615 L
.90152 .36613 L
.90171 .36611 L
.9019 .36609 L
.9021 .36608 L
.90229 .36606 L
.90248 .36604 L
.90267 .36602 L
.90286 .366 L
.90305 .36598 L
.90324 .36596 L
.90343 .36595 L
Mistroke
.90362 .36593 L
.90381 .36997 L
.904 .36995 L
.90419 .36993 L
.90438 .36991 L
.90457 .36989 L
.90476 .36987 L
.90495 .36986 L
.90514 .36984 L
.90533 .36982 L
.90552 .3698 L
.90571 .36978 L
.9059 .36976 L
.9061 .36846 L
.90629 .36844 L
.90648 .36842 L
.90667 .3684 L
.90686 .36838 L
.90705 .36836 L
.90724 .36834 L
.90743 .36833 L
.90762 .36831 L
.90781 .36829 L
.908 .37232 L
.90819 .3723 L
.90838 .37228 L
.90857 .37226 L
.90876 .37225 L
.90895 .37223 L
.90914 .37221 L
.90933 .37219 L
.90952 .37217 L
.90971 .37215 L
.9099 .37213 L
.9101 .37211 L
.91029 .3721 L
.91048 .37208 L
.91067 .37206 L
.91086 .37204 L
.91105 .37202 L
.91124 .372 L
.91143 .37198 L
.91162 .37197 L
.91181 .37195 L
.912 .37193 L
.91219 .37191 L
.91238 .37189 L
.91257 .37187 L
.91276 .37185 L
.91295 .37183 L
Mistroke
.91314 .37182 L
.91333 .3718 L
.91352 .37178 L
.91371 .3758 L
.9139 .37578 L
.9141 .37577 L
.91429 .37575 L
.91448 .37573 L
.91467 .37571 L
.91486 .37569 L
.91505 .37567 L
.91524 .37565 L
.91543 .37563 L
.91562 .37562 L
.91581 .3756 L
.916 .37558 L
.91619 .37556 L
.91638 .37554 L
.91657 .37552 L
.91676 .3755 L
.91695 .37548 L
.91714 .37418 L
.91733 .37416 L
.91752 .37414 L
.91771 .37412 L
.9179 .3741 L
.9181 .37408 L
.91829 .37407 L
.91848 .37405 L
.91867 .37403 L
.91886 .37401 L
.91905 .37399 L
.91924 .37397 L
.91943 .37395 L
.91962 .37265 L
.91981 .37263 L
.92 .37261 L
.92019 .35968 L
.92038 .35966 L
.92057 .35964 L
.92076 .35962 L
.92095 .3596 L
.92114 .35959 L
.92133 .35957 L
.92152 .35955 L
.92171 .35953 L
.9219 .35951 L
.9221 .35949 L
.92229 .35948 L
.92248 .35946 L
Mistroke
.92267 .35944 L
.92286 .3467 L
.92305 .34668 L
.92324 .34667 L
.92343 .34665 L
.92362 .34663 L
.92381 .34661 L
.924 .34659 L
.92419 .35057 L
.92438 .35055 L
.92457 .35053 L
.92476 .35051 L
.92495 .3505 L
.92514 .35048 L
.92533 .35046 L
.92552 .35044 L
.92571 .35042 L
.9259 .3504 L
.9261 .35437 L
.92629 .35435 L
.92648 .35433 L
.92667 .35432 L
.92686 .3543 L
.92705 .35428 L
.92724 .35426 L
.92743 .35424 L
.92762 .35423 L
.92781 .35421 L
.928 .35419 L
.92819 .35417 L
.92838 .35415 L
.92857 .35414 L
.92876 .35412 L
.92895 .35285 L
.92914 .35283 L
.92933 .35282 L
.92952 .3528 L
.92971 .35278 L
.9299 .35276 L
.9301 .35274 L
.93029 .3567 L
.93048 .35668 L
.93067 .35666 L
.93086 .36061 L
.93105 .3606 L
.93124 .36058 L
.93143 .36056 L
.93162 .36054 L
.93181 .36052 L
.932 .3605 L
Mistroke
.93219 .36049 L
.93238 .36047 L
.93257 .3592 L
.93276 .35918 L
.93295 .35916 L
.93314 .35915 L
.93333 .35913 L
.93352 .35911 L
.93371 .35909 L
.9339 .35907 L
.9341 .35906 L
.93429 .35904 L
.93448 .35902 L
.93467 .359 L
.93486 .35898 L
.93505 .35897 L
.93524 .35895 L
.93543 .35893 L
.93562 .35891 L
.93581 .35889 L
.936 .35888 L
.93619 .35869 L
.93638 .35868 L
.93657 .35866 L
.93676 .35864 L
.93695 .35862 L
.93714 .3586 L
.93733 .36254 L
.93752 .36253 L
.93771 .36251 L
.9379 .36249 L
.9381 .36247 L
.93829 .36121 L
.93848 .36119 L
.93867 .36117 L
.93886 .36115 L
.93905 .36113 L
.93924 .36111 L
.93943 .3611 L
.93962 .36108 L
.93981 .36106 L
.94 .36104 L
.94019 .36497 L
.94038 .36496 L
.94057 .36494 L
.94076 .36492 L
.94095 .3649 L
.94114 .36488 L
.94133 .36487 L
.94152 .36485 L
Mistroke
.94171 .36483 L
.9419 .36481 L
.9421 .36479 L
.94229 .36477 L
.94248 .3522 L
.94267 .35218 L
.94286 .35217 L
.94305 .35215 L
.94324 .35213 L
.94343 .35211 L
.94362 .35209 L
.94381 .35208 L
.944 .35206 L
.94419 .35204 L
.94438 .35202 L
.94457 .352 L
.94476 .35199 L
.94495 .35197 L
.94514 .35195 L
.94533 .35193 L
.94552 .35192 L
.94571 .35067 L
.9459 .35065 L
.9461 .35063 L
.94629 .35454 L
.94648 .35452 L
.94667 .3545 L
.94686 .35448 L
.94705 .35447 L
.94724 .35445 L
.94743 .35443 L
.94762 .35441 L
.94781 .35439 L
.948 .35438 L
.94819 .35436 L
.94838 .34198 L
.94857 .34196 L
.94876 .34194 L
.94895 .34582 L
.94914 .3458 L
.94933 .34579 L
.94952 .34577 L
.94971 .34575 L
.9499 .34573 L
.9501 .34571 L
.95029 .3457 L
.95048 .34568 L
.95067 .34566 L
.95086 .34564 L
.95105 .34563 L
Mistroke
.95124 .34561 L
.95143 .34559 L
.95162 .34557 L
.95181 .34556 L
.952 .34554 L
.95219 .34552 L
.95238 .34429 L
.95257 .34427 L
.95276 .34426 L
.95295 .34424 L
.95314 .34422 L
.95333 .3442 L
.95352 .34419 L
.95371 .34417 L
.9539 .34294 L
.9541 .34292 L
.95429 .34291 L
.95448 .34289 L
.95467 .34287 L
.95486 .34285 L
.95505 .34284 L
.95524 .34282 L
.95543 .3428 L
.95562 .34158 L
.95581 .34156 L
.956 .34154 L
.95619 .34152 L
.95638 .34151 L
.95657 .34149 L
.95676 .34147 L
.95695 .34145 L
.95714 .34144 L
.95733 .34142 L
.95752 .3414 L
.95771 .34527 L
.9579 .34525 L
.9581 .34402 L
.95829 .34401 L
.95848 .34399 L
.95867 .34397 L
.95886 .34395 L
.95905 .34394 L
.95924 .34392 L
.95943 .3439 L
.95962 .34388 L
.95981 .34387 L
.96 .34385 L
.96019 .34263 L
.96038 .34261 L
.96057 .34259 L
Mistroke
.96076 .34257 L
.96095 .34256 L
.96114 .34254 L
.96133 .34252 L
.96152 .3425 L
.96171 .34249 L
.9619 .34247 L
.9621 .34245 L
.96229 .34243 L
.96248 .34629 L
.96267 .34627 L
.96286 .34625 L
.96305 .34624 L
.96324 .34622 L
.96343 .3462 L
.96362 .34498 L
.96381 .34496 L
.964 .34881 L
.96419 .35265 L
.96438 .35263 L
.96457 .35261 L
.96476 .35259 L
.96495 .35258 L
.96514 .35256 L
.96533 .35254 L
.96552 .35252 L
.96571 .35251 L
.9659 .35249 L
.9661 .35247 L
.96629 .35245 L
.96648 .35244 L
.96667 .35242 L
.96686 .3524 L
.96705 .35238 L
.96724 .35237 L
.96743 .35235 L
.96762 .35618 L
.96781 .35616 L
.968 .35615 L
.96819 .35613 L
.96838 .35611 L
.96857 .35609 L
.96876 .35608 L
.96895 .35606 L
.96914 .35604 L
.96933 .35987 L
.96952 .35985 L
.96971 .35983 L
.9699 .35982 L
.9701 .3598 L
Mistroke
.97029 .35978 L
.97048 .35976 L
.97067 .35975 L
.97086 .35973 L
.97105 .35971 L
.97124 .35969 L
.97143 .35968 L
.97162 .35966 L
.97181 .35964 L
.972 .35962 L
.97219 .35961 L
.97238 .35959 L
.97257 .35957 L
.97276 .35955 L
.97295 .35954 L
.97314 .34735 L
.97333 .34733 L
.97352 .34731 L
.97371 .34729 L
.9739 .34728 L
.9741 .35108 L
.97429 .34987 L
.97448 .34985 L
.97467 .34983 L
.97486 .34982 L
.97505 .3498 L
.97524 .34978 L
.97543 .34976 L
.97562 .34975 L
.97581 .34973 L
.976 .34971 L
.97619 .3497 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`000?ojl2@SS_oG
001[0?ooAm?Cdoojn[34g_XC@4J2]<c<c;Ng]fIVI_Cdm2m?CfUYJKjn_/30`;C^]:JVYZf][Jo^k/c8
`8b:Q;no_eICDe5Anb:;8^KViW=cL`00oiT0I_l003_j=020P80008200?oo0820P6IVc3<cIP2009VI
od:JYb8R8_mV<`00P03oo`3o000g??mVc6=SHd92@]kNgYBDU;fm_IZDU>GPhFaXJ;8R8P00Rld009VI
V@2;000i0?o<oom0o`1^0000?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o
?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo
_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo
_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0
_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOknookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo
?oooOooo_mWEc:0Pl=WIfJj^[WMgMiW<09W<<iW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc
0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI0?nI<ol08@000?l08@000?l08@000?l08@000?l0
8@000?l08@000?l08@000?l08@0001/000<10040o`0300006`000`400@3o00<0000K00030@010?l0
0`0001/000<10040o`0300006`000`400@3o00<0000K00030@010?l00`0001/00`7o00<0000K0003
0@010?l00`0001/000<10040o`0300006`000`400@3o00<0000K00030@010?l00`0001/000<10040
o`0300006`000`400@3o00<00002008110000`40000200@10P040@@000<10040o`03000000D00@00
0@0800030@0000<000<100001@030Ol00`000005004000402@000`40000300030@0000@000<10040
o`03000000D00@000@0:00030@0000<000<100000`000`400@3o00<000001@01000100/000<10000
0`000`40000200030@010?l00`0000050040004020001040004200040@000@@000<10040o`030000
0P020@X00P4400811@000`400@3o00<0000K00030@010?l00`0001/000@10041L0000`40002>0000
6`040FX000<100000P020I00000K00040@010FX000<100000P020@03004008d0000K00040@010FX0
00<100000P060H`0000K00040@010FX000<100000P070H/0000K00040@010FX000H1004100420004
0@010H/0000K00040@010FT00P401@010@0100D00P6:00006`0010400@5Y008100<00@400`440081
RP0001/000@10041J@020@03004100811P000`40002800006`040FT010400`01000600<1R00001/0
00@10041J0020@03004100T01@4300030@000800000K00040@010FP00P400`01000:00@10`000`40
002000006`0010400@5X008100<00@002P020@030040008000<10000P00001/000@10041J0020@03
004000d000D100000@2200006`0010400@5W00814@000`4000020H80000K00040@010FL00P4A0003
0@000081PP0001/000@10041I`020A4000<100000P6200006`040FL00P4A00030@000081PP0001/0
00@10041I`000`40000@00030@010081PP0001/000@10041I`000`40000@00050@010040PP0001/0
00@10041I`000`40000@00050@010040PP0001/000@10041@@000`40000S00030@0001400P400`01
002100006`0010400@50008190020A<00P400`01002100006`0010400@5000818`020A@00P400`01
002100000P020@@000<100000P040@@000<100000`0010400@5000818`000`40000C008100<00@00
P@0000050040004020000`40000500030@0000<000@10041@0020B<000<100005P000`4000200000
00D00@000@0900040@0000D110040D000P4S00030@0001H000<10000P0000005004000402P001040
004200030@0000<000@10041@0020B<000<100005P000`400020000000D00@000@0;00060@000@01
1@0010400@5000818`000`40000F00030@00080000001@01000100P000@100010`020@D000@10041
@0050B0000<100005P000`40002000000P020@X00P4500030@0000<000@10041@0040B4000<10000
5P000`40002000006`0010400@4o00810P000`40000O00030@0001H000<10000P00001/000@10041
?`020@8000<100007`000`40000F00030@000800000K00040@010Cl00P42008180000`40000F0003
0@000800000K008100<00@00?P020@<000<100004P000`40000900030@0001H000<10000P00001/0
00@10001?P030@<000<1000020020@L00P4200811`000`40000F0081P@0001/000@10001?P000`40
000300030@0000P00P4600<10P020@L000<100005P030H00000K00040@000Ch000<100000`000`40
000800811@020@03004000<11`000`40000F0081P@0001/000@10001?@020@D000<1000020020@D0
00<100000`400`01000600030@0001H00`6000006`001040004m00030@0000@000<1000020020@D0
00<100000P4200030@0000D000<1000060000`40001n00006`001040004m00030@0000@000<10000
20020@D000<100000P4200030@0000D000<1000060000`40001n00006`001040004m00030@0000@0
00<1000020030@<00P4600030@0000D000<1000060000`40001n00006`020@03004003`000<10000
10000`40000800<100<00@400P4600030@0000D000<1000060000`40001n00006`001040004j0004
0@000@H000<1000020001P400@0100811P000`40000500030@0001P000<10000OP0001/000@10001
>P001040004600030@0000P000<100400P400`010@0600030@0000@00P4J00030@0007h0000K0004
0@000CT00`400`01000500030@0000P000@100410P020@H000<1000010000`40000I00030@0007h0
000K00040@000CT00`400`01000500030@0000P000@100010P020@H000<100000`020A/000<10000
OP0001/000@10001>0040@03004000D000<100001`020@8000<100000P4600030@0000<000<10000
6P000`40001n00006`001040004h00H11P000`40000700030@0000@000<100001@000`4000020081
70000`40001n00006`001040004h00H11P000`40000700030@0000`000<100000P000`40000K0003
0@0007h0000K008100<00@00=`030@8000<1000010000`40000700030@0000`000<100000P000`40
000K00030@0007h0000K00040@000CP00`4200030@0000@000<100001`000`40000<00030@000080
00<100006`000`40000o0081?@0001/000@10001;`000`40000600<12@000`40000700030@0000`0
00<100000P000`40000K00030@0000<00`4i0081?@0001/000@100018P000`40000:00030@0000H0
0`4900030@0000L000<1000030000`40000200030@0001/000<100000`030CP00`4m00006`001040
004R00030@0000X000<100001P030@T000<100001`000`40000<00040@0000817@000`40000300<1
=`0010400@4m00006`001040004R00030@0000X000<100001P000`400@0900030@0000L000<10000
30000`4000030Ad000<100000`000`400@0g00@1?@0001/000@1000180000`400@0700030@000080
00<100001P000`400@0900030@0000L000<1000030000`4000030Ad000<100000P401001004g00@1
?@0000800P4400030@000080104300811@001040004P00030@0100L000<100000P030@H000<10040
2@000`40000700030@0000`000<100000`4M00030@00008100@00@01=P020@8000<10000>`000005
0040004020000`40000300040@000@@00P400`01000O00030@0100H00P4400<11P000`400@090003
0@0000H00P4>00040@000Al000<100000P401001004O00030@0001@000D100000@0m000000D00@00
0@0900030@00008000@1000110001040004P00030@0100H00P4400030@0100@00`4;00030@0000H0
00<100003@001040004O00030@00008100@00@017P020AH000D100000@0m000000D00@000@0:0004
0@0000<11@001040004P00040@010@D00`4300030@0100@00`4;00030@0000H000<100003@020@03
004001h000<100000P401001004700030@0000h0104200815P001@400001008000<10000>0000005
004000402`001040004700040@000B0000@100411@030@<000<1004010030@/000<100001P000`40
000=008100<00@007P000`4000040@03004000H00P4?00D100<00@400P400`01000200030@0000`0
00H100000@020CX000001@01000100P000@100010P000`40000500040@000B00104500<10`001040
0@4300<12`000`40000600030@0000d00P400`01000N008100<00@000P400`01000600810P001040
0@4900030@00008100<00@000P400`01000200030@0000`000H100000@030CT0000200812P020@@0
0`4400040@000B00104500030@0100<000@100410P030@`000<100001P000`40000=008100<00@00
7P020@03004000<000<100001@080@<00`4300070@0000400@0200030@0100<00P4=00060@000040
104h00006`001040004P00@11@000`400@0200810P060@`000<100001P000`40000=00@180030@@0
00<1000010070@03004000800`4200810`000`400@0200@10P020@`00P4300<100<00@40>00001/0
00@1000180040@D000@100400`4200D13@000`40000600030@0000d00`4R008110000`4000040081
1P030@0400410@800P4300<10`030@801@4400030@0000800P4400810P000`40000f00006`020@03
004001l000@100411@0010400@030@801@4=00030@0000H000<100003@020B<00P4400030@0000@0
0P4600<100@00@010P020@@00P4500@100@00@410`030@<00P4800030@0003H0000K00040@000B00
00@100410`0010400@050@801@4=00030@0000H000<100003@000`40000X00030@0000@000<10000
1@000`400@020@05004000401@020@D010400`0100020@03004100810P020@T000<100008P040A00
000K00040@000B0000@100410`0010400@020@03004100801@4=00030@0000H000<100003@000`40
000X00030@0000@000<100001`030@05004100401@020@X000<1004010400`0100030@T000<10000
8@050A00000K00040@000B0000@100410`020@800P400`010@0300@13@000`40000600813P000`40
000X00030@0000@000<100002@000`4000030A4000<100000P020@03004000812P000`40000K0003
0@0000<00P4200030@000081300001/000@1000180000`400@020@800P42008100<00@401@020@d0
00<100001P000`40000h00030@0000<000<100003@000`40000F008100@00@412`000`40000K00<1
0P020@<00P400`010@0<00006`001040004P00030@0100810P020@800P4200030@0000<00P4=0003
0@0000H000<10000>0001040004200030@0003L000<100006`030@800P4400<13@0001/000@10001
80000`400@020@800P4200050@0000401@020@d00`4010010@4200030@0003P01@400`01000h0003
0@0001401045008100<00@400P4600813@0001/000@100017P0010400@030@800P4600030@0000<0
0P4=00<100<00@400P400`01000i00L1>@000`40000@00D11@00104000030AD0000K00040@000Ad0
0P400`0100030@800P4600030@0000<00P4=00<100<00@000P400`01000k00D1>@000`4000090081
0P020@0300410080104200030@0000800P4E00006`020@03004001`00`4200<10P020@d000<10000
2`050@0500410040?0020@03004103T00P4700H100@00@411P030@03004001X0000K00040@000Ad0
0`4200030@0100800P4=00030@0000/01042008100<00@00?00010400@4j00030@0000D00`4200H1
1`030A/0000K00040@000Ad00`4200030@0100800P4=00030@0000/000<1000010000`400@1k0081
1P030@80104W00006`001040004M00<10P000`400@0200816`000`40000400<1N`050@8010430081
:00001/000@100017@030@@00P400`010@0R0081OP030@800P4_00006`000`4000020A`00`440081
00<00@008P000`40001m00030@0100<1<00001/000<100000P4L00<110020@0300400:@0104`0000
6`000`4000020A`00`44008100<00@00Y@030C00000K008100<00@4070030@@0106X00030@0002h0
000K00030@00008170000`400@0400@1f@0001/000<100000P4L00030@0100D00`7I00006`000`40
00020A`000<100401@030MT0000K00030@000081700010400@4600030@000=L0000K00030@000081
70000`400@0700030@000=L0000K00050@00004070000`400@0700030@000=L00002008110000`40
000200@10`020@D000D100000@0L00030@0100L000<10000e`0000050040004020000`4000030004
0@000@@00P4200030@0001T00P400`01003P000000D00@000@0900030@00008000@1000110001@40
000101/00P7S000000D00@000@0:00030@0000800P4500050@0000406`020N<000001@01000100/0
00@100010P000`40000200050@0000406`020N<000001@01000100P000@100010P00104000440005
0@0000406`020N<0000200812P020@@00P4500050@0000406`020N<0000K00050@0000406`020N<0
000K00050@0000406`020N<0000K00810P000`40000I0081h`0001/000D100000@0>00030@0000X0
0P7S00006`001@40000100d00P4<0081h`0001/000D100000@0=008130020N<0000K00050@000040
3@030@/00P7S00006`001@40000100d00`4;0081h`0001/000D100000@0=00<12`020N<0000K0005
0@0000403@030@/00P7S00006`020@8000<100002P040@/00P7S00006`001@40000100`000@10041
2`000`40003R00006`001@40000100`000@100412`000`40003R00006`001@40000100`000@10041
2P000`40003S00006`001@40000100`000@100412P000`40003S00006`001@40000100`000@10041
2P000`40003S00006`001@40000100`000@100412P000`40003S00006`001@40000100`000@10041
2P000`40003S00006`020@8000<100002P001040004:00030@0000@01043008110020@@00P4M00@1
0`020@@00P4400817@020@@00P44008110020Al000<100000P020@@00P4400817@030@@00P440081
10020@<0000K00050@00004030001040004:00030@0000H000D100000@0200040@000@8000@10001
0P000`40000J00030@0000<000@100010P001040004200040@000A/000@100010P00104000420004
0@000@8000@100017P001@400001008000@100010P001040004200030@0001d000@100010P001040
004200040@000@800@410040000K00050@00004030001P4000400@P000<100001P001@4000010080
00@100010P001040004200030@0001/000<100000P001040004200040@000@8000@100017P001040
004200040@000@8000@100010P000`40000I00D10P001040004200040@000@8000@100017`001040
004200040@000@8000@100010P010@400@0001/000D100000@0<00030@0000<120000`4000060005
0@0000400P001040004200040@000@8000<1000070001@400001008000@100010P00104000420003
0@0001X00P4300040@000@8000@100010P001040004K00040@000@<000@100010P00104000420004
0@000A`00`4300040@000@8000@100010P001040004200006`001@40000100`000<100000`450004
0@000@P000D100000@0200040@000@8000@100010P000`40000M00040@000@8000@100010P001040
004200030@0001`000@100010P001040004200040@000@8000<100006P000`400@0300040@000@80
00@100010P001040004L00030@0000<000@100010P001040004200040@000@80000K00050@000040
30000`4000030@@00P4200030@0000@00`4300040@000@8000@100010P001040004L00040@000@80
00@100010P001040004200040@000A/000@100010P001040004200040@000@8000@100017@020@<0
00@100010P001040004200040@000A`000<100000`001040004200040@000@8000@100010P0001/0
00D100000@0<00030@0000<110020@8000<100001P000`400002008110020@@00P4N008110020@@0
0P4400817@020@@00P44008110020Al000<100000P020@@00P4400817@040@<00P44008110020@<0
000K00040@0000812`000`4000030@@00`400`01003T00006`00104000020@/000<10000104300D1
i@0001D0o`460@H0000K00040@00008110000`40000400030@0000@10P060@L000<100001`000`40
000700030@0000L000<100001`000`40000700030@0000L000<100001`000`40000700030@0000L0
00<100001P000`40000700030@0000L000<100001`000`40000700030@0000L000<100001`000`40
000700030@0000L000<100001`000`40000700030@0000L000<100002P0001/000@100000P4;0003
0@0000@10P000`400@030ND0000K00040@0000812`000`4000040@8000<100400`7U00006`001040
00020@/000<10000104010010@050N@0000K00040@0000812`000`40000200D100<00@400P7U0000
6`00104000020@/000<100000`040@0300410081i@0001/000@100000P4;00030@0000<010400`01
0@020ND0000K00810P020@/000<100000`040@800`7U00006`00104000020@/000<100000`040@<0
0P7U00006`00104000020@/000<100000`040NX0000K00040@0000812`000`40000300@1jP0001/0
00@100000P4;00030@0000D00P7Z00006`00104000020@/000<100001@020NX0000K00040@000081
2`000`40000500030@000>T0000K00040@0000812`000`40000500030@000>T0003o0240003o0240
003o0240003o0240003o0240003o0240003o0240003o02400000\
\>"],
  ImageRangeCache->{{{108, 395}, {292.938, 116}} -> {-2729.67, 0.277314, 
  20.182, 0.000612208}}]
}, Open  ]],

Cell[TextData[{
  "Compute the exact probability that you have a cold given that you've \
sneezed and not observed cat scratches on the furniture: ",
  StyleBox["Pr[2,{1,!5}]",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell["\<\
Divide[Marginalize[{{2,True},{1,True},{5,False}},network2],
       Marginalize[{{1,True},{5,False}},network2]]\
\>", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[BoxData[
    \(0.460749499951537089`\)], "Output"]
}, Open  ]],

Cell["\<\
In this experiment, we have not observed scratches on the \
furniture; note that not observing scratches is more likely than observing \
scratches so the previous query involved relatively rare evidence while this \
one involves relatively common place evidence.  Rare evidence in non root \
nodes is a problem for simple forward sampling schemes that throw away \
samples that don't agree with the evidence; likelihood weighting allows us to \
use all of our samples.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(GraphLikelihoodWeighting[{{1, True}, {5, False}}, network2, 5000, 2]
      \)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.000190476 -2.57544 6.19824 [
[.21429 .51119 -12 -9 ]
[.21429 .51119 12 0 ]
[.40476 .51119 -12 -9 ]
[.40476 .51119 12 0 ]
[.59524 .51119 -12 -9 ]
[.59524 .51119 12 0 ]
[.78571 .51119 -12 -9 ]
[.78571 .51119 12 0 ]
[.97619 .51119 -12 -9 ]
[.97619 .51119 12 0 ]
[.01131 .02783 -24 -4.5 ]
[.01131 .02783 0 4.5 ]
[.01131 .15179 -24 -4.5 ]
[.01131 .15179 0 4.5 ]
[.01131 .27576 -24 -4.5 ]
[.01131 .27576 0 4.5 ]
[.01131 .39972 -24 -4.5 ]
[.01131 .39972 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .52369 m
.21429 .52994 L
s
[(1000)] .21429 .51119 0 1 Mshowa
.40476 .52369 m
.40476 .52994 L
s
[(2000)] .40476 .51119 0 1 Mshowa
.59524 .52369 m
.59524 .52994 L
s
[(3000)] .59524 .51119 0 1 Mshowa
.78571 .52369 m
.78571 .52994 L
s
[(4000)] .78571 .51119 0 1 Mshowa
.97619 .52369 m
.97619 .52994 L
s
[(5000)] .97619 .51119 0 1 Mshowa
.125 Mabswid
.0619 .52369 m
.0619 .52744 L
s
.1 .52369 m
.1 .52744 L
s
.1381 .52369 m
.1381 .52744 L
s
.17619 .52369 m
.17619 .52744 L
s
.25238 .52369 m
.25238 .52744 L
s
.29048 .52369 m
.29048 .52744 L
s
.32857 .52369 m
.32857 .52744 L
s
.36667 .52369 m
.36667 .52744 L
s
.44286 .52369 m
.44286 .52744 L
s
.48095 .52369 m
.48095 .52744 L
s
.51905 .52369 m
.51905 .52744 L
s
.55714 .52369 m
.55714 .52744 L
s
.63333 .52369 m
.63333 .52744 L
s
.67143 .52369 m
.67143 .52744 L
s
.70952 .52369 m
.70952 .52744 L
s
.74762 .52369 m
.74762 .52744 L
s
.82381 .52369 m
.82381 .52744 L
s
.8619 .52369 m
.8619 .52744 L
s
.9 .52369 m
.9 .52744 L
s
.9381 .52369 m
.9381 .52744 L
s
.25 Mabswid
0 .52369 m
1 .52369 L
s
.02381 .02783 m
.03006 .02783 L
s
[(0.42)] .01131 .02783 1 0 Mshowa
.02381 .15179 m
.03006 .15179 L
s
[(0.44)] .01131 .15179 1 0 Mshowa
.02381 .27576 m
.03006 .27576 L
s
[(0.46)] .01131 .27576 1 0 Mshowa
.02381 .39972 m
.03006 .39972 L
s
[(0.48)] .01131 .39972 1 0 Mshowa
.125 Mabswid
.02381 .05882 m
.02756 .05882 L
s
.02381 .08981 m
.02756 .08981 L
s
.02381 .1208 m
.02756 .1208 L
s
.02381 .18278 m
.02756 .18278 L
s
.02381 .21377 m
.02756 .21377 L
s
.02381 .24476 m
.02756 .24476 L
s
.02381 .30675 m
.02756 .30675 L
s
.02381 .33774 m
.02756 .33774 L
s
.02381 .36873 m
.02756 .36873 L
s
.02381 .43071 m
.02756 .43071 L
s
.02381 .4617 m
.02756 .4617 L
s
.02381 .49269 m
.02756 .49269 L
s
.02381 .55468 m
.02756 .55468 L
s
.02381 .58567 m
.02756 .58567 L
s
.02381 .61666 m
.02756 .61666 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02504 0 m
.02506 .61803 L
s
.04052 .61803 m
.04057 .5438 L
.04076 .53988 L
.04095 .53597 L
.04114 .53207 L
.04133 .39491 L
.04152 .39135 L
.04171 .38781 L
.0419 .38427 L
.0421 .38074 L
.04229 .37722 L
.04248 .37371 L
.04267 .3702 L
.04286 .36671 L
.04305 .36322 L
.04324 .36139 L
.04343 .35791 L
.04362 .35445 L
s
.04362 .35445 m
.0438 .61803 L
s
.04484 .61803 m
.04495 .61607 L
.04514 .61265 L
.04533 .60924 L
.04552 .60583 L
.04571 .60244 L
.0459 .59904 L
.0461 .59566 L
.04629 .59228 L
.04648 .58892 L
s
.04648 .58892 m
.0465 .61803 L
s
.05626 .61803 m
.05638 .50393 L
.05657 .50154 L
.05676 .49916 L
.05695 .49678 L
.05714 .4944 L
.05733 .49203 L
.05752 .48966 L
.05771 .48729 L
.0579 .48493 L
.0581 .48257 L
.05829 .48021 L
.05848 .47786 L
.05867 .47551 L
.05886 .47317 L
.05905 .47083 L
.05924 .46849 L
.05943 .31332 L
.05962 .49551 L
.05981 .4934 L
.06 .41757 L
.06019 .28336 L
.06038 .28153 L
.06057 .2797 L
.06076 .27788 L
.06095 .27606 L
.06114 .27423 L
.06133 .27242 L
.06152 .2706 L
.06171 .26879 L
.0619 .26697 L
.0621 .14566 L
.06229 .144 L
.06248 .14234 L
.06267 .14069 L
.06286 .13904 L
.06305 .13739 L
.06324 .13574 L
.06343 .13409 L
.06362 .13245 L
.06381 .13081 L
.064 .12917 L
.06419 .12753 L
.06438 .12589 L
.06457 .12426 L
.06476 .01458 L
.06495 .01308 L
.06514 .01158 L
.06533 .19064 L
.06552 .18912 L
Mistroke
.06571 .1876 L
.0659 .18608 L
.0661 .18456 L
.06629 .18304 L
.06648 .18153 L
.06667 .18001 L
.06686 .1785 L
.06705 .07667 L
.06724 .07527 L
.06743 .07387 L
.06762 .07247 L
.06781 .07108 L
.068 .06968 L
.06819 .06829 L
.06838 .01798 L
.06857 .01664 L
.06876 .0153 L
.06895 .01397 L
.06914 .01263 L
.06933 .0113 L
Mfstroke
.06933 .0113 m
.06936 0 L
s
.07904 0 m
.07905 .00304 L
s
.07905 .00304 m
.07905 0 L
s
.08012 0 m
.08019 .0461 L
.08038 .04496 L
.08057 .04382 L
.08076 .04268 L
.08095 .04154 L
.08114 .0404 L
.08133 .03926 L
.08152 .03813 L
.08171 .03699 L
s
.08171 .03699 m
.08181 0 L
s
.08255 0 m
.08267 .07617 L
.08286 .07509 L
.08305 .07401 L
.08324 .07293 L
.08343 .07185 L
.08362 .07078 L
.08381 .0697 L
.084 .03068 L
.08419 .02964 L
.08438 .02859 L
.08457 .02755 L
.08476 .02651 L
.08495 .02547 L
.08514 .02443 L
.08533 .02339 L
.08552 .02236 L
.08571 .02132 L
.0859 .02028 L
.0861 .01925 L
.08629 .01822 L
.08648 .01718 L
.08667 .01615 L
.08686 .01512 L
.08705 .12646 L
.08724 .12541 L
.08743 .12437 L
.08762 .12333 L
.08781 .12229 L
.088 .05147 L
.08819 .05048 L
.08838 .0495 L
.08857 .04851 L
.08876 .04753 L
s
.08876 .04753 m
.0889 0 L
s
.09045 0 m
.09048 .01406 L
.09067 .01317 L
.09086 .01228 L
.09105 .01138 L
s
.09105 .01138 m
.09108 0 L
s
.09403 0 m
.0941 .03477 L
.09429 .03391 L
.09448 .03306 L
.09467 .03221 L
.09486 .03136 L
.09505 .03051 L
.09524 .02966 L
.09543 .02881 L
.09562 .02797 L
.09581 .02712 L
.096 .02627 L
.09619 .02543 L
.09638 .11666 L
.09657 .05822 L
.09676 .0574 L
.09695 .05659 L
.09714 .05577 L
.09733 .14205 L
.09752 .14123 L
.09771 .14041 L
.0979 .13959 L
.0981 .13877 L
.09829 .13795 L
.09848 .13713 L
.09867 .21931 L
.09886 .21849 L
.09905 .21766 L
.09924 .21684 L
.09943 .21602 L
.09962 .2152 L
.09981 .21438 L
.1 .21356 L
.10019 .21274 L
.10038 .21192 L
.10057 .2111 L
.10076 .21028 L
.10095 .15415 L
.10114 .15337 L
.10133 .15258 L
.10152 .1518 L
.10171 .15139 L
.1019 .1506 L
.1021 .14982 L
.10229 .14904 L
.10248 .14825 L
.10267 .14784 L
.10286 .14706 L
.10305 .22509 L
.10324 .17114 L
Mistroke
.10343 .17038 L
.10362 .16963 L
.10381 .16887 L
.104 .16812 L
.10419 .16736 L
.10438 .16661 L
.10457 .16586 L
.10476 .1651 L
.10495 .16435 L
.10514 .1636 L
.10533 .16285 L
.10552 .1621 L
.10571 .23641 L
.1059 .23566 L
.1061 .23491 L
.10629 .23415 L
.10648 .2334 L
.10667 .23265 L
.10686 .23189 L
.10705 .23114 L
.10724 .23039 L
.10743 .22964 L
.10762 .22889 L
.10781 .22814 L
.108 .22739 L
.10819 .22664 L
.10838 .22589 L
.10857 .22514 L
.10876 .22439 L
.10895 .22364 L
.10914 .22289 L
.10933 .22215 L
.10952 .2214 L
.10971 .22065 L
.1099 .21991 L
.1101 .21916 L
.11029 .21842 L
.11048 .21767 L
.11067 .21693 L
.11086 .25473 L
.11105 .25398 L
.11124 .25324 L
.11143 .25249 L
.11162 .25175 L
.11181 .251 L
.112 .25026 L
.11219 .24952 L
.11238 .24877 L
.11257 .19771 L
.11276 .19699 L
Mistroke
.11295 .19628 L
.11314 .19556 L
.11333 .19485 L
.11352 .19413 L
.11371 .19342 L
.1139 .1927 L
.1141 .19199 L
.11429 .19128 L
.11448 .19056 L
.11467 .18985 L
.11486 .18914 L
.11505 .18843 L
.11524 .18772 L
.11543 .16176 L
.11562 .16106 L
.11581 .16037 L
.116 .15967 L
.11619 .15897 L
.11638 .15828 L
.11657 .15758 L
.11676 .15689 L
.11695 .15619 L
.11714 .1555 L
.11733 .1548 L
.11752 .15411 L
.11771 .15341 L
.1179 .15272 L
.1181 .15203 L
.11829 .22116 L
.11848 .22046 L
.11867 .21977 L
.11886 .21907 L
.11905 .21838 L
.11924 .21768 L
.11943 .21699 L
.11962 .21629 L
.11981 .2156 L
.12 .21491 L
.12019 .21421 L
.12038 .21352 L
.12057 .21283 L
.12076 .21214 L
.12095 .21145 L
.12114 .21075 L
.12133 .21006 L
.12152 .20937 L
.12171 .20868 L
.1219 .20799 L
.1221 .2073 L
.12229 .20661 L
Mistroke
.12248 .20593 L
.12267 .20524 L
.12286 .20455 L
.12305 .20386 L
.12324 .20317 L
.12343 .20249 L
.12362 .2018 L
.12381 .20111 L
.124 .20043 L
.12419 .26678 L
.12438 .26609 L
.12457 .2187 L
.12476 .21804 L
.12495 .21737 L
.12514 .21671 L
.12533 .21604 L
.12552 .21538 L
.12571 .21471 L
.1259 .21405 L
.1261 .21339 L
.12629 .21272 L
.12648 .21206 L
.12667 .2114 L
.12686 .21074 L
.12705 .21007 L
.12724 .20941 L
.12743 .20875 L
.12762 .20809 L
.12781 .1626 L
.128 .16196 L
.12819 .16132 L
.12838 .16068 L
.12857 .16005 L
.12876 .1161 L
.12895 .11548 L
.12914 .11486 L
.12933 .11425 L
.12952 .11363 L
.12971 .11301 L
.1299 .1124 L
.1301 .11178 L
.13029 .11117 L
.13048 .11055 L
.13067 .17375 L
.13086 .17313 L
.13105 .23409 L
.13124 .23347 L
.13143 .23285 L
.13162 .23223 L
.13181 .29106 L
Mistroke
.132 .29044 L
.13219 .28982 L
.13238 .2892 L
.13257 .28858 L
.13276 .28796 L
.13295 .28734 L
.13314 .24457 L
.13333 .24397 L
.13352 .24337 L
.13371 .24277 L
.1339 .24217 L
.1341 .24157 L
.13429 .24097 L
.13448 .24037 L
.13467 .23977 L
.13486 .23917 L
.13505 .23857 L
.13524 .23797 L
.13543 .23737 L
.13562 .23677 L
.13581 .1955 L
.136 .19492 L
.13619 .19433 L
.13638 .19375 L
.13657 .19317 L
.13676 .19259 L
.13695 .19201 L
.13714 .19144 L
.13733 .19086 L
.13752 .15091 L
.13771 .15035 L
.1379 .14978 L
.1381 .14922 L
.13829 .14866 L
.13848 .1481 L
.13867 .14754 L
.13886 .14698 L
.13905 .14642 L
.13924 .14586 L
.13943 .14529 L
.13962 .20164 L
.13981 .20108 L
.14 .20052 L
.14019 .25506 L
.14038 .2545 L
.14057 .23388 L
.14076 .23333 L
.14095 .23277 L
.14114 .19445 L
.14133 .19391 L
Mistroke
.14152 .19337 L
.14171 .19283 L
.1419 .19229 L
.1421 .19175 L
.14229 .19121 L
.14248 .19067 L
.14267 .19013 L
.14286 .1896 L
.14305 .24223 L
.14324 .24169 L
.14343 .24115 L
.14362 .24061 L
.14381 .20336 L
.144 .20283 L
.14419 .20231 L
.14438 .18311 L
.14457 .18259 L
.14476 .18207 L
.14495 .18155 L
.14514 .18104 L
.14533 .18052 L
.14552 .18 L
.14571 .17949 L
.1459 .17897 L
.1461 .17846 L
.14629 .17794 L
.14648 .22879 L
.14667 .22827 L
.14686 .27764 L
.14705 .27712 L
.14724 .24128 L
.14743 .24078 L
.14762 .20584 L
.14781 .20535 L
.148 .25311 L
.14819 .25262 L
.14838 .25212 L
.14857 .25163 L
.14876 .21749 L
.14895 .19985 L
.14914 .19938 L
.14933 .1989 L
.14952 .19843 L
.14971 .19795 L
.1499 .19748 L
.1501 .19723 L
.15029 .19675 L
.15048 .19628 L
.15067 .1958 L
.15086 .19533 L
Mistroke
.15105 .19486 L
.15124 .19438 L
.15143 .19391 L
.15162 .19343 L
.15181 .19296 L
.152 .19249 L
.15219 .19202 L
.15238 .19154 L
.15257 .19107 L
.15276 .1906 L
.15295 .19012 L
.15314 .18965 L
.15333 .23593 L
.15352 .21858 L
.15371 .21812 L
.1539 .21765 L
.1541 .21718 L
.15429 .21671 L
.15448 .21625 L
.15467 .21578 L
.15486 .26077 L
.15505 .2603 L
.15524 .25983 L
.15543 .25937 L
.15562 .30318 L
.15581 .30271 L
.156 .30224 L
.15619 .30178 L
.15638 .30131 L
.15657 .30084 L
.15676 .30037 L
.15695 .2999 L
.15714 .29944 L
.15733 .29897 L
.15752 .2985 L
.15771 .29803 L
.1579 .29757 L
.1581 .2971 L
.15829 .29663 L
.15848 .29617 L
.15867 .2957 L
.15886 .29523 L
.15905 .29477 L
.15924 .2943 L
.15943 .26206 L
.15962 .2616 L
.15981 .26115 L
.16 .26069 L
.16019 .26024 L
.16038 .30284 L
Mistroke
.16057 .30239 L
.16076 .30193 L
.16095 .30147 L
.16114 .26995 L
.16133 .2695 L
.16152 .26905 L
.16171 .26861 L
.1619 .26816 L
.1621 .26772 L
.16229 .26727 L
.16248 .26683 L
.16267 .23605 L
.16286 .23562 L
.16305 .23518 L
.16324 .23475 L
.16343 .27615 L
.16362 .27571 L
.16381 .27528 L
.164 .27484 L
.16419 .27441 L
.16438 .27397 L
.16457 .24385 L
.16476 .24342 L
.16495 .28376 L
.16514 .28333 L
.16533 .2829 L
.16552 .28248 L
.16571 .28205 L
.1659 .26644 L
.1661 .26602 L
.16629 .2656 L
.16648 .26518 L
.16667 .26476 L
.16686 .24933 L
.16705 .24892 L
.16724 .2485 L
.16743 .24809 L
.16762 .24767 L
.16781 .24725 L
.168 .24684 L
.16819 .24642 L
.16838 .24601 L
.16857 .21723 L
.16876 .21682 L
.16895 .21642 L
.16914 .21601 L
.16933 .2156 L
.16952 .2152 L
.16971 .21479 L
.1699 .21438 L
Mistroke
.1701 .21398 L
.17029 .21357 L
.17048 .21317 L
.17067 .21276 L
.17086 .25683 L
.17105 .25642 L
.17124 .25602 L
.17143 .25561 L
.17162 .25521 L
.17181 .2548 L
.172 .29302 L
.17219 .26482 L
.17238 .26442 L
.17257 .26402 L
.17276 .26362 L
.17295 .26322 L
.17314 .26282 L
.17333 .26243 L
.17352 .26203 L
.17371 .26163 L
.1739 .26123 L
.1741 .29854 L
.17429 .29814 L
.17448 .29774 L
.17467 .29735 L
.17486 .29695 L
.17505 .33344 L
.17524 .33304 L
.17543 .33264 L
.17562 .33224 L
.17581 .33184 L
.176 .33144 L
.17619 .33104 L
.17638 .33065 L
.17657 .33025 L
.17676 .32985 L
.17695 .32945 L
.17714 .32905 L
.17733 .32866 L
.17752 .32826 L
.17771 .32786 L
.1779 .32746 L
.1781 .32707 L
.17829 .32667 L
.17848 .32627 L
.17867 .32588 L
.17886 .32548 L
.17905 .32508 L
.17924 .32469 L
.17943 .32429 L
Mistroke
.17962 .32389 L
.17981 .3235 L
.18 .29603 L
.18019 .29564 L
.18038 .33128 L
.18057 .33089 L
.18076 .37009 L
.18095 .3697 L
.18114 .36931 L
.18133 .36892 L
.18152 .36853 L
.18171 .36814 L
.1819 .36775 L
.1821 .36736 L
.18229 .36697 L
.18248 .36658 L
.18267 .36619 L
.18286 .36581 L
.18305 .36542 L
.18324 .33848 L
.18343 .33809 L
.18362 .33771 L
.18381 .33733 L
.184 .33695 L
.18419 .33657 L
.18438 .33619 L
.18457 .33581 L
.18476 .33543 L
.18495 .33505 L
.18514 .33467 L
.18533 .33429 L
.18552 .33391 L
.18571 .33353 L
.1859 .33315 L
.1861 .33277 L
.18629 .33239 L
.18648 .33201 L
.18667 .33163 L
.18686 .33125 L
.18705 .33087 L
.18724 .33049 L
.18743 .33011 L
.18762 .32973 L
.18781 .32935 L
.188 .32897 L
.18819 .32859 L
.18838 .32822 L
.18857 .32784 L
.18876 .32746 L
.18895 .32708 L
Mistroke
.18914 .3267 L
.18933 .30046 L
.18952 .30009 L
.18971 .29972 L
.1899 .29935 L
.1901 .29898 L
.19029 .33715 L
.19048 .33678 L
.19067 .33641 L
.19086 .33604 L
.19105 .33567 L
.19124 .33529 L
.19143 .33492 L
.19162 .33455 L
.19181 .33418 L
.192 .33381 L
.19219 .33344 L
.19238 .33307 L
.19257 .30735 L
.19276 .30699 L
.19295 .30662 L
.19314 .30626 L
.19333 .3059 L
.19352 .30553 L
.19371 .30517 L
.1939 .30481 L
.1941 .30444 L
.19429 .30408 L
.19448 .30372 L
.19467 .30335 L
.19486 .30299 L
.19505 .30263 L
.19524 .30226 L
.19543 .3019 L
.19562 .30154 L
.19581 .30118 L
.196 .30081 L
.19619 .28752 L
.19638 .28716 L
.19657 .2868 L
.19676 .28644 L
.19695 .28608 L
.19714 .28573 L
.19733 .28537 L
.19752 .28501 L
.19771 .28465 L
.1979 .28429 L
.1981 .28393 L
.19829 .25907 L
.19848 .25872 L
Mistroke
.19867 .29181 L
.19886 .29146 L
.19905 .29111 L
.19924 .29076 L
.19943 .2904 L
.19962 .29005 L
.19981 .2897 L
.2 .26527 L
.20019 .26492 L
.20038 .26458 L
.20057 .26423 L
.20076 .26388 L
.20095 .29629 L
.20114 .29594 L
.20133 .2956 L
.20152 .29525 L
.20171 .2949 L
.2019 .29456 L
.2021 .29421 L
.20229 .32599 L
.20248 .32565 L
.20267 .3253 L
.20286 .32495 L
.20305 .32461 L
.20324 .32426 L
.20343 .30025 L
.20362 .29991 L
.20381 .33106 L
.204 .33072 L
.20419 .33038 L
.20438 .33004 L
.20457 .3297 L
.20476 .32936 L
.20495 .32902 L
.20514 .32868 L
.20533 .32834 L
.20552 .328 L
.20571 .3044 L
.2059 .30407 L
.2061 .30373 L
.20629 .3034 L
.20648 .30307 L
.20667 .30273 L
.20686 .3024 L
.20705 .30206 L
.20724 .30173 L
.20743 .3014 L
.20762 .30106 L
.20781 .30073 L
.208 .3004 L
Mistroke
.20819 .27725 L
.20838 .27692 L
.20857 .27659 L
.20876 .27626 L
.20895 .27593 L
.20914 .25317 L
.20933 .25285 L
.20952 .25253 L
.20971 .25221 L
.2099 .25188 L
.2101 .25156 L
.21029 .25124 L
.21048 .25092 L
.21067 .2506 L
.21086 .25027 L
.21105 .24995 L
.21124 .24963 L
.21143 .24931 L
.21162 .24899 L
.21181 .24867 L
.212 .27912 L
.21219 .2788 L
.21238 .27848 L
.21257 .27815 L
.21276 .27783 L
.21295 .27751 L
.21314 .27719 L
.21333 .27687 L
.21352 .27655 L
.21371 .27622 L
.2139 .2759 L
.2141 .27558 L
.21429 .27526 L
.21448 .27494 L
.21467 .27462 L
.21486 .2743 L
.21505 .27398 L
.21524 .27366 L
.21543 .26188 L
.21562 .26157 L
.21581 .26125 L
.216 .29483 L
.21619 .29451 L
.21638 .29419 L
.21657 .29387 L
.21676 .29355 L
.21695 .29324 L
.21714 .29292 L
.21733 .2926 L
.21752 .29228 L
Mistroke
.21771 .29196 L
.2179 .29165 L
.2181 .29133 L
.21829 .29101 L
.21848 .29069 L
.21867 .29037 L
.21886 .31966 L
.21905 .31934 L
.21924 .31902 L
.21943 .31871 L
.21962 .31839 L
.21981 .34716 L
.22 .34684 L
.22019 .34653 L
.22038 .34621 L
.22057 .34589 L
.22076 .34557 L
.22095 .34525 L
.22114 .34494 L
.22133 .34462 L
.22152 .3443 L
.22171 .34398 L
.2219 .34367 L
.2221 .34335 L
.22229 .34303 L
.22248 .3713 L
.22267 .37099 L
.22286 .37067 L
.22305 .37035 L
.22324 .37003 L
.22343 .36972 L
.22362 .3694 L
.22381 .36908 L
.224 .36877 L
.22419 .36845 L
.22438 .36813 L
.22457 .36782 L
.22476 .3675 L
.22495 .36718 L
.22514 .34518 L
.22533 .34487 L
.22552 .34456 L
.22571 .34425 L
.2259 .34394 L
.2261 .34362 L
.22629 .34331 L
.22648 .343 L
.22667 .34269 L
.22686 .34238 L
.22705 .34207 L
Mistroke
.22724 .34176 L
.22743 .34145 L
.22762 .34113 L
.22781 .34082 L
.228 .34051 L
.22819 .3402 L
.22838 .33989 L
.22857 .33958 L
.22876 .33927 L
.22895 .33896 L
.22914 .33865 L
.22933 .33834 L
.22952 .33803 L
.22971 .33772 L
.2299 .33741 L
.2301 .3371 L
.23029 .33679 L
.23048 .33648 L
.23067 .33617 L
.23086 .33586 L
.23105 .31432 L
.23124 .31402 L
.23143 .31371 L
.23162 .31341 L
.23181 .30221 L
.232 .3019 L
.23219 .3016 L
.23238 .3013 L
.23257 .301 L
.23276 .30069 L
.23295 .30039 L
.23314 .30009 L
.23333 .29979 L
.23352 .29948 L
.23371 .29918 L
.2339 .29888 L
.2341 .29858 L
.23429 .29828 L
.23448 .29797 L
.23467 .29767 L
.23486 .28659 L
.23505 .28629 L
.23524 .31399 L
.23543 .34124 L
.23562 .34094 L
.23581 .34064 L
.236 .34034 L
.23619 .34004 L
.23638 .36684 L
.23657 .36654 L
Mistroke
.23676 .36623 L
.23695 .36593 L
.23714 .36563 L
.23733 .36533 L
.23752 .36503 L
.23771 .3914 L
.2379 .3911 L
.2381 .3908 L
.23829 .3905 L
.23848 .37949 L
.23867 .40542 L
.23886 .40512 L
.23905 .40482 L
.23924 .40453 L
.23943 .40423 L
.23962 .40393 L
.23981 .40363 L
.24 .40334 L
.24019 .38266 L
.24038 .38236 L
.24057 .38207 L
.24076 .38178 L
.24095 .38148 L
.24114 .38119 L
.24133 .40985 L
.24152 .40956 L
.24171 .43462 L
.2419 .41422 L
.2421 .41393 L
.24229 .41364 L
.24248 .41335 L
.24267 .41306 L
.24286 .41277 L
.24305 .41248 L
.24324 .41219 L
.24343 .43685 L
.24362 .41674 L
.24381 .41646 L
.244 .41617 L
.24419 .41589 L
.24438 .4156 L
.24457 .41532 L
.24476 .41503 L
.24495 .41475 L
.24514 .41446 L
.24533 .41418 L
.24552 .43845 L
.24571 .43817 L
.2459 .43788 L
.2461 .46179 L
Mistroke
.24629 .4615 L
.24648 .46122 L
.24667 .46094 L
.24686 .46065 L
.24705 .46037 L
.24724 .46008 L
.24743 .4598 L
.24762 .45951 L
.24781 .45923 L
.248 .45894 L
.24819 .45866 L
.24838 .45837 L
.24857 .45809 L
.24876 .45781 L
.24895 .45752 L
.24914 .45724 L
.24933 .45695 L
.24952 .45667 L
.24971 .45638 L
.2499 .4561 L
.2501 .45582 L
.25029 .45553 L
.25048 .45525 L
.25067 .45497 L
.25086 .45468 L
.25105 .4544 L
.25124 .43467 L
.25143 .43439 L
.25162 .43411 L
.25181 .43383 L
.252 .43356 L
.25219 .43328 L
.25238 .433 L
.25257 .43272 L
.25276 .43244 L
.25295 .43216 L
.25314 .43188 L
.25333 .4316 L
.25352 .43132 L
.25371 .43104 L
.2539 .43076 L
.2541 .43048 L
.25429 .4302 L
.25448 .42992 L
.25467 .42964 L
.25486 .42937 L
.25505 .42909 L
.25524 .42881 L
.25543 .42853 L
.25562 .42825 L
Mistroke
.25581 .42797 L
.256 .42769 L
.25619 .42741 L
.25638 .42714 L
.25657 .42686 L
.25676 .42658 L
.25695 .4263 L
.25714 .42602 L
.25733 .42574 L
.25752 .42546 L
.25771 .41525 L
.2579 .41497 L
.2581 .4147 L
.25829 .41442 L
.25848 .41414 L
.25867 .41387 L
.25886 .41359 L
.25905 .41332 L
.25924 .41304 L
.25943 .41276 L
.25962 .39358 L
.25981 .39331 L
.26 .39303 L
.26019 .39276 L
.26038 .39249 L
.26057 .37356 L
.26076 .37329 L
.26095 .37302 L
.26114 .39653 L
.26133 .39627 L
.26152 .396 L
.26171 .37731 L
.2619 .35885 L
.2621 .35859 L
.26229 .35833 L
.26248 .35807 L
.26267 .35781 L
.26286 .35754 L
.26305 .35728 L
.26324 .35702 L
.26343 .35676 L
.26362 .3565 L
.26381 .35624 L
.264 .35597 L
.26419 .35571 L
.26438 .35545 L
.26457 .35519 L
.26476 .35493 L
.26495 .35467 L
.26514 .35441 L
Mistroke
.26533 .35415 L
.26552 .35388 L
.26571 .35362 L
.2659 .35336 L
.2661 .3531 L
.26629 .33495 L
.26648 .33469 L
.26667 .33443 L
.26686 .33417 L
.26705 .33392 L
.26724 .33366 L
.26743 .35679 L
.26762 .35653 L
.26781 .35627 L
.268 .35601 L
.26819 .35576 L
.26838 .3463 L
.26857 .34604 L
.26876 .34579 L
.26895 .34553 L
.26914 .32774 L
.26933 .32748 L
.26952 .32723 L
.26971 .32698 L
.2699 .32673 L
.2701 .32647 L
.27029 .32622 L
.27048 .349 L
.27067 .34875 L
.27086 .3485 L
.27105 .34824 L
.27124 .34799 L
.27143 .34774 L
.27162 .34749 L
.27181 .34724 L
.272 .34698 L
.27219 .34673 L
.27238 .34648 L
.27257 .34623 L
.27276 .34597 L
.27295 .34572 L
.27314 .34547 L
.27333 .36794 L
.27352 .36768 L
.27371 .36743 L
.2739 .36718 L
.2741 .36693 L
.27429 .36667 L
.27448 .37983 L
.27467 .37958 L
Mistroke
.27486 .37933 L
.27505 .37908 L
.27524 .37882 L
.27543 .37857 L
.27562 .37832 L
.27581 .4003 L
.276 .40005 L
.27619 .3998 L
.27638 .39954 L
.27657 .39929 L
.27676 .39904 L
.27695 .39879 L
.27714 .39854 L
.27733 .39828 L
.27752 .39803 L
.27771 .39778 L
.2779 .39753 L
.2781 .39728 L
.27829 .39703 L
.27848 .39678 L
.27867 .39652 L
.27886 .39627 L
.27905 .39602 L
.27924 .39577 L
.27943 .39552 L
.27962 .39527 L
.27981 .39502 L
.28 .39477 L
.28019 .39452 L
.28038 .39426 L
.28057 .39401 L
.28076 .39376 L
.28095 .39351 L
.28114 .39326 L
.28133 .39301 L
.28152 .39276 L
.28171 .39251 L
.2819 .39226 L
.2821 .39201 L
.28229 .39176 L
.28248 .39151 L
.28267 .39126 L
.28286 .39101 L
.28305 .39075 L
.28324 .41243 L
.28343 .41218 L
.28362 .41193 L
.28381 .4118 L
.284 .41155 L
.28419 .4113 L
Mistroke
.28438 .41105 L
.28457 .4108 L
.28476 .41055 L
.28495 .4103 L
.28514 .41005 L
.28533 .4098 L
.28552 .40955 L
.28571 .39215 L
.2859 .3919 L
.2861 .39165 L
.28629 .39141 L
.28648 .39116 L
.28667 .39091 L
.28686 .39067 L
.28705 .39042 L
.28724 .39017 L
.28743 .38993 L
.28762 .37275 L
.28781 .3638 L
.288 .36355 L
.28819 .36331 L
.28838 .36307 L
.28857 .36283 L
.28876 .36258 L
.28895 .36234 L
.28914 .34549 L
.28933 .34525 L
.28952 .34501 L
.28971 .34477 L
.2899 .34453 L
.2901 .34429 L
.29029 .34405 L
.29048 .34381 L
.29067 .34357 L
.29086 .34333 L
.29105 .34309 L
.29124 .34285 L
.29143 .3642 L
.29162 .36396 L
.29181 .36372 L
.292 .36349 L
.29219 .36325 L
.29238 .36301 L
.29257 .36277 L
.29276 .36253 L
.29295 .36229 L
.29314 .36205 L
.29333 .38312 L
.29352 .38288 L
.29371 .40368 L
Mistroke
.2939 .40344 L
.2941 .4032 L
.29429 .40297 L
.29448 .40273 L
.29467 .40249 L
.29486 .40225 L
.29505 .40201 L
.29524 .40177 L
.29543 .40153 L
.29562 .40129 L
.29581 .40106 L
.296 .40082 L
.29619 .38421 L
.29638 .38397 L
.29657 .38374 L
.29676 .3835 L
.29695 .38326 L
.29714 .36685 L
.29733 .36662 L
.29752 .36638 L
.29771 .36615 L
.2979 .36592 L
.2981 .36568 L
.29829 .36545 L
.29848 .36522 L
.29867 .36499 L
.29886 .3855 L
.29905 .38527 L
.29924 .38503 L
.29943 .3848 L
.29962 .38457 L
.29981 .38433 L
.3 .3841 L
.30019 .38387 L
.30038 .38363 L
.30057 .3834 L
.30076 .38317 L
.30095 .38294 L
.30114 .3827 L
.30133 .38247 L
.30152 .38224 L
.30171 .38201 L
.3019 .38177 L
.3021 .38154 L
.30229 .38131 L
.30248 .38108 L
.30267 .38084 L
.30286 .38061 L
.30305 .38038 L
.30324 .38015 L
Mistroke
.30343 .37992 L
.30362 .37968 L
.30381 .37945 L
.304 .37922 L
.30419 .37899 L
.30438 .37876 L
.30457 .37852 L
.30476 .37829 L
.30495 .36215 L
.30514 .37416 L
.30533 .39426 L
.30552 .39403 L
.30571 .3938 L
.3059 .41364 L
.3061 .41341 L
.30629 .41318 L
.30648 .41295 L
.30667 .41273 L
.30686 .4125 L
.30705 .41227 L
.30724 .41204 L
.30743 .41181 L
.30762 .41158 L
.30781 .41135 L
.308 .43095 L
.30819 .43072 L
.30838 .43049 L
.30857 .43026 L
.30876 .43003 L
.30895 .42981 L
.30914 .42958 L
.30933 .42116 L
.30952 .42093 L
.30971 .4207 L
.3099 .42047 L
.3101 .42025 L
.31029 .42002 L
.31048 .41979 L
.31067 .41956 L
.31086 .41933 L
.31105 .41911 L
.31124 .41888 L
.31143 .41865 L
.31162 .41842 L
.31181 .4182 L
.312 .41797 L
.31219 .41774 L
.31238 .41751 L
.31257 .41729 L
.31276 .40146 L
Mistroke
.31295 .40123 L
.31314 .40101 L
.31333 .40078 L
.31352 .40056 L
.31371 .40034 L
.3139 .40011 L
.3141 .39989 L
.31429 .39966 L
.31448 .39944 L
.31467 .41879 L
.31486 .41856 L
.31505 .41834 L
.31524 .41812 L
.31543 .41789 L
.31562 .40225 L
.31581 .40203 L
.316 .40181 L
.31619 .40159 L
.31638 .4207 L
.31657 .42048 L
.31676 .42026 L
.31695 .42003 L
.31714 .41981 L
.31733 .41959 L
.31752 .41937 L
.31771 .41915 L
.3179 .41892 L
.3181 .4187 L
.31829 .41848 L
.31848 .41826 L
.31867 .41804 L
.31886 .41781 L
.31905 .41759 L
.31924 .41737 L
.31943 .40193 L
.31962 .40171 L
.31981 .39365 L
.32 .39343 L
.32019 .39321 L
.32038 .39299 L
.32057 .39277 L
.32076 .39256 L
.32095 .39234 L
.32114 .39212 L
.32133 .3919 L
.32152 .38389 L
.32171 .38368 L
.3219 .38346 L
.3221 .38324 L
.32229 .38303 L
Mistroke
.32248 .38281 L
.32267 .38259 L
.32286 .37463 L
.32305 .37442 L
.32324 .3742 L
.32343 .37399 L
.32362 .37377 L
.32381 .35878 L
.324 .35094 L
.32419 .35073 L
.32438 .35052 L
.32457 .35031 L
.32476 .3501 L
.32495 .34988 L
.32514 .34967 L
.32533 .34946 L
.32552 .34925 L
.32571 .34904 L
.3259 .34882 L
.3261 .34861 L
.32629 .33387 L
.32648 .33366 L
.32667 .31907 L
.32686 .31886 L
.32705 .31865 L
.32724 .31845 L
.32743 .31824 L
.32762 .31803 L
.32781 .31782 L
.328 .31762 L
.32819 .31741 L
.32838 .33624 L
.32857 .33603 L
.32876 .33583 L
.32895 .33562 L
.32914 .33541 L
.32933 .3278 L
.32952 .32759 L
.32971 .3462 L
.3299 .346 L
.3301 .34579 L
.33029 .34558 L
.33048 .34538 L
.33067 .34517 L
.33086 .36357 L
.33105 .36337 L
.33124 .36316 L
.33143 .36295 L
.33162 .36275 L
.33181 .36254 L
Mistroke
.332 .38074 L
.33219 .38053 L
.33238 .38032 L
.33257 .38012 L
.33276 .37991 L
.33295 .3797 L
.33314 .3795 L
.33333 .37929 L
.33352 .37909 L
.33371 .37888 L
.3339 .37867 L
.3341 .37847 L
.33429 .37826 L
.33448 .37806 L
.33467 .39605 L
.33486 .39584 L
.33505 .39564 L
.33524 .39543 L
.33543 .39522 L
.33562 .39502 L
.33581 .39481 L
.336 .39461 L
.33619 .3944 L
.33638 .3942 L
.33657 .39399 L
.33676 .39378 L
.33695 .39358 L
.33714 .37926 L
.33733 .37905 L
.33752 .37885 L
.33771 .37864 L
.3379 .37844 L
.3381 .37824 L
.33829 .37803 L
.33848 .37783 L
.33867 .37035 L
.33886 .37015 L
.33905 .36995 L
.33924 .36975 L
.33943 .36955 L
.33962 .36211 L
.33981 .37989 L
.34 .37969 L
.34019 .37948 L
.34038 .37928 L
.34057 .37908 L
.34076 .37888 L
.34095 .37868 L
.34114 .39626 L
.34133 .39606 L
Mistroke
.34152 .39586 L
.34171 .39566 L
.3419 .39546 L
.3421 .39526 L
.34229 .39505 L
.34248 .41245 L
.34267 .41225 L
.34286 .41204 L
.34305 .41184 L
.34324 .39783 L
.34343 .39763 L
.34362 .39743 L
.34381 .39723 L
.344 .39703 L
.34419 .39683 L
.34438 .39663 L
.34457 .39643 L
.34476 .39623 L
.34495 .39604 L
.34514 .39584 L
.34533 .39564 L
.34552 .39544 L
.34571 .39524 L
.3459 .39504 L
.3461 .39484 L
.34629 .41204 L
.34648 .41184 L
.34667 .41164 L
.34686 .42866 L
.34705 .4455 L
.34724 .4453 L
.34743 .4451 L
.34762 .4449 L
.34781 .4447 L
.348 .43084 L
.34819 .43064 L
.34838 .43044 L
.34857 .43025 L
.34876 .43005 L
.34895 .42985 L
.34914 .42966 L
.34933 .42946 L
.34952 .42926 L
.34971 .42907 L
.3499 .42887 L
.3501 .42867 L
.35029 .42847 L
.35048 .42828 L
.35067 .42808 L
.35086 .42788 L
Mistroke
.35105 .42769 L
.35124 .42749 L
.35143 .42729 L
.35162 .4271 L
.35181 .4269 L
.352 .4132 L
.35219 .39963 L
.35238 .39943 L
.35257 .39924 L
.35276 .39905 L
.35295 .39885 L
.35314 .39866 L
.35333 .39847 L
.35352 .39827 L
.35371 .39808 L
.3539 .38464 L
.3541 .38445 L
.35429 .38426 L
.35448 .4009 L
.35467 .40071 L
.35486 .40052 L
.35505 .40033 L
.35524 .40014 L
.35543 .39994 L
.35562 .39975 L
.35581 .39956 L
.356 .39937 L
.35619 .39918 L
.35638 .39899 L
.35657 .3988 L
.35676 .39861 L
.35695 .39842 L
.35714 .39823 L
.35733 .39804 L
.35752 .4145 L
.35771 .41431 L
.3579 .41412 L
.3581 .41393 L
.35829 .43023 L
.35848 .43004 L
.35867 .44618 L
.35886 .44599 L
.35905 .4458 L
.35924 .4456 L
.35943 .44541 L
.35962 .44522 L
.35981 .44503 L
.36 .44484 L
.36019 .44465 L
.36038 .44446 L
Mistroke
.36057 .44427 L
.36076 .44408 L
.36095 .44389 L
.36114 .43688 L
.36133 .43669 L
.36152 .4365 L
.36171 .42328 L
.3619 .4231 L
.3621 .42291 L
.36229 .42272 L
.36248 .42253 L
.36267 .42235 L
.36286 .42216 L
.36305 .42197 L
.36324 .42178 L
.36343 .42159 L
.36362 .43756 L
.36381 .43737 L
.364 .43718 L
.36419 .437 L
.36438 .43681 L
.36457 .43662 L
.36476 .42972 L
.36495 .42953 L
.36514 .42935 L
.36533 .42916 L
.36552 .41614 L
.36571 .40323 L
.3659 .40305 L
.3661 .40287 L
.36629 .40268 L
.36648 .4025 L
.36667 .40232 L
.36686 .40213 L
.36705 .40195 L
.36724 .40177 L
.36743 .40158 L
.36762 .4014 L
.36781 .40122 L
.368 .40103 L
.36819 .40085 L
.36838 .40067 L
.36857 .40048 L
.36876 .4003 L
.36895 .40012 L
.36914 .39993 L
.36933 .39975 L
.36952 .39957 L
.36971 .41536 L
.3699 .41518 L
Mistroke
.3701 .415 L
.37029 .41481 L
.37048 .43045 L
.37067 .43027 L
.37086 .43009 L
.37105 .42991 L
.37124 .42972 L
.37143 .42954 L
.37162 .42281 L
.37181 .42263 L
.372 .42244 L
.37219 .42226 L
.37238 .42208 L
.37257 .4219 L
.37276 .42172 L
.37295 .42153 L
.37314 .42135 L
.37333 .42117 L
.37352 .42099 L
.37371 .4208 L
.3739 .42062 L
.3741 .42044 L
.37429 .42026 L
.37448 .42008 L
.37467 .41989 L
.37486 .41971 L
.37505 .4352 L
.37524 .43502 L
.37543 .43484 L
.37562 .42216 L
.37581 .42198 L
.376 .4218 L
.37619 .43713 L
.37638 .43695 L
.37657 .43677 L
.37676 .43659 L
.37695 .42402 L
.37714 .41155 L
.37733 .41137 L
.37752 .4112 L
.37771 .41102 L
.3779 .41084 L
.3781 .42602 L
.37829 .42585 L
.37848 .42567 L
.37867 .42549 L
.37886 .42531 L
.37905 .42514 L
.37924 .41278 L
.37943 .41261 L
Mistroke
.37962 .42764 L
.37981 .42746 L
.38 .42729 L
.38019 .42711 L
.38038 .42694 L
.38057 .42676 L
.38076 .42659 L
.38095 .42641 L
.38114 .42624 L
.38133 .42606 L
.38152 .42588 L
.38171 .41943 L
.3819 .41925 L
.3821 .40707 L
.38229 .40689 L
.38248 .40672 L
.38267 .40655 L
.38286 .40637 L
.38305 .4062 L
.38324 .40603 L
.38343 .40585 L
.38362 .40568 L
.38381 .40551 L
.384 .40533 L
.38419 .40516 L
.38438 .39309 L
.38457 .39292 L
.38476 .39274 L
.38495 .39257 L
.38514 .3924 L
.38533 .39223 L
.38552 .40711 L
.38571 .40694 L
.3859 .40063 L
.3861 .40046 L
.38629 .40028 L
.38648 .40011 L
.38667 .39994 L
.38686 .39977 L
.38705 .41451 L
.38724 .40259 L
.38743 .40243 L
.38762 .40226 L
.38781 .40209 L
.388 .40192 L
.38819 .40175 L
.38838 .40158 L
.38857 .40141 L
.38876 .40124 L
.38895 .41584 L
Mistroke
.38914 .41567 L
.38933 .4155 L
.38952 .41533 L
.38971 .41516 L
.3899 .415 L
.3901 .41483 L
.39029 .41466 L
.39048 .41449 L
.39067 .41432 L
.39086 .41415 L
.39105 .41398 L
.39124 .41381 L
.39143 .41364 L
.39162 .41347 L
.39181 .4133 L
.392 .41313 L
.39219 .41296 L
.39238 .41279 L
.39257 .41263 L
.39276 .41246 L
.39295 .41229 L
.39314 .41212 L
.39333 .41195 L
.39352 .41178 L
.39371 .41161 L
.3939 .41144 L
.3941 .41127 L
.39429 .4111 L
.39448 .41094 L
.39467 .41077 L
.39486 .4106 L
.39505 .41043 L
.39524 .41026 L
.39543 .41009 L
.39562 .40992 L
.39581 .40975 L
.396 .40958 L
.39619 .40942 L
.39638 .40925 L
.39657 .40908 L
.39676 .40891 L
.39695 .40874 L
.39714 .40857 L
.39733 .4084 L
.39752 .40823 L
.39771 .40807 L
.3979 .4079 L
.3981 .40773 L
.39829 .40756 L
.39848 .40739 L
Mistroke
.39867 .40722 L
.39886 .40705 L
.39905 .40689 L
.39924 .40672 L
.39943 .40655 L
.39962 .40638 L
.39981 .40621 L
.4 .40604 L
.40019 .40588 L
.40038 .40571 L
.40057 .40554 L
.40076 .40537 L
.40095 .4052 L
.40114 .40503 L
.40133 .40487 L
.40152 .39313 L
.40171 .39297 L
.4019 .3928 L
.4021 .39263 L
.40229 .39247 L
.40248 .3923 L
.40267 .39213 L
.40286 .39197 L
.40305 .3918 L
.40324 .39163 L
.40343 .39147 L
.40362 .3913 L
.40381 .40576 L
.404 .4056 L
.40419 .40543 L
.40438 .40526 L
.40457 .39364 L
.40476 .39347 L
.40495 .39331 L
.40514 .39314 L
.40533 .39297 L
.40552 .39281 L
.40571 .39264 L
.4059 .39248 L
.4061 .39231 L
.40629 .39215 L
.40648 .39198 L
.40667 .39182 L
.40686 .39165 L
.40705 .39149 L
.40724 .39132 L
.40743 .39116 L
.40762 .39099 L
.40781 .39083 L
.408 .39066 L
Mistroke
.40819 .3905 L
.40838 .39033 L
.40857 .39017 L
.40876 .39 L
.40895 .38984 L
.40914 .38967 L
.40933 .38951 L
.40952 .38934 L
.40971 .38918 L
.4099 .37768 L
.4101 .37751 L
.41029 .37735 L
.41048 .39167 L
.41067 .39151 L
.41086 .39134 L
.41105 .39118 L
.41124 .39102 L
.41143 .39085 L
.41162 .39069 L
.41181 .39053 L
.412 .39036 L
.41219 .3902 L
.41238 .39004 L
.41257 .38987 L
.41276 .38971 L
.41295 .38955 L
.41314 .38938 L
.41333 .38922 L
.41352 .38906 L
.41371 .38889 L
.4139 .38873 L
.4141 .38856 L
.41429 .3884 L
.41448 .38824 L
.41467 .38807 L
.41486 .38791 L
.41505 .38191 L
.41524 .38175 L
.41543 .38158 L
.41562 .38142 L
.41581 .38126 L
.416 .36992 L
.41619 .36976 L
.41638 .3696 L
.41657 .36944 L
.41676 .36928 L
.41695 .36912 L
.41714 .36896 L
.41733 .36879 L
.41752 .36863 L
Mistroke
.41771 .36847 L
.4179 .36831 L
.4181 .36815 L
.41829 .36799 L
.41848 .36783 L
.41867 .36767 L
.41886 .35644 L
.41905 .35628 L
.41924 .35612 L
.41943 .35596 L
.41962 .3558 L
.41981 .35564 L
.42 .35548 L
.42019 .35532 L
.42038 .35516 L
.42057 .355 L
.42076 .35484 L
.42095 .35468 L
.42114 .35452 L
.42133 .35436 L
.42152 .3542 L
.42171 .35404 L
.4219 .34817 L
.4221 .34801 L
.42229 .33693 L
.42248 .33677 L
.42267 .33662 L
.42286 .33646 L
.42305 .3363 L
.42324 .33614 L
.42343 .33599 L
.42362 .33583 L
.42381 .32484 L
.424 .32468 L
.42419 .32453 L
.42438 .32437 L
.42457 .32421 L
.42476 .32406 L
.42495 .3239 L
.42514 .32374 L
.42533 .32359 L
.42552 .32343 L
.42571 .32328 L
.4259 .32312 L
.4261 .32296 L
.42629 .33713 L
.42648 .32623 L
.42667 .32607 L
.42686 .32592 L
.42705 .32576 L
Mistroke
.42724 .31495 L
.42743 .3148 L
.42762 .31464 L
.42781 .31449 L
.428 .30376 L
.42819 .3036 L
.42838 .30345 L
.42857 .3033 L
.42876 .30315 L
.42895 .30299 L
.42914 .30284 L
.42933 .30269 L
.42952 .30254 L
.42971 .30238 L
.4299 .30223 L
.4301 .30208 L
.43029 .29647 L
.43048 .29632 L
.43067 .29616 L
.43086 .29601 L
.43105 .29586 L
.43124 .29571 L
.43143 .29556 L
.43162 .28497 L
.43181 .28482 L
.432 .28466 L
.43219 .28451 L
.43238 .27897 L
.43257 .27882 L
.43276 .27867 L
.43295 .27852 L
.43314 .27837 L
.43333 .27822 L
.43352 .27807 L
.43371 .27792 L
.4339 .27777 L
.4341 .27762 L
.43429 .27747 L
.43448 .27732 L
.43467 .26687 L
.43486 .26672 L
.43505 .26657 L
.43524 .25619 L
.43543 .25604 L
.43562 .25589 L
.43581 .25575 L
.436 .2556 L
.43619 .25545 L
.43638 .2553 L
.43657 .25516 L
Mistroke
.43676 .25501 L
.43695 .25486 L
.43714 .25471 L
.43733 .25456 L
.43752 .25442 L
.43771 .26841 L
.4379 .26826 L
.4381 .25796 L
.43829 .25782 L
.43848 .25767 L
.43867 .25752 L
.43886 .25745 L
.43905 .2573 L
.43924 .25715 L
.43943 .25701 L
.43962 .25686 L
.43981 .25671 L
.44 .25657 L
.44019 .25642 L
.44038 .25628 L
.44057 .25613 L
.44076 .25598 L
.44095 .25584 L
.44114 .25569 L
.44133 .25554 L
.44152 .2554 L
.44171 .25525 L
.4419 .2551 L
.4421 .2449 L
.44229 .24476 L
.44248 .23942 L
.44267 .23927 L
.44286 .23913 L
.44305 .25298 L
.44324 .25284 L
.44343 .25269 L
.44362 .25255 L
.44381 .2524 L
.444 .25226 L
.44419 .25211 L
.44438 .25197 L
.44457 .25182 L
.44476 .25168 L
.44495 .26542 L
.44514 .26527 L
.44533 .26513 L
.44552 .26498 L
.44571 .26484 L
.4459 .26469 L
.4461 .26455 L
Mistroke
.44629 .2644 L
.44648 .26426 L
.44667 .26412 L
.44686 .25403 L
.44705 .25389 L
.44724 .25374 L
.44743 .2536 L
.44762 .26722 L
.44781 .26707 L
.448 .26693 L
.44819 .26679 L
.44838 .26664 L
.44857 .2665 L
.44876 .26636 L
.44895 .26621 L
.44914 .26607 L
.44933 .26593 L
.44952 .27943 L
.44971 .27929 L
.4499 .27915 L
.4501 .279 L
.45029 .27886 L
.45048 .27872 L
.45067 .27857 L
.45086 .27843 L
.45105 .27829 L
.45124 .27814 L
.45143 .278 L
.45162 .27786 L
.45181 .27771 L
.452 .27757 L
.45219 .29097 L
.45238 .28096 L
.45257 .28081 L
.45276 .27087 L
.45295 .27073 L
.45314 .27059 L
.45333 .27045 L
.45352 .27031 L
.45371 .27017 L
.4539 .27003 L
.4541 .2833 L
.45429 .28316 L
.45448 .28302 L
.45467 .29619 L
.45486 .29605 L
.45505 .29591 L
.45524 .29577 L
.45543 .29563 L
.45562 .29549 L
Mistroke
.45581 .29534 L
.456 .2952 L
.45619 .28533 L
.45638 .28519 L
.45657 .27539 L
.45676 .27525 L
.45695 .27511 L
.45714 .27497 L
.45733 .26524 L
.45752 .2651 L
.45771 .26496 L
.4579 .26483 L
.4581 .26469 L
.45829 .25959 L
.45848 .25946 L
.45867 .25932 L
.45886 .25918 L
.45905 .25904 L
.45924 .2589 L
.45943 .25877 L
.45962 .25863 L
.45981 .25849 L
.46 .25835 L
.46019 .25822 L
.46038 .2486 L
.46057 .24846 L
.46076 .24839 L
.46095 .24825 L
.46114 .24811 L
.46133 .24798 L
.46152 .24784 L
.46171 .2477 L
.4619 .24757 L
.4621 .24743 L
.46229 .24729 L
.46248 .24716 L
.46267 .24702 L
.46286 .24688 L
.46305 .24675 L
.46324 .24661 L
.46343 .24647 L
.46362 .24634 L
.46381 .2462 L
.464 .24606 L
.46419 .24593 L
.46438 .24579 L
.46457 .24565 L
.46476 .24552 L
.46495 .24538 L
.46514 .24524 L
Mistroke
.46533 .24511 L
.46552 .24497 L
.46571 .24483 L
.4659 .2447 L
.4661 .24456 L
.46629 .24442 L
.46648 .24429 L
.46667 .24415 L
.46686 .24401 L
.46705 .24388 L
.46724 .24374 L
.46743 .2436 L
.46762 .24347 L
.46781 .24333 L
.468 .2432 L
.46819 .24306 L
.46838 .24292 L
.46857 .24279 L
.46876 .24265 L
.46895 .24251 L
.46914 .24238 L
.46933 .24224 L
.46952 .2421 L
.46971 .24197 L
.4699 .24183 L
.4701 .2417 L
.47029 .24156 L
.47048 .24142 L
.47067 .24129 L
.47086 .24115 L
.47105 .24101 L
.47124 .24088 L
.47143 .24074 L
.47162 .24061 L
.47181 .24047 L
.472 .24033 L
.47219 .25337 L
.47238 .25323 L
.47257 .25309 L
.47276 .25296 L
.47295 .25282 L
.47314 .26575 L
.47333 .26562 L
.47352 .26548 L
.47371 .26534 L
.4739 .26521 L
.4741 .26507 L
.47429 .26493 L
.47448 .2648 L
.47467 .26466 L
Mistroke
.47486 .26453 L
.47505 .25503 L
.47524 .25489 L
.47543 .25476 L
.47562 .25462 L
.47581 .25449 L
.476 .25435 L
.47619 .25422 L
.47638 .25408 L
.47657 .25395 L
.47676 .25381 L
.47695 .25367 L
.47714 .25354 L
.47733 .2534 L
.47752 .25327 L
.47771 .25313 L
.4779 .253 L
.4781 .24803 L
.47829 .2479 L
.47848 .24776 L
.47867 .24763 L
.47886 .24749 L
.47905 .24736 L
.47924 .24723 L
.47943 .24709 L
.47962 .24696 L
.47981 .24682 L
.48 .23744 L
.48019 .23731 L
.48038 .23717 L
.48057 .23704 L
.48076 .23691 L
.48095 .23677 L
.48114 .23664 L
.48133 .23651 L
.48152 .23637 L
.48171 .23624 L
.4819 .23611 L
.4821 .23597 L
.48229 .24878 L
.48248 .23947 L
.48267 .23933 L
.48286 .2392 L
.48305 .23907 L
.48324 .25336 L
.48343 .25322 L
.48362 .25309 L
.48381 .25296 L
.484 .25283 L
.48419 .25269 L
Mistroke
.48438 .25256 L
.48457 .25243 L
.48476 .2523 L
.48495 .25216 L
.48514 .25203 L
.48533 .2519 L
.48552 .25177 L
.48571 .25163 L
.4859 .2515 L
.4861 .25137 L
.48629 .25124 L
.48648 .24199 L
.48667 .24186 L
.48686 .24173 L
.48705 .24159 L
.48724 .24146 L
.48743 .24133 L
.48762 .23649 L
.48781 .23636 L
.488 .23623 L
.48819 .2361 L
.48838 .23597 L
.48857 .23583 L
.48876 .2357 L
.48895 .23557 L
.48914 .23544 L
.48933 .23531 L
.48952 .23518 L
.48971 .23505 L
.4899 .23492 L
.4901 .23479 L
.49029 .23466 L
.49048 .24724 L
.49067 .24711 L
.49086 .24698 L
.49105 .24685 L
.49124 .24672 L
.49143 .24659 L
.49162 .25908 L
.49181 .25895 L
.492 .25882 L
.49219 .25869 L
.49238 .25856 L
.49257 .25842 L
.49276 .25829 L
.49295 .25816 L
.49314 .25803 L
.49333 .24889 L
.49352 .24876 L
.49371 .24863 L
Mistroke
.4939 .2485 L
.4941 .24837 L
.49429 .24824 L
.49448 .24811 L
.49467 .24798 L
.49486 .24785 L
.49505 .24772 L
.49524 .23864 L
.49543 .23852 L
.49562 .23839 L
.49581 .23826 L
.496 .23813 L
.49619 .25051 L
.49638 .25038 L
.49657 .25026 L
.49676 .25013 L
.49695 .25 L
.49714 .24987 L
.49733 .26216 L
.49752 .26203 L
.49771 .26191 L
.4979 .26178 L
.4981 .26165 L
.49829 .26152 L
.49848 .2525 L
.49867 .25237 L
.49886 .26457 L
.49905 .26444 L
.49924 .26431 L
.49943 .26418 L
.49962 .26406 L
.49981 .26393 L
.5 .27604 L
.50019 .27591 L
.50038 .27578 L
.50057 .27565 L
.50076 .27553 L
.50095 .2754 L
.50114 .27527 L
.50133 .27514 L
.50152 .27501 L
.50171 .27488 L
.5019 .27475 L
.5021 .27463 L
.50229 .2745 L
.50248 .27437 L
.50267 .28639 L
.50286 .28626 L
.50305 .28614 L
.50324 .28601 L
Mistroke
.50343 .28588 L
.50362 .28575 L
.50381 .28562 L
.504 .28549 L
.50419 .28537 L
.50438 .28524 L
.50457 .28511 L
.50476 .28498 L
.50495 .28485 L
.50514 .28472 L
.50533 .2846 L
.50552 .28447 L
.50571 .28434 L
.5059 .28421 L
.5061 .28408 L
.50629 .28395 L
.50648 .27501 L
.50667 .27488 L
.50686 .27475 L
.50705 .27462 L
.50724 .2745 L
.50743 .27437 L
.50762 .27424 L
.50781 .27411 L
.508 .27399 L
.50819 .27386 L
.50838 .27373 L
.50857 .27361 L
.50876 .28553 L
.50895 .28541 L
.50914 .28528 L
.50933 .28515 L
.50952 .28503 L
.50971 .29127 L
.5099 .29114 L
.5101 .29102 L
.51029 .29089 L
.51048 .29076 L
.51067 .29063 L
.51086 .29051 L
.51105 .29038 L
.51124 .29025 L
.51143 .29012 L
.51162 .29 L
.51181 .28987 L
.512 .28974 L
.51219 .28962 L
.51238 .28949 L
.51257 .28936 L
.51276 .28923 L
Mistroke
.51295 .28911 L
.51314 .28898 L
.51333 .28885 L
.51352 .28872 L
.51371 .2886 L
.5139 .3004 L
.5141 .30027 L
.51429 .30014 L
.51448 .30002 L
.51467 .29989 L
.51486 .29976 L
.51505 .29963 L
.51524 .29951 L
.51543 .29938 L
.51562 .29925 L
.51581 .29913 L
.516 .299 L
.51619 .29887 L
.51638 .29874 L
.51657 .29862 L
.51676 .29849 L
.51695 .29836 L
.51714 .29824 L
.51733 .29811 L
.51752 .29798 L
.51771 .29785 L
.5179 .29773 L
.5181 .30944 L
.51829 .30058 L
.51848 .30045 L
.51867 .30032 L
.51886 .3002 L
.51905 .30007 L
.51924 .29995 L
.51943 .31157 L
.51962 .31145 L
.51981 .31132 L
.52 .31119 L
.52019 .31107 L
.52038 .31094 L
.52057 .31081 L
.52076 .31069 L
.52095 .31056 L
.52114 .30176 L
.52133 .30163 L
.52152 .30151 L
.52171 .30138 L
.5219 .30125 L
.5221 .29251 L
.52229 .29238 L
Mistroke
.52248 .29226 L
.52267 .30379 L
.52286 .30367 L
.52305 .30354 L
.52324 .30342 L
.52343 .30329 L
.52362 .30317 L
.52381 .30304 L
.524 .30292 L
.52419 .3028 L
.52438 .30267 L
.52457 .30255 L
.52476 .30242 L
.52495 .3023 L
.52514 .30217 L
.52533 .30205 L
.52552 .30192 L
.52571 .31338 L
.5259 .31325 L
.5261 .31313 L
.52629 .31301 L
.52648 .31288 L
.52667 .31276 L
.52686 .31263 L
.52705 .31251 L
.52724 .31238 L
.52743 .31226 L
.52762 .31213 L
.52781 .31201 L
.528 .31188 L
.52819 .31176 L
.52838 .31164 L
.52857 .31151 L
.52876 .31139 L
.52895 .31126 L
.52914 .31114 L
.52933 .31101 L
.52952 .31089 L
.52971 .31077 L
.5299 .31064 L
.5301 .31052 L
.53029 .31039 L
.53048 .31027 L
.53067 .31014 L
.53086 .31002 L
.53105 .32139 L
.53124 .32127 L
.53143 .32115 L
.53162 .32102 L
.53181 .3209 L
Mistroke
.532 .32083 L
.53219 .32071 L
.53238 .32058 L
.53257 .32046 L
.53276 .32033 L
.53295 .32021 L
.53314 .32009 L
.53333 .31996 L
.53352 .31984 L
.53371 .31117 L
.5339 .31105 L
.5341 .31092 L
.53429 .3108 L
.53448 .30219 L
.53467 .30206 L
.53486 .30194 L
.53505 .30182 L
.53524 .30169 L
.53543 .29313 L
.53562 .29301 L
.53581 .29289 L
.536 .29277 L
.53619 .29265 L
.53638 .29253 L
.53657 .2924 L
.53676 .29228 L
.53695 .29216 L
.53714 .28366 L
.53733 .28354 L
.53752 .28341 L
.53771 .28329 L
.5379 .28317 L
.5381 .28305 L
.53829 .28293 L
.53848 .28281 L
.53867 .28269 L
.53886 .28257 L
.53905 .28245 L
.53924 .28232 L
.53943 .2936 L
.53962 .29348 L
.53981 .29336 L
.54 .29324 L
.54019 .29312 L
.54038 .293 L
.54057 .29288 L
.54076 .29275 L
.54095 .29263 L
.54114 .29251 L
.54133 .29239 L
Mistroke
.54152 .29227 L
.54171 .28383 L
.5419 .28371 L
.5421 .27929 L
.54229 .27917 L
.54248 .27905 L
.54267 .27893 L
.54286 .27056 L
.54305 .27044 L
.54324 .27032 L
.54343 .2702 L
.54362 .27009 L
.54381 .26997 L
.544 .26985 L
.54419 .26973 L
.54438 .26961 L
.54457 .26949 L
.54476 .26937 L
.54495 .26925 L
.54514 .26913 L
.54533 .26901 L
.54552 .2689 L
.54571 .26878 L
.5459 .26866 L
.5461 .26854 L
.54629 .26842 L
.54648 .2683 L
.54667 .26818 L
.54686 .26806 L
.54705 .26794 L
.54724 .26782 L
.54743 .26771 L
.54762 .26759 L
.54781 .26747 L
.548 .25917 L
.54819 .25905 L
.54838 .25893 L
.54857 .25882 L
.54876 .2587 L
.54895 .25858 L
.54914 .25846 L
.54933 .25834 L
.54952 .25823 L
.54971 .25811 L
.5499 .25799 L
.5501 .25787 L
.55029 .25775 L
.55048 .25764 L
.55067 .25752 L
.55086 .2574 L
Mistroke
.55105 .25728 L
.55124 .25716 L
.55143 .25705 L
.55162 .25693 L
.55181 .25681 L
.552 .25669 L
.55219 .26787 L
.55238 .26775 L
.55257 .26764 L
.55276 .26752 L
.55295 .2674 L
.55314 .26728 L
.55333 .26716 L
.55352 .26705 L
.55371 .26693 L
.5539 .26681 L
.5541 .26669 L
.55429 .26657 L
.55448 .26646 L
.55467 .27756 L
.55486 .27744 L
.55505 .28848 L
.55524 .28836 L
.55543 .28824 L
.55562 .28812 L
.55581 .288 L
.556 .28789 L
.55619 .28777 L
.55638 .28765 L
.55657 .28753 L
.55676 .28741 L
.55695 .2873 L
.55714 .28718 L
.55733 .28706 L
.55752 .28694 L
.55771 .28682 L
.5579 .28671 L
.5581 .28659 L
.55829 .27836 L
.55848 .27824 L
.55867 .27812 L
.55886 .27801 L
.55905 .27789 L
.55924 .27777 L
.55943 .26959 L
.55962 .26948 L
.55981 .26936 L
.56 .26924 L
.56019 .26913 L
.56038 .26901 L
Mistroke
.56057 .26088 L
.56076 .26077 L
.56095 .26065 L
.56114 .26054 L
.56133 .26042 L
.56152 .2603 L
.56171 .25222 L
.5619 .25211 L
.5621 .25199 L
.56229 .26293 L
.56248 .26282 L
.56267 .2627 L
.56286 .27492 L
.56305 .27481 L
.56324 .27469 L
.56343 .27458 L
.56362 .27446 L
.56381 .26643 L
.564 .26631 L
.56419 .2662 L
.56438 .27698 L
.56457 .27687 L
.56476 .27675 L
.56495 .27664 L
.56514 .27652 L
.56533 .27641 L
.56552 .27629 L
.56571 .27618 L
.5659 .27606 L
.5661 .27185 L
.56629 .27174 L
.56648 .27163 L
.56667 .27151 L
.56686 .2714 L
.56705 .27128 L
.56724 .27117 L
.56743 .27105 L
.56762 .27094 L
.56781 .27083 L
.568 .27071 L
.56819 .2706 L
.56838 .27048 L
.56857 .27037 L
.56876 .27026 L
.56895 .27014 L
.56914 .28085 L
.56933 .27289 L
.56952 .27278 L
.56971 .27266 L
.5699 .27255 L
Mistroke
.5701 .27244 L
.57029 .27232 L
.57048 .27221 L
.57067 .28417 L
.57086 .28406 L
.57105 .27989 L
.57124 .27977 L
.57143 .27966 L
.57162 .27955 L
.57181 .27944 L
.572 .27932 L
.57219 .27921 L
.57238 .2791 L
.57257 .27898 L
.57276 .27887 L
.57295 .27876 L
.57314 .27864 L
.57333 .27853 L
.57352 .27842 L
.57371 .2783 L
.5739 .27042 L
.5741 .27031 L
.57429 .28086 L
.57448 .28075 L
.57467 .28064 L
.57486 .28052 L
.57505 .28041 L
.57524 .2803 L
.57543 .29079 L
.57562 .29068 L
.57581 .29056 L
.576 .29045 L
.57619 .29034 L
.57638 .29023 L
.57657 .29011 L
.57676 .29 L
.57695 .28989 L
.57714 .28978 L
.57733 .28966 L
.57752 .28955 L
.57771 .28944 L
.5779 .28933 L
.5781 .28921 L
.57829 .2891 L
.57848 .28899 L
.57867 .28888 L
.57886 .28877 L
.57905 .28865 L
.57924 .28854 L
.57943 .28843 L
Mistroke
.57962 .28832 L
.57981 .2882 L
.58 .28809 L
.58019 .28798 L
.58038 .28787 L
.58057 .28776 L
.58076 .28764 L
.58095 .27981 L
.58114 .2797 L
.58133 .27959 L
.58152 .27948 L
.58171 .27937 L
.5819 .27925 L
.5821 .27914 L
.58229 .27903 L
.58248 .27892 L
.58267 .27881 L
.58286 .2787 L
.58305 .27859 L
.58324 .27847 L
.58343 .27836 L
.58362 .27825 L
.58381 .27814 L
.584 .28856 L
.58419 .28077 L
.58438 .28066 L
.58457 .27292 L
.58476 .27281 L
.58495 .2727 L
.58514 .27259 L
.58533 .28293 L
.58552 .28282 L
.58571 .27512 L
.5859 .27501 L
.5861 .2749 L
.58629 .27479 L
.58648 .27468 L
.58667 .27457 L
.58686 .27446 L
.58705 .27436 L
.58724 .27425 L
.58743 .27414 L
.58762 .27403 L
.58781 .27392 L
.588 .28419 L
.58819 .28408 L
.58838 .28397 L
.58857 .28386 L
.58876 .28375 L
.58895 .28364 L
Mistroke
.58914 .28353 L
.58933 .28342 L
.58952 .28332 L
.58971 .28321 L
.5899 .2831 L
.5901 .28299 L
.59029 .28288 L
.59048 .29309 L
.59067 .29298 L
.59086 .29287 L
.59105 .29276 L
.59124 .29265 L
.59143 .29254 L
.59162 .29243 L
.59181 .29232 L
.592 .29221 L
.59219 .2921 L
.59238 .28807 L
.59257 .28796 L
.59276 .28785 L
.59295 .298 L
.59314 .29789 L
.59333 .29778 L
.59352 .29767 L
.59371 .29756 L
.5939 .29745 L
.5941 .29734 L
.59429 .30743 L
.59448 .30732 L
.59467 .31735 L
.59486 .31724 L
.59505 .31713 L
.59524 .31702 L
.59543 .31691 L
.59562 .3168 L
.59581 .31669 L
.596 .31658 L
.59619 .31647 L
.59638 .31636 L
.59657 .31625 L
.59676 .30863 L
.59695 .30852 L
.59714 .30841 L
.59733 .3083 L
.59752 .30819 L
.59771 .30808 L
.5979 .30797 L
.5981 .30786 L
.59829 .30387 L
.59848 .30376 L
Mistroke
.59867 .30365 L
.59886 .30354 L
.59905 .30344 L
.59924 .30333 L
.59943 .30322 L
.59962 .30311 L
.59981 .31307 L
.6 .31296 L
.60019 .31285 L
.60038 .31274 L
.60057 .31264 L
.60076 .31253 L
.60095 .31242 L
.60114 .31231 L
.60133 .3122 L
.60152 .31209 L
.60171 .31199 L
.6019 .31193 L
.6021 .31182 L
.60229 .31171 L
.60248 .3116 L
.60267 .3115 L
.60286 .31139 L
.60305 .31128 L
.60324 .31117 L
.60343 .31106 L
.60362 .31096 L
.60381 .31085 L
.604 .31074 L
.60419 .31063 L
.60438 .31052 L
.60457 .31041 L
.60476 .31031 L
.60495 .3102 L
.60514 .31009 L
.60533 .30998 L
.60552 .30987 L
.60571 .30977 L
.6059 .30222 L
.6061 .30211 L
.60629 .30201 L
.60648 .3019 L
.60667 .30179 L
.60686 .30168 L
.60705 .30158 L
.60724 .30147 L
.60743 .30136 L
.60762 .30125 L
.60781 .30115 L
.608 .30104 L
Mistroke
.60819 .31093 L
.60838 .31083 L
.60857 .31072 L
.60876 .31061 L
.60895 .32044 L
.60914 .32034 L
.60933 .32023 L
.60952 .32012 L
.60971 .32001 L
.6099 .31991 L
.6101 .3198 L
.61029 .31969 L
.61048 .31959 L
.61067 .31948 L
.61086 .31937 L
.61105 .31926 L
.61124 .31916 L
.61143 .31905 L
.61162 .31894 L
.61181 .32872 L
.612 .32861 L
.61219 .3285 L
.61238 .321 L
.61257 .3209 L
.61276 .33061 L
.61295 .34027 L
.61314 .34016 L
.61333 .34005 L
.61352 .33995 L
.61371 .33984 L
.6139 .33973 L
.6141 .33963 L
.61429 .33952 L
.61448 .33941 L
.61467 .33931 L
.61486 .3392 L
.61505 .33909 L
.61524 .33899 L
.61543 .33888 L
.61562 .33877 L
.61581 .33867 L
.616 .33856 L
.61619 .34816 L
.61638 .34805 L
.61657 .34795 L
.61676 .34784 L
.61695 .34773 L
.61714 .35728 L
.61733 .35717 L
.61752 .35706 L
Mistroke
.61771 .35696 L
.6179 .35685 L
.6181 .35674 L
.61829 .35663 L
.61848 .35653 L
.61867 .35642 L
.61886 .35631 L
.61905 .35621 L
.61924 .3561 L
.61943 .35599 L
.61962 .35589 L
.61981 .35578 L
.62 .35567 L
.62019 .35557 L
.62038 .35546 L
.62057 .35535 L
.62076 .35525 L
.62095 .35514 L
.62114 .35503 L
.62133 .35493 L
.62152 .35482 L
.62171 .35471 L
.6219 .35461 L
.6221 .3545 L
.62229 .35439 L
.62248 .35429 L
.62267 .35418 L
.62286 .35407 L
.62305 .35397 L
.62324 .35386 L
.62343 .35375 L
.62362 .35365 L
.62381 .35354 L
.624 .35343 L
.62419 .35333 L
.62438 .35322 L
.62457 .34578 L
.62476 .34568 L
.62495 .34557 L
.62514 .34547 L
.62533 .34536 L
.62552 .34146 L
.62571 .34135 L
.6259 .34125 L
.6261 .34114 L
.62629 .34104 L
.62648 .34093 L
.62667 .34083 L
.62686 .34072 L
.62705 .34062 L
Mistroke
.62724 .34051 L
.62743 .3404 L
.62762 .33303 L
.62781 .33293 L
.628 .33282 L
.62819 .33272 L
.62838 .33261 L
.62857 .33251 L
.62876 .3324 L
.62895 .3323 L
.62914 .33219 L
.62933 .33209 L
.62952 .33198 L
.62971 .32465 L
.6299 .32454 L
.6301 .32444 L
.63029 .32434 L
.63048 .32423 L
.63067 .32413 L
.63086 .32402 L
.63105 .32392 L
.63124 .32381 L
.63143 .32371 L
.63162 .3236 L
.63181 .3235 L
.632 .3234 L
.63219 .32329 L
.63238 .32319 L
.63257 .32308 L
.63276 .32298 L
.63295 .32287 L
.63314 .32277 L
.63333 .32267 L
.63352 .32256 L
.63371 .32246 L
.6339 .33193 L
.6341 .33183 L
.63429 .33172 L
.63448 .33162 L
.63467 .33152 L
.63486 .33141 L
.63505 .33131 L
.63524 .3312 L
.63543 .3311 L
.63562 .32382 L
.63581 .32371 L
.636 .32361 L
.63619 .32351 L
.63638 .3234 L
.63657 .3233 L
Mistroke
.63676 .32319 L
.63695 .32309 L
.63714 .32299 L
.63733 .32288 L
.63752 .32278 L
.63771 .32268 L
.6379 .32257 L
.6381 .32247 L
.63829 .31523 L
.63848 .30803 L
.63867 .30792 L
.63886 .30782 L
.63905 .30772 L
.63924 .30762 L
.63943 .30751 L
.63962 .30741 L
.63981 .30731 L
.64 .30721 L
.64019 .3071 L
.64038 .307 L
.64057 .3069 L
.64076 .3068 L
.64095 .30669 L
.64114 .30659 L
.64133 .316 L
.64152 .3159 L
.64171 .3158 L
.6419 .31569 L
.6421 .31559 L
.64229 .31549 L
.64248 .31539 L
.64267 .31528 L
.64286 .31518 L
.64305 .31508 L
.64324 .31498 L
.64343 .31487 L
.64362 .31477 L
.64381 .31467 L
.644 .31457 L
.64419 .31446 L
.64438 .31436 L
.64457 .31426 L
.64476 .31416 L
.64495 .31405 L
.64514 .31395 L
.64533 .31385 L
.64552 .31375 L
.64571 .31364 L
.6459 .31354 L
.6461 .31344 L
Mistroke
.64629 .31334 L
.64648 .31324 L
.64667 .32259 L
.64686 .32249 L
.64705 .32239 L
.64724 .32229 L
.64743 .32218 L
.64762 .32208 L
.64781 .32198 L
.648 .32188 L
.64819 .32177 L
.64838 .32167 L
.64857 .32157 L
.64876 .32147 L
.64895 .32136 L
.64914 .32126 L
.64933 .32116 L
.64952 .31739 L
.64971 .31729 L
.6499 .31719 L
.6501 .31709 L
.65029 .31699 L
.65048 .31688 L
.65067 .31678 L
.65086 .31668 L
.65105 .31658 L
.65124 .32588 L
.65143 .32578 L
.65162 .31865 L
.65181 .31156 L
.652 .31146 L
.65219 .31136 L
.65238 .31126 L
.65257 .31116 L
.65276 .31106 L
.65295 .31096 L
.65314 .3202 L
.65333 .3201 L
.65352 .32 L
.65371 .3199 L
.6539 .3198 L
.6541 .3197 L
.65429 .3196 L
.65448 .3195 L
.65467 .31939 L
.65486 .31929 L
.65505 .32849 L
.65524 .32477 L
.65543 .32467 L
.65562 .32457 L
Mistroke
.65581 .32447 L
.656 .32437 L
.65619 .32427 L
.65638 .32417 L
.65657 .32406 L
.65676 .32396 L
.65695 .32386 L
.65714 .32376 L
.65733 .32366 L
.65752 .32356 L
.65771 .32346 L
.6579 .32336 L
.6581 .32326 L
.65829 .32316 L
.65848 .3323 L
.65867 .3322 L
.65886 .32849 L
.65905 .32839 L
.65924 .32829 L
.65943 .32819 L
.65962 .32809 L
.65981 .32799 L
.66 .32789 L
.66019 .32779 L
.66038 .32769 L
.66057 .32759 L
.66076 .32058 L
.66095 .32048 L
.66114 .32038 L
.66133 .32028 L
.66152 .32018 L
.66171 .32008 L
.6619 .32917 L
.6621 .32907 L
.66229 .32897 L
.66248 .32887 L
.66267 .32519 L
.66286 .3251 L
.66305 .325 L
.66324 .3249 L
.66343 .3248 L
.66362 .3247 L
.66381 .3246 L
.664 .3245 L
.66419 .3244 L
.66438 .3243 L
.66457 .3242 L
.66476 .3241 L
.66495 .324 L
.66514 .3239 L
Mistroke
.66533 .3238 L
.66552 .3237 L
.66571 .3236 L
.6659 .3235 L
.6661 .3234 L
.66629 .3233 L
.66648 .3232 L
.66667 .3231 L
.66686 .32301 L
.66705 .32291 L
.66724 .32281 L
.66743 .32271 L
.66762 .32261 L
.66781 .32251 L
.668 .31557 L
.66819 .31547 L
.66838 .31537 L
.66857 .31527 L
.66876 .31517 L
.66895 .31507 L
.66914 .31497 L
.66933 .324 L
.66952 .33298 L
.66971 .33288 L
.6699 .33278 L
.6701 .33269 L
.67029 .33259 L
.67048 .33249 L
.67067 .33239 L
.67086 .33229 L
.67105 .33219 L
.67124 .33209 L
.67143 .32518 L
.67162 .32509 L
.67181 .32499 L
.672 .32489 L
.67219 .32479 L
.67238 .32469 L
.67257 .32459 L
.67276 .3245 L
.67295 .3244 L
.67314 .3243 L
.67333 .3242 L
.67352 .3241 L
.67371 .324 L
.6739 .32391 L
.6741 .32381 L
.67429 .32371 L
.67448 .32361 L
.67467 .32351 L
Mistroke
.67486 .33244 L
.67505 .33234 L
.67524 .33224 L
.67543 .34112 L
.67562 .34103 L
.67581 .34093 L
.676 .34083 L
.67619 .34073 L
.67638 .34063 L
.67657 .34053 L
.67676 .34044 L
.67695 .34034 L
.67714 .34024 L
.67733 .34014 L
.67752 .34004 L
.67771 .33318 L
.6779 .33308 L
.6781 .33298 L
.67829 .32615 L
.67848 .32605 L
.67867 .31925 L
.67886 .31915 L
.67905 .31906 L
.67924 .31896 L
.67943 .31886 L
.67962 .31877 L
.67981 .31867 L
.68 .31857 L
.68019 .31848 L
.68038 .31838 L
.68057 .31828 L
.68076 .31819 L
.68095 .31809 L
.68114 .31799 L
.68133 .31789 L
.68152 .3178 L
.68171 .3177 L
.6819 .3176 L
.6821 .31751 L
.68229 .31741 L
.68248 .31731 L
.68267 .31722 L
.68286 .31712 L
.68305 .31702 L
.68324 .32585 L
.68343 .32575 L
.68362 .32565 L
.68381 .32556 L
.684 .32546 L
.68419 .32536 L
Mistroke
.68438 .32527 L
.68457 .32517 L
.68476 .32507 L
.68495 .32498 L
.68514 .32488 L
.68533 .32478 L
.68552 .32469 L
.68571 .32459 L
.6859 .32449 L
.6861 .3244 L
.68629 .3243 L
.68648 .3242 L
.68667 .32411 L
.68686 .32401 L
.68705 .32391 L
.68724 .32382 L
.68743 .32372 L
.68762 .32362 L
.68781 .32353 L
.688 .32343 L
.68819 .32333 L
.68838 .32324 L
.68857 .32314 L
.68876 .32304 L
.68895 .32295 L
.68914 .32285 L
.68933 .32275 L
.68952 .32266 L
.68971 .32256 L
.6899 .32246 L
.6901 .32237 L
.69029 .32227 L
.69048 .33105 L
.69067 .33095 L
.69086 .33085 L
.69105 .33076 L
.69124 .33948 L
.69143 .33939 L
.69162 .33929 L
.69181 .3392 L
.692 .3391 L
.69219 .339 L
.69238 .33891 L
.69257 .33881 L
.69276 .33206 L
.69295 .32535 L
.69314 .32183 L
.69333 .32174 L
.69352 .32164 L
.69371 .32154 L
Mistroke
.6939 .33022 L
.6941 .33013 L
.69429 .33003 L
.69448 .32993 L
.69467 .33857 L
.69486 .33847 L
.69505 .33838 L
.69524 .33828 L
.69543 .33818 L
.69562 .33809 L
.69581 .33799 L
.696 .3379 L
.69619 .3378 L
.69638 .33771 L
.69657 .33761 L
.69676 .33752 L
.69695 .33085 L
.69714 .33076 L
.69733 .33066 L
.69752 .33057 L
.69771 .33047 L
.6979 .33038 L
.6981 .33028 L
.69829 .33019 L
.69848 .33009 L
.69867 .33 L
.69886 .3299 L
.69905 .32981 L
.69924 .32971 L
.69943 .32962 L
.69962 .32952 L
.69981 .32943 L
.7 .32933 L
.70019 .32924 L
.70038 .33782 L
.70057 .33773 L
.70076 .33763 L
.70095 .33101 L
.70114 .33091 L
.70133 .33082 L
.70152 .33072 L
.70171 .33063 L
.7019 .33053 L
.7021 .33044 L
.70229 .33034 L
.70248 .33888 L
.70267 .33879 L
.70286 .33869 L
.70305 .3386 L
.70324 .3385 L
Mistroke
.70343 .33841 L
.70362 .33831 L
.70381 .33822 L
.704 .33813 L
.70419 .33803 L
.70438 .33794 L
.70457 .33784 L
.70476 .33125 L
.70495 .33116 L
.70514 .33106 L
.70533 .33097 L
.70552 .33087 L
.70571 .33078 L
.7059 .33069 L
.7061 .33059 L
.70629 .3305 L
.70648 .33041 L
.70667 .33031 L
.70686 .33022 L
.70705 .33012 L
.70724 .33003 L
.70743 .32994 L
.70762 .32984 L
.70781 .32975 L
.708 .32965 L
.70819 .32956 L
.70838 .32947 L
.70857 .32937 L
.70876 .33786 L
.70895 .33777 L
.70914 .33767 L
.70933 .33758 L
.70952 .33749 L
.70971 .33739 L
.7099 .3373 L
.7101 .3372 L
.71029 .33711 L
.71048 .33702 L
.71067 .33692 L
.71086 .33683 L
.71105 .33674 L
.71124 .33664 L
.71143 .33655 L
.71162 .34499 L
.71181 .3449 L
.712 .34988 L
.71219 .34978 L
.71238 .34969 L
.71257 .3496 L
.71276 .3495 L
Mistroke
.71295 .34941 L
.71314 .34932 L
.71333 .34922 L
.71352 .34913 L
.71371 .34903 L
.7139 .34894 L
.7141 .34885 L
.71429 .34875 L
.71448 .34866 L
.71467 .34857 L
.71486 .34847 L
.71505 .34838 L
.71524 .34828 L
.71543 .34819 L
.71562 .3481 L
.71581 .348 L
.716 .34791 L
.71619 .34782 L
.71638 .34772 L
.71657 .34763 L
.71676 .34753 L
.71695 .34744 L
.71714 .34735 L
.71733 .34725 L
.71752 .34716 L
.71771 .34707 L
.7179 .34697 L
.7181 .34688 L
.71829 .34679 L
.71848 .34669 L
.71867 .3466 L
.71886 .3465 L
.71905 .33997 L
.71924 .33987 L
.71943 .33978 L
.71962 .33969 L
.71981 .33959 L
.72 .3395 L
.72019 .33941 L
.72038 .33932 L
.72057 .33922 L
.72076 .33913 L
.72095 .33904 L
.72114 .33894 L
.72133 .33885 L
.72152 .33876 L
.72171 .33225 L
.7219 .33216 L
.7221 .33207 L
.72229 .34044 L
Mistroke
.72248 .34035 L
.72267 .34025 L
.72286 .34016 L
.72305 .34007 L
.72324 .33997 L
.72343 .33988 L
.72362 .34821 L
.72381 .3448 L
.724 .34471 L
.72419 .34962 L
.72438 .34952 L
.72457 .34943 L
.72476 .34934 L
.72495 .34925 L
.72514 .34915 L
.72533 .35741 L
.72552 .35732 L
.72571 .35723 L
.7259 .35714 L
.7261 .35704 L
.72629 .35695 L
.72648 .35686 L
.72667 .35677 L
.72686 .35667 L
.72705 .35658 L
.72724 .35649 L
.72743 .3564 L
.72762 .3563 L
.72781 .35621 L
.728 .34976 L
.72819 .34966 L
.72838 .34957 L
.72857 .34948 L
.72876 .34939 L
.72895 .3493 L
.72914 .3492 L
.72933 .34911 L
.72952 .34902 L
.72971 .34893 L
.7299 .34251 L
.7301 .34241 L
.73029 .35063 L
.73048 .35054 L
.73067 .35044 L
.73086 .35035 L
.73105 .35026 L
.73124 .35017 L
.73143 .35008 L
.73162 .34999 L
.73181 .34989 L
Mistroke
.732 .3498 L
.73219 .34971 L
.73238 .34962 L
.73257 .34953 L
.73276 .3577 L
.73295 .35761 L
.73314 .35752 L
.73333 .35743 L
.73352 .36556 L
.73371 .36547 L
.7339 .36537 L
.7341 .36528 L
.73429 .36519 L
.73448 .3651 L
.73467 .36501 L
.73486 .36492 L
.73505 .36483 L
.73524 .36473 L
.73543 .36464 L
.73562 .36455 L
.73581 .37264 L
.736 .36927 L
.73619 .36918 L
.73638 .36909 L
.73657 .369 L
.73676 .36262 L
.73695 .36253 L
.73714 .36244 L
.73733 .36235 L
.73752 .36226 L
.73771 .36217 L
.7379 .36208 L
.7381 .36199 L
.73829 .3619 L
.73848 .35555 L
.73867 .35546 L
.73886 .35537 L
.73905 .36342 L
.73924 .36333 L
.73943 .36324 L
.73962 .36315 L
.73981 .36306 L
.74 .36297 L
.74019 .36287 L
.74038 .36278 L
.74057 .36269 L
.74076 .3626 L
.74095 .36251 L
.74114 .36242 L
.74133 .36233 L
Mistroke
.74152 .36224 L
.74171 .36215 L
.7419 .36206 L
.7421 .36197 L
.74229 .36188 L
.74248 .36179 L
.74267 .36174 L
.74286 .36165 L
.74305 .36156 L
.74324 .36147 L
.74343 .36138 L
.74362 .36129 L
.74381 .3612 L
.744 .36111 L
.74419 .36102 L
.74438 .36093 L
.74457 .36084 L
.74476 .36075 L
.74495 .36066 L
.74514 .36057 L
.74533 .36048 L
.74552 .36039 L
.74571 .3603 L
.7459 .36021 L
.7461 .36012 L
.74629 .36003 L
.74648 .35994 L
.74667 .35985 L
.74686 .35976 L
.74705 .35967 L
.74724 .35958 L
.74743 .35949 L
.74762 .35318 L
.74781 .35309 L
.748 .353 L
.74819 .35291 L
.74838 .35282 L
.74857 .35273 L
.74876 .35264 L
.74895 .35255 L
.74914 .34628 L
.74933 .34619 L
.74952 .3429 L
.74971 .34281 L
.7499 .34272 L
.7501 .34264 L
.75029 .34255 L
.75048 .34246 L
.75067 .34237 L
.75086 .34228 L
Mistroke
.75105 .34219 L
.75124 .3421 L
.75143 .34201 L
.75162 .34192 L
.75181 .34183 L
.752 .34174 L
.75219 .34165 L
.75238 .34157 L
.75257 .33534 L
.75276 .33525 L
.75295 .33516 L
.75314 .33507 L
.75333 .33499 L
.75352 .3349 L
.75371 .33481 L
.7539 .33472 L
.7541 .33463 L
.75429 .34263 L
.75448 .34254 L
.75467 .34245 L
.75486 .34236 L
.75505 .34227 L
.75524 .34219 L
.75543 .3421 L
.75562 .34201 L
.75581 .34192 L
.756 .34988 L
.75619 .34979 L
.75638 .3497 L
.75657 .35762 L
.75676 .35753 L
.75695 .35744 L
.75714 .35736 L
.75733 .35727 L
.75752 .35718 L
.75771 .35709 L
.7579 .357 L
.7581 .35691 L
.75829 .35682 L
.75848 .35678 L
.75867 .35669 L
.75886 .3566 L
.75905 .35651 L
.75924 .35642 L
.75943 .35633 L
.75962 .35625 L
.75981 .35616 L
.76 .35607 L
.76019 .35598 L
.76038 .35589 L
Mistroke
.76057 .3558 L
.76076 .35254 L
.76095 .35245 L
.76114 .35236 L
.76133 .35228 L
.76152 .35219 L
.76171 .3521 L
.7619 .35201 L
.7621 .35192 L
.76229 .35183 L
.76248 .35175 L
.76267 .35166 L
.76286 .35157 L
.76305 .35148 L
.76324 .35139 L
.76343 .34814 L
.76362 .34805 L
.76381 .3419 L
.764 .34181 L
.76419 .34172 L
.76438 .34163 L
.76457 .3355 L
.76476 .33541 L
.76495 .32931 L
.76514 .32922 L
.76533 .32914 L
.76552 .32905 L
.76571 .32896 L
.7659 .32888 L
.7661 .32879 L
.76629 .3287 L
.76648 .32861 L
.76667 .32853 L
.76686 .32245 L
.76705 .32237 L
.76724 .32228 L
.76743 .32219 L
.76762 .32211 L
.76781 .32202 L
.768 .32193 L
.76819 .32185 L
.76838 .32176 L
.76857 .32963 L
.76876 .32955 L
.76895 .32946 L
.76914 .32937 L
.76933 .32929 L
.76952 .3292 L
.76971 .32911 L
.7699 .32903 L
Mistroke
.7701 .32894 L
.77029 .32885 L
.77048 .32877 L
.77067 .32868 L
.77086 .32859 L
.77105 .32851 L
.77124 .32246 L
.77143 .32238 L
.77162 .32229 L
.77181 .3222 L
.772 .32212 L
.77219 .32203 L
.77238 .32195 L
.77257 .32186 L
.77276 .32177 L
.77295 .32169 L
.77314 .3216 L
.77333 .32943 L
.77352 .32935 L
.77371 .32926 L
.7739 .32917 L
.7741 .32909 L
.77429 .329 L
.77448 .3368 L
.77467 .33671 L
.77486 .33662 L
.77505 .33654 L
.77524 .33645 L
.77543 .33637 L
.77562 .34412 L
.77581 .33811 L
.776 .33495 L
.77619 .33486 L
.77638 .33478 L
.77657 .33469 L
.77676 .33461 L
.77695 .33452 L
.77714 .32854 L
.77733 .32846 L
.77752 .3225 L
.77771 .32242 L
.7779 .32233 L
.7781 .32225 L
.77829 .32216 L
.77848 .32208 L
.77867 .32199 L
.77886 .32191 L
.77905 .32182 L
.77924 .32174 L
.77943 .32166 L
Mistroke
.77962 .32157 L
.77981 .32149 L
.78 .3214 L
.78019 .32132 L
.78038 .32123 L
.78057 .32115 L
.78076 .32106 L
.78095 .32098 L
.78114 .32089 L
.78133 .32081 L
.78152 .32072 L
.78171 .32064 L
.7819 .32055 L
.7821 .32047 L
.78229 .32038 L
.78248 .3203 L
.78267 .32021 L
.78286 .32013 L
.78305 .32004 L
.78324 .31996 L
.78343 .31987 L
.78362 .31979 L
.78381 .31971 L
.784 .31962 L
.78419 .31954 L
.78438 .31945 L
.78457 .31937 L
.78476 .31928 L
.78495 .3192 L
.78514 .31911 L
.78533 .31903 L
.78552 .31894 L
.78571 .31886 L
.7859 .31877 L
.7861 .31869 L
.78629 .3186 L
.78648 .31852 L
.78667 .31844 L
.78686 .31835 L
.78705 .31827 L
.78724 .31818 L
.78743 .3181 L
.78762 .31801 L
.78781 .31793 L
.788 .31784 L
.78819 .31776 L
.78838 .31767 L
.78857 .31759 L
.78876 .3175 L
.78895 .31742 L
Mistroke
.78914 .31734 L
.78933 .31725 L
.78952 .31134 L
.78971 .31126 L
.7899 .31117 L
.7901 .31109 L
.79029 .31101 L
.79048 .31872 L
.79067 .31863 L
.79086 .31855 L
.79105 .31545 L
.79124 .31537 L
.79143 .31528 L
.79162 .3152 L
.79181 .31511 L
.792 .31503 L
.79219 .31495 L
.79238 .30908 L
.79257 .30899 L
.79276 .30891 L
.79295 .30883 L
.79314 .30874 L
.79333 .30866 L
.79352 .31729 L
.79371 .3172 L
.7939 .31712 L
.7941 .31704 L
.79429 .31695 L
.79448 .31687 L
.79467 .31678 L
.79486 .3167 L
.79505 .31662 L
.79524 .31653 L
.79543 .31645 L
.79562 .31637 L
.79581 .31628 L
.796 .31044 L
.79619 .31036 L
.79638 .31027 L
.79657 .31019 L
.79676 .31011 L
.79695 .31002 L
.79714 .30994 L
.79733 .30986 L
.79752 .30977 L
.79771 .30969 L
.7979 .30961 L
.7981 .30952 L
.79829 .30944 L
.79848 .30936 L
Mistroke
.79867 .30928 L
.79886 .30919 L
.79905 .30911 L
.79924 .30903 L
.79943 .30894 L
.79962 .30886 L
.79981 .30878 L
.8 .30869 L
.80019 .30861 L
.80038 .30853 L
.80057 .30844 L
.80076 .3084 L
.80095 .30832 L
.80114 .30823 L
.80133 .30815 L
.80152 .30807 L
.80171 .30798 L
.8019 .31561 L
.8021 .31553 L
.80229 .31545 L
.80248 .31536 L
.80267 .31528 L
.80286 .3152 L
.80305 .31511 L
.80324 .31503 L
.80343 .31495 L
.80362 .31486 L
.80381 .31478 L
.804 .3147 L
.80419 .31461 L
.80438 .31453 L
.80457 .31445 L
.80476 .31437 L
.80495 .31428 L
.80514 .3142 L
.80533 .31412 L
.80552 .31403 L
.80571 .31395 L
.8059 .31387 L
.8061 .31378 L
.80629 .32138 L
.80648 .32129 L
.80667 .31549 L
.80686 .31244 L
.80705 .31236 L
.80724 .31228 L
.80743 .30651 L
.80762 .30643 L
.80781 .30635 L
.808 .30332 L
Mistroke
.80819 .29759 L
.80838 .29751 L
.80857 .29742 L
.80876 .29734 L
.80895 .29726 L
.80914 .29718 L
.80933 .2971 L
.80952 .29702 L
.80971 .29693 L
.8099 .30448 L
.8101 .3044 L
.81029 .30432 L
.81048 .30424 L
.81067 .30416 L
.81086 .30408 L
.81105 .30399 L
.81124 .30391 L
.81143 .30383 L
.81162 .30375 L
.81181 .30367 L
.812 .30359 L
.81219 .3035 L
.81238 .30342 L
.81257 .30334 L
.81276 .30326 L
.81295 .30318 L
.81314 .3031 L
.81333 .30301 L
.81352 .30293 L
.81371 .30285 L
.8139 .30277 L
.8141 .30269 L
.81429 .30261 L
.81448 .30252 L
.81467 .30244 L
.81486 .30236 L
.81505 .30228 L
.81524 .3022 L
.81543 .30212 L
.81562 .30203 L
.81581 .30195 L
.816 .30187 L
.81619 .30179 L
.81638 .30171 L
.81657 .30163 L
.81676 .30154 L
.81695 .30146 L
.81714 .30138 L
.81733 .3013 L
.81752 .30122 L
Mistroke
.81771 .30114 L
.8179 .30106 L
.8181 .30097 L
.81829 .30089 L
.81848 .30081 L
.81867 .30073 L
.81886 .30065 L
.81905 .30057 L
.81924 .30048 L
.81943 .308 L
.81962 .30792 L
.81981 .30784 L
.82 .30775 L
.82019 .30206 L
.82038 .30198 L
.82057 .3019 L
.82076 .30182 L
.82095 .29615 L
.82114 .29607 L
.82133 .29598 L
.82152 .2959 L
.82171 .29582 L
.8219 .29574 L
.8221 .29009 L
.82229 .29001 L
.82248 .28993 L
.82267 .28985 L
.82286 .28977 L
.82305 .28969 L
.82324 .28961 L
.82343 .28953 L
.82362 .28945 L
.82381 .28937 L
.824 .28929 L
.82419 .28921 L
.82438 .28913 L
.82457 .28905 L
.82476 .29652 L
.82495 .29644 L
.82514 .29636 L
.82533 .29074 L
.82552 .29066 L
.82571 .29058 L
.8259 .2905 L
.8261 .29042 L
.82629 .29034 L
.82648 .29026 L
.82667 .29018 L
.82686 .28458 L
.82705 .2845 L
Mistroke
.82724 .28442 L
.82743 .28434 L
.82762 .27877 L
.82781 .27869 L
.828 .27861 L
.82819 .27853 L
.82838 .27845 L
.82857 .27837 L
.82876 .27282 L
.82895 .27274 L
.82914 .27266 L
.82933 .27258 L
.82952 .2725 L
.82971 .26697 L
.8299 .26689 L
.8301 .26681 L
.83029 .26673 L
.83048 .26665 L
.83067 .26658 L
.83086 .2665 L
.83105 .26642 L
.83124 .26634 L
.83143 .26626 L
.83162 .26618 L
.83181 .2661 L
.832 .26602 L
.83219 .26594 L
.83238 .26044 L
.83257 .26036 L
.83276 .26028 L
.83295 .26021 L
.83314 .26013 L
.83333 .26005 L
.83352 .25997 L
.83371 .25989 L
.8339 .25981 L
.8341 .25433 L
.83429 .25425 L
.83448 .25418 L
.83467 .2541 L
.83486 .25402 L
.83505 .24856 L
.83524 .24848 L
.83543 .24841 L
.83562 .24833 L
.83581 .24825 L
.836 .25567 L
.83619 .2556 L
.83638 .25552 L
.83657 .25544 L
Mistroke
.83676 .25536 L
.83695 .25528 L
.83714 .25521 L
.83733 .25513 L
.83752 .25505 L
.83771 .26244 L
.8379 .26236 L
.8381 .26972 L
.83829 .26964 L
.83848 .26956 L
.83867 .26949 L
.83886 .26941 L
.83905 .26933 L
.83924 .26925 L
.83943 .26917 L
.83962 .2691 L
.83981 .26902 L
.84 .26358 L
.84019 .2635 L
.84038 .26343 L
.84057 .26335 L
.84076 .26327 L
.84095 .27059 L
.84114 .27052 L
.84133 .27781 L
.84152 .27773 L
.84171 .27765 L
.8419 .27757 L
.8421 .2775 L
.84229 .27742 L
.84248 .27734 L
.84267 .27726 L
.84286 .27719 L
.84305 .27711 L
.84324 .27703 L
.84343 .27695 L
.84362 .27688 L
.84381 .27146 L
.844 .26862 L
.84419 .26854 L
.84438 .26846 L
.84457 .26839 L
.84476 .26831 L
.84495 .26823 L
.84514 .26816 L
.84533 .26808 L
.84552 .268 L
.84571 .26793 L
.8459 .26785 L
.8461 .26777 L
Mistroke
.84629 .26769 L
.84648 .26762 L
.84667 .26754 L
.84686 .26746 L
.84705 .26739 L
.84724 .26731 L
.84743 .26723 L
.84762 .26716 L
.84781 .26708 L
.848 .267 L
.84819 .26692 L
.84838 .26685 L
.84857 .26677 L
.84876 .26669 L
.84895 .26662 L
.84914 .26654 L
.84933 .26646 L
.84952 .26639 L
.84971 .26631 L
.8499 .26623 L
.8501 .27349 L
.85029 .27341 L
.85048 .28063 L
.85067 .28782 L
.85086 .28775 L
.85105 .28767 L
.85124 .28759 L
.85143 .29475 L
.85162 .29468 L
.85181 .2946 L
.852 .29452 L
.85219 .29445 L
.85238 .29437 L
.85257 .29429 L
.85276 .29421 L
.85295 .29414 L
.85314 .29406 L
.85333 .29398 L
.85352 .29391 L
.85371 .29383 L
.8539 .30096 L
.8541 .30088 L
.85429 .30081 L
.85448 .30073 L
.85467 .30065 L
.85486 .29782 L
.85505 .29774 L
.85524 .29766 L
.85543 .29759 L
.85562 .29751 L
Mistroke
.85581 .29743 L
.856 .29736 L
.85619 .29728 L
.85638 .2972 L
.85657 .29184 L
.85676 .29176 L
.85695 .29168 L
.85714 .29161 L
.85733 .29153 L
.85752 .28619 L
.85771 .28611 L
.8579 .28603 L
.8581 .28596 L
.85829 .28588 L
.85848 .2858 L
.85867 .28573 L
.85886 .28565 L
.85905 .28558 L
.85924 .2855 L
.85943 .28542 L
.85962 .28535 L
.85981 .28527 L
.86 .28247 L
.86019 .28239 L
.86038 .28232 L
.86057 .28224 L
.86076 .28216 L
.86095 .28209 L
.86114 .27929 L
.86133 .27921 L
.86152 .27914 L
.86171 .28623 L
.8619 .28615 L
.8621 .28608 L
.86229 .286 L
.86248 .28593 L
.86267 .28585 L
.86286 .28578 L
.86305 .2857 L
.86324 .28562 L
.86343 .28555 L
.86362 .28547 L
.86381 .28017 L
.864 .2801 L
.86419 .28002 L
.86438 .27995 L
.86457 .27987 L
.86476 .2798 L
.86495 .27972 L
.86514 .27965 L
Mistroke
.86533 .27957 L
.86552 .27949 L
.86571 .27942 L
.8659 .27934 L
.8661 .27927 L
.86629 .27919 L
.86648 .27912 L
.86667 .27904 L
.86686 .27897 L
.86705 .27889 L
.86724 .27881 L
.86743 .27874 L
.86762 .27866 L
.86781 .27339 L
.868 .27332 L
.86819 .26806 L
.86838 .26799 L
.86857 .26791 L
.86876 .26784 L
.86895 .26776 L
.86914 .26769 L
.86933 .26761 L
.86952 .26754 L
.86971 .26746 L
.8699 .26739 L
.8701 .26731 L
.87029 .26724 L
.87048 .27517 L
.87067 .2751 L
.87086 .27502 L
.87105 .27495 L
.87124 .27487 L
.87143 .2748 L
.87162 .27472 L
.87181 .27465 L
.872 .27457 L
.87219 .27453 L
.87238 .27446 L
.87257 .26923 L
.87276 .26915 L
.87295 .26394 L
.87314 .26387 L
.87333 .26379 L
.87352 .26372 L
.87371 .26365 L
.8739 .26357 L
.8741 .2635 L
.87429 .26342 L
.87448 .26335 L
.87467 .26327 L
Mistroke
.87486 .2632 L
.87505 .26313 L
.87524 .25794 L
.87543 .25786 L
.87562 .25779 L
.87581 .25771 L
.876 .25764 L
.87619 .25757 L
.87638 .25749 L
.87657 .25742 L
.87676 .25735 L
.87695 .25727 L
.87714 .2572 L
.87733 .25712 L
.87752 .25705 L
.87771 .25698 L
.8779 .2569 L
.8781 .25683 L
.87829 .25675 L
.87848 .25668 L
.87867 .25661 L
.87886 .25653 L
.87905 .25646 L
.87924 .25638 L
.87943 .2634 L
.87962 .26332 L
.87981 .26325 L
.88 .26318 L
.88019 .2631 L
.88038 .26303 L
.88057 .26296 L
.88076 .26288 L
.88095 .26281 L
.88114 .26273 L
.88133 .26266 L
.88152 .26259 L
.88171 .26251 L
.8819 .26244 L
.8821 .26236 L
.88229 .26229 L
.88248 .26222 L
.88267 .26214 L
.88286 .26207 L
.88305 .25935 L
.88324 .25928 L
.88343 .2592 L
.88362 .25405 L
.88381 .25398 L
.884 .24884 L
.88419 .24877 L
Mistroke
.88438 .2487 L
.88457 .24862 L
.88476 .24855 L
.88495 .24848 L
.88514 .24841 L
.88533 .24833 L
.88552 .24826 L
.88571 .24822 L
.8859 .24815 L
.8861 .24807 L
.88629 .248 L
.88648 .24793 L
.88667 .24785 L
.88686 .24778 L
.88705 .24771 L
.88724 .24764 L
.88743 .24756 L
.88762 .24749 L
.88781 .24742 L
.888 .24734 L
.88819 .24727 L
.88838 .2472 L
.88857 .24712 L
.88876 .24705 L
.88895 .24698 L
.88914 .2469 L
.88933 .2418 L
.88952 .24172 L
.88971 .2487 L
.8899 .24863 L
.8901 .24856 L
.89029 .24848 L
.89048 .24841 L
.89067 .24834 L
.89086 .24826 L
.89105 .24819 L
.89124 .24812 L
.89143 .24805 L
.89162 .24797 L
.89181 .2479 L
.892 .24783 L
.89219 .24776 L
.89238 .24768 L
.89257 .24764 L
.89276 .24757 L
.89295 .24249 L
.89314 .24241 L
.89333 .24234 L
.89352 .24227 L
.89371 .2422 L
Mistroke
.8939 .23713 L
.8941 .24407 L
.89429 .244 L
.89448 .24393 L
.89467 .24386 L
.89486 .24378 L
.89505 .24371 L
.89524 .23866 L
.89543 .23363 L
.89562 .24054 L
.89581 .24047 L
.896 .2404 L
.89619 .24033 L
.89638 .24025 L
.89657 .24018 L
.89676 .24011 L
.89695 .24004 L
.89714 .23997 L
.89733 .2399 L
.89752 .23982 L
.89771 .23975 L
.8979 .23968 L
.8981 .23961 L
.89829 .2346 L
.89848 .23452 L
.89867 .23445 L
.89886 .23438 L
.89905 .23431 L
.89924 .23424 L
.89943 .23417 L
.89962 .2341 L
.89981 .23402 L
.9 .23395 L
.90019 .23388 L
.90038 .23381 L
.90057 .23377 L
.90076 .2337 L
.90095 .22871 L
.90114 .22864 L
.90133 .22857 L
.90152 .2285 L
.90171 .22843 L
.9019 .22345 L
.9021 .22338 L
.90229 .21843 L
.90248 .21836 L
.90267 .21829 L
.90286 .21822 L
.90305 .21814 L
.90324 .21807 L
Mistroke
.90343 .21804 L
.90362 .21797 L
.90381 .2179 L
.904 .21782 L
.90419 .21775 L
.90438 .21768 L
.90457 .21761 L
.90476 .21754 L
.90495 .21747 L
.90514 .2174 L
.90533 .21733 L
.90552 .21726 L
.90571 .21719 L
.9059 .21712 L
.9061 .21705 L
.90629 .21698 L
.90648 .21691 L
.90667 .21684 L
.90686 .21677 L
.90705 .2167 L
.90724 .21662 L
.90743 .21655 L
.90762 .21648 L
.90781 .21641 L
.908 .21634 L
.90819 .21627 L
.90838 .2162 L
.90857 .22307 L
.90876 .223 L
.90895 .21807 L
.90914 .218 L
.90933 .21793 L
.90952 .21786 L
.90971 .21779 L
.9099 .21772 L
.9101 .21765 L
.91029 .21758 L
.91048 .21751 L
.91067 .21744 L
.91086 .21737 L
.91105 .2173 L
.91124 .21723 L
.91143 .21716 L
.91162 .21709 L
.91181 .21702 L
.912 .21695 L
.91219 .21688 L
.91238 .21681 L
.91257 .21674 L
.91276 .21667 L
Mistroke
.91295 .2166 L
.91314 .21653 L
.91333 .21646 L
.91352 .21639 L
.91371 .21632 L
.9139 .21625 L
.9141 .21134 L
.91429 .21127 L
.91448 .2112 L
.91467 .21113 L
.91486 .21106 L
.91505 .21099 L
.91524 .21092 L
.91543 .21085 L
.91562 .21078 L
.91581 .20589 L
.916 .20582 L
.91619 .20575 L
.91638 .20568 L
.91657 .20561 L
.91676 .20554 L
.91695 .20547 L
.91714 .2054 L
.91733 .20533 L
.91752 .20526 L
.91771 .20519 L
.9179 .20512 L
.9181 .20505 L
.91829 .20498 L
.91848 .20491 L
.91867 .20484 L
.91886 .20477 L
.91905 .20471 L
.91924 .20464 L
.91943 .20457 L
.91962 .2045 L
.91981 .20443 L
.92 .20436 L
.92019 .20429 L
.92038 .20422 L
.92057 .20418 L
.92076 .20411 L
.92095 .20404 L
.92114 .20397 L
.92133 .2039 L
.92152 .20383 L
.92171 .20376 L
.9219 .20369 L
.9221 .20362 L
.92229 .20355 L
Mistroke
.92248 .20348 L
.92267 .20342 L
.92286 .20335 L
.92305 .20078 L
.92324 .19593 L
.92343 .19109 L
.92362 .19103 L
.92381 .19096 L
.924 .19089 L
.92419 .19082 L
.92438 .19075 L
.92457 .19068 L
.92476 .19061 L
.92495 .19054 L
.92514 .19047 L
.92533 .19041 L
.92552 .19034 L
.92571 .19027 L
.9259 .1902 L
.9261 .19013 L
.92629 .19006 L
.92648 .18999 L
.92667 .18992 L
.92686 .18985 L
.92705 .18979 L
.92724 .18972 L
.92743 .18965 L
.92762 .18483 L
.92781 .18476 L
.928 .1847 L
.92819 .18463 L
.92838 .18456 L
.92857 .18449 L
.92876 .18442 L
.92895 .18435 L
.92914 .18428 L
.92933 .18422 L
.92952 .18415 L
.92971 .18408 L
.9299 .18401 L
.9301 .18394 L
.93029 .18387 L
.93048 .1838 L
.93067 .18373 L
.93086 .18367 L
.93105 .1836 L
.93124 .18353 L
.93143 .18346 L
.93162 .18342 L
.93181 .18336 L
Mistroke
.932 .18329 L
.93219 .18322 L
.93238 .18315 L
.93257 .18308 L
.93276 .17829 L
.93295 .17822 L
.93314 .17815 L
.93333 .17808 L
.93352 .17331 L
.93371 .17324 L
.9339 .17317 L
.9341 .17311 L
.93429 .17304 L
.93448 .17297 L
.93467 .1729 L
.93486 .17283 L
.93505 .17277 L
.93524 .1727 L
.93543 .17263 L
.93562 .17012 L
.93581 .17006 L
.936 .16999 L
.93619 .16992 L
.93638 .16985 L
.93657 .16979 L
.93676 .16972 L
.93695 .16965 L
.93714 .16958 L
.93733 .16708 L
.93752 .16701 L
.93771 .16695 L
.9379 .16688 L
.9381 .16681 L
.93829 .16674 L
.93848 .16668 L
.93867 .16661 L
.93886 .16654 L
.93905 .16647 L
.93924 .1664 L
.93943 .16634 L
.93962 .16627 L
.93981 .1662 L
.94 .16613 L
.94019 .16607 L
.94038 .16357 L
.94057 .16351 L
.94076 .16344 L
.94095 .16337 L
.94114 .1633 L
.94133 .16324 L
Mistroke
.94152 .16317 L
.94171 .1631 L
.9419 .16303 L
.9421 .16296 L
.94229 .1629 L
.94248 .16283 L
.94267 .16276 L
.94286 .16269 L
.94305 .16263 L
.94324 .16256 L
.94343 .16249 L
.94362 .16242 L
.94381 .16236 L
.944 .16229 L
.94419 .16222 L
.94438 .16215 L
.94457 .16209 L
.94476 .1596 L
.94495 .15953 L
.94514 .15947 L
.94533 .1594 L
.94552 .15933 L
.94571 .15926 L
.9459 .1592 L
.9461 .15913 L
.94629 .15906 L
.94648 .15899 L
.94667 .15893 L
.94686 .15886 L
.94705 .15415 L
.94724 .15408 L
.94743 .15402 L
.94762 .15395 L
.94781 .15388 L
.948 .15141 L
.94819 .15135 L
.94838 .15128 L
.94857 .15121 L
.94876 .15114 L
.94895 .15108 L
.94914 .15101 L
.94933 .15094 L
.94952 .15088 L
.94971 .15081 L
.9499 .15074 L
.9501 .15068 L
.95029 .15061 L
.95048 .15054 L
.95067 .15047 L
.95086 .15041 L
Mistroke
.95105 .15034 L
.95124 .15027 L
.95143 .15021 L
.95162 .15014 L
.95181 .15695 L
.952 .15688 L
.95219 .15682 L
.95238 .15675 L
.95257 .15668 L
.95276 .15661 L
.95295 .15655 L
.95314 .15648 L
.95333 .15641 L
.95352 .15635 L
.95371 .15628 L
.9539 .15621 L
.9541 .15615 L
.95429 .15608 L
.95448 .16286 L
.95467 .1628 L
.95486 .15812 L
.95505 .15805 L
.95524 .15339 L
.95543 .15332 L
.95562 .15325 L
.95581 .15319 L
.956 .15312 L
.95619 .15305 L
.95638 .15299 L
.95657 .14834 L
.95676 .14827 L
.95695 .14821 L
.95714 .14814 L
.95733 .14807 L
.95752 .14801 L
.95771 .14794 L
.9579 .14788 L
.9581 .14781 L
.95829 .14774 L
.95848 .14768 L
.95867 .14761 L
.95886 .15436 L
.95905 .15429 L
.95924 .15423 L
.95943 .15179 L
.95962 .15172 L
.95981 .15166 L
.96 .15159 L
.96019 .15152 L
.96038 .15146 L
Mistroke
.96057 .15139 L
.96076 .15133 L
.96095 .15126 L
.96114 .15119 L
.96133 .15113 L
.96152 .15106 L
.96171 .15099 L
.9619 .15093 L
.9621 .15086 L
.96229 .1508 L
.96248 .15073 L
.96267 .15066 L
.96286 .1506 L
.96305 .15053 L
.96324 .15047 L
.96343 .1504 L
.96362 .15033 L
.96381 .14572 L
.964 .15243 L
.96419 .15237 L
.96438 .1523 L
.96457 .15224 L
.96476 .15217 L
.96495 .15211 L
.96514 .15204 L
.96533 .15197 L
.96552 .15191 L
.96571 .15184 L
.9659 .15178 L
.9661 .15847 L
.96629 .1584 L
.96648 .15834 L
.96667 .165 L
.96686 .16494 L
.96705 .16487 L
.96724 .1648 L
.96743 .17144 L
.96762 .17138 L
.96781 .17131 L
.968 .17125 L
.96819 .17118 L
.96838 .17111 L
.96857 .17105 L
.96876 .17098 L
.96895 .17092 L
.96914 .17085 L
.96933 .17078 L
.96952 .17072 L
.96971 .17733 L
.9699 .17727 L
Mistroke
.9701 .1772 L
.97029 .17714 L
.97048 .17707 L
.97067 .177 L
.97086 .17694 L
.97105 .17687 L
.97124 .17681 L
.97143 .17674 L
.97162 .17667 L
.97181 .18326 L
.972 .18983 L
.97219 .18976 L
.97238 .18969 L
.97257 .18963 L
.97276 .19617 L
.97295 .1961 L
.97314 .19603 L
.97333 .19597 L
.97352 .1959 L
.97371 .19584 L
.9739 .19577 L
.9741 .1957 L
.97429 .19564 L
.97448 .19557 L
.97467 .19551 L
.97486 .19544 L
.97505 .19538 L
.97524 .19531 L
.97543 .19524 L
.97562 .19518 L
.97581 .20169 L
.976 .20162 L
.97619 .20156 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`000?ojl2@SS_oG
001[0?ooAm?Cdoojn[34g_XC@4J2]<c<c;Ng]fIVI_Cdm2m?CfUYJKjn_/30`;C^]:JVYZf][Jo^k/c8
`8b:Q;no_eICDe5Anb:;8^KViW=cL`00oiT0I_l003_j=020P80008200?oo0820P6IVc3<cIP2009VI
od:JYb8R8_mV<`00P03oo`3o000g??mVc6=SHd92@]kNgYBDU;fm_IZDU>GPhFaXJ;8R8P00Rld009VI
V@2;000i0?o<oom0o`1^0000?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o
?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo
_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo
_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0
_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOknookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo
?oooOooo_mWEc:0Pl=WIfJj^[WMgMiW<09W<<iW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc
0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI0?nI<ol08@000?l08@000?l08@000?l08@000?l0
8@000?l08@000?l08@000?l08@0001/000<100002@001@400@400081lP0001/000<100002@001@40
0@400081lP0001/000<100002@001@400@400081lP0001/000<100001`030@0400410081lP0001/0
00<100001`020@801@7b00006`000`40000700810P050O800002008110000`40000400040@0000@1
10000`40000700810P060O4000001@01000100X000D100000@07008120020@801P7a000000D00@00
0@0700D10`000`40000400030@0000L00P4200H1l@000005004000401`001040004500030@0000<0
00<100001`020@801P7a000000D00@000@0800030@0100H000<100000P000`40000700810P040@03
00400?0000001@01000100T00P4300040@000@@000<100001`020@8010400`01003`00000P020@/0
00<100000P020@D000<100001`020@8010400`01003`00006`000`40000700810`030@0300400?00
000K00030@0000L00P4300810P000`40003_00006`020@P00P4300810P000`40003_00006`000`40
000700810`020@8000<10000k`0001/000<100001`020@<00P4200030@000>l0000K00030@0000L0
0P4300810P000`40003_00006`000`40000700810`020@8000<10000k`0001/000<100001`020@@0
00@10001l@0001/000<100001`020@@000@10001l@0001/000<100001`020@@000@10001l@0001/0
0P48008110001040007a00006`000`400007008110001040007a00006`000`400007008110001040
007a00006`000`400007008110001040007a00006`000`400007008110001040007a00006`000`40
0007008110001040004800030@000>H0000K00030@0000L00P4400040@000@L00P7X00006`000`40
000700811000104000470081j00001/00P4800811000104000470081j00001/000<100001`020@@0
00@100011`020NP0000K00030@0000L00P4400040@000@L00P7X00006`000`40000700811`000`40
00050081j00001/000<100001`020@L000<100001@020NP0000K00030@0000L00P4700030@0000D0
0P7X00006`000`40000700811`000`400005008100<00@00i@0000800P4400030@0000@000<10000
0`000`40000300030@0000L00P4700811P040M80104@000000D00@000@0:00030@0000<000<10000
0`020@P00P4700810`000`4000040M401P4?000000D00@000@0700D100<00@400`4400030@0000L0
0P4700810`000`4000040M4010400`01000>000000D00@000@0700040@000@8000@100011@000`40
000700811`020@<000<10000107@008100D00@000@0?000000D00@000@0800030@0100<000<10040
1@000`40000700811`030@8000<10000107?00810P001040004?000000D00@000@09008110020@D0
00<100001`020@L00`4200030@0000@1cP020@H000<100003@0000800P4;00030@0000<000<10000
0`000`40000700811`030@8000<10000107>00030@0000D00P4>00006`000`40000700811`030@80
00<10000107=008120020@d0000K00030@0000L00P4700<10P000`4000060L/000<100001`000`40
000<00006`020@P00P4700<10P000`4000060LX00P4900030@0000`0000K00030@0000L00P4700<1
0P000`4000060LX000<1000020000`40000<00006`000`40000700811`030@03004100801`780081
2P020@d0000K00030@0000L000<100001P030@03004100801`7800030@0000X000<100002`0001/0
00<100001`000`40000600<100<00@400P070LP000<100002P000`40000;00006`000`4000070003
0@0000H00`4010010@080LL00P4<0081300001/000<100001`000`40000600<100<00@402@7600<1
6P0001/000<100001`000`40000600<100<00@402@7600030@0001X0000K008120000`40000600<1
00<00@402@4400030@000;h00P4L00006`000`40000700030@0000H020400`010@050@800P6m00@1
7@0001/000<100001`000`40000600H100<00@001`420081_@000`400@0N00006`000`4000070003
0@0000L0104200030@0100H10P020Kd000<100407P0001/000<100001`000`40000800<10P000`40
0@060@800P6m00030@0001h0000K00030@0000L000<1000020030@8000<100401P420081^`000`40
0@0P00006`000`40000700030@0000P00`4200030@0100H100@00@41C`040FP00`4P00006`000`40
000700030@0000T00P4200030@0100@100<00@000`5600030@0000@000<100401@5V00818@0001/0
0P4800030@0000T00P4200040@0100<100<00@000`5600030@0000<00P400`010@040FD00P4R0000
6`000`40000700030@0000T00P42000:0@0100400@0100<12P000`40000i00040@00008100<00@00
1P5G00030@0000X00`4R00006`000`40000700030@0000T00P42000:0@0100400@0100<12@020C/0
00D100000@0200P140000`40001400030@0000X000<100008P0001/000<100001`000`400009000>
0@0000400@0100400@030@T00P4i00@10P000`4000040@0400410@d000@10001AP000`4000080081
9@0001/000<100001`000`40000=00030@0100<000<10040104600030@000081>@040@8000D10000
0@0300<130020@03004104D00P4900@1900001/000<100001`000`40000=00030@0100<000<10040
1@45008100@00@41>0040@8000D100000@0500812`020@03004104D00`4800<19@0001/000<10000
1`000`40000=00030@0100<000<100401@45008100@00@41>00010400@4200050@0000401P000`40
000900D1A0020@03004100810`030BL00002008110000`40000400030@0000800P4500030@0000L0
00<100003`001P40000100H11@020@0400410CP000<10000104:00030@0000P00P400`010@040D40
00<100000P400`01000200<19`000005004000402P001@400001008000<100000P020@L00P4A0006
0@0000401P45008100@00@41>0000`4000040@X00`4800030@0000H10P020Cd000<100000P400`01
000200<19`000005004000401`050@8000@1000110000`40000600030@00010000H100000@060@D0
0P4010010@4g00810P030@/00`4800030@0000H100<00@40?P000`4000020@0400400081:@000005
004000401`001040004300<11@000`40000600030@00010000<100000`060@D00P401001004g0003
0@0000800P4;00<11P030@800P42008100@00@41?@000`40000300040@010BX000001@01000100P0
00<100400`000`40000500030@0000H000<1000040000`40000300810P020@D00P40100100440003
0@00030000<100000P020@/00`4600810`020@801P4m00030@0000<000@10041:P00000500400040
2@020@<000<100001@000`40000600030@00010000<100000`020@800P45008100@00@0110000`40
000`00030@0000800P4=00810P000`4000020@<000<100000P030@03004103X00`4500030@0102/0
000200812`000`40000200<110000`40000600030@0001H00P4200811@020@0400400@@000<10000
<0000`40000300030@0000`000<100400`020@`00P4j00811P030B/0000K00030@0000H000<10000
5P020@<00P4200080@010@0100430081<@020@D000<1000030030@<000<1000030000`40000e0003
0@0000811`000`40000Z00006`000`40000600030@0001H00P4300810P060@03004000800P4a0003
0@0001<0104200030@0000`000<100400P4c00@120000`40000Z00006`020@L000<100005P000`40
000200810P020@030041008000@100000P4a00030@0001@00`4200030@0000`01@4700030@0002T0
0P400`01000d00006`000`40000600030@0001/00P42008100<00@400P00104000020C4000<10000
50060@h01@4700030@0002D00`401@01000103D0000K00030@0000H000<100006`020@800P400`01
0@0200040@000081<@000`40000E00@13`050@L0104S00H1>00001/000<100001P000`40000K0081
0P020@03004100800P4200815`000`40000G00030@0001P000<100003@020@800P4600@10`020@80
00<1000060060CT0000K00030@0000H000<100006`020@800P400`010@0200810P020AL000<10000
5`000`40000H00030@0001400P4400<100<00@401@4010010@020A801`4300030@0003L0000K0003
0@0000H000<100006`020@800P400`010@0300030@0000815`000`40000F0081;`00104000040@03
004100/14P040D00000K00030@0000H000<100006`020@0400410@8000D100000@02008120000`40
000;008160020Bl000@100000`4400<100@00@410P060@h0105000006`000`40000600030@0001/0
1P4200050@0000400P020@L00P4=008160020Bl000@100000`4500030@0000<10P060@8000<10000
1P001@4000400081@00001/000<100001P000`40000M00<11`000`400@020@L00P4=00815`020C00
0P4200030@0000T00P4200L100<00@001P020@0400410081@00001/00P4700030@0001d000<10000
1`001@400@0100L00P4=00815`000`40000_008100<00@402`020@801`400`010@06008100D00@00
0@1100006`000`40000600030@0001d000<100001`030@03004000H00P4400030@0000H00`4F0003
0@00030000<100405P000`400@020@D00P401@0100010440000K00030@0000H000<100007@000`40
000700810P000`40000500<10`000`40000600<15P000`40000`00030@0101H01P4300050@010040
A00001/000<100001P000`40000M00030@0000L00P4200030@0000D000<100400`000`4000060003
0@0101H000<10000<0000`400@0F00030@0000<10`000`400@020D@0000K00030@00008000D10000
0@0O00030@0000L00P4200030@0000@00P400`01000200030@0000H000<100405P000`40000`00<1
6@030@0500400@000P5500006`000`40000200050@0000407`000`40000700810P000`4000040004
0@000@<000<100001P000`400@0F00030@00030000<100006@0010400@030@03004104D0000K0003
0@00008000D100000@0O00030@0000L00P4200030@0000@000@100010P030@8000<100000P001@40
0@0101<000<10000C`040DT0000K00030@00008000D100000@0O00030@0000L00P4200030@0000@0
00@100010P001P400@400@@000D100400@0C00030@0004l00P5;00006`020@<000D100000@0O0003
0@0000L00P4200030@0000@000@100010P001P400@400@@000D100400@0C00030@00050000<10000
B@0001/000<100000P001@40000102d000<100000`020@8000D1000100020@03004000<000D10040
0@0C00030@00050000<10000B@0001/000<100000P001@40000102d00P4400810P001@40004000@1
10000`400@020A4000<10040WP0001/000<100000P001@40000102d00P4400810P001@40004000@1
0`000`4000030A4000<10040WP0001/000@100000P4300030@0002/00P4400810P001@40004000@1
0`00104000020@<000<100002`000`400@2N00006`00104000020@<000<10000:`020@@00P420005
0@010@000P400`01000200040@0000810`000`40000:00<1W`0000800P4400030@0000@000<10000
0P020@D000@100000P4300030@0002`000<100000P020@8010401@010@0100<000@1000010400`01
000600040@0000<1X0000005004000402P001@400001008000<100000P00104000020@<000<10000
;0000`40000200810P040@05004100400`00104000070@H000@100000P6Q000000D00@000@0700D1
0P001040004400810P020@<000<10000;0000`40000200040@0000@100H00@400@030@D01@400`01
0@0300810P020J4000001@01000100L000@1000110020@D000@100000P4300030@0002`000<10000
0P000`400003008100H00@000@030@D00`400`0100020@800`4010010@6Q000000D00@000@080003
0@0100<000@100011000104000020@<000<10000;0000`40000200030@0000<00P4400030@010081
1@030@03004000810P050J<000001@01000100T00P4300040@000@@000@100000P4300030@0002`0
0P4300030@0000<00P4400D11@030@030040008100<00@400`6T00000P020@/000<100000P020@D0
00@100000P4300030@0002`00P4300030@0000X01046008100<00@000P401@010@010:D0000K0004
0@0000810`000`40000/00810`000`40000:00<11`020@03004100D1Y`0001/000@100000P430003
0@0002`00P42008130000`400007008100<00@401@6W00006`00104000020@<000<10000;0020@80
00<100002`000`400007008100<00@401@6W00006`020@800P4300030@0002`00P400`010@0=0003
0@0000L000<100000P4010010@6X00006`00104000020@<000<10000;0020@03004001L000<10000
0P420081Z00001/000@100000P4300030@0002d00`4H00030@000081[00001/000@100000P430003
0@0002h00P4H00030@000081[00001/000@100000P4300030@0002h00P4H00040@000Jd0000K0004
0@0000810`000`40000^0081b@0001/000@100000P4300030@0002h00P7900006`00104000020@<0
00<10000;P000`40003800006`020@800P4300030@0002h000<10000b00001/000@100000P430003
0@000?T0000K00040@0000810`000`40003i00006`00104000020@<000<10000n@0001/000@10000
0P4300030@000?T0000K00040@0000810`000`40003i00006`00104000020@<000<10000n@0001/0
00@100000P4300030@0001X01043008110020@@00P4M00@10`020@@00P4400817@020@@00P440081
10020Al000<100000P020@@00P4400817@030@@00P44008110020@<0000K00810P020@800P4N0005
0@0000400P001040004200040@000@8000<100006P000`40000300040@000@8000@100010P001040
004K00040@000@8000@100010P001040004200040@000Ah000D100000@0200040@000@8000@10001
0P000`40000M00040@000@8000@100010P001040004200410@0100006`00104000020@800P4N0005
0@0000400P001040004200040@000@8000<100006`000`40000200040@000@8000@100010P001040
004N00040@000@8000@100010P001040004200030@0001T01@4200040@000@8000@100010P001040
004O00040@000@8000@100010P001040004200410@0100006`00104000020@8000<100007@001@40
0001008000@100010P001040004200030@0001`000D100000@0200040@000@8000@100010P000`40
000J00810`001040004200040@000@8000@100016`001040004300040@000@8000@100010P001040
004L00<10`001040004200040@000@8000@100010P0001/000@100000P4200030@0001d000D10000
0@0200040@000@8000@100010P000`40000M00040@000@8000@100010P001040004200030@0001`0
00@100010P001040004200040@000@8000<100006P000`400@0300040@000@8000@100010P001040
004L00030@0000<000@100010P001040004200040@000@80000K00040@0000810P000`40000K00<1
0`001040004200040@000@8000@1000170001040004200040@000@8000@100010P001040004K0004
0@000@8000@100010P001040004200040@000Ad00P4300040@000@8000@100010P001040004L0003
0@0000<000@100010P001040004200040@000@80000K00040@0000810P000`40000M00030@000080
0P44008110020Ah00P44008110020@@00P4M008110020@@00P4400817`000`400002008110020@@0
0P4M00@10`020@@00P4400810`0001/000@100000P4200030@000?X0000K00040@0000810P000`40
003j00005@3o0@H11P0001/000@100000P4200030@0100T000<100001P000`40000700030@0000L0
00<100001`000`40000700030@0000L000<100001`000`40000700030@0000L000<100001`000`40
000700030@0000H000<100001`000`40000700030@0000L000<100001`000`40000700030@0000L0
00<100001`000`40000700030@0000L000<100001`000`40000700030@0000X0000K00040@000081
0P000`40003j00006`00104000020@8000<10000nP0001/000@100000P4200030@000?X0000K0004
0@0000810P000`40003j00006`00104000020@8000<10000nP0001/000@100000P4200030@000?X0
000K00810P020@8000<10000nP0001/000@100000P4200030@000?X0000K00040@0000810P000`40
003j00006`00104000020@8000<10000nP0001/000@100000P4200030@000?X0000K00040@000081
0P000`40003j00006`00104000020@8000<10000nP0001/000@100000P4200030@000?X0000K0081
0P020@8000<10000nP0001/000@100000`400`01003k00006`00104000030@0300400?/0000K0004
0@0000<100<00@00n`0001/000@100000`400`01003k00006`00104000030@0300400?/0000K0004
0@0000<100<00@00n`0001/000@100000`400`01003k00006`020@800`400`01003k0000o`0Q0000
o`0Q0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000\
\>"],
  ImageRangeCache->{{{108, 395}, {687.75, 510.812}} -> {-2729.67, 0.727307, 
  20.182, 0.000620205}}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Comparing Brute-Force Marginalization and the Bucket-Brigade \
Algorithm with a Greedy Heuristic\
\>", "Section"],

Cell[TextData[{
  "Create a multiply connected network starting with the findings or \
observations:",
  StyleBox[" 1, 2, ",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "and",
  StyleBox[" 3.",
    FontFamily->"Courier",
    FontWeight->"Bold"]
}], "Text"],

Cell["ResetBayesianNetworks[] ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["network3 = CreateBayesianNetwork[8] ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[1,{4}] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[1,{!4}] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[2,{4,5}] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[2,{!4,5}] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[2,{4,!5}] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[2,{!4,!5}] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[3,{5}] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[3,{!5}] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[TextData[{
  "Next come the hidden variables or hypotheses:",
  StyleBox[" 4 ",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "and",
  StyleBox[" 5.",
    FontFamily->"Courier",
    FontWeight->"Bold"]
}], "Text"],

Cell["Pr[4,{6,7}] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[4,{!6,7}] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[4,{6,!7}] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[4,{!6,!7}] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[5,{7,8}] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[5,{!7,8}] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[5,{7,!8}] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[5,{!7,!8}] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[TextData[{
  "Finally the background or history nodes:",
  StyleBox[" 6, 7, ",
    FontFamily->"Courier",
    FontWeight->"Bold"],
  "and",
  StyleBox[" 8.",
    FontFamily->"Courier",
    FontWeight->"Bold"]
}], "Text"],

Cell["Pr[6] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[7] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["Pr[8] = 0.5 ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell["CheckDistribution[network3] ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[CellGroupData[{

Cell["ShowGraph[DependencyGraph[network3],Directed] ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0454545 0.909091 0.0454545 0.909091 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.025 w
.82141 .82141 Mdot
.5 .95455 Mdot
.17859 .82141 Mdot
.04545 .5 Mdot
.17859 .17859 Mdot
.5 .04545 Mdot
.82141 .17859 Mdot
.95455 .5 Mdot
.5 Mabswid
[ ] 0 setdash
.04545 .5 m
.82141 .82141 L
s
.04545 .5 m
.5 .95455 L
s
.17859 .17859 m
.5 .95455 L
s
.17859 .17859 m
.17859 .82141 L
s
.5 .04545 m
.04545 .5 L
s
.82141 .17859 m
.04545 .5 L
s
.82141 .17859 m
.17859 .17859 L
s
.95455 .5 m
.17859 .17859 L
s
.82141 .82141 m
.77072 .82501 L
.78811 .78302 L
F
.5 .95455 m
.45179 .93847 L
.48393 .90633 L
F
.5 .95455 m
.46161 .92125 L
.5036 .90385 L
F
.17859 .82141 m
.15586 .77596 L
.20132 .77596 L
F
.04545 .5 m
.06153 .45179 L
.09367 .48393 L
F
.04545 .5 m
.07875 .46161 L
.09615 .5036 L
F
.17859 .17859 m
.22404 .15586 L
.22404 .20132 L
F
.17859 .17859 m
.22928 .17499 L
.21189 .21698 L
F
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`4000010?ooo`000?ojl2@SS_oG
001[0?ooAm?Cdoojn[34g_XC@4J2]<c<c;Ng]fIVI_Cdm2m?CfUYJKjn_/30`;C^]:JVYZf][Jo^k/c8
`8b:Q;no_eICDe5Anb:;8^KViW=cL`00oiT0I_l003_j=020P80008200?oo0820P6IVc3<cIP2009VI
od:JYb8R8_mV<`00P03oo`3o000g??mVc6=SHd92@]kNgYBDU;fm_IZDU>GPhFaXJ;8R8P00Rld009VI
V@2;000i0?o<oom0o`1^0000?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o
?`2oO`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo
_cmoocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo
_gloogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0
_kl0oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOknookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_ol0oolo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo
?oooOooo_mWEc:0Pl=WIfJj^[WMgMiW<09W<<iW<IYW<VIW<c9W<oiWo09Wo<iWoIYWoVIWoc9Wool`0
0<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaVc<aVolbI0<bI<lbI
I/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<olco0<co<lcoI/coVLcoc<coool0<ol0I_l0VOl0c?l0oolc
0?lc<olcI_lcVOlcc?lcoomV0?mVI_mVVOmVoonI0?nI<ol08@000?l08@000?l08@000?l08@000?l0
8@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@0008h0106>0000S@060Hd0002=00H1
S@0008d01P6=0000S@060Hd0002=00D1SP0008`000<10000T@0008/000<10000TP0008X000<10000
T`0008T000<10000U00008P000<10000U@0008L000<10000UP0008H000<10000U`0008D000<10000
V00008@000<10000V@0008<000<10000VP00088000<10000V`00084000<10000W000080000<10000
W@0007l000<10000WP0007h000<10000W`0007d000<10000X00007`000<10000X@0007/000<10000
XP0007X000<10000X`0007T000<10000Y00007P000<10000Y@0007L000<10000YP0007H000<10000
Y`0007D000<10000Z00007@000<10000Z@0007<000<10000ZP00078000<10000Z`00074000<10000
[000070000<10000[@0003l000<10000;@000`40002^0000?@030Bh000<10000[`0003/01@4]0003
0@000;00000b00@10`070B`000<10000N`040C80000a00l1:`000`40001k00H1<@0003403`4Z0003
0@0007`01P4a0000<@2n0C40000a01019`000`40001m00L1<@0003803P4W00030@0007/00`4200@1
<P0003<00P400`010@080BH000<10000NP020C/0000c00040@0100T19@000`40001i0081?@0003<0
00<100400P090B<000<10000M`030Cl0000c00030@0100801`42008180000`40001f0081@P0003<0
00@100010P050@D00`4L00030@0007@00`540000<`001040004300@120020AT000<10000L`020DL0
000c00050@0000400`020@/00P4F00030@0007800P590000<`001@40000100@000<100002`030A80
00<10000L0030D/0000c00030@00008000<100004P020@l000<10000K`020Dh0000c00030@000080
00<1000050030@/000<10000K@030E00000c00030@00008000<100005`020@P000<10000K0020E<0
000c00030@0000<000<1000060020@D000<10000J`020ED0000c00030@0000<000<100006P030@03
004006X00`5G0000<`000`40000400030@0001`00P5Y0081FP0003<000<1000010000`40000K0003
0@0000<1H`030E`0000c00030@0000@000<100006P000`4000040081G`020El0000c00030@0000D0
00<1000060000`40000700<1F@030F40000c00030@0000D000<100005`000`40000;0081E@020F@0
000c00030@0000H000<100005@000`40000>0081D@020FH0000c00030@0000H000<1000050000`40
000A00<1B`030FP0000c00030@0000L000<100004P000`40000E0081A`020F/0000c00030@0000L0
00<100004@000`40000H00<1@@030Fd0000c00030@0000L000<1000040000`40000L0081?@020G00
000c00030@0000P000<100003P000`40000O0081>@020G80000c00030@0000P000<100003@000`40
000R00<1<`030G@0000c00030@0000T000<100002`000`40000V0081;`020GL0000c00030@0000T0
00<100002P000`40000Y00<1:@030GT0000c00030@0000T000<100002@000`40000]00819@020G`0
000c00030@0000X000<100001`000`40000`00<17`030Gh0000c00030@0000X000<100001P000`40
000d00816`020H40000c00030@0000/000<1000010000`40000g00815`020H<0000c00030@0000/0
00<100000`000`40000j00<14@030HD0000c00030@0000`000D100000@1000813@020HP0000c0003
0@0000`000@10001@`030@L00`6:0000<`000`40000<00030@0104L00P430081S@0003<000<10000
3@000`40001800<1S`0003<000<1000030020DP00`4010010@6=0000<`000`40000;00040@000DD0
0P470081R`0003<000<100002P001@40000104800`4;00<1R00003<000<100002@000`4000020003
0@0003h00P4A0081QP0003<000<1000020000`40000400030@0003/00P4E0081Q00003<000<10000
1`000`40000500030@0003P00`4I00<1P@0003<000<100001P000`40000700030@0003D00P4O0081
O`0003<000<100001@000`40000800030@0003800`4S00<1O00003<000<1000010000`4000090003
0@0003000P4Y0081NP0003<000<100000`000`40000;00030@0002`00`4]00<1M`0003<000<10000
0P000`40000<00030@0002X00P4c0081M@0003<000D100000@0@00030@0002L00P4g0081L`0003<0
00@100014@000`40000T00<1>`030G00000c00030@0101<000<100008@020D400P5^0000<`020A@0
00<100007P030DD00`5[0000<`000`40000C00030@0001`00P5;0081J@0003800P4F00030@0001T0
0P5?0081I`00034000<100405P000`40000F00<1D`030F@0000`00040@000AL000<100004`020ET0
0P5R0000;`001@40000101L000<1000040030Ed00`5O0000;P000`40000200030@0001D000<10000
3P020F<00P5M0000;@000`40000300030@0001H000<100002`020FL00P5K0000;0000`4000040003
0@0001H000<1000020030F/00`5H0000:`000`40000500030@0001L000<100001@020G400P5F0000
:P000`40000600030@0001L000<100000P030GD00`5C0000:@000`40000700030@0001P000@10041
N`020E40000X00030@0000P000<100005`030Gl00`5>00009`000`40000900030@0001D00P400`01
00220081C00002H000<100002P000`40000C008110000`4000220081BP0002D000<100002`000`40
000@00<11P000`40002400<1A`0002@000<1000030000`40000>00812P000`4000260081A@0002<0
00<100003@000`40000;00<130000`40002800<1@P00028000<100003P000`40000900813`000`40
002;0081@000024000<100003`000`40000700814P000`40002<0081?P00020000<1000040000`40
000400<150000`40002>00<1>`0001l000<100004@000`400002008160000`40002@0081>@0001h0
00<100004P000`400@020AX000<10000TP030CH0000M00030@0001<00P4M00030@0009D00P4d0000
70000`40000A00@17`000`40002F00<1<@0001/000<1000040020@<000<100007@000`40002I0081
;`00014000<100001P000`40000?00811@000`40000N00030@0009X00P4]00004@020@H000<10000
3@030@L000<100007P000`40002L00<1:P0001400`4400030@0000`00P4:00030@0001l000<10000
WP020BP0000@00D100<00@4030030@`000<100007`000`40002P00<19@0001001`4;00813`000`40
000O00030@000:<00P4S000040080@P00P4A00030@00020000<10000Y0020B40000?00T11@030A<0
00<1000080000`40002V00<17P0000l02P4200815P000`40000Q00030@000:P00P4L00003`0<0AP0
00<100008@000`40002Z00<16@0000l02P4J00030@00024000<10000[@020AL0000;00l16@000`40
000R00030@000:h00P4500@1300000X0404I00030@00028000<10000/0030@03004100@12`0000X0
4@4H00030@0002<000<10000/P070@/0000:014160000`40000S00030@000;<01P4;00002P070B80
00<1000090000`40002b00H12`0000/01@4010010@4O00030@0002@000<10000/`040@`0000@0004
0@0000817@000`40000T00030@000<<0000A00030@0000800P4K00030@0002D000<10000`P000180
00<100000`030AP000<100009@000`40003200004`000`40000500815P000`40000V00030@000<40
000D00030@0000H00`4C00030@0002H000<10000`@0001D000<1000020020A4000<100009P000`40
003100005P000`40000900813`000`40000W00030@000<00000G00030@0000X00`4<00030@0002L0
00<10000`00001P000<1000030020@X000<10000:0000`40002o00006@000`40000=00<11`000`40
000X00030@000;l0000J00030@0000l00P4500030@0002T000<10000_P0001/000<1000040020@<0
00<10000:@000`40002n000070000`40000A00@1:`000`40002n00007@000`40000C0081:`000`40
002m00007P000`40000B00030@010081:0000`40002m00007`000`40000A00030@0000800P4W0003
0@000;`0000P00030@00010000<1000010030B@000<10000_000024000<100003`000`4000070081
8P000`40002l00008P000`40000>00030@0000T00P4Q00030@000;/0000S00030@0000d000<10000
2`030Ah000<10000^`0002@000<1000030000`40000>00817@000`40002j00009@000`40000;0003
0@0001000`4J00030@000;X0000V00030@0000X000<100004`020AP000<10000^P0002L000<10000
2@000`40000E00815`000`40002i0000:0000`40000800030@0001L00`4D00030@000;T0000Y0003
0@0000L000<100006P020A<000<10000^00002X000<100001P000`40000L00<140000`40002h0000
:`000`40000500030@0001l00P4?00030@000;L0000/00030@0000@000<100008@030@`000<10000
]`0002d000<100000`000`40000T00812P000`40002g0000;P000`40000200030@0002H00P490003
0@000;H0000_00050@000040:P030@H000<10000]P00030000@10001;@020@D000<10000]@000340
00<10040;`030@8000<10000]@0003800P4b00<1]`0003<000<10000<P020KH0000c0081=0040K<0
000c00030@0103@000<100000P6a0000<`001040004c00030@0000800`6^0000<`001@40000103<0
00<1000010020J`0000c00030@00008000<10000<0000`4000060081ZP0003<000<100000`000`40
000_00030@0000P00`6W0000<`000`40000400030@0002l000<100002P020JD0000c00030@0000D0
00<10000;P000`40000<00<1XP0003<000<100001P000`40000^00030@0000h00P6P0000<`000`40
000700030@0002d000<1000040030Id0000c00030@0000P000<10000;0000`40000C0081V`0003<0
00<100002@000`40000/00030@0001@00P6I0000<`000`40000:00030@0002/000<100005P030IH0
000c00030@0000/000<10000:`000`40000H0081U00003<000<1000030000`40000Z00030@0001X0
0`6A0000<`000`40000=00030@0002X000<1000070020Hl0000c00030@0000h000<10000:@000`40
000N0081S@0003<000<100003`000`40000X00030@0002000`6:0000<`000`40000@00030@0002P0
00<100008P020HP0000c00030@00014000<100009`000`40000T00<1Q@0003<000<100004P000`40
000W00030@0002H00P630000<`000`40000C00030@0002H000<10000:0020H40000c00030@0001@0
00<100009@000`40000Z00<1OP0003<000<100005@000`40000U00030@0002`00P5l0000<`000`40
000F00030@0002@000<10000;P030GT0000c00030@0001L000<1000090000`40000`0081M`0003<0
00<1000060000`40000S00030@0003800`5d0000<`000`40000I00030@00028000<10000=@020G80
000c00030@0001X000<100008P000`40000f0081L00003<000<100006`000`40000Q00030@0003P0
0`5]0000<`000`40000L00030@00024000<10000>P020F/0000c00030@0001d000<1000080000`40
000l00<1J00003<000<100007P000`40000P00030@0003h00P5V0000<`000`40000O00030@0001l0
00<10000@0020F@0000c00030@00020000<100007P000`40001200<1H@0003<000<100008@000`40
000N00030@0004@00P5O0000<`000`40000R00030@0001d000<10000AP030E`0000c00030@0002<0
00<100007@000`4000180081FP0003<000<1000090000`40000L00030@0004X00`5G0000<`000`40
000U00030@0001/000<10000C@020ED0000c00030@0002H000<100006`000`40001>0081D`0002d0
304T00030@0001X000<10000D0030E00000^00/19@000`40000J00030@0005800P5>0000;P0:0BL0
00<100006@000`40001D00<13@000`40000k0000;`090BP000<100006@000`40001F00812`020C`0
000_00P1:P000`40000H00030@0005P00P4800@1>`0003001`4[00030@0001L000<10000FP030@D0
1@4j0000<0060Bd000<100005`000`40001L00810P070CT0000a00D1;P000`40000F00030@0005h0
2@4i0000<@040C0000<100005P000`40001O00P1>0000380104`00030@0001D000<10000GP0:0@03
00410081<P0003401P4`00030@0001@000<10000GP0@0C40000a00H1<@000`40000D00030@0005`0
4@4a0000<@060C8000<100004`000`40001L0141<@0003401P4c00030@0001<000<10000IP060C40
000b00@1=@000`40000B00030@0006L0104b0000K0000`40000A00030@0009d0001]00030@000140
00<10000W00006h000<1000040000`40002L0000K`000`40000@00030@0009/0001`00030@0000l0
00<10000V`00074000<100003`000`40002J0000LP000`40000>00030@0009X0001c00030@0000d0
00<10000VP0007@000<100003@000`40002I0000M@000`40000<00030@0009T0001f00030@0000`0
00<10000V00007L000<100002`000`40002H0000N0000`40000:00030@0009P0001i00030@0000X0
00<10000U`0007X000<100002@000`40002G0000N`000`40000900030@0009H0001l00030@0000P0
00<10000UP0007d000<1000020000`40002E0000OP000`40000700030@0009D0001o00030@0000H0
00<10000U@00080000<100001P000`40002D0000P@000`40000500030@0009@0002200030@0000D0
00@100000P6@0000P`000`400003008100<00@400P6@0000Q000104000080I00002500030@0100P1
T00008@0306@0000Q@0;0I00002400`1T00008<03@6@0000PP0>0I00002100l1T00008D02`6@0000
R00:0Hh0002<00L1S@0008d01P6=0000S@060Hd0002=00H1S@0008h0106>0000o`0Q0000o`0Q0000
o`0Q0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000\
\>"],
  ImageRangeCache->{{{108, 395}, {403, 116}} -> {-0.463947, 0.394598, 
  0.00383279, 0.00383279}}]
}, Open  ]],

Cell["\<\
With the following invocations you can compare brute \
marginalization and bucket brigade for any of the three networks defined \
above.\
\>", "Text"],

Cell["e = {} ; n = network3 ;", "Input"],

Cell["PrReset[] ;", "Input"],

Cell["t = TabularFormat[e,n] ;", "Input"],

Cell[CellGroupData[{

Cell["PrCount[]", "Input"],

Cell[BoxData[
    \(38\)], "Output"]
}, Open  ]],

Cell["BucketReset[] ;", "Input"],

Cell[CellGroupData[{

Cell["BucketElimination[t] ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]1\),
      SequenceForm[ "Selecting ", 1],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]3\),
      SequenceForm[ "Selecting ", 3],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]2\),
      SequenceForm[ "Selecting ", 2],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]6\),
      SequenceForm[ "Selecting ", 6],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]8\),
      SequenceForm[ "Selecting ", 8],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]4\),
      SequenceForm[ "Selecting ", 4],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]5\),
      SequenceForm[ "Selecting ", 5],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[\("Selecting "\[InvisibleSpace]7\),
      SequenceForm[ "Selecting ", 7],
      Editable->False]], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell["BucketCount[]", "Input"],

Cell[BoxData[
    \(88\)], "Output"]
}, Open  ]],

Cell["PrReset[] ;", "Input"],

Cell["Marginalize[e,n] ;", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[CellGroupData[{

Cell["PrCount[]", "Input"],

Cell[BoxData[
    \(2048\)], "Output"]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"X 3.0",
ScreenRectangle->{{0, 1280}, {0, 1024}},
AutoGeneratedPackage->None,
WindowSize->{841, 672},
WindowMargins->{{176, Automatic}, {Automatic, 102}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"Magnification"->1},
Magnification->1,
StyleDefinitions -> "Classroom.nb"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1709, 49, 47, 0, 55, "Title"],

Cell[CellGroupData[{
Cell[1781, 53, 31, 0, 57, "Section"],
Cell[1815, 55, 445, 13, 46, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[2297, 73, 66, 0, 57, "Section"],
Cell[2366, 75, 178, 6, 28, "Text"],
Cell[2547, 83, 200, 7, 47, "Input",
  InitializationCell->True],
Cell[2750, 92, 99, 1, 47, "Input",
  InitializationCell->True],
Cell[2852, 95, 171, 4, 47, "Input",
  InitializationCell->True],
Cell[3026, 101, 198, 5, 28, "Text",
  InitializationCell->True],
Cell[3227, 108, 781, 20, 257, "Input",
  InitializationCell->True],
Cell[4011, 130, 249, 8, 28, "Text",
  InitializationCell->True],
Cell[4263, 140, 1290, 34, 467, "Input",
  InitializationCell->True],
Cell[5556, 176, 302, 9, 28, "Text",
  InitializationCell->True],
Cell[5861, 187, 276, 9, 92, "Input",
  InitializationCell->True],
Cell[6140, 198, 376, 11, 28, "Text",
  InitializationCell->True],
Cell[6519, 211, 1641, 40, 557, "Input",
  InitializationCell->True],
Cell[8163, 253, 287, 5, 46, "Text"],
Cell[8453, 260, 342, 9, 152, "Input"],
Cell[8798, 271, 96, 3, 28, "Text"],
Cell[8897, 276, 98, 3, 62, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[9032, 284, 69, 0, 57, "Section"],
Cell[9104, 286, 100, 3, 28, "Text"],
Cell[9207, 291, 150, 4, 47, "Input",
  InitializationCell->True],
Cell[9360, 297, 162, 4, 47, "Input",
  InitializationCell->True],
Cell[9525, 303, 50, 0, 28, "Text"],
Cell[9578, 305, 144, 4, 47, "Input",
  InitializationCell->True],
Cell[9725, 311, 145, 4, 47, "Input",
  InitializationCell->True],
Cell[9873, 317, 145, 4, 47, "Input",
  InitializationCell->True],
Cell[10021, 323, 146, 4, 47, "Input",
  InitializationCell->True],
Cell[10170, 329, 138, 4, 47, "Input",
  InitializationCell->True],
Cell[10311, 335, 138, 4, 47, "Input",
  InitializationCell->True],
Cell[10452, 341, 88, 3, 28, "Text"],
Cell[10543, 346, 154, 4, 47, "Input",
  InitializationCell->True],

Cell[CellGroupData[{
Cell[10722, 354, 172, 4, 47, "Input",
  InitializationCell->True],
Cell[10897, 360, 7680, 146, 296, 824, 57, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[18592, 509, 126, 3, 28, "Text"],

Cell[CellGroupData[{
Cell[18743, 516, 152, 4, 47, "Input",
  InitializationCell->True],
Cell[18898, 522, 1236, 32, 47, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[20171, 559, 170, 4, 47, "Input",
  InitializationCell->True],
Cell[20344, 565, 700, 20, 47, "Output"]
}, Open  ]],
Cell[21059, 588, 92, 3, 28, "Text"],

Cell[CellGroupData[{
Cell[21176, 595, 257, 8, 77, "Input",
  InitializationCell->True],
Cell[21436, 605, 143, 3, 23, "Print"],
Cell[21582, 610, 143, 3, 23, "Print"],
Cell[21728, 615, 143, 3, 23, "Print"],
Cell[21874, 620, 143, 3, 23, "Print"],
Cell[22020, 625, 37, 1, 47, "Output"]
}, Open  ]],
Cell[22072, 629, 69, 0, 28, "Text"],

Cell[CellGroupData[{
Cell[22166, 633, 170, 4, 47, "Input",
  InitializationCell->True],
Cell[22339, 639, 143, 3, 23, "Print"],
Cell[22485, 644, 143, 3, 23, "Print"],
Cell[22631, 649, 143, 3, 23, "Print"],
Cell[22777, 654, 37, 1, 47, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[22851, 660, 198, 7, 47, "Input",
  InitializationCell->True],
Cell[23052, 669, 143, 3, 23, "Print"],
Cell[23198, 674, 39, 1, 47, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[23274, 680, 206, 7, 47, "Input",
  InitializationCell->True],
Cell[23483, 689, 40, 1, 47, "Output"]
}, Open  ]],
Cell[23538, 693, 108, 3, 28, "Text"],

Cell[CellGroupData[{
Cell[23671, 700, 149, 4, 47, "Input",
  InitializationCell->True],
Cell[23823, 706, 37, 1, 47, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[23897, 712, 167, 4, 47, "Input",
  InitializationCell->True],
Cell[24067, 718, 39, 1, 47, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[24143, 724, 175, 4, 47, "Input",
  InitializationCell->True],
Cell[24321, 730, 40, 1, 47, "Output"]
}, Open  ]],
Cell[24376, 734, 227, 7, 28, "Text"],

Cell[CellGroupData[{
Cell[24628, 745, 241, 7, 62, "Input",
  InitializationCell->True],
Cell[24872, 754, 38, 1, 47, "Output"]
}, Open  ]],
Cell[24925, 758, 519, 22, 28, "Text"],
Cell[25447, 782, 162, 4, 47, "Input",
  InitializationCell->True],
Cell[25612, 788, 144, 4, 47, "Input",
  InitializationCell->True],
Cell[25759, 794, 145, 4, 47, "Input",
  InitializationCell->True],
Cell[25907, 800, 145, 4, 47, "Input",
  InitializationCell->True],
Cell[26055, 806, 147, 4, 47, "Input",
  InitializationCell->True],
Cell[26205, 812, 139, 4, 47, "Input",
  InitializationCell->True],
Cell[26347, 818, 143, 4, 47, "Input",
  InitializationCell->True],
Cell[26493, 824, 144, 4, 47, "Input",
  InitializationCell->True],
Cell[26640, 830, 139, 4, 47, "Input",
  InitializationCell->True],
Cell[26782, 836, 142, 4, 47, "Input",
  InitializationCell->True],
Cell[26927, 842, 144, 4, 47, "Input",
  InitializationCell->True],
Cell[27074, 848, 154, 4, 47, "Input",
  InitializationCell->True],

Cell[CellGroupData[{
Cell[27253, 856, 172, 4, 47, "Input",
  InitializationCell->True],
Cell[27428, 862, 8173, 166, 296, 1032, 73, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[35616, 1031, 96, 3, 28, "Text"],

Cell[CellGroupData[{
Cell[35737, 1038, 178, 4, 47, "Input",
  InitializationCell->True],
Cell[35918, 1044, 143, 3, 23, "Print"],
Cell[36064, 1049, 143, 3, 23, "Print"],
Cell[36210, 1054, 143, 3, 23, "Print"],
Cell[36356, 1059, 143, 3, 23, "Print"],
Cell[36502, 1064, 39, 1, 47, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[36578, 1070, 157, 4, 47, "Input",
  InitializationCell->True],
Cell[36738, 1076, 39, 1, 47, "Output"]
}, Open  ]],
Cell[36792, 1080, 126, 3, 28, "Text"],

Cell[CellGroupData[{
Cell[36943, 1087, 198, 7, 47, "Input",
  InitializationCell->True],
Cell[37144, 1096, 143, 3, 23, "Print"],
Cell[37290, 1101, 143, 3, 23, "Print"],
Cell[37436, 1106, 143, 3, 23, "Print"],
Cell[37582, 1111, 57, 1, 47, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[37676, 1117, 166, 4, 47, "Input",
  InitializationCell->True],
Cell[37845, 1123, 57, 1, 47, "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[37951, 1130, 72, 0, 57, "Section"],
Cell[38026, 1132, 541, 15, 242, "Input"],

Cell[CellGroupData[{
Cell[38592, 1151, 62, 0, 47, "Input"],
Cell[38657, 1153, 53, 1, 47, "Output"]
}, Open  ]],
Cell[38725, 1157, 518, 15, 242, "Input"],

Cell[CellGroupData[{
Cell[39268, 1176, 71, 1, 47, "Input"],
Cell[39342, 1179, 58, 1, 64, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[39437, 1185, 91, 1, 47, "Input"],
Cell[39531, 1188, 50, 1, 47, "Output"]
}, Open  ]],
Cell[39596, 1192, 536, 16, 257, "Input"],

Cell[CellGroupData[{
Cell[40157, 1212, 103, 2, 47, "Input"],
Cell[40263, 1216, 91173, 5323, 186, 79764, 5178, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[131451, 6542, 241, 7, 28, "Text"],

Cell[CellGroupData[{
Cell[131717, 6553, 241, 7, 62, "Input",
  InitializationCell->True],
Cell[131961, 6562, 55, 1, 47, "Output"]
}, Open  ]],
Cell[132031, 6566, 236, 7, 28, "Text"],

Cell[CellGroupData[{
Cell[132292, 6577, 103, 2, 47, "Input"],
Cell[132398, 6581, 90582, 5181, 186, 77649, 5017, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[222995, 11765, 246, 7, 28, "Text"],

Cell[CellGroupData[{
Cell[223266, 11776, 243, 7, 62, "Input",
  InitializationCell->True],
Cell[223512, 11785, 55, 1, 47, "Output"]
}, Open  ]],
Cell[223582, 11789, 492, 8, 64, "Text"],

Cell[CellGroupData[{
Cell[224099, 11801, 108, 2, 47, "Input"],
Cell[224210, 11805, 92434, 5286, 186, 79008, 5116, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[316693, 17097, 123, 3, 57, "Section"],
Cell[316819, 17102, 266, 10, 28, "Text"],
Cell[317088, 17114, 150, 4, 47, "Input",
  InitializationCell->True],
Cell[317241, 17120, 162, 4, 47, "Input",
  InitializationCell->True],
Cell[317406, 17126, 142, 4, 47, "Input",
  InitializationCell->True],
Cell[317551, 17132, 143, 4, 47, "Input",
  InitializationCell->True],
Cell[317697, 17138, 144, 4, 47, "Input",
  InitializationCell->True],
Cell[317844, 17144, 145, 4, 47, "Input",
  InitializationCell->True],
Cell[317992, 17150, 145, 4, 47, "Input",
  InitializationCell->True],
Cell[318140, 17156, 146, 4, 47, "Input",
  InitializationCell->True],
Cell[318289, 17162, 142, 4, 47, "Input",
  InitializationCell->True],
Cell[318434, 17168, 143, 4, 47, "Input",
  InitializationCell->True],
Cell[318580, 17174, 226, 9, 28, "Text"],
Cell[318809, 17185, 144, 4, 47, "Input",
  InitializationCell->True],
Cell[318956, 17191, 145, 4, 47, "Input",
  InitializationCell->True],
Cell[319104, 17197, 145, 4, 47, "Input",
  InitializationCell->True],
Cell[319252, 17203, 146, 4, 47, "Input",
  InitializationCell->True],
Cell[319401, 17209, 144, 4, 47, "Input",
  InitializationCell->True],
Cell[319548, 17215, 145, 4, 47, "Input",
  InitializationCell->True],
Cell[319696, 17221, 145, 4, 47, "Input",
  InitializationCell->True],
Cell[319844, 17227, 146, 4, 47, "Input",
  InitializationCell->True],
Cell[319993, 17233, 225, 9, 28, "Text"],
Cell[320221, 17244, 138, 4, 47, "Input",
  InitializationCell->True],
Cell[320362, 17250, 138, 4, 47, "Input",
  InitializationCell->True],
Cell[320503, 17256, 138, 4, 47, "Input",
  InitializationCell->True],
Cell[320644, 17262, 154, 4, 47, "Input",
  InitializationCell->True],

Cell[CellGroupData[{
Cell[320823, 17270, 172, 4, 47, "Input",
  InitializationCell->True],
Cell[320998, 17276, 12369, 244, 296, 1372, 104, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[333382, 17523, 160, 4, 28, "Text"],
Cell[333545, 17529, 40, 0, 47, "Input"],
Cell[333588, 17531, 28, 0, 47, "Input"],
Cell[333619, 17533, 41, 0, 47, "Input"],

Cell[CellGroupData[{
Cell[333685, 17537, 26, 0, 47, "Input"],
Cell[333714, 17539, 36, 1, 47, "Output"]
}, Open  ]],
Cell[333765, 17543, 32, 0, 47, "Input"],

Cell[CellGroupData[{
Cell[333822, 17547, 147, 4, 47, "Input",
  InitializationCell->True],
Cell[333972, 17553, 143, 3, 23, "Print"],
Cell[334118, 17558, 143, 3, 23, "Print"],
Cell[334264, 17563, 143, 3, 23, "Print"],
Cell[334410, 17568, 143, 3, 23, "Print"],
Cell[334556, 17573, 143, 3, 23, "Print"],
Cell[334702, 17578, 143, 3, 23, "Print"],
Cell[334848, 17583, 143, 3, 23, "Print"],
Cell[334994, 17588, 143, 3, 23, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[335174, 17596, 30, 0, 47, "Input"],
Cell[335207, 17598, 36, 1, 47, "Output"]
}, Open  ]],
Cell[335258, 17602, 28, 0, 47, "Input"],
Cell[335289, 17604, 143, 4, 47, "Input",
  InitializationCell->True],

Cell[CellGroupData[{
Cell[335457, 17612, 26, 0, 47, "Input"],
Cell[335486, 17614, 38, 1, 47, "Output"]
}, Open  ]]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

